#ifndef _GOOM_TYPEDEFS_H
#define _GOOM_TYPEDEFS_H

typedef struct _PLUGIN_INFO PluginInfo;
typedef struct _SOUND_INFO SoundInfo;
typedef struct _GMLINE GMLine;
typedef struct _GMUNITPOINTER GMUnitPointer;
typedef struct _ZOOM_FILTER_DATA ZoomFilterData;
typedef struct _VISUAL_FX VisualFX;

#endif
