#ifndef DMO_GUIDS_H
#define DMO_GUIDS_H

#include "DirectShow/guids.h"

extern const GUID IID_IMediaBuffer;
extern const GUID IID_IMediaObject;
extern const GUID IID_IEnumDMO;
extern const GUID IID_IMediaObjectInPlace;
extern const GUID IID_IDMOQualityControl;
extern const GUID IID_IDMOVideoOutputOptimizations;

/* to be removed
extern const GUID DMOCATEGORY_AUDIO_DECODER;
extern const GUID DMOCATEGORY_AUDIO_ENCODER;
extern const GUID DMOCATEGORY_VIDEO_DECODER;
extern const GUID DMOCATEGORY_VIDEO_ENCODER;
extern const GUID DMOCATEGORY_AUDIO_EFFECT;
extern const GUID DMOCATEGORY_VIDEO_EFFECT;
extern const GUID DMOCATEGORY_AUDIO_CAPTURE_EFFECT;

// Acoustic Echo Canceller
// Matches KSNODETYPE_ACOUSTIC_ECHO_CANCEL
extern const GUID DMOCATEGORY_ACOUSTIC_ECHO_CANCEL;

// Noise Supression
// Matches KSNODETYPE_AUDIO_NOISE_SUPPRESS
extern const GUID DMOCATEGORY_AUDIO_NOISE_SUPPRESS;

// Automatic Gain Control
// Matches KSNODETYPE_AGC
extern const GUID DMOCATEGORY_AGC;
*/

#endif
