/*=============================================================================
 *
 *  This software has been released under the terms of the GNU Public
 *  license. See http://www.gnu.org/copyleft/gpl.html for details.
 *
 *  Copyright 2002 Anders Johansson ajh@atri.curtin.edu.au
 *
 *=============================================================================
 */

#ifndef	_DSP_H
#define	_DSP_H	1

/* Implementation of routines used for DSP */

/* Size of floating point type used in routines */
#define _ftype_t float

#include "window.h"
#include "filter.h"

#endif
