/*
   This file is based on:
   $XFree86: xc/programs/Xserver/hw/xfree86/drivers/chips/util/AsmMacros.h,v 1.1 2001/11/16 21:13:34 tsi Exp $
   Modified for readability by Nick Kurshev
*/

#ifndef __ASM_MACROS_IA64_H
#define __ASM_MACROS_IA64_H

#if defined(linux)
#include <sys/io.h>
#else
#include "sysdep/AsmMacros_generic.h"
#endif

#endif
