/*
 * Copyright (C) 2004-2007 the xine project
 *
 * This file is part of xine, a free video player.
 *
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA
 *
 */

#ifndef __HAVE_XINE_VIDEO_WINDOW_H__
#define __HAVE_XINE_VIDEO_WINDOW_H__


#import <Cocoa/Cocoa.h>


typedef enum
{
    XINE_FULLSCREEN_OVERSCAN,
    XINE_FULLSCREEN_CROP
} XineVideoWindowFullScreenMode;


@interface XineVideoWindow : NSWindow
{
    int                        width, height;
    id /* XineOpenGLView * */  xineView;
}

- (id) initWithContentSize:(NSSize)size;
- (id /* XineOpenGLView * */) xineView;

@end


/* NSWindow aspect ratio convenience methods */

@interface NSWindow (AspectRatioAdditions)

- (void) setKeepsAspectRatio:(BOOL)flag;
- (BOOL) keepsAspectRatio;

@end


#endif /* __HAVE_XINE_VIDEO_WINDOW_H__ */
