/*
 * File: pci_ids.h
 * This file was generated automatically. Don't modify it.
*/
#ifndef PCI_IDS_INCLUDED
#define PCI_IDS_INCLUDED 1

#include "pci_vendors.h"

/* Vendor: 0000: Gammagraphx, Inc. */
/* Vendor: 001a: Ascend Communications, Inc. */
/* Vendor: 0033: Paradyne corp. */
/* Vendor: 003d: Lockheed Martin-Marietta Corp */
/* Vendor: 0059: Tiger Jet Network Inc. (Wrong ID) */
/* Vendor: 0070: Hauppauge computer works Inc. */
#define DEVICE_HAUPPAUGE_WINTV_PVR_350			0x4000 /*WinTV PVR-350*/
#define DEVICE_HAUPPAUGE_WINTV_PVR_250_V1		0x4001 /*WinTV PVR-250 (v1)*/
#define DEVICE_HAUPPAUGE_WINTV_PVR_250			0x4009 /*WinTV PVR-250*/
#define DEVICE_HAUPPAUGE_WINTV_PVR_250_MCE		0x4801 /*WinTV PVR-250 MCE*/
/* Vendor: 0071: Nebula Electronics Ltd. */
/* Vendor: 0095: Silicon Image, Inc. (Wrong ID) */
#define DEVICE_SILICON_ULTRA_ATA_133_IDE		0x0680 /*Ultra ATA/133 IDE RAID CONTROLLER CARD*/
/* Vendor: 0100: Ncipher Corp Ltd */
/* Vendor: 018a: LevelOne */
#define DEVICE_LEVELONE_FPC_0106TX_MISPROGRAMMED_RTL81XX	0x0106 /*FPC-0106TX misprogrammed [RTL81xx]*/
/* Vendor: 021b: Compaq Computer Corporation */
#define DEVICE_COMPAQ_HNE_300_REALTEK_RTL8139C		0x8139 /*HNE-300 (RealTek RTL8139c) [iPaq Networking]*/
/* Vendor: 0291: Davicom Semiconductor, Inc. */
#define DEVICE_DAVICOM_DM9102A_DM9102AE_SM9102AF_ETHERNET	0x8212 /*DM9102A(DM9102AE, SM9102AF) Ethernet 100/10 MBit(Rev 40)*/
/* Vendor: 02ac: SpeedStream */
#define DEVICE_SPEEDSTREAM_1012_PCMCIA_10_100		0x1012 /*1012 PCMCIA 10/100 Ethernet Card [RTL81xx]*/
/* Vendor: 0357: TTTech AG */
#define DEVICE_TTTECH_TTP_MONITORING_CARD_V2_0		0x000a /*TTP-Monitoring Card V2.0*/
/* Vendor: 05e3: CyberDoor */
#define DEVICE_CYBERDOOR_CBD516				0x0701 /*CBD516*/
/* Vendor: 0675: Dynalink */
#define DEVICE_DYNALINK_IS64PH_ISDN_ADAPTER		0x1700 /*IS64PH ISDN Adapter*/
#define DEVICE_DYNALINK_IS64PH_ISDN_ADAPTER2		0x1702 /*IS64PH ISDN Adapter*/
/* Vendor: 0925: VIA Technologies, Inc. (Wrong ID) */
/* Vendor: 09c1: Arris */
#define DEVICE_ARRIS_CM_200E_CABLE			0x0704 /*CM 200E Cable Modem*/
/* Vendor: 0a89: BREA Technologies Inc */
/* Vendor: 0b49: ASCII Corporation */
#define DEVICE_ASCII_TRANCE_VIBRATOR			0x064f /*Trance Vibrator*/
/* Vendor: 0e11: Compaq Computer Corporation */
#define DEVICE_COMPAQ2_PCI_TO_EISA			0x0001 /*PCI to EISA Bridge*/
#define DEVICE_COMPAQ2_PCI_TO_ISA			0x0002 /*PCI to ISA Bridge*/
#define DEVICE_COMPAQ2_SMART_ARRAY_64XX			0x0046 /*Smart Array 64xx*/
	#define SUBDEVICE_COMPAQ2_SMART_ARRAY_641	0x0e11409a /*Smart Array 641*/
	#define SUBDEVICE_COMPAQ2_SMART_ARRAY_642	0x0e11409b /*Smart Array 642*/
	#define SUBDEVICE_COMPAQ2_SMART_ARRAY_6400	0x0e11409c /*Smart Array 6400*/
	#define SUBDEVICE_COMPAQ2_SMART_ARRAY_64002	0x0e11409d /*Smart Array 6400 EM*/
#define DEVICE_COMPAQ2_NC7132_GIGABIT_UPGRADE		0x0049 /*NC7132 Gigabit Upgrade Module*/
#define DEVICE_COMPAQ2_NC6136_GIGABIT_SERVER		0x004a /*NC6136 Gigabit Server Adapter*/
#define DEVICE_COMPAQ2_NC7770_1000BASETX		0x007c /*NC7770 1000BaseTX*/
#define DEVICE_COMPAQ2_NC6770_1000BASETX		0x007d /*NC6770 1000BaseTX*/
#define DEVICE_COMPAQ2_NC7780_1000BASETX		0x0085 /*NC7780 1000BaseTX*/
#define DEVICE_COMPAQ2_NC7760				0x00bb /*NC7760*/
#define DEVICE_COMPAQ2_NC7771				0x00ca /*NC7771*/
#define DEVICE_COMPAQ2_NC7781				0x00cb /*NC7781*/
#define DEVICE_COMPAQ2_NC7772				0x00cf /*NC7772*/
#define DEVICE_COMPAQ2_NC7782				0x00d0 /*NC7782*/
#define DEVICE_COMPAQ2_NC7783				0x00d1 /*NC7783*/
#define DEVICE_COMPAQ2_NC7761				0x00e3 /*NC7761*/
#define DEVICE_COMPAQ2_NETELLIGENT_4_16_TOKEN		0x0508 /*Netelligent 4/16 Token Ring*/
#define DEVICE_COMPAQ2_TRIFLEX_PENTIUM_BRIDGE_MODEL	0x1000 /*Triflex/Pentium Bridge, Model 1000*/
#define DEVICE_COMPAQ2_TRIFLEX_PENTIUM_BRIDGE_MODEL2	0x2000 /*Triflex/Pentium Bridge, Model 2000*/
#define DEVICE_COMPAQ2_QVISION_1280_P			0x3032 /*QVision 1280/p*/
#define DEVICE_COMPAQ2_QVISION_1280_P2			0x3033 /*QVision 1280/p*/
#define DEVICE_COMPAQ2_QVISION_1280_P3			0x3034 /*QVision 1280/p*/
#define DEVICE_COMPAQ2_4000_TRIFLEX			0x4000 /*4000 [Triflex]*/
#define DEVICE_COMPAQ2_SMART_2_P			0x4030 /*SMART-2/P*/
#define DEVICE_COMPAQ2_SMART_2SL			0x4031 /*SMART-2SL*/
#define DEVICE_COMPAQ2_SMART_ARRAY_3200			0x4032 /*Smart Array 3200*/
#define DEVICE_COMPAQ2_SMART_ARRAY_3100ES		0x4033 /*Smart Array 3100ES*/
#define DEVICE_COMPAQ2_SMART_ARRAY_221			0x4034 /*Smart Array 221*/
#define DEVICE_COMPAQ2_INTEGRATED_ARRAY			0x4040 /*Integrated Array*/
#define DEVICE_COMPAQ2_COMPAQ_RAID_LC2			0x4048 /*Compaq Raid LC2*/
#define DEVICE_COMPAQ2_SMART_ARRAY_4200			0x4050 /*Smart Array 4200*/
#define DEVICE_COMPAQ2_SMART_ARRAY_4250ES		0x4051 /*Smart Array 4250ES*/
#define DEVICE_COMPAQ2_SMART_ARRAY_431			0x4058 /*Smart Array 431*/
#define DEVICE_COMPAQ2_SMART_ARRAY_5300			0x4070 /*Smart Array 5300*/
#define DEVICE_COMPAQ2_SMART_ARRAY_5I			0x4080 /*Smart Array 5i*/
#define DEVICE_COMPAQ2_SMART_ARRAY_532			0x4082 /*Smart Array 532*/
#define DEVICE_COMPAQ2_SMART_ARRAY_5312			0x4083 /*Smart Array 5312*/
#define DEVICE_COMPAQ2_SMART_ARRAY_6I			0x4091 /*Smart Array 6i*/
#define DEVICE_COMPAQ2_SMART_ARRAY_641			0x409a /*Smart Array 641*/
#define DEVICE_COMPAQ2_SMART_ARRAY_642			0x409b /*Smart Array 642*/
#define DEVICE_COMPAQ2_SMART_ARRAY_6400			0x409c /*Smart Array 6400*/
#define DEVICE_COMPAQ2_SMART_ARRAY_64002		0x409d /*Smart Array 6400 EM*/
#define DEVICE_COMPAQ2_HOTPLUG_PCI_BRIDGE		0x6010 /*HotPlug PCI Bridge 6010*/
#define DEVICE_COMPAQ2_USB_CONTROLLER			0x7020 /*USB Controller*/
#define DEVICE_COMPAQ2_FIBRE_CHANNEL_HOST		0xa0ec /*Fibre Channel Host Controller*/
#define DEVICE_COMPAQ2_ADVANCED_SYSTEM_MANAGEMENT	0xa0f0 /*Advanced System Management Controller*/
#define DEVICE_COMPAQ2_TRIFLEX_PCI_TO			0xa0f3 /*Triflex PCI to ISA Bridge*/
#define DEVICE_COMPAQ2_PCI_HOTPLUG_CONTROLLER		0xa0f7 /*PCI Hotplug Controller*/
	#define SUBDEVICE_COMPAQ2_PCI_HOTPLUG_CONTROLLER	0x8086002a /*PCI Hotplug Controller A*/
	#define SUBDEVICE_COMPAQ2_PCI_HOTPLUG_CONTROLLER2	0x8086002b /*PCI Hotplug Controller B*/
#define DEVICE_COMPAQ2_ZFMICRO_CHIPSET_USB		0xa0f8 /*ZFMicro Chipset USB*/
#define DEVICE_COMPAQ2_FIBRECHANNEL_HBA_TACHYON		0xa0fc /*FibreChannel HBA Tachyon*/
#define DEVICE_COMPAQ2_SMART_2_P_RAID_CONTROLLER	0xae10 /*Smart-2/P RAID Controller*/
	#define SUBDEVICE_COMPAQ2_SMART_2_P_ARRAY_CONTROLLER	0x0e114030 /*Smart-2/P Array Controller*/
	#define SUBDEVICE_COMPAQ2_SMART_2SL_ARRAY_CONTROLLER	0x0e114031 /*Smart-2SL Array Controller*/
	#define SUBDEVICE_COMPAQ2_SMART_ARRAY_CONTROLLER	0x0e114032 /*Smart Array Controller*/
	#define SUBDEVICE_COMPAQ2_SMART_3100ES_ARRAY	0x0e114033 /*Smart 3100ES Array Controller*/
#define DEVICE_COMPAQ2_MIS_L				0xae29 /*MIS-L*/
#define DEVICE_COMPAQ2_MPC				0xae2a /*MPC*/
#define DEVICE_COMPAQ2_MIS_E				0xae2b /*MIS-E*/
#define DEVICE_COMPAQ2_SYSTEM_MANAGEMENT_CONTROLLER	0xae31 /*System Management Controller*/
#define DEVICE_COMPAQ2_NETELLIGENT_10_100_TX		0xae32 /*Netelligent 10/100 TX PCI UTP*/
#define DEVICE_COMPAQ2_TRIFLEX_DUAL_EIDE		0xae33 /*Triflex Dual EIDE Controller*/
#define DEVICE_COMPAQ2_NETELLIGENT_10_T			0xae34 /*Netelligent 10 T PCI UTP*/
#define DEVICE_COMPAQ2_INTEGRATED_NETFLEX_3_P		0xae35 /*Integrated NetFlex-3/P*/
#define DEVICE_COMPAQ2_NETELLIGENT_DUAL_10_100		0xae40 /*Netelligent Dual 10/100 TX PCI UTP*/
#define DEVICE_COMPAQ2_NETELLIGENT_INTEGRATED_10_100	0xae43 /*Netelligent Integrated 10/100 TX UTP*/
#define DEVICE_COMPAQ2_CETUS_L				0xae69 /*CETUS-L*/
#define DEVICE_COMPAQ2_NORTHSTAR			0xae6c /*Northstar*/
#define DEVICE_COMPAQ2_NORTHSTAR_CPU_TO			0xae6d /*NorthStar CPU to PCI Bridge*/
#define DEVICE_COMPAQ2_NETELLIGENT_10_100_TX2		0xb011 /*Netelligent 10/100 TX Embedded UTP*/
#define DEVICE_COMPAQ2_NETELLIGENT_10_T_2		0xb012 /*Netelligent 10 T/2 PCI UTP/Coax*/
#define DEVICE_COMPAQ2_NC3120_FAST_ETHERNET		0xb01e /*NC3120 Fast Ethernet NIC*/
#define DEVICE_COMPAQ2_NC3122_FAST_ETHERNET		0xb01f /*NC3122 Fast Ethernet NIC*/
#define DEVICE_COMPAQ2_NC1120_ETHERNET_NIC		0xb02f /*NC1120 Ethernet NIC*/
#define DEVICE_COMPAQ2_NETELLIGENT_10_100_TX3		0xb030 /*Netelligent 10/100 TX UTP*/
#define DEVICE_COMPAQ2_10_100_TX_PCI			0xb04a /*10/100 TX PCI Intel WOL UTP Controller*/
#define DEVICE_COMPAQ2_SMART_ARRAY_53002		0xb060 /*Smart Array 5300 Controller*/
#define DEVICE_COMPAQ2_NC3161_FAST_ETHERNET		0xb0c6 /*NC3161 Fast Ethernet NIC*/
#define DEVICE_COMPAQ2_NC3160_FAST_ETHERNET		0xb0c7 /*NC3160 Fast Ethernet NIC*/
#define DEVICE_COMPAQ2_NC3121_FAST_ETHERNET		0xb0d7 /*NC3121 Fast Ethernet NIC*/
#define DEVICE_COMPAQ2_NC3131_FAST_ETHERNET		0xb0dd /*NC3131 Fast Ethernet NIC*/
#define DEVICE_COMPAQ2_NC3132_FAST_ETHERNET		0xb0de /*NC3132 Fast Ethernet Module*/
#define DEVICE_COMPAQ2_NC6132_GIGABIT_MODULE		0xb0df /*NC6132 Gigabit Module*/
#define DEVICE_COMPAQ2_NC6133_GIGABIT_MODULE		0xb0e0 /*NC6133 Gigabit Module*/
#define DEVICE_COMPAQ2_NC3133_FAST_ETHERNET		0xb0e1 /*NC3133 Fast Ethernet Module*/
#define DEVICE_COMPAQ2_NC6134_GIGABIT_NIC		0xb123 /*NC6134 Gigabit NIC*/
#define DEVICE_COMPAQ2_NC3163_FAST_ETHERNET		0xb134 /*NC3163 Fast Ethernet NIC*/
#define DEVICE_COMPAQ2_NC3162_FAST_ETHERNET		0xb13c /*NC3162 Fast Ethernet NIC*/
#define DEVICE_COMPAQ2_NC3123_FAST_ETHERNET		0xb144 /*NC3123 Fast Ethernet NIC*/
#define DEVICE_COMPAQ2_NC3134_FAST_ETHERNET		0xb163 /*NC3134 Fast Ethernet NIC*/
#define DEVICE_COMPAQ2_NC3165_FAST_ETHERNET		0xb164 /*NC3165 Fast Ethernet Upgrade Module*/
#define DEVICE_COMPAQ2_SMART_ARRAY_5I_532		0xb178 /*Smart Array 5i/532*/
	#define SUBDEVICE_COMPAQ2_SMART_ARRAY_5I	0x0e114080 /*Smart Array 5i*/
	#define SUBDEVICE_COMPAQ2_SMART_ARRAY_532	0x0e114082 /*Smart Array 532*/
	#define SUBDEVICE_COMPAQ2_SMART_ARRAY_5312	0x0e114083 /*Smart Array 5312*/
#define DEVICE_COMPAQ2_NC7131_GIGABIT_SERVER		0xb1a4 /*NC7131 Gigabit Server Adapter*/
#define DEVICE_COMPAQ2_MEMORY_HOT_PLUG_CONTROLLER	0xb200 /*Memory Hot-Plug Controller*/
#define DEVICE_COMPAQ2_INTEGRATED_LIGHTS_OUT		0xb203 /*Integrated Lights Out Controller*/
#define DEVICE_COMPAQ2_INTEGRATED_LIGHTS_OUT2		0xb204 /*Integrated Lights Out  Processor*/
#define DEVICE_COMPAQ2_NETFLEX_3_P_THUNDERLAN_1_0	0xf130 /*NetFlex-3/P ThunderLAN 1.0*/
#define DEVICE_COMPAQ2_NETFLEX_3_P_THUNDERLAN_2_3	0xf150 /*NetFlex-3/P ThunderLAN 2.3*/
/* Vendor: 0e55: HaSoTec GmbH */
/* Vendor: 1000: LSI Logic / Symbios Logic */
#define DEVICE_LSI_53C810				0x0001 /*53c810*/
	#define SUBDEVICE_LSI_LSI53C810AE_PCI_TO	0x10001000 /*LSI53C810AE PCI to SCSI I/O Processor*/
#define DEVICE_LSI_53C820				0x0002 /*53c820*/
#define DEVICE_LSI_53C825				0x0003 /*53c825*/
	#define SUBDEVICE_LSI_LSI53C825AE_PCI_TO	0x10001000 /*LSI53C825AE PCI to SCSI I/O Processor (Ultra Wide)*/
#define DEVICE_LSI_53C815				0x0004 /*53c815*/
#define DEVICE_LSI_53C810AP				0x0005 /*53c810AP*/
#define DEVICE_LSI_53C860				0x0006 /*53c860*/
	#define SUBDEVICE_LSI_LSI53C860E_PCI_TO		0x10001000 /*LSI53C860E PCI to Ultra SCSI I/O Processor*/
#define DEVICE_LSI_53C1510				0x000a /*53c1510*/
	#define SUBDEVICE_LSI_LSI53C1510_PCI_TO		0x10001000 /*LSI53C1510 PCI to Dual Channel Wide Ultra2 SCSI Controller (Nonintelligent mode)*/
#define DEVICE_LSI_53C896_897				0x000b /*53C896/897*/
	#define SUBDEVICE_LSI_EOB003_SERIES_SCSI	0x0e116004 /*EOB003 Series SCSI host adapter*/
	#define SUBDEVICE_LSI_LSI53C896_7_PCI_TO	0x10001000 /*LSI53C896/7 PCI to Dual Channel Ultra2 SCSI Multifunction Controller*/
	#define SUBDEVICE_LSI_LSI22910_PCI_TO		0x10001010 /*LSI22910 PCI to Dual Channel Ultra2 SCSI host adapter*/
	#define SUBDEVICE_LSI_LSI21002_PCI_TO		0x10001020 /*LSI21002 PCI to Dual Channel Ultra2 SCSI host adapter*/
	#define SUBDEVICE_LSI_6221L_4U			0x13e91000 /*6221L-4U*/
#define DEVICE_LSI_53C895				0x000c /*53c895*/
	#define SUBDEVICE_LSI_LSI8951U_PCI_TO		0x10001010 /*LSI8951U PCI to Ultra2 SCSI host adapter*/
	#define SUBDEVICE_LSI_LSI8952U_PCI_TO		0x10001020 /*LSI8952U PCI to Ultra2 SCSI host adapter*/
	#define SUBDEVICE_LSI_DC_390U2B_SCSI_ADAPTER	0x1de13906 /*DC-390U2B SCSI adapter*/
	#define SUBDEVICE_LSI_DC_390U2W			0x1de13907 /*DC-390U2W*/
#define DEVICE_LSI_53C885				0x000d /*53c885*/
#define DEVICE_LSI_53C875				0x000f /*53c875*/
	#define SUBDEVICE_LSI_EMBEDDED_ULTRA_WIDE	0x0e117004 /*Embedded Ultra Wide SCSI Controller*/
	#define SUBDEVICE_LSI_LSI53C876_E_PCI_TO	0x10001000 /*LSI53C876/E PCI to Dual Channel SCSI Controller*/
	#define SUBDEVICE_LSI_LSI22801_PCI_TO		0x10001010 /*LSI22801 PCI to Dual Channel Ultra SCSI host adapter*/
	#define SUBDEVICE_LSI_LSI22802_PCI_TO		0x10001020 /*LSI22802 PCI to Dual Channel Ultra SCSI host adapter*/
	#define SUBDEVICE_LSI_FIREPORT_40_DUAL		0x10928760 /*FirePort 40 Dual SCSI Controller*/
	#define SUBDEVICE_LSI_DC390F_U_ULTRA_WIDE	0x1de13904 /*DC390F/U Ultra Wide SCSI Adapter*/
	#define SUBDEVICE_LSI_CC7_CR7_CP7_VC7_VP7_VR7_MAINBOARD	0x4c531000 /*CC7/CR7/CP7/VC7/VP7/VR7 mainboard*/
	#define SUBDEVICE_LSI_CT7_MAINBOARD		0x4c531050 /*CT7 mainboard*/
#define DEVICE_LSI_53C15102				0x0010 /*53C1510*/
	#define SUBDEVICE_LSI_INTEGRATED_ARRAY_CONTROLLER	0x0e114040 /*Integrated Array Controller*/
	#define SUBDEVICE_LSI_RAID_LC2_CONTROLLER	0x0e114048 /*RAID LC2 Controller*/
	#define SUBDEVICE_LSI_53C1510_PCI_TO		0x10001000 /*53C1510 PCI to Dual Channel Wide Ultra2 SCSI Controller (Intelligent mode)*/
#define DEVICE_LSI_53C895A				0x0012 /*53c895a*/
	#define SUBDEVICE_LSI_LSI53C895A_PCI_TO		0x10001000 /*LSI53C895A PCI to Ultra2 SCSI Controller*/
#define DEVICE_LSI_53C875A				0x0013 /*53c875a*/
	#define SUBDEVICE_LSI_LSI53C875A_PCI_TO		0x10001000 /*LSI53C875A PCI to Ultra SCSI Controller*/
#define DEVICE_LSI_53C1010_ULTRA3_SCSI			0x0020 /*53c1010 Ultra3 SCSI Adapter*/
	#define SUBDEVICE_LSI_LSI53C1010_33_PCI_TO	0x10001000 /*LSI53C1010-33 PCI to Dual Channel Ultra160 SCSI Controller*/
	#define SUBDEVICE_LSI_DC_390U3W			0x1de11020 /*DC-390U3W*/
#define DEVICE_LSI_53C1010_66MHZ			0x0021 /*53c1010 66MHz  Ultra3 SCSI Adapter*/
	#define SUBDEVICE_LSI_LSI53C1000_1000R_1010R_1010_66_PCI_TO	0x10001000 /*LSI53C1000/1000R/1010R/1010-66 PCI to Ultra160 SCSI Controller*/
	#define SUBDEVICE_LSI_ASUS_TR_DLS_ONBOARD	0x10001010 /*Asus TR-DLS onboard 53C1010-66*/
	#define SUBDEVICE_LSI_PMC_USCSI3		0x124b1070 /*PMC-USCSI3*/
	#define SUBDEVICE_LSI_CT8_MAINBOARD		0x4c531080 /*CT8 mainboard*/
	#define SUBDEVICE_LSI_P017_MEZZANINE_32_BIT	0x4c531300 /*P017 mezzanine (32-bit PMC)*/
	#define SUBDEVICE_LSI_P017_MEZZANINE_64_BIT	0x4c531310 /*P017 mezzanine (64-bit PMC)*/
#define DEVICE_LSI_53C1030_PCI_X_FUSION_MPT		0x0030 /*53c1030 PCI-X Fusion-MPT Dual Ultra320 SCSI*/
	#define SUBDEVICE_LSI_POWEREDGE_2600		0x10280123 /*PowerEdge 2600*/
	#define SUBDEVICE_LSI_POWEREDGE_1750		0x1028014a /*PowerEdge 1750*/
	#define SUBDEVICE_LSI_POWEREDGE_1850_MPT	0x1028016c /*PowerEdge 1850 MPT Fusion SCSI/RAID (Perc 4)*/
	#define SUBDEVICE_LSI_LSI_U320_SCSI		0x10281010 /*LSI U320 SCSI Controller*/
#define DEVICE_LSI_53C1030ZC_PCI_X_FUSION_MPT		0x0031 /*53c1030ZC PCI-X Fusion-MPT Dual Ultra320 SCSI*/
#define DEVICE_LSI_53C1035_PCI_X_FUSION_MPT		0x0032 /*53c1035 PCI-X Fusion-MPT Dual Ultra320 SCSI*/
	#define SUBDEVICE_LSI_LSI53C1020_1030_PCI_X_TO	0x10001000 /*LSI53C1020/1030 PCI-X to Ultra320 SCSI Controller*/
#define DEVICE_LSI_1030ZC_53C1035_PCI_X_FUSION_MPT	0x0033 /*1030ZC_53c1035 PCI-X Fusion-MPT Dual Ultra320 SCSI*/
#define DEVICE_LSI_53C1035_PCI_X_FUSION_MPT2		0x0040 /*53c1035 PCI-X Fusion-MPT Dual Ultra320 SCSI*/
	#define SUBDEVICE_LSI_MEGARAID_SCSI_320_2XR	0x10000033 /*MegaRAID SCSI 320-2XR*/
	#define SUBDEVICE_LSI_MEGARAID_SCSI_320_2XRWS	0x10000066 /*MegaRAID SCSI 320-2XRWS*/
#define DEVICE_LSI_53C1035ZC_PCI_X_FUSION_MPT		0x0041 /*53C1035ZC PCI-X Fusion-MPT Dual Ultra320 SCSI*/
#define DEVICE_LSI_53C875J				0x008f /*53c875J*/
	#define SUBDEVICE_LSI_FIREPORT_40_SCSI		0x10928000 /*FirePort 40 SCSI Controller*/
	#define SUBDEVICE_LSI_FIREPORT_40_DUAL2		0x10928760 /*FirePort 40 Dual SCSI Host Adapter*/
#define DEVICE_LSI_MEGARAID				0x0407 /*MegaRAID*/
	#define SUBDEVICE_LSI_MEGARAID_530_SCSI		0x10000530 /*MegaRAID 530 SCSI 320-0X RAID Controller*/
	#define SUBDEVICE_LSI_MEGARAID_531_SCSI		0x10000531 /*MegaRAID 531 SCSI 320-4X RAID Controller*/
	#define SUBDEVICE_LSI_MEGARAID_532_SCSI		0x10000532 /*MegaRAID 532 SCSI 320-2X RAID Controller*/
	#define SUBDEVICE_LSI_POWEREDGE_EXPANDABLE_RAID	0x10280531 /*PowerEdge Expandable RAID Controller 4/QC*/
	#define SUBDEVICE_LSI_POWEREDGE_EXPANDABLE_RAID2	0x10280533 /*PowerEdge Expandable RAID Controller 4/QC*/
	#define SUBDEVICE_LSI_MEGARAID_INTEL_RAID	0x80860530 /*MegaRAID Intel RAID Controller SRCZCRX*/
	#define SUBDEVICE_LSI_MEGARAID_INTEL_RAID2	0x80860532 /*MegaRAID Intel RAID Controller SRCU42X*/
#define DEVICE_LSI_MEGARAID2				0x0408 /*MegaRAID*/
	#define SUBDEVICE_LSI_MEGARAID_SCSI_320_1E	0x10000001 /*MegaRAID SCSI 320-1E RAID Controller*/
	#define SUBDEVICE_LSI_MEGARAID_SCSI_320_2E	0x10000002 /*MegaRAID SCSI 320-2E RAID Controller*/
	#define SUBDEVICE_LSI_MEGARAID_ACER_ROMB_2E	0x1025004d /*MegaRAID ACER ROMB-2E RAID Controller*/
	#define SUBDEVICE_LSI_POWEREDGE_RAID_CONTROLLER	0x10280001 /*PowerEdge RAID Controller PERC4e/SC*/
	#define SUBDEVICE_LSI_POWEREDGE_RAID_CONTROLLER2	0x10280002 /*PowerEdge RAID Controller PERC4e/DC*/
	#define SUBDEVICE_LSI_FSC_MEGARAID_PCI		0x17341065 /*FSC MegaRAID PCI Express ROMB*/
	#define SUBDEVICE_LSI_MEGARAID_INTEL_RAID3	0x80860002 /*MegaRAID Intel RAID Controller SRCU42E*/
#define DEVICE_LSI_MEGARAID3				0x0409 /*MegaRAID*/
	#define SUBDEVICE_LSI_MEGARAID_SATA_300_4X	0x10003004 /*MegaRAID SATA 300-4X RAID Controller*/
	#define SUBDEVICE_LSI_MEGARAID_SATA_300_8X	0x10003008 /*MegaRAID SATA 300-8X RAID Controller*/
	#define SUBDEVICE_LSI_MEGARAID_RAID_CONTROLLER	0x80863008 /*MegaRAID RAID Controller SRCS28X*/
	#define SUBDEVICE_LSI_MEGARAID_RAID_CONTROLLER2	0x80863431 /*MegaRAID RAID Controller Alief SROMBU42E*/
	#define SUBDEVICE_LSI_MEGARAID_RAID_CONTROLLER3	0x80863499 /*MegaRAID RAID Controller Harwich SROMBU42E*/
#define DEVICE_LSI_FC909_FIBRE_CHANNEL			0x0621 /*FC909 Fibre Channel Adapter*/
#define DEVICE_LSI_FC929_FIBRE_CHANNEL			0x0622 /*FC929 Fibre Channel Adapter*/
	#define SUBDEVICE_LSI_44929_O_DUAL		0x10001020 /*44929 O Dual Fibre Channel card*/
#define DEVICE_LSI_FC929_LAN				0x0623 /*FC929 LAN*/
#define DEVICE_LSI_FC919_FIBRE_CHANNEL			0x0624 /*FC919 Fibre Channel Adapter*/
#define DEVICE_LSI_FC919_LAN				0x0625 /*FC919 LAN*/
#define DEVICE_LSI_FC929X_FIBRE_CHANNEL			0x0626 /*FC929X Fibre Channel Adapter*/
	#define SUBDEVICE_LSI_7202_XP_LC_DUAL_FIBRE	0x10001010 /*7202-XP-LC Dual Fibre Channel card*/
#define DEVICE_LSI_FC929X_LAN				0x0627 /*FC929X LAN*/
#define DEVICE_LSI_FC919X_FIBRE_CHANNEL			0x0628 /*FC919X Fibre Channel Adapter*/
#define DEVICE_LSI_FC919X_LAN				0x0629 /*FC919X LAN*/
#define DEVICE_LSI_83C885_NT50_DIGITALSCAPE		0x0701 /*83C885 NT50 DigitalScape Fast Ethernet*/
#define DEVICE_LSI_YELLOWFIN_G_NIC_GIGABIT		0x0702 /*Yellowfin G-NIC gigabit ethernet*/
	#define SUBDEVICE_LSI_PEI100X			0x13180000 /*PEI100X*/
#define DEVICE_LSI_SA2010				0x0804 /*SA2010*/
#define DEVICE_LSI_SA2010ZC				0x0805 /*SA2010ZC*/
#define DEVICE_LSI_SA2020				0x0806 /*SA2020*/
#define DEVICE_LSI_SA2020ZC				0x0807 /*SA2020ZC*/
#define DEVICE_LSI_61C102				0x0901 /*61C102*/
#define DEVICE_LSI_63C815				0x1000 /*63C815*/
#define DEVICE_LSI_MEGARAID4				0x1960 /*MegaRAID*/
	#define SUBDEVICE_LSI_MEGARAID_518_SCSI		0x10000518 /*MegaRAID 518 SCSI 320-2 Controller*/
	#define SUBDEVICE_LSI_MEGARAID_520_SCSI		0x10000520 /*MegaRAID 520 SCSI 320-1 Controller*/
	#define SUBDEVICE_LSI_MEGARAID_522_I4		0x10000522 /*MegaRAID 522 i4 133 RAID Controller*/
	#define SUBDEVICE_LSI_MEGARAID_SATA_150_6	0x10000523 /*MegaRAID SATA 150-6 RAID Controller*/
	#define SUBDEVICE_LSI_MEGARAID_SATA_150_4	0x10004523 /*MegaRAID SATA 150-4 RAID Controller*/
	#define SUBDEVICE_LSI_MEGARAID_ZCR_SCSI		0x1000a520 /*MegaRAID ZCR SCSI 320-0 Controller*/
	#define SUBDEVICE_LSI_MEGARAID_518_DELL		0x10280518 /*MegaRAID 518 DELL PERC 4/DC RAID Controller*/
	#define SUBDEVICE_LSI_MEGARAID_520_DELL		0x10280520 /*MegaRAID 520 DELL PERC 4/SC RAID Controller*/
	#define SUBDEVICE_LSI_POWEREDGE_EXPANDABLE_RAID3	0x10280531 /*PowerEdge Expandable RAID Controller 4/QC*/
	#define SUBDEVICE_LSI_POWEREDGE_EXPANDABLE_RAID4	0x10280533 /*PowerEdge Expandable RAID Controller 4/QC*/
	#define SUBDEVICE_LSI_MEGARAIDRAID_CONTROLLER_SRCU41L	0x80860520 /*MegaRAIDRAID Controller SRCU41L*/
	#define SUBDEVICE_LSI_MEGARAID_RAID_CONTROLLER4	0x80860523 /*MegaRAID RAID Controller SRCS16*/
/* Vendor: 1001: Kolter Electronic */
#define DEVICE_KOLTER_PCI_1616_MEASUREMENT		0x0010 /*PCI 1616 Measurement card with 32 digital I/O lines*/
#define DEVICE_KOLTER_OPTO_PCI_OPTO_ISOLATED_DIGITAL	0x0011 /*OPTO-PCI Opto-Isolated digital I/O board*/
#define DEVICE_KOLTER_PCI_AD_DA_ANALOGUE_I_O		0x0012 /*PCI-AD/DA Analogue I/O board*/
#define DEVICE_KOLTER_PCI_OPTO_RELAIS_DIGITAL_I_O	0x0013 /*PCI-OPTO-RELAIS Digital I/O board with relay outputs*/
#define DEVICE_KOLTER_PCI_COUNTER_TIMER_COUNTER_TIMER	0x0014 /*PCI-Counter/Timer Counter Timer board*/
#define DEVICE_KOLTER_PCI_DAC416_ANALOGUE_OUTPUT	0x0015 /*PCI-DAC416 Analogue output board*/
#define DEVICE_KOLTER_PCI_MFB_ANALOGUE_I_O		0x0016 /*PCI-MFB Analogue I/O board*/
#define DEVICE_KOLTER_PROTO_3_PCI_PROTOTYPING		0x0017 /*PROTO-3 PCI Prototyping board*/
#define DEVICE_KOLTER_INI_9100_9100W_SCSI_HOST		0x9100 /*INI-9100/9100W SCSI Host*/
/* Vendor: 1002: ATI Technologies Inc */
#define DEVICE_ATI_M24_1P_RADEON			0x3150 /*M24 1P [Radeon Mobility X600]*/
#define DEVICE_ATI_M24_1T_FIREGL			0x3154 /*M24 1T [FireGL M24 GL]*/
#define DEVICE_ATI_RV380_0X3E50_RADEON			0x3e50 /*RV380 0x3e50 [Radeon X600]*/
#define DEVICE_ATI_RV380_0X3E54_FIREGL			0x3e54 /*RV380 0x3e54 [FireGL V3200]*/
#define DEVICE_ATI_RV380_RADEON_X600			0x3e70 /*RV380 [Radeon X600] Secondary*/
#define DEVICE_ATI_RADEON_IGP_320			0x4136 /*Radeon IGP 320 M*/
#define DEVICE_ATI_RADEON_IGP330_340_350		0x4137 /*Radeon IGP330/340/350*/
#define DEVICE_ATI_R300_AD_RADEON			0x4144 /*R300 AD [Radeon 9500 Pro]*/
#define DEVICE_ATI_R300_AE_RADEON			0x4145 /*R300 AE [Radeon 9700 Pro]*/
#define DEVICE_ATI_R300_AF_RADEON			0x4146 /*R300 AF [Radeon 9700 Pro]*/
#define DEVICE_ATI_R300_AG_FIREGL			0x4147 /*R300 AG [FireGL Z1/X1]*/
#define DEVICE_ATI_R350_AH_RADEON			0x4148 /*R350 AH [Radeon 9800]*/
#define DEVICE_ATI_R350_AI_RADEON			0x4149 /*R350 AI [Radeon 9800]*/
#define DEVICE_ATI_R350_AJ_RADEON			0x414a /*R350 AJ [Radeon 9800]*/
#define DEVICE_ATI_R350_AK_FIRE				0x414b /*R350 AK [Fire GL X2]*/
#define DEVICE_ATI_RV350_AP_RADEON			0x4150 /*RV350 AP [Radeon 9600]*/
	#define SUBDEVICE_ATI_R9600_PRO_PRIMARY		0x10020002 /*R9600 Pro primary (Asus OEM for HP)*/
	#define SUBDEVICE_ATI_R9600_PRO_SECONDARY	0x10020003 /*R9600 Pro secondary (Asus OEM for HP)*/
	#define SUBDEVICE_ATI_GIGA_BYTE_GV_R96128D_PRIMARY	0x14584024 /*Giga-Byte GV-R96128D Primary*/
	#define SUBDEVICE_ATI_POWERCOLOR_R96A_C3N	0x148c2064 /*PowerColor R96A-C3N*/
	#define SUBDEVICE_ATI_POWERCOLOR_R96A_C3N2	0x148c2066 /*PowerColor R96A-C3N*/
	#define SUBDEVICE_ATI_SAPPHIRE_ATLANTIS_RADEON	0x174b7c19 /*Sapphire Atlantis Radeon 9600 Pro*/
	#define SUBDEVICE_ATI_GC_R9600PRO_PRIMARY_SAPPHIRE	0x174b7c29 /*GC-R9600PRO Primary [Sapphire]*/
	#define SUBDEVICE_ATI_RADEON_9600_256MB		0x17ee2002 /*Radeon 9600 256Mb Primary*/
	#define SUBDEVICE_ATI_GC_R9600PRO_PRIMARY	0x18bc0101 /*GC-R9600PRO Primary*/
#define DEVICE_ATI_RV350_AQ_RADEON			0x4151 /*RV350 AQ [Radeon 9600]*/
	#define SUBDEVICE_ATI_A9600SE			0x1043c004 /*A9600SE*/
#define DEVICE_ATI_RV350_AR_RADEON			0x4152 /*RV350 AR [Radeon 9600]*/
	#define SUBDEVICE_ATI_RADEON_9600XT		0x10020002 /*Radeon 9600XT*/
	#define SUBDEVICE_ATI_RADEON_9600_XT		0x1043c002 /*Radeon 9600 XT TVD*/
#define DEVICE_ATI_RV350_AS_RADEON			0x4153 /*RV350 AS [Radeon 9600 AS]*/
#define DEVICE_ATI_RV350_AT_FIRE			0x4154 /*RV350 AT [Fire GL T2]*/
#define DEVICE_ATI_RV350_AU_FIRE			0x4155 /*RV350 AU [Fire GL T2]*/
#define DEVICE_ATI_RV350_AV_FIRE			0x4156 /*RV350 AV [Fire GL T2]*/
#define DEVICE_ATI_RV350_AW_FIRE			0x4157 /*RV350 AW [Fire GL T2]*/
#define DEVICE_ATI_68800AX_MACH32			0x4158 /*68800AX [Mach32]*/
#define DEVICE_ATI_R300_AD_RADEON2			0x4164 /*R300 AD [Radeon 9500 Pro] (Secondary)*/
#define DEVICE_ATI_R300_AE_RADEON2			0x4165 /*R300 AE [Radeon 9700 Pro] (Secondary)*/
#define DEVICE_ATI_R300_AF_RADEON2			0x4166 /*R300 AF [Radeon 9700 Pro] (Secondary)*/
#define DEVICE_ATI_RADEON_R350_RADEON			0x4168 /*Radeon R350 [Radeon 9800] (Secondary)*/
#define DEVICE_ATI_RV350_AP_RADEON2			0x4170 /*RV350 AP [Radeon 9600] (Secondary)*/
	#define SUBDEVICE_ATI_GIGA_BYTE_GV_R96128D_SECONDARY	0x14584025 /*Giga-Byte GV-R96128D Secondary*/
	#define SUBDEVICE_ATI_POWERCOLOR_R96A_C3N_SECONDARY	0x148c2067 /*PowerColor R96A-C3N (Secondary)*/
	#define SUBDEVICE_ATI_GC_R9600PRO_SECONDARY_SAPPHIRE	0x174b7c28 /*GC-R9600PRO Secondary [Sapphire]*/
	#define SUBDEVICE_ATI_RADEON_9600_256MB2	0x17ee2003 /*Radeon 9600 256Mb Secondary*/
	#define SUBDEVICE_ATI_GC_R9600PRO_SECONDARY	0x18bc0100 /*GC-R9600PRO Secondary*/
#define DEVICE_ATI_RV350_AQ_RADEON2			0x4171 /*RV350 AQ [Radeon 9600] (Secondary)*/
	#define SUBDEVICE_ATI_A9600SE_SECONDARY		0x1043c005 /*A9600SE (Secondary)*/
#define DEVICE_ATI_RV350_AR_RADEON2			0x4172 /*RV350 AR [Radeon 9600] (Secondary)*/
	#define SUBDEVICE_ATI_RADEON_9600XT_SECONDARY	0x10020003 /*Radeon 9600XT (Secondary)*/
	#define SUBDEVICE_ATI_A9600XT_SECONDARY		0x1043c003 /*A9600XT (Secondary)*/
#define DEVICE_ATI_RV350_RADEON				0x4173 /*RV350 ?? [Radeon 9550] (Secondary)*/
#define DEVICE_ATI_RADEON_7000_IGP			0x4237 /*Radeon 7000 IGP*/
#define DEVICE_ATI_R200_BB_RADEON			0x4242 /*R200 BB [Radeon All in Wonder 8500DV]*/
	#define SUBDEVICE_ATI_RADEON_8500_AIW		0x100202aa /*Radeon 8500 AIW DV Edition*/
#define DEVICE_ATI_R200_BC_RADEON			0x4243 /*R200 BC [Radeon All in Wonder 8500]*/
#define DEVICE_ATI_RADEON_MOBILITY_U1			0x4336 /*Radeon Mobility U1*/
	#define SUBDEVICE_ATI_PAVILION_ZE4400_BUILTIN	0x103c0024 /*Pavilion ze4400 builtin Video*/
#define DEVICE_ATI_RADEON_IGP_330M_340M_350M		0x4337 /*Radeon IGP 330M/340M/350M*/
	#define SUBDEVICE_ATI_THINKPAD_R40E_2684_HVG	0x1014053a /*ThinkPad R40e (2684-HVG) builtin VGA controller*/
	#define SUBDEVICE_ATI_RADEON_IGP_345M		0x103c0850 /*Radeon IGP 345M*/
#define DEVICE_ATI_IXP150_AC_97_AUDIO			0x4341 /*IXP150 AC'97 Audio Controller*/
#define DEVICE_ATI_EHCI_USB_CONTROLLER			0x4345 /*EHCI USB Controller*/
#define DEVICE_ATI_OHCI_USB_CONTROLLER			0x4347 /*OHCI USB Controller #1*/
#define DEVICE_ATI_OHCI_USB_CONTROLLER2			0x4348 /*OHCI USB Controller #2*/
#define DEVICE_ATI_IXP_AC_97_MODEM			0x434d /*IXP AC'97 Modem*/
#define DEVICE_ATI_ATI_SMBUS				0x4353 /*ATI SMBus*/
#define DEVICE_ATI_215CT_MACH64_CT			0x4354 /*215CT [Mach64 CT]*/
#define DEVICE_ATI_210888CX_MACH64_CX			0x4358 /*210888CX [Mach64 CX]*/
#define DEVICE_ATI_RADEON_MOBILITY_7000			0x4437 /*Radeon Mobility 7000 IGP*/
#define DEVICE_ATI_210888ET_MACH64_ET			0x4554 /*210888ET [Mach64 ET]*/
#define DEVICE_ATI_MACH64_VT				0x4654 /*Mach64 VT*/
#define DEVICE_ATI_3D_RAGE_PRO				0x4742 /*3D Rage Pro AGP 1X/2X*/
	#define SUBDEVICE_ATI_RAGE_PRO_TURBO		0x10020040 /*Rage Pro Turbo AGP 2X*/
	#define SUBDEVICE_ATI_RAGE_PRO_TURBO2		0x10020044 /*Rage Pro Turbo AGP 2X*/
	#define SUBDEVICE_ATI_RAGE_PRO_AIW		0x10020061 /*Rage Pro AIW AGP 2X*/
	#define SUBDEVICE_ATI_RAGE_PRO_AIW2		0x10020062 /*Rage Pro AIW AGP 2X*/
	#define SUBDEVICE_ATI_RAGE_PRO_AIW3		0x10020063 /*Rage Pro AIW AGP 2X*/
	#define SUBDEVICE_ATI_RAGE_PRO_TURBO3		0x10020080 /*Rage Pro Turbo AGP 2X*/
	#define SUBDEVICE_ATI_RAGE_PRO_TURBO4		0x10020084 /*Rage Pro Turbo AGP 2X*/
	#define SUBDEVICE_ATI_RAGE_PRO_TURBO5		0x10024742 /*Rage Pro Turbo AGP 2X*/
	#define SUBDEVICE_ATI_RAGE_PRO_TURBO6		0x10028001 /*Rage Pro Turbo AGP 2X*/
	#define SUBDEVICE_ATI_RAGE_PRO_TURBO7		0x10280082 /*Rage Pro Turbo AGP 2X*/
	#define SUBDEVICE_ATI_OPTIPLEX_GX1_ONBOARD	0x10284082 /*Optiplex GX1 Onboard Display Adapter*/
	#define SUBDEVICE_ATI_RAGE_PRO_TURBO8		0x10288082 /*Rage Pro Turbo AGP 2X*/
	#define SUBDEVICE_ATI_RAGE_PRO_TURBO9		0x1028c082 /*Rage Pro Turbo AGP 2X*/
	#define SUBDEVICE_ATI_XPERT_98D_AGP		0x80864152 /*Xpert 98D AGP 2X*/
	#define SUBDEVICE_ATI_RAGE_PRO_TURBO10		0x8086464a /*Rage Pro Turbo AGP 2X*/
#define DEVICE_ATI_3D_RAGE_PRO2				0x4744 /*3D Rage Pro AGP 1X*/
	#define SUBDEVICE_ATI_RAGE_PRO_TURBO11		0x10024744 /*Rage Pro Turbo AGP*/
#define DEVICE_ATI_3D_RAGE_PRO3				0x4747 /*3D Rage Pro*/
#define DEVICE_ATI_3D_RAGE_PRO4				0x4749 /*3D Rage Pro*/
	#define SUBDEVICE_ATI_RAGE_PRO_AIW4		0x10020061 /*Rage Pro AIW*/
	#define SUBDEVICE_ATI_RAGE_PRO_AIW5		0x10020062 /*Rage Pro AIW*/
#define DEVICE_ATI_RAGE_XC				0x474c /*Rage XC*/
#define DEVICE_ATI_RAGE_XL_AGP				0x474d /*Rage XL AGP 2X*/
	#define SUBDEVICE_ATI_XPERT_98_RXL		0x10020004 /*Xpert 98 RXL AGP 2X*/
	#define SUBDEVICE_ATI_XPERT_98_RXL2		0x10020008 /*Xpert 98 RXL AGP 2X*/
	#define SUBDEVICE_ATI_RAGE_XL_AGP		0x10020080 /*Rage XL AGP 2X*/
	#define SUBDEVICE_ATI_XPERT_98_AGP		0x10020084 /*Xpert 98 AGP 2X*/
	#define SUBDEVICE_ATI_RAGE_XL_AGP2		0x1002474d /*Rage XL AGP*/
	#define SUBDEVICE_ATI_RAGE_XL_AGP3		0x1033806a /*Rage XL AGP*/
#define DEVICE_ATI_RAGE_XC_AGP				0x474e /*Rage XC AGP*/
	#define SUBDEVICE_ATI_RAGE_XC_AGP		0x1002474e /*Rage XC AGP*/
#define DEVICE_ATI_RAGE_XL				0x474f /*Rage XL*/
	#define SUBDEVICE_ATI_RAGE_XL			0x10020008 /*Rage XL*/
	#define SUBDEVICE_ATI_RAGE_XL2			0x1002474f /*Rage XL*/
#define DEVICE_ATI_3D_RAGE_PRO5				0x4750 /*3D Rage Pro 215GP*/
	#define SUBDEVICE_ATI_RAGE_PRO_TURBO12		0x10020040 /*Rage Pro Turbo*/
	#define SUBDEVICE_ATI_RAGE_PRO_TURBO13		0x10020044 /*Rage Pro Turbo*/
	#define SUBDEVICE_ATI_RAGE_PRO_TURBO14		0x10020080 /*Rage Pro Turbo*/
	#define SUBDEVICE_ATI_RAGE_PRO_TURBO15		0x10020084 /*Rage Pro Turbo*/
	#define SUBDEVICE_ATI_RAGE_PRO_TURBO16		0x10024750 /*Rage Pro Turbo*/
#define DEVICE_ATI_3D_RAGE_PRO6				0x4751 /*3D Rage Pro 215GQ*/
#define DEVICE_ATI_RAGE_XL2				0x4752 /*Rage XL*/
	#define SUBDEVICE_ATI_RAGE_XL3			0x10020008 /*Rage XL*/
	#define SUBDEVICE_ATI_RAGE_XL4			0x10024752 /*Rage XL*/
	#define SUBDEVICE_ATI_RAGE_XL5			0x10028008 /*Rage XL*/
	#define SUBDEVICE_ATI_POWEREDGE_1400		0x102800ce /*PowerEdge 1400*/
	#define SUBDEVICE_ATI_POWEREDGE_2550		0x102800d1 /*PowerEdge 2550*/
	#define SUBDEVICE_ATI_POWEREDGE_2500		0x102800d9 /*PowerEdge 2500*/
	#define SUBDEVICE_ATI_SDS2_MAINBOARD		0x80863411 /*SDS2 Mainboard*/
	#define SUBDEVICE_ATI_S875WP1_E_MAINBOARD	0x80863427 /*S875WP1-E mainboard*/
#define DEVICE_ATI_RAGE_XC2				0x4753 /*Rage XC*/
	#define SUBDEVICE_ATI_RAGE_XC			0x10024753 /*Rage XC*/
#define DEVICE_ATI_3D_RAGE_I_II				0x4754 /*3D Rage I/II 215GT [Mach64 GT]*/
#define DEVICE_ATI_3D_RAGE_II				0x4755 /*3D Rage II+ 215GTB [Mach64 GTB]*/
#define DEVICE_ATI_3D_RAGE_IIC				0x4756 /*3D Rage IIC 215IIC [Mach64 GT IIC]*/
	#define SUBDEVICE_ATI_RAGE_IIC			0x10024756 /*Rage IIC*/
#define DEVICE_ATI_3D_RAGE_IIC2				0x4757 /*3D Rage IIC AGP*/
	#define SUBDEVICE_ATI_RAGE_IIC_AGP		0x10024757 /*Rage IIC AGP*/
	#define SUBDEVICE_ATI_RAGE_3D_IIC		0x10280089 /*Rage 3D IIC*/
	#define SUBDEVICE_ATI_RAGE_3D_IIC2		0x10284082 /*Rage 3D IIC*/
	#define SUBDEVICE_ATI_RAGE_3D_IIC3		0x10288082 /*Rage 3D IIC*/
	#define SUBDEVICE_ATI_RAGE_3D_IIC4		0x1028c082 /*Rage 3D IIC*/
#define DEVICE_ATI_210888GX_MACH64_GX			0x4758 /*210888GX [Mach64 GX]*/
#define DEVICE_ATI_3D_RAGE_IIC3				0x4759 /*3D Rage IIC*/
#define DEVICE_ATI_3D_RAGE_IIC4				0x475a /*3D Rage IIC AGP*/
	#define SUBDEVICE_ATI_RAGE_3D_PRO		0x10020084 /*Rage 3D Pro AGP 2x XPERT 98*/
	#define SUBDEVICE_ATI_RAGE_3D_IIC5		0x10020087 /*Rage 3D IIC*/
	#define SUBDEVICE_ATI_RAGE_IIC_AGP2		0x1002475a /*Rage IIC AGP*/
#define DEVICE_ATI_RADEON_RV250_ID			0x4964 /*Radeon RV250 Id [Radeon 9000]*/
#define DEVICE_ATI_RADEON_RV250_IE			0x4965 /*Radeon RV250 Ie [Radeon 9000]*/
#define DEVICE_ATI_RADEON_RV250_IF			0x4966 /*Radeon RV250 If [Radeon 9000]*/
	#define SUBDEVICE_ATI_RV250_IF_TACHYON		0x10f10002 /*RV250 If [Tachyon G9000 PRO]*/
	#define SUBDEVICE_ATI_RV250_IF_RADEON		0x148c2039 /*RV250 If [Radeon 9000 Pro \"Evil Commando\"]*/
	#define SUBDEVICE_ATI_RV250_IF_RADEON2		0x15099a00 /*RV250 If [Radeon 9000 \"AT009\"]*/
	#define SUBDEVICE_ATI_RV250_IF_3D		0x16810040 /*RV250 If [3D prophet 9000]*/
	#define SUBDEVICE_ATI_RV250_IF_SAPPHIRE		0x174b7176 /*RV250 If [Sapphire Radeon 9000 Pro]*/
	#define SUBDEVICE_ATI_RV250_IF_RADEON3		0x174b7192 /*RV250 If [Radeon 9000 \"Atlantis\"]*/
	#define SUBDEVICE_ATI_RV250_IF_EXCALIBUR	0x17af2005 /*RV250 If [Excalibur Radeon 9000 Pro]*/
	#define SUBDEVICE_ATI_RV250_IF_EXCALIBUR2	0x17af2006 /*RV250 If [Excalibur Radeon 9000]*/
#define DEVICE_ATI_RADEON_RV250_IG			0x4967 /*Radeon RV250 Ig [Radeon 9000]*/
#define DEVICE_ATI_RADEON_RV250_RADEON			0x496e /*Radeon RV250 [Radeon 9000] (Secondary)*/
#define DEVICE_ATI_R420_JH_RADEON			0x4a48 /*R420 JH [Radeon X800]*/
#define DEVICE_ATI_R420_JI_RADEON			0x4a49 /*R420 JI [Radeon X800PRO]*/
#define DEVICE_ATI_R420_JJ_RADEON			0x4a4a /*R420 JJ [Radeon X800SE]*/
#define DEVICE_ATI_R420_JK_RADEON			0x4a4b /*R420 JK [Radeon X800]*/
#define DEVICE_ATI_R420_JL_RADEON			0x4a4c /*R420 JL [Radeon X800]*/
#define DEVICE_ATI_R420_JM_FIREGL			0x4a4d /*R420 JM [FireGL X3]*/
#define DEVICE_ATI_M18_JN_RADEON			0x4a4e /*M18 JN [Radeon Mobility 9800]*/
#define DEVICE_ATI_R420_JP_RADEON			0x4a50 /*R420 JP [Radeon X800XT]*/
#define DEVICE_ATI_R420_X800XT_PE_SECONDARY		0x4a70 /*R420 [X800XT-PE] (Secondary)*/
#define DEVICE_ATI_3D_RAGE_LT				0x4c42 /*3D Rage LT Pro AGP-133*/
	#define SUBDEVICE_ATI_RAGE_LT_PRO		0x0e11b0e7 /*Rage LT Pro (Compaq Presario 5240)*/
	#define SUBDEVICE_ATI_RAGE_3D_LT		0x0e11b0e8 /*Rage 3D LT Pro*/
	#define SUBDEVICE_ATI_3D_RAGE_LT		0x0e11b10e /*3D Rage LT Pro (Compaq Armada 1750)*/
	#define SUBDEVICE_ATI_RAGE_LT_PRO2		0x10020040 /*Rage LT Pro AGP 2X*/
	#define SUBDEVICE_ATI_RAGE_LT_PRO3		0x10020044 /*Rage LT Pro AGP 2X*/
	#define SUBDEVICE_ATI_RAGE_LT_PRO4		0x10024c42 /*Rage LT Pro AGP 2X*/
	#define SUBDEVICE_ATI_RAGE_LT_PRO5		0x10028001 /*Rage LT Pro AGP 2X*/
	#define SUBDEVICE_ATI_RAGE_3D_LT2		0x10280085 /*Rage 3D LT Pro*/
#define DEVICE_ATI_3D_RAGE_LT2				0x4c44 /*3D Rage LT Pro AGP-66*/
#define DEVICE_ATI_RAGE_MOBILITY_M3			0x4c45 /*Rage Mobility M3 AGP*/
#define DEVICE_ATI_RAGE_MOBILITY_M32			0x4c46 /*Rage Mobility M3 AGP 2x*/
	#define SUBDEVICE_ATI_LATITUDE_C600		0x102800b1 /*Latitude C600*/
#define DEVICE_ATI_3D_RAGE_LT_G				0x4c47 /*3D Rage LT-G 215LG*/
#define DEVICE_ATI_3D_RAGE_LT3				0x4c49 /*3D Rage LT Pro*/
	#define SUBDEVICE_ATI_RAGE_LT_PRO6		0x10020004 /*Rage LT Pro*/
	#define SUBDEVICE_ATI_RAGE_LT_PRO7		0x10020040 /*Rage LT Pro*/
	#define SUBDEVICE_ATI_RAGE_LT_PRO8		0x10020044 /*Rage LT Pro*/
	#define SUBDEVICE_ATI_RAGE_LT_PRO9		0x10024c49 /*Rage LT Pro*/
#define DEVICE_ATI_RAGE_MOBILITY_P_M			0x4c4d /*Rage Mobility P/M AGP 2x*/
	#define SUBDEVICE_ATI_ARMADA_M700		0x0e11b111 /*Armada M700*/
	#define SUBDEVICE_ATI_ARMADA_E500		0x0e11b160 /*Armada E500*/
	#define SUBDEVICE_ATI_XPERT_98_AGP2		0x10020084 /*Xpert 98 AGP 2X (Mobility)*/
	#define SUBDEVICE_ATI_THINKPAD_A20M		0x10140154 /*ThinkPad A20m*/
	#define SUBDEVICE_ATI_LATITUDE_CPT		0x102800aa /*Latitude CPt*/
#define DEVICE_ATI_RAGE_MOBILITY_L			0x4c4e /*Rage Mobility L AGP 2x*/
#define DEVICE_ATI_3D_RAGE_LT4				0x4c50 /*3D Rage LT Pro*/
	#define SUBDEVICE_ATI_RAGE_LT_PRO10		0x10024c50 /*Rage LT Pro*/
#define DEVICE_ATI_3D_RAGE_LT5				0x4c51 /*3D Rage LT Pro*/
#define DEVICE_ATI_RAGE_MOBILITY_P_M2			0x4c52 /*Rage Mobility P/M*/
	#define SUBDEVICE_ATI_VERSA_NOTE_VXI		0x10338112 /*Versa Note VXi*/
#define DEVICE_ATI_RAGE_MOBILITY_L2			0x4c53 /*Rage Mobility L*/
#define DEVICE_ATI_264LT_MACH64_LT			0x4c54 /*264LT [Mach64 LT]*/
#define DEVICE_ATI_RADEON_MOBILITY_M7			0x4c57 /*Radeon Mobility M7 LW [Radeon Mobility 7500]*/
	#define SUBDEVICE_ATI_THINKPAD_T30		0x10140517 /*ThinkPad T30*/
	#define SUBDEVICE_ATI_RADEON_MOBILITY_M7	0x102800e6 /*Radeon Mobility M7 LW (Dell Inspiron 8100)*/
	#define SUBDEVICE_ATI_LATITUDE_C640		0x1028012a /*Latitude C640*/
	#define SUBDEVICE_ATI_RADEON_MOBILITY_M72	0x144dc006 /*Radeon Mobility M7 LW in vpr Matrix 170B4*/
#define DEVICE_ATI_RADEON_RV200_LX			0x4c58 /*Radeon RV200 LX [Mobility FireGL 7800 M7]*/
#define DEVICE_ATI_RADEON_MOBILITY_M6			0x4c59 /*Radeon Mobility M6 LY*/
	#define SUBDEVICE_ATI_THINKPAD_A30_A30P_2652_2653	0x10140235 /*ThinkPad A30/A30p (2652/2653)*/
	#define SUBDEVICE_ATI_THINKPAD_X22_X23_X24	0x10140239 /*ThinkPad X22/X23/X24*/
	#define SUBDEVICE_ATI_VAIO_PCG_GR214EP_GR214MP_GR215MP_GR314MP_GR315MP	0x104d80e7 /*VAIO PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP*/
#define DEVICE_ATI_RADEON_MOBILITY_M62			0x4c5a /*Radeon Mobility M6 LZ*/
#define DEVICE_ATI_RADEON_R250_LD			0x4c64 /*Radeon R250 Ld [Radeon Mobility 9000 M9]*/
#define DEVICE_ATI_RADEON_R250_LE			0x4c65 /*Radeon R250 Le [Radeon Mobility 9000 M9]*/
#define DEVICE_ATI_RADEON_R250_LF			0x4c66 /*Radeon R250 Lf [FireGL 9000]*/
#define DEVICE_ATI_RADEON_R250_LG			0x4c67 /*Radeon R250 Lg [Radeon Mobility 9000 M9]*/
#define DEVICE_ATI_RADEON_R250_LN			0x4c6e /*Radeon R250 Ln [Radeon Mobility 9000 M9] [Secondary]*/
#define DEVICE_ATI_RAGE_MOBILITY_M4			0x4d46 /*Rage Mobility M4 AGP*/
#define DEVICE_ATI_RAGE_MOBILITY_M42			0x4d4c /*Rage Mobility M4 AGP*/
#define DEVICE_ATI_RADEON_R300_ND			0x4e44 /*Radeon R300 ND [Radeon 9700 Pro]*/
#define DEVICE_ATI_RADEON_R300_NE			0x4e45 /*Radeon R300 NE [Radeon 9500 Pro]*/
	#define SUBDEVICE_ATI_RADEON_R300_NE		0x10020002 /*Radeon R300 NE [Radeon 9500 Pro]*/
	#define SUBDEVICE_ATI_HERCULES_3D_PROPHET	0x16810002 /*Hercules 3D Prophet 9500 PRO [Radeon 9500 Pro]*/
#define DEVICE_ATI_RV350_NF_RADEON			0x4e46 /*RV350 NF [Radeon 9600]*/
#define DEVICE_ATI_RADEON_R300_NG			0x4e47 /*Radeon R300 NG [FireGL X1]*/
#define DEVICE_ATI_RADEON_R350_RADEON2			0x4e48 /*Radeon R350 [Radeon 9800 Pro]*/
#define DEVICE_ATI_RADEON_R350_RADEON3			0x4e49 /*Radeon R350 [Radeon 9800]*/
#define DEVICE_ATI_RV350_NJ_RADEON			0x4e4a /*RV350 NJ [Radeon 9800 XT]*/
#define DEVICE_ATI_R350_NK_FIRE				0x4e4b /*R350 NK [Fire GL X2]*/
#define DEVICE_ATI_RV350_MOBILITY_RADEON		0x4e50 /*RV350 [Mobility Radeon 9600 M10]*/
	#define SUBDEVICE_ATI_TRAVELMATE_290		0x1025005a /*TravelMate 290*/
	#define SUBDEVICE_ATI_NC6000_LAPTOP		0x103c0890 /*NC6000 laptop*/
	#define SUBDEVICE_ATI_AMILO_M1420W		0x17341055 /*Amilo M1420W*/
#define DEVICE_ATI_M10_NQ_RADEON			0x4e51 /*M10 NQ [Radeon Mobility 9600]*/
#define DEVICE_ATI_RV350_MOBILITY_RADEON2		0x4e52 /*RV350 [Mobility Radeon 9600 M10]*/
#define DEVICE_ATI_M10_NS_RADEON			0x4e53 /*M10 NS [Radeon Mobility 9600]*/
#define DEVICE_ATI_M10_NT_FIREGL			0x4e54 /*M10 NT [FireGL Mobility T2]*/
#define DEVICE_ATI_M11_NV_FIREGL			0x4e56 /*M11 NV [FireGL Mobility T2e]*/
#define DEVICE_ATI_RADEON_R300_RADEON			0x4e64 /*Radeon R300 [Radeon 9700 Pro] (Secondary)*/
#define DEVICE_ATI_RADEON_R300_RADEON2			0x4e65 /*Radeon R300 [Radeon 9500 Pro] (Secondary)*/
	#define SUBDEVICE_ATI_RADEON_R300_NE2		0x10020003 /*Radeon R300 NE [Radeon 9500 Pro]*/
	#define SUBDEVICE_ATI_HERCULES_3D_PROPHET2	0x16810003 /*Hercules 3D Prophet 9500 PRO [Radeon 9500 Pro] (Secondary)*/
#define DEVICE_ATI_RV350_NF_RADEON2			0x4e66 /*RV350 NF [Radeon 9600] (Secondary)*/
#define DEVICE_ATI_RADEON_R300_FIREGL			0x4e67 /*Radeon R300 [FireGL X1] (Secondary)*/
#define DEVICE_ATI_RADEON_R350_RADEON4			0x4e68 /*Radeon R350 [Radeon 9800 Pro] (Secondary)*/
#define DEVICE_ATI_RADEON_R350_RADEON5			0x4e69 /*Radeon R350 [Radeon 9800] (Secondary)*/
#define DEVICE_ATI_RV350_NJ_RADEON2			0x4e6a /*RV350 NJ [Radeon 9800 XT] (Secondary)*/
#define DEVICE_ATI_RAGE_128_PA_PRO			0x5041 /*Rage 128 PA/PRO*/
#define DEVICE_ATI_RAGE_128_PB_PRO			0x5042 /*Rage 128 PB/PRO AGP 2x*/
#define DEVICE_ATI_RAGE_128_PC_PRO			0x5043 /*Rage 128 PC/PRO AGP 4x*/
#define DEVICE_ATI_RAGE_128_PD_PRO			0x5044 /*Rage 128 PD/PRO TMDS*/
	#define SUBDEVICE_ATI_RAGE_128_AIW		0x10020028 /*Rage 128 AIW*/
	#define SUBDEVICE_ATI_RAGE_128_AIW2		0x10020029 /*Rage 128 AIW*/
#define DEVICE_ATI_RAGE_128_PE_PRO			0x5045 /*Rage 128 PE/PRO AGP 2x TMDS*/
#define DEVICE_ATI_RAGE_128_PF_PRO			0x5046 /*Rage 128 PF/PRO AGP 4x TMDS*/
	#define SUBDEVICE_ATI_RAGE_FURY_PRO		0x10020004 /*Rage Fury Pro*/
	#define SUBDEVICE_ATI_RAGE_FURY_PRO_XPERT	0x10020008 /*Rage Fury Pro/Xpert 2000 Pro*/
	#define SUBDEVICE_ATI_RAGE_FURY_PRO2		0x10020014 /*Rage Fury Pro*/
	#define SUBDEVICE_ATI_RAGE_FURY_PRO_XPERT2	0x10020018 /*Rage Fury Pro/Xpert 2000 Pro*/
	#define SUBDEVICE_ATI_RAGE_128_PRO		0x10020028 /*Rage 128 Pro AIW AGP*/
	#define SUBDEVICE_ATI_RAGE_128_PRO2		0x1002002a /*Rage 128 Pro AIW AGP*/
	#define SUBDEVICE_ATI_RAGE_FURY_PRO3		0x10020048 /*Rage Fury Pro*/
	#define SUBDEVICE_ATI_RAGE_FURY_MAXX		0x10022000 /*Rage Fury MAXX AGP 4x (TMDS) (VGA device)*/
	#define SUBDEVICE_ATI_RAGE_FURY_MAXX2		0x10022001 /*Rage Fury MAXX AGP 4x (TMDS) (Extra device?!)*/
#define DEVICE_ATI_RAGE_128_PG_PRO			0x5047 /*Rage 128 PG/PRO*/
#define DEVICE_ATI_RAGE_128_PH_PRO			0x5048 /*Rage 128 PH/PRO AGP 2x*/
#define DEVICE_ATI_RAGE_128_PI_PRO			0x5049 /*Rage 128 PI/PRO AGP 4x*/
#define DEVICE_ATI_RAGE_128_PJ_PRO			0x504a /*Rage 128 PJ/PRO TMDS*/
#define DEVICE_ATI_RAGE_128_PK_PRO			0x504b /*Rage 128 PK/PRO AGP 2x TMDS*/
#define DEVICE_ATI_RAGE_128_PL_PRO			0x504c /*Rage 128 PL/PRO AGP 4x TMDS*/
#define DEVICE_ATI_RAGE_128_PM_PRO			0x504d /*Rage 128 PM/PRO*/
#define DEVICE_ATI_RAGE_128_PN_PRO			0x504e /*Rage 128 PN/PRO AGP 2x*/
#define DEVICE_ATI_RAGE_128_PO_PRO			0x504f /*Rage 128 PO/PRO AGP 4x*/
#define DEVICE_ATI_RAGE_128_PP_PRO			0x5050 /*Rage 128 PP/PRO TMDS [Xpert 128]*/
	#define SUBDEVICE_ATI_XPERT_128			0x10020008 /*Xpert 128*/
#define DEVICE_ATI_RAGE_128_PQ_PRO			0x5051 /*Rage 128 PQ/PRO AGP 2x TMDS*/
#define DEVICE_ATI_RAGE_128_PR_PRO			0x5052 /*Rage 128 PR/PRO AGP 4x TMDS*/
#define DEVICE_ATI_RAGE_128_PS_PRO			0x5053 /*Rage 128 PS/PRO*/
#define DEVICE_ATI_RAGE_128_PT_PRO			0x5054 /*Rage 128 PT/PRO AGP 2x*/
#define DEVICE_ATI_RAGE_128_PU_PRO			0x5055 /*Rage 128 PU/PRO AGP 4x*/
#define DEVICE_ATI_RAGE_128_PV_PRO			0x5056 /*Rage 128 PV/PRO TMDS*/
#define DEVICE_ATI_RAGE_128_PW_PRO			0x5057 /*Rage 128 PW/PRO AGP 2x TMDS*/
#define DEVICE_ATI_RAGE_128_PX_PRO			0x5058 /*Rage 128 PX/PRO AGP 4x TMDS*/
#define DEVICE_ATI_RADEON_R100_QD			0x5144 /*Radeon R100 QD [Radeon 7200]*/
	#define SUBDEVICE_ATI_RADEON_7000_RADEON_VE	0x10020008 /*Radeon 7000/Radeon VE*/
	#define SUBDEVICE_ATI_RADEON_7000_RADEON	0x10020009 /*Radeon 7000/Radeon*/
	#define SUBDEVICE_ATI_RADEON_7000_RADEON2	0x1002000a /*Radeon 7000/Radeon*/
	#define SUBDEVICE_ATI_RADEON_7000_RADEON3	0x1002001a /*Radeon 7000/Radeon*/
	#define SUBDEVICE_ATI_RADEON_AIW		0x10020029 /*Radeon AIW*/
	#define SUBDEVICE_ATI_RADEON_7000_RADEON4	0x10020038 /*Radeon 7000/Radeon*/
	#define SUBDEVICE_ATI_RADEON_7000_RADEON5	0x10020039 /*Radeon 7000/Radeon*/
	#define SUBDEVICE_ATI_RADEON_7000_RADEON6	0x1002008a /*Radeon 7000/Radeon*/
	#define SUBDEVICE_ATI_RADEON_7000_RADEON7	0x100200ba /*Radeon 7000/Radeon*/
	#define SUBDEVICE_ATI_RADEON_7000_RADEON8	0x10020139 /*Radeon 7000/Radeon*/
	#define SUBDEVICE_ATI_RADEON_7000_RADEON9	0x1002028a /*Radeon 7000/Radeon*/
	#define SUBDEVICE_ATI_RADEON_AIW2		0x100202aa /*Radeon AIW*/
	#define SUBDEVICE_ATI_RADEON_7000_RADEON10	0x1002053a /*Radeon 7000/Radeon*/
#define DEVICE_ATI_RADEON_R100_QE			0x5145 /*Radeon R100 QE*/
#define DEVICE_ATI_RADEON_R100_QF			0x5146 /*Radeon R100 QF*/
#define DEVICE_ATI_RADEON_R100_QG			0x5147 /*Radeon R100 QG*/
#define DEVICE_ATI_RADEON_R200_QH			0x5148 /*Radeon R200 QH [Radeon 8500]*/
	#define SUBDEVICE_ATI_FIREGL_8800_64MB		0x1002010a /*FireGL 8800 64Mb*/
	#define SUBDEVICE_ATI_FIREGL_8800_128MB		0x10020152 /*FireGL 8800 128Mb*/
	#define SUBDEVICE_ATI_FIREGL_8700_32MB		0x10020162 /*FireGL 8700 32Mb*/
	#define SUBDEVICE_ATI_FIREGL_8700_64MB		0x10020172 /*FireGL 8700 64Mb*/
#define DEVICE_ATI_RADEON_R200_QI			0x5149 /*Radeon R200 QI*/
#define DEVICE_ATI_RADEON_R200_QJ			0x514a /*Radeon R200 QJ*/
#define DEVICE_ATI_RADEON_R200_QK			0x514b /*Radeon R200 QK*/
#define DEVICE_ATI_RADEON_R200_QL			0x514c /*Radeon R200 QL [Radeon 8500 LE]*/
	#define SUBDEVICE_ATI_RADEON_R200_QL		0x1002003a /*Radeon R200 QL [Radeon 8500 LE]*/
	#define SUBDEVICE_ATI_RADEON_8500		0x1002013a /*Radeon 8500*/
	#define SUBDEVICE_ATI_R200_QL_RADEON		0x148c2026 /*R200 QL [Radeon 8500 Evil Master II Multi Display Edition]*/
	#define SUBDEVICE_ATI_RADEON_8500_3D		0x16810010 /*Radeon 8500 [3D Prophet 8500 128Mb]*/
	#define SUBDEVICE_ATI_RADEON_R200_QL2		0x174b7149 /*Radeon R200 QL [Sapphire Radeon 8500 LE]*/
#define DEVICE_ATI_RADEON_R200_QM			0x514d /*Radeon R200 QM [Radeon 9100]*/
#define DEVICE_ATI_RADEON_R200_QN			0x514e /*Radeon R200 QN [Radeon 8500LE]*/
#define DEVICE_ATI_RADEON_R200_QO			0x514f /*Radeon R200 QO [Radeon 8500LE]*/
#define DEVICE_ATI_R200_QT_RADEON			0x5154 /*R200 QT [Radeon 8500]*/
#define DEVICE_ATI_R200_QU_RADEON			0x5155 /*R200 QU [Radeon 9100]*/
#define DEVICE_ATI_RADEON_RV200_QW			0x5157 /*Radeon RV200 QW [Radeon 7500]*/
	#define SUBDEVICE_ATI_RADEON_7500		0x1002013a /*Radeon 7500*/
	#define SUBDEVICE_ATI_DELL_OPTIPLEX_GX260	0x1002103a /*Dell Optiplex GX260*/
	#define SUBDEVICE_ATI_RV200_QW_RADEON		0x14584000 /*RV200 QW [RADEON 7500 PRO MAYA AR]*/
	#define SUBDEVICE_ATI_RV200_QW_RADEON2		0x148c2024 /*RV200 QW [Radeon 7500LE Dual Display]*/
	#define SUBDEVICE_ATI_RV200_QW_RADEON3		0x148c2025 /*RV200 QW [Radeon 7500 Evil Master Multi Display Edition]*/
	#define SUBDEVICE_ATI_RV200_QW_RADEON4		0x148c2036 /*RV200 QW [Radeon 7500 PCI Dual Display]*/
	#define SUBDEVICE_ATI_RV200_QW_RADEON5		0x174b7146 /*RV200 QW [Radeon 7500 LE]*/
	#define SUBDEVICE_ATI_RV200_QW_SAPPHIRE		0x174b7147 /*RV200 QW [Sapphire Radeon 7500LE]*/
	#define SUBDEVICE_ATI_RADEON_RV200_QW		0x174b7161 /*Radeon RV200 QW [Radeon 7500 LE]*/
	#define SUBDEVICE_ATI_RV200_QW_EXCALIBUR	0x17af0202 /*RV200 QW [Excalibur Radeon 7500LE]*/
#define DEVICE_ATI_RADEON_RV200_QX			0x5158 /*Radeon RV200 QX [Radeon 7500]*/
#define DEVICE_ATI_RADEON_RV100_QY			0x5159 /*Radeon RV100 QY [Radeon 7000/VE]*/
	#define SUBDEVICE_ATI_RADEON_7000_RADEON_VE2	0x1002000a /*Radeon 7000/Radeon VE*/
	#define SUBDEVICE_ATI_RADEON_7000		0x1002000b /*Radeon 7000*/
	#define SUBDEVICE_ATI_RADEON_7000_RADEON_VE3	0x10020038 /*Radeon 7000/Radeon VE*/
	#define SUBDEVICE_ATI_RADEON_7000_RADEON_VE4	0x1002003a /*Radeon 7000/Radeon VE*/
	#define SUBDEVICE_ATI_RADEON_7000_RADEON_VE5	0x100200ba /*Radeon 7000/Radeon VE*/
	#define SUBDEVICE_ATI_RADEON_7000_RADEON_VE6	0x1002013a /*Radeon 7000/Radeon VE*/
	#define SUBDEVICE_ATI_RV100_QY_RADEON		0x14584002 /*RV100 QY [RADEON 7000 PRO MAYA AV Series]*/
	#define SUBDEVICE_ATI_RV100_QY_RADEON2		0x148c2003 /*RV100 QY [Radeon 7000 Multi-Display Edition]*/
	#define SUBDEVICE_ATI_RV100_QY_RADEON3		0x148c2023 /*RV100 QY [Radeon 7000 Evil Master Multi-Display]*/
	#define SUBDEVICE_ATI_RV100_QY_SAPPHIRE		0x174b7112 /*RV100 QY [Sapphire Radeon VE 7000]*/
	#define SUBDEVICE_ATI_SAPPHIRE_RADEON_VE	0x174b7c28 /*Sapphire Radeon VE 7000 DDR*/
	#define SUBDEVICE_ATI_RV100_QY_EXCALIBUR	0x17870202 /*RV100 QY [Excalibur Radeon 7000]*/
#define DEVICE_ATI_RADEON_RV100_QZ			0x515a /*Radeon RV100 QZ [Radeon 7000/VE]*/
#define DEVICE_ATI_RADEON_R200_QH2			0x5168 /*Radeon R200 Qh*/
#define DEVICE_ATI_RADEON_R200_QI2			0x5169 /*Radeon R200 Qi*/
#define DEVICE_ATI_RADEON_R200_QJ2			0x516a /*Radeon R200 Qj*/
#define DEVICE_ATI_RADEON_R200_QK2			0x516b /*Radeon R200 Qk*/
#define DEVICE_ATI_RADEON_R200_QL2			0x516c /*Radeon R200 Ql*/
#define DEVICE_ATI_RAGE_128_RE_SG			0x5245 /*Rage 128 RE/SG*/
	#define SUBDEVICE_ATI_XPERT_1282		0x10020008 /*Xpert 128*/
	#define SUBDEVICE_ATI_RAGE_128_AIW3		0x10020028 /*Rage 128 AIW*/
	#define SUBDEVICE_ATI_RAGE_128_AIW4		0x10020029 /*Rage 128 AIW*/
	#define SUBDEVICE_ATI_RAGE_128_AIW5		0x10020068 /*Rage 128 AIW*/
#define DEVICE_ATI_RAGE_128_RF_SG			0x5246 /*Rage 128 RF/SG AGP*/
	#define SUBDEVICE_ATI_MAGNUM_XPERT_128_XPERT_99	0x10020004 /*Magnum/Xpert 128/Xpert 99*/
	#define SUBDEVICE_ATI_MAGNUM_XPERT128_X99_XPERT2000	0x10020008 /*Magnum/Xpert128/X99/Xpert2000*/
	#define SUBDEVICE_ATI_RAGE_128_AIW6		0x10020028 /*Rage 128 AIW AGP*/
	#define SUBDEVICE_ATI_RAGE_FURY_XPERT_128_XPERT	0x10020044 /*Rage Fury/Xpert 128/Xpert 2000*/
	#define SUBDEVICE_ATI_RAGE_128_AIW7		0x10020068 /*Rage 128 AIW AGP*/
	#define SUBDEVICE_ATI_RAGE_FURY			0x10020448 /*Rage Fury*/
#define DEVICE_ATI_RAGE_128_RG				0x5247 /*Rage 128 RG*/
#define DEVICE_ATI_RAGE_128_RK_VR			0x524b /*Rage 128 RK/VR*/
#define DEVICE_ATI_RAGE_128_RL_VR			0x524c /*Rage 128 RL/VR AGP*/
	#define SUBDEVICE_ATI_XPERT_99_XPERT_2000	0x10020008 /*Xpert 99/Xpert 2000*/
	#define SUBDEVICE_ATI_XPERT_99			0x10020088 /*Xpert 99*/
#define DEVICE_ATI_RAGE_128_SE_4X			0x5345 /*Rage 128 SE/4x*/
#define DEVICE_ATI_RAGE_128_SF_4X			0x5346 /*Rage 128 SF/4x AGP 2x*/
	#define SUBDEVICE_ATI_RAGE_128_16MB		0x10020048 /*RAGE 128 16MB VGA TVOUT AMC PAL*/
#define DEVICE_ATI_RAGE_128_SG_4X			0x5347 /*Rage 128 SG/4x AGP 4x*/
#define DEVICE_ATI_RAGE_128_SH				0x5348 /*Rage 128 SH*/
#define DEVICE_ATI_RAGE_128_SK_4X			0x534b /*Rage 128 SK/4x*/
#define DEVICE_ATI_RAGE_128_SL_4X			0x534c /*Rage 128 SL/4x AGP 2x*/
#define DEVICE_ATI_RAGE_128_SM_4X			0x534d /*Rage 128 SM/4x AGP 4x*/
	#define SUBDEVICE_ATI_XPERT_99_XPERT_20002	0x10020008 /*Xpert 99/Xpert 2000*/
	#define SUBDEVICE_ATI_XPERT_2000		0x10020018 /*Xpert 2000*/
#define DEVICE_ATI_RAGE_128_4X				0x534e /*Rage 128 4x*/
#define DEVICE_ATI_MACH_64_VT				0x5354 /*Mach 64 VT*/
	#define SUBDEVICE_ATI_MACH_64_REFERENCE		0x10025654 /*Mach 64 reference*/
#define DEVICE_ATI_RAGE_128_PRO				0x5446 /*Rage 128 Pro Ultra TF*/
	#define SUBDEVICE_ATI_RAGE_FURY_PRO4		0x10020004 /*Rage Fury Pro*/
	#define SUBDEVICE_ATI_RAGE_FURY_PRO_XPERT3	0x10020008 /*Rage Fury Pro/Xpert 2000 Pro*/
	#define SUBDEVICE_ATI_RAGE_FURY_PRO_XPERT4	0x10020018 /*Rage Fury Pro/Xpert 2000 Pro*/
	#define SUBDEVICE_ATI_RAGE_128_AIW8		0x10020028 /*Rage 128 AIW Pro AGP*/
	#define SUBDEVICE_ATI_RAGE_128_AIW9		0x10020029 /*Rage 128 AIW*/
	#define SUBDEVICE_ATI_RAGE_128_AIW10		0x1002002a /*Rage 128 AIW Pro AGP*/
	#define SUBDEVICE_ATI_RAGE_128_AIW11		0x1002002b /*Rage 128 AIW*/
	#define SUBDEVICE_ATI_XPERT_2000_PRO		0x10020048 /*Xpert 2000 Pro*/
#define DEVICE_ATI_RAGE_128_PRO2			0x544c /*Rage 128 Pro Ultra TL*/
#define DEVICE_ATI_RAGE_128_PRO3			0x5452 /*Rage 128 Pro Ultra TR*/
	#define SUBDEVICE_ATI_RAGE_128_PRO3		0x1002001c /*Rage 128 Pro 4XL*/
	#define SUBDEVICE_ATI_RAGE_128_PRO4		0x103c1279 /*Rage 128 Pro 4XL*/
#define DEVICE_ATI_RAGE_128_PRO4			0x5453 /*Rage 128 Pro Ultra TS*/
#define DEVICE_ATI_RAGE_128_PRO5			0x5454 /*Rage 128 Pro Ultra TT*/
#define DEVICE_ATI_RAGE_128_PRO6			0x5455 /*Rage 128 Pro Ultra TU*/
#define DEVICE_ATI_M22_RADEON_MOBILITY			0x5460 /*M22 [Radeon Mobility M300]*/
#define DEVICE_ATI_M22_FIREGL_GL			0x5464 /*M22 [FireGL GL]*/
#define DEVICE_ATI_R423_UH_RADEON			0x5548 /*R423 UH [Radeon X800 (PCIE)]*/
#define DEVICE_ATI_R423_UI_RADEON			0x5549 /*R423 UI [Radeon X800PRO (PCIE)]*/
#define DEVICE_ATI_R423_UJ_RADEON			0x554a /*R423 UJ [Radeon X800LE (PCIE)]*/
#define DEVICE_ATI_R423_UK_RADEON			0x554b /*R423 UK [Radeon X800SE (PCIE)]*/
#define DEVICE_ATI_R423_UQ_FIREGL			0x5551 /*R423 UQ [FireGL V7200 (PCIE)]*/
#define DEVICE_ATI_R423_UR_FIREGL			0x5552 /*R423 UR [FireGL V5100 (PCIE)]*/
#define DEVICE_ATI_R423_UT_FIREGL			0x5554 /*R423 UT [FireGL V7100 (PCIE)]*/
#define DEVICE_ATI_RADEON_R423_UK			0x556b /*Radeon R423 UK (PCIE) [X800 SE] (Secondary)*/
#define DEVICE_ATI_264VT_MACH64_VT			0x5654 /*264VT [Mach64 VT]*/
	#define SUBDEVICE_ATI_MACH64VT_REFERENCE	0x10025654 /*Mach64VT Reference*/
#define DEVICE_ATI_264VT3_MACH64_VT3			0x5655 /*264VT3 [Mach64 VT3]*/
#define DEVICE_ATI_264VT4_MACH64_VT4			0x5656 /*264VT4 [Mach64 VT4]*/
#define DEVICE_ATI_RS300_HOST_BRIDGE			0x5830 /*RS300 Host Bridge*/
#define DEVICE_ATI_RS300_HOST_BRIDGE2			0x5831 /*RS300 Host Bridge*/
#define DEVICE_ATI_RS300_HOST_BRIDGE3			0x5832 /*RS300 Host Bridge*/
#define DEVICE_ATI_RADEON_9100_IGP			0x5833 /*Radeon 9100 IGP Host Bridge*/
#define DEVICE_ATI_RADEON_9100_IGP2			0x5834 /*Radeon 9100 IGP*/
#define DEVICE_ATI_RS300M_AGP_RADEON			0x5835 /*RS300M AGP [Radeon Mobility 9100IGP]*/
#define DEVICE_ATI_RADEON_9100_IGP3			0x5838 /*Radeon 9100 IGP AGP Bridge*/
#define DEVICE_ATI_RV280_RADEON_9200			0x5941 /*RV280 [Radeon 9200] (Secondary)*/
	#define SUBDEVICE_ATI_SAPPHIRE_RADEON_9200	0x174b7c12 /*Sapphire Radeon 9200*/
	#define SUBDEVICE_ATI_EXCALIBUR_RADEON_9200	0x17af200d /*Excalibur Radeon 9200*/
	#define SUBDEVICE_ATI_GEXCUBE_GC_R9200_C3_SECONDARY	0x18bc0050 /*GeXcube GC-R9200-C3 (Secondary)*/
#define DEVICE_ATI_RV280_RADEON_92002			0x5944 /*RV280 [Radeon 9200 SE (PCI)]*/
#define DEVICE_ATI_RV280_RADEON_92003			0x5960 /*RV280 [Radeon 9200 PRO]*/
#define DEVICE_ATI_RV280_RADEON_92004			0x5961 /*RV280 [Radeon 9200]*/
	#define SUBDEVICE_ATI_ALL_IN_WONDER_9200_SERIES	0x10022f72 /*All-in-Wonder 9200 Series*/
	#define SUBDEVICE_ATI_YUAN_SMARTVGA_RADEON	0x12ab5961 /*YUAN SMARTVGA Radeon 9200*/
	#define SUBDEVICE_ATI_GIGABYTE_RADEON_9200	0x14584018 /*Gigabyte Radeon 9200*/
	#define SUBDEVICE_ATI_SAPPHIRE_RADEON_92002	0x174b7c13 /*Sapphire Radeon 9200*/
	#define SUBDEVICE_ATI_EXCALIBUR_RADEON_92002	0x17af200c /*Excalibur Radeon 9200*/
	#define SUBDEVICE_ATI_RADEON_9200_GAME		0x18bc0050 /*Radeon 9200 Game Buster*/
	#define SUBDEVICE_ATI_GEXCUBE_GC_R9200_C3	0x18bc0051 /*GeXcube GC-R9200-C3*/
	#define SUBDEVICE_ATI_RADEON_9200_GAME2		0x18bc0053 /*Radeon 9200 Game Buster VIVO*/
#define DEVICE_ATI_RV280_RADEON_92005			0x5962 /*RV280 [Radeon 9200]*/
#define DEVICE_ATI_RV280_RADEON_92006			0x5964 /*RV280 [Radeon 9200 SE]*/
	#define SUBDEVICE_ATI_ASUS_RADEON_9200		0x1043c006 /*ASUS Radeon 9200 SE / TD / 128M*/
	#define SUBDEVICE_ATI_RADEON_9200_SE		0x14584018 /*Radeon 9200 SE*/
	#define SUBDEVICE_ATI_CN_AG92E			0x148c2073 /*CN-AG92E*/
	#define SUBDEVICE_ATI_SAPPHIRE_RADEON_92003	0x174b7c13 /*Sapphire Radeon 9200 SE*/
	#define SUBDEVICE_ATI_EXCALIBUR_9200SE_VIVO	0x17875964 /*Excalibur 9200SE VIVO 128M*/
	#define SUBDEVICE_ATI_RADEON_9200_SE2		0x17af2012 /*Radeon 9200 SE Excalibur*/
	#define SUBDEVICE_ATI_SAPPHIRE_RADEON_92004	0x18bc0170 /*Sapphire Radeon 9200 SE 128MB Game Buster*/
	#define SUBDEVICE_ATI_GC_R9200L_SE_C3H_RADEON_9200	0x18bc0173 /*GC-R9200L(SE)-C3H [Radeon 9200 Game Buster]*/
#define DEVICE_ATI_RV370_5B60_RADEON			0x5b60 /*RV370 5B60 [Radeon X300 (PCIE)]*/
	#define SUBDEVICE_ATI_EAX300SE			0x1043002a /*EAX300SE*/
#define DEVICE_ATI_RV370_5B62_RADEON			0x5b62 /*RV370 5B62 [Radeon X600 (PCIE)]*/
#define DEVICE_ATI_RV370_5B64_FIREGL			0x5b64 /*RV370 5B64 [FireGL V3100 (PCIE)]*/
#define DEVICE_ATI_RV370_5B65_FIREGL			0x5b65 /*RV370 5B65 [FireGL D1100 (PCIE)]*/
#define DEVICE_ATI_M9_5C61_RADEON			0x5c61 /*M9+ 5C61 [Radeon Mobility 9200 (AGP)]*/
#define DEVICE_ATI_M9_5C63_RADEON			0x5c63 /*M9+ 5C63 [Radeon Mobility 9200 (AGP)]*/
#define DEVICE_ATI_RV280_RADEON_92007			0x5d44 /*RV280 [Radeon 9200 SE] (Secondary)*/
	#define SUBDEVICE_ATI_RADEON_9200_SE3		0x14584019 /*Radeon 9200 SE (Secondary)*/
	#define SUBDEVICE_ATI_SAPPHIRE_RADEON_92005	0x174b7c12 /*Sapphire Radeon 9200 SE (Secondary)*/
	#define SUBDEVICE_ATI_EXCALIBUR_9200SE_VIVO2	0x17875965 /*Excalibur 9200SE VIVO 128M (Secondary)*/
	#define SUBDEVICE_ATI_RADEON_9200_SE4		0x17af2013 /*Radeon 9200 SE Excalibur (Secondary)*/
	#define SUBDEVICE_ATI_RADEON_9200_SE5		0x18bc0171 /*Radeon 9200 SE 128MB Game Buster (Secondary)*/
	#define SUBDEVICE_ATI_GC_R9200L_SE_C3H_RADEON_92002	0x18bc0172 /*GC-R9200L(SE)-C3H [Radeon 9200 Game Buster]*/
#define DEVICE_ATI_R423_5F57_RADEON			0x5d57 /*R423 5F57 [Radeon X800XT (PCIE)]*/
#define DEVICE_ATI_PCI_BRIDGE_IGP			0x700f /*PCI Bridge [IGP 320M]*/
#define DEVICE_ATI_PCI_BRIDGE_IGP2			0x7010 /*PCI Bridge [IGP 340M]*/
#define DEVICE_ATI_RADEON_9100_PRO			0x7834 /*Radeon 9100 PRO IGP*/
#define DEVICE_ATI_RADEON_MOBILITY_9200			0x7835 /*Radeon Mobility 9200 IGP*/
#define DEVICE_ATI_RV350_AQ_RADEON3			0x7c37 /*RV350 AQ [Radeon 9600 SE]*/
#define DEVICE_ATI_AGP_BRIDGE_IGP			0xcab0 /*AGP Bridge [IGP 320M]*/
#define DEVICE_ATI_RS200_RS200M_AGP_BRIDGE		0xcab2 /*RS200/RS200M AGP Bridge [IGP 340M]*/
#define DEVICE_ATI_RS200_RS200M_AGP_BRIDGE2		0xcbb2 /*RS200/RS200M AGP Bridge [IGP 340M]*/
/* Vendor: 1003: ULSI Systems */
#define DEVICE_ULSI_US201				0x0201 /*US201*/
/* Vendor: 1004: VLSI Technology Inc */
#define DEVICE_VLSI_82C592_FC1				0x0005 /*82C592-FC1*/
#define DEVICE_VLSI_82C593_FC1				0x0006 /*82C593-FC1*/
#define DEVICE_VLSI_82C594_AFC2				0x0007 /*82C594-AFC2*/
#define DEVICE_VLSI_82C596_7_WILDCAT			0x0008 /*82C596/7 [Wildcat]*/
#define DEVICE_VLSI_82C597_AFC2				0x0009 /*82C597-AFC2*/
#define DEVICE_VLSI_82C541_LYNX				0x000c /*82C541 [Lynx]*/
#define DEVICE_VLSI_82C543_LYNX				0x000d /*82C543 [Lynx]*/
#define DEVICE_VLSI_82C532				0x0101 /*82C532*/
#define DEVICE_VLSI_82C534_EAGLE			0x0102 /*82C534 [Eagle]*/
#define DEVICE_VLSI_82C538				0x0103 /*82C538*/
#define DEVICE_VLSI_82C535				0x0104 /*82C535*/
#define DEVICE_VLSI_82C147				0x0105 /*82C147*/
#define DEVICE_VLSI_82C975				0x0200 /*82C975*/
#define DEVICE_VLSI_82C925				0x0280 /*82C925*/
#define DEVICE_VLSI_QSOUND_THUNDERBIRD_PCI		0x0304 /*QSound ThunderBird PCI Audio*/
	#define SUBDEVICE_VLSI_QSOUND_THUNDERBIRD_PCI	0x10040304 /*QSound ThunderBird PCI Audio*/
	#define SUBDEVICE_VLSI_DSP368_AUDIO		0x122d1206 /*DSP368 Audio*/
	#define SUBDEVICE_VLSI_XWAVE_THUNDER_3D		0x14835020 /*XWave Thunder 3D Audio*/
#define DEVICE_VLSI_QSOUND_THUNDERBIRD_PCI2		0x0305 /*QSound ThunderBird PCI Audio Gameport*/
	#define SUBDEVICE_VLSI_QSOUND_THUNDERBIRD_PCI2	0x10040305 /*QSound ThunderBird PCI Audio Gameport*/
	#define SUBDEVICE_VLSI_DSP368_AUDIO_GAMEPORT	0x122d1207 /*DSP368 Audio Gameport*/
	#define SUBDEVICE_VLSI_XWAVE_THUNDER_3D2	0x14835021 /*XWave Thunder 3D Audio Gameport*/
#define DEVICE_VLSI_QSOUND_THUNDERBIRD_PCI3		0x0306 /*QSound ThunderBird PCI Audio Support Registers*/
	#define SUBDEVICE_VLSI_QSOUND_THUNDERBIRD_PCI3	0x10040306 /*QSound ThunderBird PCI Audio Support Registers*/
	#define SUBDEVICE_VLSI_DSP368_AUDIO_SUPPORT	0x122d1208 /*DSP368 Audio Support Registers*/
	#define SUBDEVICE_VLSI_XWAVE_THUNDER_3D3	0x14835022 /*XWave Thunder 3D Audio Support Registers*/
#define DEVICE_VLSI_THUNDERBIRD				0x0307 /*Thunderbird*/
#define DEVICE_VLSI_THUNDERBIRD2			0x0308 /*Thunderbird*/
#define DEVICE_VLSI_VAS96011_GOLDEN_GATE		0x0702 /*VAS96011 [Golden Gate II]*/
#define DEVICE_VLSI_TOLLGATE				0x0703 /*Tollgate*/
/* Vendor: 1005: Avance Logic Inc. [ALI] */
#define DEVICE_AVANCE_ALG2032_2064			0x2064 /*ALG2032/2064*/
#define DEVICE_AVANCE_ALG2364A				0x2128 /*ALG2364A*/
#define DEVICE_AVANCE_ALG2301				0x2301 /*ALG2301*/
#define DEVICE_AVANCE_ALG2302				0x2302 /*ALG2302*/
#define DEVICE_AVANCE_ALG2364				0x2364 /*ALG2364*/
#define DEVICE_AVANCE_ALG2364A2				0x2464 /*ALG2364A*/
#define DEVICE_AVANCE_ALG2564A_25128A			0x2501 /*ALG2564A/25128A*/
/* Vendor: 1006: Reply Group */
/* Vendor: 1007: NetFrame Systems Inc */
/* Vendor: 1008: Epson */
/* Vendor: 100a: Phoenix Technologies */
/* Vendor: 100b: National Semiconductor Corporation */
#define DEVICE_NATIONAL_DP83810				0x0001 /*DP83810*/
#define DEVICE_NATIONAL_87415_87560_IDE			0x0002 /*87415/87560 IDE*/
#define DEVICE_NATIONAL_87560_LEGACY_I_O		0x000e /*87560 Legacy I/O*/
#define DEVICE_NATIONAL_FIREWIRE_CONTROLLER		0x000f /*FireWire Controller*/
#define DEVICE_NATIONAL_NS87560_NATIONAL_PCI		0x0011 /*NS87560 National PCI System I/O*/
#define DEVICE_NATIONAL_USB_CONTROLLER			0x0012 /*USB Controller*/
#define DEVICE_NATIONAL_DP83815_MACPHYTER_ETHERNET	0x0020 /*DP83815 (MacPhyter) Ethernet Controller*/
	#define SUBDEVICE_NATIONAL_PAVILION_ZE4400_BUILTIN	0x103c0024 /*Pavilion ze4400 builtin Network*/
	#define SUBDEVICE_NATIONAL_FA311_FA312		0x1385f311 /*FA311 / FA312 (FA311 with WoL HW)*/
#define DEVICE_NATIONAL_DP83820_10_100_1000_ETHERNET	0x0022 /*DP83820 10/100/1000 Ethernet Controller*/
#define DEVICE_NATIONAL_CS5535_HOST_BRIDGE		0x0028 /*CS5535 Host bridge*/
#define DEVICE_NATIONAL_CS5535_ISA_BRIDGE		0x002b /*CS5535 ISA bridge*/
#define DEVICE_NATIONAL_CS5535_IDE			0x002d /*CS5535 IDE*/
#define DEVICE_NATIONAL_CS5535_AUDIO			0x002e /*CS5535 Audio*/
#define DEVICE_NATIONAL_CS5535_USB			0x002f /*CS5535 USB*/
#define DEVICE_NATIONAL_CS5535_VIDEO			0x0030 /*CS5535 Video*/
#define DEVICE_NATIONAL_DP83065_SATURN_10_100_1000	0x0035 /*DP83065 [Saturn] 10/100/1000 Ethernet Controller*/
#define DEVICE_NATIONAL_SCX200_BRIDGE			0x0500 /*SCx200 Bridge*/
#define DEVICE_NATIONAL_SCX200_SMI			0x0501 /*SCx200 SMI*/
#define DEVICE_NATIONAL_SCX200_IDE			0x0502 /*SCx200 IDE*/
#define DEVICE_NATIONAL_SCX200_AUDIO			0x0503 /*SCx200 Audio*/
#define DEVICE_NATIONAL_SCX200_VIDEO			0x0504 /*SCx200 Video*/
#define DEVICE_NATIONAL_SCX200_XBUS			0x0505 /*SCx200 XBus*/
#define DEVICE_NATIONAL_SC1100_BRIDGE			0x0510 /*SC1100 Bridge*/
#define DEVICE_NATIONAL_SC1100_SMI			0x0511 /*SC1100 SMI*/
#define DEVICE_NATIONAL_SC1100_XBUS			0x0515 /*SC1100 XBus*/
#define DEVICE_NATIONAL_87410_IDE			0xd001 /*87410 IDE*/
/* Vendor: 100c: Tseng Labs Inc */
#define DEVICE_TSENG_ET4000_W32P_REV_A			0x3202 /*ET4000/W32p rev A*/
#define DEVICE_TSENG_ET4000_W32P_REV_B			0x3205 /*ET4000/W32p rev B*/
#define DEVICE_TSENG_ET4000_W32P_REV_C			0x3206 /*ET4000/W32p rev C*/
#define DEVICE_TSENG_ET4000_W32P_REV_D			0x3207 /*ET4000/W32p rev D*/
#define DEVICE_TSENG_ET6000				0x3208 /*ET6000*/
#define DEVICE_TSENG_ET6300				0x4702 /*ET6300*/
/* Vendor: 100d: AST Research Inc */
/* Vendor: 100e: Weitek */
#define DEVICE_WEITEK_P9000_VIPER			0x9000 /*P9000 Viper*/
#define DEVICE_WEITEK_P9000_VIPER2			0x9001 /*P9000 Viper*/
#define DEVICE_WEITEK_P9000_VIPER3			0x9002 /*P9000 Viper*/
#define DEVICE_WEITEK_P9100_VIPER_PRO_SE		0x9100 /*P9100 Viper Pro/SE*/
/* Vendor: 1010: Video Logic, Ltd. */
/* Vendor: 1011: Digital Equipment Corporation */
#define DEVICE_DIGITAL_DECCHIP_21050			0x0001 /*DECchip 21050*/
#define DEVICE_DIGITAL_DECCHIP_21040_TULIP		0x0002 /*DECchip 21040 [Tulip]*/
#define DEVICE_DIGITAL_DECCHIP_21030_TGA		0x0004 /*DECchip 21030 [TGA]*/
#define DEVICE_DIGITAL_NVRAM_ZEPHYR_NVRAM		0x0007 /*NVRAM [Zephyr NVRAM]*/
#define DEVICE_DIGITAL_KZPSA_KZPSA			0x0008 /*KZPSA [KZPSA]*/
#define DEVICE_DIGITAL_DECCHIP_21140_FASTERNET		0x0009 /*DECchip 21140 [FasterNet]*/
	#define SUBDEVICE_DIGITAL_21140_FAST_ETHERNET	0x10250310 /*21140 Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_SMC9332BDT_ETHERPOWER_10_100	0x10b82001 /*SMC9332BDT EtherPower 10/100*/
	#define SUBDEVICE_DIGITAL_SMC9332BVT_ETHERPOWER_T4	0x10b82002 /*SMC9332BVT EtherPower T4 10/100*/
	#define SUBDEVICE_DIGITAL_SMC9334BDT_ETHERPOWER_10_100	0x10b82003 /*SMC9334BDT EtherPower 10/100 (1-port)*/
	#define SUBDEVICE_DIGITAL_ANA_6944A_TX_FAST_ETHERNET	0x11092400 /*ANA-6944A/TX Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_RNS2300_FAST_ETHERNET	0x11122300 /*RNS2300 Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_RNS2320_FAST_ETHERNET	0x11122320 /*RNS2320 Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_RNS2340_FAST_ETHERNET	0x11122340 /*RNS2340 Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_EN_1207_TX_FAST_ETHERNET	0x11131207 /*EN-1207-TX Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_DFE_500TX_FAST_ETHERNET	0x11861100 /*DFE-500TX Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_DFE_570TX_FAST_ETHERNET	0x11861112 /*DFE-570TX Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_DFE_660_CARDBUS_ETHERNET	0x11861140 /*DFE-660 Cardbus Ethernet 10/100*/
	#define SUBDEVICE_DIGITAL_DFE_660_CARDBUS_ETHERNET2	0x11861142 /*DFE-660 Cardbus Ethernet 10/100*/
	#define SUBDEVICE_DIGITAL_FREEDOMLINE_FAST_ETHERNET	0x11f60503 /*Freedomline Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_AEF_380TXD_FAST_ETHERNET	0x12829100 /*AEF-380TXD Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_FA310TX_FAST_ETHERNET	0x13851100 /*FA310TX Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_KNE100TX_FAST_ETHERNET	0x26460001 /*KNE100TX Fast Ethernet*/
#define DEVICE_DIGITAL_21230_VIDEO_CODEC		0x000a /*21230 Video Codec*/
#define DEVICE_DIGITAL_PBXGB_TGA2			0x000d /*PBXGB [TGA2]*/
#define DEVICE_DIGITAL_DEFPA				0x000f /*DEFPA*/
#define DEVICE_DIGITAL_DECCHIP_21041_TULIP		0x0014 /*DECchip 21041 [Tulip Pass 3]*/
	#define SUBDEVICE_DIGITAL_DE_530		0x11860100 /*DE-530+*/
#define DEVICE_DIGITAL_DGLPB_OPPO			0x0016 /*DGLPB [OPPO]*/
#define DEVICE_DIGITAL_PV_PCI_GRAPHICS_CONTROLLER	0x0017 /*PV-PCI Graphics Controller (ZLXp-L)*/
#define DEVICE_DIGITAL_DECCHIP_21142_43			0x0019 /*DECchip 21142/43*/
	#define SUBDEVICE_DIGITAL_DE500A_FAST_ETHERNET	0x1011500a /*DE500A Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_DE500B_FAST_ETHERNET	0x1011500b /*DE500B Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_10_100_ETHERJET_CARDBUS	0x10140001 /*10/100 EtherJet Cardbus*/
	#define SUBDEVICE_DIGITAL_ALN315_FAST_ETHERNET	0x10250315 /*ALN315 Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_PC_9821_CS01_100BASE_TX_INTERFACE	0x1033800c /*PC-9821-CS01 100BASE-TX Interface Card*/
	#define SUBDEVICE_DIGITAL_PC_9821NR_B06_100BASE_TX_INTERFACE	0x1033800d /*PC-9821NR-B06 100BASE-TX Interface Card*/
	#define SUBDEVICE_DIGITAL_RAPIDFIRE_2327_10_100	0x108d0016 /*Rapidfire 2327 10/100 Ethernet*/
	#define SUBDEVICE_DIGITAL_GOCARD_2250_ETHERNET	0x108d0017 /*GoCard 2250 Ethernet 10/100 Cardbus*/
	#define SUBDEVICE_DIGITAL_SMC8032DT_EXTREME_ETHERNET	0x10b82005 /*SMC8032DT Extreme Ethernet 10/100*/
	#define SUBDEVICE_DIGITAL_SMC8034_EXTREME_ETHERNET	0x10b88034 /*SMC8034 Extreme Ethernet 10/100*/
	#define SUBDEVICE_DIGITAL_CARDBUS_FAST_ETHERNET	0x10ef8169 /*Cardbus Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_ANA_6911A_TX_FAST_ETHERNET	0x11092a00 /*ANA-6911A/TX Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_ANA_6911A_TXC_FAST_ETHERNET	0x11092b00 /*ANA-6911A/TXC Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_ANA_6922_TX_FAST_ETHERNET	0x11093000 /*ANA-6922/TX Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_CHEETAH_FAST_ETHERNET	0x11131207 /*Cheetah Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_CARDBUS_FAST_ETHERNET2	0x11132220 /*Cardbus Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_CARDBUS_ETHERNET_10_100	0x115d0002 /*Cardbus Ethernet 10/100*/
	#define SUBDEVICE_DIGITAL_FAST_ETHERNET		0x11790203 /*Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_CARDBUS_FAST_ETHERNET3	0x11790204 /*Cardbus Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_DFE_500TX_FAST_ETHERNET2	0x11861100 /*DFE-500TX Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_DFE_500TX_FAST_ETHERNET3	0x11861101 /*DFE-500TX Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_DFE_500TX_FAST_ETHERNET4	0x11861102 /*DFE-500TX Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_DFE_570TX_QUAD_FAST	0x11861112 /*DFE-570TX Quad Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_AT_2800TX_FAST_ETHERNET	0x12592800 /*AT-2800Tx Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_EAGLE_FAST_ETHERMAX	0x12660004 /*Eagle Fast EtherMAX*/
	#define SUBDEVICE_DIGITAL_NETFLYER_CARDBUS_FAST	0x12af0019 /*NetFlyer Cardbus Fast Ethernet*/
	#define SUBDEVICE_DIGITAL_CARDBUS_ETHERNET_CARD	0x13740001 /*Cardbus Ethernet Card 10/100*/
	#define SUBDEVICE_DIGITAL_CARDBUS_ETHERNET_CARD2	0x13740002 /*Cardbus Ethernet Card 10/100*/
	#define SUBDEVICE_DIGITAL_CARDBUS_ETHERNET_CARD3	0x13740007 /*Cardbus Ethernet Card 10/100*/
	#define SUBDEVICE_DIGITAL_CARDBUS_ETHERNET_CARD4	0x13740008 /*Cardbus Ethernet Card 10/100*/
	#define SUBDEVICE_DIGITAL_FA510			0x13852100 /*FA510*/
	#define SUBDEVICE_DIGITAL_10_100_ETHERNET_CARDBUS	0x13950001 /*10/100 Ethernet CardBus PC Card*/
	#define SUBDEVICE_DIGITAL_ETHERFAST_10_100_CARDBUS	0x13d1ab01 /*EtherFast 10/100 Cardbus (PCMPC200)*/
	#define SUBDEVICE_DIGITAL_LNDL_100N_100BASE_TX_ETHERNET	0x14cb0100 /*LNDL-100N 100Base-TX Ethernet PC Card*/
	#define SUBDEVICE_DIGITAL_ETHEREXPRESS_PRO_100_MOBILE	0x80860001 /*EtherExpress PRO/100 Mobile CardBus 32*/
#define DEVICE_DIGITAL_FARALLON_PN9000SX_GIGABIT	0x001a /*Farallon PN9000SX Gigabit Ethernet*/
#define DEVICE_DIGITAL_DECCHIP_21052			0x0021 /*DECchip 21052*/
#define DEVICE_DIGITAL_DECCHIP_21150			0x0022 /*DECchip 21150*/
#define DEVICE_DIGITAL_DECCHIP_211502			0x0023 /*DECchip 21150*/
#define DEVICE_DIGITAL_DECCHIP_21152			0x0024 /*DECchip 21152*/
#define DEVICE_DIGITAL_DECCHIP_21153			0x0025 /*DECchip 21153*/
#define DEVICE_DIGITAL_DECCHIP_21154			0x0026 /*DECchip 21154*/
#define DEVICE_DIGITAL_56K_MODEM_CARDBUS		0x0034 /*56k Modem Cardbus*/
	#define SUBDEVICE_DIGITAL_56K_MODEM_CARDBUS	0x13740003 /*56k Modem Cardbus*/
#define DEVICE_DIGITAL_DECCHIP_21553			0x0045 /*DECchip 21553*/
#define DEVICE_DIGITAL_DECCHIP_21554			0x0046 /*DECchip 21554*/
	#define SUBDEVICE_DIGITAL_INTEGRATED_SMART_ARRAY	0x0e114050 /*Integrated Smart Array*/
	#define SUBDEVICE_DIGITAL_INTEGRATED_SMART_ARRAY2	0x0e114051 /*Integrated Smart Array*/
	#define SUBDEVICE_DIGITAL_INTEGRATED_SMART_ARRAY3	0x0e114058 /*Integrated Smart Array*/
	#define SUBDEVICE_DIGITAL_HEWLETT_PACKARD_NETRAID_4M	0x103c10c2 /*Hewlett-Packard NetRAID-4M*/
	#define SUBDEVICE_DIGITAL_IP_TELEPHONY_CARD	0x12d9000a /*IP Telephony card*/
	#define SUBDEVICE_DIGITAL_CT7_MAINBOARD		0x4c531050 /*CT7 mainboard*/
	#define SUBDEVICE_DIGITAL_CE7_MAINBOARD		0x4c531051 /*CE7 mainboard*/
	#define SUBDEVICE_DIGITAL_5400S_MUSTANG		0x90050364 /*5400S (Mustang)*/
	#define SUBDEVICE_DIGITAL_5400S_MUSTANG2	0x90050365 /*5400S (Mustang)*/
	#define SUBDEVICE_DIGITAL_DELL_POWEREDGE_RAID	0x90051364 /*Dell PowerEdge RAID Controller 2*/
	#define SUBDEVICE_DIGITAL_DELL_POWEREDGE_RAID2	0x90051365 /*Dell PowerEdge RAID Controller 2*/
	#define SUBDEVICE_DIGITAL_CC8_1_BLUES		0xe4bf1000 /*CC8-1-BLUES*/
#define DEVICE_DIGITAL_STRONGARM_DC21285		0x1065 /*StrongARM DC21285*/
	#define SUBDEVICE_DIGITAL_DAC960P_DAC1164P	0x10690020 /*DAC960P / DAC1164P*/
/* Vendor: 1012: Micronics Computers Inc */
/* Vendor: 1013: Cirrus Logic */
#define DEVICE_CIRRUS_GD_7548				0x0038 /*GD 7548*/
#define DEVICE_CIRRUS_GD_7555_FLAT			0x0040 /*GD 7555 Flat Panel GUI Accelerator*/
#define DEVICE_CIRRUS_GD_7556_VIDEO_GRAPHICS		0x004c /*GD 7556 Video/Graphics LCD/CRT Ctrlr*/
#define DEVICE_CIRRUS_GD_5430_40_ALPINE			0x00a0 /*GD 5430/40 [Alpine]*/
#define DEVICE_CIRRUS_GD_5432_ALPINE			0x00a2 /*GD 5432 [Alpine]*/
#define DEVICE_CIRRUS_GD_5434_4_ALPINE			0x00a4 /*GD 5434-4 [Alpine]*/
#define DEVICE_CIRRUS_GD_5434_8_ALPINE			0x00a8 /*GD 5434-8 [Alpine]*/
#define DEVICE_CIRRUS_GD_5436_ALPINE			0x00ac /*GD 5436 [Alpine]*/
#define DEVICE_CIRRUS_GD_5440				0x00b0 /*GD 5440*/
#define DEVICE_CIRRUS_GD_5446				0x00b8 /*GD 5446*/
#define DEVICE_CIRRUS_GD_5480				0x00bc /*GD 5480*/
	#define SUBDEVICE_CIRRUS_CL_GD5480		0x101300bc /*CL-GD5480*/
#define DEVICE_CIRRUS_GD_5462				0x00d0 /*GD 5462*/
#define DEVICE_CIRRUS_GD_5462_LAGUNA			0x00d2 /*GD 5462 [Laguna I]*/
#define DEVICE_CIRRUS_GD_5464_LAGUNA			0x00d4 /*GD 5464 [Laguna]*/
#define DEVICE_CIRRUS_GD_5464_BD			0x00d5 /*GD 5464 BD [Laguna]*/
#define DEVICE_CIRRUS_GD_5465_LAGUNA			0x00d6 /*GD 5465 [Laguna]*/
	#define SUBDEVICE_CIRRUS_BARCO_METHEUS_2	0x13ce8031 /*Barco Metheus 2 Megapixel, Dual Head*/
	#define SUBDEVICE_CIRRUS_BARCO_METHEUS_22	0x13cf8031 /*Barco Metheus 2 Megapixel, Dual Head*/
#define DEVICE_CIRRUS_GD_5436U				0x00e8 /*GD 5436U*/
#define DEVICE_CIRRUS_CL_6729				0x1100 /*CL 6729*/
#define DEVICE_CIRRUS_PD_6832_PCMCIA_CARDBUS		0x1110 /*PD 6832 PCMCIA/CardBus Ctrlr*/
#define DEVICE_CIRRUS_PD_6834_PCMCIA_CARDBUS		0x1112 /*PD 6834 PCMCIA/CardBus Ctrlr*/
#define DEVICE_CIRRUS_PD_6833_PCMCIA_CARDBUS		0x1113 /*PD 6833 PCMCIA/CardBus Ctrlr*/
#define DEVICE_CIRRUS_GD_7542_NORDIC			0x1200 /*GD 7542 [Nordic]*/
#define DEVICE_CIRRUS_GD_7543_VIKING			0x1202 /*GD 7543 [Viking]*/
#define DEVICE_CIRRUS_GD_7541_NORDIC			0x1204 /*GD 7541 [Nordic Light]*/
#define DEVICE_CIRRUS_CD_4400				0x4400 /*CD 4400*/
#define DEVICE_CIRRUS_CS_4610_11_CRYSTALCLEAR		0x6001 /*CS 4610/11 [CrystalClear SoundFusion Audio Accelerator]*/
	#define SUBDEVICE_CIRRUS_CS4610_SOUNDFUSION_AUDIO	0x10141010 /*CS4610 SoundFusion Audio Accelerator*/
#define DEVICE_CIRRUS_CS_4614_22_24_CRYSTALCLEAR	0x6003 /*CS 4614/22/24 [CrystalClear SoundFusion Audio Accelerator]*/
	#define SUBDEVICE_CIRRUS_CRYSTAL_SOUNDFUSION_PCI	0x10134280 /*Crystal SoundFusion PCI Audio Accelerator*/
	#define SUBDEVICE_CIRRUS_GAME_THEATER_XP	0x16810050 /*Game Theater XP*/
	#define SUBDEVICE_CIRRUS_FORTISSIMO_III_7_1	0x1681a011 /*Fortissimo III 7.1*/
#define DEVICE_CIRRUS_CS_4614_22_24_CRYSTALCLEAR2	0x6004 /*CS 4614/22/24 [CrystalClear SoundFusion Audio Accelerator]*/
#define DEVICE_CIRRUS_CRYSTAL_CS4281_PCI		0x6005 /*Crystal CS4281 PCI Audio*/
	#define SUBDEVICE_CIRRUS_CRYSTAL_CS4281_PCI	0x10134281 /*Crystal CS4281 PCI Audio*/
	#define SUBDEVICE_CIRRUS_CRYSTAL_CS4281_PCI2	0x10cf10a8 /*Crystal CS4281 PCI Audio*/
	#define SUBDEVICE_CIRRUS_CRYSTAL_CS4281_PCI3	0x10cf10a9 /*Crystal CS4281 PCI Audio*/
	#define SUBDEVICE_CIRRUS_CRYSTAL_CS4281_PCI4	0x10cf10aa /*Crystal CS4281 PCI Audio*/
	#define SUBDEVICE_CIRRUS_CRYSTAL_CS4281_PCI5	0x10cf10ab /*Crystal CS4281 PCI Audio*/
	#define SUBDEVICE_CIRRUS_CRYSTAL_CS4281_PCI6	0x10cf10ac /*Crystal CS4281 PCI Audio*/
	#define SUBDEVICE_CIRRUS_CRYSTAL_CS4281_PCI7	0x10cf10ad /*Crystal CS4281 PCI Audio*/
	#define SUBDEVICE_CIRRUS_CRYSTAL_CS4281_PCI8	0x10cf10b4 /*Crystal CS4281 PCI Audio*/
	#define SUBDEVICE_CIRRUS_CRYSTAL_CS4281_PCI9	0x11790001 /*Crystal CS4281 PCI Audio*/
	#define SUBDEVICE_CIRRUS_CRYSTAL_CS4281_PCI10	0x14c0000c /*Crystal CS4281 PCI Audio*/
/* Vendor: 1014: IBM */
#define DEVICE_IBM_PCI_TO_MCA				0x0002 /*PCI to MCA Bridge*/
#define DEVICE_IBM_ALTA_LITE				0x0005 /*Alta Lite*/
#define DEVICE_IBM_ALTA_MP				0x0007 /*Alta MP*/
#define DEVICE_IBM_FIRE_CORAL				0x000a /*Fire Coral*/
#define DEVICE_IBM_CPU_TO_PCI				0x0017 /*CPU to PCI Bridge*/
#define DEVICE_IBM_TR_AUTO_LANSTREAMER			0x0018 /*TR Auto LANstreamer*/
#define DEVICE_IBM_GXT_150P				0x001b /*GXT-150P*/
#define DEVICE_IBM_CARRERA				0x001c /*Carrera*/
#define DEVICE_IBM_82G2675				0x001d /*82G2675*/
#define DEVICE_IBM_GXT1000_GRAPHICS_ADAPTER		0x0020 /*GXT1000 Graphics Adapter*/
#define DEVICE_IBM_IBM27_82351				0x0022 /*IBM27-82351*/
#define DEVICE_IBM_PYTHON				0x002d /*Python*/
#define DEVICE_IBM_SCSI_RAID_ADAPTER			0x002e /*SCSI RAID Adapter [ServeRAID]*/
	#define SUBDEVICE_IBM_SERVERAID_3X		0x1014002e /*ServeRAID-3x*/
	#define SUBDEVICE_IBM_SERVERAID_4H		0x1014022e /*ServeRAID-4H*/
#define DEVICE_IBM_2_PORT_SERIAL			0x0031 /*2 Port Serial Adapter*/
	#define SUBDEVICE_IBM_2721_WAN_IOA		0x10140031 /*2721 WAN IOA - 2 Port Sync Serial Adapter*/
#define DEVICE_IBM_MIAMI					0x0036 /*Miami*/
#define DEVICE_IBM_82660_CPU_TO				0x0037 /*82660 CPU to PCI Bridge*/
#define DEVICE_IBM_CPU_TO_PCI2				0x003a /*CPU to PCI Bridge*/
#define DEVICE_IBM_GXT250P_GXT255P_GRAPHICS_ADAPTER	0x003c /*GXT250P/GXT255P Graphics Adapter*/
#define DEVICE_IBM_16_4_TOKEN_RING			0x003e /*16/4 Token ring UTP/STP controller*/
	#define SUBDEVICE_IBM_TOKEN_RING_ADAPTER	0x1014003e /*Token-Ring Adapter*/
	#define SUBDEVICE_IBM_TOKEN_RING_ADAPTER2	0x101400cd /*Token-Ring Adapter + Wake-On-LAN*/
	#define SUBDEVICE_IBM_16_4_TOKEN_RING_ADAPTER	0x101400ce /*16/4 Token-Ring Adapter 2*/
	#define SUBDEVICE_IBM_16_4_TOKEN_RING_ADAPTER2	0x101400cf /*16/4 Token-Ring Adapter Special*/
	#define SUBDEVICE_IBM_HIGH_SPEED_100_16_4_TOKEN_RING	0x101400e4 /*High-Speed 100/16/4 Token-Ring Adapter*/
	#define SUBDEVICE_IBM_16_4_TOKEN_RING_ADAPTER3	0x101400e5 /*16/4 Token-Ring Adapter 2 + Wake-On-LAN*/
	#define SUBDEVICE_IBM_ISERIES_2744_CARD		0x1014016d /*iSeries 2744 Card*/
#define DEVICE_IBM_SSA_ADAPTER				0x0045 /*SSA Adapter*/
#define DEVICE_IBM_MPIC_INTERRUPT_CONTROLLER		0x0046 /*MPIC interrupt controller*/
#define DEVICE_IBM_PCI_TO_PCI				0x0047 /*PCI to PCI Bridge*/
#define DEVICE_IBM_PCI_TO_PCI2				0x0048 /*PCI to PCI Bridge*/
#define DEVICE_IBM_WARHEAD_SCSI_CONTROLLER		0x0049 /*Warhead SCSI Controller*/
#define DEVICE_IBM_ATM_CONTROLLER_14104E00		0x004e /*ATM Controller (14104e00)*/
#define DEVICE_IBM_ATM_CONTROLLER_14104F00		0x004f /*ATM Controller (14104f00)*/
#define DEVICE_IBM_ATM_CONTROLLER_14105000		0x0050 /*ATM Controller (14105000)*/
#define DEVICE_IBM_25_MBIT_ATM				0x0053 /*25 MBit ATM Controller*/
#define DEVICE_IBM_GXT500P_GXT550P_GRAPHICS_ADAPTER	0x0054 /*GXT500P/GXT550P Graphics Adapter*/
#define DEVICE_IBM_MPEG_PCI_BRIDGE			0x0057 /*MPEG PCI Bridge*/
#define DEVICE_IBM_I82557B_10_100			0x005c /*i82557B 10/100*/
#define DEVICE_IBM_GXT800P_GRAPHICS_ADAPTER		0x005e /*GXT800P Graphics Adapter*/
#define DEVICE_IBM_ATM_CONTROLLER_14107C00		0x007c /*ATM Controller (14107c00)*/
#define DEVICE_IBM_3780IDSP_MWAVE			0x007d /*3780IDSP [MWave]*/
#define DEVICE_IBM_EADS_PCI_TO				0x008b /*EADS PCI to PCI Bridge*/
#define DEVICE_IBM_GXT3000P_GRAPHICS_ADAPTER		0x008e /*GXT3000P Graphics Adapter*/
#define DEVICE_IBM_GXT_3000P				0x0090 /*GXT 3000P*/
	#define SUBDEVICE_IBM_GXT_3000P			0x1014008e /*GXT-3000P*/
#define DEVICE_IBM_SSA_ADAPTER2				0x0091 /*SSA Adapter*/
#define DEVICE_IBM_20H2999_PCI_DOCKING			0x0095 /*20H2999 PCI Docking Bridge*/
#define DEVICE_IBM_CHUKAR_CHIPSET_SCSI			0x0096 /*Chukar chipset SCSI controller*/
	#define SUBDEVICE_IBM_ISERIES_2778_DASD		0x10140097 /*iSeries 2778 DASD IOA*/
	#define SUBDEVICE_IBM_ISERIES_2763_DASD		0x10140098 /*iSeries 2763 DASD IOA*/
	#define SUBDEVICE_IBM_ISERIES_2748_DASD		0x10140099 /*iSeries 2748 DASD IOA*/
#define DEVICE_IBM_PCI_4758_CRYPTOGRAPHIC		0x009f /*PCI 4758 Cryptographic Accelerator*/
#define DEVICE_IBM_ATM_CONTROLLER_1410A500		0x00a5 /*ATM Controller (1410a500)*/
#define DEVICE_IBM_ATM_155MBPS_MM			0x00a6 /*ATM 155MBPS MM Controller (1410a600)*/
#define DEVICE_IBM_256_BIT_GRAPHICS_RASTERIZER		0x00b7 /*256-bit Graphics Rasterizer [Fire GL1]*/
	#define SUBDEVICE_IBM_FIREGL1_AGP_32MB		0x109200b8 /*FireGL1 AGP 32Mb*/
#define DEVICE_IBM_GXT2000P_GRAPHICS_ADAPTER		0x00b8 /*GXT2000P Graphics Adapter*/
#define DEVICE_IBM_ATM_622MBPS_CONTROLLER		0x00be /*ATM 622MBPS Controller (1410be00)*/
#define DEVICE_IBM_ADVANCED_SYSTEMS_MANAGEMENT		0x00dc /*Advanced Systems Management Adapter (ASMA)*/
#define DEVICE_IBM_CPC710_DUAL_BRIDGE			0x00fc /*CPC710 Dual Bridge and Memory Controller (PCI-64)*/
#define DEVICE_IBM_GIGABIT_ETHERNET_SX_ADAPTER		0x0104 /*Gigabit Ethernet-SX Adapter*/
#define DEVICE_IBM_CPC710_DUAL_BRIDGE2			0x0105 /*CPC710 Dual Bridge and Memory Controller (PCI-32)*/
#define DEVICE_IBM_REMOTE_SUPERVISOR_ADAPTER		0x010f /*Remote Supervisor Adapter (RSA)*/
#define DEVICE_IBM_YOTTA_VIDEO_COMPOSITOR		0x0142 /*Yotta Video Compositor Input*/
	#define SUBDEVICE_IBM_YOTTA_INPUT_CONTROLLER	0x10140143 /*Yotta Input Controller (ytin)*/
#define DEVICE_IBM_YOTTA_VIDEO_COMPOSITOR2		0x0144 /*Yotta Video Compositor Output*/
	#define SUBDEVICE_IBM_YOTTA_OUTPUT_CONTROLLER	0x10140145 /*Yotta Output Controller (ytout)*/
#define DEVICE_IBM_405GP_PLB_TO				0x0156 /*405GP PLB to PCI Bridge*/
#define DEVICE_IBM_622MBPS_ATM_PCI			0x015e /*622Mbps ATM PCI Adapter*/
#define DEVICE_IBM_64BIT_66MHZ_PCI_ATM			0x0160 /*64bit/66MHz PCI ATM 155 MMF*/
#define DEVICE_IBM_GXT4000P_GRAPHICS_ADAPTER		0x016e /*GXT4000P Graphics Adapter*/
#define DEVICE_IBM_GXT6000P_GRAPHICS_ADAPTER		0x0170 /*GXT6000P Graphics Adapter*/
#define DEVICE_IBM_GXT300P_GRAPHICS_ADAPTER		0x017d /*GXT300P Graphics Adapter*/
#define DEVICE_IBM_SNIPE_CHIPSET_SCSI			0x0180 /*Snipe chipset SCSI controller*/
	#define SUBDEVICE_IBM_ISERIES_2757_DASD		0x10140241 /*iSeries 2757 DASD IOA*/
	#define SUBDEVICE_IBM_QUAD_CHANNEL_PCI_X	0x10140264 /*Quad Channel PCI-X U320 SCSI RAID Adapter (2780)*/
#define DEVICE_IBM_EADS_X_PCI_X_TO			0x0188 /*EADS-X PCI-X to PCI-X Bridge*/
#define DEVICE_IBM_PCI_X_TO_PCI_X			0x01a7 /*PCI-X to PCI-X Bridge*/
#define DEVICE_IBM_SERVERAID_CONTROLLER			0x01bd /*ServeRAID Controller*/
	#define SUBDEVICE_IBM_SERVERAID_4M		0x101401be /*ServeRAID-4M*/
	#define SUBDEVICE_IBM_SERVERAID_4L		0x101401bf /*ServeRAID-4L*/
	#define SUBDEVICE_IBM_SERVERAID_4MX		0x10140208 /*ServeRAID-4Mx*/
	#define SUBDEVICE_IBM_SERVERAID_4LX		0x1014020e /*ServeRAID-4Lx*/
	#define SUBDEVICE_IBM_SERVERAID_4H2		0x1014022e /*ServeRAID-4H*/
	#define SUBDEVICE_IBM_SERVERAID_5I		0x10140258 /*ServeRAID-5i*/
	#define SUBDEVICE_IBM_SERVERAID_5I2		0x10140259 /*ServeRAID-5i*/
#define DEVICE_IBM_64BIT_66MHZ_PCI_ATM2			0x01c1 /*64bit/66MHz PCI ATM 155 UTP*/
#define DEVICE_IBM_CRYPTOGRAPHIC_ACCELERATOR		0x01e6 /*Cryptographic Accelerator*/
#define DEVICE_IBM_10_100_MBPS_ETHERNET			0x01ff /*10/100 Mbps Ethernet*/
#define DEVICE_IBM_MULTIPORT_SERIAL_ADAPTER		0x0219 /*Multiport Serial Adapter*/
	#define SUBDEVICE_IBM_DUAL_RVX			0x1014021a /*Dual RVX*/
	#define SUBDEVICE_IBM_INTERNAL_MODEM_RVX	0x10140251 /*Internal Modem/RVX*/
	#define SUBDEVICE_IBM_QUAD_INTERNAL_MODEM	0x10140252 /*Quad Internal Modem*/
#define DEVICE_IBM_GXT6500P_GRAPHICS_ADAPTER		0x021b /*GXT6500P Graphics Adapter*/
#define DEVICE_IBM_GXT4500P_GRAPHICS_ADAPTER		0x021c /*GXT4500P Graphics Adapter*/
#define DEVICE_IBM_GXT135P_GRAPHICS_ADAPTER		0x0233 /*GXT135P Graphics Adapter*/
#define DEVICE_IBM_PCI_X_DUAL_CHANNEL			0x0266 /*PCI-X Dual Channel SCSI*/
#define DEVICE_IBM_GIGABIT_ETHERNET_SX_ADAPTER2		0x0268 /*Gigabit Ethernet-SX Adapter (PCI-X)*/
#define DEVICE_IBM_10_100_1000_BASE_TX_ETHERNET		0x0269 /*10/100/1000 Base-TX Ethernet Adapter (PCI-X)*/
#define DEVICE_IBM_WINNIPEG_PCI_X_HOST			0x0302 /*Winnipeg PCI-X Host Bridge*/
#define DEVICE_IBM_ZISC_036_NEURAL			0x0314 /*ZISC 036 Neural accelerator card*/
#define DEVICE_IBM_MPIC_2_INTERRUPT_CONTROLLER		0xffff /*MPIC-2 interrupt controller*/
/* Vendor: 1015: LSI Logic Corp of Canada */
/* Vendor: 1016: ICL Personal Systems */
/* Vendor: 1017: SPEA Software AG */
#define DEVICE_SPEA_SPEA_3D_ACCELERATOR			0x5343 /*SPEA 3D Accelerator*/
/* Vendor: 1018: Unisys Systems */
/* Vendor: 1019: Elitegroup Computer Systems */
/* Vendor: 101a: AT&T GIS (NCR) */
#define DEVICE_AT_T_100VG_ETHERNET			0x0005 /*100VG ethernet*/
/* Vendor: 101b: Vitesse Semiconductor */
/* Vendor: 101c: Western Digital */
#define DEVICE_WESTERN_33C193A				0x0193 /*33C193A*/
#define DEVICE_WESTERN_33C196A				0x0196 /*33C196A*/
#define DEVICE_WESTERN_33C197A				0x0197 /*33C197A*/
#define DEVICE_WESTERN_33C296A				0x0296 /*33C296A*/
#define DEVICE_WESTERN_7193				0x3193 /*7193*/
#define DEVICE_WESTERN_7197				0x3197 /*7197*/
#define DEVICE_WESTERN_33C296A2				0x3296 /*33C296A*/
#define DEVICE_WESTERN_34C296				0x4296 /*34C296*/
#define DEVICE_WESTERN_PIPELINE_9710			0x9710 /*Pipeline 9710*/
#define DEVICE_WESTERN_PIPELINE_9712			0x9712 /*Pipeline 9712*/
#define DEVICE_WESTERN_90C				0xc24a /*90C*/
/* Vendor: 101e: American Megatrends Inc. */
#define DEVICE_AMERICAN_MEGARAID			0x1960 /*MegaRAID*/
	#define SUBDEVICE_AMERICAN_MEGARAID_471_ENTERPRISE	0x101e0471 /*MegaRAID 471 Enterprise 1600 RAID Controller*/
	#define SUBDEVICE_AMERICAN_MEGARAID_475_EXPRESS	0x101e0475 /*MegaRAID 475 Express 500/500LC RAID Controller*/
	#define SUBDEVICE_AMERICAN_MEGARAID_477_ELITE	0x101e0477 /*MegaRAID 477 Elite 3100 RAID Controller*/
	#define SUBDEVICE_AMERICAN_MEGARAID_493_ELITE	0x101e0493 /*MegaRAID 493 Elite 1600 RAID Controller*/
	#define SUBDEVICE_AMERICAN_MEGARAID_494_ELITE	0x101e0494 /*MegaRAID 494 Elite 1650 RAID Controller*/
	#define SUBDEVICE_AMERICAN_MEGARAID_503_ENTERPRISE	0x101e0503 /*MegaRAID 503 Enterprise 1650 RAID Controller*/
	#define SUBDEVICE_AMERICAN_MEGARAID_511_I4	0x101e0511 /*MegaRAID 511 i4 IDE RAID Controller*/
	#define SUBDEVICE_AMERICAN_MEGARAID_522_I4133	0x101e0522 /*MegaRAID 522 i4133 RAID Controller*/
	#define SUBDEVICE_AMERICAN_POWEREDGE_RAID_CONTROLLER	0x10280471 /*PowerEdge RAID Controller 3/QC*/
	#define SUBDEVICE_AMERICAN_POWEREDGE_RAID_CONTROLLER2	0x10280475 /*PowerEdge RAID Controller 3/SC*/
	#define SUBDEVICE_AMERICAN_POWEREDGE_RAID_CONTROLLER3	0x10280493 /*PowerEdge RAID Controller 3/DC*/
	#define SUBDEVICE_AMERICAN_POWEREDGE_COST_EFFECTIVE	0x10280511 /*PowerEdge Cost Effective RAID Controller ATA100/4Ch*/
#define DEVICE_AMERICAN_MEGARAID_428_ULTRA		0x9010 /*MegaRAID 428 Ultra RAID Controller*/
#define DEVICE_AMERICAN_EIDE_CONTROLLER			0x9030 /*EIDE Controller*/
#define DEVICE_AMERICAN_EIDE_CONTROLLER2		0x9031 /*EIDE Controller*/
#define DEVICE_AMERICAN_EIDE_SCSI			0x9032 /*EIDE & SCSI Controller*/
#define DEVICE_AMERICAN_SCSI_CONTROLLER			0x9033 /*SCSI Controller*/
#define DEVICE_AMERICAN_MULTIMEDIA_CARD			0x9040 /*Multimedia card*/
#define DEVICE_AMERICAN_MEGARAID_434_ULTRA		0x9060 /*MegaRAID 434 Ultra GT RAID Controller*/
#define DEVICE_AMERICAN_MEGARAC				0x9063 /*MegaRAC*/
	#define SUBDEVICE_AMERICAN_DELL_REMOTE_ASSISTANT	0x101e0767 /*Dell Remote Assistant Card 2*/
/* Vendor: 101f: PictureTel */
/* Vendor: 1020: Hitachi Computer Products */
/* Vendor: 1021: OKI Electric Industry Co. Ltd. */
/* Vendor: 1022: Advanced Micro Devices [AMD] */
#define DEVICE_ADVANCED_K8_ATHLON64_OPTERON_HYPERTRANSPORT	0x1100 /*K8 [Athlon64/Opteron] HyperTransport Technology Configuration*/
#define DEVICE_ADVANCED_K8_ATHLON64_OPTERON_ADDRESS	0x1101 /*K8 [Athlon64/Opteron] Address Map*/
#define DEVICE_ADVANCED_K8_ATHLON64_OPTERON_DRAM	0x1102 /*K8 [Athlon64/Opteron] DRAM Controller*/
#define DEVICE_ADVANCED_K8_ATHLON64_OPTERON_MISCELLANEOUS	0x1103 /*K8 [Athlon64/Opteron] Miscellaneous Control*/
#define DEVICE_ADVANCED_79C970_PCNET32_LANCE		0x2000 /*79c970 [PCnet32 LANCE]*/
	#define SUBDEVICE_ADVANCED_NETFINITY_10_100_FAST	0x10142000 /*NetFinity 10/100 Fast Ethernet*/
	#define SUBDEVICE_ADVANCED_PCNET_FAST		0x10222000 /*PCnet - Fast 79C971*/
	#define SUBDEVICE_ADVANCED_ETHERNET_WITH_LAN	0x103c104c /*Ethernet with LAN remote power Adapter*/
	#define SUBDEVICE_ADVANCED_ETHERNET_WITH_LAN2	0x103c1064 /*Ethernet with LAN remote power Adapter*/
	#define SUBDEVICE_ADVANCED_ETHERNET_WITH_LAN3	0x103c1065 /*Ethernet with LAN remote power Adapter*/
	#define SUBDEVICE_ADVANCED_ETHERNET_WITH_LAN4	0x103c106c /*Ethernet with LAN remote power Adapter*/
	#define SUBDEVICE_ADVANCED_ETHERNET_WITH_LAN5	0x103c106e /*Ethernet with LAN remote power Adapter*/
	#define SUBDEVICE_ADVANCED_ETHERNET_WITH_LAN6	0x103c10ea /*Ethernet with LAN remote power Adapter*/
	#define SUBDEVICE_ADVANCED_EN1220_10_100_FAST	0x11131220 /*EN1220 10/100 Fast Ethernet*/
	#define SUBDEVICE_ADVANCED_AT_2450_10_100_FAST	0x12592450 /*AT-2450 10/100 Fast Ethernet*/
	#define SUBDEVICE_ADVANCED_AT_2450V4_10MB_ETHERNET	0x12592454 /*AT-2450v4 10Mb Ethernet Adapter*/
	#define SUBDEVICE_ADVANCED_AT_2700TX_10_100_FAST	0x12592700 /*AT-2700TX 10/100 Fast Ethernet*/
	#define SUBDEVICE_ADVANCED_AT_2700FX_100MB_ETHERNET	0x12592701 /*AT-2700FX 100Mb Ethernet*/
	#define SUBDEVICE_ADVANCED_CC7_CR7_CP7_VC7_VP7_VR7_MAINBOARD	0x4c531000 /*CC7/CR7/CP7/VC7/VP7/VR7 mainboard*/
	#define SUBDEVICE_ADVANCED_CP5_CR6_MAINBOARD	0x4c531010 /*CP5/CR6 mainboard*/
	#define SUBDEVICE_ADVANCED_VR6_MAINBOARD	0x4c531020 /*VR6 mainboard*/
	#define SUBDEVICE_ADVANCED_PC5_MAINBOARD	0x4c531030 /*PC5 mainboard*/
	#define SUBDEVICE_ADVANCED_CL7_MAINBOARD	0x4c531040 /*CL7 mainboard*/
	#define SUBDEVICE_ADVANCED_PC7_MAINBOARD	0x4c531060 /*PC7 mainboard*/
#define DEVICE_ADVANCED_79C978_HOMEPNA			0x2001 /*79c978 [HomePNA]*/
	#define SUBDEVICE_ADVANCED_MULTIMEDIA_HOME_NETWORK	0x10920a78 /*Multimedia Home Network Adapter*/
	#define SUBDEVICE_ADVANCED_ACTIONLINK_HOME_NETWORK	0x16680299 /*ActionLink Home Network Adapter*/
#define DEVICE_ADVANCED_AM_1771_MBW			0x2003 /*Am 1771 MBW [Alchemy]*/
#define DEVICE_ADVANCED_53C974_PCSCSI			0x2020 /*53c974 [PCscsi]*/
#define DEVICE_ADVANCED_79C974				0x2040 /*79c974*/
#define DEVICE_ADVANCED_ELANSC520_MICROCONTROLLER	0x3000 /*ELanSC520 Microcontroller*/
#define DEVICE_ADVANCED_AMD_751_IRONGATE_SYSTEM		0x7006 /*AMD-751 [Irongate] System Controller*/
#define DEVICE_ADVANCED_AMD_751_IRONGATE_AGP		0x7007 /*AMD-751 [Irongate] AGP Bridge*/
#define DEVICE_ADVANCED_AMD_IGR4_AGP_HOST		0x700a /*AMD-IGR4 AGP Host to PCI Bridge*/
#define DEVICE_ADVANCED_AMD_IGR4_PCI_TO			0x700b /*AMD-IGR4 PCI to PCI Bridge*/
#define DEVICE_ADVANCED_AMD_760_MP_IGD4_2P		0x700c /*AMD-760 MP [IGD4-2P] System Controller*/
#define DEVICE_ADVANCED_AMD_760_MP_IGD4_2P2		0x700d /*AMD-760 MP [IGD4-2P] AGP Bridge*/
#define DEVICE_ADVANCED_AMD_760_IGD4_1P_SYSTEM		0x700e /*AMD-760 [IGD4-1P] System Controller*/
#define DEVICE_ADVANCED_AMD_760_IGD4_1P_AGP		0x700f /*AMD-760 [IGD4-1P] AGP Bridge*/
#define DEVICE_ADVANCED_AMD_755_COBRA_ISA		0x7400 /*AMD-755 [Cobra] ISA*/
#define DEVICE_ADVANCED_AMD_755_COBRA_IDE		0x7401 /*AMD-755 [Cobra] IDE*/
#define DEVICE_ADVANCED_AMD_755_COBRA_ACPI		0x7403 /*AMD-755 [Cobra] ACPI*/
#define DEVICE_ADVANCED_AMD_755_COBRA_USB		0x7404 /*AMD-755 [Cobra] USB*/
#define DEVICE_ADVANCED_AMD_756_VIPER_ISA		0x7408 /*AMD-756 [Viper] ISA*/
#define DEVICE_ADVANCED_AMD_756_VIPER_IDE		0x7409 /*AMD-756 [Viper] IDE*/
#define DEVICE_ADVANCED_AMD_756_VIPER_ACPI		0x740b /*AMD-756 [Viper] ACPI*/
#define DEVICE_ADVANCED_AMD_756_VIPER_USB		0x740c /*AMD-756 [Viper] USB*/
#define DEVICE_ADVANCED_AMD_766_VIPERPLUS_ISA		0x7410 /*AMD-766 [ViperPlus] ISA*/
#define DEVICE_ADVANCED_AMD_766_VIPERPLUS_IDE		0x7411 /*AMD-766 [ViperPlus] IDE*/
#define DEVICE_ADVANCED_AMD_766_VIPERPLUS_ACPI		0x7413 /*AMD-766 [ViperPlus] ACPI*/
#define DEVICE_ADVANCED_AMD_766_VIPERPLUS_USB		0x7414 /*AMD-766 [ViperPlus] USB*/
#define DEVICE_ADVANCED_AMD_768_OPUS_ISA		0x7440 /*AMD-768 [Opus] ISA*/
	#define SUBDEVICE_ADVANCED_A7M_D_MAINBOARD	0x10438044 /*A7M-D Mainboard*/
#define DEVICE_ADVANCED_AMD_768_OPUS_IDE		0x7441 /*AMD-768 [Opus] IDE*/
#define DEVICE_ADVANCED_AMD_768_OPUS_ACPI		0x7443 /*AMD-768 [Opus] ACPI*/
	#define SUBDEVICE_ADVANCED_A7M_D_MAINBOARD2	0x10438044 /*A7M-D Mainboard*/
#define DEVICE_ADVANCED_AMD_768_OPUS_AUDIO		0x7445 /*AMD-768 [Opus] Audio*/
#define DEVICE_ADVANCED_AMD_768_OPUS_MC97		0x7446 /*AMD-768 [Opus] MC97 Modem (Smart Link HAMR5600 compatible)*/
#define DEVICE_ADVANCED_AMD_768_OPUS_PCI		0x7448 /*AMD-768 [Opus] PCI*/
#define DEVICE_ADVANCED_AMD_768_OPUS_USB		0x7449 /*AMD-768 [Opus] USB*/
#define DEVICE_ADVANCED_AMD_8131_PCI_X_BRIDGE		0x7450 /*AMD-8131 PCI-X Bridge*/
#define DEVICE_ADVANCED_AMD_8131_PCI_X_APIC		0x7451 /*AMD-8131 PCI-X APIC*/
#define DEVICE_ADVANCED_AMD_8151_SYSTEM_CONTROLLER	0x7454 /*AMD-8151 System Controller*/
#define DEVICE_ADVANCED_AMD_8151_AGP_BRIDGE		0x7455 /*AMD-8151 AGP Bridge*/
#define DEVICE_ADVANCED_AMD_8111_PCI			0x7460 /*AMD-8111 PCI*/
	#define SUBDEVICE_ADVANCED_HDAMB		0x161f3017 /*HDAMB*/
#define DEVICE_ADVANCED_AMD_8111_USB			0x7461 /*AMD-8111 USB*/
#define DEVICE_ADVANCED_AMD_8111_ETHERNET		0x7462 /*AMD-8111 Ethernet*/
#define DEVICE_ADVANCED_AMD_8111_USB2			0x7464 /*AMD-8111 USB*/
	#define SUBDEVICE_ADVANCED_HDAMB2		0x161f3017 /*HDAMB*/
#define DEVICE_ADVANCED_AMD_8111_LPC			0x7468 /*AMD-8111 LPC*/
	#define SUBDEVICE_ADVANCED_HDAMB3		0x161f3017 /*HDAMB*/
#define DEVICE_ADVANCED_AMD_8111_IDE			0x7469 /*AMD-8111 IDE*/
	#define SUBDEVICE_ADVANCED_HDAMB4		0x161f3017 /*HDAMB*/
#define DEVICE_ADVANCED_AMD_8111_SMBUS_2_0		0x746a /*AMD-8111 SMBus 2.0*/
#define DEVICE_ADVANCED_AMD_8111_ACPI			0x746b /*AMD-8111 ACPI*/
	#define SUBDEVICE_ADVANCED_HDAMB5		0x161f3017 /*HDAMB*/
#define DEVICE_ADVANCED_AMD_8111_AC97_AUDIO		0x746d /*AMD-8111 AC97 Audio*/
	#define SUBDEVICE_ADVANCED_HDAMB6		0x161f3017 /*HDAMB*/
#define DEVICE_ADVANCED_AMD_8111_MC97_MODEM		0x746e /*AMD-8111 MC97 Modem*/
#define DEVICE_ADVANCED_AMD_8111_ACPI2			0x756b /*AMD-8111 ACPI*/
/* Vendor: 1023: Trident Microsystems */
#define DEVICE_TRIDENT_82C194				0x0194 /*82C194*/
#define DEVICE_TRIDENT_4DWAVE_DX			0x2000 /*4DWave DX*/
#define DEVICE_TRIDENT_4DWAVE_NX			0x2001 /*4DWave NX*/
	#define SUBDEVICE_TRIDENT_TRIDENT_PCI288_Q3DII_NX	0x122d1400 /*Trident PCI288-Q3DII (NX)*/
#define DEVICE_TRIDENT_CYBERBLADE_XP4M32		0x2100 /*CyberBlade XP4m32*/
#define DEVICE_TRIDENT_CYBERBLADE_I7			0x8400 /*CyberBlade/i7*/
	#define SUBDEVICE_TRIDENT_CYBERBLADE_I7_AGP	0x10238400 /*CyberBlade i7 AGP*/
#define DEVICE_TRIDENT_CYBERBLADE_I7D			0x8420 /*CyberBlade/i7d*/
	#define SUBDEVICE_TRIDENT_CYBERBLADE_I7_AGP2	0x0e11b15a /*CyberBlade i7 AGP*/
#define DEVICE_TRIDENT_CYBERBLADE_I1			0x8500 /*CyberBlade/i1*/
#define DEVICE_TRIDENT_CYBERBLADE_I12			0x8520 /*CyberBlade i1*/
	#define SUBDEVICE_TRIDENT_CYBERBLADE_I1_AGP	0x0e11b16e /*CyberBlade i1 AGP*/
	#define SUBDEVICE_TRIDENT_CYBERBLADE_I1_AGP2	0x10238520 /*CyberBlade i1 AGP*/
#define DEVICE_TRIDENT_CYBERBLADE_I13			0x8620 /*CyberBlade/i1*/
	#define SUBDEVICE_TRIDENT_THINKPAD_R30_T30	0x10140502 /*ThinkPad R30/T30*/
#define DEVICE_TRIDENT_CYBERBLADE_XPAI1			0x8820 /*CyberBlade XPAi1*/
#define DEVICE_TRIDENT_TGUI_9320			0x9320 /*TGUI 9320*/
#define DEVICE_TRIDENT_GUI_ACCELERATOR			0x9350 /*GUI Accelerator*/
#define DEVICE_TRIDENT_FLAT_PANEL_GUI			0x9360 /*Flat panel GUI Accelerator*/
#define DEVICE_TRIDENT_CYBER_9382_REFERENCE		0x9382 /*Cyber 9382 [Reference design]*/
#define DEVICE_TRIDENT_CYBER_9383_REFERENCE		0x9383 /*Cyber 9383 [Reference design]*/
#define DEVICE_TRIDENT_CYBER_9385_REFERENCE		0x9385 /*Cyber 9385 [Reference design]*/
#define DEVICE_TRIDENT_CYBER_9386			0x9386 /*Cyber 9386*/
#define DEVICE_TRIDENT_CYBER_9388			0x9388 /*Cyber 9388*/
#define DEVICE_TRIDENT_CYBER_9397			0x9397 /*Cyber 9397*/
#define DEVICE_TRIDENT_CYBER_9397DVD			0x939a /*Cyber 9397DVD*/
#define DEVICE_TRIDENT_TGUI_9420			0x9420 /*TGUI 9420*/
#define DEVICE_TRIDENT_TGUI_9430			0x9430 /*TGUI 9430*/
#define DEVICE_TRIDENT_TGUI_9440			0x9440 /*TGUI 9440*/
#define DEVICE_TRIDENT_TGUI_9460			0x9460 /*TGUI 9460*/
#define DEVICE_TRIDENT_TGUI_9470			0x9470 /*TGUI 9470*/
#define DEVICE_TRIDENT_CYBER_9520			0x9520 /*Cyber 9520*/
#define DEVICE_TRIDENT_CYBER_9525			0x9525 /*Cyber 9525*/
	#define SUBDEVICE_TRIDENT_LIFEBOOK_C6155	0x10cf1094 /*Lifebook C6155*/
#define DEVICE_TRIDENT_CYBER_9540			0x9540 /*Cyber 9540*/
#define DEVICE_TRIDENT_TGUI_9660_938X_968X		0x9660 /*TGUI 9660/938x/968x*/
#define DEVICE_TRIDENT_TGUI_9680			0x9680 /*TGUI 9680*/
#define DEVICE_TRIDENT_TGUI_9682			0x9682 /*TGUI 9682*/
#define DEVICE_TRIDENT_TGUI_9683			0x9683 /*TGUI 9683*/
#define DEVICE_TRIDENT_PROVIDIA_9685			0x9685 /*ProVIDIA 9685*/
#define DEVICE_TRIDENT_3DIMAGE_9750			0x9750 /*3DImage 9750*/
	#define SUBDEVICE_TRIDENT_3DIMAGE_9750		0x10149750 /*3DImage 9750*/
	#define SUBDEVICE_TRIDENT_3DIMAGE_97502		0x10239750 /*3DImage 9750*/
#define DEVICE_TRIDENT_TGUI_9753			0x9753 /*TGUI 9753*/
#define DEVICE_TRIDENT_TGUI_9754			0x9754 /*TGUI 9754*/
#define DEVICE_TRIDENT_TGUI_975				0x9759 /*TGUI 975*/
#define DEVICE_TRIDENT_TGUI_9783			0x9783 /*TGUI 9783*/
#define DEVICE_TRIDENT_TGUI_9785			0x9785 /*TGUI 9785*/
#define DEVICE_TRIDENT_3DIMAGE_9850			0x9850 /*3DImage 9850*/
#define DEVICE_TRIDENT_BLADE_3D_PCI_AGP			0x9880 /*Blade 3D PCI/AGP*/
	#define SUBDEVICE_TRIDENT_BLADE_3D		0x10239880 /*Blade 3D*/
#define DEVICE_TRIDENT_CYBERBLADE_XP			0x9910 /*CyberBlade/XP*/
#define DEVICE_TRIDENT_CYBERBLADE_XPM			0x9930 /*CyberBlade/XPm*/
/* Vendor: 1024: Zenith Data Systems */
/* Vendor: 1025: Acer Incorporated [ALI] */
#define DEVICE_ACER_M1435				0x1435 /*M1435*/
#define DEVICE_ACER_M1445				0x1445 /*M1445*/
#define DEVICE_ACER_M1449				0x1449 /*M1449*/
#define DEVICE_ACER_M1451				0x1451 /*M1451*/
#define DEVICE_ACER_M1461				0x1461 /*M1461*/
#define DEVICE_ACER_M1489				0x1489 /*M1489*/
#define DEVICE_ACER_M1511				0x1511 /*M1511*/
#define DEVICE_ACER_ALI_M1512_ALADDIN			0x1512 /*ALI M1512 Aladdin*/
#define DEVICE_ACER_M1513				0x1513 /*M1513*/
#define DEVICE_ACER_ALI_M1521_ALADDIN			0x1521 /*ALI M1521 Aladdin III CPU Bridge*/
	#define SUBDEVICE_ACER_ALI_M1521_ALADDIN	0x10b91521 /*ALI M1521 Aladdin III CPU Bridge*/
#define DEVICE_ACER_ALI_M1523_ISA			0x1523 /*ALI M1523 ISA Bridge*/
	#define SUBDEVICE_ACER_ALI_M1523_ISA		0x10b91523 /*ALI M1523 ISA Bridge*/
#define DEVICE_ACER_M1531_NORTHBRIDGE_ALADDIN		0x1531 /*M1531 Northbridge [Aladdin IV/IV+]*/
#define DEVICE_ACER_M1533_PCI_TO_ISA_BRIDGE		0x1533 /*M1533 PCI-to-ISA Bridge*/
	#define SUBDEVICE_ACER_ALI_M1533_ALADDIN	0x10b91533 /*ALI M1533 Aladdin IV/V ISA South Bridge*/
#define DEVICE_ACER_M1535_PCI_BRIDGE			0x1535 /*M1535 PCI Bridge + Super I/O + FIR*/
#define DEVICE_ACER_M1541_NORTHBRIDGE_ALADDIN		0x1541 /*M1541 Northbridge [Aladdin V]*/
	#define SUBDEVICE_ACER_ALI_M1541_ALADDIN	0x10b91541 /*ALI M1541 Aladdin V/V+ AGP+PCI North Bridge*/
#define DEVICE_ACER_M1542_NORTHBRIDGE_ALADDIN		0x1542 /*M1542 Northbridge [Aladdin V]*/
#define DEVICE_ACER_M1543_PCI_TO_ISA_BRIDGE		0x1543 /*M1543 PCI-to-ISA Bridge + Super I/O + FIR*/
#define DEVICE_ACER_M1561_NORTHBRIDGE_ALADDIN		0x1561 /*M1561 Northbridge [Aladdin 7]*/
#define DEVICE_ACER_M1621_NORTHBRIDGE_ALADDIN_PRO	0x1621 /*M1621 Northbridge [Aladdin-Pro II]*/
#define DEVICE_ACER_M1631_NORTHBRIDGE_3D_GRAPHICS	0x1631 /*M1631 Northbridge+3D Graphics [Aladdin TNT2]*/
#define DEVICE_ACER_M1641_NORTHBRIDGE_ALADDIN_PRO	0x1641 /*M1641 Northbridge [Aladdin-Pro IV]*/
#define DEVICE_ACER_M1647_MAGIK1_PCI			0x1647 /*M1647 [MaGiK1] PCI North Bridge*/
#define DEVICE_ACER_M1671_NORTHBRIDGE_ALADDIN_P4	0x1671 /*M1671 Northbridge [ALADDiN-P4]*/
#define DEVICE_ACER_NORTHBRIDGE_CYBERALADDIN_P4		0x1672 /*Northbridge [CyberALADDiN-P4]*/
#define DEVICE_ACER_M3141				0x3141 /*M3141*/
#define DEVICE_ACER_M3143				0x3143 /*M3143*/
#define DEVICE_ACER_M3145				0x3145 /*M3145*/
#define DEVICE_ACER_M3147				0x3147 /*M3147*/
#define DEVICE_ACER_M3149				0x3149 /*M3149*/
#define DEVICE_ACER_M3151				0x3151 /*M3151*/
#define DEVICE_ACER_M3307_MPEG_I_VIDEO			0x3307 /*M3307 MPEG-I Video Controller*/
#define DEVICE_ACER_M3309_MPEG_II_VIDEO			0x3309 /*M3309 MPEG-II Video w/ Software Audio Decoder*/
#define DEVICE_ACER_M3321_MPEG_II_AUDIO_VIDEO		0x3321 /*M3321 MPEG-II Audio/Video Decoder*/
#define DEVICE_ACER_M4803				0x5212 /*M4803*/
#define DEVICE_ACER_ALI_PCI_EIDE			0x5215 /*ALI PCI EIDE Controller*/
#define DEVICE_ACER_M5217H				0x5217 /*M5217H*/
#define DEVICE_ACER_M5219				0x5219 /*M5219*/
#define DEVICE_ACER_M5225				0x5225 /*M5225*/
#define DEVICE_ACER_M5229				0x5229 /*M5229*/
#define DEVICE_ACER_M5235				0x5235 /*M5235*/
#define DEVICE_ACER_M5237_PCI_USB			0x5237 /*M5237 PCI USB Host Controller*/
#define DEVICE_ACER_EIDE_CONTROLLER			0x5240 /*EIDE Controller*/
#define DEVICE_ACER_PCMCIA_BRIDGE			0x5241 /*PCMCIA Bridge*/
#define DEVICE_ACER_GENERAL_PURPOSE_CONTROLLER		0x5242 /*General Purpose Controller*/
#define DEVICE_ACER_PCI_TO_PCI				0x5243 /*PCI to PCI Bridge Controller*/
#define DEVICE_ACER_FLOPPY_DISK_CONTROLLER		0x5244 /*Floppy Disk Controller*/
#define DEVICE_ACER_M1541_PCI_TO			0x5247 /*M1541 PCI to PCI Bridge*/
#define DEVICE_ACER_M5251_P1394_CONTROLLER		0x5251 /*M5251 P1394 Controller*/
#define DEVICE_ACER_PCI_TO_AGP				0x5427 /*PCI to AGP Bridge*/
#define DEVICE_ACER_M5451_PCI_AC_LINK			0x5451 /*M5451 PCI AC-Link Controller Audio Device*/
#define DEVICE_ACER_M5453_PCI_AC_LINK			0x5453 /*M5453 PCI AC-Link Controller Modem Device*/
#define DEVICE_ACER_M7101_PCI_PMU			0x7101 /*M7101 PCI PMU Power Management Controller*/
	#define SUBDEVICE_ACER_M7101_PCI_PMU		0x10b97101 /*M7101 PCI PMU Power Management Controller*/
/* Vendor: 1028: Dell */
#define DEVICE_DELL_POWEREDGE_EXPANDABLE_RAID		0x0001 /*PowerEdge Expandable RAID Controller 2/Si*/
	#define SUBDEVICE_DELL_POWEREDGE_2400		0x10280001 /*PowerEdge 2400*/
#define DEVICE_DELL_POWEREDGE_EXPANDABLE_RAID2		0x0002 /*PowerEdge Expandable RAID Controller 3/Di*/
	#define SUBDEVICE_DELL_POWEREDGE_4400		0x10280002 /*PowerEdge 4400*/
#define DEVICE_DELL_POWEREDGE_EXPANDABLE_RAID3		0x0003 /*PowerEdge Expandable RAID Controller 3/Si*/
	#define SUBDEVICE_DELL_POWEREDGE_2450		0x10280003 /*PowerEdge 2450*/
#define DEVICE_DELL_POWEREDGE_EXPANDABLE_RAID4		0x0006 /*PowerEdge Expandable RAID Controller 3/Di*/
#define DEVICE_DELL_REMOTE_ACCESS_CARD			0x0007 /*Remote Access Card III*/
#define DEVICE_DELL_REMOTE_ACCESS_CARD2			0x0008 /*Remote Access Card III*/
#define DEVICE_DELL_REMOTE_ACCESS_CARD3			0x0009 /*Remote Access Card III: BMC/SMIC device not present*/
#define DEVICE_DELL_POWEREDGE_EXPANDABLE_RAID5		0x000a /*PowerEdge Expandable RAID Controller 3/Di*/
#define DEVICE_DELL_EMBEDDED_REMOTE_ACCESS		0x000c /*Embedded Remote Access or ERA/O*/
#define DEVICE_DELL_EMBEDDED_REMOTE_ACCESS2		0x000d /*Embedded Remote Access: BMC/SMIC device*/
#define DEVICE_DELL_POWEREDGE_EXPANDABLE_RAID6		0x000e /*PowerEdge Expandable RAID controller 4/Di*/
#define DEVICE_DELL_POWEREDGE_EXPANDABLE_RAID7		0x000f /*PowerEdge Expandable RAID controller 4/Di*/
#define DEVICE_DELL_REMOTE_ACCESS_CARD4			0x0010 /*Remote Access Card 4*/
#define DEVICE_DELL_REMOTE_ACCESS_CARD5			0x0011 /*Remote Access Card 4 Daughter Card*/
#define DEVICE_DELL_REMOTE_ACCESS_CARD6			0x0012 /*Remote Access Card 4 Daughter Card Virtual UART*/
#define DEVICE_DELL_POWEREDGE_EXPANDABLE_RAID8		0x0013 /*PowerEdge Expandable RAID controller 4*/
	#define SUBDEVICE_DELL_POWEREDGE_EXPANDABLE_RAID	0x1028016c /*PowerEdge Expandable RAID Controller 4e/Si*/
	#define SUBDEVICE_DELL_POWEREDGE_EXPANDABLE_RAID2	0x1028016d /*PowerEdge Expandable RAID Controller 4e/Di*/
	#define SUBDEVICE_DELL_POWEREDGE_EXPANDABLE_RAID3	0x1028016e /*PowerEdge Expandable RAID Controller 4e/Di*/
	#define SUBDEVICE_DELL_POWEREDGE_EXPANDABLE_RAID4	0x1028016f /*PowerEdge Expandable RAID Controller 4e/Di*/
	#define SUBDEVICE_DELL_POWEREDGE_EXPANDABLE_RAID5	0x10280170 /*PowerEdge Expandable RAID Controller 4e/Di*/
#define DEVICE_DELL_REMOTE_ACCESS_CARD7			0x0014 /*Remote Access Card 4 Daughter Card SMIC interface*/
/* Vendor: 1029: Siemens Nixdorf IS */
/* Vendor: 102a: LSI Logic */
#define DEVICE_LSI3_HYDRA				0x0000 /*HYDRA*/
#define DEVICE_LSI3_ASPEN				0x0010 /*ASPEN*/
#define DEVICE_LSI3_AHA_2940U2_U2W_7890_7891_SCSI	0x001f /*AHA-2940U2/U2W /7890/7891 SCSI Controllers*/
	#define SUBDEVICE_LSI3_2940U2W_SCSI_CONTROLLER	0x9005000f /*2940U2W SCSI Controller*/
	#define SUBDEVICE_LSI3_2940U2W_SCSI_CONTROLLER2	0x90050106 /*2940U2W SCSI Controller*/
	#define SUBDEVICE_LSI3_2940U2W_SCSI_CONTROLLER3	0x9005a180 /*2940U2W SCSI Controller*/
#define DEVICE_LSI3_AIC_7899_U160_M_SCSI		0x00c5 /*AIC-7899 U160/m SCSI Controller*/
	#define SUBDEVICE_LSI3_POWEREDGE_2550_2650_4600	0x102800c5 /*PowerEdge 2550/2650/4600*/
#define DEVICE_LSI3_AIC_7899P_U160_M			0x00cf /*AIC-7899P U160/m*/
	#define SUBDEVICE_LSI3_POWEREDGE_4600		0x10280106 /*PowerEdge 4600*/
	#define SUBDEVICE_LSI3_POWEREDGE_2650		0x10280121 /*PowerEdge 2650*/
/* Vendor: 102b: Matrox Graphics, Inc. */
#define DEVICE_MATROX_MGA_I_IMPRESSION			0x0010 /*MGA-I [Impression?]*/
#define DEVICE_MATROX_MGA_1064SG_MYSTIQUE		0x0100 /*MGA 1064SG [Mystique]*/
#define DEVICE_MATROX_MGA_II_ATHENA			0x0518 /*MGA-II [Athena]*/
#define DEVICE_MATROX_MGA_2064W_MILLENNIUM		0x0519 /*MGA 2064W [Millennium]*/
#define DEVICE_MATROX_MGA_1064SG_MYSTIQUE2		0x051a /*MGA 1064SG [Mystique]*/
	#define SUBDEVICE_MATROX_MGA_1064SG_MYSTIQUE	0x102b0100 /*MGA-1064SG Mystique*/
	#define SUBDEVICE_MATROX_MGA_1084SG_MYSTIQUE	0x102b1100 /*MGA-1084SG Mystique*/
	#define SUBDEVICE_MATROX_MGA_1084SG_MYSTIQUE2	0x102b1200 /*MGA-1084SG Mystique*/
	#define SUBDEVICE_MATROX_MGA_1084SG_MYSTIQUE3	0x1100102b /*MGA-1084SG Mystique*/
	#define SUBDEVICE_MATROX_SCENIC_PRO_C5		0x110a0018 /*Scenic Pro C5 (D1025)*/
#define DEVICE_MATROX_MGA_2164W_MILLENNIUM		0x051b /*MGA 2164W [Millennium II]*/
	#define SUBDEVICE_MATROX_MGA_2164W_MILLENNIUM_II	0x102b051b /*MGA-2164W Millennium II*/
	#define SUBDEVICE_MATROX_MGA_2164W_MILLENNIUM_II2	0x102b1100 /*MGA-2164W Millennium II*/
	#define SUBDEVICE_MATROX_MGA_2164W_MILLENNIUM_II3	0x102b1200 /*MGA-2164W Millennium II*/
#define DEVICE_MATROX_MGA_1064SG_MYSTIQUE3		0x051e /*MGA 1064SG [Mystique] AGP*/
#define DEVICE_MATROX_MGA_2164W_MILLENNIUM2		0x051f /*MGA 2164W [Millennium II] AGP*/
#define DEVICE_MATROX_MGA_G200				0x0520 /*MGA G200*/
	#define SUBDEVICE_MATROX_G200_MULTI_MONITOR	0x102bdbc2 /*G200 Multi-Monitor*/
	#define SUBDEVICE_MATROX_G200_MULTI_MONITOR2	0x102bdbc8 /*G200 Multi-Monitor*/
	#define SUBDEVICE_MATROX_G200_MULTI_MONITOR3	0x102bdbe2 /*G200 Multi-Monitor*/
	#define SUBDEVICE_MATROX_G200_MULTI_MONITOR4	0x102bdbe8 /*G200 Multi-Monitor*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G200_SD	0x102bff03 /*Millennium G200 SD*/
	#define SUBDEVICE_MATROX_MARVEL_G200		0x102bff04 /*Marvel G200*/
#define DEVICE_MATROX_MGA_G200_AGP			0x0521 /*MGA G200 AGP*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G200_AGP	0x1014ff03 /*Millennium G200 AGP*/
	#define SUBDEVICE_MATROX_MYSTIQUE_G200_AGP	0x102b48e9 /*Mystique G200 AGP*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G200_SD2	0x102b48f8 /*Millennium G200 SD AGP*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G200_LE	0x102b4a60 /*Millennium G200 LE AGP*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G200_AGP2	0x102b4a64 /*Millennium G200 AGP*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G200_AGP3	0x102bc93c /*Millennium G200 AGP*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G200_AGP4	0x102bc9b0 /*Millennium G200 AGP*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G200_AGP5	0x102bc9bc /*Millennium G200 AGP*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G250_LE	0x102bca60 /*Millennium G250 LE AGP*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G250_AGP	0x102bca6c /*Millennium G250 AGP*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G200_AGP6	0x102bdbbc /*Millennium G200 AGP*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G200_MMS	0x102bdbc2 /*Millennium G200 MMS (Dual G200)*/
	#define SUBDEVICE_MATROX_G200_MULTI_MONITOR5	0x102bdbc3 /*G200 Multi-Monitor*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G200_MMS2	0x102bdbc8 /*Millennium G200 MMS (Dual G200)*/
	#define SUBDEVICE_MATROX_G200_MULTI_MONITOR6	0x102bdbd2 /*G200 Multi-Monitor*/
	#define SUBDEVICE_MATROX_G200_MULTI_MONITOR7	0x102bdbd3 /*G200 Multi-Monitor*/
	#define SUBDEVICE_MATROX_G200_MULTI_MONITOR8	0x102bdbd4 /*G200 Multi-Monitor*/
	#define SUBDEVICE_MATROX_G200_MULTI_MONITOR9	0x102bdbd5 /*G200 Multi-Monitor*/
	#define SUBDEVICE_MATROX_G200_MULTI_MONITOR10	0x102bdbd8 /*G200 Multi-Monitor*/
	#define SUBDEVICE_MATROX_G200_MULTI_MONITOR11	0x102bdbd9 /*G200 Multi-Monitor*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G200_MMS3	0x102bdbe2 /*Millennium G200 MMS (Quad G200)*/
	#define SUBDEVICE_MATROX_G200_MULTI_MONITOR12	0x102bdbe3 /*G200 Multi-Monitor*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G200_MMS4	0x102bdbe8 /*Millennium G200 MMS (Quad G200)*/
	#define SUBDEVICE_MATROX_G200_MULTI_MONITOR13	0x102bdbf2 /*G200 Multi-Monitor*/
	#define SUBDEVICE_MATROX_G200_MULTI_MONITOR14	0x102bdbf3 /*G200 Multi-Monitor*/
	#define SUBDEVICE_MATROX_G200_MULTI_MONITOR15	0x102bdbf4 /*G200 Multi-Monitor*/
	#define SUBDEVICE_MATROX_G200_MULTI_MONITOR16	0x102bdbf5 /*G200 Multi-Monitor*/
	#define SUBDEVICE_MATROX_G200_MULTI_MONITOR17	0x102bdbf8 /*G200 Multi-Monitor*/
	#define SUBDEVICE_MATROX_G200_MULTI_MONITOR18	0x102bdbf9 /*G200 Multi-Monitor*/
	#define SUBDEVICE_MATROX_MYSTIQUE_G200_VIDEO	0x102bf806 /*Mystique G200 Video AGP*/
	#define SUBDEVICE_MATROX_MGA_G200_AGP		0x102bff00 /*MGA-G200 AGP*/
	#define SUBDEVICE_MATROX_MYSTIQUE_G200_AGP2	0x102bff02 /*Mystique G200 AGP*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G200_AGP7	0x102bff03 /*Millennium G200 AGP*/
	#define SUBDEVICE_MATROX_MARVEL_G200_AGP	0x102bff04 /*Marvel G200 AGP*/
	#define SUBDEVICE_MATROX_MGA_G200_AGP2		0x110a0032 /*MGA-G200 AGP*/
#define DEVICE_MATROX_MGA_G400_AGP			0x0525 /*MGA G400 AGP*/
	#define SUBDEVICE_MATROX_MGA_G400_AGP		0x0e11b16f /*MGA-G400 AGP*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G400_16MB	0x102b0328 /*Millennium G400 16Mb SDRAM*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G400_16MB2	0x102b0338 /*Millennium G400 16Mb SDRAM*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G400_32MB	0x102b0378 /*Millennium G400 32Mb SDRAM*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G450_DUAL	0x102b0541 /*Millennium G450 Dual Head*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G450_DUAL2	0x102b0542 /*Millennium G450 Dual Head LX*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G450_SINGLE	0x102b0543 /*Millennium G450 Single Head LX*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G450_32MB	0x102b0641 /*Millennium G450 32Mb SDRAM Dual Head*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G450_32MB2	0x102b0642 /*Millennium G450 32Mb SDRAM Dual Head LX*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G450_32MB3	0x102b0643 /*Millennium G450 32Mb SDRAM Single Head LX*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G450_DUAL3	0x102b07c0 /*Millennium G450 Dual Head LE*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G450_SDR	0x102b07c1 /*Millennium G450 SDR Dual Head LE*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G450_DUAL4	0x102b0d41 /*Millennium G450 Dual Head PCI*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G450_DUAL5	0x102b0d42 /*Millennium G450 Dual Head LX PCI*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G450_32MB4	0x102b0d43 /*Millennium G450 32Mb Dual Head PCI*/
	#define SUBDEVICE_MATROX_MARVEL_G450_ETV	0x102b0e00 /*Marvel G450 eTV*/
	#define SUBDEVICE_MATROX_MARVEL_G450_ETV2	0x102b0e01 /*Marvel G450 eTV*/
	#define SUBDEVICE_MATROX_MARVEL_G450_ETV3	0x102b0e02 /*Marvel G450 eTV*/
	#define SUBDEVICE_MATROX_MARVEL_G450_ETV4	0x102b0e03 /*Marvel G450 eTV*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G450_LOW	0x102b0f80 /*Millennium G450 Low Profile*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G450_LOW2	0x102b0f81 /*Millennium G450 Low Profile*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G450_LOW3	0x102b0f82 /*Millennium G450 Low Profile DVI*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G450_LOW4	0x102b0f83 /*Millennium G450 Low Profile DVI*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G400_16MB3	0x102b19d8 /*Millennium G400 16Mb SGRAM*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G400_32MB2	0x102b19f8 /*Millennium G400 32Mb SGRAM*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G400_DUAL	0x102b2159 /*Millennium G400 Dual Head 16Mb*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G400_MAX_DUAL	0x102b2179 /*Millennium G400 MAX/Dual Head 32Mb*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G400_DUAL2	0x102b217d /*Millennium G400 Dual Head Max*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G450	0x102b23c0 /*Millennium G450*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G4502	0x102b23c1 /*Millennium G450*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G450_DVI	0x102b23c2 /*Millennium G450 DVI*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G450_DVI2	0x102b23c3 /*Millennium G450 DVI*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G400	0x102b2f58 /*Millennium G400*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G4002	0x102b2f78 /*Millennium G400*/
	#define SUBDEVICE_MATROX_MARVEL_G400_AGP	0x102b3693 /*Marvel G400 AGP*/
	#define SUBDEVICE_MATROX_4SIGHT_II		0x102b5dd0 /*4Sight II*/
	#define SUBDEVICE_MATROX_4SIGHT_II2		0x102b5f50 /*4Sight II*/
	#define SUBDEVICE_MATROX_4SIGHT_II3		0x102b5f51 /*4Sight II*/
	#define SUBDEVICE_MATROX_4SIGHT_II4		0x102b5f52 /*4Sight II*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G400_DUAL3	0x102b9010 /*Millennium G400 Dual Head*/
	#define SUBDEVICE_MATROX_GA_G400		0x14580400 /*GA-G400*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G450_32MB5	0x17050001 /*Millennium G450 32MB SGRAM*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G450_16MB	0x17050002 /*Millennium G450 16MB SGRAM*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G450_32MB6	0x17050003 /*Millennium G450 32MB*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G450_16MB2	0x17050004 /*Millennium G450 16MB*/
#define DEVICE_MATROX_MGA_PARHELIA_AGP			0x0527 /*MGA Parhelia AGP*/
	#define SUBDEVICE_MATROX_PARHELIA_128MB		0x102b0840 /*Parhelia 128Mb*/
#define DEVICE_MATROX_MGA_ULTIMA_IMPRESSION		0x0d10 /*MGA Ultima/Impression*/
#define DEVICE_MATROX_MGA_G100_PRODUCTIVA		0x1000 /*MGA G100 [Productiva]*/
	#define SUBDEVICE_MATROX_PRODUCTIVA_G100	0x102bff01 /*Productiva G100*/
	#define SUBDEVICE_MATROX_PRODUCTIVA_G100_MULTI_MONITOR	0x102bff05 /*Productiva G100 Multi-Monitor*/
#define DEVICE_MATROX_MGA_G100_PRODUCTIVA2		0x1001 /*MGA G100 [Productiva] AGP*/
	#define SUBDEVICE_MATROX_MGA_G100_AGP		0x102b1001 /*MGA-G100 AGP*/
	#define SUBDEVICE_MATROX_MGA_G100_AGP2		0x102bff00 /*MGA-G100 AGP*/
	#define SUBDEVICE_MATROX_MGA_G100_PRODUCTIVA_AGP	0x102bff01 /*MGA-G100 Productiva AGP*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G100_AGP	0x102bff03 /*Millennium G100 AGP*/
	#define SUBDEVICE_MATROX_MGA_G100_AGP3		0x102bff04 /*MGA-G100 AGP*/
	#define SUBDEVICE_MATROX_MGA_G100_PRODUCTIVA_AGP2	0x102bff05 /*MGA-G100 Productiva AGP Multi-Monitor*/
	#define SUBDEVICE_MATROX_MGA_G100_AGP4		0x110a001e /*MGA-G100 AGP*/
#define DEVICE_MATROX_MGA_MISTRAL			0x2007 /*MGA Mistral*/
#define DEVICE_MATROX_MGA_G550_AGP			0x2527 /*MGA G550 AGP*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G550	0x102b0f83 /*Millennium G550*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G550_DUAL	0x102b0f84 /*Millennium G550 Dual Head DDR 32Mb*/
	#define SUBDEVICE_MATROX_MILLENNIUM_G5502	0x102b1e41 /*Millennium G550*/
#define DEVICE_MATROX_MGA_G650_AGP			0x2537 /*MGA G650 AGP*/
#define DEVICE_MATROX_VIA_FRAMEGRABBER			0x4536 /*VIA Framegrabber*/
#define DEVICE_MATROX_SHARK_10_100_MULTIPORT		0x6573 /*Shark 10/100 Multiport SwitchNIC*/
/* Vendor: 102c: Chips and Technologies */
#define DEVICE_CHIPS_F64310				0x00b8 /*F64310*/
#define DEVICE_CHIPS_F69000_HIQVIDEO			0x00c0 /*F69000 HiQVideo*/
	#define SUBDEVICE_CHIPS_F69000_HIQVIDEO		0x102c00c0 /*F69000 HiQVideo*/
	#define SUBDEVICE_CHIPS_CC7_CR7_CP7_VC7_VP7_VR7_MAINBOARD	0x4c531000 /*CC7/CR7/CP7/VC7/VP7/VR7 mainboard*/
	#define SUBDEVICE_CHIPS_CP5_CR6_MAINBOARD	0x4c531010 /*CP5/CR6 mainboard*/
	#define SUBDEVICE_CHIPS_VR6_MAINBOARD		0x4c531020 /*VR6 mainboard*/
	#define SUBDEVICE_CHIPS_PC5_MAINBOARD		0x4c531030 /*PC5 mainboard*/
	#define SUBDEVICE_CHIPS_CT7_MAINBOARD		0x4c531050 /*CT7 mainboard*/
	#define SUBDEVICE_CHIPS_CE7_MAINBOARD		0x4c531051 /*CE7 mainboard*/
#define DEVICE_CHIPS_F65545				0x00d0 /*F65545*/
#define DEVICE_CHIPS_F655452				0x00d8 /*F65545*/
#define DEVICE_CHIPS_F65548				0x00dc /*F65548*/
#define DEVICE_CHIPS_F65550				0x00e0 /*F65550*/
#define DEVICE_CHIPS_F65554				0x00e4 /*F65554*/
#define DEVICE_CHIPS_F65555_HIQVPRO			0x00e5 /*F65555 HiQVPro*/
	#define SUBDEVICE_CHIPS_ARMADA_1700_LAPTOP	0x0e11b049 /*Armada 1700 Laptop Display Controller*/
#define DEVICE_CHIPS_F68554				0x00f0 /*F68554*/
#define DEVICE_CHIPS_F68554_HIQVISION			0x00f4 /*F68554 HiQVision*/
#define DEVICE_CHIPS_F68555				0x00f5 /*F68555*/
#define DEVICE_CHIPS_F69030				0x0c30 /*F69030*/
	#define SUBDEVICE_CHIPS_CC7_CR7_CP7_VC7_VP7_VR7_MAINBOARD2	0x4c531000 /*CC7/CR7/CP7/VC7/VP7/VR7 mainboard*/
	#define SUBDEVICE_CHIPS_CT7_MAINBOARD2		0x4c531050 /*CT7 mainboard*/
	#define SUBDEVICE_CHIPS_CE7_MAINBOARD2		0x4c531051 /*CE7 mainboard*/
	#define SUBDEVICE_CHIPS_CT8_MAINBOARD		0x4c531080 /*CT8 mainboard*/
/* Vendor: 102d: Wyse Technology Inc. */
#define DEVICE_WYSE_3328_AUDIO				0x50dc /*3328 Audio*/
/* Vendor: 102e: Olivetti Advanced Technology */
/* Vendor: 102f: Toshiba America */
#define DEVICE_TOSHIBA_R4X00				0x0009 /*r4x00*/
#define DEVICE_TOSHIBA_TX3927_MIPS_RISC			0x000a /*TX3927 MIPS RISC PCI Controller*/
#define DEVICE_TOSHIBA_ATM_METEOR_155			0x0020 /*ATM Meteor 155*/
	#define SUBDEVICE_TOSHIBA_ATM_METEOR_155	0x102f00f8 /*ATM Meteor 155*/
#define DEVICE_TOSHIBA_TC35815CF_PCI_10_100		0x0030 /*TC35815CF PCI 10/100 Mbit Ethernet Controller*/
#define DEVICE_TOSHIBA_TC35815CF_PCI_10_1002		0x0031 /*TC35815CF PCI 10/100 Mbit Ethernet Controller with WOL*/
#define DEVICE_TOSHIBA_TC86C001_GOKU_S_IDE		0x0105 /*TC86C001 [goku-s] IDE*/
#define DEVICE_TOSHIBA_TC86C001_GOKU_S_USB		0x0106 /*TC86C001 [goku-s] USB 1.1 Host*/
#define DEVICE_TOSHIBA_TC86C001_GOKU_S_USB2		0x0107 /*TC86C001 [goku-s] USB Device Controller*/
#define DEVICE_TOSHIBA_TC86C001_GOKU_S_I2C_SIO_GPIO	0x0108 /*TC86C001 [goku-s] I2C/SIO/GPIO Controller*/
#define DEVICE_TOSHIBA_TX4927_38_MIPS_RISC		0x0180 /*TX4927/38 MIPS RISC PCI Controller*/
#define DEVICE_TOSHIBA_TX4925_MIPS_RISC			0x0181 /*TX4925 MIPS RISC PCI Controller*/
#define DEVICE_TOSHIBA_TX4937_MIPS_RISC			0x0182 /*TX4937 MIPS RISC PCI Controller*/
/* Vendor: 1030: TMC Research */
/* Vendor: 1031: Miro Computer Products AG */
#define DEVICE_MIRO_DC20_ASIC				0x5601 /*DC20 ASIC*/
#define DEVICE_MIRO_VIDEO_I_O				0x5607 /*Video I/O & motion JPEG compressor*/
#define DEVICE_MIRO_MEDIA_3D				0x5631 /*Media 3D*/
#define DEVICE_MIRO_MIROVIDEO_DC10_DC30			0x6057 /*MiroVideo DC10/DC30+*/
/* Vendor: 1032: Compaq */
/* Vendor: 1033: NEC Corporation */
#define DEVICE_NEC_VR4181A_USB_HOST			0x0000 /*Vr4181A USB Host or Function Control Unit*/
#define DEVICE_NEC_PCI_TO_486_LIKE			0x0001 /*PCI to 486-like bus Bridge*/
#define DEVICE_NEC_PCI_TO_VL98				0x0002 /*PCI to VL98 Bridge*/
#define DEVICE_NEC_ATM_CONTROLLER			0x0003 /*ATM Controller*/
#define DEVICE_NEC_R4000_PCI_BRIDGE			0x0004 /*R4000 PCI Bridge*/
#define DEVICE_NEC_PCI_TO_486_LIKE2			0x0005 /*PCI to 486-like bus Bridge*/
#define DEVICE_NEC_PC_9800_GRAPHIC_ACCELERATOR		0x0006 /*PC-9800 Graphic Accelerator*/
#define DEVICE_NEC_PCI_TO_UX_BUS			0x0007 /*PCI to UX-Bus Bridge*/
#define DEVICE_NEC_PC_9800_GRAPHIC_ACCELERATOR2		0x0008 /*PC-9800 Graphic Accelerator*/
#define DEVICE_NEC_PCI_TO_PC9800			0x0009 /*PCI to PC9800 Core-Graph Bridge*/
#define DEVICE_NEC_PCI_TO_VL				0x0016 /*PCI to VL Bridge*/
#define DEVICE_NEC__NILE_II				0x001a /*[Nile II]*/
#define DEVICE_NEC_VRC4373_NILE_I			0x0021 /*Vrc4373 [Nile I]*/
#define DEVICE_NEC_POWERVR_PCX1				0x0029 /*PowerVR PCX1*/
#define DEVICE_NEC_POWERVR_3D				0x002a /*PowerVR 3D*/
#define DEVICE_NEC_STAR_ALPHA_2				0x002c /*Star Alpha 2*/
#define DEVICE_NEC_PCI_TO_C_BUS				0x002d /*PCI to C-bus Bridge*/
#define DEVICE_NEC_USB					0x0035 /*USB*/
	#define SUBDEVICE_NEC_USB			0x11790001 /*USB*/
	#define SUBDEVICE_NEC_ROOT_HUB			0x12ee7000 /*Root Hub*/
	#define SUBDEVICE_NEC_ROOT_HUB2			0x17990001 /*Root Hub*/
	#define SUBDEVICE_NEC_PCI_USB2_OHCI_SUBSYSTEM	0x807d0035 /*PCI-USB2 (OHCI subsystem)*/
#define DEVICE_NEC_PCI_TO_C_BUS2			0x003b /*PCI to C-bus Bridge*/
#define DEVICE_NEC_NAPCCARD_CARDBUS_CONTROLLER		0x003e /*NAPCCARD Cardbus Controller*/
#define DEVICE_NEC_POWERVR_PCX2_MIDAS			0x0046 /*PowerVR PCX2 [midas]*/
#define DEVICE_NEC_VRC5074_NILE_4			0x005a /*Vrc5074 [Nile 4]*/
#define DEVICE_NEC_FIREWARDEN				0x0063 /*Firewarden*/
#define DEVICE_NEC_POWERVR_NEON_250			0x0067 /*PowerVR Neon 250 Chipset*/
	#define SUBDEVICE_NEC_POWERVR_NEON_250		0x10100020 /*PowerVR Neon 250 AGP 32Mb*/
	#define SUBDEVICE_NEC_POWERVR_NEON_2502		0x10100080 /*PowerVR Neon 250 AGP 16Mb*/
	#define SUBDEVICE_NEC_POWERVR_NEON_2503		0x10100088 /*PowerVR Neon 250 16Mb*/
	#define SUBDEVICE_NEC_POWERVR_NEON_2504		0x10100090 /*PowerVR Neon 250 AGP 16Mb*/
	#define SUBDEVICE_NEC_POWERVR_NEON_2505		0x10100098 /*PowerVR Neon 250 16Mb*/
	#define SUBDEVICE_NEC_POWERVR_NEON_2506		0x101000a0 /*PowerVR Neon 250 AGP 32Mb*/
	#define SUBDEVICE_NEC_POWERVR_NEON_2507		0x101000a8 /*PowerVR Neon 250 32Mb*/
	#define SUBDEVICE_NEC_POWERVR_NEON_2508		0x10100120 /*PowerVR Neon 250 AGP 32Mb*/
#define DEVICE_NEC_56K_VOICE_MODEM			0x0074 /*56k Voice Modem*/
	#define SUBDEVICE_NEC_RCV56ACF_56K_VOICE	0x10338014 /*RCV56ACF 56k Voice Modem*/
#define DEVICE_NEC_VRC5476				0x009b /*Vrc5476*/
#define DEVICE_NEC_VRC4173				0x00a5 /*VRC4173*/
#define DEVICE_NEC_VRC5477_AC97				0x00a6 /*VRC5477 AC97*/
#define DEVICE_NEC_IEEE_1394_ORANGELINK			0x00cd /*IEEE 1394 [OrangeLink] Host Controller*/
	#define SUBDEVICE_NEC_ROOT_HUB3			0x12ee8011 /*Root hub*/
#define DEVICE_NEC_IEEE_1394_HOST			0x00ce /*IEEE 1394 Host Controller*/
#define DEVICE_NEC_VR4131				0x00df /*Vr4131*/
#define DEVICE_NEC_USB_2_0				0x00e0 /*USB 2.0*/
	#define SUBDEVICE_NEC_SITECOM_IEEE_1394		0x0ee43383 /*Sitecom IEEE 1394 / USB2.0 Combo Card*/
	#define SUBDEVICE_NEC_ROOT_HUB4			0x12ee7001 /*Root hub*/
	#define SUBDEVICE_NEC_ROOT_HUB5			0x17990002 /*Root Hub*/
	#define SUBDEVICE_NEC_PCI_USB2_EHCI_SUBSYSTEM	0x807d1043 /*PCI-USB2 (EHCI subsystem)*/
#define DEVICE_NEC_IEEE_1394_HOST2			0x00e7 /*IEEE 1394 Host Controller*/
#define DEVICE_NEC_UPD72874_IEEE1394_OHCI		0x00f2 /*uPD72874 IEEE1394 OHCI 1.1 3-port PHY-Link Ctrlr*/
#define DEVICE_NEC_UPD6113X_MULTIMEDIA_DECODER_PROCESSOR	0x00f3 /*uPD6113x Multimedia Decoder/Processor [EMMA2]*/
#define DEVICE_NEC_VR7701				0x010c /*VR7701*/
/* Vendor: 1034: Framatome Connectors USA Inc. */
/* Vendor: 1035: Comp. & Comm. Research Lab */
/* Vendor: 1036: Future Domain Corp. */
#define DEVICE_FUTURE_TMC_18C30_36C70			0x0000 /*TMC-18C30 [36C70]*/
/* Vendor: 1037: Hitachi Micro Systems */
/* Vendor: 1038: AMP, Inc */
/* Vendor: 1039: Silicon Integrated Systems [SiS] */
#define DEVICE_SILICON2_VIRTUAL_PCI_TO_PCI_BRIDGE	0x0001 /*Virtual PCI-to-PCI bridge (AGP)*/
#define DEVICE_SILICON2_SG86C202			0x0002 /*SG86C202*/
#define DEVICE_SILICON2_85C501_2_3			0x0006 /*85C501/2/3*/
#define DEVICE_SILICON2_SIS85C503_5513_LPC_BRIDGE	0x0008 /*SiS85C503/5513 (LPC Bridge)*/
#define DEVICE_SILICON2_ACPI				0x0009 /*ACPI*/
#define DEVICE_SILICON2_SIS961_2_SMBUS_CONTROLLER	0x0016 /*SiS961/2 SMBus Controller*/
#define DEVICE_SILICON2_SIS85C503_5513_LPC_BRIDGE2	0x0018 /*SiS85C503/5513 (LPC Bridge)*/
#define DEVICE_SILICON2_RAID_BUS_CONTROLLER		0x0180 /*RAID bus controller 180 SATA/PATA  [SiS]*/
#define DEVICE_SILICON2_SIS_SATA			0x0181 /*SiS SATA*/
#define DEVICE_SILICON2_5597_5598_6326_VGA		0x0200 /*5597/5598/6326 VGA*/
	#define SUBDEVICE_SILICON2_SIS5597_SVGA_SHARED	0x10390000 /*SiS5597 SVGA (Shared RAM)*/
#define DEVICE_SILICON2_82C204				0x0204 /*82C204*/
#define DEVICE_SILICON2_SG86C205			0x0205 /*SG86C205*/
#define DEVICE_SILICON2_300_305_PCI_AGP_VGA		0x0300 /*300/305 PCI/AGP VGA Display Adapter*/
	#define SUBDEVICE_SILICON2_LEADTEK_WINFAST_VR300	0x107d2720 /*Leadtek WinFast VR300*/
#define DEVICE_SILICON2_315H_PCI_AGP_VGA		0x0310 /*315H PCI/AGP VGA Display Adapter*/
#define DEVICE_SILICON2_315_PCI_AGP_VGA			0x0315 /*315 PCI/AGP VGA Display Adapter*/
#define DEVICE_SILICON2_315PRO_PCI_AGP_VGA		0x0325 /*315PRO PCI/AGP VGA Display Adapter*/
#define DEVICE_SILICON2_330_XABRE_PCI_AGP		0x0330 /*330 [Xabre] PCI/AGP VGA Display Adapter*/
#define DEVICE_SILICON2_85C501_2			0x0406 /*85C501/2*/
#define DEVICE_SILICON2_85C496				0x0496 /*85C496*/
#define DEVICE_SILICON2_530_HOST			0x0530 /*530 Host*/
#define DEVICE_SILICON2_540_HOST			0x0540 /*540 Host*/
#define DEVICE_SILICON2_550_HOST			0x0550 /*550 Host*/
#define DEVICE_SILICON2_5513C				0x0597 /*5513C*/
#define DEVICE_SILICON2_85C601				0x0601 /*85C601*/
#define DEVICE_SILICON2_620_HOST			0x0620 /*620 Host*/
#define DEVICE_SILICON2_630_HOST			0x0630 /*630 Host*/
#define DEVICE_SILICON2_633_HOST			0x0633 /*633 Host*/
#define DEVICE_SILICON2_635_HOST			0x0635 /*635 Host*/
#define DEVICE_SILICON2_SIS645_HOST			0x0645 /*SiS645 Host & Memory & AGP Controller*/
#define DEVICE_SILICON2_SIS645DX_HOST			0x0646 /*SiS645DX Host & Memory & AGP Controller*/
#define DEVICE_SILICON2_SIS_645XX			0x0648 /*SiS 645xx*/
#define DEVICE_SILICON2_650_M650_HOST			0x0650 /*650/M650 Host*/
#define DEVICE_SILICON2_651_HOST			0x0651 /*651 Host*/
#define DEVICE_SILICON2_655_HOST			0x0655 /*655 Host*/
#define DEVICE_SILICON2_660_HOST			0x0660 /*660 Host*/
#define DEVICE_SILICON2_661FX_M661FX_M661MX_HOST	0x0661 /*661FX/M661FX/M661MX Host*/
#define DEVICE_SILICON2_730_HOST			0x0730 /*730 Host*/
#define DEVICE_SILICON2_733_HOST			0x0733 /*733 Host*/
#define DEVICE_SILICON2_735_HOST			0x0735 /*735 Host*/
#define DEVICE_SILICON2_740_HOST			0x0740 /*740 Host*/
#define DEVICE_SILICON2_741_741GX_M741_HOST		0x0741 /*741/741GX/M741 Host*/
#define DEVICE_SILICON2_745_HOST			0x0745 /*745 Host*/
#define DEVICE_SILICON2_746_HOST			0x0746 /*746 Host*/
#define DEVICE_SILICON2_755_HOST			0x0755 /*755 Host*/
#define DEVICE_SILICON2_760_M760_HOST			0x0760 /*760/M760 Host*/
#define DEVICE_SILICON2_SIS900_PCI_FAST			0x0900 /*SiS900 PCI Fast Ethernet*/
	#define SUBDEVICE_SILICON2_K7S5A_MOTHERBOARD	0x10190a14 /*K7S5A motherboard*/
	#define SUBDEVICE_SILICON2_SIS900_10_100_ETHERNET	0x10390900 /*SiS900 10/100 Ethernet Adapter*/
	#define SUBDEVICE_SILICON2_CUSI_FX_MOTHERBOARD	0x10438035 /*CUSI-FX motherboard*/
#define DEVICE_SILICON2_SIS961_MUTIOL_MEDIA		0x0961 /*SiS961 [MuTIOL Media IO]*/
#define DEVICE_SILICON2_SIS962_MUTIOL_MEDIA		0x0962 /*SiS962 [MuTIOL Media IO]*/
#define DEVICE_SILICON2_SIS963_MUTIOL_MEDIA		0x0963 /*SiS963 [MuTIOL Media IO]*/
#define DEVICE_SILICON2_SIS964_MUTIOL_MEDIA		0x0964 /*SiS964 [MuTIOL Media IO]*/
#define DEVICE_SILICON2_SIS965_MUTIOL_MEDIA		0x0965 /*SiS965 [MuTIOL Media IO]*/
#define DEVICE_SILICON2_83C602				0x3602 /*83C602*/
#define DEVICE_SILICON2_5107				0x5107 /*5107*/
#define DEVICE_SILICON2_SIS540_PCI_DISPLAY		0x5300 /*SiS540 PCI Display Adapter*/
#define DEVICE_SILICON2_550_PCI_AGP_VGA			0x5315 /*550 PCI/AGP VGA Display Adapter*/
#define DEVICE_SILICON2_486_PCI_CHIPSET			0x5401 /*486 PCI Chipset*/
#define DEVICE_SILICON2_5511_5512			0x5511 /*5511/5512*/
#define DEVICE_SILICON2_5513_IDE			0x5513 /*5513 [IDE]*/
	#define SUBDEVICE_SILICON2_P6STP_FL_MOTHERBOARD	0x10190970 /*P6STP-FL motherboard*/
	#define SUBDEVICE_SILICON2_SIS5513_EIDE_CONTROLLER	0x10395513 /*SiS5513 EIDE Controller (A,B step)*/
	#define SUBDEVICE_SILICON2_CUSI_FX_MOTHERBOARD2	0x10438035 /*CUSI-FX motherboard*/
#define DEVICE_SILICON2_5517				0x5517 /*5517*/
#define DEVICE_SILICON2_5571				0x5571 /*5571*/
#define DEVICE_SILICON2_5581_PENTIUM_CHIPSET		0x5581 /*5581 Pentium Chipset*/
#define DEVICE_SILICON2_5582				0x5582 /*5582*/
#define DEVICE_SILICON2_5591_5592_HOST			0x5591 /*5591/5592 Host*/
#define DEVICE_SILICON2_5596_PENTIUM_CHIPSET		0x5596 /*5596 Pentium Chipset*/
#define DEVICE_SILICON2_5597_SIS5582			0x5597 /*5597 [SiS5582]*/
#define DEVICE_SILICON2_5600_HOST			0x5600 /*5600 Host*/
#define DEVICE_SILICON2_VIDEO_DECODER			0x6204 /*Video decoder & MPEG interface*/
#define DEVICE_SILICON2_VGA_CONTROLLER			0x6205 /*VGA Controller*/
#define DEVICE_SILICON2_6236_3D_AGP			0x6236 /*6236 3D-AGP*/
#define DEVICE_SILICON2_630_730_PCI_AGP_VGA		0x6300 /*630/730 PCI/AGP VGA Display Adapter*/
	#define SUBDEVICE_SILICON2_P6STP_FL_MOTHERBOARD2	0x10190970 /*P6STP-FL motherboard*/
	#define SUBDEVICE_SILICON2_CUSI_FX_MOTHERBOARD3	0x10438035 /*CUSI-FX motherboard*/
#define DEVICE_SILICON2_530_620_PCI_AGP_VGA		0x6306 /*530/620 PCI/AGP VGA Display Adapter*/
	#define SUBDEVICE_SILICON2_SIS530_620_GUI_ACCELERATOR_3D	0x10396306 /*SiS530,620 GUI Accelerator+3D*/
#define DEVICE_SILICON2_65X_M650_740_PCI_AGP_VGA	0x6325 /*65x/M650/740 PCI/AGP VGA Display Adapter*/
#define DEVICE_SILICON2_86C326_5598_6326		0x6326 /*86C326 5598/6326*/
	#define SUBDEVICE_SILICON2_SIS6326_GUI_ACCELERATOR	0x10396326 /*SiS6326 GUI Accelerator*/
	#define SUBDEVICE_SILICON2_SPEEDSTAR_A50	0x10920a50 /*SpeedStar A50*/
	#define SUBDEVICE_SILICON2_SPEEDSTAR_A70	0x10920a70 /*SpeedStar A70*/
	#define SUBDEVICE_SILICON2_SPEEDSTAR_A702	0x10924910 /*SpeedStar A70*/
	#define SUBDEVICE_SILICON2_SPEEDSTAR_A703	0x10924920 /*SpeedStar A70*/
	#define SUBDEVICE_SILICON2_SIS6326_GUI_ACCELERATOR2	0x15696326 /*SiS6326 GUI Accelerator*/
#define DEVICE_SILICON2_661_741_760_PCI_AGP_VGA		0x6330 /*661/741/760 PCI/AGP VGA Display Adapter*/
	#define SUBDEVICE_SILICON2__M_661XX_M_741_GX_M_760_PCI_AGP_VGA	0x10396330 /*[M]661xX/[M]741[GX]/[M]760 PCI/AGP VGA Adapter*/
#define DEVICE_SILICON2_USB_1_0_CONTROLLER		0x7001 /*USB 1.0 Controller*/
	#define SUBDEVICE_SILICON2_K7S5A_MOTHERBOARD2	0x10190a14 /*K7S5A motherboard*/
	#define SUBDEVICE_SILICON2_ONBOARD_USB_CONTROLLER	0x10397000 /*Onboard USB Controller*/
#define DEVICE_SILICON2_USB_2_0_CONTROLLER		0x7002 /*USB 2.0 Controller*/
	#define SUBDEVICE_SILICON2_ONBOARD_USB_CONTROLLER2	0x15097002 /*Onboard USB Controller*/
#define DEVICE_SILICON2_FIREWIRE_CONTROLLER		0x7007 /*FireWire Controller*/
#define DEVICE_SILICON2_SOUND_CONTROLLER		0x7012 /*Sound Controller*/
#define DEVICE_SILICON2_AC_97_MODEM_CONTROLLER		0x7013 /*AC'97 Modem Controller*/
#define DEVICE_SILICON2_SIS7016_PCI_FAST		0x7016 /*SiS7016 PCI Fast Ethernet Adapter*/
	#define SUBDEVICE_SILICON2_SIS7016_10_100_ETHERNET	0x10397016 /*SiS7016 10/100 Ethernet Adapter*/
#define DEVICE_SILICON2_SIS_PCI_AUDIO			0x7018 /*SiS PCI Audio Accelerator*/
	#define SUBDEVICE_SILICON2_SIS_PCI_AUDIO	0x101401b6 /*SiS PCI Audio Accelerator*/
	#define SUBDEVICE_SILICON2_SIS_PCI_AUDIO2	0x101401b7 /*SiS PCI Audio Accelerator*/
	#define SUBDEVICE_SILICON2_SIS_PCI_AUDIO3	0x10197018 /*SiS PCI Audio Accelerator*/
	#define SUBDEVICE_SILICON2_SIS_PCI_AUDIO4	0x1025000e /*SiS PCI Audio Accelerator*/
	#define SUBDEVICE_SILICON2_SIS_PCI_AUDIO5	0x10250018 /*SiS PCI Audio Accelerator*/
	#define SUBDEVICE_SILICON2_SIS_PCI_AUDIO6	0x10397018 /*SiS PCI Audio Accelerator*/
	#define SUBDEVICE_SILICON2_SIS_PCI_AUDIO7	0x1043800b /*SiS PCI Audio Accelerator*/
	#define SUBDEVICE_SILICON2_SIS_PCI_AUDIO8	0x10547018 /*SiS PCI Audio Accelerator*/
	#define SUBDEVICE_SILICON2_SIS_PCI_AUDIO9	0x107d5330 /*SiS PCI Audio Accelerator*/
	#define SUBDEVICE_SILICON2_SIS_PCI_AUDIO10	0x107d5350 /*SiS PCI Audio Accelerator*/
	#define SUBDEVICE_SILICON2_SIS_PCI_AUDIO11	0x11703209 /*SiS PCI Audio Accelerator*/
	#define SUBDEVICE_SILICON2_SIS_PCI_AUDIO12	0x1462400a /*SiS PCI Audio Accelerator*/
	#define SUBDEVICE_SILICON2_SIS_PCI_AUDIO13	0x14a42089 /*SiS PCI Audio Accelerator*/
	#define SUBDEVICE_SILICON2_SIS_PCI_AUDIO14	0x14cd2194 /*SiS PCI Audio Accelerator*/
	#define SUBDEVICE_SILICON2_SIS_PCI_AUDIO15	0x14ff1100 /*SiS PCI Audio Accelerator*/
	#define SUBDEVICE_SILICON2_SIS_PCI_AUDIO16	0x152d8808 /*SiS PCI Audio Accelerator*/
	#define SUBDEVICE_SILICON2_SIS_PCI_AUDIO17	0x15581103 /*SiS PCI Audio Accelerator*/
	#define SUBDEVICE_SILICON2_SIS_PCI_AUDIO18	0x15582200 /*SiS PCI Audio Accelerator*/
	#define SUBDEVICE_SILICON2_SIS_PCI_AUDIO19	0x15637018 /*SiS PCI Audio Accelerator*/
	#define SUBDEVICE_SILICON2_SIS_PCI_AUDIO20	0x15c50111 /*SiS PCI Audio Accelerator*/
	#define SUBDEVICE_SILICON2_SIS_PCI_AUDIO21	0x270fa171 /*SiS PCI Audio Accelerator*/
	#define SUBDEVICE_SILICON2_SIS_PCI_AUDIO22	0xa0a00022 /*SiS PCI Audio Accelerator*/
#define DEVICE_SILICON2_SIS7019_AUDIO_ACCELERATOR	0x7019 /*SiS7019 Audio Accelerator*/
/* Vendor: 103a: Seiko Epson Corporation */
/* Vendor: 103b: Tatung Co. of America */
/* Vendor: 103c: Hewlett-Packard Company */
#define DEVICE_HEWLETT_PACKARD_A4977A_VISUALIZE_EG	0x1005 /*A4977A Visualize EG*/
#define DEVICE_HEWLETT_PACKARD_VISUALIZE_FX6		0x1006 /*Visualize FX6*/
#define DEVICE_HEWLETT_PACKARD_VISUALIZE_FX4		0x1008 /*Visualize FX4*/
#define DEVICE_HEWLETT_PACKARD_VISUALIZE_FX2		0x100a /*Visualize FX2*/
#define DEVICE_HEWLETT_PACKARD_TACH_TL_FIBRE		0x1028 /*Tach TL Fibre Channel Host Adapter*/
#define DEVICE_HEWLETT_PACKARD_TACH_XL2_FIBRE		0x1029 /*Tach XL2 Fibre Channel Host Adapter*/
	#define SUBDEVICE_HEWLETT_PACKARD_INTERPHASE_5560_FIBRE	0x107e000f /*Interphase 5560 Fibre Channel Adapter*/
	#define SUBDEVICE_HEWLETT_PACKARD_1GB_2GB_FAMILY_FIBRE	0x90049210 /*1Gb/2Gb Family Fibre Channel Controller*/
	#define SUBDEVICE_HEWLETT_PACKARD_1GB_2GB_FAMILY_FIBRE2	0x90049211 /*1Gb/2Gb Family Fibre Channel Controller*/
#define DEVICE_HEWLETT_PACKARD_TACH_TS_FIBRE		0x102a /*Tach TS Fibre Channel Host Adapter*/
	#define SUBDEVICE_HEWLETT_PACKARD_INTERPHASE_5540_5541_FIBRE	0x107e000e /*Interphase 5540/5541 Fibre Channel Adapter*/
	#define SUBDEVICE_HEWLETT_PACKARD_1GB_2GB_FAMILY_FIBRE3	0x90049110 /*1Gb/2Gb Family Fibre Channel Controller*/
	#define SUBDEVICE_HEWLETT_PACKARD_1GB_2GB_FAMILY_FIBRE4	0x90049111 /*1Gb/2Gb Family Fibre Channel Controller*/
#define DEVICE_HEWLETT_PACKARD_J2585A_DESKDIRECT_10_100VG	0x1030 /*J2585A DeskDirect 10/100VG NIC*/
#define DEVICE_HEWLETT_PACKARD_J2585B_HP_10_100VG	0x1031 /*J2585B HP 10/100VG PCI LAN Adapter*/
	#define SUBDEVICE_HEWLETT_PACKARD_J2973A_DESKDIRECT_10BASET	0x103c1040 /*J2973A DeskDirect 10BaseT NIC*/
	#define SUBDEVICE_HEWLETT_PACKARD_J2585B_DESKDIRECT_10_100VG	0x103c1041 /*J2585B DeskDirect 10/100VG NIC*/
	#define SUBDEVICE_HEWLETT_PACKARD_J2970A_DESKDIRECT_10BASET_2	0x103c1042 /*J2970A DeskDirect 10BaseT/2 NIC*/
#define DEVICE_HEWLETT_PACKARD_J2973A_DESKDIRECT_10BASET	0x1040 /*J2973A DeskDirect 10BaseT NIC*/
#define DEVICE_HEWLETT_PACKARD_J2585B_DESKDIRECT_10_100	0x1041 /*J2585B DeskDirect 10/100 NIC*/
#define DEVICE_HEWLETT_PACKARD_J2970A_DESKDIRECT_10BASET_2	0x1042 /*J2970A DeskDirect 10BaseT/2 NIC*/
#define DEVICE_HEWLETT_PACKARD_DIVA_SERIAL_GSP		0x1048 /*Diva Serial [GSP] Multiport UART*/
	#define SUBDEVICE_HEWLETT_PACKARD_TOSCA_CONSOLE	0x103c1049 /*Tosca Console*/
	#define SUBDEVICE_HEWLETT_PACKARD_TOSCA_SECONDARY	0x103c104a /*Tosca Secondary*/
	#define SUBDEVICE_HEWLETT_PACKARD_MAESTRO_SP2	0x103c104b /*Maestro SP2*/
	#define SUBDEVICE_HEWLETT_PACKARD_SUPERDOME_CONSOLE	0x103c1223 /*Superdome Console*/
	#define SUBDEVICE_HEWLETT_PACKARD_KEYSTONE_SP2	0x103c1226 /*Keystone SP2*/
	#define SUBDEVICE_HEWLETT_PACKARD_POWERBAR_SP2	0x103c1227 /*Powerbar SP2*/
	#define SUBDEVICE_HEWLETT_PACKARD_EVEREST_SP2	0x103c1282 /*Everest SP2*/
#define DEVICE_HEWLETT_PACKARD_PCI_LOCAL_BUS		0x1054 /*PCI Local Bus Adapter*/
#define DEVICE_HEWLETT_PACKARD_79C970_PCNET_ETHERNET	0x1064 /*79C970 PCnet Ethernet Controller*/
#define DEVICE_HEWLETT_PACKARD_VISUALIZE_FXE		0x108b /*Visualize FXe*/
#define DEVICE_HEWLETT_PACKARD_NETSERVER_SMART_IRQ	0x10c1 /*NetServer Smart IRQ Router*/
#define DEVICE_HEWLETT_PACKARD_TOPTOOLS_REMOTE_CONTROL	0x10ed /*TopTools Remote Control*/
#define DEVICE_HEWLETT_PACKARD_82557B_10_100_NIC	0x1200 /*82557B 10/100 NIC*/
#define DEVICE_HEWLETT_PACKARD_NETSERVER_PCI_HOT_PLUG	0x1219 /*NetServer PCI Hot-Plug Controller*/
#define DEVICE_HEWLETT_PACKARD_NETSERVER_SMIC_CONTROLLER	0x121a /*NetServer SMIC Controller*/
#define DEVICE_HEWLETT_PACKARD_NETSERVER_LEGACY_COM	0x121b /*NetServer Legacy COM Port Decoder*/
#define DEVICE_HEWLETT_PACKARD_NETSERVER_PCI_COM	0x121c /*NetServer PCI COM Port Decoder*/
#define DEVICE_HEWLETT_PACKARD_ZX1_SYSTEM_BUS		0x1229 /*zx1 System Bus Adapter*/
#define DEVICE_HEWLETT_PACKARD_ZX1_I_O_CONTROLLER	0x122a /*zx1 I/O Controller*/
#define DEVICE_HEWLETT_PACKARD_ZX1_LOCAL_BUS		0x122e /*zx1 Local Bus Adapter*/
#define DEVICE_HEWLETT_PACKARD_SX1000_I_O_CONTROLLER	0x127c /*sx1000 I/O Controller*/
#define DEVICE_HEWLETT_PACKARD_AUXILIARY_DIVA_SERIAL	0x1290 /*Auxiliary Diva Serial Port*/
#define DEVICE_HEWLETT_PACKARD_E2910A_PCIBUS_EXERCISER	0x2910 /*E2910A PCIBus Exerciser*/
#define DEVICE_HEWLETT_PACKARD_E2925A_32_BIT		0x2925 /*E2925A 32 Bit, 33 MHzPCI Exerciser & Analyzer*/
/* Vendor: 103e: Solliday Engineering */
/* Vendor: 103f: Synopsys/Logic Modeling Group */
/* Vendor: 1040: Accelgraphics Inc. */
/* Vendor: 1041: Computrend */
/* Vendor: 1042: Micron */
#define DEVICE_MICRON_PC_TECH_RZ1000			0x1000 /*PC Tech RZ1000*/
#define DEVICE_MICRON_PC_TECH_RZ1001			0x1001 /*PC Tech RZ1001*/
#define DEVICE_MICRON_SAMURAI_0				0x3000 /*Samurai_0*/
#define DEVICE_MICRON_SAMURAI_1				0x3010 /*Samurai_1*/
#define DEVICE_MICRON_SAMURAI_IDE			0x3020 /*Samurai_IDE*/
/* Vendor: 1043: ASUSTeK Computer Inc. */
#define DEVICE_ASUSTEK_ISDNLINK_P_IN100_ST_D		0x0675 /*ISDNLink P-IN100-ST-D*/
#define DEVICE_ASUSTEK_V7100_SDRAM_GEFORCE2		0x4015 /*v7100 SDRAM [GeForce2 MX]*/
#define DEVICE_ASUSTEK_V7100_COMBO_DELUXE		0x4021 /*v7100 Combo Deluxe [GeForce2 MX + TV tuner]*/
#define DEVICE_ASUSTEK_V8200_GEFORCE_3			0x4057 /*v8200 GeForce 3*/
#define DEVICE_ASUSTEK_V8240_PAL_128M			0x8043 /*v8240 PAL 128M [P4T] Motherboard*/
#define DEVICE_ASUSTEK_V9280_TD_GEFORCE4_TI4200		0x807b /*v9280/TD [Geforce4 TI4200 8X With TV-Out and DVI]*/
#define DEVICE_ASUSTEK_V9180_MAGIC_T_GEFORCE4		0x80bb /*v9180 Magic/T [GeForce4 MX440 AGP 8x 64MB TV-out]*/
/* Vendor: 1044: Adaptec (formerly DPT) */
#define DEVICE_ADAPTEC_DOMINO_RAID_ENGINE		0x1012 /*Domino RAID Engine*/
#define DEVICE_ADAPTEC_SMARTCACHE_RAID_I_IV_CONTROLLER	0xa400 /*SmartCache/Raid I-IV Controller*/
#define DEVICE_ADAPTEC_PCI_BRIDGE			0xa500 /*PCI Bridge*/
#define DEVICE_ADAPTEC_SMARTRAID_V_CONTROLLER		0xa501 /*SmartRAID V Controller*/
	#define SUBDEVICE_ADAPTEC_PM1554U2_ULTRA2_SINGLE	0x1044c001 /*PM1554U2 Ultra2 Single Channel*/
	#define SUBDEVICE_ADAPTEC_PM1654U2_ULTRA2_SINGLE	0x1044c002 /*PM1654U2 Ultra2 Single Channel*/
	#define SUBDEVICE_ADAPTEC_PM1564U3_ULTRA3_SINGLE	0x1044c003 /*PM1564U3 Ultra3 Single Channel*/
	#define SUBDEVICE_ADAPTEC_PM1564U3_ULTRA3_DUAL	0x1044c004 /*PM1564U3 Ultra3 Dual Channel*/
	#define SUBDEVICE_ADAPTEC_PM1554U2_ULTRA2_SINGLE2	0x1044c005 /*PM1554U2 Ultra2 Single Channel (NON ACPI)*/
	#define SUBDEVICE_ADAPTEC_PM2554U2_ULTRA2_SINGLE	0x1044c00a /*PM2554U2 Ultra2 Single Channel*/
	#define SUBDEVICE_ADAPTEC_PM2654U2_ULTRA2_SINGLE	0x1044c00b /*PM2654U2 Ultra2 Single Channel*/
	#define SUBDEVICE_ADAPTEC_PM2664U3_ULTRA3_SINGLE	0x1044c00c /*PM2664U3 Ultra3 Single Channel*/
	#define SUBDEVICE_ADAPTEC_PM2664U3_ULTRA3_DUAL	0x1044c00d /*PM2664U3 Ultra3 Dual Channel*/
	#define SUBDEVICE_ADAPTEC_PM2554U2_ULTRA2_SINGLE2	0x1044c00e /*PM2554U2 Ultra2 Single Channel (NON ACPI)*/
	#define SUBDEVICE_ADAPTEC_PM2654U2_ULTRA2_SINGLE2	0x1044c00f /*PM2654U2 Ultra2 Single Channel (NON ACPI)*/
	#define SUBDEVICE_ADAPTEC_PM3754U2_ULTRA2_SINGLE	0x1044c014 /*PM3754U2 Ultra2 Single Channel (NON ACPI)*/
	#define SUBDEVICE_ADAPTEC_PM3755U2B_ULTRA2_SINGLE	0x1044c015 /*PM3755U2B Ultra2 Single Channel (NON ACPI)*/
	#define SUBDEVICE_ADAPTEC_PM3755F_FIBRE_CHANNEL	0x1044c016 /*PM3755F Fibre Channel (NON ACPI)*/
	#define SUBDEVICE_ADAPTEC_PM3757U2_ULTRA2_SINGLE	0x1044c01e /*PM3757U2 Ultra2 Single Channel*/
	#define SUBDEVICE_ADAPTEC_PM3757U2_ULTRA2_DUAL	0x1044c01f /*PM3757U2 Ultra2 Dual Channel*/
	#define SUBDEVICE_ADAPTEC_PM3767U3_ULTRA3_DUAL	0x1044c020 /*PM3767U3 Ultra3 Dual Channel*/
	#define SUBDEVICE_ADAPTEC_PM3767U3_ULTRA3_QUAD	0x1044c021 /*PM3767U3 Ultra3 Quad Channel*/
	#define SUBDEVICE_ADAPTEC_PM2865U3_ULTRA3_SINGLE	0x1044c028 /*PM2865U3 Ultra3 Single Channel*/
	#define SUBDEVICE_ADAPTEC_PM2865U3_ULTRA3_DUAL	0x1044c029 /*PM2865U3 Ultra3 Dual Channel*/
	#define SUBDEVICE_ADAPTEC_PM2865F_FIBRE_CHANNEL	0x1044c02a /*PM2865F Fibre Channel*/
	#define SUBDEVICE_ADAPTEC_2000S_ULTRA3_SINGLE	0x1044c03c /*2000S Ultra3 Single Channel*/
	#define SUBDEVICE_ADAPTEC_2000S_ULTRA3_DUAL	0x1044c03d /*2000S Ultra3 Dual Channel*/
	#define SUBDEVICE_ADAPTEC_2000F_FIBRE_CHANNEL	0x1044c03e /*2000F Fibre Channel*/
	#define SUBDEVICE_ADAPTEC_3000S_ULTRA3_SINGLE	0x1044c046 /*3000S Ultra3 Single Channel*/
	#define SUBDEVICE_ADAPTEC_3000S_ULTRA3_DUAL	0x1044c047 /*3000S Ultra3 Dual Channel*/
	#define SUBDEVICE_ADAPTEC_3000F_FIBRE_CHANNEL	0x1044c048 /*3000F Fibre Channel*/
	#define SUBDEVICE_ADAPTEC_5000S_ULTRA3_SINGLE	0x1044c050 /*5000S Ultra3 Single Channel*/
	#define SUBDEVICE_ADAPTEC_5000S_ULTRA3_DUAL	0x1044c051 /*5000S Ultra3 Dual Channel*/
	#define SUBDEVICE_ADAPTEC_5000F_FIBRE_CHANNEL	0x1044c052 /*5000F Fibre Channel*/
	#define SUBDEVICE_ADAPTEC_2400A_UDMA_FOUR	0x1044c05a /*2400A UDMA Four Channel*/
	#define SUBDEVICE_ADAPTEC_2400A_UDMA_FOUR2	0x1044c05b /*2400A UDMA Four Channel DAC*/
	#define SUBDEVICE_ADAPTEC_3010S_ULTRA3_DUAL	0x1044c064 /*3010S Ultra3 Dual Channel*/
	#define SUBDEVICE_ADAPTEC_3010S_ULTRA3_FOUR	0x1044c065 /*3010S Ultra3 Four Channel*/
	#define SUBDEVICE_ADAPTEC_3010S_FIBRE_CHANNEL	0x1044c066 /*3010S Fibre Channel*/
#define DEVICE_ADAPTEC_SMARTRAID_V_CONTROLLER2		0xa511 /*SmartRAID V Controller*/
	#define SUBDEVICE_ADAPTEC_ASR_2005S_I2O_ZERO	0x1044c032 /*ASR-2005S I2O Zero Channel*/
/* Vendor: 1045: OPTi Inc. */
#define DEVICE_OPTI_82C750_VENDETTA_USB			0xa0f8 /*82C750 [Vendetta] USB Controller*/
#define DEVICE_OPTI_92C264				0xc101 /*92C264*/
#define DEVICE_OPTI_92C178				0xc178 /*92C178*/
#define DEVICE_OPTI_82X556_VIPER			0xc556 /*82X556 [Viper]*/
#define DEVICE_OPTI_82C557_VIPER_M			0xc557 /*82C557 [Viper-M]*/
#define DEVICE_OPTI_82C558_VIPER_M_ISA_IDE		0xc558 /*82C558 [Viper-M ISA+IDE]*/
#define DEVICE_OPTI_82C750_VENDETTA_DEVICE		0xc567 /*82C750 [Vendetta], device 0*/
#define DEVICE_OPTI_82C750_VENDETTA_DEVICE2		0xc568 /*82C750 [Vendetta], device 1*/
#define DEVICE_OPTI_82C579_VIPER_XPRESS			0xc569 /*82C579 [Viper XPress+ Chipset]*/
#define DEVICE_OPTI_82C621_VIPER_M_N			0xc621 /*82C621 [Viper-M/N+]*/
#define DEVICE_OPTI_82C700_FIRESTAR			0xc700 /*82C700 [FireStar]*/
#define DEVICE_OPTI_82C701_FIRESTAR_PLUS		0xc701 /*82C701 [FireStar Plus]*/
#define DEVICE_OPTI_82C814_FIREBRIDGE_1			0xc814 /*82C814 [Firebridge 1]*/
#define DEVICE_OPTI_82C822				0xc822 /*82C822*/
#define DEVICE_OPTI_82C824				0xc824 /*82C824*/
#define DEVICE_OPTI_82C825_FIREBRIDGE_2			0xc825 /*82C825 [Firebridge 2]*/
#define DEVICE_OPTI_82C832				0xc832 /*82C832*/
#define DEVICE_OPTI_82C861				0xc861 /*82C861*/
#define DEVICE_OPTI_82C895				0xc895 /*82C895*/
#define DEVICE_OPTI_EV1935_ECTIVA_MACHONE		0xc935 /*EV1935 ECTIVA MachOne PCIAudio*/
#define DEVICE_OPTI_82C825_FIREBRIDGE_22		0xd568 /*82C825 [Firebridge 2]*/
#define DEVICE_OPTI_IDE_FIRESTAR			0xd721 /*IDE [FireStar]*/
/* Vendor: 1046: IPC Corporation, Ltd. */
/* Vendor: 1047: Genoa Systems Corp */
/* Vendor: 1048: Elsa AG */
#define DEVICE_ELSA_GLADIAC_MX				0x0c60 /*Gladiac MX*/
#define DEVICE_ELSA_QUADRO4_900XGL_ELSA			0x0d22 /*Quadro4 900XGL [ELSA GLoria4 900XGL]*/
#define DEVICE_ELSA_QUICKSTEP_1000			0x1000 /*QuickStep 1000*/
#define DEVICE_ELSA_QUICKSTEP_3000			0x3000 /*QuickStep 3000*/
#define DEVICE_ELSA_GLORIA_XL				0x8901 /*Gloria XL*/
/* Vendor: 1049: Fountain Technologies, Inc. */
/* Vendor: 104a: STMicroelectronics */
#define DEVICE_STMICROELECTRONICS_STG_2000X		0x0008 /*STG 2000X*/
#define DEVICE_STMICROELECTRONICS_STG_1764X		0x0009 /*STG 1764X*/
#define DEVICE_STMICROELECTRONICS_STG4000_3D_PROPHET	0x0010 /*STG4000 [3D Prophet Kyro Series]*/
#define DEVICE_STMICROELECTRONICS_STPC_CONSUMER_INDUSTRIAL_NORTH	0x0209 /*STPC Consumer/Industrial North- and Southbridge*/
#define DEVICE_STMICROELECTRONICS_STPC_ATLAS_CONSUMERS_CONSUMER_IIA	0x020a /*STPC Atlas/ConsumerS/Consumer IIA Northbridge*/
#define DEVICE_STMICROELECTRONICS_STPC_ATLAS_ISA	0x0210 /*STPC Atlas ISA Bridge*/
#define DEVICE_STMICROELECTRONICS_STPC_CONSUMER_S	0x021a /*STPC Consumer S Southbridge*/
#define DEVICE_STMICROELECTRONICS_STPC_CONSUMER_IIA	0x021b /*STPC Consumer IIA Southbridge*/
#define DEVICE_STMICROELECTRONICS_ST70137_UNICORN_ADSL	0x0500 /*ST70137 [Unicorn] ADSL DMT Transceiver*/
#define DEVICE_STMICROELECTRONICS_STPC_CLIENT_NORTHBRIDGE	0x0564 /*STPC Client Northbridge*/
#define DEVICE_STMICROELECTRONICS_21X4X_DEC_TULIP_COMPATIBLE	0x0981 /*21x4x DEC-Tulip compatible 10/100 Ethernet*/
#define DEVICE_STMICROELECTRONICS_STG_1764X2		0x1746 /*STG 1764X*/
#define DEVICE_STMICROELECTRONICS_21X4X_DEC_TULIP_COMPATIBLE2	0x2774 /*21x4x DEC-Tulip compatible 10/100 Ethernet*/
#define DEVICE_STMICROELECTRONICS_MPEG_II_DECODER_CARD	0x3520 /*MPEG-II decoder card*/
#define DEVICE_STMICROELECTRONICS_STPC_CLIENT_SOUTHBRIDGE	0x55cc /*STPC Client Southbridge*/
/* Vendor: 104b: BusLogic */
#define DEVICE_BUSLOGIC_BT_946C_OLD_MULTIMASTER		0x0140 /*BT-946C (old) [multimaster  01]*/
#define DEVICE_BUSLOGIC_BT_946C_BA80C30_MULTIMASTER	0x1040 /*BT-946C (BA80C30) [MultiMaster 10]*/
#define DEVICE_BUSLOGIC_FLASHPOINT_LT			0x8130 /*Flashpoint LT*/
/* Vendor: 104c: Texas Instruments */
#define DEVICE_TEXAS_100_MBIT_LAN			0x0500 /*100 MBit LAN Controller*/
#define DEVICE_TEXAS_TMS380C2X_COMPRESSOR_INTERFACE	0x0508 /*TMS380C2X Compressor Interface*/
#define DEVICE_TEXAS_EAGLE_I_F_AS			0x1000 /*Eagle i/f AS*/
#define DEVICE_TEXAS_PCI1510_PC_CARD			0x104c /*PCI1510 PC card Cardbus Controller*/
#define DEVICE_TEXAS_TVP4010_PERMEDIA			0x3d04 /*TVP4010 [Permedia]*/
#define DEVICE_TEXAS_TVP4020_PERMEDIA_2			0x3d07 /*TVP4020 [Permedia 2]*/
	#define SUBDEVICE_TEXAS_COMET			0x10114d10 /*Comet*/
	#define SUBDEVICE_TEXAS_ACCELSTAR_II		0x1040000f /*AccelStar II*/
	#define SUBDEVICE_TEXAS_ACCELSTAR_II2		0x10400011 /*AccelStar II*/
	#define SUBDEVICE_TEXAS_WINNER_2000		0x10480a31 /*WINNER 2000*/
	#define SUBDEVICE_TEXAS_GLORIA_SYNERGY		0x10480a32 /*GLoria Synergy*/
	#define SUBDEVICE_TEXAS_GLORIA_SYNERGY2		0x10480a35 /*GLoria Synergy*/
	#define SUBDEVICE_TEXAS_WINFAST_3D_L2300	0x107d2633 /*WinFast 3D L2300*/
	#define SUBDEVICE_TEXAS_FIRE_GL_1000		0x10920127 /*FIRE GL 1000 PRO*/
	#define SUBDEVICE_TEXAS_FIRE_GL_10002		0x10920136 /*FIRE GL 1000 PRO*/
	#define SUBDEVICE_TEXAS_FIRE_GL_10003		0x10920141 /*FIRE GL 1000 PRO*/
	#define SUBDEVICE_TEXAS_FIRE_GL_10004		0x10920146 /*FIRE GL 1000 PRO*/
	#define SUBDEVICE_TEXAS_FIRE_GL_10005		0x10920148 /*FIRE GL 1000 PRO*/
	#define SUBDEVICE_TEXAS_FIRE_GL_10006		0x10920149 /*FIRE GL 1000 PRO*/
	#define SUBDEVICE_TEXAS_FIRE_GL_10007		0x10920152 /*FIRE GL 1000 PRO*/
	#define SUBDEVICE_TEXAS_FIRE_GL_10008		0x10920154 /*FIRE GL 1000 PRO*/
	#define SUBDEVICE_TEXAS_FIRE_GL_10009		0x10920155 /*FIRE GL 1000 PRO*/
	#define SUBDEVICE_TEXAS_FIRE_GL_100010		0x10920156 /*FIRE GL 1000 PRO*/
	#define SUBDEVICE_TEXAS_FIRE_GL_100011		0x10920157 /*FIRE GL 1000 PRO*/
	#define SUBDEVICE_TEXAS_JERONIMO_PRO		0x10973d01 /*Jeronimo Pro*/
	#define SUBDEVICE_TEXAS_GRAPHICS_BLASTER_EXTREME	0x1102100f /*Graphics Blaster Extreme*/
	#define SUBDEVICE_TEXAS_REFERENCE_PERMEDIA_2	0x3d3d0100 /*Reference Permedia 2 3D*/
#define DEVICE_TEXAS_PCILYNX_PCILYNX2_IEEE_1394		0x8000 /*PCILynx/PCILynx2 IEEE 1394 Link Layer Controller*/
	#define SUBDEVICE_TEXAS_CF1_1_SNARE		0xe4bf1010 /*CF1-1-SNARE*/
	#define SUBDEVICE_TEXAS_CF1_2_SNARE		0xe4bf1020 /*CF1-2-SNARE*/
#define DEVICE_TEXAS_FIREWIRE_CONTROLLER		0x8009 /*FireWire Controller*/
	#define SUBDEVICE_TEXAS_8032_OHCI_I_LINK	0x104d8032 /*8032 OHCI i.LINK (IEEE 1394) Controller*/
#define DEVICE_TEXAS_PCI4410_FIREWIRE_CONTROLLER	0x8017 /*PCI4410 FireWire Controller*/
#define DEVICE_TEXAS_TSB12LV23_IEEE_1394_CONTROLLER	0x8019 /*TSB12LV23 IEEE-1394 Controller*/
	#define SUBDEVICE_TEXAS_STUDIO_DV500_1394	0x11bd000a /*Studio DV500-1394*/
	#define SUBDEVICE_TEXAS_STUDIO_DV		0x11bd000e /*Studio DV*/
	#define SUBDEVICE_TEXAS_CF2_1_CYMBAL		0xe4bf1010 /*CF2-1-CYMBAL*/
#define DEVICE_TEXAS_TSB12LV26_IEEE_1394_CONTROLLER	0x8020 /*TSB12LV26 IEEE-1394 Controller (Link)*/
#define DEVICE_TEXAS_TSB43AA22_IEEE_1394_CONTROLLER	0x8021 /*TSB43AA22 IEEE-1394 Controller (PHY/Link Integrated)*/
	#define SUBDEVICE_TEXAS_VAIO_PCG_FX403		0x104d80df /*Vaio PCG-FX403*/
	#define SUBDEVICE_TEXAS_VAIO_PCG_GR214EP_GR214MP_GR215MP_GR314MP_GR315MP	0x104d80e7 /*VAIO PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP*/
#define DEVICE_TEXAS_TSB43AB22_IEEE_1394A_2000_CONTROLLER	0x8022 /*TSB43AB22 IEEE-1394a-2000 Controller (PHY/Link)*/
#define DEVICE_TEXAS_TSB43AB22_A_IEEE_1394A_2000_CONTROLLER	0x8023 /*TSB43AB22/A IEEE-1394a-2000 Controller (PHY/Link)*/
#define DEVICE_TEXAS_TSB43AB23_IEEE_1394A_2000_CONTROLLER	0x8024 /*TSB43AB23 IEEE-1394a-2000 Controller (PHY/Link)*/
#define DEVICE_TEXAS_TSB82AA2_IEEE_1394B_LINK		0x8025 /*TSB82AA2 IEEE-1394b Link Layer Controller*/
	#define SUBDEVICE_TEXAS_FIREWIRE_800_PCI	0x55aa55aa /*FireWire 800 PCI Card*/
#define DEVICE_TEXAS_TSB43AB21_IEEE_1394A_2000_CONTROLLER	0x8026 /*TSB43AB21 IEEE-1394a-2000 Controller (PHY/Link)*/
#define DEVICE_TEXAS_PCI4451_IEEE_1394_CONTROLLER	0x8027 /*PCI4451 IEEE-1394 Controller*/
	#define SUBDEVICE_TEXAS_PCI4451_IEEE_1394_CONTROLLER	0x102800e6 /*PCI4451 IEEE-1394 Controller (Dell Inspiron 8100)*/
#define DEVICE_TEXAS_PCI4510_IEEE_1394_CONTROLLER	0x8029 /*PCI4510 IEEE-1394 Controller*/
	#define SUBDEVICE_TEXAS_LATITUDE_D505		0x10280163 /*Latitude D505*/
	#define SUBDEVICE_TEXAS_MIM2900			0x10718160 /*MIM2900*/
#define DEVICE_TEXAS_PCI7X20_1394A_2000_OHCI		0x802e /*PCI7x20 1394a-2000 OHCI Two-Port PHY/Link-Layer Controller*/
#define DEVICE_TEXAS_PCI1620_FIRMWARE_LOADING		0x8201 /*PCI1620 Firmware Loading Function*/
#define DEVICE_TEXAS_ACX_100_22MBPS			0x8400 /*ACX 100 22Mbps Wireless Interface*/
	#define SUBDEVICE_TEXAS_U_S_ROBOTICS_22		0x00fc16ec /*U.S. Robotics 22 Mbps Wireless PC Card (model 2210)*/
	#define SUBDEVICE_TEXAS_U_S_ROBOTICS_22MBPS	0x00fd16ec /*U.S. Robotics 22Mbps Wireless PCI Adapter (model 2216)*/
	#define SUBDEVICE_TEXAS_DWL_650_PC_CARD		0x11863b00 /*DWL-650+ PC Card cardbus 22Mbs Wireless Adapter [AirPlus]*/
	#define SUBDEVICE_TEXAS_DWL_520_22MBPS_PCI	0x11863b01 /*DWL-520+ 22Mbps PCI Wireless Adapter*/
#define DEVICE_TEXAS_ACX_100_22MBPS2			0x8401 /*ACX 100 22Mbps Wireless Interface*/
#define DEVICE_TEXAS_WIRELESS_INTERFACE_OF		0x9000 /*Wireless Interface (of unknown type)*/
#define DEVICE_TEXAS_ACX_111_54MBPS			0x9066 /*ACX 111 54Mbps Wireless Interface*/
#define DEVICE_TEXAS_TDC1570				0xa001 /*TDC1570*/
#define DEVICE_TEXAS_TDC1561				0xa100 /*TDC1561*/
#define DEVICE_TEXAS_TNETA1575_HYPERSAR_PLUS		0xa102 /*TNETA1575 HyperSAR Plus w/PCI Host i/f & UTOPIA i/f*/
#define DEVICE_TEXAS_TMS320C6205_FIXED_POINT		0xa106 /*TMS320C6205 Fixed Point DSP*/
	#define SUBDEVICE_TEXAS_ASI50XX_AUDIO_ADAPTER	0x175c5000 /*ASI50xx Audio Adapter*/
	#define SUBDEVICE_TEXAS_ASI87XX_RADIO_TUNER	0x175c8700 /*ASI87xx Radio Tuner card*/
#define DEVICE_TEXAS_PCI1050				0xac10 /*PCI1050*/
#define DEVICE_TEXAS_PCI1053				0xac11 /*PCI1053*/
#define DEVICE_TEXAS_PCI1130				0xac12 /*PCI1130*/
#define DEVICE_TEXAS_PCI1031				0xac13 /*PCI1031*/
#define DEVICE_TEXAS_PCI1131				0xac15 /*PCI1131*/
#define DEVICE_TEXAS_PCI1250				0xac16 /*PCI1250*/
	#define SUBDEVICE_TEXAS_THINKPAD_600		0x10140092 /*ThinkPad 600*/
#define DEVICE_TEXAS_PCI1220				0xac17 /*PCI1220*/
#define DEVICE_TEXAS_PCI1260				0xac18 /*PCI1260*/
#define DEVICE_TEXAS_PCI1221				0xac19 /*PCI1221*/
#define DEVICE_TEXAS_PCI1210				0xac1a /*PCI1210*/
#define DEVICE_TEXAS_PCI1450				0xac1b /*PCI1450*/
	#define SUBDEVICE_TEXAS_ARMADA_M700		0x0e11b113 /*Armada M700*/
#define DEVICE_TEXAS_PCI1225				0xac1c /*PCI1225*/
	#define SUBDEVICE_TEXAS_ARMADA_E500		0x0e11b121 /*Armada E500*/
	#define SUBDEVICE_TEXAS_DELL_COMPUTER		0x10280088 /*Dell  Computer Corporation  Latitude CPi A400XT*/
#define DEVICE_TEXAS_PCI1251A				0xac1d /*PCI1251A*/
#define DEVICE_TEXAS_PCI1211				0xac1e /*PCI1211*/
#define DEVICE_TEXAS_PCI1251B				0xac1f /*PCI1251B*/
#define DEVICE_TEXAS_TI_2030				0xac20 /*TI 2030*/
#define DEVICE_TEXAS_PCI2031				0xac21 /*PCI2031*/
#define DEVICE_TEXAS_PCI2032_PCI_DOCKING		0xac22 /*PCI2032 PCI Docking Bridge*/
#define DEVICE_TEXAS_PCI2250_PCI_TO_PCI_BRIDGE		0xac23 /*PCI2250 PCI-to-PCI Bridge*/
#define DEVICE_TEXAS_PCI2050_PCI_TO_PCI_BRIDGE		0xac28 /*PCI2050 PCI-to-PCI Bridge*/
#define DEVICE_TEXAS_PCI1260_PC_CARD			0xac30 /*PCI1260 PC card Cardbus Controller*/
#define DEVICE_TEXAS_PCI4450_PC_CARD			0xac40 /*PCI4450 PC card Cardbus Controller*/
#define DEVICE_TEXAS_PCI4410_PC_CARD			0xac41 /*PCI4410 PC card Cardbus Controller*/
#define DEVICE_TEXAS_PCI4451_PC_CARD			0xac42 /*PCI4451 PC card Cardbus Controller*/
	#define SUBDEVICE_TEXAS_PCI4451_PC_CARD		0x102800e6 /*PCI4451 PC card CardBus Controller (Dell Inspiron 8100)*/
#define DEVICE_TEXAS_PCI4510_PC_CARD			0xac44 /*PCI4510 PC card Cardbus Controller*/
	#define SUBDEVICE_TEXAS_LATITUDE_D5052		0x10280163 /*Latitude D505*/
	#define SUBDEVICE_TEXAS_MIM2000			0x10718160 /*MIM2000*/
#define DEVICE_TEXAS_PCI4520_PC_CARD			0xac46 /*PCI4520 PC card Cardbus Controller*/
#define DEVICE_TEXAS_PCI1410_PC_CARD			0xac50 /*PCI1410 PC card Cardbus Controller*/
#define DEVICE_TEXAS_PCI1420				0xac51 /*PCI1420*/
	#define SUBDEVICE_TEXAS_THINKPAD_T23_2647_4MG	0x1014023b /*ThinkPad T23 (2647-4MG)*/
	#define SUBDEVICE_TEXAS_LATITUDE_C600		0x102800b1 /*Latitude C600*/
	#define SUBDEVICE_TEXAS_LATITUDE_C640		0x1028012a /*Latitude C640*/
	#define SUBDEVICE_TEXAS_VERSA_NOTE_VXI		0x103380cd /*Versa Note VXi*/
	#define SUBDEVICE_TEXAS_LIFEBOOK_C6155		0x10cf1095 /*Lifebook C6155*/
	#define SUBDEVICE_TEXAS_CP2_2_HIPHOP		0xe4bf1000 /*CP2-2-HIPHOP*/
#define DEVICE_TEXAS_PCI1451_PC_CARD			0xac52 /*PCI1451 PC card Cardbus Controller*/
#define DEVICE_TEXAS_PCI1421_PC_CARD			0xac53 /*PCI1421 PC card Cardbus Controller*/
#define DEVICE_TEXAS_PCI1620_PC_CARD			0xac54 /*PCI1620 PC Card Controller*/
#define DEVICE_TEXAS_PCI1520_PC_CARD			0xac55 /*PCI1520 PC card Cardbus Controller*/
	#define SUBDEVICE_TEXAS_THINKPAD_T30_T40	0x10140512 /*ThinkPad T30/T40*/
#define DEVICE_TEXAS_PCI1510_PC_CARD2			0xac56 /*PCI1510 PC card Cardbus Controller*/
	#define SUBDEVICE_TEXAS_THINKPAD_R40E_2684_HVG	0x10140528 /*ThinkPad R40e (2684-HVG) Cardbus Controller*/
#define DEVICE_TEXAS_PCI2040_PCI_TO			0xac60 /*PCI2040 PCI to DSP Bridge Controller*/
	#define SUBDEVICE_TEXAS_ASI51XX_AUDIO_ADAPTER	0x175c5100 /*ASI51xx Audio Adapter*/
	#define SUBDEVICE_TEXAS_ASI61XX_AUDIO_ADAPTER	0x175c6100 /*ASI61xx Audio Adapter*/
	#define SUBDEVICE_TEXAS_ASI62XX_AUDIO_ADAPTER	0x175c6200 /*ASI62xx Audio Adapter*/
#define DEVICE_TEXAS_PCI_7620				0xac8d /*PCI 7620*/
#define DEVICE_TEXAS_PCI7420_CARDBUS_CONTROLLER		0xac8e /*PCI7420 CardBus Controller*/
#define DEVICE_TEXAS_PCI7420_PCI7620_DUAL_SOCKET	0xac8f /*PCI7420/PCI7620 Dual Socket CardBus and Smart Card Cont. w/ 1394a-2000 OHCI Two-Port  PHY/Link-Layer Cont. and SD/MS-Pro Sockets*/
#define DEVICE_TEXAS_FIREWIRE_HOST_CONTROLLER		0xfe00 /*FireWire Host Controller*/
#define DEVICE_TEXAS_12C01A_FIREWIRE_HOST		0xfe03 /*12C01A FireWire Host Controller*/
/* Vendor: 104d: Sony Corporation */
#define DEVICE_SONY_CXD1947Q_I_LINK_CONTROLLER		0x8009 /*CXD1947Q i.LINK Controller*/
#define DEVICE_SONY_CXD3222_I_LINK_CONTROLLER		0x8039 /*CXD3222 i.LINK Controller*/
#define DEVICE_SONY_ROCKWELL_HCF_56K			0x8056 /*Rockwell HCF 56K modem*/
#define DEVICE_SONY_MEMORY_STICK_CONTROLLER		0x808a /*Memory Stick Controller*/
/* Vendor: 104e: Oak Technology, Inc */
#define DEVICE_OAK_OTI_64017				0x0017 /*OTI-64017*/
#define DEVICE_OAK_OTI_107_SPITFIRE			0x0107 /*OTI-107 [Spitfire]*/
#define DEVICE_OAK_VIDEO_ADAPTER			0x0109 /*Video Adapter*/
#define DEVICE_OAK_OTI_64111_SPITFIRE			0x0111 /*OTI-64111 [Spitfire]*/
#define DEVICE_OAK_OTI_64217				0x0217 /*OTI-64217*/
#define DEVICE_OAK_OTI_64317				0x0317 /*OTI-64317*/
/* Vendor: 104f: Co-time Computer Ltd */
/* Vendor: 1050: Winbond Electronics Corp */
#define DEVICE_WINBOND_NE2000				0x0000 /*NE2000*/
#define DEVICE_WINBOND_W83769F				0x0001 /*W83769F*/
#define DEVICE_WINBOND_W82C105				0x0105 /*W82C105*/
#define DEVICE_WINBOND_W89C840				0x0840 /*W89C840*/
	#define SUBDEVICE_WINBOND_W89C840_ETHERNET_ADAPTER	0x10500001 /*W89C840 Ethernet Adapter*/
	#define SUBDEVICE_WINBOND_W89C840_ETHERNET_ADAPTER2	0x10500840 /*W89C840 Ethernet Adapter*/
#define DEVICE_WINBOND_W89C940				0x0940 /*W89C940*/
#define DEVICE_WINBOND_W89C940F				0x5a5a /*W89C940F*/
#define DEVICE_WINBOND_W6692				0x6692 /*W6692*/
#define DEVICE_WINBOND_W99200F_MPEG_1_VIDEO		0x9921 /*W99200F MPEG-1 Video Encoder*/
#define DEVICE_WINBOND_W99200F_W9922PF_MPEG_1_2_VIDEO	0x9922 /*W99200F/W9922PF MPEG-1/2 Video Encoder*/
#define DEVICE_WINBOND_W9970CF				0x9970 /*W9970CF*/
/* Vendor: 1051: Anigma, Inc. */
/* Vendor: 1052: ?Young Micro Systems */
/* Vendor: 1053: Young Micro Systems */
/* Vendor: 1054: Hitachi, Ltd */
/* Vendor: 1055: Efar Microsystems */
#define DEVICE_EFAR_SLC90E66_VICTORY66_IDE		0x9130 /*SLC90E66 [Victory66] IDE*/
#define DEVICE_EFAR_SLC90E66_VICTORY66_ISA		0x9460 /*SLC90E66 [Victory66] ISA*/
#define DEVICE_EFAR_SLC90E66_VICTORY66_USB		0x9462 /*SLC90E66 [Victory66] USB*/
#define DEVICE_EFAR_SLC90E66_VICTORY66_ACPI		0x9463 /*SLC90E66 [Victory66] ACPI*/
/* Vendor: 1056: ICL */
/* Vendor: 1057: Motorola */
#define DEVICE_MOTOROLA_MPC105_EAGLE			0x0001 /*MPC105 [Eagle]*/
#define DEVICE_MOTOROLA_MPC106_GRACKLE			0x0002 /*MPC106 [Grackle]*/
#define DEVICE_MOTOROLA_MPC8240_KAHLUA			0x0003 /*MPC8240 [Kahlua]*/
#define DEVICE_MOTOROLA_MPC107				0x0004 /*MPC107*/
#define DEVICE_MOTOROLA_MPC8245_UNITY			0x0006 /*MPC8245 [Unity]*/
#define DEVICE_MOTOROLA_MPC8540				0x0008 /*MPC8540*/
#define DEVICE_MOTOROLA_MPC8560				0x0009 /*MPC8560*/
#define DEVICE_MOTOROLA_MC145575_HFC_PCI		0x0100 /*MC145575 [HFC-PCI]*/
#define DEVICE_MOTOROLA_KTI829C_100VG			0x0431 /*KTI829c 100VG*/
#define DEVICE_MOTOROLA_DSP56301_DIGITAL_SIGNAL		0x1801 /*DSP56301 Digital Signal Processor*/
	#define SUBDEVICE_MOTOROLA_TRANSAS_RADAR_IMITATOR	0x14fb0101 /*Transas Radar Imitator Board [RIM]*/
	#define SUBDEVICE_MOTOROLA_TRANSAS_RADAR_IMITATOR2	0x14fb0102 /*Transas Radar Imitator Board [RIM-2]*/
	#define SUBDEVICE_MOTOROLA_TRANSAS_RADAR_INTEGRATOR	0x14fb0202 /*Transas Radar Integrator Board [RIB-2]*/
	#define SUBDEVICE_MOTOROLA_1_CHANNEL_CAN	0x14fb0611 /*1 channel CAN bus Controller [CanPci-1]*/
	#define SUBDEVICE_MOTOROLA_2_CHANNELS_CAN	0x14fb0612 /*2 channels CAN bus Controller [CanPci-2]*/
	#define SUBDEVICE_MOTOROLA_3_CHANNELS_CAN	0x14fb0613 /*3 channels CAN bus Controller [CanPci-3]*/
	#define SUBDEVICE_MOTOROLA_4_CHANNELS_CAN	0x14fb0614 /*4 channels CAN bus Controller [CanPci-4]*/
	#define SUBDEVICE_MOTOROLA_1_CHANNEL_CAN2	0x14fb0621 /*1 channel CAN bus Controller [CanPci2-1]*/
	#define SUBDEVICE_MOTOROLA_2_CHANNELS_CAN2	0x14fb0622 /*2 channels CAN bus Controller [CanPci2-2]*/
	#define SUBDEVICE_MOTOROLA_TRANSAS_VTS_RADAR	0x14fb0810 /*Transas VTS Radar Integrator Board [RIB-4]*/
	#define SUBDEVICE_MOTOROLA_ASI4215_AUDIO_ADAPTER	0x175c4200 /*ASI4215 Audio Adapter*/
	#define SUBDEVICE_MOTOROLA_ASI43XX_AUDIO_ADAPTER	0x175c4300 /*ASI43xx Audio Adapter*/
	#define SUBDEVICE_MOTOROLA_ASI4401_AUDIO_ADAPTER	0x175c4400 /*ASI4401 Audio Adapter*/
	#define SUBDEVICE_MOTOROLA_LAYLA		0xecc00030 /*Layla*/
#define DEVICE_MOTOROLA_MPC8265A_MPC8266		0x18c0 /*MPC8265A/MPC8266*/
#define DEVICE_MOTOROLA_MPC8271_MPC8272			0x18c1 /*MPC8271/MPC8272*/
#define DEVICE_MOTOROLA_RAVEN				0x4801 /*Raven*/
#define DEVICE_MOTOROLA_FALCON				0x4802 /*Falcon*/
#define DEVICE_MOTOROLA_HAWK				0x4803 /*Hawk*/
#define DEVICE_MOTOROLA_CPX8216				0x4806 /*CPX8216*/
#define DEVICE_MOTOROLA_20268				0x4d68 /*20268*/
#define DEVICE_MOTOROLA_SM56_PCI_MODEM			0x5600 /*SM56 PCI Modem*/
	#define SUBDEVICE_MOTOROLA_SM56_PCI_SPEAKERPHONE	0x10570300 /*SM56 PCI Speakerphone Modem*/
	#define SUBDEVICE_MOTOROLA_SM56_PCI_VOICE	0x10570301 /*SM56 PCI Voice Modem*/
	#define SUBDEVICE_MOTOROLA_SM56_PCI_FAX		0x10570302 /*SM56 PCI Fax Modem*/
	#define SUBDEVICE_MOTOROLA_SM56_PCI_VOICE2	0x10575600 /*SM56 PCI Voice modem*/
	#define SUBDEVICE_MOTOROLA_SM56_PCI_SPEAKERPHONE2	0x13d20300 /*SM56 PCI Speakerphone Modem*/
	#define SUBDEVICE_MOTOROLA_SM56_PCI_VOICE3	0x13d20301 /*SM56 PCI Voice modem*/
	#define SUBDEVICE_MOTOROLA_SM56_PCI_FAX2	0x13d20302 /*SM56 PCI Fax Modem*/
	#define SUBDEVICE_MOTOROLA_SM56_PCI_SPEAKERPHONE3	0x14360300 /*SM56 PCI Speakerphone Modem*/
	#define SUBDEVICE_MOTOROLA_SM56_PCI_VOICE4	0x14360301 /*SM56 PCI Voice modem*/
	#define SUBDEVICE_MOTOROLA_SM56_PCI_FAX3	0x14360302 /*SM56 PCI Fax Modem*/
	#define SUBDEVICE_MOTOROLA_SM56_PCI_FAX4	0x144f100c /*SM56 PCI Fax Modem*/
	#define SUBDEVICE_MOTOROLA_SM56_PCI_SPEAKERPHONE4	0x14940300 /*SM56 PCI Speakerphone Modem*/
	#define SUBDEVICE_MOTOROLA_SM56_PCI_VOICE5	0x14940301 /*SM56 PCI Voice modem*/
	#define SUBDEVICE_MOTOROLA_SM56_PCI_SPEAKERPHONE5	0x14c80300 /*SM56 PCI Speakerphone Modem*/
	#define SUBDEVICE_MOTOROLA_SM56_PCI_FAX5	0x14c80302 /*SM56 PCI Fax Modem*/
	#define SUBDEVICE_MOTOROLA_SM56_PCI_SPEAKERPHONE6	0x16680300 /*SM56 PCI Speakerphone Modem*/
	#define SUBDEVICE_MOTOROLA_SM56_PCI_FAX6	0x16680302 /*SM56 PCI Fax Modem*/
#define DEVICE_MOTOROLA_MPC5200				0x5803 /*MPC5200*/
#define DEVICE_MOTOROLA_MPC190_SECURITY_PROCESSOR	0x6400 /*MPC190 Security Processor (S1 family, encryption)*/
#define DEVICE_MOTOROLA_MPC184_SECURITY_PROCESSOR	0x6405 /*MPC184 Security Processor (S1 family)*/
/* Vendor: 1058: Electronics & Telecommunications RSH */
/* Vendor: 1059: Teknor Industrial Computers Inc */
/* Vendor: 105a: Promise Technology, Inc. */
#define DEVICE_PROMISE_PDC20265_FASTTRAK100_LITE_ULTRA100	0x0d30 /*PDC20265 (FastTrak100 Lite/Ultra100)*/
	#define SUBDEVICE_PROMISE_ULTRA100		0x105a4d33 /*Ultra100*/
#define DEVICE_PROMISE_20263				0x0d38 /*20263*/
	#define SUBDEVICE_PROMISE_FASTTRAK66		0x105a4d39 /*Fasttrak66*/
#define DEVICE_PROMISE_20275				0x1275 /*20275*/
#define DEVICE_PROMISE_PDC20318_SATA150_TX4		0x3318 /*PDC20318 (SATA150 TX4)*/
#define DEVICE_PROMISE_PDC20319_FASTTRAK_S150		0x3319 /*PDC20319 (FastTrak S150 TX4)*/
	#define SUBDEVICE_PROMISE_S875WP1_E_MAINBOARD	0x80863427 /*S875WP1-E mainboard*/
#define DEVICE_PROMISE_PDC20371_FASTTRAK_S150		0x3371 /*PDC20371 (FastTrak S150 TX2plus)*/
#define DEVICE_PROMISE_PDC20378_FASTTRAK_378_SATA	0x3373 /*PDC20378 (FastTrak 378/SATA 378)*/
	#define SUBDEVICE_PROMISE_PC_DL_DELUXE_MOTHERBOARD	0x104380f5 /*PC-DL Deluxe motherboard*/
	#define SUBDEVICE_PROMISE_K8T_NEO_FIS2R		0x1462702e /*K8T NEO FIS2R motherboard*/
#define DEVICE_PROMISE_PDC20375_SATA150_TX2PLUS		0x3375 /*PDC20375 (SATA150 TX2plus)*/
#define DEVICE_PROMISE_PDC20376_FASTTRAK_376		0x3376 /*PDC20376 (FastTrak 376)*/
	#define SUBDEVICE_PROMISE_A7V8X_MOTHERBOARD	0x1043809e /*A7V8X motherboard*/
#define DEVICE_PROMISE_PDC20579_SATAII_150		0x3574 /*PDC20579 SATAII 150 IDE Controller*/
#define DEVICE_PROMISE_PDC20518_SATAII_150		0x3d18 /*PDC20518 SATAII 150 IDE Controller*/
#define DEVICE_PROMISE_PDC20267_FASTTRAK100_ULTRA100	0x4d30 /*PDC20267 (FastTrak100/Ultra100)*/
	#define SUBDEVICE_PROMISE_ULTRA1002		0x105a4d33 /*Ultra100*/
	#define SUBDEVICE_PROMISE_FASTTRAK100		0x105a4d39 /*FastTrak100*/
#define DEVICE_PROMISE_20246				0x4d33 /*20246*/
	#define SUBDEVICE_PROMISE_20246_IDE_CONTROLLER	0x105a4d33 /*20246 IDE Controller*/
#define DEVICE_PROMISE_PDC20262_FASTTRAK66_ULTRA66	0x4d38 /*PDC20262 (FastTrak66/Ultra66)*/
	#define SUBDEVICE_PROMISE_ULTRA_DEVICE_ON	0x105a4d30 /*Ultra Device on SuperTrak*/
	#define SUBDEVICE_PROMISE_ULTRA66		0x105a4d33 /*Ultra66*/
	#define SUBDEVICE_PROMISE_FASTTRAK662		0x105a4d39 /*FastTrak66*/
#define DEVICE_PROMISE_PDC20268_ULTRA100_TX2		0x4d68 /*PDC20268 (Ultra100 TX2)*/
	#define SUBDEVICE_PROMISE_ULTRA100TX2		0x105a4d68 /*Ultra100TX2*/
#define DEVICE_PROMISE_20269				0x4d69 /*20269*/
	#define SUBDEVICE_PROMISE_ULTRA133TX2		0x105a4d68 /*Ultra133TX2*/
#define DEVICE_PROMISE_PDC20276_MBFASTTRAK133_LITE	0x5275 /*PDC20276 (MBFastTrak133 Lite)*/
	#define SUBDEVICE_PROMISE_SUPERTRAK_SX6000_IDE	0x105a0275 /*SuperTrak SX6000 IDE*/
	#define SUBDEVICE_PROMISE_MBFASTTRAK133_LITE_TM	0x105a1275 /*MBFastTrak133 Lite (tm) Controller (RAID mode)*/
	#define SUBDEVICE_PROMISE_MBULTRA_133		0x1458b001 /*MBUltra 133*/
#define DEVICE_PROMISE_DC5300				0x5300 /*DC5300*/
#define DEVICE_PROMISE_PDC20270_FASTTRAK100_LP_TX2_TX4	0x6268 /*PDC20270 (FastTrak100 LP/TX2/TX4)*/
	#define SUBDEVICE_PROMISE_FASTTRAK100_TX2	0x105a4d68 /*FastTrak100 TX2*/
#define DEVICE_PROMISE_PDC20271_FASTTRAK_TX2000		0x6269 /*PDC20271 (FastTrak TX2000)*/
	#define SUBDEVICE_PROMISE_FASTTRAK_TX2_TX2000	0x105a6269 /*FastTrak TX2/TX2000*/
#define DEVICE_PROMISE_PDC20621_FASTTRAK_S150		0x6621 /*PDC20621 (FastTrak S150 SX4/FastTrak SX4000 lite)*/
#define DEVICE_PROMISE_PDC20621_SATA150_SX4		0x6622 /*PDC20621 [SATA150 SX4] 4 Channel IDE RAID Controller*/
#define DEVICE_PROMISE_PDC20618_ULTRA_618		0x6626 /*PDC20618 (Ultra 618)*/
#define DEVICE_PROMISE_PDC20619_FASTTRAK_TX4000		0x6629 /*PDC20619 (FastTrak TX4000)*/
#define DEVICE_PROMISE_PDC20277_SBFASTTRAK133_LITE	0x7275 /*PDC20277 (SBFastTrak133 Lite)*/
/* Vendor: 105b: Foxconn International, Inc. */
/* Vendor: 105c: Wipro Infotech Limited */
/* Vendor: 105d: Number 9 Computer Company */
#define DEVICE_NUMBER_IMAGINE_128			0x2309 /*Imagine 128*/
#define DEVICE_NUMBER_IMAGINE_128_II			0x2339 /*Imagine 128-II*/
	#define SUBDEVICE_NUMBER_IMAGINE_128_SERIES	0x105d0000 /*Imagine 128 series 2 4Mb VRAM*/
	#define SUBDEVICE_NUMBER_IMAGINE_128_SERIES2	0x105d0001 /*Imagine 128 series 2 4Mb VRAM*/
	#define SUBDEVICE_NUMBER_IMAGINE_128_SERIES3	0x105d0002 /*Imagine 128 series 2 4Mb VRAM*/
	#define SUBDEVICE_NUMBER_IMAGINE_128_SERIES4	0x105d0003 /*Imagine 128 series 2 4Mb VRAM*/
	#define SUBDEVICE_NUMBER_IMAGINE_128_SERIES5	0x105d0004 /*Imagine 128 series 2 4Mb VRAM*/
	#define SUBDEVICE_NUMBER_IMAGINE_128_SERIES6	0x105d0005 /*Imagine 128 series 2 4Mb VRAM*/
	#define SUBDEVICE_NUMBER_IMAGINE_128_SERIES7	0x105d0006 /*Imagine 128 series 2 4Mb VRAM*/
	#define SUBDEVICE_NUMBER_IMAGINE_128_SERIES8	0x105d0007 /*Imagine 128 series 2 4Mb VRAM*/
	#define SUBDEVICE_NUMBER_IMAGINE_128_SERIES9	0x105d0008 /*Imagine 128 series 2e 4Mb DRAM*/
	#define SUBDEVICE_NUMBER_IMAGINE_128_SERIES10	0x105d0009 /*Imagine 128 series 2e 4Mb DRAM*/
	#define SUBDEVICE_NUMBER_IMAGINE_128_SERIES11	0x105d000a /*Imagine 128 series 2 8Mb VRAM*/
	#define SUBDEVICE_NUMBER_IMAGINE_128_SERIES12	0x105d000b /*Imagine 128 series 2 8Mb H-VRAM*/
	#define SUBDEVICE_NUMBER_BARCO_METHEUS_5	0x11a4000a /*Barco Metheus 5 Megapixel*/
	#define SUBDEVICE_NUMBER_BARCO_METHEUS_52	0x13cc0000 /*Barco Metheus 5 Megapixel*/
	#define SUBDEVICE_NUMBER_BARCO_METHEUS_53	0x13cc0004 /*Barco Metheus 5 Megapixel*/
	#define SUBDEVICE_NUMBER_BARCO_METHEUS_54	0x13cc0005 /*Barco Metheus 5 Megapixel*/
	#define SUBDEVICE_NUMBER_BARCO_METHEUS_55	0x13cc0006 /*Barco Metheus 5 Megapixel*/
	#define SUBDEVICE_NUMBER_BARCO_METHEUS_56	0x13cc0008 /*Barco Metheus 5 Megapixel*/
	#define SUBDEVICE_NUMBER_BARCO_METHEUS_57	0x13cc0009 /*Barco Metheus 5 Megapixel*/
	#define SUBDEVICE_NUMBER_BARCO_METHEUS_58	0x13cc000a /*Barco Metheus 5 Megapixel*/
	#define SUBDEVICE_NUMBER_BARCO_METHEUS_59	0x13cc000c /*Barco Metheus 5 Megapixel*/
#define DEVICE_NUMBER_IMAGINE_128_T2R			0x493d /*Imagine 128 T2R [Ticket to Ride]*/
	#define SUBDEVICE_NUMBER_BARCO_METHEUS_510	0x11a4000a /*Barco Metheus 5 Megapixel, Dual Head*/
	#define SUBDEVICE_NUMBER_BARCO_METHEUS_511	0x11a4000b /*Barco Metheus 5 Megapixel, Dual Head*/
	#define SUBDEVICE_NUMBER_BARCO_METHEUS_4	0x13cc0002 /*Barco Metheus 4 Megapixel, Dual Head*/
	#define SUBDEVICE_NUMBER_BARCO_METHEUS_512	0x13cc0003 /*Barco Metheus 5 Megapixel, Dual Head*/
	#define SUBDEVICE_NUMBER_BARCO_METHEUS_513	0x13cc0007 /*Barco Metheus 5 Megapixel, Dual Head*/
	#define SUBDEVICE_NUMBER_BARCO_METHEUS_514	0x13cc0008 /*Barco Metheus 5 Megapixel, Dual Head*/
	#define SUBDEVICE_NUMBER_BARCO_METHEUS_515	0x13cc0009 /*Barco Metheus 5 Megapixel, Dual Head*/
	#define SUBDEVICE_NUMBER_BARCO_METHEUS_516	0x13cc000a /*Barco Metheus 5 Megapixel, Dual Head*/
#define DEVICE_NUMBER_REVOLUTION_4			0x5348 /*Revolution 4*/
	#define SUBDEVICE_NUMBER_REVOLUTION_IV_FP_AGP	0x105d0037 /*Revolution IV-FP AGP (For SGI 1600SW)*/
/* Vendor: 105e: Vtech Computers Ltd */
/* Vendor: 105f: Infotronic America Inc */
/* Vendor: 1060: United Microelectronics [UMC] */
#define DEVICE_UNITED_UM82C881				0x0001 /*UM82C881*/
#define DEVICE_UNITED_UM82C886				0x0002 /*UM82C886*/
#define DEVICE_UNITED_UM8673F				0x0101 /*UM8673F*/
#define DEVICE_UNITED_UM8881				0x0881 /*UM8881*/
#define DEVICE_UNITED_UM8886F				0x0886 /*UM8886F*/
#define DEVICE_UNITED_UM8891A				0x0891 /*UM8891A*/
#define DEVICE_UNITED_UM886A				0x1001 /*UM886A*/
#define DEVICE_UNITED_UM8886BF				0x673a /*UM8886BF*/
#define DEVICE_UNITED_EIDE_MASTER_DMA			0x673b /*EIDE Master/DMA*/
#define DEVICE_UNITED_UM8710				0x8710 /*UM8710*/
#define DEVICE_UNITED_UM8886A				0x886a /*UM8886A*/
#define DEVICE_UNITED_UM8881F				0x8881 /*UM8881F*/
#define DEVICE_UNITED_UM8886F2				0x8886 /*UM8886F*/
#define DEVICE_UNITED_UM8886A2				0x888a /*UM8886A*/
#define DEVICE_UNITED_UM8891A2				0x8891 /*UM8891A*/
#define DEVICE_UNITED_UM9017F				0x9017 /*UM9017F*/
#define DEVICE_UNITED_UM9018				0x9018 /*UM9018*/
#define DEVICE_UNITED_UM9026				0x9026 /*UM9026*/
#define DEVICE_UNITED_UM8881N				0xe881 /*UM8881N*/
#define DEVICE_UNITED_UM8886N				0xe886 /*UM8886N*/
#define DEVICE_UNITED_UM8886N2				0xe88a /*UM8886N*/
#define DEVICE_UNITED_UM8891N				0xe891 /*UM8891N*/
/* Vendor: 1061: I.I.T. */
#define DEVICE_I_I_T_AGX016				0x0001 /*AGX016*/
#define DEVICE_I_I_T_IIT3204_3501			0x0002 /*IIT3204/3501*/
/* Vendor: 1062: Maspar Computer Corp */
/* Vendor: 1063: Ocean Office Automation */
/* Vendor: 1064: Alcatel */
/* Vendor: 1065: Texas Microsystems */
/* Vendor: 1066: PicoPower Technology */
#define DEVICE_PICOPOWER_PT80C826			0x0000 /*PT80C826*/
#define DEVICE_PICOPOWER_PT86C521_VESUVIUS_V1		0x0001 /*PT86C521 [Vesuvius v1] Host Bridge*/
#define DEVICE_PICOPOWER_PT86C523_VESUVIUS_V3		0x0002 /*PT86C523 [Vesuvius v3] PCI-ISA Bridge Master*/
#define DEVICE_PICOPOWER_PT86C524_NILE_PCI_TO_PCI	0x0003 /*PT86C524 [Nile] PCI-to-PCI Bridge*/
#define DEVICE_PICOPOWER_PT86C525_NILE_II_PCI_TO_PCI	0x0004 /*PT86C525 [Nile-II] PCI-to-PCI Bridge*/
#define DEVICE_PICOPOWER_NATIONAL_PC87550_SYSTEM	0x0005 /*National PC87550 System Controller*/
#define DEVICE_PICOPOWER_PT86C523_VESUVIUS_V32		0x8002 /*PT86C523 [Vesuvius v3] PCI-ISA Bridge Slave*/
/* Vendor: 1067: Mitsubishi Electric */
#define DEVICE_MITSUBISHI_ACCELGRAPHICS_ACCELECLIPSE	0x0301 /*AccelGraphics AccelECLIPSE*/
#define DEVICE_MITSUBISHI_ACCELGALAXY_A2100_OEM		0x0304 /*AccelGALAXY A2100 [OEM Evans & Sutherland]*/
#define DEVICE_MITSUBISHI_TORNADO_3000_OEM		0x0308 /*Tornado 3000 [OEM Evans & Sutherland]*/
#define DEVICE_MITSUBISHI_VG500_VOLUMEPRO_VOLUME	0x1002 /*VG500 [VolumePro Volume Rendering Accelerator]*/
/* Vendor: 1068: Diversified Technology */
/* Vendor: 1069: Mylex Corporation */
#define DEVICE_MYLEX_DAC960P				0x0001 /*DAC960P*/
#define DEVICE_MYLEX_DAC960PD				0x0002 /*DAC960PD*/
#define DEVICE_MYLEX_DAC960PG				0x0010 /*DAC960PG*/
#define DEVICE_MYLEX_DAC960LA				0x0020 /*DAC960LA*/
#define DEVICE_MYLEX_ACCELERAID_352_170_160_SUPPORT	0x0050 /*AcceleRAID 352/170/160 support Device*/
#define DEVICE_MYLEX_GEMSTONE_CHIPSET_SCSI		0xb166 /*Gemstone chipset SCSI controller*/
	#define SUBDEVICE_MYLEX_ISERIES_2872_DASD	0x10140242 /*iSeries 2872 DASD IOA*/
	#define SUBDEVICE_MYLEX_DUAL_CHANNEL_PCI_X	0x10140266 /*Dual Channel PCI-X U320 SCSI Adapter*/
	#define SUBDEVICE_MYLEX_DUAL_CHANNEL_PCI_X2	0x10140278 /*Dual Channel PCI-X U320 SCSI RAID Adapter*/
#define DEVICE_MYLEX_EXTREMERAID_1100_SUPPORT		0xba55 /*eXtremeRAID 1100 support Device*/
#define DEVICE_MYLEX_EXTREMERAID_2000_3000_SUPPORT	0xba56 /*eXtremeRAID 2000/3000 support Device*/
/* Vendor: 106a: Aten Research Inc */
/* Vendor: 106b: Apple Computer Inc. */
#define DEVICE_APPLE_BANDIT_POWERPC_HOST		0x0001 /*Bandit PowerPC host bridge*/
#define DEVICE_APPLE_GRAND_CENTRAL_I_O			0x0002 /*Grand Central I/O*/
#define DEVICE_APPLE_CONTROL_VIDEO			0x0003 /*Control Video*/
#define DEVICE_APPLE_PLANB_VIDEO_IN			0x0004 /*PlanB Video-In*/
#define DEVICE_APPLE_O_HARE_I_O				0x0007 /*O'Hare I/O*/
#define DEVICE_APPLE_HYDRA_MAC_I_O			0x000e /*Hydra Mac I/O*/
#define DEVICE_APPLE_HEATHROW_MAC_I_O			0x0010 /*Heathrow Mac I/O*/
#define DEVICE_APPLE_PADDINGTON_MAC_I_O			0x0017 /*Paddington Mac I/O*/
#define DEVICE_APPLE_UNINORTH_FIREWIRE			0x0018 /*UniNorth FireWire*/
#define DEVICE_APPLE_KEYLARGO_USB			0x0019 /*KeyLargo USB*/
#define DEVICE_APPLE_UNINORTH_INTERNAL_PCI		0x001e /*UniNorth Internal PCI*/
#define DEVICE_APPLE_UNINORTH_PCI			0x001f /*UniNorth PCI*/
#define DEVICE_APPLE_UNINORTH_AGP			0x0020 /*UniNorth AGP*/
#define DEVICE_APPLE_UNINORTH_GMAC_SUN			0x0021 /*UniNorth GMAC (Sun GEM)*/
#define DEVICE_APPLE_KEYLARGO_MAC_I_O			0x0022 /*KeyLargo Mac I/O*/
#define DEVICE_APPLE_UNINORTH_PANGEA_GMAC_SUN		0x0024 /*UniNorth/Pangea GMAC (Sun GEM)*/
#define DEVICE_APPLE_KEYLARGO_PANGEA_MAC_I_O		0x0025 /*KeyLargo/Pangea Mac I/O*/
#define DEVICE_APPLE_KEYLARGO_PANGEA_USB		0x0026 /*KeyLargo/Pangea USB*/
#define DEVICE_APPLE_UNINORTH_PANGEA_AGP		0x0027 /*UniNorth/Pangea AGP*/
#define DEVICE_APPLE_UNINORTH_PANGEA_PCI		0x0028 /*UniNorth/Pangea PCI*/
#define DEVICE_APPLE_UNINORTH_PANGEA_INTERNAL_PCI	0x0029 /*UniNorth/Pangea Internal PCI*/
#define DEVICE_APPLE_UNINORTH_1_5_AGP			0x002d /*UniNorth 1.5 AGP*/
#define DEVICE_APPLE_UNINORTH_1_5_PCI			0x002e /*UniNorth 1.5 PCI*/
#define DEVICE_APPLE_UNINORTH_1_5_INTERNAL		0x002f /*UniNorth 1.5 Internal PCI*/
#define DEVICE_APPLE_UNINORTH_PANGEA_FIREWIRE		0x0030 /*UniNorth/Pangea FireWire*/
#define DEVICE_APPLE_UNINORTH_2_FIREWIRE		0x0031 /*UniNorth 2 FireWire*/
#define DEVICE_APPLE_UNINORTH_2_GMAC			0x0032 /*UniNorth 2 GMAC (Sun GEM)*/
#define DEVICE_APPLE_UNINORTH_2_ATA_100			0x0033 /*UniNorth 2 ATA/100*/
#define DEVICE_APPLE_UNINORTH_2_AGP			0x0034 /*UniNorth 2 AGP*/
#define DEVICE_APPLE_UNINORTH_2_PCI			0x0035 /*UniNorth 2 PCI*/
#define DEVICE_APPLE_UNINORTH_2_INTERNAL		0x0036 /*UniNorth 2 Internal PCI*/
#define DEVICE_APPLE_UNINORTH_INTREPID_ATA_100		0x003b /*UniNorth/Intrepid ATA/100*/
#define DEVICE_APPLE_KEYLARGO_INTREPID_MAC_I_O		0x003e /*KeyLargo/Intrepid Mac I/O*/
#define DEVICE_APPLE_KEYLARGO_INTREPID_USB		0x003f /*KeyLargo/Intrepid USB*/
#define DEVICE_APPLE_K2_KEYLARGO_USB			0x0040 /*K2 KeyLargo USB*/
#define DEVICE_APPLE_K2_KEYLARGO_MAC_IO			0x0041 /*K2 KeyLargo Mac/IO*/
#define DEVICE_APPLE_K2_FIREWIRE			0x0042 /*K2 FireWire*/
#define DEVICE_APPLE_K2_ATA_100				0x0043 /*K2 ATA/100*/
#define DEVICE_APPLE_K2_HT_PCI_BRIDGE			0x0045 /*K2 HT-PCI Bridge*/
#define DEVICE_APPLE_K2_HT_PCI_BRIDGE2			0x0046 /*K2 HT-PCI Bridge*/
#define DEVICE_APPLE_K2_HT_PCI_BRIDGE3			0x0047 /*K2 HT-PCI Bridge*/
#define DEVICE_APPLE_K2_HT_PCI_BRIDGE4			0x0048 /*K2 HT-PCI Bridge*/
#define DEVICE_APPLE_K2_HT_PCI_BRIDGE5			0x0049 /*K2 HT-PCI Bridge*/
#define DEVICE_APPLE_U3_AGP				0x004b /*U3 AGP*/
#define DEVICE_APPLE_K2_GMAC_SUN			0x004c /*K2 GMAC (Sun GEM)*/
#define DEVICE_APPLE_SHASTA_MAC_I_O			0x004f /*Shasta Mac I/O*/
#define DEVICE_APPLE_SHASTA_IDE				0x0050 /*Shasta IDE*/
#define DEVICE_APPLE_SHASTA_SUN_GEM			0x0051 /*Shasta (Sun GEM)*/
#define DEVICE_APPLE_SHASTA_FIREWIRE			0x0052 /*Shasta Firewire*/
#define DEVICE_APPLE_SHASTA_PCI_BRIDGE			0x0053 /*Shasta PCI Bridge*/
#define DEVICE_APPLE_SHASTA_PCI_BRIDGE2			0x0054 /*Shasta PCI Bridge*/
#define DEVICE_APPLE_SHASTA_PCI_BRIDGE3			0x0055 /*Shasta PCI Bridge*/
#define DEVICE_APPLE_U3L_AGP_BRIDGE			0x0058 /*U3L AGP Bridge*/
#define DEVICE_APPLE_TIGON3_GIGABIT_ETHERNET		0x1645 /*Tigon3 Gigabit Ethernet NIC (BCM5701)*/
/* Vendor: 106c: Hynix Semiconductor */
#define DEVICE_HYNIX_DUAL_PENTIUM_ISA_PCI		0x8801 /*Dual Pentium ISA/PCI Motherboard*/
#define DEVICE_HYNIX_POWERPC_ISA_PCI_MOTHERBOARD	0x8802 /*PowerPC ISA/PCI Motherboard*/
#define DEVICE_HYNIX_DUAL_WINDOW_GRAPHICS		0x8803 /*Dual Window Graphics Accelerator*/
#define DEVICE_HYNIX_LAN_CONTROLLER			0x8804 /*LAN Controller*/
#define DEVICE_HYNIX_100_BASET_LAN			0x8805 /*100-BaseT LAN*/
/* Vendor: 106d: Sequent Computer Systems */
/* Vendor: 106e: DFI, Inc */
/* Vendor: 106f: City Gate Development Ltd */
/* Vendor: 1070: Daewoo Telecom Ltd */
/* Vendor: 1071: Mitac */
#define DEVICE_MITAC_MITAC_8060B_MOBILE			0x8160 /*Mitac 8060B Mobile Platform*/
/* Vendor: 1072: GIT Co Ltd */
/* Vendor: 1073: Yamaha Corporation */
#define DEVICE_YAMAHA_3D_GUI_ACCELERATOR		0x0001 /*3D GUI Accelerator*/
#define DEVICE_YAMAHA_YGV615_RPA3_3D_GRAPHICS		0x0002 /*YGV615 [RPA3 3D-Graphics Controller]*/
#define DEVICE_YAMAHA_YMF_740				0x0003 /*YMF-740*/
#define DEVICE_YAMAHA_YMF_724				0x0004 /*YMF-724*/
	#define SUBDEVICE_YAMAHA_YMF724_BASED_PCI_AUDIO	0x10730004 /*YMF724-Based PCI Audio Adapter*/
#define DEVICE_YAMAHA_DS1_AUDIO				0x0005 /*DS1 Audio*/
	#define SUBDEVICE_YAMAHA_DS_XG_PCI_AUDIO	0x10730005 /*DS-XG PCI Audio CODEC*/
#define DEVICE_YAMAHA_DS1_AUDIO2			0x0006 /*DS1 Audio*/
#define DEVICE_YAMAHA_DS1_AUDIO3			0x0008 /*DS1 Audio*/
	#define SUBDEVICE_YAMAHA_DS_XG_PCI_AUDIO2	0x10730008 /*DS-XG PCI Audio CODEC*/
#define DEVICE_YAMAHA_DS1L_AUDIO			0x000a /*DS1L Audio*/
	#define SUBDEVICE_YAMAHA_DS_XG_PCI_AUDIO3	0x10730004 /*DS-XG PCI Audio CODEC*/
	#define SUBDEVICE_YAMAHA_DS_XG_PCI_AUDIO4	0x1073000a /*DS-XG PCI Audio CODEC*/
#define DEVICE_YAMAHA_YMF_740C_DS_1L_AUDIO		0x000c /*YMF-740C [DS-1L Audio Controller]*/
	#define SUBDEVICE_YAMAHA_DS_XG_PCI_AUDIO5	0x107a000c /*DS-XG PCI Audio CODEC*/
#define DEVICE_YAMAHA_YMF_724F_DS_1_AUDIO		0x000d /*YMF-724F [DS-1 Audio Controller]*/
	#define SUBDEVICE_YAMAHA_DS_XG_PCI_AUDIO6	0x1073000d /*DS-XG PCI Audio CODEC*/
#define DEVICE_YAMAHA_YMF_744B_DS_1S_AUDIO		0x0010 /*YMF-744B [DS-1S Audio Controller]*/
	#define SUBDEVICE_YAMAHA_DS_XG_PCI_AUDIO7	0x10730006 /*DS-XG PCI Audio CODEC*/
	#define SUBDEVICE_YAMAHA_DS_XG_PCI_AUDIO8	0x10730010 /*DS-XG PCI Audio CODEC*/
#define DEVICE_YAMAHA_YMF_754_DS_1E_AUDIO		0x0012 /*YMF-754 [DS-1E Audio Controller]*/
	#define SUBDEVICE_YAMAHA_DS_XG_PCI_AUDIO9	0x10730012 /*DS-XG PCI Audio Codec*/
#define DEVICE_YAMAHA_DS_1_AUDIO			0x0020 /*DS-1 Audio*/
#define DEVICE_YAMAHA_DS2416_DIGITAL_MIXING		0x2000 /*DS2416 Digital Mixing Card*/
	#define SUBDEVICE_YAMAHA_DS2416_DIGITAL_MIXING	0x10732000 /*DS2416 Digital Mixing Card*/
/* Vendor: 1074: NexGen Microsystems */
#define DEVICE_NEXGEN_82C500_1				0x4e78 /*82c500/1*/
/* Vendor: 1075: Advanced Integrations Research */
/* Vendor: 1076: Chaintech Computer Co. Ltd */
/* Vendor: 1077: QLogic Corp. */
#define DEVICE_QLOGIC_ISP10160_SINGLE_CHANNEL		0x1016 /*ISP10160 Single Channel Ultra3 SCSI Processor*/
#define DEVICE_QLOGIC_ISP1020_FAST_WIDE_SCSI		0x1020 /*ISP1020 Fast-wide SCSI*/
#define DEVICE_QLOGIC_ISP1022_FAST_WIDE_SCSI		0x1022 /*ISP1022 Fast-wide SCSI*/
#define DEVICE_QLOGIC_ISP1080_SCSI_HOST			0x1080 /*ISP1080 SCSI Host Adapter*/
#define DEVICE_QLOGIC_ISP12160_DUAL_CHANNEL		0x1216 /*ISP12160 Dual Channel Ultra3 SCSI Processor*/
	#define SUBDEVICE_QLOGIC_QLA12160_ON_AMI	0x101e8471 /*QLA12160 on AMI MegaRAID*/
	#define SUBDEVICE_QLOGIC_QLA12160_ON_AMI2	0x101e8493 /*QLA12160 on AMI MegaRAID*/
#define DEVICE_QLOGIC_ISP1240_SCSI_HOST			0x1240 /*ISP1240 SCSI Host Adapter*/
#define DEVICE_QLOGIC_ISP1280_SCSI_HOST			0x1280 /*ISP1280 SCSI Host Adapter*/
#define DEVICE_QLOGIC_ISP2020A_FAST_SCSI_BASIC		0x2020 /*ISP2020A Fast!SCSI Basic Adapter*/
#define DEVICE_QLOGIC_QLA2100_64_BIT_FIBRE		0x2100 /*QLA2100 64-bit Fibre Channel Adapter*/
	#define SUBDEVICE_QLOGIC_QLA2100_64_BIT_FIBRE	0x10770001 /*QLA2100 64-bit Fibre Channel Adapter*/
#define DEVICE_QLOGIC_QLA2200_64_BIT_FIBRE		0x2200 /*QLA2200 64-bit Fibre Channel Adapter*/
	#define SUBDEVICE_QLOGIC_QLA2200		0x10770002 /*QLA2200*/
#define DEVICE_QLOGIC_QLA2300_64_BIT_FIBRE		0x2300 /*QLA2300 64-bit Fibre Channel Adapter*/
#define DEVICE_QLOGIC_QLA2312_FIBRE_CHANNEL		0x2312 /*QLA2312 Fibre Channel Adapter*/
/* Vendor: 1078: Cyrix Corporation */
#define DEVICE_CYRIX_5510_GRAPPA			0x0000 /*5510 [Grappa]*/
#define DEVICE_CYRIX_PCI_MASTER				0x0001 /*PCI Master*/
#define DEVICE_CYRIX_5520_COGNAC			0x0002 /*5520 [Cognac]*/
#define DEVICE_CYRIX_5530_LEGACY_KAHLUA			0x0100 /*5530 Legacy [Kahlua]*/
#define DEVICE_CYRIX_5530_SMI_KAHLUA			0x0101 /*5530 SMI [Kahlua]*/
#define DEVICE_CYRIX_5530_IDE_KAHLUA			0x0102 /*5530 IDE [Kahlua]*/
#define DEVICE_CYRIX_5530_AUDIO_KAHLUA			0x0103 /*5530 Audio [Kahlua]*/
#define DEVICE_CYRIX_5530_VIDEO_KAHLUA			0x0104 /*5530 Video [Kahlua]*/
#define DEVICE_CYRIX_ZFMICRO_PCI_BRIDGE			0x0400 /*ZFMicro PCI Bridge*/
#define DEVICE_CYRIX_ZFMICRO_CHIPSET_SMI		0x0401 /*ZFMicro Chipset SMI*/
#define DEVICE_CYRIX_ZFMICRO_CHIPSET_IDE		0x0402 /*ZFMicro Chipset IDE*/
#define DEVICE_CYRIX_ZFMICRO_EXPANSION_BUS		0x0403 /*ZFMicro Expansion Bus*/
/* Vendor: 1079: I-Bus */
/* Vendor: 107a: NetWorth */
/* Vendor: 107b: Gateway 2000 */
/* Vendor: 107c: LG Electronics [Lucky Goldstar Co. Ltd] */
/* Vendor: 107d: LeadTek Research Inc. */
#define DEVICE_LEADTEK_P86C850				0x0000 /*P86C850*/
#define DEVICE_LEADTEK_WINFAST_3D_S320			0x2134 /*WinFast 3D S320 II*/
#define DEVICE_LEADTEK__GEFORCE_FX_5900			0x2971 /*[GeForce FX 5900] WinFast A350 TDH MyViVo*/
/* Vendor: 107e: Interphase Corporation */
#define DEVICE_INTERPHASE_5515_ATM_ADAPTER		0x0001 /*5515 ATM Adapter [Flipper]*/
#define DEVICE_INTERPHASE_100_VG_ANYLAN			0x0002 /*100 VG AnyLan Controller*/
#define DEVICE_INTERPHASE_5526_FIBRE_CHANNEL		0x0004 /*5526 Fibre Channel Host Adapter*/
#define DEVICE_INTERPHASE_X526_FIBRE_CHANNEL		0x0005 /*x526 Fibre Channel Host Adapter*/
#define DEVICE_INTERPHASE_5525_5575_ATM_ADAPTER		0x0008 /*5525/5575 ATM Adapter (155 Mbit) [Atlantic]*/
#define DEVICE_INTERPHASE_5535_4P_BRI_ST		0x9003 /*5535-4P-BRI-ST*/
#define DEVICE_INTERPHASE_5535_4P_BRI_U			0x9007 /*5535-4P-BRI-U*/
#define DEVICE_INTERPHASE_5535_1P_SR			0x9008 /*5535-1P-SR*/
#define DEVICE_INTERPHASE_5535_1P_SR_ST			0x900c /*5535-1P-SR-ST*/
#define DEVICE_INTERPHASE_5535_1P_SR_U			0x900e /*5535-1P-SR-U*/
#define DEVICE_INTERPHASE_5535_1P_PRI			0x9011 /*5535-1P-PRI*/
#define DEVICE_INTERPHASE_5535_2P_PRI			0x9013 /*5535-2P-PRI*/
#define DEVICE_INTERPHASE_5536_4P_BRI_ST		0x9023 /*5536-4P-BRI-ST*/
#define DEVICE_INTERPHASE_5536_4P_BRI_U			0x9027 /*5536-4P-BRI-U*/
#define DEVICE_INTERPHASE_5536_1P_PRI			0x9031 /*5536-1P-PRI*/
#define DEVICE_INTERPHASE_5536_2P_PRI			0x9033 /*5536-2P-PRI*/
/* Vendor: 107f: Data Technology Corporation */
#define DEVICE_DATA_SL82C105				0x0802 /*SL82C105*/
/* Vendor: 1080: Contaq Microsystems */
#define DEVICE_CONTAQ_82C599				0x0600 /*82C599*/
#define DEVICE_CONTAQ_CYPRESS_CY82C691			0xc691 /*Cypress CY82C691*/
#define DEVICE_CONTAQ_82C693				0xc693 /*82c693*/
/* Vendor: 1081: Supermac Technology */
#define DEVICE_SUPERMAC_RADIUS_PCI_TO			0x0d47 /*Radius PCI to NuBUS Bridge*/
/* Vendor: 1082: EFA Corporation of America */
/* Vendor: 1083: Forex Computer Corporation */
#define DEVICE_FOREX_FR710				0x0001 /*FR710*/
/* Vendor: 1084: Parador */
/* Vendor: 1085: Tulip Computers Int.B.V. */
/* Vendor: 1086: J. Bond Computer Systems */
/* Vendor: 1087: Cache Computer */
/* Vendor: 1088: Microcomputer Systems (M) Son */
/* Vendor: 1089: Data General Corporation */
/* Vendor: 108a: SBS Technologies */
#define DEVICE_SBS_VME_BRIDGE_MODEL			0x0001 /*VME Bridge Model 617*/
#define DEVICE_SBS_VME_BRIDGE_MODEL2			0x0010 /*VME Bridge Model 618*/
#define DEVICE_SBS_DATABLIZZARD				0x0040 /*dataBLIZZARD*/
#define DEVICE_SBS_VME_BRIDGE_MODEL3			0x3000 /*VME Bridge Model 2706*/
/* Vendor: 108c: Oakleigh Systems Inc. */
/* Vendor: 108d: Olicom */
#define DEVICE_OLICOM_TOKEN_RING_16_4_PCI		0x0001 /*Token-Ring 16/4 PCI Adapter (3136/3137)*/
#define DEVICE_OLICOM_16_4_TOKEN_RING			0x0002 /*16/4 Token Ring*/
#define DEVICE_OLICOM_RAPIDFIRE_3139_TOKEN_RING		0x0004 /*RapidFire 3139 Token-Ring 16/4 PCI Adapter*/
	#define SUBDEVICE_OLICOM_OC_3139_3140_RAPIDFIRE_TOKEN_RING	0x108d0004 /*OC-3139/3140 RapidFire Token-Ring 16/4 Adapter*/
#define DEVICE_OLICOM_GOCARD_3250_TOKEN_RING		0x0005 /*GoCard 3250 Token-Ring 16/4 CardBus PC Card*/
#define DEVICE_OLICOM_OC_3530_RAPIDFIRE_TOKEN_RING	0x0006 /*OC-3530 RapidFire Token-Ring 100*/
#define DEVICE_OLICOM_RAPIDFIRE_3141_TOKEN_RING		0x0007 /*RapidFire 3141 Token-Ring 16/4 PCI Fiber Adapter*/
	#define SUBDEVICE_OLICOM_OC_3141_RAPIDFIRE_TOKEN_RING	0x108d0007 /*OC-3141 RapidFire Token-Ring 16/4 Adapter*/
#define DEVICE_OLICOM_RAPIDFIRE_3540_HSTR		0x0008 /*RapidFire 3540 HSTR 100/16/4 PCI Adapter*/
	#define SUBDEVICE_OLICOM_OC_3540_RAPIDFIRE_HSTR	0x108d0008 /*OC-3540 RapidFire HSTR 100/16/4 Adapter*/
#define DEVICE_OLICOM_OC_2315				0x0011 /*OC-2315*/
#define DEVICE_OLICOM_OC_2325				0x0012 /*OC-2325*/
#define DEVICE_OLICOM_OC_2183_2185			0x0013 /*OC-2183/2185*/
#define DEVICE_OLICOM_OC_2326				0x0014 /*OC-2326*/
#define DEVICE_OLICOM_OC_2327_2250_10_100_ETHERNET	0x0019 /*OC-2327/2250 10/100 Ethernet Adapter*/
	#define SUBDEVICE_OLICOM_OC_2327_RAPIDFIRE_10_100	0x108d0016 /*OC-2327 Rapidfire 10/100 Ethernet Adapter*/
	#define SUBDEVICE_OLICOM_OC_2250_GOCARD_10_100	0x108d0017 /*OC-2250 GoCard 10/100 Ethernet Adapter*/
#define DEVICE_OLICOM_OC_6151_6152_RAPIDFIRE_ATM	0x0021 /*OC-6151/6152 [RapidFire ATM 155]*/
#define DEVICE_OLICOM_ATM_ADAPTER			0x0022 /*ATM Adapter*/
/* Vendor: 108e: Sun Microsystems Computer Corp. */
#define DEVICE_SUN_EBUS					0x0001 /*EBUS*/
#define DEVICE_SUN_EBUS2					0x1000 /*EBUS*/
#define DEVICE_SUN_HAPPY_MEAL				0x1001 /*Happy Meal*/
#define DEVICE_SUN_RIO_EBUS				0x1100 /*RIO EBUS*/
#define DEVICE_SUN_RIO_GEM				0x1101 /*RIO GEM*/
#define DEVICE_SUN_RIO_1394				0x1102 /*RIO 1394*/
#define DEVICE_SUN_RIO_USB				0x1103 /*RIO USB*/
#define DEVICE_SUN_GEM					0x2bad /*GEM*/
#define DEVICE_SUN_SIMBA_ADVANCED_PCI			0x5000 /*Simba Advanced PCI Bridge*/
#define DEVICE_SUN_SUNPCI_CO_PROCESSOR			0x5043 /*SunPCI Co-processor*/
#define DEVICE_SUN_PSYCHO_PCI_BUS			0x8000 /*Psycho PCI Bus Module*/
#define DEVICE_SUN_SCHIZO_PCI_BUS			0x8001 /*Schizo PCI Bus Module*/
#define DEVICE_SUN_ULTRA_III				0xa000 /*Ultra IIi*/
#define DEVICE_SUN_ULTRA_IIE				0xa001 /*Ultra IIe*/
#define DEVICE_SUN_TOMATILLO_PCI_BUS			0xa801 /*Tomatillo PCI Bus Module*/
#define DEVICE_SUN_CASSINI_10_100_1000			0xabba /*Cassini 10/100/1000*/
/* Vendor: 108f: Systemsoft */
/* Vendor: 1090: Encore Computer Corporation */
/* Vendor: 1091: Intergraph Corporation */
#define DEVICE_INTERGRAPH_3D_GRAPHICS_PROCESSOR		0x0020 /*3D graphics processor*/
#define DEVICE_INTERGRAPH_3D_GRAPHICS_PROCESSOR2	0x0021 /*3D graphics processor w/Texturing*/
#define DEVICE_INTERGRAPH_3D_GRAPHICS_FRAME		0x0040 /*3D graphics frame buffer*/
#define DEVICE_INTERGRAPH_3D_GRAPHICS_FRAME2		0x0041 /*3D graphics frame buffer*/
#define DEVICE_INTERGRAPH_PROPRIETARY_BUS_BRIDGE	0x0060 /*Proprietary bus bridge*/
#define DEVICE_INTERGRAPH_POWERSTORM_4D50T		0x00e4 /*Powerstorm 4D50T*/
#define DEVICE_INTERGRAPH_MOTION_JPEG_CODEC		0x0720 /*Motion JPEG codec*/
#define DEVICE_INTERGRAPH_SUN_EXPERT3D_LITE_GRAPHICS	0x07a0 /*Sun Expert3D-Lite Graphics Accelerator*/
#define DEVICE_INTERGRAPH_SUN_EXPERT3D_GRAPHICS		0x1091 /*Sun Expert3D Graphics Accelerator*/
/* Vendor: 1092: Diamond Multimedia Systems */
#define DEVICE_DIAMOND_SPEEDSTAR_PRO_SE			0x00a0 /*Speedstar Pro SE*/
#define DEVICE_DIAMOND_SPEEDSTAR_64			0x00a8 /*Speedstar 64*/
#define DEVICE_DIAMOND_VIPER_V550			0x0550 /*Viper V550*/
#define DEVICE_DIAMOND_SUPRA_2260_MODEM			0x08d4 /*Supra 2260 Modem*/
#define DEVICE_DIAMOND_SUPRAEXPRESS_56I_PRO		0x094c /*SupraExpress 56i Pro*/
#define DEVICE_DIAMOND_VIPER_V330			0x1092 /*Viper V330*/
#define DEVICE_DIAMOND_MAXIMUM_DVD			0x6120 /*Maximum DVD*/
#define DEVICE_DIAMOND_STEALTH_SE			0x8810 /*Stealth SE*/
#define DEVICE_DIAMOND_STEALTH_64_SE			0x8811 /*Stealth 64/SE*/
#define DEVICE_DIAMOND_STEALTH				0x8880 /*Stealth*/
#define DEVICE_DIAMOND_STEALTH2				0x8881 /*Stealth*/
#define DEVICE_DIAMOND_STEALTH_64			0x88b0 /*Stealth 64*/
#define DEVICE_DIAMOND_STEALTH_642			0x88b1 /*Stealth 64*/
#define DEVICE_DIAMOND_STEALTH_643			0x88c0 /*Stealth 64*/
#define DEVICE_DIAMOND_STEALTH_644			0x88c1 /*Stealth 64*/
#define DEVICE_DIAMOND_STEALTH_645			0x88d0 /*Stealth 64*/
#define DEVICE_DIAMOND_STEALTH_646			0x88d1 /*Stealth 64*/
#define DEVICE_DIAMOND_STEALTH_647			0x88f0 /*Stealth 64*/
#define DEVICE_DIAMOND_STEALTH_648			0x88f1 /*Stealth 64*/
#define DEVICE_DIAMOND_DMD_I0928_1_MONSTER_SOUND	0x9999 /*DMD-I0928-1 \"Monster sound\" sound chip*/
/* Vendor: 1093: National Instruments */
#define DEVICE_NATIONAL2_PCI_DIO_96			0x0160 /*PCI-DIO-96*/
#define DEVICE_NATIONAL2_PCI_MIO_16XE_50		0x0162 /*PCI-MIO-16XE-50*/
#define DEVICE_NATIONAL2_PCI_MIO_16XE_10		0x1170 /*PCI-MIO-16XE-10*/
#define DEVICE_NATIONAL2_PCI_MIO_16E_1			0x1180 /*PCI-MIO-16E-1*/
#define DEVICE_NATIONAL2_PCI_MIO_16E_4			0x1190 /*PCI-MIO-16E-4*/
#define DEVICE_NATIONAL2_PCI_6031E			0x1330 /*PCI-6031E*/
#define DEVICE_NATIONAL2_PCI_6071E			0x1350 /*PCI-6071E*/
#define DEVICE_NATIONAL2_PCI_6110			0x14e0 /*PCI-6110*/
#define DEVICE_NATIONAL2_PCI_6111			0x14f0 /*PCI-6111*/
#define DEVICE_NATIONAL2_PCI_6503			0x17d0 /*PCI-6503*/
#define DEVICE_NATIONAL2_PCI_6713			0x1870 /*PCI-6713*/
#define DEVICE_NATIONAL2_PCI_6711			0x1880 /*PCI-6711*/
#define DEVICE_NATIONAL2_PCI_6052E			0x18b0 /*PCI-6052E*/
#define DEVICE_NATIONAL2_PCI_6733			0x2410 /*PCI-6733*/
#define DEVICE_NATIONAL2_PCI_6036E			0x2890 /*PCI-6036E*/
#define DEVICE_NATIONAL2_PCI_6023E			0x2a60 /*PCI-6023E*/
#define DEVICE_NATIONAL2_PCI_6024E			0x2a70 /*PCI-6024E*/
#define DEVICE_NATIONAL2_PCI_6025E			0x2a80 /*PCI-6025E*/
#define DEVICE_NATIONAL2_PCI_6035E			0x2c80 /*PCI-6035E*/
#define DEVICE_NATIONAL2_PCI_6034E			0x2ca0 /*PCI-6034E*/
#define DEVICE_NATIONAL2_IMAQ_PCI_1408			0xb001 /*IMAQ-PCI-1408*/
#define DEVICE_NATIONAL2_IMAQ_PXI_1408			0xb011 /*IMAQ-PXI-1408*/
#define DEVICE_NATIONAL2_IMAQ_PCI_1424			0xb021 /*IMAQ-PCI-1424*/
#define DEVICE_NATIONAL2_IMAQ_PCI_1413			0xb031 /*IMAQ-PCI-1413*/
#define DEVICE_NATIONAL2_IMAQ_PCI_1407			0xb041 /*IMAQ-PCI-1407*/
#define DEVICE_NATIONAL2_IMAQ_PXI_1407			0xb051 /*IMAQ-PXI-1407*/
#define DEVICE_NATIONAL2_IMAQ_PCI_1411			0xb061 /*IMAQ-PCI-1411*/
#define DEVICE_NATIONAL2_IMAQ_PCI_1422			0xb071 /*IMAQ-PCI-1422*/
#define DEVICE_NATIONAL2_IMAQ_PXI_1422			0xb081 /*IMAQ-PXI-1422*/
#define DEVICE_NATIONAL2_IMAQ_PXI_1411			0xb091 /*IMAQ-PXI-1411*/
#define DEVICE_NATIONAL2_PCI_GPIB			0xc801 /*PCI-GPIB*/
#define DEVICE_NATIONAL2_PCI_GPIB_BRIDGE		0xc831 /*PCI-GPIB bridge*/
/* Vendor: 1094: First International Computers [FIC] */
/* Vendor: 1095: Silicon Image, Inc. (formerly CMD Technology Inc) */
#define DEVICE_SILICON3_ADAPTEC_AAR_1210SA_SATA		0x0240 /*Adaptec AAR-1210SA SATA HostRAID Controller*/
#define DEVICE_SILICON3_PCI0640				0x0640 /*PCI0640*/
#define DEVICE_SILICON3_PCI0643				0x0643 /*PCI0643*/
#define DEVICE_SILICON3_PCI0646				0x0646 /*PCI0646*/
#define DEVICE_SILICON3_PCI0647				0x0647 /*PCI0647*/
#define DEVICE_SILICON3_PCI0648				0x0648 /*PCI0648*/
#define DEVICE_SILICON3_SII_0649_ULTRA			0x0649 /*SiI 0649 Ultra ATA/100 PCI to ATA Host Controller*/
	#define SUBDEVICE_SILICON3_INTEGRATED_ULTRA_ATA_100	0x0e11005d /*Integrated Ultra ATA-100 Dual Channel Controller*/
	#define SUBDEVICE_SILICON3_INTEGRATED_ULTRA_ATA_1002	0x0e11007e /*Integrated Ultra ATA-100 IDE RAID Controller*/
	#define SUBDEVICE_SILICON3_AMI_MEGARAID_IDE	0x101e0649 /*AMI MegaRAID IDE 100 Controller*/
#define DEVICE_SILICON3_PBC0650A			0x0650 /*PBC0650A*/
#define DEVICE_SILICON3_USB0670				0x0670 /*USB0670*/
	#define SUBDEVICE_SILICON3_USB0670		0x10950670 /*USB0670*/
#define DEVICE_SILICON3_USB0673				0x0673 /*USB0673*/
#define DEVICE_SILICON3_PCI0680_ULTRA_ATA_133		0x0680 /*PCI0680 Ultra ATA-133 Host Controller*/
	#define SUBDEVICE_SILICON3_WINIC_W_680_SILICON	0x10953680 /*Winic W-680 (Silicon Image 680 based)*/
#define DEVICE_SILICON3_SII_3112_SATALINK_SATARAID	0x3112 /*SiI 3112 [SATALink/SATARaid] Serial ATA Controller*/
	#define SUBDEVICE_SILICON3_SII_3112_SATALINK	0x10953112 /*SiI 3112 SATALink Controller*/
	#define SUBDEVICE_SILICON3_SII_3112_SATARAID	0x10956112 /*SiI 3112 SATARaid Controller*/
#define DEVICE_SILICON3_SII_3114_SATALINK_SATARAID	0x3114 /*SiI 3114 [SATALink/SATARaid] Serial ATA Controller*/
	#define SUBDEVICE_SILICON3_SII_3114_SATALINK	0x10953114 /*SiI 3114 SATALink Controller*/
	#define SUBDEVICE_SILICON3_SII_3114_SATARAID	0x10956114 /*SiI 3114 SATARaid Controller*/
#define DEVICE_SILICON3_SII_3124_PCI_X			0x3124 /*SiI 3124 PCI-X Serial ATA Controller*/
	#define SUBDEVICE_SILICON3_SII_3124_PCI_X	0x10953124 /*SiI 3124 PCI-X Serial ATA Controller*/
#define DEVICE_SILICON3_SII_3512_SATALINK_SATARAID	0x3512 /*SiI 3512 [SATALink/SATARaid] Serial ATA Controller*/
	#define SUBDEVICE_SILICON3_SII_3512_SATALINK	0x10953512 /*SiI 3512 SATALink Controller*/
	#define SUBDEVICE_SILICON3_SII_3512_SATARAID	0x10956512 /*SiI 3512 SATARaid Controller*/
/* Vendor: 1096: Alacron */
/* Vendor: 1097: Appian Technology */
/* Vendor: 1098: Quantum Designs (H.K.) Ltd */
#define DEVICE_QUANTUM_QD_8500				0x0001 /*QD-8500*/
#define DEVICE_QUANTUM_QD_8580				0x0002 /*QD-8580*/
/* Vendor: 1099: Samsung Electronics Co., Ltd */
/* Vendor: 109a: Packard Bell */
/* Vendor: 109b: Gemlight Computer Ltd. */
/* Vendor: 109c: Megachips Corporation */
/* Vendor: 109d: Zida Technologies Ltd. */
/* Vendor: 109e: Brooktree Corporation */
#define DEVICE_BROOKTREE_BT848_VIDEO_CAPTURE		0x0350 /*Bt848 Video Capture*/
#define DEVICE_BROOKTREE_BT849A_VIDEO_CAPTURE		0x0351 /*Bt849A Video capture*/
#define DEVICE_BROOKTREE_BT878_VIDEO_CAPTURE		0x0369 /*Bt878 Video Capture*/
	#define SUBDEVICE_BROOKTREE_TV_WONDER		0x10020001 /*TV-Wonder*/
	#define SUBDEVICE_BROOKTREE_TV_WONDER_VE	0x10020003 /*TV-Wonder/VE*/
#define DEVICE_BROOKTREE_BT879_VIDEO_CAPTURE		0x036c /*Bt879(?) Video Capture*/
	#define SUBDEVICE_BROOKTREE_WIN_TV_VIDEO_SECTION	0x13e90070 /*Win/TV (Video Section)*/
#define DEVICE_BROOKTREE_BT878_VIDEO_CAPTURE2		0x036e /*Bt878 Video Capture*/
	#define SUBDEVICE_BROOKTREE_WINTV_SERIES	0x007013eb /*WinTV Series*/
	#define SUBDEVICE_BROOKTREE_VIEWCAST_OSPREY_200	0x0070ff01 /*Viewcast Osprey 200*/
	#define SUBDEVICE_BROOKTREE_DIGITV_PCI		0x00710101 /*DigiTV PCI*/
	#define SUBDEVICE_BROOKTREE_WINFAST_TV_2000	0x107d6606 /*WinFast TV 2000*/
	#define SUBDEVICE_BROOKTREE_PCTV_PRO_TV		0x11bd0012 /*PCTV pro (TV + FM stereo receiver)*/
	#define SUBDEVICE_BROOKTREE_PCTV_SAT_DBC	0x11bd001c /*PCTV Sat (DBC receiver)*/
	#define SUBDEVICE_BROOKTREE_BT878_MEDIASTREAM_CONTROLLER	0x127a0001 /*Bt878 Mediastream Controller NTSC*/
	#define SUBDEVICE_BROOKTREE_BT878_MEDIASTREAM_CONTROLLER2	0x127a0002 /*Bt878 Mediastream Controller PAL BG*/
	#define SUBDEVICE_BROOKTREE_BT878A_MEDIASTREAM_CONTROLLER	0x127a0003 /*Bt878a Mediastream Controller PAL BG*/
	#define SUBDEVICE_BROOKTREE_BT878_832_MEDIASTREAM_CONTROLLER	0x127a0048 /*Bt878/832 Mediastream Controller*/
	#define SUBDEVICE_BROOKTREE_MAGICTVIEW_CPH060	0x144f3000 /*MagicTView CPH060 - Video*/
	#define SUBDEVICE_BROOKTREE_TV98_SERIES_TV_NO	0x14610002 /*TV98 Series (TV/No FM/Remote)*/
	#define SUBDEVICE_BROOKTREE_AVERTV_WDM_VIDEO	0x14610004 /*AVerTV WDM Video Capture*/
	#define SUBDEVICE_BROOKTREE_AVERTV_DVB_T	0x14610761 /*AverTV DVB-T*/
	#define SUBDEVICE_BROOKTREE_BT878_MEDIASTREAM_CONTROLLER3	0x14f10001 /*Bt878 Mediastream Controller NTSC*/
	#define SUBDEVICE_BROOKTREE_BT878_MEDIASTREAM_CONTROLLER4	0x14f10002 /*Bt878 Mediastream Controller PAL BG*/
	#define SUBDEVICE_BROOKTREE_BT878A_MEDIASTREAM_CONTROLLER2	0x14f10003 /*Bt878a Mediastream Controller PAL BG*/
	#define SUBDEVICE_BROOKTREE_BT878_832_MEDIASTREAM_CONTROLLER2	0x14f10048 /*Bt878/832 Mediastream Controller*/
	#define SUBDEVICE_BROOKTREE_VISIONPLUS_DVB_CARD	0x18220001 /*VisionPlus DVB card*/
	#define SUBDEVICE_BROOKTREE_FLYVIDEO_98_VIDEO	0x18511850 /*FlyVideo'98 - Video*/
	#define SUBDEVICE_BROOKTREE_FLYVIDEO_II		0x18511851 /*FlyVideo II*/
	#define SUBDEVICE_BROOKTREE_FLYVIDEO_98_VIDEO2	0x18521852 /*FlyVideo'98 - Video (with FM Tuner)*/
	#define SUBDEVICE_BROOKTREE_DIGITOP_DTT_1000	0x270ffc00 /*Digitop DTT-1000*/
	#define SUBDEVICE_BROOKTREE_PCTV_PRO_TV2	0xbd111200 /*PCTV pro (TV + FM stereo receiver)*/
#define DEVICE_BROOKTREE_BT879_VIDEO_CAPTURE2		0x036f /*Bt879 Video Capture*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE	0x127a0044 /*Bt879 Video Capture NTSC*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE2	0x127a0122 /*Bt879 Video Capture PAL I*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE3	0x127a0144 /*Bt879 Video Capture NTSC*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE4	0x127a0222 /*Bt879 Video Capture PAL BG*/
	#define SUBDEVICE_BROOKTREE_BT879A_VIDEO_CAPTURE	0x127a0244 /*Bt879a Video Capture NTSC*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE5	0x127a0322 /*Bt879 Video Capture NTSC*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE6	0x127a0422 /*Bt879 Video Capture NTSC*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE7	0x127a1122 /*Bt879 Video Capture PAL I*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE8	0x127a1222 /*Bt879 Video Capture PAL BG*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE9	0x127a1322 /*Bt879 Video Capture NTSC*/
	#define SUBDEVICE_BROOKTREE_BT879A_VIDEO_CAPTURE2	0x127a1522 /*Bt879a Video Capture PAL I*/
	#define SUBDEVICE_BROOKTREE_BT879A_VIDEO_CAPTURE3	0x127a1622 /*Bt879a Video Capture PAL BG*/
	#define SUBDEVICE_BROOKTREE_BT879A_VIDEO_CAPTURE4	0x127a1722 /*Bt879a Video Capture NTSC*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE10	0x14f10044 /*Bt879 Video Capture NTSC*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE11	0x14f10122 /*Bt879 Video Capture PAL I*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE12	0x14f10144 /*Bt879 Video Capture NTSC*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE13	0x14f10222 /*Bt879 Video Capture PAL BG*/
	#define SUBDEVICE_BROOKTREE_BT879A_VIDEO_CAPTURE5	0x14f10244 /*Bt879a Video Capture NTSC*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE14	0x14f10322 /*Bt879 Video Capture NTSC*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE15	0x14f10422 /*Bt879 Video Capture NTSC*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE16	0x14f11122 /*Bt879 Video Capture PAL I*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE17	0x14f11222 /*Bt879 Video Capture PAL BG*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE18	0x14f11322 /*Bt879 Video Capture NTSC*/
	#define SUBDEVICE_BROOKTREE_BT879A_VIDEO_CAPTURE6	0x14f11522 /*Bt879a Video Capture PAL I*/
	#define SUBDEVICE_BROOKTREE_BT879A_VIDEO_CAPTURE7	0x14f11622 /*Bt879a Video Capture PAL BG*/
	#define SUBDEVICE_BROOKTREE_BT879A_VIDEO_CAPTURE8	0x14f11722 /*Bt879a Video Capture NTSC*/
	#define SUBDEVICE_BROOKTREE_FLYVIDEO_98_VIDEO3	0x18511850 /*FlyVideo'98 - Video*/
	#define SUBDEVICE_BROOKTREE_FLYVIDEO_II2	0x18511851 /*FlyVideo II*/
	#define SUBDEVICE_BROOKTREE_FLYVIDEO_98_VIDEO4	0x18521852 /*FlyVideo'98 - Video (with FM Tuner)*/
#define DEVICE_BROOKTREE_BT880_VIDEO_CAPTURE		0x0370 /*Bt880 Video Capture*/
	#define SUBDEVICE_BROOKTREE_FLYVIDEO_98		0x18511850 /*FlyVideo'98*/
	#define SUBDEVICE_BROOKTREE_FLYVIDEO_98_EZ	0x18511851 /*FlyVideo'98 EZ - video*/
	#define SUBDEVICE_BROOKTREE_FLYVIDEO_98_WITH_FM	0x18521852 /*FlyVideo'98 (with FM Tuner)*/
#define DEVICE_BROOKTREE_BT878_AUDIO_CAPTURE		0x0878 /*Bt878 Audio Capture*/
	#define SUBDEVICE_BROOKTREE_WINTV_SERIES2	0x007013eb /*WinTV Series*/
	#define SUBDEVICE_BROOKTREE_VIEWCAST_OSPREY_2002	0x0070ff01 /*Viewcast Osprey 200*/
	#define SUBDEVICE_BROOKTREE_DIGITV_PCI2		0x00710101 /*DigiTV PCI*/
	#define SUBDEVICE_BROOKTREE_TV_WONDER2		0x10020001 /*TV-Wonder*/
	#define SUBDEVICE_BROOKTREE_TV_WONDER_VE2	0x10020003 /*TV-Wonder/VE*/
	#define SUBDEVICE_BROOKTREE_PCTV_PRO_TV3	0x11bd0012 /*PCTV pro (TV + FM stereo receiver, audio section)*/
	#define SUBDEVICE_BROOKTREE_PCTV_SAT_DBC2	0x11bd001c /*PCTV Sat (DBC receiver)*/
	#define SUBDEVICE_BROOKTREE_BT878_VIDEO_CAPTURE	0x127a0001 /*Bt878 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT878_VIDEO_CAPTURE2	0x127a0002 /*Bt878 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT878_VIDEO_CAPTURE3	0x127a0003 /*Bt878 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT878_VIDEO_CAPTURE4	0x127a0048 /*Bt878 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_WIN_TV_AUDIO_SECTION	0x13e90070 /*Win/TV (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_MAGICTVIEW_CPH0602	0x144f3000 /*MagicTView CPH060 - Audio*/
	#define SUBDEVICE_BROOKTREE_AVERTV_WDM_AUDIO	0x14610004 /*AVerTV WDM Audio Capture*/
	#define SUBDEVICE_BROOKTREE_AVERTV_DVB_T2	0x14610761 /*AVerTV DVB-T*/
	#define SUBDEVICE_BROOKTREE_BT878_VIDEO_CAPTURE5	0x14f10001 /*Bt878 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT878_VIDEO_CAPTURE6	0x14f10002 /*Bt878 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT878_VIDEO_CAPTURE7	0x14f10003 /*Bt878 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT878_VIDEO_CAPTURE8	0x14f10048 /*Bt878 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_VISIONPLUS_DVB_CARD2	0x18220001 /*VisionPlus DVB Card*/
	#define SUBDEVICE_BROOKTREE_DIGITOP_DTT_10002	0x270ffc00 /*Digitop DTT-1000*/
	#define SUBDEVICE_BROOKTREE_PCTV_PRO_TV4	0xbd111200 /*PCTV pro (TV + FM stereo receiver, audio section)*/
#define DEVICE_BROOKTREE_BT879_AUDIO_CAPTURE		0x0879 /*Bt879 Audio Capture*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE19	0x127a0044 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE20	0x127a0122 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE21	0x127a0144 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE22	0x127a0222 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE23	0x127a0244 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE24	0x127a0322 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE25	0x127a0422 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE26	0x127a1122 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE27	0x127a1222 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE28	0x127a1322 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE29	0x127a1522 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE30	0x127a1622 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE31	0x127a1722 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE32	0x14f10044 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE33	0x14f10122 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE34	0x14f10144 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE35	0x14f10222 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE36	0x14f10244 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE37	0x14f10322 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE38	0x14f10422 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE39	0x14f11122 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE40	0x14f11222 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE41	0x14f11322 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE42	0x14f11522 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE43	0x14f11622 /*Bt879 Video Capture (Audio Section)*/
	#define SUBDEVICE_BROOKTREE_BT879_VIDEO_CAPTURE44	0x14f11722 /*Bt879 Video Capture (Audio Section)*/
#define DEVICE_BROOKTREE_BT880_AUDIO_CAPTURE		0x0880 /*Bt880 Audio Capture*/
#define DEVICE_BROOKTREE_BTV_2115_MEDIASTREAM		0x2115 /*BtV 2115 Mediastream controller*/
#define DEVICE_BROOKTREE_BTV_2125_MEDIASTREAM		0x2125 /*BtV 2125 Mediastream controller*/
#define DEVICE_BROOKTREE_BTV_2164			0x2164 /*BtV 2164*/
#define DEVICE_BROOKTREE_BTV_2165			0x2165 /*BtV 2165*/
#define DEVICE_BROOKTREE_BT8230_ATM_SEGMENT_REASSEMBLY	0x8230 /*Bt8230 ATM Segment/Reassembly Ctrlr (SRC)*/
#define DEVICE_BROOKTREE_BT8472				0x8472 /*Bt8472*/
#define DEVICE_BROOKTREE_BT8474				0x8474 /*Bt8474*/
/* Vendor: 109f: Trigem Computer Inc. */
/* Vendor: 10a0: Meidensha Corporation */
/* Vendor: 10a1: Juko Electronics Ind. Co. Ltd */
/* Vendor: 10a2: Quantum Corporation */
/* Vendor: 10a3: Everex Systems Inc */
/* Vendor: 10a4: Globe Manufacturing Sales */
/* Vendor: 10a5: Smart Link Ltd. */
#define DEVICE_SMART_SMARTPCI562_56K_MODEM		0x3052 /*SmartPCI562 56K Modem*/
#define DEVICE_SMART_SMARTPCI561_MODEM			0x5449 /*SmartPCI561 modem*/
/* Vendor: 10a6: Informtech Industrial Ltd. */
/* Vendor: 10a7: Benchmarq Microelectronics */
/* Vendor: 10a8: Sierra Semiconductor */
#define DEVICE_SIERRA_STB_HORIZON_64			0x0000 /*STB Horizon 64*/
/* Vendor: 10a9: Silicon Graphics, Inc. */
#define DEVICE_SILICON4_CROSSTALK_TO_PCI		0x0001 /*Crosstalk to PCI Bridge*/
#define DEVICE_SILICON4_LINC_I_O_CONTROLLER		0x0002 /*Linc I/O controller*/
#define DEVICE_SILICON4_IOC3_I_O_CONTROLLER		0x0003 /*IOC3 I/O controller*/
#define DEVICE_SILICON4_O2_MACE				0x0004 /*O2 MACE*/
#define DEVICE_SILICON4_RAD_AUDIO			0x0005 /*RAD Audio*/
#define DEVICE_SILICON4_HPCEX				0x0006 /*HPCEX*/
#define DEVICE_SILICON4_RPCEX				0x0007 /*RPCEX*/
#define DEVICE_SILICON4_DIVO_VIP			0x0008 /*DiVO VIP*/
#define DEVICE_SILICON4_ACENIC_GIGABIT_ETHERNET		0x0009 /*AceNIC Gigabit Ethernet*/
	#define SUBDEVICE_SILICON4_ACENIC_GIGABIT_ETHERNET	0x10a98002 /*AceNIC Gigabit Ethernet*/
#define DEVICE_SILICON4_AMP_VIDEO_I_O			0x0010 /*AMP Video I/O*/
#define DEVICE_SILICON4_GRIP				0x0011 /*GRIP*/
#define DEVICE_SILICON4_SGH_PSHAC_GSN			0x0012 /*SGH PSHAC GSN*/
#define DEVICE_SILICON4_MAGIC_CARPET			0x1001 /*Magic Carpet*/
#define DEVICE_SILICON4_LITHIUM				0x1002 /*Lithium*/
#define DEVICE_SILICON4_DUAL_JPEG_1			0x1003 /*Dual JPEG 1*/
#define DEVICE_SILICON4_DUAL_JPEG_2			0x1004 /*Dual JPEG 2*/
#define DEVICE_SILICON4_DUAL_JPEG_3			0x1005 /*Dual JPEG 3*/
#define DEVICE_SILICON4_DUAL_JPEG_4			0x1006 /*Dual JPEG 4*/
#define DEVICE_SILICON4_DUAL_JPEG_5			0x1007 /*Dual JPEG 5*/
#define DEVICE_SILICON4_CESIUM				0x1008 /*Cesium*/
#define DEVICE_SILICON4_IOC4_I_O_CONTROLLER		0x100a /*IOC4 I/O controller*/
#define DEVICE_SILICON4_FIBRE_CHANNEL			0x2001 /*Fibre Channel*/
#define DEVICE_SILICON4_ASDE				0x2002 /*ASDE*/
#define DEVICE_SILICON4_O2_1394				0x8001 /*O2 1394*/
#define DEVICE_SILICON4_G_NET_NT			0x8002 /*G-net NT*/
/* Vendor: 10aa: ACC Microelectronics */
#define DEVICE_ACC_ACCM_2188				0x0000 /*ACCM 2188*/
/* Vendor: 10ab: Digicom */
/* Vendor: 10ac: Honeywell IAC */
/* Vendor: 10ad: Symphony Labs */
#define DEVICE_SYMPHONY_W83769F				0x0001 /*W83769F*/
#define DEVICE_SYMPHONY_SL82C103			0x0003 /*SL82C103*/
#define DEVICE_SYMPHONY_SL82C105			0x0005 /*SL82C105*/
#define DEVICE_SYMPHONY_SL82C1032			0x0103 /*SL82c103*/
#define DEVICE_SYMPHONY_SL82C1052			0x0105 /*SL82c105*/
#define DEVICE_SYMPHONY_W83C553				0x0565 /*W83C553*/
/* Vendor: 10ae: Cornerstone Technology */
/* Vendor: 10af: Micro Computer Systems Inc */
/* Vendor: 10b0: CardExpert Technology */
/* Vendor: 10b1: Cabletron Systems Inc */
/* Vendor: 10b2: Raytheon Company */
/* Vendor: 10b3: Databook Inc */
#define DEVICE_DATABOOK_DB87144				0x3106 /*DB87144*/
#define DEVICE_DATABOOK_DB871442			0xb106 /*DB87144*/
/* Vendor: 10b4: STB Systems Inc */
#define DEVICE_STB_VELOCITY_128_3D			0x1b1d /*Velocity 128 3D*/
	#define SUBDEVICE_STB_VELOCITY_4400		0x10b4237e /*Velocity 4400*/
/* Vendor: 10b5: PLX Technology, Inc. */
#define DEVICE_PLX_I960_PCI_BUS				0x0001 /*i960 PCI bus interface*/
#define DEVICE_PLX_VSCOM_800_8				0x1076 /*VScom 800 8 port serial adaptor*/
#define DEVICE_PLX_VSCOM_400_4				0x1077 /*VScom 400 4 port serial adaptor*/
#define DEVICE_PLX_VSCOM_210_2				0x1078 /*VScom 210 2 port serial and 1 port parallel adaptor*/
#define DEVICE_PLX_VSCOM_200_2				0x1103 /*VScom 200 2 port serial adaptor*/
#define DEVICE_PLX_VSCOM_010_1				0x1146 /*VScom 010 1 port parallel adaptor*/
#define DEVICE_PLX_VSCOM_020_2				0x1147 /*VScom 020 2 port parallel adaptor*/
#define DEVICE_PLX_THALES_PCSM_SECURITY			0x2724 /*Thales PCSM Security Card*/
#define DEVICE_PLX_PCI_IOBUS				0x9030 /*PCI <-> IOBus Bridge Hot Swap*/
	#define SUBDEVICE_PLX_ALPERMANN_VELTE_PCL_PCI	0x10b52862 /*Alpermann+Velte PCL PCI LV (3V/5V): Timecode Reader Board*/
	#define SUBDEVICE_PLX_ALPERMANN_VELTE_PCI_TS	0x10b52906 /*Alpermann+Velte PCI TS (3V/5V): Time Synchronisation Board*/
	#define SUBDEVICE_PLX_ALPERMANN_VELTE_PCL_PCI2	0x10b52940 /*Alpermann+Velte PCL PCI D (3V/5V): Timecode Reader Board*/
	#define SUBDEVICE_PLX_MCCS_8_PORT_SERIAL	0x15ed1002 /*MCCS 8-port Serial Hot Swap*/
	#define SUBDEVICE_PLX_MCCS_16_PORT_SERIAL	0x15ed1003 /*MCCS 16-port Serial Hot Swap*/
#define DEVICE_PLX_9036					0x9036 /*9036*/
#define DEVICE_PLX_PCI_IOBUS2				0x9050 /*PCI <-> IOBus Bridge*/
	#define SUBDEVICE_PLX_IXXAT_CAN_I165		0x10b51067 /*IXXAT CAN i165*/
	#define SUBDEVICE_PLX_IK220_HEIDENHAIN		0x10b51172 /*IK220 (Heidenhain)*/
	#define SUBDEVICE_PLX_SATPAK_GPS		0x10b52036 /*SatPak GPS*/
	#define SUBDEVICE_PLX_ALPERMANN_VELTE_PCL_PCI3	0x10b52221 /*Alpermann+Velte PCL PCI LV: Timecode Reader Board*/
	#define SUBDEVICE_PLX_SH_ARC_SOHARD_ARCNET	0x10b52273 /*SH-ARC SoHard ARCnet card*/
	#define SUBDEVICE_PLX_ALPERMANN_VELTE_PCL_PCI4	0x10b52431 /*Alpermann+Velte PCL PCI D: Timecode Reader Board*/
	#define SUBDEVICE_PLX_ALPERMANN_VELTE_PCI_TS2	0x10b52905 /*Alpermann+Velte PCI TS: Time Synchronisation Board*/
	#define SUBDEVICE_PLX_MP9050			0x10b59050 /*MP9050*/
	#define SUBDEVICE_PLX_TPMC866_8_CHANNEL		0x14980362 /*TPMC866 8 Channel Serial Card*/
	#define SUBDEVICE_PLX_ROCKFORCE_4_PORT		0x15220001 /*RockForce 4 Port V.90 Data/Fax/Voice Modem*/
	#define SUBDEVICE_PLX_ROCKFORCE_2_PORT		0x15220002 /*RockForce 2 Port V.90 Data/Fax/Voice Modem*/
	#define SUBDEVICE_PLX_ROCKFORCE_6_PORT		0x15220003 /*RockForce 6 Port V.90 Data/Fax/Voice Modem*/
	#define SUBDEVICE_PLX_ROCKFORCE_8_PORT		0x15220004 /*RockForce 8 Port V.90 Data/Fax/Voice Modem*/
	#define SUBDEVICE_PLX_ROCKFORCE2000_4_PORT	0x15220010 /*RockForce2000 4 Port V.90 Data/Fax/Voice Modem*/
	#define SUBDEVICE_PLX_ROCKFORCE2000_2_PORT	0x15220020 /*RockForce2000 2 Port V.90 Data/Fax/Voice Modem*/
	#define SUBDEVICE_PLX_MACROLINK_MCCS_8_PORT	0x15ed1000 /*Macrolink MCCS 8-port Serial*/
	#define SUBDEVICE_PLX_MACROLINK_MCCS_16_PORT	0x15ed1001 /*Macrolink MCCS 16-port Serial*/
	#define SUBDEVICE_PLX_MACROLINK_MCCS_8_PORT2	0x15ed1002 /*Macrolink MCCS 8-port Serial Hot Swap*/
	#define SUBDEVICE_PLX_MACROLINK_MCCS_16_PORT2	0x15ed1003 /*Macrolink MCCS 16-port Serial Hot Swap*/
	#define SUBDEVICE_PLX_OPENSWITCH_6_TELEPHONY	0x56542036 /*OpenSwitch 6 Telephony card*/
	#define SUBDEVICE_PLX_OPENSWITCH_12_TELEPHONY	0x56543132 /*OpenSwitch 12 Telephony card*/
	#define SUBDEVICE_PLX_OPENLINE4_TELEPHONY_CARD	0x56545634 /*OpenLine4 Telephony Card*/
	#define SUBDEVICE_PLX_PCINTELLICAN_2XSJA1000_CAN	0xd531c002 /*PCIntelliCAN 2xSJA1000 CAN bus*/
	#define SUBDEVICE_PLX_EX_4006_1P		0xd84d4006 /*EX-4006 1P*/
	#define SUBDEVICE_PLX_EX_4008_1P_EPP_ECP	0xd84d4008 /*EX-4008 1P EPP/ECP*/
	#define SUBDEVICE_PLX_EX_4014_2P		0xd84d4014 /*EX-4014 2P*/
	#define SUBDEVICE_PLX_EX_4018_3P_EPP_ECP	0xd84d4018 /*EX-4018 3P EPP/ECP*/
	#define SUBDEVICE_PLX_EX_4025_1S_16C550_RS_232	0xd84d4025 /*EX-4025 1S(16C550) RS-232*/
	#define SUBDEVICE_PLX_EX_4027_1S_16C650_RS_232	0xd84d4027 /*EX-4027 1S(16C650) RS-232*/
	#define SUBDEVICE_PLX_EX_4028_1S_16C850_RS_232	0xd84d4028 /*EX-4028 1S(16C850) RS-232*/
	#define SUBDEVICE_PLX_EX_4036_2S_16C650_RS_232	0xd84d4036 /*EX-4036 2S(16C650) RS-232*/
	#define SUBDEVICE_PLX_EX_4037_2S_16C650_RS_232	0xd84d4037 /*EX-4037 2S(16C650) RS-232*/
	#define SUBDEVICE_PLX_EX_4038_2S_16C850_RS_232	0xd84d4038 /*EX-4038 2S(16C850) RS-232*/
	#define SUBDEVICE_PLX_EX_4052_1S_16C550_RS_422_485	0xd84d4052 /*EX-4052 1S(16C550) RS-422/485*/
	#define SUBDEVICE_PLX_EX_4053_2S_16C550_RS_422_485	0xd84d4053 /*EX-4053 2S(16C550) RS-422/485*/
	#define SUBDEVICE_PLX_EX_4055_4S_16C550_RS_232	0xd84d4055 /*EX-4055 4S(16C550) RS-232*/
	#define SUBDEVICE_PLX_EX_4055_4S_16C650_RS_232	0xd84d4058 /*EX-4055 4S(16C650) RS-232*/
	#define SUBDEVICE_PLX_EX_4065_8S_16C550_RS_232	0xd84d4065 /*EX-4065 8S(16C550) RS-232*/
	#define SUBDEVICE_PLX_EX_4068_8S_16C650_RS_232	0xd84d4068 /*EX-4068 8S(16C650) RS-232*/
	#define SUBDEVICE_PLX_EX_4078_2S_16C552_RS_232_1P	0xd84d4078 /*EX-4078 2S(16C552) RS-232+1P*/
#define DEVICE_PLX_PCI_IOBUS3				0x9054 /*PCI <-> IOBus Bridge*/
	#define SUBDEVICE_PLX_WESSEX_TECHOLOGY_PHIL_PCI	0x10b52455 /*Wessex Techology PHIL-PCI*/
	#define SUBDEVICE_PLX_INNES_CORP_AM		0x10b52696 /*Innes Corp AM Radcap card*/
	#define SUBDEVICE_PLX_INNES_CORP_AURICON	0x10b52717 /*Innes Corp Auricon card*/
	#define SUBDEVICE_PLX_INNES_CORP_TVS		0x10b52844 /*Innes Corp TVS Encoder card*/
	#define SUBDEVICE_PLX_PCI_PROSODY_CARD		0x12d90002 /*PCI Prosody Card rev 1.5*/
	#define SUBDEVICE_PLX_PIKA_PRIMENET_MM		0x16df0011 /*PIKA PrimeNet MM PCI*/
	#define SUBDEVICE_PLX_PIKA_PRIMENET_MM2		0x16df0012 /*PIKA PrimeNet MM cPCI 8*/
	#define SUBDEVICE_PLX_PIKA_PRIMENET_MM3		0x16df0013 /*PIKA PrimeNet MM cPCI 8 (without CAS Signaling Option)*/
	#define SUBDEVICE_PLX_PIKA_PRIMENET_MM4		0x16df0014 /*PIKA PrimeNet MM cPCI 4*/
	#define SUBDEVICE_PLX_PIKA_DAYTONA_MM		0x16df0015 /*PIKA Daytona MM*/
	#define SUBDEVICE_PLX_PIKA_INLINE_MM		0x16df0016 /*PIKA InLine MM*/
#define DEVICE_PLX_FRANCOIS				0x9056 /*Francois*/
	#define SUBDEVICE_PLX_CELLINKBLADE_11		0x10b52979 /*CellinkBlade 11 - CPCI board VoATM AAL1*/
#define DEVICE_PLX_9060					0x9060 /*9060*/
#define DEVICE_PLX_9060SD				0x906d /*9060SD*/
	#define SUBDEVICE_PLX_ARIES_16000P		0x125c0640 /*Aries 16000P*/
#define DEVICE_PLX_9060ES				0x906e /*9060ES*/
#define DEVICE_PLX_9080					0x9080 /*9080*/
	#define SUBDEVICE_PLX__AGILENT_E2777B_83K	0x103c10eb /*(Agilent) E2777B 83K Series PCI based Optical Communication Interface*/
	#define SUBDEVICE_PLX__AGILENT_E6978_66442_PCI	0x103c10ec /*(Agilent) E6978-66442 PCI CIC*/
	#define SUBDEVICE_PLX_9080_REAL_SUBSYSTEM	0x10b59080 /*9080 [real subsystem ID not set]*/
	#define SUBDEVICE_PLX_ACULAB_PCI_PROSIDY	0x129d0002 /*Aculab PCI Prosidy card*/
	#define SUBDEVICE_PLX_PCI_PROSODY_CARD2		0x12d90002 /*PCI Prosody Card*/
	#define SUBDEVICE_PLX_4422PCI_DO_ALL_TELEMETRY	0x12df4422 /*4422PCI [\"Do-All\" Telemetry Data Aquisition System]*/
#define DEVICE_PLX_B_B_3PCIOSD1A_ISOLATED		0xbb04 /*B&B 3PCIOSD1A Isolated PCI Serial*/
/* Vendor: 10b6: Madge Networks */
#define DEVICE_MADGE_SMART_16_4_PCI			0x0001 /*Smart 16/4 PCI Ringnode*/
#define DEVICE_MADGE_SMART_16_4_PCI2			0x0002 /*Smart 16/4 PCI Ringnode Mk2*/
	#define SUBDEVICE_MADGE_SMART_16_4_PCI		0x10b60002 /*Smart 16/4 PCI Ringnode Mk2*/
	#define SUBDEVICE_MADGE_16_4_CARDBUS_ADAPTER	0x10b60006 /*16/4 CardBus Adapter*/
#define DEVICE_MADGE_SMART_16_4_PCI3			0x0003 /*Smart 16/4 PCI Ringnode Mk3*/
	#define SUBDEVICE_MADGE_COMPAQ_NC4621_PCI	0x0e11b0fd /*Compaq NC4621 PCI, 4/16, WOL*/
	#define SUBDEVICE_MADGE_SMART_16_4_PCI2		0x10b60003 /*Smart 16/4 PCI Ringnode Mk3*/
	#define SUBDEVICE_MADGE_PRESTO_PCI_PLUS		0x10b60007 /*Presto PCI Plus Adapter*/
#define DEVICE_MADGE_SMART_16_4_PCI4			0x0004 /*Smart 16/4 PCI Ringnode Mk1*/
#define DEVICE_MADGE_16_4_CARDBUS_ADAPTER		0x0006 /*16/4 Cardbus Adapter*/
	#define SUBDEVICE_MADGE_16_4_CARDBUS_ADAPTER2	0x10b60006 /*16/4 CardBus Adapter*/
#define DEVICE_MADGE_PRESTO_PCI_ADAPTER			0x0007 /*Presto PCI Adapter*/
	#define SUBDEVICE_MADGE_PRESTO_PCI		0x10b60007 /*Presto PCI*/
#define DEVICE_MADGE_SMART_100_16_4_PCI_HS		0x0009 /*Smart 100/16/4 PCI-HS Ringnode*/
	#define SUBDEVICE_MADGE_SMART_100_16_4_PCI_HS	0x10b60009 /*Smart 100/16/4 PCI-HS Ringnode*/
#define DEVICE_MADGE_SMART_100_16_4_PCI			0x000a /*Smart 100/16/4 PCI Ringnode*/
	#define SUBDEVICE_MADGE_SMART_100_16_4_PCI	0x10b6000a /*Smart 100/16/4 PCI Ringnode*/
#define DEVICE_MADGE_16_4_CARDBUS_ADAPTER2		0x000b /*16/4 CardBus Adapter Mk2*/
	#define SUBDEVICE_MADGE_16_4_CARDBUS_ADAPTER3	0x10b60008 /*16/4 CardBus Adapter Mk2*/
	#define SUBDEVICE_MADGE_16_4_CARDBUS_ADAPTER4	0x10b6000b /*16/4 Cardbus Adapter Mk2*/
#define DEVICE_MADGE_RAPIDFIRE_3140V2_16_4		0x000c /*RapidFire 3140V2 16/4 TR Adapter*/
	#define SUBDEVICE_MADGE_RAPIDFIRE_3140V2_16_4	0x10b6000c /*RapidFire 3140V2 16/4 TR Adapter*/
#define DEVICE_MADGE_COLLAGE_25_155_ATM			0x1000 /*Collage 25/155 ATM Client Adapter*/
#define DEVICE_MADGE_COLLAGE_155_ATM			0x1001 /*Collage 155 ATM Server Adapter*/
/* Vendor: 10b7: 3Com Corporation */
#define DEVICE_3COM_3C985_1000BASESX_SX_TX		0x0001 /*3c985 1000BaseSX (SX/TX)*/
#define DEVICE_3COM_AR5212_802_11ABG_NIC		0x0013 /*AR5212 802.11abg NIC (3CRDAG675)*/
	#define SUBDEVICE_3COM_3CRDAG675_11A_B_G_WIRELESS	0x10b72031 /*3CRDAG675 11a/b/g Wireless PCI Adapter*/
#define DEVICE_3COM_3C910_A01				0x0910 /*3C910-A01*/
#define DEVICE_3COM_MINI_PCI_TYPE			0x1006 /*MINI PCI type 3B Data Fax Modem*/
#define DEVICE_3COM_MINI_PCI_56K			0x1007 /*Mini PCI 56k Winmodem*/
	#define SUBDEVICE_3COM_MINI_PCI_56K		0x10b7615c /*Mini PCI 56K Modem*/
#define DEVICE_3COM_3C982_TXM_10_100BASETX_DUAL		0x1201 /*3c982-TXM 10/100baseTX Dual Port A [Hydra]*/
#define DEVICE_3COM_3C982_TXM_10_100BASETX_DUAL2	0x1202 /*3c982-TXM 10/100baseTX Dual Port B [Hydra]*/
#define DEVICE_3COM_3C940_10_100_1000BASE_T_MARVELL	0x1700 /*3c940 10/100/1000Base-T [Marvell]*/
	#define SUBDEVICE_3COM_P4P800_MAINBOARD		0x104380eb /*P4P800 Mainboard*/
	#define SUBDEVICE_3COM_3C940_GIGABIT_LOM	0x10b70010 /*3C940 Gigabit LOM Ethernet Adapter*/
	#define SUBDEVICE_3COM_3C941_GIGABIT_LOM	0x10b70020 /*3C941 Gigabit LOM Ethernet Adapter*/
	#define SUBDEVICE_3COM_KV8_MAX3_MOTHERBOARD	0x147b1407 /*KV8-MAX3 motherboard*/
#define DEVICE_3COM_3C339_TOKENLINK_VELOCITY		0x3390 /*3c339 TokenLink Velocity*/
#define DEVICE_3COM_3C359_TOKENLINK_VELOCITY		0x3590 /*3c359 TokenLink Velocity XL*/
	#define SUBDEVICE_3COM_TOKENLINK_VELOCITY_XL	0x10b73590 /*TokenLink Velocity XL Adapter (3C359/359B)*/
#define DEVICE_3COM_3C450_HOMEPNA_TORNADO		0x4500 /*3c450 HomePNA [Tornado]*/
#define DEVICE_3COM_3C555_LAPTOP_HURRICANE		0x5055 /*3c555 Laptop Hurricane*/
#define DEVICE_3COM_3C575_MEGAHERTZ_10_100		0x5057 /*3c575 Megahertz 10/100 LAN CardBus [Boomerang]*/
	#define SUBDEVICE_3COM_3C575_MEGAHERTZ_10_100	0x10b75a57 /*3C575 Megahertz 10/100 LAN Cardbus PC Card*/
#define DEVICE_3COM_3CCFE575BT_MEGAHERTZ_10_100		0x5157 /*3cCFE575BT Megahertz 10/100 LAN CardBus [Cyclone]*/
	#define SUBDEVICE_3COM_3C575_MEGAHERTZ_10_1002	0x10b75b57 /*3C575 Megahertz 10/100 LAN Cardbus PC Card*/
#define DEVICE_3COM_3CCFE575CT_CARDBUS_CYCLONE		0x5257 /*3cCFE575CT CardBus [Cyclone]*/
	#define SUBDEVICE_3COM_FE575C_3COM_10_100_LAN	0x10b75c57 /*FE575C-3Com 10/100 LAN CardBus-Fast Ethernet*/
#define DEVICE_3COM_3C590_10BASET_VORTEX		0x5900 /*3c590 10BaseT [Vortex]*/
#define DEVICE_3COM_3C592_EISA_10MBPS			0x5920 /*3c592 EISA 10mbps Demon/Vortex*/
#define DEVICE_3COM_3C595_100BASETX_VORTEX		0x5950 /*3c595 100BaseTX [Vortex]*/
#define DEVICE_3COM_3C595_100BASET4_VORTEX		0x5951 /*3c595 100BaseT4 [Vortex]*/
#define DEVICE_3COM_3C595_100BASE_MII_VORTEX		0x5952 /*3c595 100Base-MII [Vortex]*/
#define DEVICE_3COM_3C597_EISA_FAST			0x5970 /*3c597 EISA Fast Demon/Vortex*/
#define DEVICE_3COM_3C595_MEGAHERTZ_10_100		0x5b57 /*3c595 Megahertz 10/100 LAN CardBus [Boomerang]*/
	#define SUBDEVICE_3COM_3C575_MEGAHERTZ_10_1003	0x10b75b57 /*3C575 Megahertz 10/100 LAN Cardbus PC Card*/
#define DEVICE_3COM_3CRSHPW796_OFFICECONNECT_WIRELESS	0x6000 /*3CRSHPW796 [OfficeConnect Wireless CardBus]*/
#define DEVICE_3COM_3COM_3CRWE154G72_OFFICE		0x6001 /*3com 3CRWE154G72 [Office Connect Wireless LAN Adapter]*/
#define DEVICE_3COM_3C556_HURRICANE_CARDBUS		0x6055 /*3c556 Hurricane CardBus [Cyclone]*/
#define DEVICE_3COM_3C556B_CARDBUS_TORNADO		0x6056 /*3c556B CardBus [Tornado]*/
	#define SUBDEVICE_3COM_10_100_MINI_PCI		0x10b76556 /*10/100 Mini PCI Ethernet Adapter*/
#define DEVICE_3COM_3CCFE656_CARDBUS_CYCLONE		0x6560 /*3cCFE656 CardBus [Cyclone]*/
	#define SUBDEVICE_3COM_3CCFEM656_10_100_LAN_56K	0x10b7656a /*3CCFEM656 10/100 LAN+56K Modem CardBus*/
#define DEVICE_3COM_3CCFEM656_10_100_LAN_56K		0x6561 /*3cCFEM656 10/100 LAN+56K Modem CardBus*/
	#define SUBDEVICE_3COM_3CCFEM656_10_100_LAN_56K2	0x10b7656b /*3CCFEM656 10/100 LAN+56K Modem CardBus*/
#define DEVICE_3COM_3CCFEM656B_10_100_LAN_WINMODEM	0x6562 /*3cCFEM656B 10/100 LAN+Winmodem CardBus [Cyclone]*/
	#define SUBDEVICE_3COM_3CCFEM656B_10_100_LAN_56K	0x10b7656b /*3CCFEM656B 10/100 LAN+56K Modem CardBus*/
#define DEVICE_3COM_3CCFEM656B_10_100_LAN_56K		0x6563 /*3cCFEM656B 10/100 LAN+56K Modem CardBus*/
	#define SUBDEVICE_3COM_3CCFEM656_10_100_LAN_56K3	0x10b7656b /*3CCFEM656 10/100 LAN+56K Modem CardBus*/
#define DEVICE_3COM_3CXFEM656C_10_100_LAN_WINMODEM	0x6564 /*3cXFEM656C 10/100 LAN+Winmodem CardBus [Tornado]*/
#define DEVICE_3COM_3CSOHO100_TX_HURRICANE		0x7646 /*3cSOHO100-TX Hurricane*/
#define DEVICE_3COM_3CRWE777_PCI_PLX_WIRELESS		0x7770 /*3CRWE777 PCI(PLX) Wireless Adaptor [Airconnect]*/
#define DEVICE_3COM_3C803_FDDILINK_UTP			0x7940 /*3c803 FDDILink UTP Controller*/
#define DEVICE_3COM_3C804_FDDILINK_SAS			0x7980 /*3c804 FDDILink SAS Controller*/
#define DEVICE_3COM_3C805_FDDILINK_DAS			0x7990 /*3c805 FDDILink DAS Controller*/
#define DEVICE_3COM_3C940B_10_100_1000BASE_T		0x80eb /*3c940B 10/100/1000Base-T*/
#define DEVICE_3COM_TOKEN_RING				0x8811 /*Token ring*/
#define DEVICE_3COM_3C900_10BASET_BOOMERANG		0x9000 /*3c900 10BaseT [Boomerang]*/
#define DEVICE_3COM_3C900_10MBPS_COMBO			0x9001 /*3c900 10Mbps Combo [Boomerang]*/
#define DEVICE_3COM_3C900B_TPO_ETHERLINK_XL		0x9004 /*3c900B-TPO Etherlink XL [Cyclone]*/
	#define SUBDEVICE_3COM_3C900B_TPO_ETHERLINK_XL	0x10b79004 /*3C900B-TPO Etherlink XL TPO 10Mb*/
#define DEVICE_3COM_3C900B_COMBO_ETHERLINK_XL		0x9005 /*3c900B-Combo Etherlink XL [Cyclone]*/
	#define SUBDEVICE_3COM_3C900B_COMBO_ETHERLINK_XL	0x10b79005 /*3C900B-Combo Etherlink XL Combo*/
#define DEVICE_3COM_3C900B_TPC_ETHERLINK_XL		0x9006 /*3c900B-TPC Etherlink XL [Cyclone]*/
#define DEVICE_3COM_3C900B_FL_10BASE_FL_CYCLONE		0x900a /*3c900B-FL 10base-FL [Cyclone]*/
#define DEVICE_3COM_3C905_100BASETX_BOOMERANG		0x9050 /*3c905 100BaseTX [Boomerang]*/
#define DEVICE_3COM_3C905_100BASET4_BOOMERANG		0x9051 /*3c905 100BaseT4 [Boomerang]*/
#define DEVICE_3COM_3C905B_100BASETX_CYCLONE		0x9055 /*3c905B 100BaseTX [Cyclone]*/
	#define SUBDEVICE_3COM_3C905B_FAST_ETHERLINK	0x10280080 /*3C905B Fast Etherlink XL 10/100*/
	#define SUBDEVICE_3COM_3C905B_FAST_ETHERLINK2	0x10280081 /*3C905B Fast Etherlink XL 10/100*/
	#define SUBDEVICE_3COM_3C905B_FAST_ETHERLINK3	0x10280082 /*3C905B Fast Etherlink XL 10/100*/
	#define SUBDEVICE_3COM_3C905B_FAST_ETHERLINK4	0x10280083 /*3C905B Fast Etherlink XL 10/100*/
	#define SUBDEVICE_3COM_3C905B_FAST_ETHERLINK5	0x10280084 /*3C905B Fast Etherlink XL 10/100*/
	#define SUBDEVICE_3COM_3C905B_FAST_ETHERLINK6	0x10280085 /*3C905B Fast Etherlink XL 10/100*/
	#define SUBDEVICE_3COM_3C905B_FAST_ETHERLINK7	0x10280086 /*3C905B Fast Etherlink XL 10/100*/
	#define SUBDEVICE_3COM_3C905B_FAST_ETHERLINK8	0x10280087 /*3C905B Fast Etherlink XL 10/100*/
	#define SUBDEVICE_3COM_3C905B_FAST_ETHERLINK9	0x10280088 /*3C905B Fast Etherlink XL 10/100*/
	#define SUBDEVICE_3COM_3C905B_FAST_ETHERLINK10	0x10280089 /*3C905B Fast Etherlink XL 10/100*/
	#define SUBDEVICE_3COM_3C905B_FAST_ETHERLINK11	0x10280090 /*3C905B Fast Etherlink XL 10/100*/
	#define SUBDEVICE_3COM_3C905B_FAST_ETHERLINK12	0x10280091 /*3C905B Fast Etherlink XL 10/100*/
	#define SUBDEVICE_3COM_3C905B_FAST_ETHERLINK13	0x10280092 /*3C905B Fast Etherlink XL 10/100*/
	#define SUBDEVICE_3COM_3C905B_FAST_ETHERLINK14	0x10280093 /*3C905B Fast Etherlink XL 10/100*/
	#define SUBDEVICE_3COM_3C905B_FAST_ETHERLINK15	0x10280094 /*3C905B Fast Etherlink XL 10/100*/
	#define SUBDEVICE_3COM_3C905B_FAST_ETHERLINK16	0x10280095 /*3C905B Fast Etherlink XL 10/100*/
	#define SUBDEVICE_3COM_3C905B_FAST_ETHERLINK17	0x10280096 /*3C905B Fast Etherlink XL 10/100*/
	#define SUBDEVICE_3COM_3C905B_FAST_ETHERLINK18	0x10280097 /*3C905B Fast Etherlink XL 10/100*/
	#define SUBDEVICE_3COM_3C905B_FAST_ETHERLINK19	0x10280098 /*3C905B Fast Etherlink XL 10/100*/
	#define SUBDEVICE_3COM_3C905B_FAST_ETHERLINK20	0x10280099 /*3C905B Fast Etherlink XL 10/100*/
	#define SUBDEVICE_3COM_3C905B_FAST_ETHERLINK21	0x10b79055 /*3C905B Fast Etherlink XL 10/100*/
#define DEVICE_3COM_3C905B_T4_FAST_ETHERLINK		0x9056 /*3c905B-T4 Fast EtherLink XL [Cyclone]*/
#define DEVICE_3COM_3C905B_DELUXE_ETHERLINK		0x9058 /*3c905B Deluxe Etherlink 10/100/BNC [Cyclone]*/
#define DEVICE_3COM_3C905B_FX_FAST_ETHERLINK		0x905a /*3c905B-FX Fast Etherlink XL FX 100baseFx [Cyclone]*/
#define DEVICE_3COM_3C905C_TX_TX_M_TORNADO		0x9200 /*3c905C-TX/TX-M [Tornado]*/
	#define SUBDEVICE_3COM_3C920_INTEGRATED_FAST	0x10280095 /*3C920 Integrated Fast Ethernet Controller*/
	#define SUBDEVICE_3COM_3C920_INTEGRATED_FAST2	0x10280097 /*3C920 Integrated Fast Ethernet Controller*/
	#define SUBDEVICE_3COM_OPTIPLEX_GX240		0x102800fe /*Optiplex GX240*/
	#define SUBDEVICE_3COM_3C920_INTEGRATED_FAST3	0x1028012a /*3C920 Integrated Fast Ethernet Controller [Latitude C640]*/
	#define SUBDEVICE_3COM_3C905C_TX_FAST_ETHERLINK	0x10b71000 /*3C905C-TX Fast Etherlink for PC Management NIC*/
	#define SUBDEVICE_3COM_10_100_MINI_PCI2		0x10b77000 /*10/100 Mini PCI Ethernet Adapter*/
	#define SUBDEVICE_3COM_TIGER_MPX_S2466		0x10f12466 /*Tiger MPX S2466 (3C920 Integrated Fast Ethernet Controller)*/
#define DEVICE_3COM_3C920B_EMB_INTEGRATED_FAST		0x9201 /*3C920B-EMB Integrated Fast Ethernet Controller [Tornado]*/
	#define SUBDEVICE_3COM_A7N8X_DELUXE_ONBOARD	0x104380ab /*A7N8X Deluxe onboard 3C920B-EMB Integrated Fast Ethernet Controller*/
#define DEVICE_3COM_3COM_3C920B_EMB_WNM_INTEGRATED	0x9202 /*3Com 3C920B-EMB-WNM Integrated Fast Ethernet Controller*/
#define DEVICE_3COM_3C920B_EMB_WNM_INTEGRATED_FAST	0x9210 /*3C920B-EMB-WNM Integrated Fast Ethernet Controller*/
#define DEVICE_3COM_3CSOHO100B_TX_910_A01_TULIP		0x9300 /*3CSOHO100B-TX 910-A01 [tulip]*/
#define DEVICE_3COM_3C980_TX_FAST_ETHERLINK		0x9800 /*3c980-TX Fast Etherlink XL Server Adapter [Cyclone]*/
	#define SUBDEVICE_3COM_3C980_TX_FAST_ETHERLINK	0x10b79800 /*3c980-TX Fast Etherlink XL Server Adapter*/
#define DEVICE_3COM_3C980_C_10_100BASETX_NIC		0x9805 /*3c980-C 10/100baseTX NIC [Python-T]*/
	#define SUBDEVICE_3COM_ETHERLINK_SERVER_10_100	0x10b71201 /*EtherLink Server 10/100 Dual Port A*/
	#define SUBDEVICE_3COM_ETHERLINK_SERVER_10_1002	0x10b71202 /*EtherLink Server 10/100 Dual Port B*/
	#define SUBDEVICE_3COM_3C980_10_100BASETX_NIC	0x10b79805 /*3c980 10/100baseTX NIC [Python-T]*/
	#define SUBDEVICE_3COM_THUNDER_K7_S2462		0x10f12462 /*Thunder K7 S2462*/
#define DEVICE_3COM_3C990_TX_TYPHOON			0x9900 /*3C990-TX [Typhoon]*/
#define DEVICE_3COM_3CR990_TX_95_TYPHOON_56_BIT		0x9902 /*3CR990-TX-95 [Typhoon 56-bit]*/
#define DEVICE_3COM_3CR990_TX_97_TYPHOON_168_BIT	0x9903 /*3CR990-TX-97 [Typhoon 168-bit]*/
#define DEVICE_3COM_3C990B_TX_M_3C990BSVR_TYPHOON2	0x9904 /*3C990B-TX-M/3C990BSVR [Typhoon2]*/
	#define SUBDEVICE_3COM_3CR990B_TX_M_TYPHOON2	0x10b71000 /*3CR990B-TX-M [Typhoon2]*/
	#define SUBDEVICE_3COM_3CR990BSVR_TYPHOON2_SERVER	0x10b72000 /*3CR990BSVR [Typhoon2 Server]*/
#define DEVICE_3COM_3CR990_FX_95_97_95_TYPHON_FIBER	0x9905 /*3CR990-FX-95/97/95 [Typhon Fiber]*/
	#define SUBDEVICE_3COM_3CR990_FX_95_TYPHOON_FIBER	0x10b71101 /*3CR990-FX-95 [Typhoon Fiber 56-bit]*/
	#define SUBDEVICE_3COM_3CR990_FX_97_TYPHOON_FIBER	0x10b71102 /*3CR990-FX-97 [Typhoon Fiber 168-bit]*/
	#define SUBDEVICE_3COM_3CR990_FX_95_SERVER_TYPHOON	0x10b72101 /*3CR990-FX-95 Server [Typhoon Fiber 56-bit]*/
	#define SUBDEVICE_3COM_3CR990_FX_97_SERVER_TYPHOON	0x10b72102 /*3CR990-FX-97 Server [Typhoon Fiber 168-bit]*/
#define DEVICE_3COM_3CR990SVR95_TYPHOON_SERVER		0x9908 /*3CR990SVR95 [Typhoon Server 56-bit]*/
#define DEVICE_3COM_3CR990SVR97_TYPHOON_SERVER		0x9909 /*3CR990SVR97 [Typhoon Server 168-bit]*/
#define DEVICE_3COM_3C990SVR_TYPHOON_SERVER		0x990a /*3C990SVR [Typhoon Server]*/
#define DEVICE_3COM_3C990SVR_TYPHOON_SERVER2		0x990b /*3C990SVR [Typhoon Server]*/
/* Vendor: 10b8: Standard Microsystems Corp [SMC] */
#define DEVICE_STANDARD_83C170_EPIC_100_FAST		0x0005 /*83c170 EPIC/100 Fast Ethernet Adapter*/
	#define SUBDEVICE_STANDARD_LANEPIC_10_100_EVB171Q_PCI	0x1055e000 /*LANEPIC 10/100 [EVB171Q-PCI]*/
	#define SUBDEVICE_STANDARD_LANEPIC_10_100_EVB171G_PCI	0x1055e002 /*LANEPIC 10/100 [EVB171G-PCI]*/
	#define SUBDEVICE_STANDARD_ETHERPOWER_II_10_100	0x10b8a011 /*EtherPower II 10/100*/
	#define SUBDEVICE_STANDARD_ETHERPOWER_II_10_1002	0x10b8a014 /*EtherPower II 10/100*/
	#define SUBDEVICE_STANDARD_ETHERPOWER_II_10_1003	0x10b8a015 /*EtherPower II 10/100*/
	#define SUBDEVICE_STANDARD_ETHERPOWER_II_10_1004	0x10b8a016 /*EtherPower II 10/100*/
	#define SUBDEVICE_STANDARD_ETHERPOWER_II_10_1005	0x10b8a017 /*EtherPower II 10/100*/
#define DEVICE_STANDARD_83C175_EPIC_100_FAST		0x0006 /*83c175 EPIC/100 Fast Ethernet Adapter*/
	#define SUBDEVICE_STANDARD_LANEPIC_CARDBUS_FAST	0x1055e100 /*LANEPIC Cardbus Fast Ethernet Adapter*/
	#define SUBDEVICE_STANDARD_LANEPIC_CARDBUS_FAST2	0x1055e102 /*LANEPIC Cardbus Fast Ethernet Adapter*/
	#define SUBDEVICE_STANDARD_LANEPIC_CARDBUS_FAST3	0x1055e300 /*LANEPIC Cardbus Fast Ethernet Adapter*/
	#define SUBDEVICE_STANDARD_LANEPIC_CARDBUS_FAST4	0x1055e302 /*LANEPIC Cardbus Fast Ethernet Adapter*/
	#define SUBDEVICE_STANDARD_LANEPIC_CARDBUS_FAST5	0x10b8a012 /*LANEPIC Cardbus Fast Ethernet Adapter*/
	#define SUBDEVICE_STANDARD_LANEPIC_CARDBUS_FAST6	0x13a28002 /*LANEPIC Cardbus Fast Ethernet Adapter*/
	#define SUBDEVICE_STANDARD_LANEPIC_CARDBUS_FAST7	0x13a28006 /*LANEPIC Cardbus Fast Ethernet Adapter*/
#define DEVICE_STANDARD_FDC_37C665			0x1000 /*FDC 37c665*/
#define DEVICE_STANDARD_FDC_37C922			0x1001 /*FDC 37C922*/
#define DEVICE_STANDARD_SMC2802W_EZ_CONNECT		0x2802 /*SMC2802W [EZ Connect g]*/
#define DEVICE_STANDARD_83C170QF			0xa011 /*83C170QF*/
#define DEVICE_STANDARD_SMC34C90			0xb106 /*SMC34C90*/
/* Vendor: 10b9: ALi Corporation */
#define DEVICE_ALI_CMI8338_C3DX_PCI_AUDIO		0x0101 /*CMI8338/C3DX PCI Audio Device*/
#define DEVICE_ALI_C_MEDIA_CMI8738_C3DX_AUDIO		0x0111 /*C-Media CMI8738/C3DX Audio Device (OEM)*/
	#define SUBDEVICE_ALI_C_MEDIA_CMI8738_C3DX_AUDIO	0x10b90111 /*C-Media CMI8738/C3DX Audio Device (OEM)*/
#define DEVICE_ALI_MULTI_IO_CARD			0x0780 /*Multi-IO Card*/
#define DEVICE_ALI_MULTI_IO_CARD2			0x0782 /*Multi-IO Card*/
#define DEVICE_ALI_M1435					0x1435 /*M1435*/
#define DEVICE_ALI_M1445					0x1445 /*M1445*/
#define DEVICE_ALI_M1449					0x1449 /*M1449*/
#define DEVICE_ALI_M1451					0x1451 /*M1451*/
#define DEVICE_ALI_M1461					0x1461 /*M1461*/
#define DEVICE_ALI_M1489					0x1489 /*M1489*/
#define DEVICE_ALI_M1511_ALADDIN			0x1511 /*M1511 [Aladdin]*/
#define DEVICE_ALI_M1512_ALADDIN			0x1512 /*M1512 [Aladdin]*/
#define DEVICE_ALI_M1513_ALADDIN			0x1513 /*M1513 [Aladdin]*/
#define DEVICE_ALI_M1521_ALADDIN_III			0x1521 /*M1521 [Aladdin III]*/
	#define SUBDEVICE_ALI_ALI_M1521_ALADDIN		0x10b91521 /*ALI M1521 Aladdin III CPU Bridge*/
#define DEVICE_ALI_M1523					0x1523 /*M1523*/
	#define SUBDEVICE_ALI_ALI_M1523_ISA		0x10b91523 /*ALI M1523 ISA Bridge*/
#define DEVICE_ALI_M1531_ALADDIN_IV			0x1531 /*M1531 [Aladdin IV]*/
#define DEVICE_ALI_M1533_PCI_TO				0x1533 /*M1533 PCI to ISA Bridge [Aladdin IV]*/
	#define SUBDEVICE_ALI_THINKPAD_R40E_2684_HVG	0x1014053b /*ThinkPad R40e (2684-HVG) PCI to ISA Bridge*/
	#define SUBDEVICE_ALI_ALI_M1533_ALADDIN		0x10b91533 /*ALI M1533 Aladdin IV ISA Bridge*/
#define DEVICE_ALI_M1541					0x1541 /*M1541*/
	#define SUBDEVICE_ALI_ALI_M1541_ALADDIN		0x10b91541 /*ALI M1541 Aladdin V/V+ AGP System Controller*/
#define DEVICE_ALI_M1543					0x1543 /*M1543*/
#define DEVICE_ALI_M1563_HYPERTRANSPORT_SOUTH		0x1563 /*M1563 HyperTransport South Bridge*/
#define DEVICE_ALI_M1621					0x1621 /*M1621*/
#define DEVICE_ALI_ALI_M1631_PCI			0x1631 /*ALI M1631 PCI North Bridge Aladdin Pro III*/
#define DEVICE_ALI_M1632M_NORTHBRIDGE_TRIDENT		0x1632 /*M1632M Northbridge+Trident*/
#define DEVICE_ALI_ALI_M1641_PCI			0x1641 /*ALI M1641 PCI North Bridge Aladdin Pro IV*/
#define DEVICE_ALI_M1644_M1644T_NORTHBRIDGE_TRIDENT	0x1644 /*M1644/M1644T Northbridge+Trident*/
#define DEVICE_ALI_M1646_NORTHBRIDGE_TRIDENT		0x1646 /*M1646 Northbridge+Trident*/
#define DEVICE_ALI_M1647_NORTHBRIDGE_MAGIK		0x1647 /*M1647 Northbridge [MAGiK 1 / MobileMAGiK 1]*/
#define DEVICE_ALI_M1651_M1651T_NORTHBRIDGE_ALADDIN_PRO	0x1651 /*M1651/M1651T Northbridge [Aladdin-Pro 5/5M,Aladdin-Pro 5T/5TM]*/
#define DEVICE_ALI_M1671_SUPER_P4			0x1671 /*M1671 Super P4 Northbridge [AGP4X,PCI and SDR/DDR]*/
#define DEVICE_ALI_M1672_NORTHBRIDGE_CYBERALADDIN_P4	0x1672 /*M1672 Northbridge [CyberALADDiN-P4]*/
#define DEVICE_ALI_M1681_P4_NORTHBRIDGE			0x1681 /*M1681 P4 Northbridge [AGP8X,HyperTransport and SDR/DDR]*/
#define DEVICE_ALI_M1687_K8_NORTHBRIDGE			0x1687 /*M1687 K8 Northbridge [AGP8X and HyperTransport]*/
#define DEVICE_ALI_M1689_K8_NORTHBRIDGE			0x1689 /*M1689 K8 Northbridge [Super K8 Single Chip]*/
#define DEVICE_ALI_M3141					0x3141 /*M3141*/
#define DEVICE_ALI_M3143					0x3143 /*M3143*/
#define DEVICE_ALI_M3145					0x3145 /*M3145*/
#define DEVICE_ALI_M3147					0x3147 /*M3147*/
#define DEVICE_ALI_M3149					0x3149 /*M3149*/
#define DEVICE_ALI_M3151					0x3151 /*M3151*/
#define DEVICE_ALI_M3307					0x3307 /*M3307*/
#define DEVICE_ALI_M3309					0x3309 /*M3309*/
#define DEVICE_ALI_M3325_VIDEO_AUDIO_DECODER		0x3323 /*M3325 Video/Audio Decoder*/
#define DEVICE_ALI_M4803					0x5212 /*M4803*/
#define DEVICE_ALI_MS4803				0x5215 /*MS4803*/
#define DEVICE_ALI_M5217H				0x5217 /*M5217H*/
#define DEVICE_ALI_M5219					0x5219 /*M5219*/
#define DEVICE_ALI_M5225					0x5225 /*M5225*/
#define DEVICE_ALI_M5229_IDE				0x5229 /*M5229 IDE*/
	#define SUBDEVICE_ALI_THINKPAD_R30		0x1014050f /*ThinkPad R30*/
	#define SUBDEVICE_ALI_THINKPAD_R40E_2684_HVG2	0x1014053d /*ThinkPad R40e (2684-HVG) builtin IDE*/
	#define SUBDEVICE_ALI_PAVILION_ZE4400_BUILTIN	0x103c0024 /*Pavilion ze4400 builtin IDE*/
	#define SUBDEVICE_ALI_A7A266_MOTHERBOARD_IDE	0x10438053 /*A7A266 Motherboard IDE*/
#define DEVICE_ALI_M52252				0x5235 /*M5225*/
#define DEVICE_ALI_USB_1_1_CONTROLLER			0x5237 /*USB 1.1 Controller*/
	#define SUBDEVICE_ALI_THINKPAD_R40E_2684_HVG3	0x10140540 /*ThinkPad R40e (2684-HVG) builtin USB*/
	#define SUBDEVICE_ALI_PAVILION_ZE4400_BUILTIN2	0x103c0024 /*Pavilion ze4400 builtin USB*/
#define DEVICE_ALI_USB_2_0_CONTROLLER			0x5239 /*USB 2.0 Controller*/
#define DEVICE_ALI_M1541_PCI_TO				0x5243 /*M1541 PCI to AGP Controller*/
#define DEVICE_ALI_AGP8X_CONTROLLER			0x5246 /*AGP8X Controller*/
#define DEVICE_ALI_PCI_TO_AGP				0x5247 /*PCI to AGP Controller*/
#define DEVICE_ALI_M5249_HTT_TO				0x5249 /*M5249 HTT to PCI Bridge*/
#define DEVICE_ALI_M5251_P1394_OHCI			0x5251 /*M5251 P1394 OHCI 1.0 Controller*/
#define DEVICE_ALI_M5253_P1394_OHCI			0x5253 /*M5253 P1394 OHCI 1.1 Controller*/
#define DEVICE_ALI_M5261_ETHERNET_CONTROLLER		0x5261 /*M5261 Ethernet Controller*/
#define DEVICE_ALI_M5263_ETHERNET_CONTROLLER		0x5263 /*M5263 Ethernet Controller*/
#define DEVICE_ALI_ALI_M5281_SERIAL			0x5281 /*ALi M5281 Serial ATA / RAID Host Controller*/
#define DEVICE_ALI_LUCENT_TECHNOLOGIES_SOFT		0x5450 /*Lucent Technologies Soft Modem AMR*/
#define DEVICE_ALI_M5451_PCI_AC_LINK			0x5451 /*M5451 PCI AC-Link Controller Audio Device*/
	#define SUBDEVICE_ALI_THINKPAD_R302		0x10140506 /*ThinkPad R30*/
	#define SUBDEVICE_ALI_THINKPAD_R40E_2684_HVG4	0x1014053e /*ThinkPad R40e (2684-HVG) builtin Audio*/
	#define SUBDEVICE_ALI_PAVILION_ZE4400_BUILTIN3	0x103c0024 /*Pavilion ze4400 builtin Audio*/
	#define SUBDEVICE_ALI_HP_COMPAQ_NC4010		0x10b95451 /*HP Compaq nc4010 (DY885AA#ABN)*/
#define DEVICE_ALI_M5453_PCI_AC_LINK			0x5453 /*M5453 PCI AC-Link Controller Modem Device*/
#define DEVICE_ALI_M5455_PCI_AC_LINK			0x5455 /*M5455 PCI AC-Link Controller Audio Device*/
#define DEVICE_ALI_M5457_AC_97_MODEM			0x5457 /*M5457 AC'97 Modem Controller*/
	#define SUBDEVICE_ALI_THINKPAD_R40E_2684_HVG5	0x10140535 /*ThinkPad R40e (2684-HVG) builtin modem*/
	#define SUBDEVICE_ALI_PAVILION_ZE4400_BUILTIN4	0x103c0024 /*Pavilion ze4400 builtin Modem Device*/
#define DEVICE_ALI_SMARTLINK_SMARTPCI561_56K		0x5459 /*SmartLink SmartPCI561 56K Modem*/
#define DEVICE_ALI_SMARTLINK_SMARTPCI563_56K		0x545a /*SmartLink SmartPCI563 56K Modem*/
#define DEVICE_ALI_M5471_MEMORY_STICK			0x5471 /*M5471 Memory Stick Controller*/
#define DEVICE_ALI_M5473_SD_MMC_CONTROLLER		0x5473 /*M5473 SD-MMC Controller*/
#define DEVICE_ALI_M7101_POWER_MANAGEMENT		0x7101 /*M7101 Power Management Controller [PMU]*/
	#define SUBDEVICE_ALI_THINKPAD_R303		0x10140510 /*ThinkPad R30*/
	#define SUBDEVICE_ALI_THINKPAD_R40E_2684_HVG6	0x1014053c /*ThinkPad R40e (2684-HVG) Power Management Controller*/
	#define SUBDEVICE_ALI_PAVILION_ZE4400		0x103c0024 /*Pavilion ze4400*/
/* Vendor: 10ba: Mitsubishi Electric Corp. */
#define DEVICE_MITSUBISHI2_ACCELGRAPHICS_ACCELECLIPSE	0x0301 /*AccelGraphics AccelECLIPSE*/
#define DEVICE_MITSUBISHI2_ACCELGALAXY_A2100_OEM	0x0304 /*AccelGALAXY A2100 [OEM Evans & Sutherland]*/
#define DEVICE_MITSUBISHI2_TORNADO_3000_OEM		0x0308 /*Tornado 3000 [OEM Evans & Sutherland]*/
#define DEVICE_MITSUBISHI2_VG500_VOLUMEPRO_VOLUME	0x1002 /*VG500 [VolumePro Volume Rendering Accelerator]*/
/* Vendor: 10bb: Dapha Electronics Corporation */
/* Vendor: 10bc: Advanced Logic Research */
/* Vendor: 10bd: Surecom Technology */
#define DEVICE_SURECOM_NE_34				0x0e34 /*NE-34*/
/* Vendor: 10be: Tseng Labs International Co. */
/* Vendor: 10bf: Most Inc */
/* Vendor: 10c0: Boca Research Inc. */
/* Vendor: 10c1: ICM Co., Ltd. */
/* Vendor: 10c2: Auspex Systems Inc. */
/* Vendor: 10c3: Samsung Semiconductors, Inc. */
#define DEVICE_SAMSUNG2_SMARTETHER100_SC1100_LAN	0x1100 /*Smartether100 SC1100 LAN Adapter (i82557B)*/
/* Vendor: 10c4: Award Software International Inc. */
/* Vendor: 10c5: Xerox Corporation */
/* Vendor: 10c6: Rambus Inc. */
/* Vendor: 10c7: Media Vision */
/* Vendor: 10c8: Neomagic Corporation */
#define DEVICE_NEOMAGIC_NM2070_MAGICGRAPH_128		0x0001 /*NM2070 [MagicGraph 128]*/
#define DEVICE_NEOMAGIC_NM2090_MAGICGRAPH_128V		0x0002 /*NM2090 [MagicGraph 128V]*/
#define DEVICE_NEOMAGIC_NM2093_MAGICGRAPH_128ZV		0x0003 /*NM2093 [MagicGraph 128ZV]*/
#define DEVICE_NEOMAGIC_NM2160_MAGICGRAPH_128XD		0x0004 /*NM2160 [MagicGraph 128XD]*/
	#define SUBDEVICE_NEOMAGIC_MAGICGRAPH_128XD	0x101400ba /*MagicGraph 128XD*/
	#define SUBDEVICE_NEOMAGIC_MAGICGRAPH_128XD2	0x10251007 /*MagicGraph 128XD*/
	#define SUBDEVICE_NEOMAGIC_MAGICGRAPH_128XD3	0x10280074 /*MagicGraph 128XD*/
	#define SUBDEVICE_NEOMAGIC_MAGICGRAPH_128XD4	0x10280075 /*MagicGraph 128XD*/
	#define SUBDEVICE_NEOMAGIC_MAGICGRAPH_128XD5	0x1028007d /*MagicGraph 128XD*/
	#define SUBDEVICE_NEOMAGIC_MAGICGRAPH_128XD6	0x1028007e /*MagicGraph 128XD*/
	#define SUBDEVICE_NEOMAGIC_MAGICGRAPH_128XD7	0x1033802f /*MagicGraph 128XD*/
	#define SUBDEVICE_NEOMAGIC_MAGICGRAPH_128XD8	0x104d801b /*MagicGraph 128XD*/
	#define SUBDEVICE_NEOMAGIC_MAGICGRAPH_128XD9	0x104d802f /*MagicGraph 128XD*/
	#define SUBDEVICE_NEOMAGIC_MAGICGRAPH_128XD10	0x104d830b /*MagicGraph 128XD*/
	#define SUBDEVICE_NEOMAGIC_MAGICGRAPH_128XD11	0x10ba0e00 /*MagicGraph 128XD*/
	#define SUBDEVICE_NEOMAGIC_MAGICGRAPH_128XD12	0x10c80004 /*MagicGraph 128XD*/
	#define SUBDEVICE_NEOMAGIC_MAGICGRAPH_128XD13	0x10cf1029 /*MagicGraph 128XD*/
	#define SUBDEVICE_NEOMAGIC_MAGICGRAPH_128XD14	0x10f78308 /*MagicGraph 128XD*/
	#define SUBDEVICE_NEOMAGIC_MAGICGRAPH_128XD15	0x10f78309 /*MagicGraph 128XD*/
	#define SUBDEVICE_NEOMAGIC_MAGICGRAPH_128XD16	0x10f7830b /*MagicGraph 128XD*/
	#define SUBDEVICE_NEOMAGIC_MAGICGRAPH_128XD17	0x10f7830d /*MagicGraph 128XD*/
	#define SUBDEVICE_NEOMAGIC_MAGICGRAPH_128XD18	0x10f78312 /*MagicGraph 128XD*/
#define DEVICE_NEOMAGIC_NM2200_MAGICGRAPH_256AV		0x0005 /*NM2200 [MagicGraph 256AV]*/
	#define SUBDEVICE_NEOMAGIC_THINKPAD_570		0x101400dd /*ThinkPad 570*/
	#define SUBDEVICE_NEOMAGIC_LATITUDE_CPI_A	0x10280088 /*Latitude CPi A*/
#define DEVICE_NEOMAGIC_NM2360_MAGICMEDIA_256ZX		0x0006 /*NM2360 [MagicMedia 256ZX]*/
#define DEVICE_NEOMAGIC_NM2380_MAGICMEDIA_256XL		0x0016 /*NM2380 [MagicMedia 256XL+]*/
	#define SUBDEVICE_NEOMAGIC_MAGICMEDIA_256XL	0x10c80016 /*MagicMedia 256XL+*/
#define DEVICE_NEOMAGIC_NM2230_MAGICGRAPH_256AV		0x0025 /*NM2230 [MagicGraph 256AV+]*/
#define DEVICE_NEOMAGIC_NM2093_MAGICGRAPH_128ZV2	0x0083 /*NM2093 [MagicGraph 128ZV+]*/
#define DEVICE_NEOMAGIC_NM2200_MAGICMEDIA_256AV		0x8005 /*NM2200 [MagicMedia 256AV Audio]*/
	#define SUBDEVICE_NEOMAGIC_MAGICMEDIA_256AV_AUDIO	0x0e11b0d1 /*MagicMedia 256AV Audio Device on Discovery*/
	#define SUBDEVICE_NEOMAGIC_MAGICMEDIA_256AV_AUDIO2	0x0e11b126 /*MagicMedia 256AV Audio Device on Durango*/
	#define SUBDEVICE_NEOMAGIC_MAGICMEDIA_256AV_AUDIO3	0x101400dd /*MagicMedia 256AV Audio Device on BlackTip Thinkpad*/
	#define SUBDEVICE_NEOMAGIC_MAGICMEDIA_256AV_AUDIO4	0x10251003 /*MagicMedia 256AV Audio Device on TravelMate 720*/
	#define SUBDEVICE_NEOMAGIC_LATITUDE_CPI_A2	0x10280088 /*Latitude CPi A*/
	#define SUBDEVICE_NEOMAGIC_MAGICMEDIA_256AV_AUDIO5	0x1028008f /*MagicMedia 256AV Audio Device on Colorado Inspiron*/
	#define SUBDEVICE_NEOMAGIC_MAGICMEDIA_256AV_AUDIO6	0x103c0007 /*MagicMedia 256AV Audio Device on Voyager II*/
	#define SUBDEVICE_NEOMAGIC_MAGICMEDIA_256AV_AUDIO7	0x103c0008 /*MagicMedia 256AV Audio Device on Voyager III*/
	#define SUBDEVICE_NEOMAGIC_MAGICMEDIA_256AV_AUDIO8	0x103c000d /*MagicMedia 256AV Audio Device on Omnibook 900*/
	#define SUBDEVICE_NEOMAGIC_MAGICMEDIA_256AV_AUDIO9	0x10c88005 /*MagicMedia 256AV Audio Device on FireAnt*/
	#define SUBDEVICE_NEOMAGIC_MAGICMEDIA_256AV_AUDIO10	0x110a8005 /*MagicMedia 256AV Audio Device*/
	#define SUBDEVICE_NEOMAGIC_MAGICMEDIA_256AV_AUDIO11	0x14c00004 /*MagicMedia 256AV Audio Device*/
#define DEVICE_NEOMAGIC_NM2360_MAGICMEDIA_256ZX2	0x8006 /*NM2360 [MagicMedia 256ZX Audio]*/
#define DEVICE_NEOMAGIC_NM2380_MAGICMEDIA_256XL2	0x8016 /*NM2380 [MagicMedia 256XL+ Audio]*/
/* Vendor: 10c9: Dataexpert Corporation */
/* Vendor: 10ca: Fujitsu Microelectr., Inc. */
/* Vendor: 10cb: Omron Corporation */
/* Vendor: 10cc: Mai Logic Incorporated */
#define DEVICE_MAI_ARTICIA_S_HOST			0x0660 /*Articia S Host Bridge*/
#define DEVICE_MAI_ARTICIA_S_PCI			0x0661 /*Articia S PCI Bridge*/
/* Vendor: 10cd: Advanced System Products, Inc */
#define DEVICE_ADVANCED4_ASC1100			0x1100 /*ASC1100*/
#define DEVICE_ADVANCED4_ASC1200_ABP940_FAST		0x1200 /*ASC1200 [(abp940) Fast SCSI-II]*/
#define DEVICE_ADVANCED4_ABP940_U_ABP960_U		0x1300 /*ABP940-U / ABP960-U*/
	#define SUBDEVICE_ADVANCED4_ASC1300_SCSI_ADAPTER	0x10cd1310 /*ASC1300 SCSI Adapter*/
#define DEVICE_ADVANCED4_ABP940_UW			0x2300 /*ABP940-UW*/
#define DEVICE_ADVANCED4_ABP940_U2W			0x2500 /*ABP940-U2W*/
/* Vendor: 10ce: Radius */
/* Vendor: 10cf: Fujitsu Limited. */
#define DEVICE_FUJITSU2_MB86605				0x2001 /*mb86605*/
/* Vendor: 10d1: FuturePlus Systems Corp. */
/* Vendor: 10d2: Molex Incorporated */
/* Vendor: 10d3: Jabil Circuit Inc */
/* Vendor: 10d4: Hualon Microelectronics */
/* Vendor: 10d5: Autologic Inc. */
/* Vendor: 10d6: Cetia */
/* Vendor: 10d7: BCM Advanced Research */
/* Vendor: 10d8: Advanced Peripherals Labs */
/* Vendor: 10d9: Macronix, Inc. [MXIC] */
#define DEVICE_MACRONIX_MX98713				0x0512 /*MX98713*/
#define DEVICE_MACRONIX_MX987X5				0x0531 /*MX987x5*/
	#define SUBDEVICE_MACRONIX_DFE_540TX_PROFAST_10_100	0x11861200 /*DFE-540TX ProFAST 10/100 Adapter*/
#define DEVICE_MACRONIX_MX86250				0x8625 /*MX86250*/
#define DEVICE_MACRONIX_MX86200				0x8888 /*MX86200*/
/* Vendor: 10da: Compaq IPG-Austin */
#define DEVICE_COMPAQ4_TC4048_TOKEN_RING		0x0508 /*TC4048 Token Ring 4/16*/
#define DEVICE_COMPAQ4_TL3C3X9				0x3390 /*Tl3c3x9*/
/* Vendor: 10db: Rohm LSI Systems, Inc. */
/* Vendor: 10dc: CERN/ECP/EDU */
#define DEVICE_CERN_ECP_EDU_STAR_RD24_SCI_PCI_PMC	0x0001 /*STAR/RD24 SCI-PCI (PMC)*/
#define DEVICE_CERN_ECP_EDU_TAR_RD24_SCI_PCI_PMC	0x0002 /*TAR/RD24 SCI-PCI (PMC)*/
#define DEVICE_CERN_ECP_EDU_HIPPI_DESTINATION		0x0021 /*HIPPI destination*/
#define DEVICE_CERN_ECP_EDU_HIPPI_SOURCE		0x0022 /*HIPPI source*/
#define DEVICE_CERN_ECP_EDU_ATT2C15_3_FPGA		0x10dc /*ATT2C15-3 FPGA*/
/* Vendor: 10dd: Evans & Sutherland */
/* Vendor: 10de: nVidia Corporation */
#define DEVICE_NVIDIA_NV1_EDGE_3D			0x0008 /*NV1 [EDGE 3D]*/
#define DEVICE_NVIDIA_NV1_EDGE_3D2			0x0009 /*NV1 [EDGE 3D]*/
#define DEVICE_NVIDIA_NV2_MUTARA_V08			0x0010 /*NV2 [Mutara V08]*/
#define DEVICE_NVIDIA_NV4_RIVA_TNT			0x0020 /*NV4 [RIVA TNT]*/
	#define SUBDEVICE_NVIDIA_V3400_TNT		0x10430200 /*V3400 TNT*/
	#define SUBDEVICE_NVIDIA_ERAZOR_II_SGRAM	0x10480c18 /*Erazor II SGRAM*/
	#define SUBDEVICE_NVIDIA_ERAZOR_II		0x10480c1b /*Erazor II*/
	#define SUBDEVICE_NVIDIA_VIPER_V550		0x10920550 /*Viper V550*/
	#define SUBDEVICE_NVIDIA_VIPER_V5502		0x10920552 /*Viper V550*/
	#define SUBDEVICE_NVIDIA_VIPER_V5503		0x10924804 /*Viper V550*/
	#define SUBDEVICE_NVIDIA_VIPER_V5504		0x10924808 /*Viper V550*/
	#define SUBDEVICE_NVIDIA_VIPER_V5505		0x10924810 /*Viper V550*/
	#define SUBDEVICE_NVIDIA_VIPER_V5506		0x10924812 /*Viper V550*/
	#define SUBDEVICE_NVIDIA_VIPER_V5507		0x10924815 /*Viper V550*/
	#define SUBDEVICE_NVIDIA_VIPER_V550_WITH	0x10924820 /*Viper V550 with TV out*/
	#define SUBDEVICE_NVIDIA_VIPER_V5508		0x10924822 /*Viper V550*/
	#define SUBDEVICE_NVIDIA_VIPER_V5509		0x10924904 /*Viper V550*/
	#define SUBDEVICE_NVIDIA_VIPER_V55010		0x10924914 /*Viper V550*/
	#define SUBDEVICE_NVIDIA_VIPER_V55011		0x10928225 /*Viper V550*/
	#define SUBDEVICE_NVIDIA_VELOCITY_4400		0x10b4273d /*Velocity 4400*/
	#define SUBDEVICE_NVIDIA_VELOCITY_44002		0x10b4273e /*Velocity 4400*/
	#define SUBDEVICE_NVIDIA_VELOCITY_44003		0x10b42740 /*Velocity 4400*/
	#define SUBDEVICE_NVIDIA_RIVA_TNT		0x10de0020 /*Riva TNT*/
	#define SUBDEVICE_NVIDIA_GRAPHICS_BLASTER_CT6710	0x11021015 /*Graphics Blaster CT6710*/
	#define SUBDEVICE_NVIDIA_GRAPHICS_BLASTER_RIVA	0x11021016 /*Graphics Blaster RIVA TNT*/
#define DEVICE_NVIDIA_NV5_RIVA_TNT2_TNT2		0x0028 /*NV5 [RIVA TNT2/TNT2 Pro]*/
	#define SUBDEVICE_NVIDIA_AGP_V3800_SGRAM	0x10430200 /*AGP-V3800 SGRAM*/
	#define SUBDEVICE_NVIDIA_AGP_V3800_SDRAM	0x10430201 /*AGP-V3800 SDRAM*/
	#define SUBDEVICE_NVIDIA_PCI_V3800		0x10430205 /*PCI-V3800*/
	#define SUBDEVICE_NVIDIA_AGP_V3800PRO		0x10434000 /*AGP-V3800PRO*/
	#define SUBDEVICE_NVIDIA_SYNERGY_II		0x10480c21 /*Synergy II*/
	#define SUBDEVICE_NVIDIA_ERAZOR_III		0x10480c31 /*Erazor III*/
	#define SUBDEVICE_NVIDIA_WINFAST_3D_S320	0x107d2134 /*WinFast 3D S320 II + TV-Out*/
	#define SUBDEVICE_NVIDIA_VIPER_V770		0x10924804 /*Viper V770*/
	#define SUBDEVICE_NVIDIA_VIPER_V7702		0x10924a00 /*Viper V770*/
	#define SUBDEVICE_NVIDIA_VIPER_V770_ULTRA	0x10924a02 /*Viper V770 Ultra*/
	#define SUBDEVICE_NVIDIA_RIVA_TNT2_TNT2_PRO	0x10925a00 /*RIVA TNT2/TNT2 Pro*/
	#define SUBDEVICE_NVIDIA_VIPER_V770_ULTRA2	0x10926a02 /*Viper V770 Ultra*/
	#define SUBDEVICE_NVIDIA_VIPER_V770_ULTRA3	0x10927a02 /*Viper V770 Ultra*/
	#define SUBDEVICE_NVIDIA_RIVA_TNT2_PRO		0x10de0005 /*RIVA TNT2 Pro*/
	#define SUBDEVICE_NVIDIA_COMPAQ_NVIDIA_TNT2	0x10de000f /*Compaq NVIDIA TNT2 Pro*/
	#define SUBDEVICE_NVIDIA_3D_BLASTER_RIVA	0x11021020 /*3D Blaster RIVA TNT2*/
	#define SUBDEVICE_NVIDIA_3D_BLASTER_RIVA2	0x11021026 /*3D Blaster RIVA TNT2 Digital*/
	#define SUBDEVICE_NVIDIA_MAXI_GAMER_XENTOR	0x14af5810 /*Maxi Gamer Xentor*/
#define DEVICE_NVIDIA_NV5_RIVA_TNT2			0x0029 /*NV5 [RIVA TNT2 Ultra]*/
	#define SUBDEVICE_NVIDIA_AGP_V3800_DELUXE	0x10430200 /*AGP-V3800 Deluxe*/
	#define SUBDEVICE_NVIDIA_AGP_V3800_ULTRA_SDRAM	0x10430201 /*AGP-V3800 Ultra SDRAM*/
	#define SUBDEVICE_NVIDIA_PCI_V3800_ULTRA	0x10430205 /*PCI-V3800 Ultra*/
	#define SUBDEVICE_NVIDIA_3D_BLASTER_RIVA3	0x11021021 /*3D Blaster RIVA TNT2 Ultra*/
	#define SUBDEVICE_NVIDIA_3D_BLASTER_RIVA4	0x11021029 /*3D Blaster RIVA TNT2 Ultra*/
	#define SUBDEVICE_NVIDIA_3D_BLASTER_RIVA5	0x1102102f /*3D Blaster RIVA TNT2 Ultra*/
	#define SUBDEVICE_NVIDIA_MAXI_GAMER_XENTOR2	0x14af5820 /*Maxi Gamer Xentor 32*/
#define DEVICE_NVIDIA_NV5_RIVA_TNT22			0x002a /*NV5 [Riva TnT2]*/
#define DEVICE_NVIDIA_NV5_RIVA_TNT23			0x002b /*NV5 [Riva TnT2]*/
#define DEVICE_NVIDIA_NV6_VANTA_VANTA_LT		0x002c /*NV6 [Vanta/Vanta LT]*/
	#define SUBDEVICE_NVIDIA_AGP_V3800_COMBAT_SDRAM	0x10430200 /*AGP-V3800 Combat SDRAM*/
	#define SUBDEVICE_NVIDIA_AGP_V3800_COMBAT	0x10430201 /*AGP-V3800 Combat*/
	#define SUBDEVICE_NVIDIA_VIPER_V730		0x10926820 /*Viper V730*/
	#define SUBDEVICE_NVIDIA_CT6938_VANTA_8MB	0x11021031 /*CT6938 VANTA 8MB*/
	#define SUBDEVICE_NVIDIA_CT6894_VANTA_16MB	0x11021034 /*CT6894 VANTA 16MB*/
	#define SUBDEVICE_NVIDIA_MAXI_GAMER_PHOENIX	0x14af5008 /*Maxi Gamer Phoenix 2*/
#define DEVICE_NVIDIA_NV5M64_RIVA_TNT2			0x002d /*NV5M64 [RIVA TNT2 Model 64/Model 64 Pro]*/
	#define SUBDEVICE_NVIDIA_AGP_V3800M		0x10430200 /*AGP-V3800M*/
	#define SUBDEVICE_NVIDIA_AGP_V3800M2		0x10430201 /*AGP-V3800M*/
	#define SUBDEVICE_NVIDIA_ERAZOR_III_LT		0x10480c3a /*Erazor III LT*/
	#define SUBDEVICE_NVIDIA_M64_AGP4X		0x10de001e /*M64 AGP4x*/
	#define SUBDEVICE_NVIDIA_CT6892_RIVA_TNT2	0x11021023 /*CT6892 RIVA TNT2 Value*/
	#define SUBDEVICE_NVIDIA_CT6932_RIVA_TNT2	0x11021024 /*CT6932 RIVA TNT2 Value 32Mb*/
	#define SUBDEVICE_NVIDIA_CT6931_RIVA_TNT2	0x1102102c /*CT6931 RIVA TNT2 Value [Jumper]*/
	#define SUBDEVICE_NVIDIA_MSI_8808		0x14628808 /*MSI-8808*/
	#define SUBDEVICE_NVIDIA_PIXELVIEW_RIVA_TNT2	0x15541041 /*Pixelview RIVA TNT2 M64*/
#define DEVICE_NVIDIA_NV6_VANTA				0x002e /*NV6 [Vanta]*/
#define DEVICE_NVIDIA_NV6_VANTA2			0x002f /*NV6 [Vanta]*/
#define DEVICE_NVIDIA_MCP04_SMBUS			0x0034 /*MCP04 SMBus*/
#define DEVICE_NVIDIA_MCP04_IDE				0x0035 /*MCP04 IDE*/
#define DEVICE_NVIDIA_MCP04_SERIAL_ATA			0x0036 /*MCP04 Serial ATA Controller*/
#define DEVICE_NVIDIA_MCP04_ETHERNET_CONTROLLER		0x0037 /*MCP04 Ethernet Controller*/
#define DEVICE_NVIDIA_MCP04_ETHERNET_CONTROLLER2	0x0038 /*MCP04 Ethernet Controller*/
#define DEVICE_NVIDIA_MCP04_AC_97_AUDIO			0x003a /*MCP04 AC'97 Audio Controller*/
#define DEVICE_NVIDIA_MCP04_USB_CONTROLLER		0x003b /*MCP04 USB Controller*/
#define DEVICE_NVIDIA_MCP04_USB_CONTROLLER2		0x003c /*MCP04 USB Controller*/
#define DEVICE_NVIDIA_MCP04_PCI_BRIDGE			0x003d /*MCP04 PCI Bridge*/
#define DEVICE_NVIDIA_MCP04_SERIAL_ATA2			0x003e /*MCP04 Serial ATA Controller*/
#define DEVICE_NVIDIA_NV40_GEFORCE_6800			0x0040 /*nv40 [GeForce 6800 Ultra]*/
#define DEVICE_NVIDIA_NV40_GEFORCE_68002		0x0041 /*NV40 [GeForce 6800]*/
#define DEVICE_NVIDIA_NV40_2				0x0042 /*NV40.2*/
#define DEVICE_NVIDIA_NV40_3				0x0043 /*NV40.3*/
#define DEVICE_NVIDIA_NV40_GEFORCE_68003		0x0045 /*NV40 [GeForce 6800 GT]*/
#define DEVICE_NVIDIA_NV40GL				0x0049 /*NV40GL*/
#define DEVICE_NVIDIA_NV40GL_QUADRO_FX			0x004e /*NV40GL [Quadro FX 4000]*/
#define DEVICE_NVIDIA_CK804_SMBUS			0x0052 /*CK804 SMBus*/
#define DEVICE_NVIDIA_CK804_IDE				0x0053 /*CK804 IDE*/
#define DEVICE_NVIDIA_CK804_SERIAL_ATA			0x0054 /*CK804 Serial ATA Controller*/
#define DEVICE_NVIDIA_CK804_SERIAL_ATA2			0x0055 /*CK804 Serial ATA Controller*/
#define DEVICE_NVIDIA_CK804_ETHERNET_CONTROLLER		0x0056 /*CK804 Ethernet Controller*/
#define DEVICE_NVIDIA_CK804_ETHERNET_CONTROLLER2	0x0057 /*CK804 Ethernet Controller*/
#define DEVICE_NVIDIA_CK804_AC_97_AUDIO			0x0059 /*CK804 AC'97 Audio Controller*/
#define DEVICE_NVIDIA_CK804_USB_CONTROLLER		0x005a /*CK804 USB Controller*/
#define DEVICE_NVIDIA_CK804_USB_CONTROLLER2		0x005b /*CK804 USB Controller*/
#define DEVICE_NVIDIA_CK804_PCI_BRIDGE			0x005c /*CK804 PCI Bridge*/
#define DEVICE_NVIDIA_CK804_PCIE_BRIDGE			0x005d /*CK804 PCIE Bridge*/
#define DEVICE_NVIDIA_CK804_MEMORY_CONTROLLER		0x005e /*CK804 Memory Controller*/
#define DEVICE_NVIDIA_NFORCE2_ISA_BRIDGE		0x0060 /*nForce2 ISA Bridge*/
	#define SUBDEVICE_NVIDIA_A7N8X_MAINBOARD	0x104380ad /*A7N8X Mainboard*/
#define DEVICE_NVIDIA_NFORCE2_SMBUS_MCP			0x0064 /*nForce2 SMBus (MCP)*/
#define DEVICE_NVIDIA_NFORCE2_IDE			0x0065 /*nForce2 IDE*/
#define DEVICE_NVIDIA_NFORCE2_ETHERNET_CONTROLLER	0x0066 /*nForce2 Ethernet Controller*/
	#define SUBDEVICE_NVIDIA_A7N8X_MAINBOARD_ONBOARD	0x104380a7 /*A7N8X Mainboard onboard nForce2 Ethernet*/
#define DEVICE_NVIDIA_NFORCE2_USB_CONTROLLER		0x0067 /*nForce2 USB Controller*/
	#define SUBDEVICE_NVIDIA_A7N8X_MAINBOARD2	0x10430c11 /*A7N8X Mainboard*/
#define DEVICE_NVIDIA_NFORCE2_USB_CONTROLLER2		0x0068 /*nForce2 USB Controller*/
	#define SUBDEVICE_NVIDIA_A7N8X_MAINBOARD3	0x10430c11 /*A7N8X Mainboard*/
#define DEVICE_NVIDIA_NFORCE2_AC97_AUDIO		0x006a /*nForce2 AC97 Audio Controler (MCP)*/
#define DEVICE_NVIDIA_NFORCE_MULTIMEDIA_AUDIO		0x006b /*nForce MultiMedia audio [Via VT82C686B]*/
	#define SUBDEVICE_NVIDIA_NFORCE2_MCP_AUDIO	0x10de006b /*nForce2 MCP Audio Processing Unit*/
#define DEVICE_NVIDIA_NFORCE2_EXTERNAL_PCI		0x006c /*nForce2 External PCI Bridge*/
#define DEVICE_NVIDIA_NFORCE2_PCI_BRIDGE		0x006d /*nForce2 PCI Bridge*/
#define DEVICE_NVIDIA_NFORCE2_FIREWIRE_IEEE		0x006e /*nForce2 FireWire (IEEE 1394) Controller*/
#define DEVICE_NVIDIA_MCP2A_SMBUS			0x0084 /*MCP2A SMBus*/
#define DEVICE_NVIDIA_MCP2A_IDE				0x0085 /*MCP2A IDE*/
#define DEVICE_NVIDIA_MCP2A_ETHERNET_CONTROLLER		0x0086 /*MCP2A Ethernet Controller*/
#define DEVICE_NVIDIA_MCP2A_USB_CONTROLLER		0x0087 /*MCP2A USB Controller*/
#define DEVICE_NVIDIA_MCP2A_USB_CONTROLLER2		0x0088 /*MCP2A USB Controller*/
#define DEVICE_NVIDIA_MCP2S_AC_97_AUDIO			0x008a /*MCP2S AC'97 Audio Controller*/
#define DEVICE_NVIDIA_MCP2A_PCI_BRIDGE			0x008b /*MCP2A PCI Bridge*/
#define DEVICE_NVIDIA_MCP2A_ETHERNET_CONTROLLER2	0x008c /*MCP2A Ethernet Controller*/
#define DEVICE_NVIDIA_NFORCE2_SERIAL_ATA		0x008e /*nForce2 Serial ATA Controller*/
#define DEVICE_NVIDIA_NV5_ALADDIN_TNT2			0x00a0 /*NV5 [Aladdin TNT2]*/
	#define SUBDEVICE_NVIDIA_MAXI_GAMER_XENTOR3	0x14af5810 /*Maxi Gamer Xentor*/
#define DEVICE_NVIDIA_NV41_0				0x00c0 /*NV41.0*/
#define DEVICE_NVIDIA_NV41_1				0x00c1 /*NV41.1*/
#define DEVICE_NVIDIA_NV41_2				0x00c2 /*NV41.2*/
#define DEVICE_NVIDIA_NV41_8				0x00c8 /*NV41.8*/
#define DEVICE_NVIDIA_NV41GL				0x00ce /*NV41GL*/
#define DEVICE_NVIDIA_NFORCE3_LPC_BRIDGE		0x00d0 /*nForce3 LPC Bridge*/
#define DEVICE_NVIDIA_NFORCE3_HOST_BRIDGE		0x00d1 /*nForce3 Host Bridge*/
#define DEVICE_NVIDIA_NFORCE3_AGP_BRIDGE		0x00d2 /*nForce3 AGP Bridge*/
#define DEVICE_NVIDIA_CK804_MEMORY_CONTROLLER2		0x00d3 /*CK804 Memory Controller*/
#define DEVICE_NVIDIA_NFORCE3_SMBUS			0x00d4 /*nForce3 SMBus*/
#define DEVICE_NVIDIA_NFORCE3_IDE			0x00d5 /*nForce3 IDE*/
#define DEVICE_NVIDIA_NFORCE3_ETHERNET			0x00d6 /*nForce3 Ethernet*/
#define DEVICE_NVIDIA_NFORCE3_USB_1_1			0x00d7 /*nForce3 USB 1.1*/
#define DEVICE_NVIDIA_NFORCE3_USB_2_0			0x00d8 /*nForce3 USB 2.0*/
#define DEVICE_NVIDIA_NFORCE3_AUDIO			0x00da /*nForce3 Audio*/
#define DEVICE_NVIDIA_NFORCE3_PCI_BRIDGE		0x00dd /*nForce3 PCI Bridge*/
#define DEVICE_NVIDIA_CK8S_ETHERNET_CONTROLLER		0x00df /*CK8S Ethernet Controller*/
#define DEVICE_NVIDIA_NFORCE3_250GB_HOST		0x00e1 /*nForce3 250Gb Host Bridge*/
#define DEVICE_NVIDIA_NFORCE3_250GB_AGP			0x00e2 /*nForce3 250Gb AGP Host to PCI Bridge*/
#define DEVICE_NVIDIA_CK8S_SERIAL_ATA			0x00e3 /*CK8S Serial ATA Controller (v2.5)*/
#define DEVICE_NVIDIA_NFORCE_250GB_PCI			0x00e4 /*nForce 250Gb PCI System Management*/
#define DEVICE_NVIDIA_CK8S_PARALLEL_ATA			0x00e5 /*CK8S Parallel ATA Controller (v2.5)*/
#define DEVICE_NVIDIA_CK8S_ETHERNET_CONTROLLER2		0x00e6 /*CK8S Ethernet Controller*/
#define DEVICE_NVIDIA_CK8S_USB_CONTROLLER		0x00e7 /*CK8S USB Controller*/
#define DEVICE_NVIDIA_CK8S_USB_CONTROLLER2		0x00e8 /*CK8S USB Controller*/
#define DEVICE_NVIDIA_NFORCE3_250GB_AC_97		0x00ea /*nForce3 250Gb AC'97 Audio Controller*/
#define DEVICE_NVIDIA_NFORCE3_250GB_PCI_TO_PCI		0x00ed /*nForce3 250Gb PCI-to-PCI Bridge*/
#define DEVICE_NVIDIA_CK8S_SERIAL_ATA2			0x00ee /*CK8S Serial ATA Controller (v2.5)*/
#define DEVICE_NVIDIA_NV40_GEFORCE_6800_GEFORCE		0x00f0 /*NV40 [GeForce 6800/GeForce 6800 Ultra]*/
#define DEVICE_NVIDIA_NV43_GEFORCE_6600_GEFORCE		0x00f1 /*NV43 [GeForce 6600/GeForce 6600 GT]*/
#define DEVICE_NVIDIA_NV43_GEFORCE_6600			0x00f2 /*NV43 [GeForce 6600 GT]*/
#define DEVICE_NVIDIA_NV45GL_QUADRO_FX			0x00f8 /*NV45GL [Quadro FX 3400]*/
#define DEVICE_NVIDIA_NV40_GEFORCE_68004		0x00f9 /*NV40 [GeForce 6800 Ultra]*/
#define DEVICE_NVIDIA_NV36_GEFORCE_PCX			0x00fa /*NV36 [GeForce PCX 5750]*/
#define DEVICE_NVIDIA_NV35_GEFORCE_PCX			0x00fb /*NV35 [GeForce PCX 5900]*/
#define DEVICE_NVIDIA_NV37GL_QUADRO_FX			0x00fc /*NV37GL [Quadro FX 330/GeForce PCX 5300]*/
#define DEVICE_NVIDIA_NV37GL_QUADRO_FX2			0x00fd /*NV37GL [Quadro FX 330]*/
#define DEVICE_NVIDIA_NV38GL_QUADRO_FX			0x00fe /*NV38GL [Quadro FX 1300]*/
#define DEVICE_NVIDIA_NV18_GEFORCE_PCX			0x00ff /*NV18 [GeForce PCX 4300]*/
#define DEVICE_NVIDIA_NV10_GEFORCE_256			0x0100 /*NV10 [GeForce 256 SDR]*/
	#define SUBDEVICE_NVIDIA_AGP_V6600_SGRAM	0x10430200 /*AGP-V6600 SGRAM*/
	#define SUBDEVICE_NVIDIA_AGP_V6600_SDRAM	0x10430201 /*AGP-V6600 SDRAM*/
	#define SUBDEVICE_NVIDIA_AGP_V6600_SGRAM2	0x10434008 /*AGP-V6600 SGRAM*/
	#define SUBDEVICE_NVIDIA_AGP_V6600_SDRAM2	0x10434009 /*AGP-V6600 SDRAM*/
	#define SUBDEVICE_NVIDIA_CT6941_GEFORCE_256	0x1102102d /*CT6941 GeForce 256*/
	#define SUBDEVICE_NVIDIA_3D_PROPHET_SE		0x14af5022 /*3D Prophet SE*/
#define DEVICE_NVIDIA_NV10DDR_GEFORCE_256		0x0101 /*NV10DDR [GeForce 256 DDR]*/
	#define SUBDEVICE_NVIDIA_AGP_V6800_DDR		0x10430202 /*AGP-V6800 DDR*/
	#define SUBDEVICE_NVIDIA_AGP_V6800_DDR_SGRAM	0x1043400a /*AGP-V6800 DDR SGRAM*/
	#define SUBDEVICE_NVIDIA_AGP_V6800_DDR_SDRAM	0x1043400b /*AGP-V6800 DDR SDRAM*/
	#define SUBDEVICE_NVIDIA_WINFAST_GEFORCE_256	0x107d2822 /*WinFast GeForce 256*/
	#define SUBDEVICE_NVIDIA_CT6971_GEFORCE_256	0x1102102e /*CT6971 GeForce 256 DDR*/
	#define SUBDEVICE_NVIDIA_3D_PROPHET_DDR_DVI	0x14af5021 /*3D Prophet DDR-DVI*/
#define DEVICE_NVIDIA_NV10GL_QUADRO			0x0103 /*NV10GL [Quadro]*/
#define DEVICE_NVIDIA_NV11_GEFORCE2_MX_MX		0x0110 /*NV11 [GeForce2 MX/MX 400]*/
	#define SUBDEVICE_NVIDIA_AGP_V7100_PRO		0x10434015 /*AGP-V7100 Pro*/
	#define SUBDEVICE_NVIDIA_V7100_PRO_WITH		0x10434031 /*V7100 Pro with TV output*/
	#define SUBDEVICE_NVIDIA_DELL_OEM_GEFORCE	0x10de0091 /*Dell OEM GeForce 2 MX 400*/
	#define SUBDEVICE_NVIDIA_MSI_GEFORCE2_MX400	0x14628817 /*MSI GeForce2 MX400 Pro32S [MS-8817]*/
	#define SUBDEVICE_NVIDIA_3D_PROPHET_II		0x14af7102 /*3D Prophet II MX*/
	#define SUBDEVICE_NVIDIA_3D_PROPHET_II2		0x14af7103 /*3D Prophet II MX Dual-Display*/
#define DEVICE_NVIDIA_NV11DDR_GEFORCE2_MX		0x0111 /*NV11DDR [GeForce2 MX 100 DDR/200 DDR]*/
#define DEVICE_NVIDIA_NV11_GEFORCE2_GO			0x0112 /*NV11 [GeForce2 Go]*/
#define DEVICE_NVIDIA_NV11GL_QUADRO2_MXR_EX		0x0113 /*NV11GL [Quadro2 MXR/EX]*/
#define DEVICE_NVIDIA_NV15_GEFORCE2_GTS_PRO		0x0150 /*NV15 [GeForce2 GTS/Pro]*/
	#define SUBDEVICE_NVIDIA_V7700_AGP_VIDEO	0x10434016 /*V7700 AGP Video Card*/
	#define SUBDEVICE_NVIDIA_WINFAST_GEFORCE2_GTS	0x107d2840 /*WinFast GeForce2 GTS with TV output*/
	#define SUBDEVICE_NVIDIA_WINFAST_GEFORCE_2	0x107d2842 /*WinFast GeForce 2 Pro*/
	#define SUBDEVICE_NVIDIA_CREATIVE_GEFORCE2_PRO	0x14628831 /*Creative GeForce2 Pro*/
#define DEVICE_NVIDIA_NV15DDR_GEFORCE2_TI		0x0151 /*NV15DDR [GeForce2 Ti]*/
	#define SUBDEVICE_NVIDIA_V7700TI		0x1043405f /*V7700Ti*/
	#define SUBDEVICE_NVIDIA_CREATIVE_3D_BLASTER	0x14625506 /*Creative 3D Blaster Geforce2 Titanium*/
#define DEVICE_NVIDIA_NV15BR_GEFORCE2_ULTRA		0x0152 /*NV15BR [GeForce2 Ultra, Bladerunner]*/
	#define SUBDEVICE_NVIDIA_GLADIAC_ULTRA		0x10480c56 /*GLADIAC Ultra*/
#define DEVICE_NVIDIA_NV15GL_QUADRO2_PRO		0x0153 /*NV15GL [Quadro2 Pro]*/
#define DEVICE_NVIDIA_NV17_GEFORCE4_MX			0x0170 /*NV17 [GeForce4 MX 460]*/
#define DEVICE_NVIDIA_NV17_GEFORCE4_MX2			0x0171 /*NV17 [GeForce4 MX 440]*/
	#define SUBDEVICE_NVIDIA_GAINWARD_PRO_600_TV	0x10b00002 /*Gainward Pro/600 TV*/
	#define SUBDEVICE_NVIDIA_G4MX440_VTP		0x14628661 /*G4MX440-VTP*/
	#define SUBDEVICE_NVIDIA_MX440SES_T_MS_8873	0x14628730 /*MX440SES-T (MS-8873)*/
	#define SUBDEVICE_NVIDIA_ABIT_SILURO_GEFORCE4MX440	0x147b8f00 /*Abit Siluro GeForce4MX440*/
#define DEVICE_NVIDIA_NV17_GEFORCE4_MX3			0x0172 /*NV17 [GeForce4 MX 420]*/
#define DEVICE_NVIDIA_NV17_GEFORCE4_MX4			0x0173 /*NV17 [GeForce4 MX 440-SE]*/
#define DEVICE_NVIDIA_NV17_GEFORCE4_440			0x0174 /*NV17 [GeForce4 440 Go]*/
#define DEVICE_NVIDIA_NV17_GEFORCE4_420			0x0175 /*NV17 [GeForce4 420 Go]*/
#define DEVICE_NVIDIA_NV17_GEFORCE4_4202		0x0176 /*NV17 [GeForce4 420 Go 32M]*/
	#define SUBDEVICE_NVIDIA_CX9_VX9		0x4c531090 /*Cx9 / Vx9 mainboard*/
#define DEVICE_NVIDIA_NV17_GEFORCE4_460			0x0177 /*NV17 [GeForce4 460 Go]*/
#define DEVICE_NVIDIA_NV17GL_QUADRO4_550		0x0178 /*NV17GL [Quadro4 550 XGL]*/
#define DEVICE_NVIDIA_NV17_GEFORCE4_4402		0x0179 /*NV17 [GeForce4 440 Go 64M]*/
	#define SUBDEVICE_NVIDIA_GEFORCE4_MX_MAC	0x10de0179 /*GeForce4 MX (Mac)*/
#define DEVICE_NVIDIA_NV17GL_QUADRO4_200_400		0x017a /*NV17GL [Quadro4 200/400 NVS]*/
#define DEVICE_NVIDIA_NV17GL_QUADRO4_5502		0x017b /*NV17GL [Quadro4 550 XGL]*/
#define DEVICE_NVIDIA_NV17GL_QUADRO4_5503		0x017c /*NV17GL [Quadro4 550 GoGL]*/
#define DEVICE_NVIDIA_NV17_GEFORCE4_410			0x017d /*NV17 [GeForce4 410 Go 16M]*/
#define DEVICE_NVIDIA_NV18_GEFORCE4_MX			0x0181 /*NV18 [GeForce4 MX 440 AGP 8x]*/
	#define SUBDEVICE_NVIDIA_V9180_MAGIC		0x1043806f /*V9180 Magic*/
	#define SUBDEVICE_NVIDIA_MS_STARFORCE_GEFORCE4_MX	0x14628880 /*MS-StarForce GeForce4 MX 440 with AGP8X*/
	#define SUBDEVICE_NVIDIA_MS_8890_GEFORCE_4	0x14628900 /*MS-8890 GeForce 4 MX440 AGP8X*/
	#define SUBDEVICE_NVIDIA_MSI_GEFORCE4_MX	0x14629350 /*MSI Geforce4 MX T8X with AGP8X*/
	#define SUBDEVICE_NVIDIA_SILURO_GF4_MX_8X	0x147b8f0d /*Siluro GF4 MX-8X*/
#define DEVICE_NVIDIA_NV18_GEFORCE4_MX2			0x0182 /*NV18 [GeForce4 MX 440SE AGP 8x]*/
#define DEVICE_NVIDIA_NV18_GEFORCE4_MX3			0x0183 /*NV18 [GeForce4 MX 420 AGP 8x]*/
#define DEVICE_NVIDIA_NV18_GEFORCE4_MX4			0x0185 /*NV18 [GeForce4 MX 4000 AGP 8x]*/
#define DEVICE_NVIDIA_NV18M_GEFORCE4_448		0x0186 /*NV18M [GeForce4 448 Go]*/
#define DEVICE_NVIDIA_NV18M_GEFORCE4_488		0x0187 /*NV18M [GeForce4 488 Go]*/
#define DEVICE_NVIDIA_NV18GL_QUADRO4_580		0x0188 /*NV18GL [Quadro4 580 XGL]*/
#define DEVICE_NVIDIA_NV18GL_QUADRO4_NVS		0x018a /*NV18GL [Quadro4 NVS AGP 8x]*/
#define DEVICE_NVIDIA_NV18GL_QUADRO4_380		0x018b /*NV18GL [Quadro4 380 XGL]*/
#define DEVICE_NVIDIA_NV18M_GEFORCE4_4482		0x018d /*NV18M [GeForce4 448 Go]*/
#define DEVICE_NVIDIA_NVCRUSH11_GEFORCE2_MX		0x01a0 /*NVCrush11 [GeForce2 MX Integrated Graphics]*/
#define DEVICE_NVIDIA_NFORCE_CPU_BRIDGE			0x01a4 /*nForce CPU bridge*/
#define DEVICE_NVIDIA_NFORCE_420_MEMORY			0x01ab /*nForce 420 Memory Controller (DDR)*/
#define DEVICE_NVIDIA_NFORCE_220_420_MEMORY		0x01ac /*nForce 220/420 Memory Controller*/
#define DEVICE_NVIDIA_NFORCE_220_420_MEMORY2		0x01ad /*nForce 220/420 Memory Controller*/
#define DEVICE_NVIDIA_NFORCE_AUDIO			0x01b0 /*nForce Audio*/
#define DEVICE_NVIDIA_NFORCE_AUDIO2			0x01b1 /*nForce Audio*/
#define DEVICE_NVIDIA_NFORCE_ISA_BRIDGE			0x01b2 /*nForce ISA Bridge*/
#define DEVICE_NVIDIA_NFORCE_PCI_SYSTEM			0x01b4 /*nForce PCI System Management*/
#define DEVICE_NVIDIA_NFORCE_AGP_TO			0x01b7 /*nForce AGP to PCI Bridge*/
#define DEVICE_NVIDIA_NFORCE_PCI_TO_PCI_BRIDGE		0x01b8 /*nForce PCI-to-PCI bridge*/
#define DEVICE_NVIDIA_NFORCE_IDE			0x01bc /*nForce IDE*/
#define DEVICE_NVIDIA_NFORCE_AC_97_MODEM		0x01c1 /*nForce AC'97 Modem Controller*/
#define DEVICE_NVIDIA_NFORCE_USB_CONTROLLER		0x01c2 /*nForce USB Controller*/
#define DEVICE_NVIDIA_NFORCE_ETHERNET_CONTROLLER	0x01c3 /*nForce Ethernet Controller*/
#define DEVICE_NVIDIA_NFORCE2_AGP_DIFFERENT		0x01e0 /*nForce2 AGP (different version?)*/
#define DEVICE_NVIDIA_NFORCE2_AGP			0x01e8 /*nForce2 AGP*/
#define DEVICE_NVIDIA_NFORCE2_MEMORY_CONTROLLER		0x01ea /*nForce2 Memory Controller 0*/
#define DEVICE_NVIDIA_NFORCE2_MEMORY_CONTROLLER2	0x01eb /*nForce2 Memory Controller 1*/
#define DEVICE_NVIDIA_NFORCE2_MEMORY_CONTROLLER3	0x01ec /*nForce2 Memory Controller 2*/
#define DEVICE_NVIDIA_NFORCE2_MEMORY_CONTROLLER4	0x01ed /*nForce2 Memory Controller 3*/
#define DEVICE_NVIDIA_NFORCE2_MEMORY_CONTROLLER5	0x01ee /*nForce2 Memory Controller 4*/
#define DEVICE_NVIDIA_NFORCE2_MEMORY_CONTROLLER6	0x01ef /*nForce2 Memory Controller 5*/
#define DEVICE_NVIDIA_NV18_GEFORCE4_MX5			0x01f0 /*NV18 [GeForce4 MX - nForce GPU]*/
#define DEVICE_NVIDIA_NV20_GEFORCE3			0x0200 /*NV20 [GeForce3]*/
	#define SUBDEVICE_NVIDIA_AGP_V8200_DDR		0x1043402f /*AGP-V8200 DDR*/
#define DEVICE_NVIDIA_NV20_GEFORCE3_TI			0x0201 /*NV20 [GeForce3 Ti 200]*/
#define DEVICE_NVIDIA_NV20_GEFORCE3_TI2			0x0202 /*NV20 [GeForce3 Ti 500]*/
	#define SUBDEVICE_NVIDIA_V8200_T5		0x1043405b /*V8200 T5*/
	#define SUBDEVICE_NVIDIA_XTASY_6964		0x1545002f /*Xtasy 6964*/
#define DEVICE_NVIDIA_NV20DCC_QUADRO_DCC		0x0203 /*NV20DCC [Quadro DCC]*/
#define DEVICE_NVIDIA_NV25_GEFORCE4_TI			0x0250 /*NV25 [GeForce4 Ti 4600]*/
#define DEVICE_NVIDIA_NV25_GEFORCE4_TI2			0x0251 /*NV25 [GeForce4 Ti 4400]*/
	#define SUBDEVICE_NVIDIA_V8440_GEFORCE_4	0x10438023 /*v8440 GeForce 4 Ti4400*/
#define DEVICE_NVIDIA_NV25_GEFORCE4_TI3			0x0252 /*NV25 [GeForce4 Ti]*/
#define DEVICE_NVIDIA_NV25_GEFORCE4_TI4			0x0253 /*NV25 [GeForce4 Ti 4200]*/
	#define SUBDEVICE_NVIDIA_WINFAST_A250_LE	0x107d2896 /*WinFast A250 LE TD (Dual VGA/TV-out/DVI)*/
	#define SUBDEVICE_NVIDIA_SILURO_DUAL_VGA_TV_OUT_DVI	0x147b8f09 /*Siluro (Dual VGA/TV-out/DVI)*/
#define DEVICE_NVIDIA_NV25GL_QUADRO4_900		0x0258 /*NV25GL [Quadro4 900 XGL]*/
#define DEVICE_NVIDIA_NV25GL_QUADRO4_750		0x0259 /*NV25GL [Quadro4 750 XGL]*/
#define DEVICE_NVIDIA_NV25GL_QUADRO4_700		0x025b /*NV25GL [Quadro4 700 XGL]*/
#define DEVICE_NVIDIA_NV28_GEFORCE4_TI			0x0280 /*NV28 [GeForce4 Ti 4800]*/
#define DEVICE_NVIDIA_NV28_GEFORCE4_TI2			0x0281 /*NV28 [GeForce4 Ti 4200 AGP 8x]*/
#define DEVICE_NVIDIA_NV28_GEFORCE4_TI3			0x0282 /*NV28 [GeForce4 Ti 4800 SE]*/
#define DEVICE_NVIDIA_NV28_GEFORCE4_TI4			0x0286 /*NV28 [GeForce4 Ti 4200 Go AGP 8x]*/
#define DEVICE_NVIDIA_NV28GL_QUADRO4_980		0x0288 /*NV28GL [Quadro4 980 XGL]*/
#define DEVICE_NVIDIA_NV28GL_QUADRO4_780		0x0289 /*NV28GL [Quadro4 780 XGL]*/
#define DEVICE_NVIDIA_NV28GLM_QUADRO4_700		0x028c /*NV28GLM [Quadro4 700 GoGL]*/
#define DEVICE_NVIDIA_NV30_GEFORCE_FX			0x0300 /*NV30 [GeForce FX]*/
#define DEVICE_NVIDIA_NV30_GEFORCE_FX2			0x0301 /*NV30 [GeForce FX 5800 Ultra]*/
#define DEVICE_NVIDIA_NV30_GEFORCE_FX3			0x0302 /*NV30 [GeForce FX 5800]*/
#define DEVICE_NVIDIA_NV30GL_QUADRO_FX			0x0308 /*NV30GL [Quadro FX 2000]*/
#define DEVICE_NVIDIA_NV30GL_QUADRO_FX2			0x0309 /*NV30GL [Quadro FX 1000]*/
#define DEVICE_NVIDIA_NV31_GEFORCE_FX			0x0311 /*NV31 [GeForce FX 5600 Ultra]*/
#define DEVICE_NVIDIA_NV31_GEFORCE_FX2			0x0312 /*NV31 [GeForce FX 5600]*/
#define DEVICE_NVIDIA_NV31				0x0313 /*NV31*/
#define DEVICE_NVIDIA_NV31_GEFORCE_FX3			0x0314 /*NV31 [GeForce FX 5600XT]*/
	#define SUBDEVICE_NVIDIA_V9560XT_TD		0x1043814a /*V9560XT/TD*/
#define DEVICE_NVIDIA_NV312				0x0316 /*NV31*/
#define DEVICE_NVIDIA_NV313				0x0317 /*NV31*/
#define DEVICE_NVIDIA_NV31M_GEFORCE_FX			0x031a /*NV31M [GeForce FX Go 5600]*/
#define DEVICE_NVIDIA_NV31M_GEFORCE_FX2			0x031b /*NV31M [GeForce FX Go5650]*/
#define DEVICE_NVIDIA_NVIDIA_QUADRO_FX			0x031c /*NVIDIA Quadro FX 700 Go*/
#define DEVICE_NVIDIA_NV314				0x031d /*NV31*/
#define DEVICE_NVIDIA_NV315				0x031e /*NV31*/
#define DEVICE_NVIDIA_NV316				0x031f /*NV31*/
#define DEVICE_NVIDIA_NV34_GEFORCE_FX			0x0320 /*NV34 [GeForce FX 5200]*/
#define DEVICE_NVIDIA_NV34_GEFORCE_FX2			0x0321 /*NV34 [GeForce FX 5200 Ultra]*/
#define DEVICE_NVIDIA_NV34_GEFORCE_FX3			0x0322 /*NV34 [GeForce FX 5200]*/
	#define SUBDEVICE_NVIDIA_MS_8917_FX5200_T128	0x14629171 /*MS-8917 (FX5200-T128)*/
#define DEVICE_NVIDIA_NV34_GEFORCE_FX4			0x0323 /*NV34 [GeForce FX 5200LE]*/
#define DEVICE_NVIDIA_NV34M_GEFORCE_FX			0x0324 /*NV34M [GeForce FX Go 5200]*/
	#define SUBDEVICE_NVIDIA_MIM2000		0x10718160 /*MIM2000*/
#define DEVICE_NVIDIA_NV34M_GEFORCE_FX2			0x0325 /*NV34M [GeForce FX Go5250]*/
#define DEVICE_NVIDIA_NV34_GEFORCE_FX5			0x0326 /*NV34 [GeForce FX 5500]*/
#define DEVICE_NVIDIA_NV34_GEFORCE_FX6			0x0327 /*NV34 [GeForce FX 5100]*/
#define DEVICE_NVIDIA_NV34M_GEFORCE_FX3			0x0328 /*NV34M [GeForce FX Go 5200]*/
#define DEVICE_NVIDIA_NV34M_GEFORCE_FX4			0x0329 /*NV34M [GeForce FX Go5200]*/
#define DEVICE_NVIDIA_NV34GL_QUADRO_NVS			0x032a /*NV34GL [Quadro NVS 280 PCI]*/
#define DEVICE_NVIDIA_NV34GL_QUADRO_FX			0x032b /*NV34GL [Quadro FX 500/600 PCI]*/
#define DEVICE_NVIDIA_NV34GLM_GEFORCE_FX		0x032c /*NV34GLM [GeForce FX Go 5300]*/
#define DEVICE_NVIDIA_NV34_GEFORCE_FX7			0x032d /*NV34 [GeForce FX Go5100]*/
#define DEVICE_NVIDIA_NV34				0x032f /*NV34*/
#define DEVICE_NVIDIA_NV35_GEFORCE_FX			0x0330 /*NV35 [GeForce FX 5900 Ultra]*/
#define DEVICE_NVIDIA_NV35_GEFORCE_FX2			0x0331 /*NV35 [GeForce FX 5900]*/
	#define SUBDEVICE_NVIDIA_V9950GE		0x10438145 /*V9950GE*/
#define DEVICE_NVIDIA_NV35_GEFORCE_FX3			0x0332 /*NV35 [GeForce FX 5900XT]*/
#define DEVICE_NVIDIA_NV38_GEFORCE_FX			0x0333 /*NV38 [GeForce FX 5950 Ultra]*/
#define DEVICE_NVIDIA_NV35_GEFORCE_FX4			0x0334 /*NV35 [GeForce FX 5900ZT]*/
#define DEVICE_NVIDIA_NV35GL_QUADRO_FX			0x0338 /*NV35GL [Quadro FX 3000]*/
#define DEVICE_NVIDIA_NV35GL_QUADRO_FX2			0x033f /*NV35GL [Quadro FX 700]*/
#define DEVICE_NVIDIA_NV36_1_GEFORCE_FX			0x0341 /*NV36.1 [GeForce FX 5700 Ultra]*/
#define DEVICE_NVIDIA_NV36_2_GEFORCE_FX			0x0342 /*NV36.2 [GeForce FX 5700]*/
#define DEVICE_NVIDIA_NV36_GEFORCE_FX			0x0343 /*NV36 [GeForce FX 5700LE]*/
#define DEVICE_NVIDIA_NV36_4_GEFORCE_FX			0x0344 /*NV36.4 [GeForce FX 5700VE]*/
#define DEVICE_NVIDIA_NV36_5				0x0345 /*NV36.5*/
#define DEVICE_NVIDIA_NV36_GEFORCE_FX2			0x0347 /*NV36 [GeForce FX Go5700]*/
#define DEVICE_NVIDIA_NV36_GEFORCE_FX3			0x0348 /*NV36 [GeForce FX Go5700]*/
#define DEVICE_NVIDIA_NV36				0x0349 /*NV36*/
#define DEVICE_NVIDIA_NV362				0x034b /*NV36*/
#define DEVICE_NVIDIA_NV36_QUADRO_FX			0x034c /*NV36 [Quadro FX Go1000]*/
#define DEVICE_NVIDIA_NV36GL_QUADRO_FX			0x034e /*NV36GL [Quadro FX 1100]*/
#define DEVICE_NVIDIA_NV36GL				0x034f /*NV36GL*/
/* Vendor: 10df: Emulex Corporation */
#define DEVICE_EMULEX_LP6000_FIBRE_CHANNEL		0x1ae5 /*LP6000 Fibre Channel Host Adapter*/
#define DEVICE_EMULEX_LP_8000_FIBRE			0x1ae6 /*LP 8000 Fibre Channel Host Adapter Alternate ID (JX1:2-3, JX2:1-2)*/
#define DEVICE_EMULEX_LP_8000_FIBRE2			0x1ae7 /*LP 8000 Fibre Channel Host Adapter Alternate ID (JX1:2-3, JX2:2-3)*/
#define DEVICE_EMULEX_LP1150E				0xf015 /*LP1150e*/
#define DEVICE_EMULEX_LP850_FIBRE_CHANNEL		0xf085 /*LP850 Fibre Channel Adapter*/
#define DEVICE_EMULEX_LP952_FIBRE_CHANNEL		0xf095 /*LP952 Fibre Channel Adapter*/
#define DEVICE_EMULEX_LP982_FIBRE_CHANNEL		0xf098 /*LP982 Fibre Channel Adapter*/
#define DEVICE_EMULEX_LIGHTPULSE_FIBRE_CHANNEL		0xf0a1 /*LightPulse Fibre Channel Adapter*/
#define DEVICE_EMULEX_LP1050				0xf0a5 /*LP1050*/
#define DEVICE_EMULEX_LP1150				0xf0d5 /*LP1150*/
#define DEVICE_EMULEX_LP11000E				0xf100 /*LP11000e*/
#define DEVICE_EMULEX_LP7000_FIBRE_CHANNEL		0xf700 /*LP7000 Fibre Channel Host Adapter*/
#define DEVICE_EMULEX_LP_7000EFIBRE_CHANNEL		0xf701 /*LP 7000EFibre Channel Host Adapter Alternate ID (JX1:2-3, JX2:1-2)*/
#define DEVICE_EMULEX_LP8000_FIBRE_CHANNEL		0xf800 /*LP8000 Fibre Channel Host Adapter*/
#define DEVICE_EMULEX_LP_8000_FIBRE3			0xf801 /*LP 8000 Fibre Channel Host Adapter Alternate ID (JX1:2-3, JX2:1-2)*/
#define DEVICE_EMULEX_LP9000_FIBRE_CHANNEL		0xf900 /*LP9000 Fibre Channel Host Adapter*/
#define DEVICE_EMULEX_LP_9000_FIBRE			0xf901 /*LP 9000 Fibre Channel Host Adapter Alternate ID (JX1:2-3, JX2:1-2)*/
#define DEVICE_EMULEX_LP9802_FIBRE_CHANNEL		0xf980 /*LP9802 Fibre Channel Adapter*/
#define DEVICE_EMULEX_LP_9802_FIBRE			0xf981 /*LP 9802 Fibre Channel Host Adapter Alternate ID*/
#define DEVICE_EMULEX_LP_9802_FIBRE2			0xf982 /*LP 9802 Fibre Channel Host Adapter Alternate ID*/
#define DEVICE_EMULEX_LP10000_FIBRE_CHANNEL		0xfa00 /*LP10000 Fibre Channel Host Adapter*/
#define DEVICE_EMULEX_LP101				0xfa01 /*LP101*/
#define DEVICE_EMULEX_LIGHTPULSE_FIBRE_CHANNEL2		0xfb00 /*LightPulse Fibre Channel Adapter*/
#define DEVICE_EMULEX_LP11000				0xfd00 /*LP11000*/
/* Vendor: 10e0: Integrated Micro Solutions Inc. */
#define DEVICE_INTEGRATED_IMS5026_27_28			0x5026 /*IMS5026/27/28*/
#define DEVICE_INTEGRATED_IMS5027			0x5027 /*IMS5027*/
#define DEVICE_INTEGRATED_IMS5028			0x5028 /*IMS5028*/
#define DEVICE_INTEGRATED_IMS8849			0x8849 /*IMS8849*/
#define DEVICE_INTEGRATED_IMS8853			0x8853 /*IMS8853*/
#define DEVICE_INTEGRATED_IMS9128_TWIN_TURBO		0x9128 /*IMS9128 [Twin turbo 128]*/
/* Vendor: 10e1: Tekram Technology Co.,Ltd. */
#define DEVICE_TEKRAM_TRM_S1040				0x0391 /*TRM-S1040*/
	#define SUBDEVICE_TEKRAM_DC_315U_SCSI_3_HOST	0x10e10391 /*DC-315U SCSI-3 Host Adapter*/
#define DEVICE_TEKRAM_DC_690C				0x690c /*DC-690c*/
#define DEVICE_TEKRAM_DC_290				0xdc29 /*DC-290*/
/* Vendor: 10e2: Aptix Corporation */
/* Vendor: 10e3: Tundra Semiconductor Corp. */
#define DEVICE_TUNDRA_CA91C042_UNIVERSE			0x0000 /*CA91C042 [Universe]*/
#define DEVICE_TUNDRA_CA91C860_QSPAN			0x0860 /*CA91C860 [QSpan]*/
#define DEVICE_TUNDRA_CA91C862A_QSPAN_II		0x0862 /*CA91C862A [QSpan-II]*/
#define DEVICE_TUNDRA_CA91L8200B_DUAL_PCI		0x8260 /*CA91L8200B [Dual PCI PowerSpan II]*/
#define DEVICE_TUNDRA_CA91L8260B_SINGLE_PCI		0x8261 /*CA91L8260B [Single PCI PowerSpan II]*/
/* Vendor: 10e4: Tandem Computers */
/* Vendor: 10e5: Micro Industries Corporation */
/* Vendor: 10e6: Gainbery Computer Products Inc. */
/* Vendor: 10e7: Vadem */
/* Vendor: 10e8: Applied Micro Circuits Corp. */
#define DEVICE_APPLIED_INES_GPIB_PCI_AMCC5920		0x1072 /*INES GPIB-PCI (AMCC5920 based)*/
#define DEVICE_APPLIED_Q_MOTION_VIDEO_CAPTURE_EDIT	0x2011 /*Q-Motion Video Capture/Edit board*/
#define DEVICE_APPLIED_S5930_MATCHMAKER			0x4750 /*S5930 [Matchmaker]*/
#define DEVICE_APPLIED_S5920				0x5920 /*S5920*/
#define DEVICE_APPLIED_LANAI4_X_MYRINET_LANAI		0x8043 /*LANai4.x [Myrinet LANai interface chip]*/
#define DEVICE_APPLIED_S5933_PARASTATION		0x8062 /*S5933_PARASTATION*/
#define DEVICE_APPLIED_S5933_MATCHMAKER			0x807d /*S5933 [Matchmaker]*/
#define DEVICE_APPLIED_KONGSBERG_SPACETEC_FORMAT	0x8088 /*Kongsberg Spacetec Format Synchronizer*/
#define DEVICE_APPLIED_KONGSBERG_SPACETEC_SERIAL	0x8089 /*Kongsberg Spacetec Serial Output Board*/
#define DEVICE_APPLIED_S5933_HEPC3			0x809c /*S5933_HEPC3*/
#define DEVICE_APPLIED_PCI_9112				0x80d7 /*PCI-9112*/
#define DEVICE_APPLIED_PCI_9118				0x80d9 /*PCI-9118*/
#define DEVICE_APPLIED_PCI_9812				0x80da /*PCI-9812*/
#define DEVICE_APPLIED_PCI_IEEE1355_DS_DE_INTERFACE	0x811a /*PCI-IEEE1355-DS-DE Interface*/
#define DEVICE_APPLIED_FASTCOM_ESCC_PCI_COMMTECH	0x814c /*Fastcom ESCC-PCI (Commtech, Inc.)*/
#define DEVICE_APPLIED_S5933_MATCHMAKER_CHIPSET		0x8170 /*S5933 [Matchmaker] (Chipset Development Tool)*/
#define DEVICE_APPLIED_MULTIMEDIA_VIDEO_CONTROLLER	0x81e6 /*Multimedia video controller*/
#define DEVICE_APPLIED_FASTCOM_232_8_PCI_COMMTECH	0x8291 /*Fastcom 232/8-PCI (Commtech, Inc.)*/
#define DEVICE_APPLIED_FASTCOM_422_4_PCI_COMMTECH	0x82c4 /*Fastcom 422/4-PCI (Commtech, Inc.)*/
#define DEVICE_APPLIED_FASTCOM_422_2_PCI_COMMTECH	0x82c5 /*Fastcom 422/2-PCI (Commtech, Inc.)*/
#define DEVICE_APPLIED_FASTCOM_IG422_1_PCI_COMMTECH	0x82c6 /*Fastcom IG422/1-PCI (Commtech, Inc.)*/
#define DEVICE_APPLIED_FASTCOM_IG232_2_PCI_COMMTECH	0x82c7 /*Fastcom IG232/2-PCI (Commtech, Inc.)*/
#define DEVICE_APPLIED_FASTCOM_232_4_PCI_COMMTECH	0x82ca /*Fastcom 232/4-PCI (Commtech, Inc.)*/
#define DEVICE_APPLIED_AJA_HDNTV_HD			0x82db /*AJA HDNTV HD SDI Framestore*/
#define DEVICE_APPLIED_FASTCOM_DIO24H_PCI_COMMTECH	0x82e2 /*Fastcom DIO24H-PCI (Commtech, Inc.)*/
#define DEVICE_APPLIED_S5933_ON_INNES			0x8851 /*S5933 on Innes Corp FM Radio Capture card*/
/* Vendor: 10e9: Alps Electric Co., Ltd. */
/* Vendor: 10ea: Intergraphics Systems */
#define DEVICE_INTERGRAPHICS_IGA_1680			0x1680 /*IGA-1680*/
#define DEVICE_INTERGRAPHICS_IGA_1682			0x1682 /*IGA-1682*/
#define DEVICE_INTERGRAPHICS_IGA_1683			0x1683 /*IGA-1683*/
#define DEVICE_INTERGRAPHICS_CYBERPRO_2000		0x2000 /*CyberPro 2000*/
#define DEVICE_INTERGRAPHICS_CYBERPRO_2000A		0x2010 /*CyberPro 2000A*/
#define DEVICE_INTERGRAPHICS_CYBERPRO_5000		0x5000 /*CyberPro 5000*/
#define DEVICE_INTERGRAPHICS_CYBERPRO_5050		0x5050 /*CyberPro 5050*/
#define DEVICE_INTERGRAPHICS_CYBERPRO_5202		0x5202 /*CyberPro 5202*/
#define DEVICE_INTERGRAPHICS_CYBERPRO5252		0x5252 /*CyberPro5252*/
/* Vendor: 10eb: Artists Graphics */
#define DEVICE_ARTISTS_3GA				0x0101 /*3GA*/
#define DEVICE_ARTISTS_TWIST3_FRAME_GRABBER		0x8111 /*Twist3 Frame Grabber*/
/* Vendor: 10ec: Realtek Semiconductor Co., Ltd. */
#define DEVICE_REALTEK_RTL_8029_AS			0x8029 /*RTL-8029(AS)*/
	#define SUBDEVICE_REALTEK_EZ_CARD_SMC1208	0x10b82011 /*EZ-Card (SMC1208)*/
	#define SUBDEVICE_REALTEK_RTL_8029_AS		0x10ec8029 /*RTL-8029(AS)*/
	#define SUBDEVICE_REALTEK_EN1208		0x11131208 /*EN1208*/
	#define SUBDEVICE_REALTEK_DE_528		0x11860300 /*DE-528*/
	#define SUBDEVICE_REALTEK_AT_2400		0x12592400 /*AT-2400*/
#define DEVICE_REALTEK_RTL_8129				0x8129 /*RTL-8129*/
	#define SUBDEVICE_REALTEK_RT8129_FAST_ETHERNET	0x10ec8129 /*RT8129 Fast Ethernet Adapter*/
#define DEVICE_REALTEK_RT8139_B_C_CARDBUS		0x8138 /*RT8139 (B/C) Cardbus Fast Ethernet Adapter*/
	#define SUBDEVICE_REALTEK_RT8139_B_C_FAST	0x10ec8138 /*RT8139 (B/C) Fast Ethernet Adapter*/
#define DEVICE_REALTEK_RTL_8139_8139C_8139C		0x8139 /*RTL-8139/8139C/8139C+*/
	#define SUBDEVICE_REALTEK_TTP_MONITORING_CARD_V2_0	0x0357000a /*TTP-Monitoring Card V2.0*/
	#define SUBDEVICE_REALTEK_TRAVELMATE_290	0x1025005a /*TravelMate 290*/
	#define SUBDEVICE_REALTEK_ALN_325		0x10258920 /*ALN-325*/
	#define SUBDEVICE_REALTEK_ALN_3252		0x10258921 /*ALN-325*/
	#define SUBDEVICE_REALTEK_MIM2000		0x10718160 /*MIM2000*/
	#define SUBDEVICE_REALTEK_EP_320X_R		0x10bd0320 /*EP-320X-R*/
	#define SUBDEVICE_REALTEK_RT8139		0x10ec8139 /*RT8139*/
	#define SUBDEVICE_REALTEK_FNC_0107TX		0x1113ec01 /*FNC-0107TX*/
	#define SUBDEVICE_REALTEK_DFE_538TX		0x11861300 /*DFE-538TX*/
	#define SUBDEVICE_REALTEK_SN5200		0x11861320 /*SN5200*/
	#define SUBDEVICE_REALTEK_DRN_32TX		0x11868139 /*DRN-32TX*/
	#define SUBDEVICE_REALTEK_FN22_3_A_LINXPRO_ETHERNET	0x11f68139 /*FN22-3(A) LinxPRO Ethernet Adapter*/
	#define SUBDEVICE_REALTEK_AT_2500TX		0x12592500 /*AT-2500TX*/
	#define SUBDEVICE_REALTEK_AT_2500TX_ACPI	0x12592503 /*AT-2500TX/ACPI*/
	#define SUBDEVICE_REALTEK_ND010			0x1429d010 /*ND010*/
	#define SUBDEVICE_REALTEK_EN_9130TX		0x14329130 /*EN-9130TX*/
	#define SUBDEVICE_REALTEK_RT81392		0x14368139 /*RT8139*/
	#define SUBDEVICE_REALTEK_GA_7VM400M_7VT600_MOTHERBOARD	0x1458e000 /*GA-7VM400M/7VT600 Motherboard*/
	#define SUBDEVICE_REALTEK_FE_1439TX		0x146c1439 /*FE-1439TX*/
	#define SUBDEVICE_REALTEK_GF100TXRII		0x14896001 /*GF100TXRII*/
	#define SUBDEVICE_REALTEK_GF100TXRA		0x14896002 /*GF100TXRA*/
	#define SUBDEVICE_REALTEK_LFE_8139ATX		0x149c139a /*LFE-8139ATX*/
	#define SUBDEVICE_REALTEK_LFE_8139TX		0x149c8139 /*LFE-8139TX*/
	#define SUBDEVICE_REALTEK_LNR_100_FAMILY_10_100	0x14cb0200 /*LNR-100 Family 10/100 Base-TX Ethernet*/
	#define SUBDEVICE_REALTEK_F5D5000_PCI_CARD_DESKTOP	0x17995000 /*F5D5000 PCI Card/Desktop Network PCI Card*/
	#define SUBDEVICE_REALTEK_ETHERX		0x26460001 /*EtheRx*/
	#define SUBDEVICE_REALTEK_KF_230TX		0x8e2e7000 /*KF-230TX*/
	#define SUBDEVICE_REALTEK_KF_230TX_2		0x8e2e7100 /*KF-230TX/2*/
	#define SUBDEVICE_REALTEK_ALN_325C		0xa0a00007 /*ALN-325C*/
#define DEVICE_REALTEK_RTL_8169_GIGABIT_ETHERNET	0x8169 /*RTL-8169 Gigabit Ethernet*/
	#define SUBDEVICE_REALTEK_CG_LAPCIGT		0x1259c107 /*CG-LAPCIGT*/
	#define SUBDEVICE_REALTEK_PROG_2000L		0x1371434e /*ProG-2000L*/
	#define SUBDEVICE_REALTEK_GA_K8VT800_PRO_MOTHERBOARD	0x1458e000 /*GA-K8VT800 Pro Motherboard*/
	#define SUBDEVICE_REALTEK_K8T_NEO_2		0x1462702c /*K8T NEO 2 motherboard*/
#define DEVICE_REALTEK_RTL8180L_802_11B_MAC		0x8180 /*RTL8180L 802.11b MAC*/
#define DEVICE_REALTEK_SMARTLAN56_56K_MODEM		0x8197 /*SmartLAN56 56K Modem*/
/* Vendor: 10ed: Ascii Corporation */
#define DEVICE_ASCII2_V7310				0x7310 /*V7310*/
/* Vendor: 10ee: Xilinx Corporation */
#define DEVICE_XILINX_RME_DIGI96			0x3fc0 /*RME Digi96*/
#define DEVICE_XILINX_RME_DIGI96_8			0x3fc1 /*RME Digi96/8*/
#define DEVICE_XILINX_RME_DIGI96_8_PRO			0x3fc2 /*RME Digi96/8 Pro*/
#define DEVICE_XILINX_RME_DIGI96_8_PAD			0x3fc3 /*RME Digi96/8 Pad*/
#define DEVICE_XILINX_RME_DIGI9652_HAMMERFALL		0x3fc4 /*RME Digi9652 (Hammerfall)*/
#define DEVICE_XILINX_RME_HAMMERFALL_DSP		0x3fc5 /*RME Hammerfall DSP*/
#define DEVICE_XILINX_RME_HAMMERFALL_DSP2		0x3fc6 /*RME Hammerfall DSP MADI*/
#define DEVICE_XILINX_ELLIPS_SANTOS_FRAME		0x8381 /*Ellips Santos Frame Grabber*/
/* Vendor: 10ef: Racore Computer Products, Inc. */
#define DEVICE_RACORE_M815X_TOKEN_RING			0x8154 /*M815x Token Ring Adapter*/
/* Vendor: 10f0: Peritek Corporation */
/* Vendor: 10f1: Tyan Computer */
/* Vendor: 10f2: Achme Computer, Inc. */
/* Vendor: 10f3: Alaris, Inc. */
/* Vendor: 10f4: S-MOS Systems, Inc. */
/* Vendor: 10f5: NKK Corporation */
#define DEVICE_NKK_NDR4000_NR4600_BRIDGE		0xa001 /*NDR4000 [NR4600 Bridge]*/
/* Vendor: 10f6: Creative Electronic Systems SA */
/* Vendor: 10f7: Matsushita Electric Industrial Co., Ltd. */
/* Vendor: 10f8: Altos India Ltd */
/* Vendor: 10f9: PC Direct */
/* Vendor: 10fa: Truevision */
#define DEVICE_TRUEVISION_TARGA_1000			0x000c /*TARGA 1000*/
/* Vendor: 10fb: Thesys Gesellschaft für Mikroelektronik mbH */
#define DEVICE_THESYS_TH_6255				0x186f /*TH 6255*/
/* Vendor: 10fc: I-O Data Device, Inc. */
#define DEVICE_I_O_CARDBUS_IDE_CONTROLLER		0x0003 /*Cardbus IDE Controller*/
#define DEVICE_I_O_CARDBUS_SCSI_CBSC			0x0005 /*Cardbus SCSI CBSC II*/
/* Vendor: 10fd: Soyo Computer, Inc */
/* Vendor: 10fe: Fast Multimedia AG */
/* Vendor: 10ff: NCube */
/* Vendor: 1100: Jazz Multimedia */
/* Vendor: 1101: Initio Corporation */
#define DEVICE_INITIO_INI_A100U2W			0x1060 /*INI-A100U2W*/
#define DEVICE_INITIO_INI_9100_9100W			0x9100 /*INI-9100/9100W*/
#define DEVICE_INITIO_INI_940				0x9400 /*INI-940*/
#define DEVICE_INITIO_INI_950				0x9401 /*INI-950*/
#define DEVICE_INITIO_360P				0x9500 /*360P*/
#define DEVICE_INITIO_INITIO_INI_9100UW_ULTRA		0x9502 /*Initio INI-9100UW Ultra Wide SCSI Controller INIC-950P chip*/
/* Vendor: 1102: Creative Labs */
#define DEVICE_CREATIVE2_SB_LIVE_EMU10K1		0x0002 /*SB Live! EMU10k1*/
	#define SUBDEVICE_CREATIVE2_CT4850_SBLIVE_VALUE	0x11020020 /*CT4850 SBLive! Value*/
	#define SUBDEVICE_CREATIVE2_CT4620_SBLIVE	0x11020021 /*CT4620 SBLive!*/
	#define SUBDEVICE_CREATIVE2_SBLIVE_MAINBOARD_IMPLEMENTATION	0x1102002f /*SBLive! mainboard implementation*/
	#define SUBDEVICE_CREATIVE2_E_MU_APS		0x11024001 /*E-mu APS*/
	#define SUBDEVICE_CREATIVE2_CT4780_SBLIVE_VALUE	0x11028022 /*CT4780 SBLive! Value*/
	#define SUBDEVICE_CREATIVE2_CT4790_SOUNDBLASTER_PCI512	0x11028023 /*CT4790 SoundBlaster PCI512*/
	#define SUBDEVICE_CREATIVE2_CT4760_SBLIVE	0x11028024 /*CT4760 SBLive!*/
	#define SUBDEVICE_CREATIVE2_SBLIVE_MAINBOARD_IMPLEMENTATION2	0x11028025 /*SBLive! Mainboard Implementation*/
	#define SUBDEVICE_CREATIVE2_CT4830_SBLIVE_VALUE	0x11028026 /*CT4830 SBLive! Value*/
	#define SUBDEVICE_CREATIVE2_CT4832_SBLIVE_VALUE	0x11028027 /*CT4832 SBLive! Value*/
	#define SUBDEVICE_CREATIVE2_CT4760_SBLIVE_OEM	0x11028028 /*CT4760 SBLive! OEM version*/
	#define SUBDEVICE_CREATIVE2_CT4831_SBLIVE_VALUE	0x11028031 /*CT4831 SBLive! Value*/
	#define SUBDEVICE_CREATIVE2_CT4760_SBLIVE2	0x11028040 /*CT4760 SBLive!*/
	#define SUBDEVICE_CREATIVE2_CT4850_SBLIVE_VALUE2	0x11028051 /*CT4850 SBLive! Value*/
	#define SUBDEVICE_CREATIVE2_SBLIVE_PLAYER_5_1	0x11028061 /*SBLive! Player 5.1*/
	#define SUBDEVICE_CREATIVE2_SB_LIVE_5_1		0x11028064 /*SB Live! 5.1 Model SB0100*/
	#define SUBDEVICE_CREATIVE2_SBLIVE_5_1_DIGITAL	0x11028065 /*SBLive! 5.1 Digital Model SB0220*/
	#define SUBDEVICE_CREATIVE2_SBLIVE_5_1_EMICRO	0x11028067 /*SBLive! 5.1 eMicro 28028*/
#define DEVICE_CREATIVE2_SB_AUDIGY			0x0004 /*SB Audigy*/
	#define SUBDEVICE_CREATIVE2_SB0090_AUDIGY_PLAYER	0x11020051 /*SB0090 Audigy Player*/
	#define SUBDEVICE_CREATIVE2_SB0090_AUDIGY_PLAYER_OEM	0x11020053 /*SB0090 Audigy Player/OEM*/
	#define SUBDEVICE_CREATIVE2_SB0090_AUDIGY_PLAYER_OEM2	0x11020058 /*SB0090 Audigy Player/OEM*/
	#define SUBDEVICE_CREATIVE2_SB_AUDIGY_2		0x11022002 /*SB Audigy 2 ZS (SB0350)*/
#define DEVICE_CREATIVE2__SB_LIVE_VALUE			0x0006 /*[SB Live! Value] EMU10k1X*/
#define DEVICE_CREATIVE2_SB_AUDIGY_LS			0x0007 /*SB Audigy LS*/
	#define SUBDEVICE_CREATIVE2_SB0310_AUDIGY_LS	0x11021001 /*SB0310 Audigy LS*/
	#define SUBDEVICE_CREATIVE2_SB0312_AUDIGY_LS	0x11021002 /*SB0312 Audigy LS*/
#define DEVICE_CREATIVE2_SB0400_AUDIGY2_VALUE		0x0008 /*SB0400 Audigy2 Value*/
#define DEVICE_CREATIVE2_SB_AUDIGY_FIREWIRE		0x4001 /*SB Audigy FireWire Port*/
	#define SUBDEVICE_CREATIVE2_SB_AUDIGY_FIREWIRE	0x11020010 /*SB Audigy FireWire Port*/
#define DEVICE_CREATIVE2_SB_LIVE_MIDI_GAME		0x7002 /*SB Live! MIDI/Game Port*/
	#define SUBDEVICE_CREATIVE2_GAMEPORT_JOYSTICK	0x11020020 /*Gameport Joystick*/
#define DEVICE_CREATIVE2_SB_AUDIGY_MIDI_GAME		0x7003 /*SB Audigy MIDI/Game port*/
	#define SUBDEVICE_CREATIVE2_SB_AUDIGY_MIDI_GAME	0x11020040 /*SB Audigy MIDI/Game Port*/
#define DEVICE_CREATIVE2__SB_LIVE_VALUE2		0x7004 /*[SB Live! Value] Input device controller*/
#define DEVICE_CREATIVE2_SB_AUDIGY_LS2			0x7005 /*SB Audigy LS MIDI/Game port*/
	#define SUBDEVICE_CREATIVE2_SB0310_AUDIGY_LS2	0x11021001 /*SB0310 Audigy LS MIDI/Game port*/
	#define SUBDEVICE_CREATIVE2_SB0312_AUDIGY_LS2	0x11021002 /*SB0312 Audigy LS MIDI/Game port*/
#define DEVICE_CREATIVE2_SB0100_SBLIVE_5_1		0x8064 /*SB0100 [SBLive! 5.1 OEM]*/
#define DEVICE_CREATIVE2_ECTIVA_EV1938			0x8938 /*Ectiva EV1938*/
/* Vendor: 1103: Triones Technologies, Inc. */
#define DEVICE_TRIONES_HPT343				0x0003 /*HPT343*/
#define DEVICE_TRIONES_HPT366_368_370_370A_372		0x0004 /*HPT366/368/370/370A/372*/
	#define SUBDEVICE_TRIONES_HPT370A		0x11030001 /*HPT370A*/
	#define SUBDEVICE_TRIONES_HPT343_HPT345		0x11030003 /*HPT343 / HPT345 / HPT363 UDMA33*/
	#define SUBDEVICE_TRIONES_HPT366_UDMA66_R1	0x11030004 /*HPT366 UDMA66 (r1) / HPT368 UDMA66 (r2) / HPT370 UDMA100 (r3) / HPT370 UDMA100 RAID (r4)*/
	#define SUBDEVICE_TRIONES_HPT370_UDMA100	0x11030005 /*HPT370 UDMA100*/
	#define SUBDEVICE_TRIONES_HPT302		0x11030006 /*HPT302*/
	#define SUBDEVICE_TRIONES_HPT371_UDMA133	0x11030007 /*HPT371 UDMA133*/
	#define SUBDEVICE_TRIONES_HPT374_UDMA_ATA133_RAID	0x11030008 /*HPT374 UDMA/ATA133 RAID Controller*/
#define DEVICE_TRIONES_HPT372A				0x0005 /*HPT372A*/
#define DEVICE_TRIONES_HPT302				0x0006 /*HPT302*/
#define DEVICE_TRIONES_HPT371				0x0007 /*HPT371*/
#define DEVICE_TRIONES_HPT374				0x0008 /*HPT374*/
#define DEVICE_TRIONES_HPT372N				0x0009 /*HPT372N*/
/* Vendor: 1104: RasterOps Corp. */
/* Vendor: 1105: Sigma Designs, Inc. */
#define DEVICE_SIGMA_REALMAGIC_XCARD_MPEG		0x1105 /*REALmagic Xcard MPEG 1/2/3/4 DVD Decoder*/
#define DEVICE_SIGMA_REALMAGIC_HOLLYWOOD_PLUS		0x8300 /*REALmagic Hollywood Plus DVD Decoder*/
#define DEVICE_SIGMA_EM840X_REALMAGIC_DVD_MPEG_2	0x8400 /*EM840x REALmagic DVD/MPEG-2 Audio/Video Decoder*/
#define DEVICE_SIGMA_EM8401_REALMAGIC_DVD_MPEG_2	0x8401 /*EM8401 REALmagic DVD/MPEG-2 A/V Decoder*/
#define DEVICE_SIGMA_EM8470_REALMAGIC_DVD_MPEG_4	0x8470 /*EM8470 REALmagic DVD/MPEG-4 A/V Decoder*/
#define DEVICE_SIGMA_EM8471_REALMAGIC_DVD_MPEG_4	0x8471 /*EM8471 REALmagic DVD/MPEG-4 A/V Decoder*/
#define DEVICE_SIGMA_EM8475_REALMAGIC_DVD_MPEG_4	0x8475 /*EM8475 REALmagic DVD/MPEG-4 A/V Decoder*/
#define DEVICE_SIGMA_EM8476_REALMAGIC_DVD_MPEG_4	0x8476 /*EM8476 REALmagic DVD/MPEG-4 A/V Decoder*/
#define DEVICE_SIGMA_EM8485_REALMAGIC_DVD_MPEG_4	0x8485 /*EM8485 REALmagic DVD/MPEG-4 A/V Decoder*/
#define DEVICE_SIGMA_EM8486_REALMAGIC_DVD_MPEG_4	0x8486 /*EM8486 REALmagic DVD/MPEG-4 A/V Decoder*/
/* Vendor: 1106: VIA Technologies, Inc. */
#define DEVICE_VIA2_EMBEDDED_VIA_ETHERNET		0x0102 /*Embedded VIA Ethernet Controller*/
#define DEVICE_VIA2_VT6305_1394_A_CONTROLLER		0x0130 /*VT6305 1394.A Controller*/
#define DEVICE_VIA2_VT8363_8365_KT133_KM133		0x0305 /*VT8363/8365 [KT133/KM133]*/
	#define SUBDEVICE_VIA2_A7V_MAINBOARD		0x10438033 /*A7V Mainboard*/
	#define SUBDEVICE_VIA2_A7V_E_MAINBOARD		0x1043803e /*A7V-E Mainboard*/
	#define SUBDEVICE_VIA2_A7V133_A7V133_C_MAINBOARD	0x10438042 /*A7V133/A7V133-C Mainboard*/
	#define SUBDEVICE_VIA2_KT7_KT7_RAID_KT7A_KT7A_RAID_MAINBOARD	0x147ba401 /*KT7/KT7-RAID/KT7A/KT7A-RAID Mainboard*/
#define DEVICE_VIA2_VT8371_KX133			0x0391 /*VT8371 [KX133]*/
#define DEVICE_VIA2_VT8501_APOLLO_MVP4			0x0501 /*VT8501 [Apollo MVP4]*/
#define DEVICE_VIA2_VT82C505				0x0505 /*VT82C505*/
#define DEVICE_VIA2_VT82C576MV				0x0561 /*VT82C576MV*/
#define DEVICE_VIA2_VT82C586A_B_VT82C686_A_B_VT823X_A_C_PIPC_BUS	0x0571 /*VT82C586A/B/VT82C686/A/B/VT823x/A/C PIPC Bus Master IDE*/
	#define SUBDEVICE_VIA2_P6VXA_MOTHERBOARD	0x10190985 /*P6VXA Motherboard*/
	#define SUBDEVICE_VIA2_L7VTA_V1_0_MOTHERBOARD	0x10190a81 /*L7VTA v1.0 Motherboard (KT400-8235)*/
	#define SUBDEVICE_VIA2_VT8233A_BUS_MASTER	0x10438052 /*VT8233A Bus Master ATA100/66/33 IDE*/
	#define SUBDEVICE_VIA2_A7V8X_MOTHERBOARD	0x1043808c /*A7V8X motherboard*/
	#define SUBDEVICE_VIA2_A7V8X_X_MOTHERBOARD_REV	0x104380a1 /*A7V8X-X motherboard rev. 1.01*/
	#define SUBDEVICE_VIA2_A7V600_MOTHERBOARD	0x104380ed /*A7V600 motherboard*/
	#define SUBDEVICE_VIA2_VT82C586_B_VT82C686_A_B_VT8233_A_C_VT8235_PIPC_BUS	0x11060571 /*VT82C586/B/VT82C686/A/B/VT8233/A/C/VT8235 PIPC Bus Master IDE*/
	#define SUBDEVICE_VIA2_MAGNIA_Z310		0x11790001 /*Magnia Z310*/
	#define SUBDEVICE_VIA2_FX41_MOTHERBOARD		0x1297f641 /*FX41 motherboard*/
	#define SUBDEVICE_VIA2_GA_7VAX_MAINBOARD	0x14585002 /*GA-7VAX Mainboard*/
	#define SUBDEVICE_VIA2_K8T_NEO_2		0x14627020 /*K8T NEO 2 motherboard*/
	#define SUBDEVICE_VIA2_KV8_MAX3_MOTHERBOARD	0x147b1407 /*KV8-MAX3 motherboard*/
	#define SUBDEVICE_VIA2_K7VT2_MOTHERBOARD	0x18490571 /*K7VT2 motherboard*/
#define DEVICE_VIA2_VT82C576_3V_APOLLO			0x0576 /*VT82C576 3V [Apollo Master]*/
#define DEVICE_VIA2_VT82C585VP_APOLLO_VP1_VPX		0x0585 /*VT82C585VP [Apollo VP1/VPX]*/
#define DEVICE_VIA2_VT82C586_A_B_PCI_TO_ISA_APOLLO	0x0586 /*VT82C586/A/B PCI-to-ISA [Apollo VP]*/
	#define SUBDEVICE_VIA2_MVP3_ISA_BRIDGE		0x11060000 /*MVP3 ISA Bridge*/
#define DEVICE_VIA2_VT82C595_APOLLO_VP2			0x0595 /*VT82C595 [Apollo VP2]*/
#define DEVICE_VIA2_VT82C596_ISA_MOBILE			0x0596 /*VT82C596 ISA [Mobile South]*/
	#define SUBDEVICE_VIA2_VT82C596_A_B_PCI_TO	0x11060000 /*VT82C596/A/B PCI to ISA Bridge*/
	#define SUBDEVICE_VIA2_VT82C596_A_B_PCI_TO2	0x14580596 /*VT82C596/A/B PCI to ISA Bridge*/
#define DEVICE_VIA2_VT82C597_APOLLO_VP3			0x0597 /*VT82C597 [Apollo VP3]*/
#define DEVICE_VIA2_VT82C598_APOLLO_MVP3		0x0598 /*VT82C598 [Apollo MVP3]*/
#define DEVICE_VIA2_VT8601_APOLLO_PROMEDIA		0x0601 /*VT8601 [Apollo ProMedia]*/
#define DEVICE_VIA2_VT8605_PROSAVAGE_PM133		0x0605 /*VT8605 [ProSavage PM133]*/
	#define SUBDEVICE_VIA2_CUV4X_MAINBOARD		0x1043802c /*CUV4X mainboard*/
#define DEVICE_VIA2_VT82C680_APOLLO_P6			0x0680 /*VT82C680 [Apollo P6]*/
#define DEVICE_VIA2_VT82C686_APOLLO_SUPER		0x0686 /*VT82C686 [Apollo Super South]*/
	#define SUBDEVICE_VIA2_P6VXA_MOTHERBOARD2	0x10190985 /*P6VXA Motherboard*/
	#define SUBDEVICE_VIA2_CUV4X_MAINBOARD2		0x1043802c /*CUV4X mainboard*/
	#define SUBDEVICE_VIA2_A7V_MAINBOARD2		0x10438033 /*A7V Mainboard*/
	#define SUBDEVICE_VIA2_A7V_E_MAINBOARD2		0x1043803e /*A7V-E Mainboard*/
	#define SUBDEVICE_VIA2_A7M266_MAINBOARD		0x10438040 /*A7M266 Mainboard*/
	#define SUBDEVICE_VIA2_A7V133_A7V133_C_MAINBOARD2	0x10438042 /*A7V133/A7V133-C Mainboard*/
	#define SUBDEVICE_VIA2_VT82C686_A_PCI_TO	0x11060000 /*VT82C686/A PCI to ISA Bridge*/
	#define SUBDEVICE_VIA2_VT82C686_A_PCI_TO2	0x11060686 /*VT82C686/A PCI to ISA Bridge*/
	#define SUBDEVICE_VIA2_MAGNIA_Z3102		0x11790001 /*Magnia Z310*/
	#define SUBDEVICE_VIA2_KG7_LITE_MAINBOARD	0x147ba702 /*KG7-Lite Mainboard*/
#define DEVICE_VIA2_VT82C693A_694X_APOLLO_PRO133X	0x0691 /*VT82C693A/694x [Apollo PRO133x]*/
	#define SUBDEVICE_VIA2_P6VXA_MOTHERBOARD3	0x10190985 /*P6VXA Motherboard*/
	#define SUBDEVICE_VIA2_MAGNIA_Z3103		0x11790001 /*Magnia Z310*/
	#define SUBDEVICE_VIA2_VT82C691_APOLLO_PRO	0x14580691 /*VT82C691 Apollo Pro System Controller*/
#define DEVICE_VIA2_VT82C693_APOLLO_PRO			0x0693 /*VT82C693 [Apollo Pro Plus]*/
#define DEVICE_VIA2_VT82C693A_APOLLO_PRO133		0x0698 /*VT82C693A [Apollo Pro133 AGP]*/
#define DEVICE_VIA2_VT82C926_AMAZON			0x0926 /*VT82C926 [Amazon]*/
#define DEVICE_VIA2_VT82C570MV				0x1000 /*VT82C570MV*/
#define DEVICE_VIA2_VT82C570MV2				0x1106 /*VT82C570MV*/
#define DEVICE_VIA2_VT82C576M_VT82C586			0x1571 /*VT82C576M/VT82C586*/
#define DEVICE_VIA2_VT82C595_97_APOLLO_VP2_97		0x1595 /*VT82C595/97 [Apollo VP2/97]*/
#define DEVICE_VIA2_CLE266				0x3022 /*CLE266*/
#define DEVICE_VIA2_VT82XXXXX_UHCI_USB			0x3038 /*VT82xxxxx UHCI USB 1.1 Controller*/
	#define SUBDEVICE_VIA2_USB_CONTROLLER		0x09251234 /*USB Controller*/
	#define SUBDEVICE_VIA2_P6VXA_MOTHERBOARD4	0x10190985 /*P6VXA Motherboard*/
	#define SUBDEVICE_VIA2_L7VTA_V1_0_MOTHERBOARD2	0x10190a81 /*L7VTA v1.0 Motherboard (KT400-8235)*/
	#define SUBDEVICE_VIA2_VT6202_USB2_0_4		0x1043808c /*VT6202 USB2.0 4 port controller*/
	#define SUBDEVICE_VIA2_A7V8X_X_MOTHERBOARD	0x104380a1 /*A7V8X-X motherboard*/
	#define SUBDEVICE_VIA2_A7V600_MOTHERBOARD2	0x104380ed /*A7V600 motherboard*/
	#define SUBDEVICE_VIA2_MAGNIA_Z3104		0x11790001 /*Magnia Z310*/
	#define SUBDEVICE_VIA2_GA_7VAX_MAINBOARD2	0x14585004 /*GA-7VAX Mainboard*/
	#define SUBDEVICE_VIA2_K8T_NEO_22		0x14627020 /*K8T NEO 2 motherboard*/
	#define SUBDEVICE_VIA2_KV8_MAX3_MOTHERBOARD2	0x147b1407 /*KV8-MAX3 motherboard*/
#define DEVICE_VIA2_VT82C586B_ACPI			0x3040 /*VT82C586B ACPI*/
#define DEVICE_VIA2_VT86C100A_RHINE			0x3043 /*VT86C100A [Rhine]*/
	#define SUBDEVICE_VIA2_VT86C100A_FAST_ETHERNET	0x10bd0000 /*VT86C100A Fast Ethernet Adapter*/
	#define SUBDEVICE_VIA2_VT86C100A_FAST_ETHERNET2	0x11060100 /*VT86C100A Fast Ethernet Adapter*/
	#define SUBDEVICE_VIA2_DFE_530TX_REV_A		0x11861400 /*DFE-530TX rev A*/
#define DEVICE_VIA2_IEEE_1394_HOST			0x3044 /*IEEE 1394 Host Controller*/
	#define SUBDEVICE_VIA2_TRAVELMATE_290		0x1025005a /*TravelMate 290*/
	#define SUBDEVICE_VIA2_GA_7VT600_1394_MOTHERBOARD	0x14581000 /*GA-7VT600-1394 Motherboard*/
	#define SUBDEVICE_VIA2_K8T_NEO_23		0x1462702d /*K8T NEO 2 motherboard*/
#define DEVICE_VIA2_VT82C596_POWER_MANAGEMENT		0x3050 /*VT82C596 Power Management*/
#define DEVICE_VIA2_VT82C596_POWER_MANAGEMENT2		0x3051 /*VT82C596 Power Management*/
#define DEVICE_VIA2_VT6105M_RHINE_III			0x3053 /*VT6105M [Rhine-III]*/
#define DEVICE_VIA2_VT82C686_APOLLO_SUPER2		0x3057 /*VT82C686 [Apollo Super ACPI]*/
	#define SUBDEVICE_VIA2_P6VXA_MOTHERBOARD5	0x10190985 /*P6VXA Motherboard*/
	#define SUBDEVICE_VIA2_A7V_MAINBOARD3		0x10438033 /*A7V Mainboard*/
	#define SUBDEVICE_VIA2_A7V_E_MAINBOARD3		0x1043803e /*A7V-E Mainboard*/
	#define SUBDEVICE_VIA2_A7M266_MAINBOARD2	0x10438040 /*A7M266 Mainboard*/
	#define SUBDEVICE_VIA2_A7V133_A7V133_C_MAINBOARD3	0x10438042 /*A7V133/A7V133-C Mainboard*/
	#define SUBDEVICE_VIA2_MAGNIA_Z3105		0x11790001 /*Magnia Z310*/
#define DEVICE_VIA2_VT82C686_AC97_AUDIO			0x3058 /*VT82C686 AC97 Audio Controller*/
	#define SUBDEVICE_VIA2_SOUNDMAX_DIGITAL_INTEGRATED	0x0e110097 /*SoundMax Digital Integrated Audio*/
	#define SUBDEVICE_VIA2_SOUNDMAX_INTEGRATED_DIGITAL	0x0e11b194 /*Soundmax integrated digital audio*/
	#define SUBDEVICE_VIA2_P6VXA_MOTHERBOARD6	0x10190985 /*P6VXA Motherboard*/
	#define SUBDEVICE_VIA2_A7V133_A7V133_C_MAINBOARD4	0x10431106 /*A7V133/A7V133-C Mainboard*/
	#define SUBDEVICE_VIA2_ONBOARD_AUDIO_ON		0x11064511 /*Onboard Audio on EP7KXA*/
	#define SUBDEVICE_VIA2_ONBOARD_AUDIO		0x14587600 /*Onboard Audio*/
	#define SUBDEVICE_VIA2_MS_6309_ONBOARD_AUDIO	0x14623091 /*MS-6309 Onboard Audio*/
	#define SUBDEVICE_VIA2_MS_6330_ONBOARD_AUDIO	0x14623300 /*MS-6330 Onboard Audio*/
	#define SUBDEVICE_VIA2_ONBOARD_AUDIO2		0x15dd7609 /*Onboard Audio*/
#define DEVICE_VIA2_VT8233_A_8235_8237_AC97_AUDIO	0x3059 /*VT8233/A/8235/8237 AC97 Audio Controller*/
	#define SUBDEVICE_VIA2_L7VTA_V1_0_MOTHERBOARD3	0x10190a81 /*L7VTA v1.0 Motherboard (KT400-8235)*/
	#define SUBDEVICE_VIA2_A7V8X_MOTHERBOARD_REALTEK	0x10438095 /*A7V8X Motherboard (Realtek ALC650 codec)*/
	#define SUBDEVICE_VIA2_A7V8X_X_MOTHERBOARD2	0x104380a1 /*A7V8X-X Motherboard*/
	#define SUBDEVICE_VIA2_A7V600_MOTHERBOARD_ADI	0x104380b0 /*A7V600 motherboard (ADI AD1980 codec [SoundMAX])*/
	#define SUBDEVICE_VIA2_L7VMM2_MOTHERBOARD	0x11063059 /*L7VMM2 Motherboard*/
	#define SUBDEVICE_VIA2_K7VT2_MOTHERBOARD2	0x11064161 /*K7VT2 motherboard*/
	#define SUBDEVICE_VIA2_FX41_MOTHERBOARD_REALTEK	0x1297c160 /*FX41 motherboard (Realtek ALC650 codec)*/
	#define SUBDEVICE_VIA2_GA_7VAX_ONBOARD_AUDIO	0x1458a002 /*GA-7VAX Onboard Audio (Realtek ALC650)*/
	#define SUBDEVICE_VIA2_K8T_NEO_24		0x14620080 /*K8T NEO 2 motherboard*/
	#define SUBDEVICE_VIA2_KT266_ONBOARD_AUDIO	0x14623800 /*KT266 onboard audio*/
	#define SUBDEVICE_VIA2_KV8_MAX3_MOTHERBOARD3	0x147b1407 /*KV8-MAX3 motherboard*/
#define DEVICE_VIA2_VT6102_RHINE_II			0x3065 /*VT6102 [Rhine-II]*/
	#define SUBDEVICE_VIA2_A7V8X_X_MOTHERBOARD3	0x104380a1 /*A7V8X-X Motherboard*/
	#define SUBDEVICE_VIA2_VT6102_RHINE_II		0x11060102 /*VT6102 [Rhine II] Embeded Ethernet Controller on VT8235*/
	#define SUBDEVICE_VIA2_DFE_530TX_REV_A2		0x11861400 /*DFE-530TX rev A*/
	#define SUBDEVICE_VIA2_DFE_530TX_REV_B		0x11861401 /*DFE-530TX rev B*/
	#define SUBDEVICE_VIA2_LD_10_100AL_PCI_FAST	0x13b91421 /*LD-10/100AL PCI Fast Ethernet Adapter (rev.B)*/
#define DEVICE_VIA2_AC_97_MODEM_CONTROLLER		0x3068 /*AC'97 Modem Controller*/
	#define SUBDEVICE_VIA2_MS_6309_SATURN_MOTHERBOARD	0x1462309e /*MS-6309 Saturn Motherboard*/
#define DEVICE_VIA2_VT8233_PCI_TO			0x3074 /*VT8233 PCI to ISA Bridge*/
	#define SUBDEVICE_VIA2_VT8233A			0x10438052 /*VT8233A*/
#define DEVICE_VIA2_VT8633_APOLLO_PRO266		0x3091 /*VT8633 [Apollo Pro266]*/
#define DEVICE_VIA2_VT8366_A_7_APOLLO_KT266_A_333	0x3099 /*VT8366/A/7 [Apollo KT266/A/333]*/
	#define SUBDEVICE_VIA2_A7V266_E_MAINBOARD	0x10438064 /*A7V266-E Mainboard*/
	#define SUBDEVICE_VIA2_A7V333_MAINBOARD		0x1043807f /*A7V333 Mainboard*/
	#define SUBDEVICE_VIA2_K7VT2_MOTHERBOARD3	0x18493099 /*K7VT2 motherboard*/
#define DEVICE_VIA2_VT8653_HOST_BRIDGE			0x3101 /*VT8653 Host Bridge*/
#define DEVICE_VIA2_VT8662_HOST_BRIDGE			0x3102 /*VT8662 Host Bridge*/
#define DEVICE_VIA2_VT8615_HOST_BRIDGE			0x3103 /*VT8615 Host Bridge*/
#define DEVICE_VIA2_USB_2_0				0x3104 /*USB 2.0*/
	#define SUBDEVICE_VIA2_L7VTA_V1_0_MOTHERBOARD4	0x10190a81 /*L7VTA v1.0 Motherboard (KT400-8235)*/
	#define SUBDEVICE_VIA2_A7V8X_MOTHERBOARD2	0x1043808c /*A7V8X motherboard*/
	#define SUBDEVICE_VIA2_A7V8X_X_MOTHERBOARD_REV2	0x104380a1 /*A7V8X-X motherboard rev 1.01*/
	#define SUBDEVICE_VIA2_A7V600_MOTHERBOARD3	0x104380ed /*A7V600 motherboard*/
	#define SUBDEVICE_VIA2_FX41_MOTHERBOARD2	0x1297f641 /*FX41 motherboard*/
	#define SUBDEVICE_VIA2_GA_7VAX_MAINBOARD3	0x14585004 /*GA-7VAX Mainboard*/
	#define SUBDEVICE_VIA2_K8T_NEO_25		0x14627020 /*K8T NEO 2 motherboard*/
	#define SUBDEVICE_VIA2_KV8_MAX3_MOTHERBOARD4	0x147b1407 /*KV8-MAX3 motherboard*/
#define DEVICE_VIA2_VT6105_RHINE_III			0x3106 /*VT6105 [Rhine-III]*/
	#define SUBDEVICE_VIA2_DFE_530TX_REV_C		0x11861403 /*DFE-530TX rev C*/
#define DEVICE_VIA2_S3_UNICHROME_PRO			0x3108 /*S3 Unichrome Pro VGA Adapter*/
#define DEVICE_VIA2_VT8233C_PCI_TO			0x3109 /*VT8233C PCI to ISA Bridge*/
#define DEVICE_VIA2_VT8361_KLE133_HOST			0x3112 /*VT8361 [KLE133] Host Bridge*/
#define DEVICE_VIA2_VT8375_KM266_KL266_HOST		0x3116 /*VT8375 [KM266/KL266] Host Bridge*/
	#define SUBDEVICE_VIA2_FX41_MOTHERBOARD3	0x1297f641 /*FX41 motherboard*/
#define DEVICE_VIA2_S3_UNICHROME_PRO2			0x3118 /*S3 Unichrome Pro VGA Adapter*/
#define DEVICE_VIA2_VT6120_VT6121_VT6122_GIGABIT_ETHERNET	0x3119 /*VT6120/VT6121/VT6122 Gigabit Ethernet Adapter*/
#define DEVICE_VIA2_VT8623_APOLLO_CLE266		0x3122 /*VT8623 [Apollo CLE266] integrated CastleRock graphics*/
#define DEVICE_VIA2_VT8623_APOLLO_CLE2662		0x3123 /*VT8623 [Apollo CLE266]*/
#define DEVICE_VIA2_VT8753_P4X266_AGP			0x3128 /*VT8753 [P4X266 AGP]*/
#define DEVICE_VIA2_VT3133_HOST_BRIDGE			0x3133 /*VT3133 Host Bridge*/
#define DEVICE_VIA2_VT8233A_ISA_BRIDGE			0x3147 /*VT8233A ISA Bridge*/
#define DEVICE_VIA2_P4M266_HOST_BRIDGE			0x3148 /*P4M266 Host Bridge*/
#define DEVICE_VIA2_VIA_VT6420_SATA			0x3149 /*VIA VT6420 SATA RAID Controller*/
	#define SUBDEVICE_VIA2_A7V600_MOTHERBOARD4	0x104380ed /*A7V600 motherboard*/
	#define SUBDEVICE_VIA2_GA_7VM400AM_F_MOTHERBOARD	0x1458b003 /*GA-7VM400AM(F) Motherboard*/
	#define SUBDEVICE_VIA2_K8T_NEO_26		0x14627020 /*K8T Neo 2 Motherboard*/
#define DEVICE_VIA2_P_KN266_HOST_BRIDGE			0x3156 /*P/KN266 Host Bridge*/
#define DEVICE_VIA2_VT6410_ATA133_RAID			0x3164 /*VT6410 ATA133 RAID controller*/
#define DEVICE_VIA2_VT8374_P4X400_HOST			0x3168 /*VT8374 P4X400 Host Controller/AGP Bridge*/
#define DEVICE_VIA2_VT8235_ISA_BRIDGE			0x3177 /*VT8235 ISA Bridge*/
	#define SUBDEVICE_VIA2_L7VTA_V1_0_MOTHERBOARD5	0x10190a81 /*L7VTA v1.0 Motherboard (KT400-8235)*/
	#define SUBDEVICE_VIA2_A7V8X_MOTHERBOARD3	0x1043808c /*A7V8X motherboard*/
	#define SUBDEVICE_VIA2_A7V8X_X_MOTHERBOARD4	0x104380a1 /*A7V8X-X motherboard*/
	#define SUBDEVICE_VIA2_FX41_MOTHERBOARD4	0x1297f641 /*FX41 motherboard*/
	#define SUBDEVICE_VIA2_GA_7VAX_MAINBOARD4	0x14585001 /*GA-7VAX Mainboard*/
	#define SUBDEVICE_VIA2_K7VT2_MOTHERBOARD4	0x18493177 /*K7VT2 motherboard*/
#define DEVICE_VIA2_VT8385_K8T800_AGP			0x3188 /*VT8385 [K8T800 AGP] Host Bridge*/
	#define SUBDEVICE_VIA2_KV8_MAX3_MOTHERBOARD5	0x147b1407 /*KV8-MAX3 motherboard*/
#define DEVICE_VIA2_VT8377_KT400_KT600_AGP		0x3189 /*VT8377 [KT400/KT600 AGP] Host Bridge*/
	#define SUBDEVICE_VIA2_A7V8X_MOTHERBOARD4	0x1043807f /*A7V8X motherboard*/
	#define SUBDEVICE_VIA2_GA_7VAX_MAINBOARD5	0x14585000 /*GA-7VAX Mainboard*/
#define DEVICE_VIA2_K8M800				0x3204 /*K8M800*/
#define DEVICE_VIA2_VT8378_KM400_A_CHIPSET		0x3205 /*VT8378 [KM400/A] Chipset Host Bridge*/
	#define SUBDEVICE_VIA2_GA_7VM400M_MOTHERBOARD	0x14585000 /*GA-7VM400M Motherboard*/
#define DEVICE_VIA2_VT8237_ISA_BRIDGE			0x3227 /*VT8237 ISA bridge [KT600/K8T800 South]*/
	#define SUBDEVICE_VIA2_A7V600_MOTHERBOARD5	0x104380ed /*A7V600 motherboard*/
	#define SUBDEVICE_VIA2_DFI_KT600_AL_MOTHERBOARD	0x11063227 /*DFI KT600-AL Motherboard*/
	#define SUBDEVICE_VIA2_GA_7VT600_MOTHERBOARD	0x14585001 /*GA-7VT600 Motherboard*/
	#define SUBDEVICE_VIA2_KV8_MAX3_MOTHERBOARD6	0x147b1407 /*KV8-MAX3 motherboard*/
#define DEVICE_VIA2_VIA_VT6420_ATA133			0x4149 /*VIA VT6420 (ATA133) Controller*/
#define DEVICE_VIA2_VT82C596_ACPI_APOLLO		0x5030 /*VT82C596 ACPI [Apollo PRO]*/
#define DEVICE_VIA2_VT85C100A_RHINE_II			0x6100 /*VT85C100A [Rhine II]*/
#define DEVICE_VIA2_K8M8002				0x7204 /*K8M800*/
#define DEVICE_VIA2_VT8378_S3_UNICHROME			0x7205 /*VT8378 [S3 UniChrome] Integrated Video*/
	#define SUBDEVICE_VIA2_GIGABYTE_GA_7VM400_A_M_F_MOTHERBOARD	0x1458d000 /*Gigabyte GA-7VM400(A)M(F) Motherboard*/
#define DEVICE_VIA2_VT8231_PCI_TO_ISA_BRIDGE		0x8231 /*VT8231 [PCI-to-ISA Bridge]*/
#define DEVICE_VIA2_VT8235_ACPI				0x8235 /*VT8235 ACPI*/
#define DEVICE_VIA2_VT8363_8365_KT133_KM133_AGP		0x8305 /*VT8363/8365 [KT133/KM133 AGP]*/
#define DEVICE_VIA2_VT8371_KX133_AGP			0x8391 /*VT8371 [KX133 AGP]*/
#define DEVICE_VIA2_VT8501_APOLLO_MVP42			0x8501 /*VT8501 [Apollo MVP4 AGP]*/
#define DEVICE_VIA2_VT82C596_APOLLO_PRO			0x8596 /*VT82C596 [Apollo PRO AGP]*/
#define DEVICE_VIA2_VT82C597_APOLLO_VP32		0x8597 /*VT82C597 [Apollo VP3 AGP]*/
#define DEVICE_VIA2_VT82C598_694X_APOLLO_MVP3_PRO133X	0x8598 /*VT82C598/694x [Apollo MVP3/Pro133x AGP]*/
	#define SUBDEVICE_VIA2_P6VXA_MOTHERBOARD7	0x10190985 /*P6VXA Motherboard*/
#define DEVICE_VIA2_VT8601_APOLLO_PROMEDIA2		0x8601 /*VT8601 [Apollo ProMedia AGP]*/
#define DEVICE_VIA2_VT8605_PM133_AGP			0x8605 /*VT8605 [PM133 AGP]*/
#define DEVICE_VIA2_VT82C691_APOLLO_PRO			0x8691 /*VT82C691 [Apollo Pro]*/
#define DEVICE_VIA2_VT82C693_APOLLO_PRO2		0x8693 /*VT82C693 [Apollo Pro Plus] PCI Bridge*/
#define DEVICE_VIA2_VT8633_APOLLO_PRO2662		0xb091 /*VT8633 [Apollo Pro266 AGP]*/
#define DEVICE_VIA2_VT8366_A_7_APOLLO_KT266_A_3332	0xb099 /*VT8366/A/7 [Apollo KT266/A/333 AGP]*/
#define DEVICE_VIA2_VT8653_AGP_BRIDGE			0xb101 /*VT8653 AGP Bridge*/
#define DEVICE_VIA2_VT8362_AGP_BRIDGE			0xb102 /*VT8362 AGP Bridge*/
#define DEVICE_VIA2_VT8615_AGP_BRIDGE			0xb103 /*VT8615 AGP Bridge*/
#define DEVICE_VIA2_VT8361_KLE133_AGP			0xb112 /*VT8361 [KLE133] AGP Bridge*/
#define DEVICE_VIA2_VT8235_PCI_BRIDGE			0xb168 /*VT8235 PCI Bridge*/
#define DEVICE_VIA2_VT8237_PCI_BRIDGE			0xb188 /*VT8237 PCI bridge [K8T800 South]*/
	#define SUBDEVICE_VIA2_KV8_MAX3_MOTHERBOARD7	0x147b1407 /*KV8-MAX3 motherboard*/
#define DEVICE_VIA2_VT8237_PCI_BRIDGE2			0xb198 /*VT8237 PCI Bridge*/
#define DEVICE_VIA2_VT8237_INTEGRATED_FAST		0xd104 /*VT8237 Integrated Fast Ethernet Controller*/
/* Vendor: 1107: Stratus Computers */
#define DEVICE_STRATUS_VIA_VT82C570MV_APOLLO		0x0576 /*VIA VT82C570MV [Apollo] (Wrong vendor ID!)*/
/* Vendor: 1108: Proteon, Inc. */
#define DEVICE_PROTEON_P1690PLUS_AA			0x0100 /*p1690plus_AA*/
#define DEVICE_PROTEON_P1690PLUS_AB			0x0101 /*p1690plus_AB*/
#define DEVICE_PROTEON_P1690PLUS			0x0105 /*P1690Plus*/
#define DEVICE_PROTEON_P1690PLUS2			0x0108 /*P1690Plus*/
#define DEVICE_PROTEON_P1690PLUS3			0x0138 /*P1690Plus*/
#define DEVICE_PROTEON_P1690PLUS4			0x0139 /*P1690Plus*/
#define DEVICE_PROTEON_P1690PLUS5			0x013c /*P1690Plus*/
#define DEVICE_PROTEON_P1690PLUS6			0x013d /*P1690Plus*/
/* Vendor: 1109: Cogent Data Technologies, Inc. */
#define DEVICE_COGENT_EM110TX_EX110TX			0x1400 /*EM110TX [EX110TX]*/
/* Vendor: 110a: Siemens Nixdorf AG */
#define DEVICE_SIEMENS2_PIRAHNA_2_PORT			0x0002 /*Pirahna 2-port*/
#define DEVICE_SIEMENS2_TULIP_CONTROLLER_POWER		0x0005 /*Tulip controller, power management, switch extender*/
#define DEVICE_SIEMENS2_FSC_PINC_I_O_APIC		0x0006 /*FSC PINC (I/O-APIC)*/
#define DEVICE_SIEMENS2_FSC_MULTIPROCESSOR_INTERRUPT	0x0015 /*FSC Multiprocessor Interrupt Controller*/
#define DEVICE_SIEMENS2_FSC_COPERNICUS_MANAGEMENT	0x001d /*FSC Copernicus Management Controller*/
#define DEVICE_SIEMENS2_FSC_REMOTE_SERVICE		0x007b /*FSC Remote Service Controller, mailbox device*/
#define DEVICE_SIEMENS2_FSC_REMOTE_SERVICE2		0x007c /*FSC Remote Service Controller, shared memory device*/
#define DEVICE_SIEMENS2_FSC_REMOTE_SERVICE3		0x007d /*FSC Remote Service Controller, SMIC device*/
#define DEVICE_SIEMENS2_DSCC4_WAN_ADAPTER		0x2102 /*DSCC4 WAN adapter*/
#define DEVICE_SIEMENS2_EICON_DIVA_2_02			0x2104 /*Eicon Diva 2.02 compatible passive ISDN card*/
#define DEVICE_SIEMENS2_SIMATIC_NET_CP			0x3142 /*SIMATIC NET CP 5613A1 (Profibus Adapter)*/
#define DEVICE_SIEMENS2_SIMATIC_NET_CP2			0x4021 /*SIMATIC NET CP 5512 (Profibus and MPI Cardbus Adapter)*/
#define DEVICE_SIEMENS2_SIMATIC_NET_CP3			0x4029 /*SIMATIC NET CP 5613A2 (Profibus Adapter)*/
#define DEVICE_SIEMENS2_FPGA_I_BUS_TRACER		0x4942 /*FPGA I-Bus Tracer for MBD*/
#define DEVICE_SIEMENS2_SZB6120				0x6120 /*SZB6120*/
/* Vendor: 110b: Chromatic Research Inc. */
#define DEVICE_CHROMATIC_MPACT_MEDIA_PROCESSOR		0x0001 /*Mpact Media Processor*/
#define DEVICE_CHROMATIC_MPACT_2			0x0004 /*Mpact 2*/
/* Vendor: 110c: Mini-Max Technology, Inc. */
/* Vendor: 110d: Znyx Advanced Systems */
/* Vendor: 110e: CPU Technology */
/* Vendor: 110f: Ross Technology */
/* Vendor: 1110: Powerhouse Systems */
#define DEVICE_POWERHOUSE_FIREPOWER_POWERIZED_SMP	0x6037 /*Firepower Powerized SMP I/O ASIC*/
#define DEVICE_POWERHOUSE_FIREPOWER_POWERIZED_SMP2	0x6073 /*Firepower Powerized SMP I/O ASIC*/
/* Vendor: 1111: Santa Cruz Operation */
/* Vendor: 1112: Osicom Technologies Inc */
#define DEVICE_OSICOM_FDDI_ADAPTER			0x2200 /*FDDI Adapter*/
#define DEVICE_OSICOM_FAST_ETHERNET_ADAPTER		0x2300 /*Fast Ethernet Adapter*/
#define DEVICE_OSICOM_4_PORT_FAST			0x2340 /*4 Port Fast Ethernet Adapter*/
#define DEVICE_OSICOM_ATM_ADAPTER			0x2400 /*ATM Adapter*/
/* Vendor: 1113: Accton Technology Corporation */
#define DEVICE_ACCTON_SMC2_1211TX			0x1211 /*SMC2-1211TX*/
	#define SUBDEVICE_ACCTON_EN_1207D_FAST_ETHERNET	0x103c1207 /*EN-1207D Fast Ethernet Adapter*/
	#define SUBDEVICE_ACCTON_EN_1207D_FAST_ETHERNET2	0x11131211 /*EN-1207D Fast Ethernet Adapter*/
#define DEVICE_ACCTON_EN_1216_ETHERNET_ADAPTER		0x1216 /*EN-1216 Ethernet Adapter*/
	#define SUBDEVICE_ACCTON_EN2242_10_100_ETHERNET	0x11132242 /*EN2242 10/100 Ethernet Mini-PCI Card*/
	#define SUBDEVICE_ACCTON_SPEEDSTREAM_1020_PCI	0x111a1020 /*SpeedStream 1020 PCI 10/100 Ethernet Adaptor [EN-1207F-TX ?]*/
#define DEVICE_ACCTON_EN_1217_ETHERNET_ADAPTER		0x1217 /*EN-1217 Ethernet Adapter*/
#define DEVICE_ACCTON_10MBPS_NETWORK_CARD		0x5105 /*10Mbps Network card*/
#define DEVICE_ACCTON_EN_1207D_FAST_ETHERNET		0x9211 /*EN-1207D Fast Ethernet Adapter*/
	#define SUBDEVICE_ACCTON_EN_1207D_FAST_ETHERNET3	0x11139211 /*EN-1207D Fast Ethernet Adapter*/
#define DEVICE_ACCTON_21X4X_DEC_TULIP_COMPATIBLE	0x9511 /*21x4x DEC-Tulip compatible Fast Ethernet*/
#define DEVICE_ACCTON_CPWNA100_PHILIPS_WIRELESS		0xd301 /*CPWNA100 (Philips wireless PCMCIA)*/
#define DEVICE_ACCTON_SMC_1244TX_V3			0xec02 /*SMC 1244TX v3*/
/* Vendor: 1114: Atmel Corporation */
#define DEVICE_ATMEL_802_11B_WIRELESS_NETWORK		0x0506 /*802.11b Wireless Network Adaptor (at76c506)*/
/* Vendor: 1115: 3D Labs */
/* Vendor: 1116: Data Translation */
#define DEVICE_DATA3_DT3001				0x0022 /*DT3001*/
#define DEVICE_DATA3_DT3002				0x0023 /*DT3002*/
#define DEVICE_DATA3_DT3003				0x0024 /*DT3003*/
#define DEVICE_DATA3_DT3004				0x0025 /*DT3004*/
#define DEVICE_DATA3_DT3005				0x0026 /*DT3005*/
#define DEVICE_DATA3_DT3001_PGL				0x0027 /*DT3001-PGL*/
#define DEVICE_DATA3_DT3003_PGL				0x0028 /*DT3003-PGL*/
/* Vendor: 1117: Datacube, Inc */
#define DEVICE_DATACUBE_MAX_1C_SVGA_CARD		0x9500 /*Max-1C SVGA card*/
#define DEVICE_DATACUBE_MAX_1C_IMAGE_PROCESSING		0x9501 /*Max-1C image processing*/
/* Vendor: 1118: Berg Electronics */
/* Vendor: 1119: ICP Vortex Computersysteme GmbH */
#define DEVICE_ICP_GDT_6000_6020_6050			0x0000 /*GDT 6000/6020/6050*/
#define DEVICE_ICP_GDT_6000B_6010			0x0001 /*GDT 6000B/6010*/
#define DEVICE_ICP_GDT_6110_6510			0x0002 /*GDT 6110/6510*/
#define DEVICE_ICP_GDT_6120_6520			0x0003 /*GDT 6120/6520*/
#define DEVICE_ICP_GDT_6530				0x0004 /*GDT 6530*/
#define DEVICE_ICP_GDT_6550				0x0005 /*GDT 6550*/
#define DEVICE_ICP_GDT_6117_6517			0x0006 /*GDT 6117/6517*/
#define DEVICE_ICP_GDT_6127_6527			0x0007 /*GDT 6127/6527*/
#define DEVICE_ICP_GDT_6537				0x0008 /*GDT 6537*/
#define DEVICE_ICP_GDT_6557_6557_ECC			0x0009 /*GDT 6557/6557-ECC*/
#define DEVICE_ICP_GDT_6115_6515			0x000a /*GDT 6115/6515*/
#define DEVICE_ICP_GDT_6125_6525			0x000b /*GDT 6125/6525*/
#define DEVICE_ICP_GDT_6535				0x000c /*GDT 6535*/
#define DEVICE_ICP_GDT_6555				0x000d /*GDT 6555*/
#define DEVICE_ICP_GDT_6115_65152			0x0010 /*GDT 6115/6515*/
#define DEVICE_ICP_GDT_6125_65252			0x0011 /*GDT 6125/6525*/
#define DEVICE_ICP_GDT_65352				0x0012 /*GDT 6535*/
#define DEVICE_ICP_GDT_6555_6555_ECC			0x0013 /*GDT 6555/6555-ECC*/
#define DEVICE_ICP_GDT_6117RP_6517RP			0x0100 /*GDT 6117RP/6517RP*/
#define DEVICE_ICP_GDT_6127RP_6527RP			0x0101 /*GDT 6127RP/6527RP*/
#define DEVICE_ICP_GDT_6537RP				0x0102 /*GDT 6537RP*/
#define DEVICE_ICP_GDT_6557RP				0x0103 /*GDT 6557RP*/
#define DEVICE_ICP_GDT_6111RP_6511RP			0x0104 /*GDT 6111RP/6511RP*/
#define DEVICE_ICP_GDT_6121RP_6521RP			0x0105 /*GDT 6121RP/6521RP*/
#define DEVICE_ICP_GDT_6117RD_6517RD			0x0110 /*GDT 6117RD/6517RD*/
#define DEVICE_ICP_GDT_6127RD_6527RD			0x0111 /*GDT 6127RD/6527RD*/
#define DEVICE_ICP_GDT_6537RD				0x0112 /*GDT 6537RD*/
#define DEVICE_ICP_GDT_6557RD				0x0113 /*GDT 6557RD*/
#define DEVICE_ICP_GDT_6111RD_6511RD			0x0114 /*GDT 6111RD/6511RD*/
#define DEVICE_ICP_GDT_6121RD_6521RD			0x0115 /*GDT 6121RD/6521RD*/
#define DEVICE_ICP_GDT_6118RD_6518RD_6618RD		0x0118 /*GDT 6118RD/6518RD/6618RD*/
#define DEVICE_ICP_GDT_6128RD_6528RD_6628RD		0x0119 /*GDT 6128RD/6528RD/6628RD*/
#define DEVICE_ICP_GDT_6538RD_6638RD			0x011a /*GDT 6538RD/6638RD*/
#define DEVICE_ICP_GDT_6558RD_6658RD			0x011b /*GDT 6558RD/6658RD*/
#define DEVICE_ICP_GDT_6117RP2_6517RP2			0x0120 /*GDT 6117RP2/6517RP2*/
#define DEVICE_ICP_GDT_6127RP2_6527RP2			0x0121 /*GDT 6127RP2/6527RP2*/
#define DEVICE_ICP_GDT_6537RP2				0x0122 /*GDT 6537RP2*/
#define DEVICE_ICP_GDT_6557RP2				0x0123 /*GDT 6557RP2*/
#define DEVICE_ICP_GDT_6111RP2_6511RP2			0x0124 /*GDT 6111RP2/6511RP2*/
#define DEVICE_ICP_GDT_6121RP2_6521RP2			0x0125 /*GDT 6121RP2/6521RP2*/
#define DEVICE_ICP_GDT_6113RS_6513RS			0x0136 /*GDT 6113RS/6513RS*/
#define DEVICE_ICP_GDT_6123RS_6523RS			0x0137 /*GDT 6123RS/6523RS*/
#define DEVICE_ICP_GDT_6118RS_6518RS_6618RS		0x0138 /*GDT 6118RS/6518RS/6618RS*/
#define DEVICE_ICP_GDT_6128RS_6528RS_6628RS		0x0139 /*GDT 6128RS/6528RS/6628RS*/
#define DEVICE_ICP_GDT_6538RS_6638RS			0x013a /*GDT 6538RS/6638RS*/
#define DEVICE_ICP_GDT_6558RS_6658RS			0x013b /*GDT 6558RS/6658RS*/
#define DEVICE_ICP_GDT_6533RS_6633RS			0x013c /*GDT 6533RS/6633RS*/
#define DEVICE_ICP_GDT_6543RS_6643RS			0x013d /*GDT 6543RS/6643RS*/
#define DEVICE_ICP_GDT_6553RS_6653RS			0x013e /*GDT 6553RS/6653RS*/
#define DEVICE_ICP_GDT_6563RS_6663RS			0x013f /*GDT 6563RS/6663RS*/
#define DEVICE_ICP_GDT_7113RN_7513RN_7613RN		0x0166 /*GDT 7113RN/7513RN/7613RN*/
#define DEVICE_ICP_GDT_7123RN_7523RN_7623RN		0x0167 /*GDT 7123RN/7523RN/7623RN*/
#define DEVICE_ICP_GDT_7118RN_7518RN_7518RN		0x0168 /*GDT 7118RN/7518RN/7518RN*/
#define DEVICE_ICP_GDT_7128RN_7528RN_7628RN		0x0169 /*GDT 7128RN/7528RN/7628RN*/
#define DEVICE_ICP_GDT_7538RN_7638RN			0x016a /*GDT 7538RN/7638RN*/
#define DEVICE_ICP_GDT_7558RN_7658RN			0x016b /*GDT 7558RN/7658RN*/
#define DEVICE_ICP_GDT_7533RN_7633RN			0x016c /*GDT 7533RN/7633RN*/
#define DEVICE_ICP_GDT_7543RN_7643RN			0x016d /*GDT 7543RN/7643RN*/
#define DEVICE_ICP_GDT_7553RN_7653RN			0x016e /*GDT 7553RN/7653RN*/
#define DEVICE_ICP_GDT_7563RN_7663RN			0x016f /*GDT 7563RN/7663RN*/
#define DEVICE_ICP_GDT_4X13RZ				0x01d6 /*GDT 4x13RZ*/
#define DEVICE_ICP_GDT_4X23RZ				0x01d7 /*GDT 4x23RZ*/
#define DEVICE_ICP_GDT_8X13RZ				0x01f6 /*GDT 8x13RZ*/
#define DEVICE_ICP_GDT_8X23RZ				0x01f7 /*GDT 8x23RZ*/
#define DEVICE_ICP_GDT_8X33RZ				0x01fc /*GDT 8x33RZ*/
#define DEVICE_ICP_GDT_8X43RZ				0x01fd /*GDT 8x43RZ*/
#define DEVICE_ICP_GDT_8X53RZ				0x01fe /*GDT 8x53RZ*/
#define DEVICE_ICP_GDT_8X63RZ				0x01ff /*GDT 8x63RZ*/
#define DEVICE_ICP_GDT_6519RD_6619RD			0x0210 /*GDT 6519RD/6619RD*/
#define DEVICE_ICP_GDT_6529RD_6629RD			0x0211 /*GDT 6529RD/6629RD*/
#define DEVICE_ICP_GDT_7519RN_7619RN			0x0260 /*GDT 7519RN/7619RN*/
#define DEVICE_ICP_GDT_7529RN_7629RN			0x0261 /*GDT 7529RN/7629RN*/
#define DEVICE_ICP_GDT_MAXRP				0x02ff /*GDT MAXRP*/
#define DEVICE_ICP_GDT_NEWRX				0x0300 /*GDT NEWRX*/
/* Vendor: 111a: Efficient Networks, Inc */
#define DEVICE_EFFICIENT_155P_MF1_FPGA			0x0000 /*155P-MF1 (FPGA)*/
#define DEVICE_EFFICIENT_155P_MF1_ASIC			0x0002 /*155P-MF1 (ASIC)*/
#define DEVICE_EFFICIENT_ENI_25P_ATM			0x0003 /*ENI-25P ATM*/
	#define SUBDEVICE_EFFICIENT_ENI_25P_MINIPORT_ATM	0x111a0000 /*ENI-25p Miniport ATM Adapter*/
#define DEVICE_EFFICIENT_SPEEDSTREAM_LANAI		0x0005 /*SpeedStream (LANAI)*/
	#define SUBDEVICE_EFFICIENT_ENI_3010_ATM	0x111a0001 /*ENI-3010 ATM*/
	#define SUBDEVICE_EFFICIENT_ENI_3060_ADSL_VPI_0	0x111a0009 /*ENI-3060 ADSL (VPI=0)*/
	#define SUBDEVICE_EFFICIENT_ENI_3010_ATM2	0x111a0101 /*ENI-3010 ATM*/
	#define SUBDEVICE_EFFICIENT_ENI_3060CO_ADSL_VPI_0	0x111a0109 /*ENI-3060CO ADSL (VPI=0)*/
	#define SUBDEVICE_EFFICIENT_ENI_3060_ADSL_VPI_02	0x111a0809 /*ENI-3060 ADSL (VPI=0 or 8)*/
	#define SUBDEVICE_EFFICIENT_ENI_3060CO_ADSL_VPI_02	0x111a0909 /*ENI-3060CO ADSL (VPI=0 or 8)*/
	#define SUBDEVICE_EFFICIENT_ENI_3060_ADSL_VPI_0_15	0x111a0a09 /*ENI-3060 ADSL (VPI=<0..15>)*/
#define DEVICE_EFFICIENT_SPEEDSTREAM_ADSL		0x0007 /*SpeedStream ADSL*/
	#define SUBDEVICE_EFFICIENT_ENI_3061_ADSL_ASIC	0x111a1001 /*ENI-3061 ADSL [ASIC]*/
#define DEVICE_EFFICIENT_SPEEDSTREAM_1023_WIRELESS	0x1203 /*SpeedStream 1023 Wireless PCI Adapter*/
/* Vendor: 111b: Teledyne Electronic Systems */
/* Vendor: 111c: Tricord Systems Inc. */
#define DEVICE_TRICORD_POWERBIS_BRIDGE			0x0001 /*Powerbis Bridge*/
/* Vendor: 111d: Integrated Device Technology, Inc. */
#define DEVICE_INTEGRATED2_IDT77201_77211_155MBPS_ATM	0x0001 /*IDT77201/77211 155Mbps ATM SAR Controller [NICStAR]*/
#define DEVICE_INTEGRATED2_IDT77222_77252_155MBPS_ATM	0x0003 /*IDT77222/77252 155Mbps ATM MICRO ABR SAR Controller*/
#define DEVICE_INTEGRATED2_IDT77V252_155MBPS_ATM	0x0004 /*IDT77V252 155Mbps ATM MICRO ABR SAR Controller*/
#define DEVICE_INTEGRATED2_IDT77V222_155MBPS_ATM	0x0005 /*IDT77V222 155Mbps ATM MICRO ABR SAR Controller*/
/* Vendor: 111e: Eldec */
/* Vendor: 111f: Precision Digital Images */
#define DEVICE_PRECISION_PRECISION_MX_VIDEO		0x4a47 /*Precision MX Video engine interface*/
#define DEVICE_PRECISION_FRAME_CAPTURE_BUS		0x5243 /*Frame capture bus interface*/
/* Vendor: 1120: EMC Corporation */
/* Vendor: 1121: Zilog */
/* Vendor: 1122: Multi-tech Systems, Inc. */
/* Vendor: 1123: Excellent Design, Inc. */
/* Vendor: 1124: Leutron Vision AG */
/* Vendor: 1125: Eurocore */
/* Vendor: 1126: Vigra */
/* Vendor: 1127: FORE Systems Inc */
#define DEVICE_FORE_FORERUNNER_PCA_200_ATM		0x0200 /*ForeRunner PCA-200 ATM*/
#define DEVICE_FORE_PCA_200PC				0x0210 /*PCA-200PC*/
#define DEVICE_FORE_ATM					0x0250 /*ATM*/
#define DEVICE_FORE_FORERUNNER_PCA_200EPC_ATM		0x0300 /*ForeRunner PCA-200EPC ATM*/
#define DEVICE_FORE_ATM2					0x0310 /*ATM*/
#define DEVICE_FORE_FORERUNNERHE_ATM_ADAPTER		0x0400 /*ForeRunnerHE ATM Adapter*/
	#define SUBDEVICE_FORE_FORERUNNERHE_ATM		0x11270400 /*ForeRunnerHE ATM*/
/* Vendor: 1129: Firmworks */
/* Vendor: 112a: Hermes Electronics Company, Ltd. */
/* Vendor: 112b: Linotype - Hell AG */
/* Vendor: 112c: Zenith Data Systems */
/* Vendor: 112d: Ravicad */
/* Vendor: 112e: Infomedia Microelectronics Inc. */
/* Vendor: 112f: Imaging Technology Inc */
#define DEVICE_IMAGING_MVC_IC_PCI			0x0000 /*MVC IC-PCI*/
#define DEVICE_IMAGING_MVC_IM_PCI_VIDEO			0x0001 /*MVC IM-PCI Video frame grabber/processor*/
/* Vendor: 1130: Computervision */
/* Vendor: 1131: Philips Semiconductors */
#define DEVICE_PHILIPS_USB_1_1_HOST			0x1561 /*USB 1.1 Host Controller*/
#define DEVICE_PHILIPS_USB_2_0_HOST			0x1562 /*USB 2.0 Host Controller*/
#define DEVICE_PHILIPS_SMARTPCI56_UCB1500_56K_MODEM	0x3400 /*SmartPCI56(UCB1500) 56K Modem*/
#define DEVICE_PHILIPS_TRIMEDIA_TM1000_1100		0x5400 /*TriMedia TM1000/1100*/
#define DEVICE_PHILIPS_TRIMEDIA_TM_1300			0x5402 /*TriMedia TM-1300*/
#define DEVICE_PHILIPS_SAA7130_VIDEO_BROADCAST		0x7130 /*SAA7130 Video Broadcast Decoder*/
	#define SUBDEVICE_PHILIPS_LIVEVIEW_FLYVIDEO_2000	0x51680138 /*LiveView FlyVideo 2000*/
#define DEVICE_PHILIPS_SAA713X_AUDIO_VIDEO_BROADCAST	0x7133 /*SAA713X Audio+video broadcast decoder*/
	#define SUBDEVICE_PHILIPS_LIFEVIEW_FLYVIDEO_3000	0x51680138 /*LifeView FlyVideo 3000*/
	#define SUBDEVICE_PHILIPS_LIFEVIEW_FLYTV_PLATINUM	0x51680212 /*LifeView FlyTV Platinum mini*/
#define DEVICE_PHILIPS_SAA7134				0x7134 /*SAA7134*/
#define DEVICE_PHILIPS_SAA7135_AUDIO_VIDEO_BROADCAST	0x7135 /*SAA7135 Audio+video broadcast decoder*/
#define DEVICE_PHILIPS_SAA7145				0x7145 /*SAA7145*/
#define DEVICE_PHILIPS_SAA7146				0x7146 /*SAA7146*/
	#define SUBDEVICE_PHILIPS_FUJITSU_SIEMENS_DVB_C_CARD	0x110a0000 /*Fujitsu/Siemens DVB-C card rev1.5*/
	#define SUBDEVICE_PHILIPS_FUJITSU_SIEMENS_DVB_C_CARD2	0x110affff /*Fujitsu/Siemens DVB-C card rev1.5*/
	#define SUBDEVICE_PHILIPS_KNC1_DVB_S_BUDGET	0x11314f56 /*KNC1 DVB-S Budget*/
	#define SUBDEVICE_PHILIPS_FUJITSU_SIEMENS_ACTIVY_DVB_S	0x11314f61 /*Fujitsu-Siemens Activy DVB-S Budget*/
	#define SUBDEVICE_PHILIPS_DVRAPTOR_VIDEO_EDIT_CAPTURE	0x114b2003 /*DVRaptor Video Edit/Capture Card*/
	#define SUBDEVICE_PHILIPS_DV500_OVERLAY		0x11bd0006 /*DV500 Overlay*/
	#define SUBDEVICE_PHILIPS_DV500_OVERLAY2	0x11bd000a /*DV500 Overlay*/
	#define SUBDEVICE_PHILIPS_SIEMENS_TECHNOTREND_HAUPPAUGE_DVB_CARD	0x13c20000 /*Siemens/Technotrend/Hauppauge DVB card rev1.3 or rev1.5*/
	#define SUBDEVICE_PHILIPS_TECHNOTREND_HAUPPAUGE_DVB_CARD	0x13c20001 /*Technotrend/Hauppauge DVB card rev1.3 or rev1.6*/
	#define SUBDEVICE_PHILIPS_TECHNOTREND_HAUPPAUGE_DVB_CARD2	0x13c20002 /*Technotrend/Hauppauge DVB card rev2.1*/
	#define SUBDEVICE_PHILIPS_TECHNOTREND_HAUPPAUGE_DVB_CARD3	0x13c20003 /*Technotrend/Hauppauge DVB card rev2.1*/
	#define SUBDEVICE_PHILIPS_TECHNOTREND_HAUPPAUGE_DVB_CARD4	0x13c20004 /*Technotrend/Hauppauge DVB card rev2.1*/
	#define SUBDEVICE_PHILIPS_TECHNOTREND_HAUPPAUGE_DVB_CARD5	0x13c20006 /*Technotrend/Hauppauge DVB card rev1.3 or rev1.6*/
	#define SUBDEVICE_PHILIPS_TECHNOTREND_HAUPPAUGE_DVB_T	0x13c20008 /*Technotrend/Hauppauge DVB-T*/
	#define SUBDEVICE_PHILIPS_OCTAL_TECHNOTREND_DVB_C_FOR	0x13c2000a /*Octal/Technotrend DVB-C for iTV*/
	#define SUBDEVICE_PHILIPS_TECHNOTREND_BUDGET_HAUPPAUGE	0x13c21003 /*Technotrend-Budget / Hauppauge WinTV-NOVA-S DVB card*/
	#define SUBDEVICE_PHILIPS_TECHNOTREND_BUDGET_HAUPPAUGE2	0x13c21004 /*Technotrend-Budget / Hauppauge WinTV-NOVA-C DVB card*/
	#define SUBDEVICE_PHILIPS_TECHNOTREND_BUDGET_HAUPPAUGE3	0x13c21005 /*Technotrend-Budget / Hauppauge WinTV-NOVA-T DVB card*/
	#define SUBDEVICE_PHILIPS_TECHNOTREND_BUDGET_HAUPPAUGE4	0x13c2100c /*Technotrend-Budget / Hauppauge WinTV-NOVA-CI DVB card*/
	#define SUBDEVICE_PHILIPS_TECHNOTREND_BUDGET_HAUPPAUGE5	0x13c2100f /*Technotrend-Budget / Hauppauge WinTV-NOVA-CI DVB card*/
	#define SUBDEVICE_PHILIPS_TECHNOTREND_BUDGET_HAUPPAUGE6	0x13c21011 /*Technotrend-Budget / Hauppauge WinTV-NOVA-T DVB card*/
	#define SUBDEVICE_PHILIPS_SATELCO_MULTIMEDIA_DVB	0x13c21013 /*SATELCO Multimedia DVB*/
	#define SUBDEVICE_PHILIPS_TECHNOTREND_HAUPPAUGE_DVB_CARD6	0x13c21102 /*Technotrend/Hauppauge DVB card rev2.1*/
/* Vendor: 1132: Mitel Corp. */
/* Vendor: 1133: Eicon Networks Corporation */
#define DEVICE_EICON_EICONCARD_S90			0x7901 /*EiconCard S90*/
#define DEVICE_EICON_EICONCARD_S902			0x7902 /*EiconCard S90*/
#define DEVICE_EICON_EICONCARD_S91			0x7911 /*EiconCard S91*/
#define DEVICE_EICON_EICONCARD_S912			0x7912 /*EiconCard S91*/
#define DEVICE_EICON_EICONCARD_S94			0x7941 /*EiconCard S94*/
#define DEVICE_EICON_EICONCARD_S942			0x7942 /*EiconCard S94*/
#define DEVICE_EICON_EICONCARD_S943			0x7943 /*EiconCard S94*/
#define DEVICE_EICON_EICONCARD_S944			0x7944 /*EiconCard S94*/
#define DEVICE_EICON_EICONCARD_P92			0xb921 /*EiconCard P92*/
#define DEVICE_EICON_EICONCARD_P922			0xb922 /*EiconCard P92*/
#define DEVICE_EICON_EICONCARD_P923			0xb923 /*EiconCard P92*/
#define DEVICE_EICON_DIVA_PRO_2_0			0xe001 /*Diva Pro 2.0 S/T*/
#define DEVICE_EICON_DIVA_2_0_S_T			0xe002 /*Diva 2.0 S/T PCI*/
#define DEVICE_EICON_DIVA_PRO_2_02			0xe003 /*Diva Pro 2.0 U*/
#define DEVICE_EICON_DIVA_2_0_U				0xe004 /*Diva 2.0 U PCI*/
#define DEVICE_EICON_DIVA_2_01_S_T			0xe005 /*Diva 2.01 S/T PCI*/
#define DEVICE_EICON_DIVA_CT_S_T			0xe006 /*Diva CT S/T PCI*/
#define DEVICE_EICON_DIVA_CT_U				0xe007 /*Diva CT U PCI*/
#define DEVICE_EICON_DIVA_CT_LITE			0xe008 /*Diva CT Lite S/T PCI*/
#define DEVICE_EICON_DIVA_CT_LITE2			0xe009 /*Diva CT Lite U PCI*/
#define DEVICE_EICON_DIVA_ISDN_V_90_PCI			0xe00a /*Diva ISDN+V.90 PCI*/
#define DEVICE_EICON_DIVA_2_02_PCI			0xe00b /*Diva 2.02 PCI S/T*/
#define DEVICE_EICON_DIVA_2_02_PCI2			0xe00c /*Diva 2.02 PCI U*/
#define DEVICE_EICON_DIVA_ISDN_PRO			0xe00d /*Diva ISDN Pro 3.0 PCI*/
#define DEVICE_EICON_DIVA_ISDN_CT_S_T			0xe00e /*Diva ISDN+CT S/T PCI Rev 2*/
#define DEVICE_EICON_DIVA_SERVER_BRI_2M			0xe010 /*Diva Server BRI-2M PCI*/
	#define SUBDEVICE_EICON_FUJITSU_SIEMENS_ISDN	0x110a0021 /*Fujitsu Siemens ISDN S0*/
	#define SUBDEVICE_EICON_DIVA_SERVER_BRI_2M	0x80010014 /*Diva Server BRI-2M PCI Cornet NQ*/
#define DEVICE_EICON_DIVA_SERVER_BRI			0xe011 /*Diva Server BRI S/T Rev 2*/
#define DEVICE_EICON_DIVA_SERVER_4BRI_8M		0xe012 /*Diva Server 4BRI-8M PCI*/
	#define SUBDEVICE_EICON_DIVA_SERVER_4BRI_8M	0x80010014 /*Diva Server 4BRI-8M PCI Cornet NQ*/
#define DEVICE_EICON_DIVA_SERVER_4BRI			0xe013 /*Diva Server 4BRI Rev 2*/
	#define SUBDEVICE_EICON_DIVA_SERVER_V_4BRI_8	0x11331300 /*Diva Server V-4BRI-8*/
	#define SUBDEVICE_EICON_DIVA_SERVER_4BRI_8M2	0x1133e013 /*Diva Server 4BRI-8M 2.0 PCI*/
	#define SUBDEVICE_EICON_DIVA_SERVER_4BRI_8M3	0x80010014 /*Diva Server 4BRI-8M 2.0 PCI Cornet NQ*/
#define DEVICE_EICON_DIVA_SERVER_PRI_30M		0xe014 /*Diva Server PRI-30M PCI*/
	#define SUBDEVICE_EICON_DIVA_SERVER_PRI_30M	0x00080100 /*Diva Server PRI-30M PCI*/
	#define SUBDEVICE_EICON_DIVA_SERVER_PRI_30M2	0x80010014 /*Diva Server PRI-30M PCI Cornet NQ*/
#define DEVICE_EICON_DIVA_SERVER_PRI			0xe015 /*DIVA Server PRI Rev 2*/
	#define SUBDEVICE_EICON_DIVA_SERVER_PRI		0x1133e015 /*Diva Server PRI 2.0 PCI*/
	#define SUBDEVICE_EICON_DIVA_SERVER_PRI2	0x80010014 /*Diva Server PRI 2.0 PCI Cornet NQ*/
#define DEVICE_EICON_DIVA_SERVER_VOICE			0xe016 /*Diva Server Voice 4BRI PCI*/
	#define SUBDEVICE_EICON_DIVA_SERVER_PRI3	0x80010014 /*Diva Server PRI Cornet NQ*/
#define DEVICE_EICON_DIVA_SERVER_VOICE2			0xe017 /*Diva Server Voice 4BRI Rev 2*/
	#define SUBDEVICE_EICON_DIVA_SERVER_VOICE	0x1133e017 /*Diva Server Voice 4BRI-8M 2.0 PCI*/
	#define SUBDEVICE_EICON_DIVA_SERVER_VOICE2	0x80010014 /*Diva Server Voice 4BRI-8M 2.0 PCI Cornet NQ*/
#define DEVICE_EICON_DIVA_SERVER_BRI_2M2		0xe018 /*Diva Server BRI-2M 2.0 PCI*/
	#define SUBDEVICE_EICON_DIVA_SERVER_V_BRI_2	0x11331800 /*Diva Server V-BRI-2*/
	#define SUBDEVICE_EICON_DIVA_SERVER_BRI_2M2	0x1133e018 /*Diva Server BRI-2M 2.0 PCI*/
	#define SUBDEVICE_EICON_DIVA_SERVER_BRI_2M3	0x80010014 /*Diva Server BRI-2M 2.0 PCI Cornet NQ*/
#define DEVICE_EICON_DIVA_SERVER_VOICE3			0xe019 /*Diva Server Voice PRI Rev 2*/
	#define SUBDEVICE_EICON_DIVA_SERVER_VOICE3	0x1133e019 /*Diva Server Voice PRI 2.0 PCI*/
	#define SUBDEVICE_EICON_DIVA_SERVER_VOICE4	0x80010014 /*Diva Server Voice PRI 2.0 PCI Cornet NQ*/
#define DEVICE_EICON_DIVA_SERVER_2FX			0xe01a /*Diva Server 2FX*/
#define DEVICE_EICON_DIVA_SERVER_VOICE4			0xe01b /*Diva Server Voice BRI-2M 2.0 PCI*/
	#define SUBDEVICE_EICON_DIVA_SERVER_VOICE5	0x1133e01b /*Diva Server Voice BRI-2M 2.0 PCI*/
	#define SUBDEVICE_EICON_DIVA_SERVER_VOICE6	0x80010014 /*Diva Server Voice BRI-2M 2.0 PCI Cornet NQ*/
#define DEVICE_EICON_DIVA_SERVER_PRI2			0xe01c /*Diva Server PRI Rev 3*/
	#define SUBDEVICE_EICON_DIVA_SERVER_PRI_E1_T1_8	0x11331c01 /*Diva Server PRI/E1/T1-8*/
	#define SUBDEVICE_EICON_DIVA_SERVER_PRI_T1_24	0x11331c02 /*Diva Server PRI/T1-24*/
	#define SUBDEVICE_EICON_DIVA_SERVER_PRI_E1_30	0x11331c03 /*Diva Server PRI/E1-30*/
	#define SUBDEVICE_EICON_DIVA_SERVER_PRI_E1_T1	0x11331c04 /*Diva Server PRI/E1/T1*/
	#define SUBDEVICE_EICON_DIVA_SERVER_V_PRI_T1_24	0x11331c05 /*Diva Server V-PRI/T1-24*/
	#define SUBDEVICE_EICON_DIVA_SERVER_V_PRI_E1_30	0x11331c06 /*Diva Server V-PRI/E1-30*/
	#define SUBDEVICE_EICON_DIVA_SERVER_PRI_E1_T1_82	0x11331c07 /*Diva Server PRI/E1/T1-8 Cornet NQ*/
	#define SUBDEVICE_EICON_DIVA_SERVER_PRI_T1_242	0x11331c08 /*Diva Server PRI/T1-24 Cornet NQ*/
	#define SUBDEVICE_EICON_DIVA_SERVER_PRI_E1_302	0x11331c09 /*Diva Server PRI/E1-30 Cornet NQ*/
	#define SUBDEVICE_EICON_DIVA_SERVER_PRI_E1_T12	0x11331c0a /*Diva Server PRI/E1/T1 Cornet NQ*/
	#define SUBDEVICE_EICON_DIVA_SERVER_V_PRI_T1_242	0x11331c0b /*Diva Server V-PRI/T1-24 Cornet NQ*/
	#define SUBDEVICE_EICON_DIVA_SERVER_V_PRI_E1_302	0x11331c0c /*Diva Server V-PRI/E1-30 Cornet NQ*/
#define DEVICE_EICON_DIVA_SERVER_2PRI			0xe01e /*Diva Server 2PRI*/
	#define SUBDEVICE_EICON_DIVA_SERVER_V_2PRI_E1_60	0x11331e00 /*Diva Server V-2PRI/E1-60*/
	#define SUBDEVICE_EICON_DIVA_SERVER_V_2PRI_T1_48	0x11331e01 /*Diva Server V-2PRI/T1-48*/
	#define SUBDEVICE_EICON_DIVA_SERVER_2PRI_E1_60	0x11331e02 /*Diva Server 2PRI/E1-60*/
	#define SUBDEVICE_EICON_DIVA_SERVER_2PRI_T1_48	0x11331e03 /*Diva Server 2PRI/T1-48*/
#define DEVICE_EICON_DIVA_SERVER_4PRI			0xe020 /*Diva Server 4PRI*/
	#define SUBDEVICE_EICON_DIVA_SERVER_V_4PRI_E1_120	0x11332000 /*Diva Server V-4PRI/E1-120*/
	#define SUBDEVICE_EICON_DIVA_SERVER_V_4PRI_T1_96	0x11332001 /*Diva Server V-4PRI/T1-96*/
	#define SUBDEVICE_EICON_DIVA_SERVER_4PRI_E1_120	0x11332002 /*Diva Server 4PRI/E1-120*/
	#define SUBDEVICE_EICON_DIVA_SERVER_4PRI_T1_96	0x11332003 /*Diva Server 4PRI/T1-96*/
#define DEVICE_EICON_DIVA_SERVER_ANALOG_4P		0xe024 /*Diva Server Analog-4P*/
	#define SUBDEVICE_EICON_DIVA_SERVER_V_ANALOG_4P	0x11332400 /*Diva Server V-Analog-4P*/
	#define SUBDEVICE_EICON_DIVA_SERVER_ANALOG_4P	0x1133e024 /*Diva Server Analog-4P*/
#define DEVICE_EICON_DIVA_SERVER_ANALOG_8P		0xe028 /*Diva Server Analog-8P*/
	#define SUBDEVICE_EICON_DIVA_SERVER_V_ANALOG_8P	0x11332800 /*Diva Server V-Analog-8P*/
	#define SUBDEVICE_EICON_DIVA_SERVER_ANALOG_8P	0x1133e028 /*Diva Server Analog-8P*/
/* Vendor: 1134: Mercury Computer Systems */
#define DEVICE_MERCURY_RACEWAY_BRIDGE			0x0001 /*Raceway Bridge*/
#define DEVICE_MERCURY_DUAL_PCI_TO			0x0002 /*Dual PCI to RapidIO Bridge*/
/* Vendor: 1135: Fuji Xerox Co Ltd */
#define DEVICE_FUJI_PRINTER_CONTROLLER			0x0001 /*Printer controller*/
/* Vendor: 1136: Momentum Data Systems */
/* Vendor: 1137: Cisco Systems Inc */
/* Vendor: 1138: Ziatech Corporation */
#define DEVICE_ZIATECH_8905_STD_32			0x8905 /*8905 [STD 32 Bridge]*/
/* Vendor: 1139: Dynamic Pictures, Inc */
#define DEVICE_DYNAMIC_VGA_COMPATABLE_3D		0x0001 /*VGA Compatable 3D Graphics*/
/* Vendor: 113a: FWB Inc */
/* Vendor: 113b: Network Computing Devices */
/* Vendor: 113c: Cyclone Microsystems, Inc. */
#define DEVICE_CYCLONE_PCI_9060_I960_BRIDGE		0x0000 /*PCI-9060 i960 Bridge*/
#define DEVICE_CYCLONE_PCI_SDK_PCI_I960			0x0001 /*PCI-SDK [PCI i960 Evaluation Platform]*/
#define DEVICE_CYCLONE_PCI_911_I960JX_BASED_INTELLIGENT	0x0911 /*PCI-911 [i960Jx-based Intelligent I/O Controller]*/
#define DEVICE_CYCLONE_PCI_912_I960CF_BASED_INTELLIGENT	0x0912 /*PCI-912 [i960CF-based Intelligent I/O Controller]*/
#define DEVICE_CYCLONE_PCI_913				0x0913 /*PCI-913*/
#define DEVICE_CYCLONE_PCI_914_I_O_CONTROLLER		0x0914 /*PCI-914 [I/O Controller w/ secondary PCI bus]*/
/* Vendor: 113d: Leading Edge Products Inc */
/* Vendor: 113e: Sanyo Electric Co - Computer Engineering Dept */
/* Vendor: 113f: Equinox Systems, Inc. */
#define DEVICE_EQUINOX_SST_64P_ADAPTER			0x0808 /*SST-64P Adapter*/
#define DEVICE_EQUINOX_SST_128P_ADAPTER			0x1010 /*SST-128P Adapter*/
#define DEVICE_EQUINOX_SST_16P_DB_ADAPTER		0x80c0 /*SST-16P DB Adapter*/
#define DEVICE_EQUINOX_SST_16P_RJ_ADAPTER		0x80c4 /*SST-16P RJ Adapter*/
#define DEVICE_EQUINOX_SST_16P_ADAPTER			0x80c8 /*SST-16P Adapter*/
#define DEVICE_EQUINOX_SST_4P_ADAPTER			0x8888 /*SST-4P Adapter*/
#define DEVICE_EQUINOX_SST_8P_ADAPTER			0x9090 /*SST-8P Adapter*/
/* Vendor: 1140: Intervoice Inc */
/* Vendor: 1141: Crest Microsystem Inc */
/* Vendor: 1142: Alliance Semiconductor Corporation */
#define DEVICE_ALLIANCE_AP6410				0x3210 /*AP6410*/
#define DEVICE_ALLIANCE_PROVIDEO_6422			0x6422 /*ProVideo 6422*/
#define DEVICE_ALLIANCE_PROVIDEO_6424			0x6424 /*ProVideo 6424*/
#define DEVICE_ALLIANCE_PROMOTION_AT25			0x6425 /*ProMotion AT25*/
#define DEVICE_ALLIANCE_PROMOTION_AT3D			0x643d /*ProMotion AT3D*/
/* Vendor: 1143: NetPower, Inc */
/* Vendor: 1144: Cincinnati Milacron */
#define DEVICE_CINCINNATI_NOSERVO_CONTROLLER		0x0001 /*Noservo controller*/
/* Vendor: 1145: Workbit Corporation */
#define DEVICE_WORKBIT_NINJASCSI_32_WORKBIT		0x8007 /*NinjaSCSI-32 Workbit*/
#define DEVICE_WORKBIT_NINJASCSI_32_KME			0xf007 /*NinjaSCSI-32 KME*/
#define DEVICE_WORKBIT_NINJASCSI_32_WORKBIT2		0xf010 /*NinjaSCSI-32 Workbit*/
#define DEVICE_WORKBIT_NINJASCSI_32_LOGITEC		0xf012 /*NinjaSCSI-32 Logitec*/
#define DEVICE_WORKBIT_NINJASCSI_32_LOGITEC2		0xf013 /*NinjaSCSI-32 Logitec*/
#define DEVICE_WORKBIT_NINJASCSI_32_MELCO		0xf015 /*NinjaSCSI-32 Melco*/
/* Vendor: 1146: Force Computers */
/* Vendor: 1147: Interface Corp */
/* Vendor: 1148: SysKonnect */
#define DEVICE_SYSKONNECT_FDDI_ADAPTER			0x4000 /*FDDI Adapter*/
	#define SUBDEVICE_SYSKONNECT_NETELLIGENT_100_FDDI	0x0e11b03b /*Netelligent 100 FDDI DAS Fibre SC*/
	#define SUBDEVICE_SYSKONNECT_NETELLIGENT_100_FDDI2	0x0e11b03c /*Netelligent 100 FDDI SAS Fibre SC*/
	#define SUBDEVICE_SYSKONNECT_NETELLIGENT_100_FDDI3	0x0e11b03d /*Netelligent 100 FDDI DAS UTP*/
	#define SUBDEVICE_SYSKONNECT_NETELLIGENT_100_FDDI4	0x0e11b03e /*Netelligent 100 FDDI SAS UTP*/
	#define SUBDEVICE_SYSKONNECT_NETELLIGENT_100_FDDI5	0x0e11b03f /*Netelligent 100 FDDI SAS Fibre MIC*/
	#define SUBDEVICE_SYSKONNECT_FDDI_SK_5521_SK_NET	0x11485521 /*FDDI SK-5521 (SK-NET FDDI-UP)*/
	#define SUBDEVICE_SYSKONNECT_FDDI_SK_5522_SK_NET	0x11485522 /*FDDI SK-5522 (SK-NET FDDI-UP DAS)*/
	#define SUBDEVICE_SYSKONNECT_FDDI_SK_5541_SK_NET	0x11485541 /*FDDI SK-5541 (SK-NET FDDI-FP)*/
	#define SUBDEVICE_SYSKONNECT_FDDI_SK_5543_SK_NET	0x11485543 /*FDDI SK-5543 (SK-NET FDDI-LP)*/
	#define SUBDEVICE_SYSKONNECT_FDDI_SK_5544_SK_NET	0x11485544 /*FDDI SK-5544 (SK-NET FDDI-LP DAS)*/
	#define SUBDEVICE_SYSKONNECT_FDDI_SK_5821_SK_NET	0x11485821 /*FDDI SK-5821 (SK-NET FDDI-UP64)*/
	#define SUBDEVICE_SYSKONNECT_FDDI_SK_5822_SK_NET	0x11485822 /*FDDI SK-5822 (SK-NET FDDI-UP64 DAS)*/
	#define SUBDEVICE_SYSKONNECT_FDDI_SK_5841_SK_NET	0x11485841 /*FDDI SK-5841 (SK-NET FDDI-FP64)*/
	#define SUBDEVICE_SYSKONNECT_FDDI_SK_5843_SK_NET	0x11485843 /*FDDI SK-5843 (SK-NET FDDI-LP64)*/
	#define SUBDEVICE_SYSKONNECT_FDDI_SK_5844_SK_NET	0x11485844 /*FDDI SK-5844 (SK-NET FDDI-LP64 DAS)*/
#define DEVICE_SYSKONNECT_TOKEN_RING_ADAPTER		0x4200 /*Token Ring adapter*/
#define DEVICE_SYSKONNECT_SK_98XX_GIGABIT_ETHERNET	0x4300 /*SK-98xx Gigabit Ethernet Server Adapter*/
	#define SUBDEVICE_SYSKONNECT_SK_9821_GIGABIT_ETHERNET	0x11489821 /*SK-9821 Gigabit Ethernet Server Adapter (SK-NET GE-T)*/
	#define SUBDEVICE_SYSKONNECT_SK_9822_GIGABIT_ETHERNET	0x11489822 /*SK-9822 Gigabit Ethernet Server Adapter (SK-NET GE-T dual link)*/
	#define SUBDEVICE_SYSKONNECT_SK_9841_GIGABIT_ETHERNET	0x11489841 /*SK-9841 Gigabit Ethernet Server Adapter (SK-NET GE-LX)*/
	#define SUBDEVICE_SYSKONNECT_SK_9842_GIGABIT_ETHERNET	0x11489842 /*SK-9842 Gigabit Ethernet Server Adapter (SK-NET GE-LX dual link)*/
	#define SUBDEVICE_SYSKONNECT_SK_9843_GIGABIT_ETHERNET	0x11489843 /*SK-9843 Gigabit Ethernet Server Adapter (SK-NET GE-SX)*/
	#define SUBDEVICE_SYSKONNECT_SK_9844_GIGABIT_ETHERNET	0x11489844 /*SK-9844 Gigabit Ethernet Server Adapter (SK-NET GE-SX dual link)*/
	#define SUBDEVICE_SYSKONNECT_SK_9861_GIGABIT_ETHERNET	0x11489861 /*SK-9861 Gigabit Ethernet Server Adapter (SK-NET GE-SX Volition)*/
	#define SUBDEVICE_SYSKONNECT_SK_9862_GIGABIT_ETHERNET	0x11489862 /*SK-9862 Gigabit Ethernet Server Adapter (SK-NET GE-SX Volition dual link)*/
	#define SUBDEVICE_SYSKONNECT_SK_9871_GIGABIT_ETHERNET	0x11489871 /*SK-9871 Gigabit Ethernet Server Adapter (SK-NET GE-ZX)*/
	#define SUBDEVICE_SYSKONNECT_SK_9872_GIGABIT_ETHERNET	0x11489872 /*SK-9872 Gigabit Ethernet Server Adapter (SK-NET GE-ZX dual link)*/
	#define SUBDEVICE_SYSKONNECT_AT_2970SX_GIGABIT_ETHERNET	0x12592970 /*AT-2970SX Gigabit Ethernet Adapter*/
	#define SUBDEVICE_SYSKONNECT_AT_2970LX_GIGABIT_ETHERNET	0x12592971 /*AT-2970LX Gigabit Ethernet Adapter*/
	#define SUBDEVICE_SYSKONNECT_AT_2970TX_GIGABIT_ETHERNET	0x12592972 /*AT-2970TX Gigabit Ethernet Adapter*/
	#define SUBDEVICE_SYSKONNECT_AT_2971SX_GIGABIT_ETHERNET	0x12592973 /*AT-2971SX Gigabit Ethernet Adapter*/
	#define SUBDEVICE_SYSKONNECT_AT_2971T_GIGABIT_ETHERNET	0x12592974 /*AT-2971T Gigabit Ethernet Adapter*/
	#define SUBDEVICE_SYSKONNECT_AT_2970SX_2SC_GIGABIT_ETHERNET	0x12592975 /*AT-2970SX/2SC Gigabit Ethernet Adapter*/
	#define SUBDEVICE_SYSKONNECT_AT_2970LX_2SC_GIGABIT_ETHERNET	0x12592976 /*AT-2970LX/2SC Gigabit Ethernet Adapter*/
	#define SUBDEVICE_SYSKONNECT_AT_2970TX_2TX_GIGABIT_ETHERNET	0x12592977 /*AT-2970TX/2TX Gigabit Ethernet Adapter*/
#define DEVICE_SYSKONNECT_SK_98XX_V2_0_GIGABIT		0x4320 /*SK-98xx V2.0 Gigabit Ethernet Adapter*/
	#define SUBDEVICE_SYSKONNECT_MARVELL_RDK_8001_ADAPTER	0x11480121 /*Marvell RDK-8001 Adapter*/
	#define SUBDEVICE_SYSKONNECT_MARVELL_RDK_8002_ADAPTER	0x11480221 /*Marvell RDK-8002 Adapter*/
	#define SUBDEVICE_SYSKONNECT_MARVELL_RDK_8003_ADAPTER	0x11480321 /*Marvell RDK-8003 Adapter*/
	#define SUBDEVICE_SYSKONNECT_MARVELL_RDK_8004_ADAPTER	0x11480421 /*Marvell RDK-8004 Adapter*/
	#define SUBDEVICE_SYSKONNECT_MARVELL_RDK_8006_ADAPTER	0x11480621 /*Marvell RDK-8006 Adapter*/
	#define SUBDEVICE_SYSKONNECT_MARVELL_RDK_8007_ADAPTER	0x11480721 /*Marvell RDK-8007 Adapter*/
	#define SUBDEVICE_SYSKONNECT_MARVELL_RDK_8008_ADAPTER	0x11480821 /*Marvell RDK-8008 Adapter*/
	#define SUBDEVICE_SYSKONNECT_MARVELL_RDK_8009_ADAPTER	0x11480921 /*Marvell RDK-8009 Adapter*/
	#define SUBDEVICE_SYSKONNECT_MARVELL_RDK_8011_ADAPTER	0x11481121 /*Marvell RDK-8011 Adapter*/
	#define SUBDEVICE_SYSKONNECT_MARVELL_RDK_8012_ADAPTER	0x11481221 /*Marvell RDK-8012 Adapter*/
	#define SUBDEVICE_SYSKONNECT_SK_9521_V2_0_10_100_1000BASE_T	0x11483221 /*SK-9521 V2.0 10/100/1000Base-T Adapter*/
	#define SUBDEVICE_SYSKONNECT_SK_9821_V2_0_GIGABIT	0x11485021 /*SK-9821 V2.0 Gigabit Ethernet 10/100/1000Base-T Adapter*/
	#define SUBDEVICE_SYSKONNECT_SK_9841_V2_0_GIGABIT	0x11485041 /*SK-9841 V2.0 Gigabit Ethernet 1000Base-LX Adapter*/
	#define SUBDEVICE_SYSKONNECT_SK_9843_V2_0_GIGABIT	0x11485043 /*SK-9843 V2.0 Gigabit Ethernet 1000Base-SX Adapter*/
	#define SUBDEVICE_SYSKONNECT_SK_9851_V2_0_GIGABIT	0x11485051 /*SK-9851 V2.0 Gigabit Ethernet 1000Base-SX Adapter*/
	#define SUBDEVICE_SYSKONNECT_SK_9861_V2_0_GIGABIT	0x11485061 /*SK-9861 V2.0 Gigabit Ethernet 1000Base-SX Adapter*/
	#define SUBDEVICE_SYSKONNECT_SK_9871_V2_0_GIGABIT	0x11485071 /*SK-9871 V2.0 Gigabit Ethernet 1000Base-ZX Adapter*/
	#define SUBDEVICE_SYSKONNECT_SK_9521_10_100_1000BASE_T_ADAPTER	0x11489521 /*SK-9521 10/100/1000Base-T Adapter*/
#define DEVICE_SYSKONNECT_SK_9DXX_GIGABIT_ETHERNET	0x4400 /*SK-9Dxx Gigabit Ethernet Adapter*/
#define DEVICE_SYSKONNECT_SK_9MXX_GIGABIT_ETHERNET	0x4500 /*SK-9Mxx Gigabit Ethernet Adapter*/
#define DEVICE_SYSKONNECT_SK_9EXX_10_100_1000BASE_T_ADAPTER	0x9e00 /*SK-9Exx 10/100/1000Base-T Adapter*/
	#define SUBDEVICE_SYSKONNECT_SK_9E21_SERVER_ADAPTER	0x11482100 /*SK-9E21 Server Adapter*/
	#define SUBDEVICE_SYSKONNECT_SK_9E21D_10_100_1000BASE_T_ADAPTER	0x114821d0 /*SK-9E21D 10/100/1000Base-T Adapter*/
	#define SUBDEVICE_SYSKONNECT_SK_9E22_SERVER_ADAPTER	0x11482200 /*SK-9E22 Server Adapter*/
	#define SUBDEVICE_SYSKONNECT_SK_9E81_SERVER_ADAPTER	0x11488100 /*SK-9E81 Server Adapter*/
	#define SUBDEVICE_SYSKONNECT_SK_9E82_SERVER_ADAPTER	0x11488200 /*SK-9E82 Server Adapter*/
	#define SUBDEVICE_SYSKONNECT_SK_9E91_SERVER_ADAPTER	0x11489100 /*SK-9E91 Server Adapter*/
	#define SUBDEVICE_SYSKONNECT_SK_9E92_SERVER_ADAPTER	0x11489200 /*SK-9E92 Server Adapter*/
/* Vendor: 1149: Win System Corporation */
/* Vendor: 114a: VMIC */
#define DEVICE_VMIC_VMIPCI_5579_REFLECTIVE_MEMORY	0x5579 /*VMIPCI-5579 (Reflective Memory Card)*/
#define DEVICE_VMIC_VMIPCI_5587_REFLECTIVE_MEMORY	0x5587 /*VMIPCI-5587 (Reflective Memory Card)*/
#define DEVICE_VMIC_VMIC_PCI_7755			0x6504 /*VMIC PCI 7755 FPGA*/
#define DEVICE_VMIC_VMIVME_7587				0x7587 /*VMIVME-7587*/
/* Vendor: 114b: Canopus Co., Ltd */
/* Vendor: 114c: Annabooks */
/* Vendor: 114d: IC Corporation */
/* Vendor: 114e: Nikon Systems Inc */
/* Vendor: 114f: Digi International */
#define DEVICE_DIGI_ACCELEPORT_EPC			0x0002 /*AccelePort EPC*/
#define DEVICE_DIGI_RIGHTSWITCH_SE_6			0x0003 /*RightSwitch SE-6*/
#define DEVICE_DIGI_ACCELEPORT_XEM			0x0004 /*AccelePort Xem*/
#define DEVICE_DIGI_ACCELEPORT_XR			0x0005 /*AccelePort Xr*/
#define DEVICE_DIGI_ACCELEPORT_XR_C_X			0x0006 /*AccelePort Xr,C/X*/
#define DEVICE_DIGI_ACCELEPORT_XR_J			0x0009 /*AccelePort Xr/J*/
#define DEVICE_DIGI_ACCELEPORT_EPC_J			0x000a /*AccelePort EPC/J*/
#define DEVICE_DIGI_DATAFIREPRIME_T1_1_PORT		0x000c /*DataFirePRIme T1 (1-port)*/
#define DEVICE_DIGI_SYNCPORT_2_PORT_X_25_FR		0x000d /*SyncPort 2-Port (x.25/FR)*/
#define DEVICE_DIGI_ACCELEPORT_8R_EIA_232		0x0011 /*AccelePort 8r EIA-232 (IBM)*/
#define DEVICE_DIGI_ACCELEPORT_8R_EIA_422		0x0012 /*AccelePort 8r EIA-422*/
#define DEVICE_DIGI_ACCELEPORT_XR2			0x0013 /*AccelePort Xr*/
#define DEVICE_DIGI_ACCELEPORT_8R_EIA_4222		0x0014 /*AccelePort 8r EIA-422*/
#define DEVICE_DIGI_ACCELEPORT_XEM2			0x0015 /*AccelePort Xem*/
#define DEVICE_DIGI_ACCELEPORT_EPC_X			0x0016 /*AccelePort EPC/X*/
#define DEVICE_DIGI_ACCELEPORT_C_X			0x0017 /*AccelePort C/X*/
#define DEVICE_DIGI_DATAFIREPRIME_E1_1_PORT		0x001a /*DataFirePRIme E1 (1-port)*/
#define DEVICE_DIGI_ACCELEPORT_C_X_IBM			0x001b /*AccelePort C/X (IBM)*/
#define DEVICE_DIGI_DATAFIRE_RAS_T1_E1_PRI		0x001d /*DataFire RAS T1/E1/PRI*/
	#define SUBDEVICE_DIGI_DATAFIRE_RAS_E1		0x114f0050 /*DataFire RAS E1 Adapter*/
	#define SUBDEVICE_DIGI_DATAFIRE_RAS_DUAL	0x114f0051 /*DataFire RAS Dual E1 Adapter*/
	#define SUBDEVICE_DIGI_DATAFIRE_RAS_T1		0x114f0052 /*DataFire RAS T1 Adapter*/
	#define SUBDEVICE_DIGI_DATAFIRE_RAS_DUAL2	0x114f0053 /*DataFire RAS Dual T1 Adapter*/
#define DEVICE_DIGI_ACCELEPORT_RAS			0x0023 /*AccelePort RAS*/
#define DEVICE_DIGI_DATAFIRE_RAS_B4			0x0024 /*DataFire RAS B4 ST/U*/
	#define SUBDEVICE_DIGI_DATAFIRE_RAS_BRI		0x114f0030 /*DataFire RAS BRI U Adapter*/
	#define SUBDEVICE_DIGI_DATAFIRE_RAS_BRI2	0x114f0031 /*DataFire RAS BRI S/T Adapter*/
#define DEVICE_DIGI_ACCELEPORT_4R_920			0x0026 /*AccelePort 4r 920*/
#define DEVICE_DIGI_ACCELEPORT_XR_920			0x0027 /*AccelePort Xr 920*/
#define DEVICE_DIGI_CLASSICBOARD_4			0x0028 /*ClassicBoard 4*/
#define DEVICE_DIGI_CLASSICBOARD_8			0x0029 /*ClassicBoard 8*/
#define DEVICE_DIGI_ACCELEPORT_2R_920			0x0034 /*AccelePort 2r 920*/
#define DEVICE_DIGI_DATAFIRE_DSP_T1_E1_PRI		0x0035 /*DataFire DSP T1/E1/PRI cPCI*/
#define DEVICE_DIGI_ACCELEPORT_XP			0x0040 /*AccelePort Xp*/
#define DEVICE_DIGI_ACCELEPORT_2P			0x0042 /*AccelePort 2p*/
#define DEVICE_DIGI_ACCELEPORT_4P			0x0043 /*AccelePort 4p*/
#define DEVICE_DIGI_ACCELEPORT_8P			0x0044 /*AccelePort 8p*/
#define DEVICE_DIGI_ACCELEPORT_16P			0x0045 /*AccelePort 16p*/
#define DEVICE_DIGI_ACCELEPORT_32P			0x004e /*AccelePort 32p*/
#define DEVICE_DIGI_DATAFIRE_MICRO_V			0x0070 /*Datafire Micro V IOM2 (Europe)*/
#define DEVICE_DIGI_DATAFIRE_MICRO_V2			0x0071 /*Datafire Micro V (Europe)*/
#define DEVICE_DIGI_DATAFIRE_MICRO_V3			0x0072 /*Datafire Micro V IOM2 (North America)*/
#define DEVICE_DIGI_DATAFIRE_MICRO_V4			0x0073 /*Datafire Micro V (North America)*/
#define DEVICE_DIGI_DIGI_NEO_4				0x00b0 /*Digi Neo 4*/
#define DEVICE_DIGI_DIGI_NEO_8				0x00b1 /*Digi Neo 8*/
#define DEVICE_DIGI_DIGI_NEO_2				0x00c8 /*Digi Neo 2 DB9*/
#define DEVICE_DIGI_DIGI_NEO_22				0x00c9 /*Digi Neo 2 DB9 PRI*/
#define DEVICE_DIGI_DIGI_NEO_23				0x00ca /*Digi Neo 2 RJ45*/
#define DEVICE_DIGI_DIGI_NEO_24				0x00cb /*Digi Neo 2 RJ45 PRI*/
#define DEVICE_DIGI_CLASSICBOARD_4_422			0x00d0 /*ClassicBoard 4 422*/
#define DEVICE_DIGI_CLASSICBOARD_8_422			0x00d1 /*ClassicBoard 8 422*/
#define DEVICE_DIGI_AVANSTAR				0x6001 /*Avanstar*/
/* Vendor: 1150: Thinking Machines Corp */
/* Vendor: 1151: JAE Electronics Inc. */
/* Vendor: 1152: Megatek */
/* Vendor: 1153: Land Win Electronic Corp */
/* Vendor: 1154: Melco Inc */
/* Vendor: 1155: Pine Technology Ltd */
/* Vendor: 1156: Periscope Engineering */
/* Vendor: 1157: Avsys Corporation */
/* Vendor: 1158: Voarx R & D Inc */
#define DEVICE_VOARX_TOKENET_VG_1001_10M_ANYLAN		0x3011 /*Tokenet/vg 1001/10m anylan*/
#define DEVICE_VOARX_LANFLEET_TRUEVALUE			0x9050 /*Lanfleet/Truevalue*/
#define DEVICE_VOARX_LANFLEET_TRUEVALUE2		0x9051 /*Lanfleet/Truevalue*/
/* Vendor: 1159: Mutech Corp */
#define DEVICE_MUTECH_MV_1000				0x0001 /*MV-1000*/
/* Vendor: 115a: Harlequin Ltd */
/* Vendor: 115b: Parallax Graphics */
/* Vendor: 115c: Photron Ltd. */
/* Vendor: 115d: Xircom */
#define DEVICE_XIRCOM_CARDBUS_ETHERNET_10_100		0x0003 /*Cardbus Ethernet 10/100*/
	#define SUBDEVICE_XIRCOM_10_100_ETHERJET_CARDBUS	0x10140181 /*10/100 EtherJet Cardbus Adapter*/
	#define SUBDEVICE_XIRCOM_10_100_ETHERJET_CARDBUS2	0x10141181 /*10/100 EtherJet Cardbus Adapter*/
	#define SUBDEVICE_XIRCOM_10_100_ETHERJET_CARDBUS3	0x10148181 /*10/100 EtherJet Cardbus Adapter*/
	#define SUBDEVICE_XIRCOM_10_100_ETHERJET_CARDBUS4	0x10149181 /*10/100 EtherJet Cardbus Adapter*/
	#define SUBDEVICE_XIRCOM_CARDBUS_ETHERNET_10_100	0x115d0181 /*Cardbus Ethernet 10/100*/
	#define SUBDEVICE_XIRCOM_CARDBUS_ETHERNET_10_1002	0x115d1181 /*Cardbus Ethernet 10/100*/
	#define SUBDEVICE_XIRCOM_CARDBUS_ETHERNET_10_1003	0x11790181 /*Cardbus Ethernet 10/100*/
	#define SUBDEVICE_XIRCOM_ETHEREXPRESS_PRO_100_MOBILE	0x80868181 /*EtherExpress PRO/100 Mobile CardBus 32 Adapter*/
	#define SUBDEVICE_XIRCOM_ETHEREXPRESS_PRO_100_MOBILE2	0x80869181 /*EtherExpress PRO/100 Mobile CardBus 32 Adapter*/
#define DEVICE_XIRCOM_CARDBUS_ETHERNET_10_1002		0x0005 /*Cardbus Ethernet 10/100*/
	#define SUBDEVICE_XIRCOM_10_100_ETHERJET_CARDBUS5	0x10140182 /*10/100 EtherJet Cardbus Adapter*/
	#define SUBDEVICE_XIRCOM_10_100_ETHERJET_CARDBUS6	0x10141182 /*10/100 EtherJet Cardbus Adapter*/
	#define SUBDEVICE_XIRCOM_CARDBUS_ETHERNET_10_1004	0x115d0182 /*Cardbus Ethernet 10/100*/
	#define SUBDEVICE_XIRCOM_CARDBUS_ETHERNET_10_1005	0x115d1182 /*Cardbus Ethernet 10/100*/
#define DEVICE_XIRCOM_CARDBUS_ETHERNET_10_1003		0x0007 /*Cardbus Ethernet 10/100*/
	#define SUBDEVICE_XIRCOM_10_100_ETHERJET_CARDBUS7	0x10140182 /*10/100 EtherJet Cardbus Adapter*/
	#define SUBDEVICE_XIRCOM_10_100_ETHERJET_CARDBUS8	0x10141182 /*10/100 EtherJet Cardbus Adapter*/
	#define SUBDEVICE_XIRCOM_CARDBUS_ETHERNET_10_1006	0x115d0182 /*Cardbus Ethernet 10/100*/
	#define SUBDEVICE_XIRCOM_CARDBUS_ETHERNET_10_1007	0x115d1182 /*Cardbus Ethernet 10/100*/
#define DEVICE_XIRCOM_CARDBUS_ETHERNET_10_1004		0x000b /*Cardbus Ethernet 10/100*/
	#define SUBDEVICE_XIRCOM_10_100_ETHERJET_CARDBUS9	0x10140183 /*10/100 EtherJet Cardbus Adapter*/
	#define SUBDEVICE_XIRCOM_CARDBUS_ETHERNET_10_1008	0x115d0183 /*Cardbus Ethernet 10/100*/
#define DEVICE_XIRCOM_MINI_PCI_V_90_56K			0x000c /*Mini-PCI V.90 56k Modem*/
#define DEVICE_XIRCOM_CARDBUS_ETHERNET_10_1005		0x000f /*Cardbus Ethernet 10/100*/
	#define SUBDEVICE_XIRCOM_10_100_ETHERJET_CARDBUS10	0x10140183 /*10/100 EtherJet Cardbus Adapter*/
	#define SUBDEVICE_XIRCOM_CARDBUS_ETHERNET_10_1009	0x115d0183 /*Cardbus Ethernet 10/100*/
#define DEVICE_XIRCOM_MINI_PCI_K56FLEX_MODEM		0x00d4 /*Mini-PCI K56Flex Modem*/
#define DEVICE_XIRCOM_CARDBUS_56K_MODEM			0x0101 /*Cardbus 56k modem*/
	#define SUBDEVICE_XIRCOM_CARDBUS_56K_MODEM	0x115d1081 /*Cardbus 56k Modem*/
#define DEVICE_XIRCOM_CARDBUS_ETHERNET			0x0103 /*Cardbus Ethernet + 56k Modem*/
	#define SUBDEVICE_XIRCOM_CARDBUS_56K_MODEM2	0x10149181 /*Cardbus 56k Modem*/
	#define SUBDEVICE_XIRCOM_CARDBUS_ETHERNET_100	0x11151181 /*Cardbus Ethernet 100 + 56k Modem*/
	#define SUBDEVICE_XIRCOM_CBEM56G_100_ETHERNET	0x115d1181 /*CBEM56G-100 Ethernet + 56k Modem*/
	#define SUBDEVICE_XIRCOM_PRO_100_LAN		0x80869181 /*PRO/100 LAN + Modem56 CardBus*/
/* Vendor: 115e: Peer Protocols Inc */
/* Vendor: 115f: Maxtor Corporation */
/* Vendor: 1160: Megasoft Inc */
/* Vendor: 1161: PFU Limited */
/* Vendor: 1162: OA Laboratory Co Ltd */
/* Vendor: 1163: Rendition */
#define DEVICE_RENDITION_VERITE_1000			0x0001 /*Verite 1000*/
#define DEVICE_RENDITION_VERITE_V2000_V2100_V2200	0x2000 /*Verite V2000/V2100/V2200*/
	#define SUBDEVICE_RENDITION_STEALTH_II_S220	0x10922000 /*Stealth II S220*/
/* Vendor: 1164: Advanced Peripherals Technologies */
/* Vendor: 1165: Imagraph Corporation */
#define DEVICE_IMAGRAPH_MOTION_TPEG_RECORDER_PLAYER	0x0001 /*Motion TPEG Recorder/Player with audio*/
/* Vendor: 1166: ServerWorks */
#define DEVICE_SERVERWORKS_CMIC_LE			0x0000 /*CMIC-LE*/
#define DEVICE_SERVERWORKS_CNB20_LE_HOST_BRIDGE		0x0005 /*CNB20-LE Host Bridge*/
#define DEVICE_SERVERWORKS_CNB20HE_HOST_BRIDGE		0x0006 /*CNB20HE Host Bridge*/
#define DEVICE_SERVERWORKS_CNB20_LE_HOST_BRIDGE2	0x0007 /*CNB20-LE Host Bridge*/
#define DEVICE_SERVERWORKS_CNB20HE_HOST_BRIDGE2		0x0008 /*CNB20HE Host Bridge*/
#define DEVICE_SERVERWORKS_CNB20LE_HOST_BRIDGE		0x0009 /*CNB20LE Host Bridge*/
#define DEVICE_SERVERWORKS_CIOB30			0x0010 /*CIOB30*/
#define DEVICE_SERVERWORKS_CMIC_HE			0x0011 /*CMIC-HE*/
#define DEVICE_SERVERWORKS_CMIC_WS_HOST_BRIDGE		0x0012 /*CMIC-WS Host Bridge (GC-LE chipset)*/
#define DEVICE_SERVERWORKS_CNB20_HE_HOST_BRIDGE		0x0013 /*CNB20-HE Host Bridge*/
#define DEVICE_SERVERWORKS_CMIC_LE_HOST_BRIDGE		0x0014 /*CMIC-LE Host Bridge (GC-LE chipset)*/
#define DEVICE_SERVERWORKS_CMIC_GC_HOST_BRIDGE		0x0015 /*CMIC-GC Host Bridge*/
#define DEVICE_SERVERWORKS_CMIC_GC_HOST_BRIDGE2		0x0016 /*CMIC-GC Host Bridge*/
#define DEVICE_SERVERWORKS_GCNB_LE_HOST_BRIDGE		0x0017 /*GCNB-LE Host Bridge*/
#define DEVICE_SERVERWORKS_CIOB_X2_PCI_X_I_O		0x0101 /*CIOB-X2 PCI-X I/O Bridge*/
#define DEVICE_SERVERWORKS_CIOB_E_I_O_BRIDGE		0x0110 /*CIOB-E I/O Bridge with Gigabit Ethernet*/
#define DEVICE_SERVERWORKS_OSB4_SOUTH_BRIDGE		0x0200 /*OSB4 South Bridge*/
#define DEVICE_SERVERWORKS_CSB5_SOUTH_BRIDGE		0x0201 /*CSB5 South Bridge*/
	#define SUBDEVICE_SERVERWORKS_CT8_MAINBOARD	0x4c531080 /*CT8 mainboard*/
#define DEVICE_SERVERWORKS_CSB6_SOUTH_BRIDGE		0x0203 /*CSB6 South Bridge*/
#define DEVICE_SERVERWORKS_OSB4_IDE_CONTROLLER		0x0211 /*OSB4 IDE Controller*/
#define DEVICE_SERVERWORKS_CSB5_IDE_CONTROLLER		0x0212 /*CSB5 IDE Controller*/
	#define SUBDEVICE_SERVERWORKS_CT8_MAINBOARD2	0x4c531080 /*CT8 mainboard*/
#define DEVICE_SERVERWORKS_CSB6_RAID_IDE_CONTROLLER	0x0213 /*CSB6 RAID/IDE Controller*/
#define DEVICE_SERVERWORKS_CSB6_IDE_CONTROLLER		0x0217 /*CSB6 IDE Controller*/
#define DEVICE_SERVERWORKS_OSB4_CSB5_OHCI_USB		0x0220 /*OSB4/CSB5 OHCI USB Controller*/
	#define SUBDEVICE_SERVERWORKS_CT8_MAINBOARD3	0x4c531080 /*CT8 mainboard*/
#define DEVICE_SERVERWORKS_CSB6_OHCI_USB		0x0221 /*CSB6 OHCI USB Controller*/
#define DEVICE_SERVERWORKS_CSB5_LPC_BRIDGE		0x0225 /*CSB5 LPC bridge*/
	#define SUBDEVICE_SERVERWORKS_CT8_MAINBOARD4	0x4c531080 /*CT8 mainboard*/
#define DEVICE_SERVERWORKS_GCLE_2_HOST_BRIDGE		0x0227 /*GCLE-2 Host Bridge*/
#define DEVICE_SERVERWORKS_CSB5_LPC_BRIDGE2		0x0230 /*CSB5 LPC bridge*/
	#define SUBDEVICE_SERVERWORKS_CT8_MAINBOARD5	0x4c531080 /*CT8 mainboard*/
#define DEVICE_SERVERWORKS_K2_SATA			0x0240 /*K2 SATA*/
/* Vendor: 1167: Mutoh Industries Inc */
/* Vendor: 1168: Thine Electronics Inc */
/* Vendor: 1169: Centre for Development of Advanced Computing */
/* Vendor: 116a: Polaris Communications */
#define DEVICE_POLARIS_BUS_TAG_CHANNEL			0x6100 /*Bus/Tag Channel*/
#define DEVICE_POLARIS_ESCON_CHANNEL			0x6800 /*Escon Channel*/
#define DEVICE_POLARIS_BUS_TAG_CHANNEL2			0x7100 /*Bus/Tag Channel*/
#define DEVICE_POLARIS_ESCON_CHANNEL2			0x7800 /*Escon Channel*/
/* Vendor: 116b: Connectware Inc */
/* Vendor: 116c: Intelligent Resources Integrated Systems */
/* Vendor: 116d: Martin-Marietta */
/* Vendor: 116e: Electronics for Imaging */
/* Vendor: 116f: Workstation Technology */
/* Vendor: 1170: Inventec Corporation */
/* Vendor: 1171: Loughborough Sound Images Plc */
/* Vendor: 1172: Altera Corporation */
/* Vendor: 1173: Adobe Systems, Inc */
/* Vendor: 1174: Bridgeport Machines */
/* Vendor: 1175: Mitron Computer Inc. */
/* Vendor: 1176: SBE Incorporated */
/* Vendor: 1177: Silicon Engineering */
/* Vendor: 1178: Alfa, Inc. */
#define DEVICE_ALFA_FAST_ETHERNET_ADAPTER		0xafa1 /*Fast Ethernet Adapter*/
/* Vendor: 1179: Toshiba America Info Systems */
#define DEVICE_TOSHIBA2_EX_IDE_TYPE_B			0x0103 /*EX-IDE Type-B*/
#define DEVICE_TOSHIBA2_DVD_DECODER_CARD		0x0404 /*DVD Decoder card*/
#define DEVICE_TOSHIBA2_TECRA_VIDEO_CAPTURE		0x0406 /*Tecra Video Capture device*/
#define DEVICE_TOSHIBA2_DVD_DECODER_CARD2		0x0407 /*DVD Decoder card (Version 2)*/
#define DEVICE_TOSHIBA2_601				0x0601 /*601*/
#define DEVICE_TOSHIBA2_TOPIC95_PCI_TO			0x0603 /*ToPIC95 PCI to CardBus Bridge for Notebooks*/
#define DEVICE_TOSHIBA2_TOPIC95				0x060a /*ToPIC95*/
#define DEVICE_TOSHIBA2_TOPIC97				0x060f /*ToPIC97*/
#define DEVICE_TOSHIBA2_TOPIC100_PCI_TO			0x0617 /*ToPIC100 PCI to Cardbus Bridge with ZV Support*/
#define DEVICE_TOSHIBA2_CPU_TO_PCI			0x0618 /*CPU to PCI and PCI to ISA bridge*/
#define DEVICE_TOSHIBA2_FIR_PORT			0x0701 /*FIR Port*/
#define DEVICE_TOSHIBA2_TC6371AF_SMARTMEDIA_CONTROLLER	0x0804 /*TC6371AF SmartMedia Controller*/
#define DEVICE_TOSHIBA2_SD_TYPA_CONTROLLER		0x0805 /*SD TypA Controller*/
#define DEVICE_TOSHIBA2_FIR_PORT_TYPE_DO		0x0d01 /*FIR Port Type-DO*/
	#define SUBDEVICE_TOSHIBA2_FIR_PORT_TYPE_DO	0x11790001 /*FIR Port Type-DO*/
/* Vendor: 117a: A-Trend Technology */
/* Vendor: 117b: L G Electronics, Inc. */
/* Vendor: 117c: Atto Technology */
/* Vendor: 117d: Becton & Dickinson */
/* Vendor: 117e: T/R Systems */
/* Vendor: 117f: Integrated Circuit Systems */
/* Vendor: 1180: Ricoh Co Ltd */
#define DEVICE_RICOH_RL5C465				0x0465 /*RL5c465*/
#define DEVICE_RICOH_RL5C466				0x0466 /*RL5c466*/
#define DEVICE_RICOH_RL5C475				0x0475 /*RL5c475*/
	#define SUBDEVICE_RICOH_VPR_MATRIX_170B4	0x144dc006 /*vpr Matrix 170B4 CardBus bridge*/
#define DEVICE_RICOH_RL5C476_II				0x0476 /*RL5c476 II*/
	#define SUBDEVICE_RICOH_THINKPAD_A_T_X_SERIES	0x10140185 /*ThinkPad A/T/X Series*/
	#define SUBDEVICE_RICOH_VAIO_PCG_FX403		0x104d80df /*Vaio PCG-FX403*/
	#define SUBDEVICE_RICOH_VAIO_PCG_GR214EP_GR214MP_GR215MP_GR314MP_GR315MP	0x104d80e7 /*VAIO PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP*/
	#define SUBDEVICE_RICOH_PCD_RP_220S		0x14ef0220 /*PCD-RP-220S*/
#define DEVICE_RICOH_RL5C477				0x0477 /*RL5c477*/
#define DEVICE_RICOH_RL5C478				0x0478 /*RL5c478*/
	#define SUBDEVICE_RICOH_THINKPAD_A30P_2653_64G	0x10140184 /*ThinkPad A30p (2653-64G)*/
#define DEVICE_RICOH_R5C522_IEEE_1394			0x0522 /*R5C522 IEEE 1394 Controller*/
	#define SUBDEVICE_RICOH_THINKPAD_A30P_2653_64G2	0x101401cf /*ThinkPad A30p (2653-64G)*/
#define DEVICE_RICOH_R5C551_IEEE_1394			0x0551 /*R5C551 IEEE 1394 Controller*/
	#define SUBDEVICE_RICOH_VPR_MATRIX_170B42	0x144dc006 /*vpr Matrix 170B4*/
#define DEVICE_RICOH_R5C552_IEEE_1394			0x0552 /*R5C552 IEEE 1394 Controller*/
	#define SUBDEVICE_RICOH_THINKPAD_A_T_X_SERIES2	0x10140511 /*ThinkPad A/T/X Series*/
/* Vendor: 1181: Telmatics International */
/* Vendor: 1183: Fujikura Ltd */
/* Vendor: 1184: Forks Inc */
/* Vendor: 1185: Dataworld International Ltd */
/* Vendor: 1186: D-Link System Inc */
#define DEVICE_D_LINK_DC21041				0x0100 /*DC21041*/
#define DEVICE_D_LINK_DL10050_SUNDANCE_ETHERNET		0x1002 /*DL10050 Sundance Ethernet*/
	#define SUBDEVICE_D_LINK_DFE_550TX		0x11861002 /*DFE-550TX*/
	#define SUBDEVICE_D_LINK_DFE_580TX		0x11861012 /*DFE-580TX*/
#define DEVICE_D_LINK_AIRPLUS_XTREME_G			0x1025 /*AirPlus Xtreme G DWL-G650 Adapter*/
#define DEVICE_D_LINK_AIRXPERT_DWL_AG650_WIRELESS	0x1026 /*AirXpert DWL-AG650 Wireless Cardbus Adapter*/
#define DEVICE_D_LINK_AIRXPERT_DWL_AG650_WIRELESS2	0x1043 /*AirXpert DWL-AG650 Wireless Cardbus Adapter*/
#define DEVICE_D_LINK_RTL8139_ETHERNET			0x1300 /*RTL8139 Ethernet*/
	#define SUBDEVICE_D_LINK_DFE_538TX_10_100_ETHERNET	0x11861300 /*DFE-538TX 10/100 Ethernet Adapter*/
	#define SUBDEVICE_D_LINK_DFE_530TX_10_100_ETHERNET	0x11861301 /*DFE-530TX+ 10/100 Ethernet Adapter*/
#define DEVICE_D_LINK_DFE_690TXD_CARDBUS_PC		0x1340 /*DFE-690TXD CardBus PC Card*/
#define DEVICE_D_LINK_DFE_680TXD_CARDBUS_PC		0x1541 /*DFE-680TXD CardBus PC Card*/
#define DEVICE_D_LINK_DRP_32TXD_CARDBUS_PC		0x1561 /*DRP-32TXD Cardbus PC Card*/
#define DEVICE_D_LINK_AIRPLUS_XTREME_G2			0x2027 /*AirPlus Xtreme G DWL-G520 Adapter*/
#define DEVICE_D_LINK_AIRPLUS_XTREME_G3			0x3203 /*AirPlus Xtreme G DWL-G520 Adapter*/
#define DEVICE_D_LINK_DWL_510_2_4GHZ_WIRELESS		0x3300 /*DWL-510 2.4GHz Wireless PCI Adapter*/
#define DEVICE_D_LINK_AIRPRO_DWL_A650_WIRELESS		0x3a03 /*AirPro DWL-A650 Wireless Cardbus Adapter(rev.B)*/
#define DEVICE_D_LINK_AIRPRO_DWL_AB650_MULTIMODE	0x3a04 /*AirPro DWL-AB650 Multimode Wireless Cardbus Adapter*/
#define DEVICE_D_LINK_AIRPRO_DWL_AB520_MULTIMODE	0x3a05 /*AirPro DWL-AB520 Multimode Wireless PCI Adapter*/
#define DEVICE_D_LINK_AIRXPERT_DWL_AG650_WIRELESS3	0x3a07 /*AirXpert DWL-AG650 Wireless Cardbus Adapter*/
#define DEVICE_D_LINK_AIRXPERT_DWL_AG520_WIRELESS	0x3a08 /*AirXpert DWL-AG520 Wireless PCI Adapter*/
#define DEVICE_D_LINK_AIRXPERT_DWL_AG650_WIRELESS4	0x3a10 /*AirXpert DWL-AG650 Wireless Cardbus Adapter(rev.B)*/
#define DEVICE_D_LINK_AIRXPERT_DWL_AG520_WIRELESS2	0x3a11 /*AirXpert DWL-AG520 Wireless PCI Adapter(rev.B)*/
#define DEVICE_D_LINK_AIRPLUS_DWL_G650_WIRELESS		0x3a12 /*AirPlus DWL-G650 Wireless Cardbus Adapter(rev.C)*/
#define DEVICE_D_LINK_AIRPLUS_DWL_G520_WIRELESS		0x3a13 /*AirPlus DWL-G520 Wireless PCI Adapter(rev.B)*/
#define DEVICE_D_LINK_AIRPREMIER_DWL_AG530_WIRELESS	0x3a14 /*AirPremier DWL-AG530 Wireless PCI Adapter*/
#define DEVICE_D_LINK_AIRXPERT_DWL_AG660_WIRELESS	0x3a63 /*AirXpert DWL-AG660 Wireless Cardbus Adapter*/
#define DEVICE_D_LINK_DWL_G650_CARDBUS_PC		0x3b05 /*DWL-G650+ CardBus PC Card*/
#define DEVICE_D_LINK_DL2000_BASED_GIGABIT_ETHERNET	0x4000 /*DL2000-based Gigabit Ethernet*/
#define DEVICE_D_LINK_GIGABIT_ETHERNET_ADAPTER		0x4c00 /*Gigabit Ethernet Adapter*/
	#define SUBDEVICE_D_LINK_DGE_530T_GIGABIT_ETHERNET	0x11864c00 /*DGE-530T Gigabit Ethernet Adapter*/
#define DEVICE_D_LINK_D_LINK_DWL_650_CARDBUS		0x8400 /*D-Link DWL-650+ CardBus PC Card*/
/* Vendor: 1187: Advanced Technology Laboratories, Inc. */
/* Vendor: 1188: Shima Seiki Manufacturing Ltd. */
/* Vendor: 1189: Matsushita Electronics Co Ltd */
/* Vendor: 118a: Hilevel Technology */
/* Vendor: 118b: Hypertec Pty Limited */
/* Vendor: 118c: Corollary, Inc */
#define DEVICE_COROLLARY_PCIB_C_BUS_II			0x0014 /*PCIB [C-bus II to PCI bus host bridge chip]*/
#define DEVICE_COROLLARY_INTEL_8_WAY_XEON		0x1117 /*Intel 8-way XEON Profusion Chipset [Cache Coherency Filter]*/
/* Vendor: 118d: BitFlow Inc */
#define DEVICE_BITFLOW_RAPTOR_PCI_FRAMEGRABBER		0x0001 /*Raptor-PCI framegrabber*/
#define DEVICE_BITFLOW_MODEL_12_ROAD			0x0012 /*Model 12 Road Runner Frame Grabber*/
#define DEVICE_BITFLOW_MODEL_14_ROAD			0x0014 /*Model 14 Road Runner Frame Grabber*/
#define DEVICE_BITFLOW_MODEL_24_ROAD			0x0024 /*Model 24 Road Runner Frame Grabber*/
#define DEVICE_BITFLOW_MODEL_44_ROAD			0x0044 /*Model 44 Road Runner Frame Grabber*/
#define DEVICE_BITFLOW_MODEL_12_ROAD2			0x0112 /*Model 12 Road Runner Frame Grabber*/
#define DEVICE_BITFLOW_MODEL_14_ROAD2			0x0114 /*Model 14 Road Runner Frame Grabber*/
#define DEVICE_BITFLOW_MODEL_24_ROAD2			0x0124 /*Model 24 Road Runner Frame Grabber*/
#define DEVICE_BITFLOW_MODEL_44_ROAD2			0x0144 /*Model 44 Road Runner Frame Grabber*/
#define DEVICE_BITFLOW_MODEL_12_ROAD3			0x0212 /*Model 12 Road Runner Frame Grabber*/
#define DEVICE_BITFLOW_MODEL_14_ROAD3			0x0214 /*Model 14 Road Runner Frame Grabber*/
#define DEVICE_BITFLOW_MODEL_24_ROAD3			0x0224 /*Model 24 Road Runner Frame Grabber*/
#define DEVICE_BITFLOW_MODEL_44_ROAD3			0x0244 /*Model 44 Road Runner Frame Grabber*/
#define DEVICE_BITFLOW_MODEL_12_ROAD4			0x0312 /*Model 12 Road Runner Frame Grabber*/
#define DEVICE_BITFLOW_MODEL_14_ROAD4			0x0314 /*Model 14 Road Runner Frame Grabber*/
#define DEVICE_BITFLOW_MODEL_24_ROAD4			0x0324 /*Model 24 Road Runner Frame Grabber*/
#define DEVICE_BITFLOW_MODEL_44_ROAD4			0x0344 /*Model 44 Road Runner Frame Grabber*/
/* Vendor: 118e: Hermstedt GmbH */
/* Vendor: 118f: Green Logic */
/* Vendor: 1190: Tripace */
#define DEVICE_TRIPACE_TP_910_920_940_PCI_ULTRA_WIDE	0xc731 /*TP-910/920/940 PCI Ultra(Wide) SCSI Adapter*/
/* Vendor: 1191: Artop Electronic Corp */
#define DEVICE_ARTOP_SCSI_CACHE_HOST			0x0003 /*SCSI Cache Host Adapter*/
#define DEVICE_ARTOP_ATP8400				0x0004 /*ATP8400*/
#define DEVICE_ARTOP_ATP850UF				0x0005 /*ATP850UF*/
#define DEVICE_ARTOP_ATP860_NO_BIOS			0x0006 /*ATP860 NO-BIOS*/
#define DEVICE_ARTOP_ATP860				0x0007 /*ATP860*/
#define DEVICE_ARTOP_ATP865_NO_ROM			0x0008 /*ATP865 NO-ROM*/
#define DEVICE_ARTOP_ATP865				0x0009 /*ATP865*/
#define DEVICE_ARTOP_AEC6710_SCSI_2_HOST		0x8002 /*AEC6710 SCSI-2 Host Adapter*/
#define DEVICE_ARTOP_AEC6712UW_SCSI			0x8010 /*AEC6712UW SCSI*/
#define DEVICE_ARTOP_AEC6712U_SCSI			0x8020 /*AEC6712U SCSI*/
#define DEVICE_ARTOP_AEC6712S_SCSI			0x8030 /*AEC6712S SCSI*/
#define DEVICE_ARTOP_AEC6712D_SCSI			0x8040 /*AEC6712D SCSI*/
#define DEVICE_ARTOP_AEC6712SUW_SCSI			0x8050 /*AEC6712SUW SCSI*/
/* Vendor: 1192: Densan Company Ltd */
/* Vendor: 1193: Zeitnet Inc. */
#define DEVICE_ZEITNET_1221				0x0001 /*1221*/
#define DEVICE_ZEITNET_1225				0x0002 /*1225*/
/* Vendor: 1194: Toucan Technology */
/* Vendor: 1195: Ratoc System Inc */
/* Vendor: 1196: Hytec Electronics Ltd */
/* Vendor: 1197: Gage Applied Sciences, Inc. */
#define DEVICE_GAGE_COMPUSCOPE_82G_8BIT			0x010c /*CompuScope 82G 8bit 2GS/s Analog Input Card*/
/* Vendor: 1198: Lambda Systems Inc */
/* Vendor: 1199: Attachmate Corporation */
/* Vendor: 119a: Mind Share, Inc. */
/* Vendor: 119b: Omega Micro Inc. */
#define DEVICE_OMEGA_82C092G				0x1221 /*82C092G*/
/* Vendor: 119c: Information Technology Inst. */
/* Vendor: 119d: Bug, Inc. Sapporo Japan */
/* Vendor: 119e: Fujitsu Microelectronics Ltd. */
#define DEVICE_FUJITSU3_FIRESTREAM_155			0x0001 /*FireStream 155*/
#define DEVICE_FUJITSU3_FIRESTREAM_50			0x0003 /*FireStream 50*/
/* Vendor: 119f: Bull HN Information Systems */
/* Vendor: 11a0: Convex Computer Corporation */
/* Vendor: 11a1: Hamamatsu Photonics K.K. */
/* Vendor: 11a2: Sierra Research and Technology */
/* Vendor: 11a3: Deuretzbacher GmbH & Co. Eng. KG */
/* Vendor: 11a4: Barco Graphics NV */
/* Vendor: 11a5: Microunity Systems Eng. Inc */
/* Vendor: 11a6: Pure Data Ltd. */
/* Vendor: 11a7: Power Computing Corp. */
/* Vendor: 11a8: Systech Corp. */
/* Vendor: 11a9: InnoSys Inc. */
#define DEVICE_INNOSYS_AMCC_S933Q_INTELLIGENT		0x4240 /*AMCC S933Q Intelligent Serial Card*/
/* Vendor: 11aa: Actel */
/* Vendor: 11ab: Marvell Technology Group Ltd. */
#define DEVICE_MARVELL_GT_64010_64010A_SYSTEM_CONTROLLER	0x0146 /*GT-64010/64010A System Controller*/
#define DEVICE_MARVELL_W8300_802_11_ADAPTER		0x138f /*W8300 802.11 Adapter (rev 07)*/
#define DEVICE_MARVELL_MARVELL_W8300_802_11		0x1fa6 /*Marvell W8300 802.11 Adapter*/
#define DEVICE_MARVELL_GIGABIT_ETHERNET_CONTROLLER	0x4320 /*Gigabit Ethernet Controller*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8001_GIGABIT	0x10190f38 /*Marvell 88E8001 Gigabit Ethernet Controller (ECS)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8001_GIGABIT2	0x10198001 /*Marvell 88E8001 Gigabit Ethernet Controller (ECS)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8001_GIGABIT3	0x1043173c /*Marvell 88E8001 Gigabit Ethernet Controller (Asus)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8001_GIGABIT4	0x1043811a /*Marvell 88E8001 Gigabit Ethernet Controller (Asus)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8001_GIGABIT5	0x105b0c19 /*Marvell 88E8001 Gigabit Ethernet Controller (Foxconn)*/
	#define SUBDEVICE_MARVELL_SMC_EZ_CARD		0x10b8b452 /*SMC EZ Card 1000 (SMC9452TXV.2)*/
	#define SUBDEVICE_MARVELL_MARVELL_RDK_8001	0x11ab0121 /*Marvell RDK-8001*/
	#define SUBDEVICE_MARVELL_MARVELL_RDK_8003	0x11ab0321 /*Marvell RDK-8003*/
	#define SUBDEVICE_MARVELL_MARVELL_RDK_8010	0x11ab1021 /*Marvell RDK-8010*/
	#define SUBDEVICE_MARVELL_MARVELL_YUKON_GIGABIT	0x11ab5021 /*Marvell Yukon Gigabit Ethernet 10/100/1000Base-T Controller (64 bit)*/
	#define SUBDEVICE_MARVELL_MARVELL_YUKON_GIGABIT2	0x11ab9521 /*Marvell Yukon Gigabit Ethernet 10/100/1000Base-T Controller (32 bit)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8001_GIGABIT6	0x1458e000 /*Marvell 88E8001 Gigabit Ethernet Controller (Gigabyte)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8001_GIGABIT7	0x147b1406 /*Marvell 88E8001 Gigabit Ethernet Controller (Abit)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8001_GIGABIT8	0x15d40047 /*Marvell 88E8001 Gigabit Ethernet Controller (Iwill)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8001_GIGABIT9	0x16959025 /*Marvell 88E8001 Gigabit Ethernet Controller (Epox)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8001_GIGABIT10	0x17f21c03 /*Marvell 88E8001 Gigabit Ethernet Controller (Albatron)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8001_GIGABIT11	0x270f2803 /*Marvell 88E8001 Gigabit Ethernet Controller (Chaintech)*/
#define DEVICE_MARVELL_FAST_ETHERNET_CONTROLLER		0x4350 /*Fast Ethernet Controller*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8035_FAST	0x11790001 /*Marvell 88E8035 Fast Ethernet Controller (Toshiba)*/
	#define SUBDEVICE_MARVELL_MARVELL_RDK_8035	0x11ab3521 /*Marvell RDK-8035*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8035_FAST2	0x1854000d /*Marvell 88E8035 Fast Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8035_FAST3	0x1854000e /*Marvell 88E8035 Fast Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8035_FAST4	0x1854000f /*Marvell 88E8035 Fast Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8035_FAST5	0x18540011 /*Marvell 88E8035 Fast Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8035_FAST6	0x18540012 /*Marvell 88E8035 Fast Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8035_FAST7	0x18540016 /*Marvell 88E8035 Fast Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8035_FAST8	0x18540017 /*Marvell 88E8035 Fast Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8035_FAST9	0x18540018 /*Marvell 88E8035 Fast Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8035_FAST10	0x18540019 /*Marvell 88E8035 Fast Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8035_FAST11	0x1854001c /*Marvell 88E8035 Fast Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8035_FAST12	0x1854001e /*Marvell 88E8035 Fast Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8035_FAST13	0x18540020 /*Marvell 88E8035 Fast Ethernet Controller (LGE)*/
#define DEVICE_MARVELL_FAST_ETHERNET_CONTROLLER2	0x4351 /*Fast Ethernet Controller*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8036_FAST	0x107b4009 /*Marvell 88E8036 Fast Ethernet Controller (Wistron)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8036_FAST2	0x10f78338 /*Marvell 88E8036 Fast Ethernet Controller (Panasonic)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8036_FAST3	0x11790001 /*Marvell 88E8036 Fast Ethernet Controller (Toshiba)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8036_FAST4	0x1179ff00 /*Marvell 88E8036 Fast Ethernet Controller (Compal)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8036_FAST5	0x1179ff10 /*Marvell 88E8036 Fast Ethernet Controller (Inventec)*/
	#define SUBDEVICE_MARVELL_MARVELL_RDK_8036	0x11ab3621 /*Marvell RDK-8036*/
	#define SUBDEVICE_MARVELL_ABOCOM_EFE3K		0x13d1ac12 /*Abocom EFE3K - 10/100 Ethernet Expresscard*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8036_FAST6	0x161f203d /*Marvell 88E8036 Fast Ethernet Controller (Arima)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8036_FAST7	0x1854000d /*Marvell 88E8036 Fast Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8036_FAST8	0x1854000e /*Marvell 88E8036 Fast Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8036_FAST9	0x1854000f /*Marvell 88E8036 Fast Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8036_FAST10	0x18540011 /*Marvell 88E8036 Fast Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8036_FAST11	0x18540012 /*Marvell 88E8036 Fast Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8036_FAST12	0x18540016 /*Marvell 88E8036 Fast Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8036_FAST13	0x18540017 /*Marvell 88E8036 Fast Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8036_FAST14	0x18540018 /*Marvell 88E8036 Fast Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8036_FAST15	0x18540019 /*Marvell 88E8036 Fast Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8036_FAST16	0x1854001c /*Marvell 88E8036 Fast Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8036_FAST17	0x1854001e /*Marvell 88E8036 Fast Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8036_FAST18	0x18540020 /*Marvell 88E8036 Fast Ethernet Controller (LGE)*/
#define DEVICE_MARVELL_GIGABIT_ETHERNET_CONTROLLER2	0x4360 /*Gigabit Ethernet Controller*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8052_GIGABIT	0x10438134 /*Marvell 88E8052 Gigabit Ethernet Controller (Asus)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8052_GIGABIT2	0x107b4009 /*Marvell 88E8052 Gigabit Ethernet Controller (Wistron)*/
	#define SUBDEVICE_MARVELL_MARVELL_RDK_8052	0x11ab5221 /*Marvell RDK-8052*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8052_GIGABIT3	0x1458e000 /*Marvell 88E8052 Gigabit Ethernet Controller (Gigabyte)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8052_GIGABIT4	0x1462052c /*Marvell 88E8052 Gigabit Ethernet Controller (MSI)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8052_GIGABIT5	0x18498052 /*Marvell 88E8052 Gigabit Ethernet Controller (ASRock)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8052_GIGABIT6	0x1940e000 /*Marvell 88E8052 Gigabit Ethernet Controller (Gigabyte)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8052_GIGABIT7	0xa0a00509 /*Marvell 88E8052 Gigabit Ethernet Controller (Aopen)*/
#define DEVICE_MARVELL_GIGABIT_ETHERNET_CONTROLLER3	0x4361 /*Gigabit Ethernet Controller*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8050_GIGABIT	0x107b3015 /*Marvell 88E8050 Gigabit Ethernet Controller (Gateway)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8050_GIGABIT2	0x11ab5021 /*Marvell 88E8050 Gigabit Ethernet Controller (Intel)*/
	#define SUBDEVICE_MARVELL_D925XCVLK_MAINBOARD	0x80863063 /*D925XCVLK mainboard*/
#define DEVICE_MARVELL_GIGABIT_ETHERNET_CONTROLLER4	0x4362 /*Gigabit Ethernet Controller*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT	0x103c2a0d /*Marvell 88E8053 Gigabit Ethernet Controller (Asus)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT2	0x10438142 /*Marvell 88E8053 Gigabit Ethernet Controller (Asus)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT3	0x109f3197 /*Marvell 88E8053 Gigabit Ethernet Controller (Trigem)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT4	0x10f78338 /*Marvell 88E8053 Gigabit Ethernet Controller (Panasonic)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT5	0x10fda430 /*Marvell 88E8053 Gigabit Ethernet Controller (SOYO)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT6	0x11790001 /*Marvell 88E8053 Gigabit Ethernet Controller (Toshiba)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT7	0x1179ff00 /*Marvell 88E8053 Gigabit Ethernet Controller (Compal)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT8	0x1179ff10 /*Marvell 88E8053 Gigabit Ethernet Controller (Inventec)*/
	#define SUBDEVICE_MARVELL_MARVELL_RDK_8053	0x11ab5321 /*Marvell RDK-8053*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT9	0x1297c240 /*Marvell 88E8053 Gigabit Ethernet Controller (Shuttle)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT10	0x1297c241 /*Marvell 88E8053 Gigabit Ethernet Controller (Shuttle)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT11	0x1297c242 /*Marvell 88E8053 Gigabit Ethernet Controller (Shuttle)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT12	0x1297c243 /*Marvell 88E8053 Gigabit Ethernet Controller (Shuttle)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT13	0x1297c244 /*Marvell 88E8053 Gigabit Ethernet Controller (Shuttle)*/
	#define SUBDEVICE_MARVELL_ABOCOM_EGE5K		0x13d1ac11 /*Abocom EGE5K - Giga Ethernet Expresscard*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT14	0x1458e000 /*Marvell 88E8053 Gigabit Ethernet Controller (Gigabyte)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT15	0x1462058c /*Marvell 88E8053 Gigabit Ethernet Controller (MSI)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT16	0x14c00012 /*Marvell 88E8053 Gigabit Ethernet Controller (Compal)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT17	0x155804a0 /*Marvell 88E8053 Gigabit Ethernet Controller (Clevo)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT18	0x15bd1003 /*Marvell 88E8053 Gigabit Ethernet Controller (DFI)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT19	0x161f203c /*Marvell 88E8053 Gigabit Ethernet Controller (Arima)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT20	0x161f203d /*Marvell 88E8053 Gigabit Ethernet Controller (Arima)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT21	0x16959029 /*Marvell 88E8053 Gigabit Ethernet Controller (Epox)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT22	0x17f22c08 /*Marvell 88E8053 Gigabit Ethernet Controller (Albatron)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT23	0x17ff0585 /*Marvell 88E8053 Gigabit Ethernet Controller (Quanta)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT24	0x18498053 /*Marvell 88E8053 Gigabit Ethernet Controller (ASRock)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT25	0x1854000b /*Marvell 88E8053 Gigabit Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT26	0x1854000c /*Marvell 88E8053 Gigabit Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT27	0x18540010 /*Marvell 88E8053 Gigabit Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT28	0x18540013 /*Marvell 88E8053 Gigabit Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT29	0x18540014 /*Marvell 88E8053 Gigabit Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT30	0x18540015 /*Marvell 88E8053 Gigabit Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT31	0x1854001a /*Marvell 88E8053 Gigabit Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT32	0x1854001b /*Marvell 88E8053 Gigabit Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT33	0x1854001d /*Marvell 88E8053 Gigabit Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT34	0x1854001f /*Marvell 88E8053 Gigabit Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT35	0x18540021 /*Marvell 88E8053 Gigabit Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT36	0x18540022 /*Marvell 88E8053 Gigabit Ethernet Controller (LGE)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT37	0x1940e000 /*Marvell 88E8053 Gigabit Ethernet Controller (Gigabyte)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT38	0x270f2801 /*Marvell 88E8053 Gigabit Ethernet Controller (Chaintech)*/
	#define SUBDEVICE_MARVELL_MARVELL_88E8053_GIGABIT39	0xa0a00506 /*Marvell 88E8053 Gigabit Ethernet Controller (Aopen)*/
#define DEVICE_MARVELL_GT_64115_SYSTEM_CONTROLLER	0x4611 /*GT-64115 System Controller*/
#define DEVICE_MARVELL_GT_64120_64120A_64121A_SYSTEM_CONTROLLER	0x4620 /*GT-64120/64120A/64121A System Controller*/
#define DEVICE_MARVELL_GT_48001				0x4801 /*GT-48001*/
#define DEVICE_MARVELL_MV88SX5040_4_PORT_SATA		0x5040 /*MV88SX5040 4-port SATA I PCI-X Controller*/
#define DEVICE_MARVELL_MV88SX5041_4_PORT_SATA		0x5041 /*MV88SX5041 4-port SATA I PCI-X Controller*/
#define DEVICE_MARVELL_MV88SX5080_8_PORT_SATA		0x5080 /*MV88SX5080 8-port SATA I PCI-X Controller*/
#define DEVICE_MARVELL_MV88SX5081_8_PORT_SATA		0x5081 /*MV88SX5081 8-port SATA I PCI-X Controller*/
#define DEVICE_MARVELL_MV88SX6041_4_PORT_SATA		0x6041 /*MV88SX6041 4-port SATA II PCI-X Controller*/
#define DEVICE_MARVELL_MV88SX6081_8_PORT_SATA		0x6081 /*MV88SX6081 8-port SATA II PCI-X Controller*/
#define DEVICE_MARVELL_MV64360_64361_64362_SYSTEM_CONTROLLER	0x6460 /*MV64360/64361/64362 System Controller*/
#define DEVICE_MARVELL_GT_64010_PRIMARY_IMAGE		0xf003 /*GT-64010 Primary Image Piranha Image Generator*/
/* Vendor: 11ac: Canon Information Systems Research Aust. */
/* Vendor: 11ad: Lite-On Communications Inc */
#define DEVICE_LITE_ON_LNE100TX				0x0002 /*LNE100TX*/
	#define SUBDEVICE_LITE_ON_LNE100TX		0x11ad0002 /*LNE100TX*/
	#define SUBDEVICE_LITE_ON_LNE100TX2		0x11ad0003 /*LNE100TX*/
	#define SUBDEVICE_LITE_ON_LNE100TX3		0x11adf003 /*LNE100TX*/
	#define SUBDEVICE_LITE_ON_LNE100TX4		0x11adffff /*LNE100TX*/
	#define SUBDEVICE_LITE_ON_FA310TX		0x1385f004 /*FA310TX*/
#define DEVICE_LITE_ON_LNE100TX_LINKSYS_ETHERFAST	0xc115 /*LNE100TX [Linksys EtherFast 10/100]*/
	#define SUBDEVICE_LITE_ON_LNE100TX_VER_2_0	0x11adc001 /*LNE100TX [ver 2.0]*/
/* Vendor: 11ae: Aztech System Ltd */
/* Vendor: 11af: Avid Technology Inc. */
#define DEVICE_AVID__CINEMA				0x0001 /*[Cinema]*/
/* Vendor: 11b0: V3 Semiconductor Inc. */
#define DEVICE_V3_SEMICONDUCTOR_V300PSC			0x0002 /*V300PSC*/
#define DEVICE_V3_SEMICONDUCTOR_V292PBC_AM29030_40_BRIDGE	0x0292 /*V292PBC [Am29030/40 Bridge]*/
#define DEVICE_V3_SEMICONDUCTOR_V96XPBC			0x0960 /*V96xPBC*/
#define DEVICE_V3_SEMICONDUCTOR_V96DPC			0xc960 /*V96DPC*/
/* Vendor: 11b1: Apricot Computers */
/* Vendor: 11b2: Eastman Kodak */
/* Vendor: 11b3: Barr Systems Inc. */
/* Vendor: 11b4: Leitch Technology International */
/* Vendor: 11b5: Radstone Technology Plc */
/* Vendor: 11b6: United Video Corp */
/* Vendor: 11b7: Motorola */
/* Vendor: 11b8: XPoint Technologies, Inc */
#define DEVICE_XPOINT_QUAD_PEERMASTER			0x0001 /*Quad PeerMaster*/
/* Vendor: 11b9: Pathlight Technology Inc. */
#define DEVICE_PATHLIGHT_SSA_CONTROLLER			0xc0ed /*SSA Controller*/
/* Vendor: 11ba: Videotron Corp */
/* Vendor: 11bb: Pyramid Technology */
/* Vendor: 11bc: Network Peripherals Inc */
#define DEVICE_NETWORK2_NP_PCI				0x0001 /*NP-PCI*/
/* Vendor: 11bd: Pinnacle Systems Inc. */
/* Vendor: 11be: International Microcircuits Inc */
/* Vendor: 11bf: Astrodesign, Inc. */
/* Vendor: 11c0: Hewlett Packard */
/* Vendor: 11c1: Agere Systems (former Lucent Microelectronics) */
#define DEVICE_AGERE_56K_WINMODEM			0x0440 /*56k WinModem*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K		0x10338015 /*LT WinModem 56k Data+Fax+Voice+Dsvd*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K2	0x10338047 /*LT WinModem 56k Data+Fax+Voice+Dsvd*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K3	0x1033804f /*LT WinModem 56k Data+Fax+Voice+Dsvd*/
	#define SUBDEVICE_AGERE_LB_LT_MODEM		0x10cf102c /*LB LT Modem V.90 56k*/
	#define SUBDEVICE_AGERE_BIBLO_LT_MODEM		0x10cf104a /*BIBLO LT Modem 56k*/
	#define SUBDEVICE_AGERE_LB2_LT_MODEM		0x10cf105f /*LB2 LT Modem V.90 56k*/
	#define SUBDEVICE_AGERE_INTERNAL_V_90_MODEM	0x11790001 /*Internal V.90 Modem*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K4	0x11c10440 /*LT WinModem 56k Data+Fax+Voice+Dsvd*/
	#define SUBDEVICE_AGERE_MDP7800_U_MODEM		0x122d4101 /*MDP7800-U Modem*/
	#define SUBDEVICE_AGERE_MDP7800SP_U_MODEM	0x122d4102 /*MDP7800SP-U Modem*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K5	0x13e00040 /*LT WinModem 56k Data+Fax+Voice+Dsvd*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K6	0x13e00440 /*LT WinModem 56k Data+Fax+Voice+Dsvd*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K7	0x13e00441 /*LT WinModem 56k Data+Fax+Voice+Dsvd*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K8	0x13e00450 /*LT WinModem 56k Data+Fax+Voice+Dsvd*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K9	0x13e0f100 /*LT WinModem 56k Data+Fax+Voice+Dsvd*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K10	0x13e0f101 /*LT WinModem 56k Data+Fax+Voice+Dsvd*/
	#define SUBDEVICE_AGERE_LT56PV_MODEM		0x144d2101 /*LT56PV Modem*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K11	0x149f0440 /*LT WinModem 56k Data+Fax+Voice+Dsvd*/
#define DEVICE_AGERE_56K_WINMODEM2			0x0441 /*56k WinModem*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K12	0x1033804d /*LT WinModem 56k Data+Fax*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K13	0x10338065 /*LT WinModem 56k Data+Fax*/
	#define SUBDEVICE_AGERE_SUPRA_56I		0x10920440 /*Supra 56i*/
	#define SUBDEVICE_AGERE_INTERNAL_V_90_MODEM2	0x11790001 /*Internal V.90 Modem*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K14	0x11c10440 /*LT WinModem 56k Data+Fax*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K15	0x11c10441 /*LT WinModem 56k Data+Fax*/
	#define SUBDEVICE_AGERE_MDP7800_U_MODEM2	0x122d4100 /*MDP7800-U Modem*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K16	0x13e00040 /*LT WinModem 56k Data+Fax*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K17	0x13e00100 /*LT WinModem 56k Data+Fax*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K18	0x13e00410 /*LT WinModem 56k Data+Fax*/
	#define SUBDEVICE_AGERE_TELEPATH_INTERNET_56K	0x13e00420 /*TelePath Internet 56k WinModem*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K19	0x13e00440 /*LT WinModem 56k Data+Fax*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K20	0x13e00443 /*LT WinModem 56k Data+Fax*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K21	0x13e0f102 /*LT WinModem 56k Data+Fax*/
	#define SUBDEVICE_AGERE_COMMWAVE_56K_MODEM	0x14169804 /*CommWave 56k Modem*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K22	0x141d0440 /*LT WinModem 56k Data+Fax*/
	#define SUBDEVICE_AGERE_LUCENT_56K_V_90		0x144f0441 /*Lucent 56k V.90 DF Modem*/
	#define SUBDEVICE_AGERE_LUCENT_56K_V_902	0x144f0449 /*Lucent 56k V.90 DF Modem*/
	#define SUBDEVICE_AGERE_LUCENT_WIN_MODEM	0x144f110d /*Lucent Win Modem*/
	#define SUBDEVICE_AGERE_PRESARIO_56K_V_90	0x14680441 /*Presario 56k V.90 DF Modem*/
	#define SUBDEVICE_AGERE_LUCENT_WIN_MODEM2	0x16680440 /*Lucent Win Modem*/
#define DEVICE_AGERE_56K_WINMODEM3			0x0442 /*56k WinModem*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K23	0x11c10440 /*LT WinModem 56k Data+Fax+Voice+VoiceView+Dsvd*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K24	0x11c10442 /*LT WinModem 56k Data+Fax+Voice+VoiceView+Dsvd*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K25	0x13e00412 /*LT WinModem 56k Data+Fax+Voice+VoiceView+Dsvd*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K26	0x13e00442 /*LT WinModem 56k Data+Fax+Voice+VoiceView+Dsvd*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K27	0x13fc2471 /*LT WinModem 56k Data+Fax+Voice+VoiceView+Dsvd*/
	#define SUBDEVICE_AGERE_LT56PT_MODEM		0x144d2104 /*LT56PT Modem*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K28	0x144f1104 /*LT WinModem 56k Data+Fax+Voice+VoiceView+Dsvd*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K29	0x149f0440 /*LT WinModem 56k Data+Fax+Voice+VoiceView+Dsvd*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K30	0x16680440 /*LT WinModem 56k Data+Fax+Voice+VoiceView+Dsvd*/
#define DEVICE_AGERE_LT_WINMODEM			0x0443 /*LT WinModem*/
#define DEVICE_AGERE_LT_WINMODEM2			0x0444 /*LT WinModem*/
#define DEVICE_AGERE_LT_WINMODEM3			0x0445 /*LT WinModem*/
	#define SUBDEVICE_AGERE_PRO_100_MINIPCI_PROBABLY	0x80862203 /*PRO/100+ MiniPCI (probably an Ambit U98.003.C.00 combo card)*/
	#define SUBDEVICE_AGERE_PRO_100_MINIPCI_ON	0x80862204 /*PRO/100+ MiniPCI on Armada E500*/
#define DEVICE_AGERE_LT_WINMODEM4			0x0446 /*LT WinModem*/
#define DEVICE_AGERE_LT_WINMODEM5			0x0447 /*LT WinModem*/
#define DEVICE_AGERE_WINMODEM_56K			0x0448 /*WinModem 56k*/
	#define SUBDEVICE_AGERE_LUCENT_WIN_MODEM3	0x10140131 /*Lucent Win Modem*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K31	0x10338066 /*LT WinModem 56k Data+Fax+Voice+Dsvd*/
	#define SUBDEVICE_AGERE_56K_VOICE_MODEM		0x13e00030 /*56k Voice Modem*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K32	0x13e00040 /*LT WinModem 56k Data+Fax+Voice+Dsvd*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K33	0x16682400 /*LT WinModem 56k (MiniPCI Ethernet+Modem)*/
#define DEVICE_AGERE_WINMODEM_56K2			0x0449 /*WinModem 56k*/
	#define SUBDEVICE_AGERE_56K_V_90_MODEM		0x0e11b14d /*56k V.90 Modem*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K34	0x13e00020 /*LT WinModem 56k Data+Fax*/
	#define SUBDEVICE_AGERE_TELEPATH_INTERNET_56K2	0x13e00041 /*TelePath Internet 56k WinModem*/
	#define SUBDEVICE_AGERE_LUCENT_WIN_MODEM4	0x14360440 /*Lucent Win Modem*/
	#define SUBDEVICE_AGERE_LUCENT_56K_V_903	0x144f0449 /*Lucent 56k V.90 DFi Modem*/
	#define SUBDEVICE_AGERE_IBM_THINKPAD_T23	0x14680410 /*IBM ThinkPad T23 (2647-4MG)*/
	#define SUBDEVICE_AGERE_LUCENT_WIN_MODEM5	0x14680440 /*Lucent Win Modem*/
	#define SUBDEVICE_AGERE_PRESARIO_56K_V_902	0x14680449 /*Presario 56k V.90 DFi Modem*/
#define DEVICE_AGERE_F_1156IV_WINMODEM_V90		0x044a /*F-1156IV WinModem (V90, 56KFlex)*/
	#define SUBDEVICE_AGERE_LB_GLOBAL_LT		0x10cf1072 /*LB Global LT Modem*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K35	0x13e00012 /*LT WinModem 56k Data+Fax+Voice+VoiceView+Dsvd*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K36	0x13e00042 /*LT WinModem 56k Data+Fax+Voice+VoiceView+Dsvd*/
	#define SUBDEVICE_AGERE_LT_WINMODEM_56K37	0x144f1005 /*LT WinModem 56k Data+Fax+Voice+VoiceView+Dsvd*/
#define DEVICE_AGERE_LT_WINMODEM6			0x044b /*LT WinModem*/
#define DEVICE_AGERE_LT_WINMODEM7			0x044c /*LT WinModem*/
#define DEVICE_AGERE_LT_WINMODEM8			0x044d /*LT WinModem*/
#define DEVICE_AGERE_LT_WINMODEM9			0x044e /*LT WinModem*/
#define DEVICE_AGERE_V90_WILDWIRE_MODEM			0x044f /*V90 WildWire Modem*/
#define DEVICE_AGERE_LT_WINMODEM10			0x0450 /*LT WinModem*/
	#define SUBDEVICE_AGERE_VERSA_NOTE_VXI		0x103380a8 /*Versa Note Vxi*/
	#define SUBDEVICE_AGERE_MAGNIA_SG20		0x144f4005 /*Magnia SG20*/
#define DEVICE_AGERE_LT_WINMODEM11			0x0451 /*LT WinModem*/
#define DEVICE_AGERE_LT_WINMODEM12			0x0452 /*LT WinModem*/
#define DEVICE_AGERE_LT_WINMODEM13			0x0453 /*LT WinModem*/
#define DEVICE_AGERE_LT_WINMODEM14			0x0454 /*LT WinModem*/
#define DEVICE_AGERE_LT_WINMODEM15			0x0455 /*LT WinModem*/
#define DEVICE_AGERE_LT_WINMODEM16			0x0456 /*LT WinModem*/
#define DEVICE_AGERE_LT_WINMODEM17			0x0457 /*LT WinModem*/
#define DEVICE_AGERE_LT_WINMODEM18			0x0458 /*LT WinModem*/
#define DEVICE_AGERE_LT_WINMODEM19			0x0459 /*LT WinModem*/
#define DEVICE_AGERE_LT_WINMODEM20			0x045a /*LT WinModem*/
#define DEVICE_AGERE_LT_WINMODEM21			0x045c /*LT WinModem*/
#define DEVICE_AGERE_V90_WILDWIRE_MODEM2		0x0461 /*V90 WildWire Modem*/
#define DEVICE_AGERE_V90_WILDWIRE_MODEM3		0x0462 /*V90 WildWire Modem*/
#define DEVICE_AGERE_VENUS_MODEM_V90			0x0480 /*Venus Modem (V90, 56KFlex)*/
#define DEVICE_AGERE_V_92_56K_WINMODEM			0x048c /*V.92 56K WinModem*/
#define DEVICE_AGERE_V_92_56K_WINMODEM2			0x048f /*V.92 56k WinModem*/
#define DEVICE_AGERE_USB					0x5801 /*USB*/
#define DEVICE_AGERE_USS_312_USB_CONTROLLER		0x5802 /*USS-312 USB Controller*/
#define DEVICE_AGERE_USS_344S_USB_CONTROLLER		0x5803 /*USS-344S USB Controller*/
#define DEVICE_AGERE_FW323				0x5811 /*FW323*/
	#define SUBDEVICE_AGERE_D865PERL_MAINBOARD	0x8086524c /*D865PERL mainboard*/
	#define SUBDEVICE_AGERE_FIREWIRE_HOST_BUS	0xdead0800 /*FireWire Host Bus Adapter*/
#define DEVICE_AGERE_WL60010_WIRELESS_LAN		0xab10 /*WL60010 Wireless LAN MAC*/
#define DEVICE_AGERE_WL60040_MULTIMODE_WIRELES		0xab11 /*WL60040 Multimode Wireles LAN MAC*/
	#define SUBDEVICE_AGERE_WAVELAN_11ABG_CARDBUS	0x11c1ab12 /*WaveLAN 11abg Cardbus card (Model 1102)*/
	#define SUBDEVICE_AGERE_WAVELAN_11ABG_MINIPCI	0x11c1ab13 /*WaveLAN 11abg MiniPCI card (Model 0512)*/
	#define SUBDEVICE_AGERE_WAVELAN_11ABG_CARDBUS2	0x11c1ab15 /*WaveLAN 11abg Cardbus card (Model 1106)*/
	#define SUBDEVICE_AGERE_WAVELAN_11ABG_MINIPCI2	0x11c1ab16 /*WaveLAN 11abg MiniPCI card (Model 0516)*/
#define DEVICE_AGERE_ORINOCO_PCI_ADAPTER		0xab20 /*ORiNOCO PCI Adapter*/
#define DEVICE_AGERE_AGERE_WIRELESS_PCI			0xab21 /*Agere Wireless PCI Adapter*/
#define DEVICE_AGERE_HERMES2_MINI_PCI_WAVELAN		0xab30 /*Hermes2 Mini-PCI WaveLAN a/b/g*/
	#define SUBDEVICE_AGERE_HERMES2_MINI_PCI_WAVELAN	0x14cd2012 /*Hermes2 Mini-PCI WaveLAN a/b/g*/
/* Vendor: 11c2: Sand Microelectronics */
/* Vendor: 11c3: NEC Corporation */
/* Vendor: 11c4: Document Technologies, Inc */
/* Vendor: 11c5: Shiva Corporation */
/* Vendor: 11c6: Dainippon Screen Mfg. Co. Ltd */
/* Vendor: 11c7: D.C.M. Data Systems */
/* Vendor: 11c8: Dolphin Interconnect Solutions AS */
#define DEVICE_DOLPHIN_PSB32_SCI_ADAPTER_D31X		0x0658 /*PSB32 SCI-Adapter D31x*/
#define DEVICE_DOLPHIN_PSB64_SCI_ADAPTER_D32X		0xd665 /*PSB64 SCI-Adapter D32x*/
#define DEVICE_DOLPHIN_PSB66_SCI_ADAPTER_D33X		0xd667 /*PSB66 SCI-Adapter D33x*/
/* Vendor: 11c9: Magma */
#define DEVICE_MAGMA_16_LINE_SERIAL_PORT		0x0010 /*16-line serial port w/- DMA*/
#define DEVICE_MAGMA_4_LINE_SERIAL_PORT			0x0011 /*4-line serial port w/- DMA*/
/* Vendor: 11ca: LSI Systems, Inc */
/* Vendor: 11cb: Specialix Research Ltd. */
#define DEVICE_SPECIALIX_PCI_9050			0x2000 /*PCI_9050*/
	#define SUBDEVICE_SPECIALIX_SX			0x11cb0200 /*SX*/
	#define SUBDEVICE_SPECIALIX_I_O8		0x11cbb008 /*I/O8+*/
#define DEVICE_SPECIALIX_SUPI_1				0x4000 /*SUPI_1*/
#define DEVICE_SPECIALIX_T225				0x8000 /*T225*/
/* Vendor: 11cc: Michels & Kleberhoff Computer GmbH */
/* Vendor: 11cd: HAL Computer Systems, Inc. */
/* Vendor: 11ce: Netaccess */
/* Vendor: 11cf: Pioneer Electronic Corporation */
/* Vendor: 11d0: Lockheed Martin Federal Systems-Manassas */
/* Vendor: 11d1: Auravision */
#define DEVICE_AURAVISION_VXP524			0x01f7 /*VxP524*/
/* Vendor: 11d2: Intercom Inc. */
/* Vendor: 11d3: Trancell Systems Inc */
/* Vendor: 11d4: Analog Devices */
#define DEVICE_ANALOG_BLACKFIN_BF535_PROCESSOR		0x1535 /*Blackfin BF535 processor*/
#define DEVICE_ANALOG_SM56_PCI_MODEM			0x1805 /*SM56 PCI modem*/
#define DEVICE_ANALOG_AD1889_SOUND_CHIP			0x1889 /*AD1889 sound chip*/
/* Vendor: 11d5: Ikon Corporation */
#define DEVICE_IKON_10115				0x0115 /*10115*/
#define DEVICE_IKON_10117				0x0117 /*10117*/
/* Vendor: 11d6: Tekelec Telecom */
/* Vendor: 11d7: Trenton Technology, Inc. */
/* Vendor: 11d8: Image Technologies Development */
/* Vendor: 11d9: TEC Corporation */
/* Vendor: 11da: Novell */
/* Vendor: 11db: Sega Enterprises Ltd */
/* Vendor: 11dc: Questra Corporation */
/* Vendor: 11dd: Crosfield Electronics Limited */
/* Vendor: 11de: Zoran Corporation */
#define DEVICE_ZORAN_ZR36057PQC_VIDEO_CUTTING		0x6057 /*ZR36057PQC Video cutting chipset*/
	#define SUBDEVICE_ZORAN_DC10_PLUS		0x10317efe /*DC10 Plus*/
	#define SUBDEVICE_ZORAN_MIROVIDEO_DC50_MOTION	0x1031fc00 /*MiroVIDEO DC50, Motion JPEG Capture/CODEC Board*/
	#define SUBDEVICE_ZORAN_JPEG_TV_CARD		0x13ca4231 /*JPEG/TV Card*/
#define DEVICE_ZORAN_ZR36120				0x6120 /*ZR36120*/
	#define SUBDEVICE_ZORAN_CINEMASTER_C_DVD	0x1328f001 /*Cinemaster C DVD Decoder*/
/* Vendor: 11df: New Wave PDG */
/* Vendor: 11e0: Cray Communications A/S */
/* Vendor: 11e1: GEC Plessey Semi Inc. */
/* Vendor: 11e2: Samsung Information Systems America */
/* Vendor: 11e3: Quicklogic Corporation */
#define DEVICE_QUICKLOGIC_PC_WATCHDOG			0x5030 /*PC Watchdog*/
/* Vendor: 11e4: Second Wave Inc */
/* Vendor: 11e5: IIX Consulting */
/* Vendor: 11e6: Mitsui-Zosen System Research */
/* Vendor: 11e7: Toshiba America, Elec. Company */
/* Vendor: 11e8: Digital Processing Systems Inc. */
/* Vendor: 11e9: Highwater Designs Ltd. */
/* Vendor: 11ea: Elsag Bailey */
/* Vendor: 11eb: Formation Inc. */
/* Vendor: 11ec: Coreco Inc */
/* Vendor: 11ed: Mediamatics */
/* Vendor: 11ee: Dome Imaging Systems Inc */
/* Vendor: 11ef: Nicolet Technologies B.V. */
/* Vendor: 11f0: Compu-Shack */
#define DEVICE_COMPU_SHACK_FDDI				0x4231 /*FDDI*/
#define DEVICE_COMPU_SHACK_FASTLINE_UTP_QUATTRO		0x4232 /*FASTline UTP Quattro*/
#define DEVICE_COMPU_SHACK_FASTLINE_FO			0x4233 /*FASTline FO*/
#define DEVICE_COMPU_SHACK_FASTLINE_UTP			0x4234 /*FASTline UTP*/
#define DEVICE_COMPU_SHACK_FASTLINE_II_UTP		0x4235 /*FASTline-II UTP*/
#define DEVICE_COMPU_SHACK_FASTLINE_II_FO		0x4236 /*FASTline-II FO*/
#define DEVICE_COMPU_SHACK_GIGALINE			0x4731 /*GIGAline*/
/* Vendor: 11f1: Symbios Logic Inc */
/* Vendor: 11f2: Picture Tel Japan K.K. */
/* Vendor: 11f3: Keithley Metrabyte */
/* Vendor: 11f4: Kinetic Systems Corporation */
#define DEVICE_KINETIC_CAMAC_CONTROLLER			0x2915 /*CAMAC controller*/
/* Vendor: 11f5: Computing Devices International */
/* Vendor: 11f6: Compex */
#define DEVICE_COMPEX_ENET100VG4			0x0112 /*ENet100VG4*/
#define DEVICE_COMPEX_FREEDOMLINE_100			0x0113 /*FreedomLine 100*/
#define DEVICE_COMPEX_READYLINK_2000			0x1401 /*ReadyLink 2000*/
#define DEVICE_COMPEX_RL100_ATX_10_100			0x2011 /*RL100-ATX 10/100*/
	#define SUBDEVICE_COMPEX_RL100_ATX		0x11f62011 /*RL100-ATX*/
#define DEVICE_COMPEX_READYLINK_100TX_WINBOND		0x2201 /*ReadyLink 100TX (Winbond W89C840)*/
	#define SUBDEVICE_COMPEX_READYLINK_100TX	0x11f62011 /*ReadyLink 100TX*/
#define DEVICE_COMPEX_RL100TX_FAST_ETHERNET		0x9881 /*RL100TX Fast Ethernet*/
/* Vendor: 11f7: Scientific Atlanta */
/* Vendor: 11f8: PMC-Sierra Inc. */
#define DEVICE_PMC_SIERRA_PM7375_LASAR_155_ATM		0x7375 /*PM7375 [LASAR-155 ATM SAR]*/
/* Vendor: 11f9: I-Cube Inc */
/* Vendor: 11fa: Kasan Electronics Company, Ltd. */
/* Vendor: 11fb: Datel Inc */
/* Vendor: 11fc: Silicon Magic */
/* Vendor: 11fd: High Street Consultants */
/* Vendor: 11fe: Comtrol Corporation */
#define DEVICE_COMTROL_ROCKETPORT_32_PORT		0x0001 /*RocketPort 32 port w/external I/F*/
#define DEVICE_COMTROL_ROCKETPORT_8_PORT		0x0002 /*RocketPort 8 port w/external I/F*/
#define DEVICE_COMTROL_ROCKETPORT_16_PORT		0x0003 /*RocketPort 16 port w/external I/F*/
#define DEVICE_COMTROL_ROCKETPORT_4_PORT		0x0004 /*RocketPort 4 port w/quad cable*/
#define DEVICE_COMTROL_ROCKETPORT_8_PORT2		0x0005 /*RocketPort 8 port w/octa cable*/
#define DEVICE_COMTROL_ROCKETPORT_8_PORT3		0x0006 /*RocketPort 8 port w/RJ11 connectors*/
#define DEVICE_COMTROL_ROCKETPORT_4_PORT2		0x0007 /*RocketPort 4 port w/RJ11 connectors*/
#define DEVICE_COMTROL_ROCKETPORT_8_PORT4		0x0008 /*RocketPort 8 port w/ DB78 SNI (Siemens) connector*/
#define DEVICE_COMTROL_ROCKETPORT_16_PORT2		0x0009 /*RocketPort 16 port w/ DB78 SNI (Siemens) connector*/
#define DEVICE_COMTROL_ROCKETPORT_PLUS_4		0x000a /*RocketPort Plus 4 port*/
#define DEVICE_COMTROL_ROCKETPORT_PLUS_8		0x000b /*RocketPort Plus 8 port*/
#define DEVICE_COMTROL_ROCKETMODEM_6_PORT		0x000c /*RocketModem 6 port*/
#define DEVICE_COMTROL_ROCKETMODEM_4_PORT		0x000d /*RocketModem 4-port*/
#define DEVICE_COMTROL_ROCKETPORT_PLUS_2		0x000e /*RocketPort Plus 2 port RS232*/
#define DEVICE_COMTROL_ROCKETPORT_PLUS_22		0x000f /*RocketPort Plus 2 port RS422*/
#define DEVICE_COMTROL_ROCKETPORT_UPCI_32		0x0801 /*RocketPort UPCI 32 port w/external I/F*/
#define DEVICE_COMTROL_ROCKETPORT_UPCI_8		0x0802 /*RocketPort UPCI 8 port w/external I/F*/
#define DEVICE_COMTROL_ROCKETPORT_UPCI_16		0x0803 /*RocketPort UPCI 16 port w/external I/F*/
#define DEVICE_COMTROL_ROCKETPORT_UPCI_82		0x0805 /*RocketPort UPCI 8 port w/octa cable*/
#define DEVICE_COMTROL_ROCKETMODEM_III_8		0x080c /*RocketModem III 8 port*/
#define DEVICE_COMTROL_ROCKETMODEM_III_4		0x080d /*RocketModem III 4 port*/
#define DEVICE_COMTROL_ROCKETPORT_COMPACT_PCI		0x0903 /*RocketPort Compact PCI 16 port w/external I/F*/
#define DEVICE_COMTROL_ROCKETPORT_4_PORT_UART		0x8015 /*RocketPort 4-port UART 16954*/
/* Vendor: 11ff: Scion Corporation */
#define DEVICE_SCION_AG_5				0x0003 /*AG-5*/
/* Vendor: 1200: CSS Corporation */
/* Vendor: 1201: Vista Controls Corp */
/* Vendor: 1202: Network General Corp. */
#define DEVICE_NETWORK3_GIGABIT_ETHERNET_ADAPTER	0x4300 /*Gigabit Ethernet Adapter*/
	#define SUBDEVICE_NETWORK3_SK_9841_LX		0x12029841 /*SK-9841 LX*/
	#define SUBDEVICE_NETWORK3_SK_9841_LX_DUAL	0x12029842 /*SK-9841 LX dual link*/
	#define SUBDEVICE_NETWORK3_SK_9843_SX		0x12029843 /*SK-9843 SX*/
	#define SUBDEVICE_NETWORK3_SK_9843_SX_DUAL	0x12029844 /*SK-9843 SX dual link*/
/* Vendor: 1203: Bayer Corporation, Agfa Division */
/* Vendor: 1204: Lattice Semiconductor Corporation */
/* Vendor: 1205: Array Corporation */
/* Vendor: 1206: Amdahl Corporation */
/* Vendor: 1208: Parsytec GmbH */
#define DEVICE_PARSYTEC_HS_LINK_DEVICE			0x4853 /*HS-Link Device*/
/* Vendor: 1209: SCI Systems Inc */
/* Vendor: 120a: Synaptel */
/* Vendor: 120b: Adaptive Solutions */
/* Vendor: 120c: Technical Corp. */
/* Vendor: 120d: Compression Labs, Inc. */
/* Vendor: 120e: Cyclades Corporation */
#define DEVICE_CYCLADES_CYCLOM_Y_BELOW_FIRST		0x0100 /*Cyclom-Y below first megabyte*/
#define DEVICE_CYCLADES_CYCLOM_Y_ABOVE_FIRST		0x0101 /*Cyclom-Y above first megabyte*/
#define DEVICE_CYCLADES_CYCLOM_4Y_BELOW_FIRST		0x0102 /*Cyclom-4Y below first megabyte*/
#define DEVICE_CYCLADES_CYCLOM_4Y_ABOVE_FIRST		0x0103 /*Cyclom-4Y above first megabyte*/
#define DEVICE_CYCLADES_CYCLOM_8Y_BELOW_FIRST		0x0104 /*Cyclom-8Y below first megabyte*/
#define DEVICE_CYCLADES_CYCLOM_8Y_ABOVE_FIRST		0x0105 /*Cyclom-8Y above first megabyte*/
#define DEVICE_CYCLADES_CYCLADES_Z_BELOW_FIRST		0x0200 /*Cyclades-Z below first megabyte*/
#define DEVICE_CYCLADES_CYCLADES_Z_ABOVE_FIRST		0x0201 /*Cyclades-Z above first megabyte*/
#define DEVICE_CYCLADES_PC300_RSV_OR_X21		0x0300 /*PC300/RSV or /X21 (2 ports)*/
#define DEVICE_CYCLADES_PC300_RSV_OR_X212		0x0301 /*PC300/RSV or /X21 (1 port)*/
#define DEVICE_CYCLADES_PC300_TE_2_PORTS		0x0310 /*PC300/TE (2 ports)*/
#define DEVICE_CYCLADES_PC300_TE_1_PORT			0x0311 /*PC300/TE (1 port)*/
#define DEVICE_CYCLADES_PC300_TE_M_2_PORTS		0x0320 /*PC300/TE-M (2 ports)*/
#define DEVICE_CYCLADES_PC300_TE_M_1_PORT		0x0321 /*PC300/TE-M (1 port)*/
#define DEVICE_CYCLADES_PC400				0x0400 /*PC400*/
/* Vendor: 120f: Essential Communications */
#define DEVICE_ESSENTIAL_ROADRUNNER_SERIAL_HIPPI	0x0001 /*Roadrunner serial HIPPI*/
/* Vendor: 1210: Hyperparallel Technologies */
/* Vendor: 1211: Braintech Inc */
/* Vendor: 1212: Kingston Technology Corp. */
/* Vendor: 1213: Applied Intelligent Systems, Inc. */
/* Vendor: 1214: Performance Technologies, Inc. */
/* Vendor: 1215: Interware Co., Ltd */
/* Vendor: 1216: Purup Prepress A/S */
/* Vendor: 1217: O2 Micro, Inc. */
#define DEVICE_O2_MICRO_OZ6729				0x6729 /*OZ6729*/
#define DEVICE_O2_MICRO_OZ6730				0x673a /*OZ6730*/
#define DEVICE_O2_MICRO_OZ6832_6833_CARDBUS_CONTROLLER	0x6832 /*OZ6832/6833 Cardbus Controller*/
#define DEVICE_O2_MICRO_OZ6836_6860_CARDBUS_CONTROLLER	0x6836 /*OZ6836/6860 Cardbus Controller*/
#define DEVICE_O2_MICRO_OZ6812_CARDBUS_CONTROLLER	0x6872 /*OZ6812 Cardbus Controller*/
#define DEVICE_O2_MICRO_OZ6922_CARDBUS_CONTROLLER	0x6925 /*OZ6922 Cardbus Controller*/
#define DEVICE_O2_MICRO_OZ6933_CARDBUS_CONTROLLER	0x6933 /*OZ6933 Cardbus Controller*/
	#define SUBDEVICE_O2_MICRO_TRAVELMATE_612_TX	0x10251016 /*Travelmate 612 TX*/
#define DEVICE_O2_MICRO_OZ6912_CARDBUS_CONTROLLER	0x6972 /*OZ6912 Cardbus Controller*/
	#define SUBDEVICE_O2_MICRO_THINKPAD_R30		0x1014020c /*ThinkPad R30*/
	#define SUBDEVICE_O2_MICRO_MAGNIA_Z310		0x11790001 /*Magnia Z310*/
#define DEVICE_O2_MICRO_OZ711MX_MULTIMEDIABAY_ACCELERATOR	0x7110 /*OZ711Mx MultiMediaBay Accelerator*/
	#define SUBDEVICE_O2_MICRO_NC6000_LAPTOP	0x103c0890 /*NC6000 laptop*/
#define DEVICE_O2_MICRO_OZ711EC1_M1_SMARTCARDBUS_MULTIMEDIABAY	0x7112 /*OZ711EC1/M1 SmartCardBus MultiMediaBay Controller*/
#define DEVICE_O2_MICRO_OZ711EC1_SMARTCARDBUS_CONTROLLER	0x7113 /*OZ711EC1 SmartCardBus Controller*/
#define DEVICE_O2_MICRO_OZ711M1_SMARTCARDBUS_MULTIMEDIABAY	0x7114 /*OZ711M1 SmartCardBus MultiMediaBay Controller*/
#define DEVICE_O2_MICRO_OZ711E2_SMARTCARDBUS_CONTROLLER	0x71e2 /*OZ711E2 SmartCardBus Controller*/
#define DEVICE_O2_MICRO_OZ711M2_SMARTCARDBUS_MULTIMEDIABAY	0x7212 /*OZ711M2 SmartCardBus MultiMediaBay Controller*/
#define DEVICE_O2_MICRO_OZ6933E_CARDBUS_CONTROLLER	0x7213 /*OZ6933E CardBus Controller*/
#define DEVICE_O2_MICRO_OZ711M3_SMARTCARDBUS_MULTIMEDIABAY	0x7223 /*OZ711M3 SmartCardBus MultiMediaBay Controller*/
	#define SUBDEVICE_O2_MICRO_NC6000_LAPTOP2	0x103c0890 /*NC6000 laptop*/
/* Vendor: 1218: Hybricon Corp. */
/* Vendor: 1219: First Virtual Corporation */
/* Vendor: 121a: 3Dfx Interactive, Inc. */
#define DEVICE_3DFX_VOODOO				0x0001 /*Voodoo*/
#define DEVICE_3DFX_VOODOO_2				0x0002 /*Voodoo 2*/
#define DEVICE_3DFX_VOODOO_BANSHEE			0x0003 /*Voodoo Banshee*/
	#define SUBDEVICE_3DFX_MONSTER_FUSION		0x10920003 /*Monster Fusion*/
	#define SUBDEVICE_3DFX_MONSTER_FUSION2		0x10924000 /*Monster Fusion*/
	#define SUBDEVICE_3DFX_MONSTER_FUSION3		0x10924002 /*Monster Fusion*/
	#define SUBDEVICE_3DFX_MONSTER_FUSION_AGP	0x10924801 /*Monster Fusion AGP*/
	#define SUBDEVICE_3DFX_MONSTER_FUSION_AGP2	0x10924803 /*Monster Fusion AGP*/
	#define SUBDEVICE_3DFX_MONSTER_FUSION4		0x10928030 /*Monster Fusion*/
	#define SUBDEVICE_3DFX_MONSTER_FUSION_AGP3	0x10928035 /*Monster Fusion AGP*/
	#define SUBDEVICE_3DFX_DRAGON_4000		0x10b00001 /*Dragon 4000*/
	#define SUBDEVICE_3DFX_3D_BLASTER_BANSHEE	0x11021018 /*3D Blaster Banshee VE*/
	#define SUBDEVICE_3DFX_VOODOO_BANSHEE_AGP	0x121a0001 /*Voodoo Banshee AGP*/
	#define SUBDEVICE_3DFX_VOODOO_BANSHEE_AGP2	0x121a0003 /*Voodoo Banshee AGP SGRAM*/
	#define SUBDEVICE_3DFX_VOODOO_BANSHEE		0x121a0004 /*Voodoo Banshee*/
	#define SUBDEVICE_3DFX_RAVEN			0x139c0016 /*Raven*/
	#define SUBDEVICE_3DFX_RAVEN2			0x139c0017 /*Raven*/
	#define SUBDEVICE_3DFX_MAXI_GAMER_PHOENIX	0x14af0002 /*Maxi Gamer Phoenix*/
#define DEVICE_3DFX_VOODOO_BANSHEE_VELOCITY		0x0004 /*Voodoo Banshee [Velocity 100]*/
#define DEVICE_3DFX_VOODOO_3				0x0005 /*Voodoo 3*/
	#define SUBDEVICE_3DFX_VOODOO3_AGP		0x121a0004 /*Voodoo3 AGP*/
	#define SUBDEVICE_3DFX_VOODOO3_AGP2		0x121a0030 /*Voodoo3 AGP*/
	#define SUBDEVICE_3DFX_VOODOO3_AGP3		0x121a0031 /*Voodoo3 AGP*/
	#define SUBDEVICE_3DFX_VOODOO3_AGP4		0x121a0034 /*Voodoo3 AGP*/
	#define SUBDEVICE_3DFX_VOODOO3_2000_PCI		0x121a0036 /*Voodoo3 2000 PCI*/
	#define SUBDEVICE_3DFX_VOODOO3_AGP5		0x121a0037 /*Voodoo3 AGP*/
	#define SUBDEVICE_3DFX_VOODOO3_AGP6		0x121a0038 /*Voodoo3 AGP*/
	#define SUBDEVICE_3DFX_VOODOO3_AGP7		0x121a003a /*Voodoo3 AGP*/
	#define SUBDEVICE_3DFX_VOODOO3			0x121a0044 /*Voodoo3*/
	#define SUBDEVICE_3DFX_VELOCITY_100		0x121a004b /*Velocity 100*/
	#define SUBDEVICE_3DFX_VELOCITY_200		0x121a004c /*Velocity 200*/
	#define SUBDEVICE_3DFX_VOODOO3_AGP8		0x121a004d /*Voodoo3 AGP*/
	#define SUBDEVICE_3DFX_VOODOO3_AGP9		0x121a004e /*Voodoo3 AGP*/
	#define SUBDEVICE_3DFX_VOODOO3_AGP10		0x121a0051 /*Voodoo3 AGP*/
	#define SUBDEVICE_3DFX_VOODOO3_AGP11		0x121a0052 /*Voodoo3 AGP*/
	#define SUBDEVICE_3DFX_VOODOO3_3500_TV		0x121a0060 /*Voodoo3 3500 TV (NTSC)*/
	#define SUBDEVICE_3DFX_VOODOO3_3500_TV2		0x121a0061 /*Voodoo3 3500 TV (PAL)*/
	#define SUBDEVICE_3DFX_VOODOO3_3500_TV3		0x121a0062 /*Voodoo3 3500 TV (SECAM)*/
#define DEVICE_3DFX_VOODOO_4				0x0009 /*Voodoo 4 / Voodoo 5*/
	#define SUBDEVICE_3DFX_VOODOO5_PCI_5500		0x121a0003 /*Voodoo5 PCI 5500*/
	#define SUBDEVICE_3DFX_VOODOO5_AGP_5500_6000	0x121a0009 /*Voodoo5 AGP 5500/6000*/
#define DEVICE_3DFX_VOODOO_3_3000_AVENGER		0x0057 /*Voodoo 3/3000 [Avenger]*/
/* Vendor: 121b: Advanced Telecommunications Modules */
/* Vendor: 121c: Nippon Texaco., Ltd */
/* Vendor: 121d: Lippert Automationstechnik GmbH */
/* Vendor: 121e: CSPI */
/* Vendor: 121f: Arcus Technology, Inc. */
/* Vendor: 1220: Ariel Corporation */
#define DEVICE_ARIEL_AMCC_5933_TMS320C80		0x1220 /*AMCC 5933 TMS320C80 DSP/Imaging board*/
/* Vendor: 1221: Contec Co., Ltd */
/* Vendor: 1222: Ancor Communications, Inc. */
/* Vendor: 1223: Artesyn Communication Products */
#define DEVICE_ARTESYN_PM_LINK				0x0003 /*PM/Link*/
#define DEVICE_ARTESYN_PM_T1				0x0004 /*PM/T1*/
#define DEVICE_ARTESYN_PM_E1				0x0005 /*PM/E1*/
#define DEVICE_ARTESYN_PM_SLS				0x0008 /*PM/SLS*/
#define DEVICE_ARTESYN_BAJASPAN_RESOURCE_TARGET		0x0009 /*BajaSpan Resource Target*/
#define DEVICE_ARTESYN_BAJASPAN_SECTION_0		0x000a /*BajaSpan Section 0*/
#define DEVICE_ARTESYN_BAJASPAN_SECTION_1		0x000b /*BajaSpan Section 1*/
#define DEVICE_ARTESYN_BAJASPAN_SECTION_2		0x000c /*BajaSpan Section 2*/
#define DEVICE_ARTESYN_BAJASPAN_SECTION_3		0x000d /*BajaSpan Section 3*/
#define DEVICE_ARTESYN_PM_PPC				0x000e /*PM/PPC*/
/* Vendor: 1224: Interactive Images */
/* Vendor: 1225: Power I/O, Inc. */
/* Vendor: 1227: Tech-Source */
#define DEVICE_TECH_SOURCE_RAPTOR_GFX_8P		0x0006 /*Raptor GFX 8P*/
/* Vendor: 1228: Norsk Elektro Optikk A/S */
/* Vendor: 1229: Data Kinesis Inc. */
/* Vendor: 122a: Integrated Telecom */
/* Vendor: 122b: LG Industrial Systems Co., Ltd */
/* Vendor: 122c: Sican GmbH */
/* Vendor: 122d: Aztech System Ltd */
#define DEVICE_AZTECH2_368DSP				0x1206 /*368DSP*/
#define DEVICE_AZTECH2_TRIDENT_PCI288_Q3DII_NX		0x1400 /*Trident PCI288-Q3DII (NX)*/
#define DEVICE_AZTECH2_3328_AUDIO			0x50dc /*3328 Audio*/
	#define SUBDEVICE_AZTECH2_3328_AUDIO		0x122d0001 /*3328 Audio*/
#define DEVICE_AZTECH2_3328_AUDIO2			0x80da /*3328 Audio*/
	#define SUBDEVICE_AZTECH2_3328_AUDIO2		0x122d0001 /*3328 Audio*/
/* Vendor: 122e: Xyratex */
/* Vendor: 122f: Andrew Corporation */
/* Vendor: 1230: Fishcamp Engineering */
/* Vendor: 1231: Woodward McCoach, Inc. */
/* Vendor: 1232: GPT Limited */
/* Vendor: 1233: Bus-Tech, Inc. */
/* Vendor: 1234: Technical Corp. */
/* Vendor: 1235: Risq Modular Systems, Inc. */
/* Vendor: 1236: Sigma Designs Corporation */
#define DEVICE_SIGMA2_REALMAGIC64_GX			0x0000 /*RealMagic64/GX*/
#define DEVICE_SIGMA2_REALMAGIC_64_GX_SD		0x6401 /*REALmagic 64/GX (SD 6425)*/
/* Vendor: 1237: Alta Technology Corporation */
/* Vendor: 1238: Adtran */
/* Vendor: 1239: 3DO Company */
/* Vendor: 123a: Visicom Laboratories, Inc. */
/* Vendor: 123b: Seeq Technology, Inc. */
/* Vendor: 123c: Century Systems, Inc. */
/* Vendor: 123d: Engineering Design Team, Inc. */
#define DEVICE_ENGINEERING_EASYCONNECT_8_32		0x0000 /*EasyConnect 8/32*/
#define DEVICE_ENGINEERING_EASYCONNECT_8_64		0x0002 /*EasyConnect 8/64*/
#define DEVICE_ENGINEERING_EASYIO			0x0003 /*EasyIO*/
/* Vendor: 123e: Simutech, Inc. */
/* Vendor: 123f: C-Cube Microsystems */
#define DEVICE_C_CUBE_MPEG				0x00e4 /*MPEG*/
#define DEVICE_C_CUBE_E4					0x8120 /*E4?*/
	#define SUBDEVICE_C_CUBE_DV500_E4		0x11bd0006 /*DV500 E4*/
	#define SUBDEVICE_C_CUBE_DV500_E42		0x11bd000a /*DV500 E4*/
#define DEVICE_C_CUBE_CINEMASTER_C_3_0			0x8888 /*Cinemaster C 3.0 DVD Decoder*/
	#define SUBDEVICE_C_CUBE_CINEMASTER_C_3_0	0x10020001 /*Cinemaster C 3.0 DVD Decoder*/
	#define SUBDEVICE_C_CUBE_CINEMASTER_C_3_02	0x10020002 /*Cinemaster C 3.0 DVD Decoder*/
	#define SUBDEVICE_C_CUBE_CINEMASTER_C_3_03	0x13280001 /*Cinemaster C 3.0 DVD Decoder*/
/* Vendor: 1240: Marathon Technologies Corp. */
/* Vendor: 1241: DSC Communications */
/* Vendor: 1242: JNI Corporation */
#define DEVICE_JNI_JNIC_1560_PCI_X_FIBRE		0x1560 /*JNIC-1560 PCI-X Fibre Channel Controller*/
	#define SUBDEVICE_JNI_FCX2_6562_DUAL_CHANNEL	0x12426562 /*FCX2-6562 Dual Channel PCI-X Fibre Channel Adapter*/
	#define SUBDEVICE_JNI_FCX_6562_PCI_X_FIBRE	0x1242656a /*FCX-6562 PCI-X Fibre Channel Adapter*/
#define DEVICE_JNI_FCI_1063_FIBRE_CHANNEL		0x4643 /*FCI-1063 Fibre Channel Adapter*/
#define DEVICE_JNI_FCX2_6562_DUAL_CHANNEL		0x6562 /*FCX2-6562 Dual Channel PCI-X Fibre Channel Adapter*/
#define DEVICE_JNI_FCX_6562_PCI_X_FIBRE			0x656a /*FCX-6562 PCI-X Fibre Channel Adapter*/
/* Vendor: 1243: Delphax */
/* Vendor: 1244: AVM Audiovisuelles MKTG & Computer System GmbH */
#define DEVICE_AVM_B1_ISDN				0x0700 /*B1 ISDN*/
#define DEVICE_AVM_C4_ISDN				0x0800 /*C4 ISDN*/
#define DEVICE_AVM_A1_ISDN_FRITZ			0x0a00 /*A1 ISDN [Fritz]*/
	#define SUBDEVICE_AVM_FRITZ_CARD_ISDN_CONTROLLER	0x12440a00 /*FRITZ!Card ISDN Controller*/
#define DEVICE_AVM_FRITZ_PCI_V2_0_ISDN			0x0e00 /*Fritz!PCI v2.0 ISDN*/
#define DEVICE_AVM_C2_ISDN				0x1100 /*C2 ISDN*/
#define DEVICE_AVM_T1_ISDN				0x1200 /*T1 ISDN*/
#define DEVICE_AVM_FRITZ_CARD_DSL_SL			0x2700 /*Fritz!Card DSL SL*/
#define DEVICE_AVM_FRITZ_CARD_DSL_V2_0			0x2900 /*Fritz!Card DSL v2.0*/
/* Vendor: 1245: A.P.D., S.A. */
/* Vendor: 1246: Dipix Technologies, Inc. */
/* Vendor: 1247: Xylon Research, Inc. */
/* Vendor: 1248: Central Data Corporation */
/* Vendor: 1249: Samsung Electronics Co., Ltd. */
/* Vendor: 124a: AEG Electrocom GmbH */
/* Vendor: 124b: SBS/Greenspring Modular I/O */
#define DEVICE_SBS_GREENSPRING_PCI_40A_OR_CPCI_200	0x0040 /*PCI-40A or cPCI-200 Quad IndustryPack carrier*/
	#define SUBDEVICE_SBS_GREENSPRING_PCI9080_BRIDGE	0x124b9080 /*PCI9080 Bridge*/
/* Vendor: 124c: Solitron Technologies, Inc. */
/* Vendor: 124d: Stallion Technologies, Inc. */
#define DEVICE_STALLION_EASYCONNECTION_8_32		0x0000 /*EasyConnection 8/32*/
#define DEVICE_STALLION_EASYCONNECTION_8_64		0x0002 /*EasyConnection 8/64*/
#define DEVICE_STALLION_EASYIO				0x0003 /*EasyIO*/
#define DEVICE_STALLION_EASYCONNECTION_RA		0x0004 /*EasyConnection/RA*/
/* Vendor: 124e: Cylink */
/* Vendor: 124f: Infotrend Technology, Inc. */
#define DEVICE_INFOTREND_IFT_2000_SERIES_RAID		0x0041 /*IFT-2000 Series RAID Controller*/
/* Vendor: 1250: Hitachi Microcomputer System Ltd */
/* Vendor: 1251: VLSI Solutions Oy */
/* Vendor: 1253: Guzik Technical Enterprises */
/* Vendor: 1254: Linear Systems Ltd. */
/* Vendor: 1255: Optibase Ltd */
#define DEVICE_OPTIBASE_MPEG_FORGE			0x1110 /*MPEG Forge*/
#define DEVICE_OPTIBASE_MPEG_FUSION			0x1210 /*MPEG Fusion*/
#define DEVICE_OPTIBASE_VIDEOPLEX			0x2110 /*VideoPlex*/
#define DEVICE_OPTIBASE_VIDEOPLEX_CC			0x2120 /*VideoPlex CC*/
#define DEVICE_OPTIBASE_VIDEOQUEST			0x2130 /*VideoQuest*/
/* Vendor: 1256: Perceptive Solutions, Inc. */
#define DEVICE_PERCEPTIVE_PCI_2220I			0x4201 /*PCI-2220I*/
#define DEVICE_PERCEPTIVE_PCI_2240I			0x4401 /*PCI-2240I*/
#define DEVICE_PERCEPTIVE_PCI_2000			0x5201 /*PCI-2000*/
/* Vendor: 1257: Vertex Networks, Inc. */
/* Vendor: 1258: Gilbarco, Inc. */
/* Vendor: 1259: Allied Telesyn International */
#define DEVICE_ALLIED_AT_2560_FAST_ETHERNET		0x2560 /*AT-2560 Fast Ethernet Adapter (i82557B)*/
#define DEVICE_ALLIED_RTL81XX_FAST_ETHERNET		0xa117 /*RTL81xx Fast Ethernet*/
#define DEVICE_ALLIED_21X4X_DEC_TULIP_COMPATIBLE	0xa120 /*21x4x DEC-Tulip compatible 10/100 Ethernet*/
/* Vendor: 125a: ABB Power Systems */
/* Vendor: 125b: Asix Electronics Corporation */
#define DEVICE_ASIX_ALFA_GFC2204_FAST			0x1400 /*ALFA GFC2204 Fast Ethernet*/
/* Vendor: 125c: Aurora Technologies, Inc. */
#define DEVICE_AURORA_SATURN_4520P			0x0101 /*Saturn 4520P*/
#define DEVICE_AURORA_ARIES_16000P			0x0640 /*Aries 16000P*/
/* Vendor: 125d: ESS Technology */
#define DEVICE_ESS_ES336H_FAX_MODEM			0x0000 /*ES336H Fax Modem (Early Model)*/
#define DEVICE_ESS_SOLO					0x1948 /*Solo?*/
#define DEVICE_ESS_ES1968_MAESTRO_2			0x1968 /*ES1968 Maestro 2*/
	#define SUBDEVICE_ESS_ES1968_MAESTRO_2_PCI	0x10280085 /*ES1968 Maestro-2 PCI*/
	#define SUBDEVICE_ESS_ES1968_MAESTRO_2_AUDIODRIVE	0x10338051 /*ES1968 Maestro-2 Audiodrive*/
#define DEVICE_ESS_ES1969_SOLO_1_AUDIODRIVE		0x1969 /*ES1969 Solo-1 Audiodrive*/
	#define SUBDEVICE_ESS_ES1969_SOLO_1_AUDIODRIVE	0x10140166 /*ES1969 SOLO-1 AudioDrive on IBM Aptiva Mainboard*/
	#define SUBDEVICE_ESS_SOLO_1_AUDIO_ADAPTER	0x125d8888 /*Solo-1 Audio Adapter*/
#define DEVICE_ESS_ES1978_MAESTRO_2E			0x1978 /*ES1978 Maestro 2E*/
	#define SUBDEVICE_ESS_ARMADA_M700_E500		0x0e11b112 /*Armada M700/E500*/
	#define SUBDEVICE_ESS_ES1978_MAESTRO_2E_AUDIODRIVE	0x1033803c /*ES1978 Maestro-2E Audiodrive*/
	#define SUBDEVICE_ESS_ES1978_MAESTRO_2E_AUDIODRIVE2	0x10338058 /*ES1978 Maestro-2E Audiodrive*/
	#define SUBDEVICE_ESS_MONSTER_SOUND_MX400	0x10924000 /*Monster Sound MX400*/
	#define SUBDEVICE_ESS_ES1978_MAESTRO_2E_AUDIODRIVE3	0x11790001 /*ES1978 Maestro-2E Audiodrive*/
#define DEVICE_ESS_ES1988_ALLEGRO_1			0x1988 /*ES1988 Allegro-1*/
	#define SUBDEVICE_ESS_SONIC_IMPACT_S100		0x10924100 /*Sonic Impact S100*/
	#define SUBDEVICE_ESS_ESS_ALLEGRO_1_AUDIODRIVE	0x125d1988 /*ESS Allegro-1 Audiodrive*/
#define DEVICE_ESS_ESS_MODEM				0x1989 /*ESS Modem*/
	#define SUBDEVICE_ESS_ESS_MODEM			0x125d1989 /*ESS Modem*/
#define DEVICE_ESS_ES1983S_MAESTRO_3I_PCI		0x1998 /*ES1983S Maestro-3i PCI Audio Accelerator*/
	#define SUBDEVICE_ESS_LATITUDE_C600		0x102800b1 /*Latitude C600*/
	#define SUBDEVICE_ESS_ES1983S_MAESTRO_3I_DELL	0x102800e6 /*ES1983S Maestro-3i (Dell Inspiron 8100)*/
#define DEVICE_ESS_ES1983S_MAESTRO_3I_PCI2		0x1999 /*ES1983S Maestro-3i PCI Modem Accelerator*/
#define DEVICE_ESS_ES1983S_MAESTRO_3I_PCI3		0x199a /*ES1983S Maestro-3i PCI Audio Accelerator*/
#define DEVICE_ESS_ES1983S_MAESTRO_3I_PCI4		0x199b /*ES1983S Maestro-3i PCI Modem Accelerator*/
#define DEVICE_ESS_ES336H_FAX_MODEM2			0x2808 /*ES336H Fax Modem (Later Model)*/
#define DEVICE_ESS_ES2838_2839_SUPERLINK_MODEM		0x2838 /*ES2838/2839 SuperLink Modem*/
#define DEVICE_ESS_ES2898_MODEM				0x2898 /*ES2898 Modem*/
	#define SUBDEVICE_ESS_ES56_PI_DATA_FAX		0x125d0424 /*ES56-PI Data Fax Modem*/
	#define SUBDEVICE_ESS_ES56T_PI_DATA_FAX		0x125d0425 /*ES56T-PI Data Fax Modem*/
	#define SUBDEVICE_ESS_ES56V_PI_DATA_FAX		0x125d0426 /*ES56V-PI Data Fax Modem*/
	#define SUBDEVICE_ESS_VW_PI_DATA_FAX		0x125d0427 /*VW-PI Data Fax Modem*/
	#define SUBDEVICE_ESS_ES56ST_PI_DATA_FAX	0x125d0428 /*ES56ST-PI Data Fax Modem*/
	#define SUBDEVICE_ESS_ES56SV_PI_DATA_FAX	0x125d0429 /*ES56SV-PI Data Fax Modem*/
	#define SUBDEVICE_ESS_ES56_PI_DATA_FAX2		0x147ac001 /*ES56-PI Data Fax Modem*/
	#define SUBDEVICE_ESS_ES56_PI_DATA_FAX3		0x14fe0428 /*ES56-PI Data Fax Modem*/
	#define SUBDEVICE_ESS_ES56_PI_DATA_FAX4		0x14fe0429 /*ES56-PI Data Fax Modem*/
/* Vendor: 125e: Specialvideo Engineering SRL */
/* Vendor: 125f: Concurrent Technologies, Inc. */
/* Vendor: 1260: Intersil Corporation */
#define DEVICE_INTERSIL_PRISM_2_5_WAVELAN		0x3872 /*Prism 2.5 Wavelan chipset*/
	#define SUBDEVICE_INTERSIL_LAN_EXPRESS_IEEE_802_11B	0x14680202 /*LAN-Express IEEE 802.11b Wireless LAN*/
#define DEVICE_INTERSIL_PRISM_2_5_WAVELAN2		0x3873 /*Prism 2.5 Wavelan chipset*/
	#define SUBDEVICE_INTERSIL_DWL_520_WIRELESS_PCI	0x11863501 /*DWL-520 Wireless PCI Adapter*/
	#define SUBDEVICE_INTERSIL_DWL_520_WIRELESS_PCI2	0x11863700 /*DWL-520 Wireless PCI Adapter, Rev E1*/
	#define SUBDEVICE_INTERSIL_MA311_802_11B_WIRELESS	0x13854105 /*MA311 802.11b wireless adapter*/
	#define SUBDEVICE_INTERSIL_HWP01170_01_802_11B_PCI	0x16680414 /*HWP01170-01 802.11b PCI Wireless Adapter*/
	#define SUBDEVICE_INTERSIL_AIR_MATE_PC_400_PCI	0x16a51601 /*AIR.mate PC-400 PCI Wireless LAN Adapter*/
	#define SUBDEVICE_INTERSIL_WMP11_WIRELESS_802_11B	0x17373874 /*WMP11 Wireless 802.11b PCI Adapter*/
	#define SUBDEVICE_INTERSIL_WIRELESS_802_11B_MINIPCI	0x80862513 /*Wireless 802.11b MiniPCI Adapter*/
#define DEVICE_INTERSIL_ISL3886_PRISM_JAVELIN_PRISM	0x3886 /*ISL3886 [Prism Javelin/Prism Xbow]*/
	#define SUBDEVICE_INTERSIL_Z_COM_XG_901_AND	0x17cf0037 /*Z-Com XG-901 and clones Wireless Adapter*/
#define DEVICE_INTERSIL_INTERSIL_ISL3890_PRISM		0x3890 /*Intersil ISL3890 [Prism GT/Prism Duette]*/
	#define SUBDEVICE_INTERSIL_SMC2802W_WIRELESS_PCI	0x10b82802 /*SMC2802W Wireless PCI Adapter*/
	#define SUBDEVICE_INTERSIL_SMC2835W_WIRELESS_CARDBUS	0x10b82835 /*SMC2835W Wireless Cardbus Adapter*/
	#define SUBDEVICE_INTERSIL_SMC2835W_V2_WIRELESS	0x10b8a835 /*SMC2835W V2 Wireless Cardbus Adapter*/
	#define SUBDEVICE_INTERSIL_SMC2802W_V2_WIRELESS	0x1113ee03 /*SMC2802W V2 Wireless PCI Adapter*/
	#define SUBDEVICE_INTERSIL_DWL_G650_A1_WIRELESS	0x11863202 /*DWL-G650 A1 Wireless Adapter*/
	#define SUBDEVICE_INTERSIL_CG_WLCB54GT_WIRELESS_ADAPTER	0x1259c104 /*CG-WLCB54GT Wireless Adapter*/
	#define SUBDEVICE_INTERSIL_WG511_WIRELESS_ADAPTER	0x13854800 /*WG511 Wireless Adapter*/
	#define SUBDEVICE_INTERSIL_ALLNET_ALL0271_WIRELESS	0x16a51605 /*ALLNET ALL0271 Wireless PCI Adapter*/
	#define SUBDEVICE_INTERSIL_Z_COM_XG_600_AND	0x17cf0014 /*Z-Com XG-600 and clones Wireless Adapter*/
	#define SUBDEVICE_INTERSIL_Z_COM_XG_900_AND	0x17cf0020 /*Z-Com XG-900 and clones Wireless Adapter*/
#define DEVICE_INTERSIL_HMP8130_NTSC_PAL_VIDEO		0x8130 /*HMP8130 NTSC/PAL Video Decoder*/
#define DEVICE_INTERSIL_HMP8131_NTSC_PAL_VIDEO		0x8131 /*HMP8131 NTSC/PAL Video Decoder*/
/* Vendor: 1261: Matsushita-Kotobuki Electronics Industries, Ltd. */
/* Vendor: 1262: ES Computer Company, Ltd. */
/* Vendor: 1263: Sonic Solutions */
/* Vendor: 1264: Aval Nagasaki Corporation */
/* Vendor: 1265: Casio Computer Co., Ltd. */
/* Vendor: 1266: Microdyne Corporation */
#define DEVICE_MICRODYNE_NE10_100_ADAPTER_I82557B	0x0001 /*NE10/100 Adapter (i82557B)*/
#define DEVICE_MICRODYNE_NE2000PLUS_RT8029_ETHERNET	0x1910 /*NE2000Plus (RT8029) Ethernet Adapter*/
	#define SUBDEVICE_MICRODYNE_NE2000PLUS_ETHERNET_ADAPTER	0x12661910 /*NE2000Plus Ethernet Adapter*/
/* Vendor: 1267: S. A. Telecommunications */
#define DEVICE_S_A_PCR2101				0x5352 /*PCR2101*/
#define DEVICE_S_A_TELSAT_TURBO				0x5a4b /*Telsat Turbo*/
/* Vendor: 1268: Tektronix */
/* Vendor: 1269: Thomson-CSF/TTM */
/* Vendor: 126a: Lexmark International, Inc. */
/* Vendor: 126b: Adax, Inc. */
/* Vendor: 126c: Northern Telecom */
#define DEVICE_NORTHERN_10_100BASETX_RTL81XX		0x1211 /*10/100BaseTX [RTL81xx]*/
#define DEVICE_NORTHERN_802_11B_WIRELESS_ETHERNET	0x126c /*802.11b Wireless Ethernet Adapter*/
/* Vendor: 126d: Splash Technology, Inc. */
/* Vendor: 126e: Sumitomo Metal Industries, Ltd. */
/* Vendor: 126f: Silicon Motion, Inc. */
#define DEVICE_SILICON7_SM501_VOYAGERGX			0x0501 /*SM501 VoyagerGX*/
#define DEVICE_SILICON7_SM710_LYNXEM			0x0710 /*SM710 LynxEM*/
#define DEVICE_SILICON7_SM712_LYNXEM			0x0712 /*SM712 LynxEM+*/
#define DEVICE_SILICON7_SM720_LYNX3DM			0x0720 /*SM720 Lynx3DM*/
#define DEVICE_SILICON7_SM731_COUGAR3DR			0x0730 /*SM731 Cougar3DR*/
#define DEVICE_SILICON7_SM810_LYNXE			0x0810 /*SM810 LynxE*/
#define DEVICE_SILICON7_SM811_LYNXE			0x0811 /*SM811 LynxE*/
#define DEVICE_SILICON7_SM820_LYNX3D			0x0820 /*SM820 Lynx3D*/
#define DEVICE_SILICON7_SM910				0x0910 /*SM910*/
/* Vendor: 1270: Olympus Optical Co., Ltd. */
/* Vendor: 1271: GW Instruments */
/* Vendor: 1272: Telematics International */
/* Vendor: 1273: Hughes Network Systems */
#define DEVICE_HUGHES_DIRECPC				0x0002 /*DirecPC*/
/* Vendor: 1274: Ensoniq */
#define DEVICE_ENSONIQ_ES1373_AUDIOPCI_ALSO		0x1171 /*ES1373 [AudioPCI] (also Creative Labs CT5803)*/
#define DEVICE_ENSONIQ_ES1371_AUDIOPCI_97		0x1371 /*ES1371 [AudioPCI-97]*/
	#define SUBDEVICE_ENSONIQ_AUDIOPCI_ON_MOTHERBOARD	0x0e110024 /*AudioPCI on Motherboard Compaq Deskpro*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI	0x0e11b1a7 /*ES1371, ES1373 AudioPCI*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI2	0x103380ac /*ES1371, ES1373 AudioPCI*/
	#define SUBDEVICE_ENSONIQ_TAZER			0x10421854 /*Tazer*/
	#define SUBDEVICE_ENSONIQ_TABOR2		0x107b8054 /*Tabor2*/
	#define SUBDEVICE_ENSONIQ_CREATIVE_SOUND_BLASTER	0x12741371 /*Creative Sound Blaster AudioPCI64V, AudioPCI128*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI3	0x14626470 /*ES1371, ES1373 AudioPCI On Motherboard MS-6147 1.1A*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI4	0x14626560 /*ES1371, ES1373 AudioPCI On Motherboard MS-6156 1.10*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI5	0x14626630 /*ES1371, ES1373 AudioPCI On Motherboard MS-6163BX 1.0A*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI6	0x14626631 /*ES1371, ES1373 AudioPCI On Motherboard MS-6163VIA 1.0A*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI7	0x14626632 /*ES1371, ES1373 AudioPCI On Motherboard MS-6163BX 2.0A*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI8	0x14626633 /*ES1371, ES1373 AudioPCI On Motherboard MS-6163VIA 2.0A*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI9	0x14626820 /*ES1371, ES1373 AudioPCI On Motherboard MS-6182 1.00*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI10	0x14626822 /*ES1371, ES1373 AudioPCI On Motherboard MS-6182 1.00A*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI11	0x14626830 /*ES1371, ES1373 AudioPCI On Motherboard MS-6183 1.00*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI12	0x14626880 /*ES1371, ES1373 AudioPCI On Motherboard MS-6188 1.00*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI13	0x14626900 /*ES1371, ES1373 AudioPCI On Motherboard MS-6190 1.00*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI14	0x14626910 /*ES1371, ES1373 AudioPCI On Motherboard MS-6191*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI15	0x14626930 /*ES1371, ES1373 AudioPCI On Motherboard MS-6193*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI16	0x14626990 /*ES1371, ES1373 AudioPCI On Motherboard MS-6199BX 2.0A*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI17	0x14626991 /*ES1371, ES1373 AudioPCI On Motherboard MS-6199VIA 2.0A*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI18	0x14a42077 /*ES1371, ES1373 AudioPCI On Motherboard KR639*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI19	0x14a42105 /*ES1371, ES1373 AudioPCI On Motherboard MR800*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI20	0x14a42107 /*ES1371, ES1373 AudioPCI On Motherboard MR801*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI21	0x14a42172 /*ES1371, ES1373 AudioPCI On Motherboard DR739*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI22	0x15099902 /*ES1371, ES1373 AudioPCI On Motherboard KW11*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI23	0x15099903 /*ES1371, ES1373 AudioPCI On Motherboard KW31*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI24	0x15099904 /*ES1371, ES1373 AudioPCI On Motherboard KA11*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI25	0x15099905 /*ES1371, ES1373 AudioPCI On Motherboard KC13*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI26	0x152d8801 /*ES1371, ES1373 AudioPCI On Motherboard CP810E*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI27	0x152d8802 /*ES1371, ES1373 AudioPCI On Motherboard CP810*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI28	0x152d8803 /*ES1371, ES1373 AudioPCI On Motherboard P3810E*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI29	0x152d8804 /*ES1371, ES1373 AudioPCI On Motherboard P3810-S*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI30	0x152d8805 /*ES1371, ES1373 AudioPCI On Motherboard P3820-S*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI31	0x270f2001 /*ES1371, ES1373 AudioPCI On Motherboard 6CTR*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI32	0x270f2200 /*ES1371, ES1373 AudioPCI On Motherboard 6WTX*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI33	0x270f3000 /*ES1371, ES1373 AudioPCI On Motherboard 6WSV*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI34	0x270f3100 /*ES1371, ES1373 AudioPCI On Motherboard 6WIV2*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI35	0x270f3102 /*ES1371, ES1373 AudioPCI On Motherboard 6WIV*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI36	0x270f7060 /*ES1371, ES1373 AudioPCI On Motherboard 6ASA2*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI37	0x80864249 /*ES1371, ES1373 AudioPCI On Motherboard BI440ZX*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI38	0x8086424c /*ES1371, ES1373 AudioPCI On Motherboard BL440ZX*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI39	0x8086425a /*ES1371, ES1373 AudioPCI On Motherboard BZ440ZX*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI40	0x80864341 /*ES1371, ES1373 AudioPCI On Motherboard Cayman*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI41	0x80864343 /*ES1371, ES1373 AudioPCI On Motherboard Cape Cod*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI42	0x80864649 /*ES1371, ES1373 AudioPCI On Motherboard Fire Island*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI43	0x8086464a /*ES1371, ES1373 AudioPCI On Motherboard FJ440ZX*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI44	0x80864d4f /*ES1371, ES1373 AudioPCI On Motherboard Montreal*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI45	0x80864f43 /*ES1371, ES1373 AudioPCI On Motherboard OC440LX*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI46	0x80865243 /*ES1371, ES1373 AudioPCI On Motherboard RC440BX*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI47	0x80865352 /*ES1371, ES1373 AudioPCI On Motherboard SunRiver*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI48	0x80865643 /*ES1371, ES1373 AudioPCI On Motherboard Vancouver*/
	#define SUBDEVICE_ENSONIQ_ES1371_ES1373_AUDIOPCI49	0x80865753 /*ES1371, ES1373 AudioPCI On Motherboard WS440BX*/
#define DEVICE_ENSONIQ_ES1370_AUDIOPCI			0x5000 /*ES1370 [AudioPCI]*/
#define DEVICE_ENSONIQ_5880_AUDIOPCI			0x5880 /*5880 AudioPCI*/
	#define SUBDEVICE_ENSONIQ_CREATIVE_SOUND_BLASTER2	0x12742000 /*Creative Sound Blaster AudioPCI128*/
	#define SUBDEVICE_ENSONIQ_CREATIVE_SOUNDBLASTER_AUDIOPCI	0x12742003 /*Creative SoundBlaster AudioPCI 128*/
	#define SUBDEVICE_ENSONIQ_CREATIVE_SOUND_BLASTER3	0x12745880 /*Creative Sound Blaster AudioPCI128*/
	#define SUBDEVICE_ENSONIQ_SOUND_BLASTER_16PCI	0x12748001 /*Sound Blaster 16PCI 4.1ch*/
	#define SUBDEVICE_ENSONIQ_5880_AUDIOPCI_ON	0x1458a000 /*5880 AudioPCI On Motherboard 6OXET*/
	#define SUBDEVICE_ENSONIQ_5880_AUDIOPCI_ON2	0x14626880 /*5880 AudioPCI On Motherboard MS-6188 1.00*/
	#define SUBDEVICE_ENSONIQ_5880_AUDIOPCI_ON3	0x270f2001 /*5880 AudioPCI On Motherboard 6CTR*/
	#define SUBDEVICE_ENSONIQ_5880_AUDIOPCI_ON4	0x270f2200 /*5880 AudioPCI On Motherboard 6WTX*/
	#define SUBDEVICE_ENSONIQ_5880_AUDIOPCI_ON5	0x270f7040 /*5880 AudioPCI On Motherboard 6ATA4*/
/* Vendor: 1275: Network Appliance Corporation */
/* Vendor: 1276: Switched Network Technologies, Inc. */
/* Vendor: 1277: Comstream */
/* Vendor: 1278: Transtech Parallel Systems Ltd. */
#define DEVICE_TRANSTECH_TPE3_TM3_POWERPC_NODE		0x0701 /*TPE3/TM3 PowerPC Node*/
#define DEVICE_TRANSTECH_TPE5_POWERPC_PCI		0x0710 /*TPE5 PowerPC PCI board*/
/* Vendor: 1279: Transmeta Corporation */
#define DEVICE_TRANSMETA_NORTHBRIDGE			0x0295 /*Northbridge*/
#define DEVICE_TRANSMETA_LONGRUN_NORTHBRIDGE		0x0395 /*LongRun Northbridge*/
#define DEVICE_TRANSMETA_SDRAM_CONTROLLER		0x0396 /*SDRAM controller*/
#define DEVICE_TRANSMETA_BIOS_SCRATCHPAD		0x0397 /*BIOS scratchpad*/
/* Vendor: 127a: Rockwell International */
#define DEVICE_ROCKWELL_HCF_56K_DATA_FAX		0x1002 /*HCF 56k Data/Fax Modem*/
	#define SUBDEVICE_ROCKWELL_SUPRAEXPRESS_56I_PRO	0x1092094c /*SupraExpress 56i PRO [Diamond SUP2380]*/
	#define SUBDEVICE_ROCKWELL_HPG_MDP3858_U	0x122d4002 /*HPG / MDP3858-U*/
	#define SUBDEVICE_ROCKWELL_MDP3858_E		0x122d4005 /*MDP3858-E*/
	#define SUBDEVICE_ROCKWELL_MDP3858_A_NZ		0x122d4007 /*MDP3858-A/-NZ*/
	#define SUBDEVICE_ROCKWELL_MDP3858_SA		0x122d4012 /*MDP3858-SA*/
	#define SUBDEVICE_ROCKWELL_MDP3858_W		0x122d4017 /*MDP3858-W*/
	#define SUBDEVICE_ROCKWELL_MDP3858_W2		0x122d4018 /*MDP3858-W*/
	#define SUBDEVICE_ROCKWELL_ROCKWELL_56K_D_F	0x127a1002 /*Rockwell 56K D/F HCF Modem*/
#define DEVICE_ROCKWELL_HCF_56K_DATA_FAX2		0x1003 /*HCF 56k Data/Fax Modem*/
	#define SUBDEVICE_ROCKWELL_229_DF_ZEPHYR	0x0e11b0bc /*229-DF Zephyr*/
	#define SUBDEVICE_ROCKWELL_229_DF_CHEETAH	0x0e11b114 /*229-DF Cheetah*/
	#define SUBDEVICE_ROCKWELL_229_DF		0x1033802b /*229-DF*/
	#define SUBDEVICE_ROCKWELL_PCI56RX_MODEM	0x13df1003 /*PCI56RX Modem*/
	#define SUBDEVICE_ROCKWELL_IBM			0x13e00117 /*IBM*/
	#define SUBDEVICE_ROCKWELL_IBM_F_1156IV_R3_SPAIN	0x13e00147 /*IBM F-1156IV+/R3 Spain V.90 Modem*/
	#define SUBDEVICE_ROCKWELL_IBM2			0x13e00197 /*IBM*/
	#define SUBDEVICE_ROCKWELL_IBM_F_1156IV_R3_WW	0x13e001c7 /*IBM F-1156IV+/R3 WW V.90 Modem*/
	#define SUBDEVICE_ROCKWELL_IBM3			0x13e001f7 /*IBM*/
	#define SUBDEVICE_ROCKWELL_IBM4			0x14361003 /*IBM*/
	#define SUBDEVICE_ROCKWELL_IBM_5614PM3G_V_90	0x14361103 /*IBM 5614PM3G V.90 Modem*/
	#define SUBDEVICE_ROCKWELL_COMPAQ_229_DF_DUCATI	0x14361602 /*Compaq 229-DF Ducati*/
#define DEVICE_ROCKWELL_HCF_56K_DATA_FAX_VOICE		0x1004 /*HCF 56k Data/Fax/Voice Modem*/
	#define SUBDEVICE_ROCKWELL_MICROLINK_56K_MODEM	0x10481500 /*MicroLink 56k Modem*/
	#define SUBDEVICE_ROCKWELL_FUJITSU_229_DFRT	0x10cf1059 /*Fujitsu 229-DFRT*/
#define DEVICE_ROCKWELL_HCF_56K_DATA_FAX_VOICE_SPKP	0x1005 /*HCF 56k Data/Fax/Voice/Spkp (w/Handset) Modem*/
	#define SUBDEVICE_ROCKWELL_229_DFSV		0x10338029 /*229-DFSV*/
	#define SUBDEVICE_ROCKWELL_MODEM		0x10338054 /*Modem*/
	#define SUBDEVICE_ROCKWELL_FUJITSU		0x10cf103c /*Fujitsu*/
	#define SUBDEVICE_ROCKWELL_FUJITSU_229_DFSV	0x10cf1055 /*Fujitsu 229-DFSV*/
	#define SUBDEVICE_ROCKWELL_FUJITSU_229_DFSV2	0x10cf1056 /*Fujitsu 229-DFSV*/
	#define SUBDEVICE_ROCKWELL_MDP3858SP_U		0x122d4003 /*MDP3858SP-U*/
	#define SUBDEVICE_ROCKWELL_PACKARD_BELL_MDP3858V_E	0x122d4006 /*Packard Bell MDP3858V-E*/
	#define SUBDEVICE_ROCKWELL_MDP3858SP_A_SP_NZ	0x122d4008 /*MDP3858SP-A/SP-NZ*/
	#define SUBDEVICE_ROCKWELL_MDP3858SP_E		0x122d4009 /*MDP3858SP-E*/
	#define SUBDEVICE_ROCKWELL_MDP3858V_U		0x122d4010 /*MDP3858V-U*/
	#define SUBDEVICE_ROCKWELL_MDP3858SP_SA		0x122d4011 /*MDP3858SP-SA*/
	#define SUBDEVICE_ROCKWELL_MDP3858V_A_V_NZ	0x122d4013 /*MDP3858V-A/V-NZ*/
	#define SUBDEVICE_ROCKWELL_MDP3858SP_W		0x122d4015 /*MDP3858SP-W*/
	#define SUBDEVICE_ROCKWELL_MDP3858V_W		0x122d4016 /*MDP3858V-W*/
	#define SUBDEVICE_ROCKWELL_MDP3858V_SA		0x122d4019 /*MDP3858V-SA*/
	#define SUBDEVICE_ROCKWELL_PCI56RVP_MODEM	0x13df1005 /*PCI56RVP Modem*/
	#define SUBDEVICE_ROCKWELL_IBM5			0x13e00187 /*IBM*/
	#define SUBDEVICE_ROCKWELL_IBM6			0x13e001a7 /*IBM*/
	#define SUBDEVICE_ROCKWELL_IBM_DF_1156IV_R3_SPAIN	0x13e001b7 /*IBM DF-1156IV+/R3 Spain V.90 Modem*/
	#define SUBDEVICE_ROCKWELL_IBM_DF_1156IV_R3_WW	0x13e001d7 /*IBM DF-1156IV+/R3 WW V.90 Modem*/
	#define SUBDEVICE_ROCKWELL_IBM7			0x14361005 /*IBM*/
	#define SUBDEVICE_ROCKWELL_IBM8			0x14361105 /*IBM*/
	#define SUBDEVICE_ROCKWELL_IBM_5614PS3G_V_90	0x14371105 /*IBM 5614PS3G V.90 Modem*/
#define DEVICE_ROCKWELL_HCF_56K_MODEM			0x1022 /*HCF 56k Modem*/
	#define SUBDEVICE_ROCKWELL_M3_5614PM3G_V_90_MODEM	0x14361303 /*M3-5614PM3G V.90 Modem*/
#define DEVICE_ROCKWELL_HCF_56K_DATA_FAX3		0x1023 /*HCF 56k Data/Fax Modem*/
	#define SUBDEVICE_ROCKWELL_PACKARD_BELL_MDP3858_WE	0x122d4020 /*Packard Bell MDP3858-WE*/
	#define SUBDEVICE_ROCKWELL_MDP3858_UE		0x122d4023 /*MDP3858-UE*/
	#define SUBDEVICE_ROCKWELL_IBM_F_1156IV_R6_SPAIN	0x13e00247 /*IBM F-1156IV+/R6 Spain V.90 Modem*/
	#define SUBDEVICE_ROCKWELL_IBM9			0x13e00297 /*IBM*/
	#define SUBDEVICE_ROCKWELL_IBM_F_1156IV_R6_WW	0x13e002c7 /*IBM F-1156IV+/R6 WW V.90 Modem*/
	#define SUBDEVICE_ROCKWELL_IBM10		0x14361203 /*IBM*/
	#define SUBDEVICE_ROCKWELL_IBM11		0x14361303 /*IBM*/
#define DEVICE_ROCKWELL_HCF_56K_DATA_FAX_VOICE2		0x1024 /*HCF 56k Data/Fax/Voice Modem*/
#define DEVICE_ROCKWELL_HCF_56K_DATA_FAX_VOICE_SPKP2	0x1025 /*HCF 56k Data/Fax/Voice/Spkp (w/Handset) Modem*/
	#define SUBDEVICE_ROCKWELL_FUJITSU_235_DFSV	0x10cf106a /*Fujitsu 235-DFSV*/
	#define SUBDEVICE_ROCKWELL_PACKARD_BELL_MDP3858V_WE	0x122d4021 /*Packard Bell MDP3858V-WE*/
	#define SUBDEVICE_ROCKWELL_MDP3858SP_WE		0x122d4022 /*MDP3858SP-WE*/
	#define SUBDEVICE_ROCKWELL_MDP3858V_UE		0x122d4024 /*MDP3858V-UE*/
	#define SUBDEVICE_ROCKWELL_MDP3858SP_UE		0x122d4025 /*MDP3858SP-UE*/
#define DEVICE_ROCKWELL_HCF_56K_PCI			0x1026 /*HCF 56k PCI Speakerphone Modem*/
#define DEVICE_ROCKWELL_HCF_56K_MODEM2			0x1032 /*HCF 56k Modem*/
#define DEVICE_ROCKWELL_HCF_56K_MODEM3			0x1033 /*HCF 56k Modem*/
#define DEVICE_ROCKWELL_HCF_56K_MODEM4			0x1034 /*HCF 56k Modem*/
#define DEVICE_ROCKWELL_HCF_56K_PCI2			0x1035 /*HCF 56k PCI Speakerphone Modem*/
#define DEVICE_ROCKWELL_HCF_56K_MODEM5			0x1036 /*HCF 56k Modem*/
#define DEVICE_ROCKWELL_HCF_56K_VOLCANO			0x1085 /*HCF 56k Volcano PCI Modem*/
#define DEVICE_ROCKWELL_HCF_56K_DATA_FAX4		0x2005 /*HCF 56k Data/Fax Modem*/
	#define SUBDEVICE_ROCKWELL_229_DFSV2		0x104d8044 /*229-DFSV*/
	#define SUBDEVICE_ROCKWELL_229_DFSV3		0x104d8045 /*229-DFSV*/
	#define SUBDEVICE_ROCKWELL_PBE_AZTECH_235W_DFSV	0x104d8055 /*PBE/Aztech 235W-DFSV*/
	#define SUBDEVICE_ROCKWELL_235_DFSV		0x104d8056 /*235-DFSV*/
	#define SUBDEVICE_ROCKWELL_MODEM2		0x104d805a /*Modem*/
	#define SUBDEVICE_ROCKWELL_MODEM3		0x104d805f /*Modem*/
	#define SUBDEVICE_ROCKWELL_MODEM4		0x104d8074 /*Modem*/
#define DEVICE_ROCKWELL_HSF_56K_DATA_FAX		0x2013 /*HSF 56k Data/Fax Modem*/
	#define SUBDEVICE_ROCKWELL_MODEM5		0x11790001 /*Modem*/
	#define SUBDEVICE_ROCKWELL_MODEM6		0x1179ff00 /*Modem*/
#define DEVICE_ROCKWELL_HSF_56K_DATA_FAX_VOICE		0x2014 /*HSF 56k Data/Fax/Voice Modem*/
	#define SUBDEVICE_ROCKWELL_FUJITSU_CITICORP_III	0x10cf1057 /*Fujitsu Citicorp III*/
	#define SUBDEVICE_ROCKWELL_MSP3880_U		0x122d4050 /*MSP3880-U*/
	#define SUBDEVICE_ROCKWELL_MSP3880_W		0x122d4055 /*MSP3880-W*/
#define DEVICE_ROCKWELL_HSF_56K_DATA_FAX_VOICE_SPKP	0x2015 /*HSF 56k Data/Fax/Voice/Spkp (w/Handset) Modem*/
	#define SUBDEVICE_ROCKWELL_FUJITSU2		0x10cf1063 /*Fujitsu*/
	#define SUBDEVICE_ROCKWELL_FUJITSU3		0x10cf1064 /*Fujitsu*/
	#define SUBDEVICE_ROCKWELL_FUJITSU4		0x14682015 /*Fujitsu*/
#define DEVICE_ROCKWELL_HSF_56K_DATA_FAX_VOICE_SPKP2	0x2016 /*HSF 56k Data/Fax/Voice/Spkp Modem*/
	#define SUBDEVICE_ROCKWELL_MSP3880V_W		0x122d4051 /*MSP3880V-W*/
	#define SUBDEVICE_ROCKWELL_MSP3880SP_W		0x122d4052 /*MSP3880SP-W*/
	#define SUBDEVICE_ROCKWELL_MSP3880V_U		0x122d4054 /*MSP3880V-U*/
	#define SUBDEVICE_ROCKWELL_MSP3880SP_U		0x122d4056 /*MSP3880SP-U*/
	#define SUBDEVICE_ROCKWELL_MSP3880SP_A		0x122d4057 /*MSP3880SP-A*/
#define DEVICE_ROCKWELL_RIPTIDE_HSF_56K			0x4311 /*Riptide HSF 56k PCI Modem*/
	#define SUBDEVICE_ROCKWELL_RING_MODULAR_RIPTIDE	0x127a4311 /*Ring Modular? Riptide HSF RT HP Dom*/
	#define SUBDEVICE_ROCKWELL_HP_GVC		0x13e00210 /*HP-GVC*/
#define DEVICE_ROCKWELL_RIPTIDE_PCI_AUDIO		0x4320 /*Riptide PCI Audio Controller*/
	#define SUBDEVICE_ROCKWELL_RIPTIDE_PCI_AUDIO	0x12354320 /*Riptide PCI Audio Controller*/
#define DEVICE_ROCKWELL_RIPTIDE_HCF_56K			0x4321 /*Riptide HCF 56k PCI Modem*/
	#define SUBDEVICE_ROCKWELL_HEWLETT_PACKARD_DF	0x12354321 /*Hewlett Packard DF*/
	#define SUBDEVICE_ROCKWELL_HEWLETT_PACKARD_DF2	0x12354324 /*Hewlett Packard DF*/
	#define SUBDEVICE_ROCKWELL_HEWLETT_PACKARD_DF3	0x13e00210 /*Hewlett Packard DF*/
	#define SUBDEVICE_ROCKWELL_RIPTIDE		0x144d2321 /*Riptide*/
#define DEVICE_ROCKWELL_RIPTIDE_PCI_GAME		0x4322 /*Riptide PCI Game Controller*/
	#define SUBDEVICE_ROCKWELL_RIPTIDE_PCI_GAME	0x12354322 /*Riptide PCI Game Controller*/
#define DEVICE_ROCKWELL_RAPIDFIRE_616X_ATM155		0x8234 /*RapidFire 616X ATM155 Adapter*/
	#define SUBDEVICE_ROCKWELL_RAPIDFIRE_616X_ATM155	0x108d0022 /*RapidFire 616X ATM155 Adapter*/
	#define SUBDEVICE_ROCKWELL_RAPIDFIRE_616X_ATM1552	0x108d0027 /*RapidFire 616X ATM155 Adapter*/
/* Vendor: 127b: Pixera Corporation */
/* Vendor: 127c: Crosspoint Solutions, Inc. */
/* Vendor: 127d: Vela Research */
/* Vendor: 127e: Winnov, L.P. */
/* Vendor: 127f: Fujifilm */
/* Vendor: 1280: Photoscript Group Ltd. */
/* Vendor: 1281: Yokogawa Electric Corporation */
/* Vendor: 1282: Davicom Semiconductor, Inc. */
#define DEVICE_DAVICOM2_ETHERNET_100_10_MBIT		0x9009 /*Ethernet 100/10 MBit*/
#define DEVICE_DAVICOM2_21X4X_DEC_TULIP_COMPATIBLE	0x9100 /*21x4x DEC-Tulip compatible 10/100 Ethernet*/
#define DEVICE_DAVICOM2_21X4X_DEC_TULIP_COMPATIBLE2	0x9102 /*21x4x DEC-Tulip compatible 10/100 Ethernet*/
#define DEVICE_DAVICOM2_ETHERNET_100_10_MBIT2		0x9132 /*Ethernet 100/10 MBit*/
/* Vendor: 1283: Integrated Technology Express, Inc. */
#define DEVICE_INTEGRATED5_IT8330G			0x673a /*IT8330G*/
#define DEVICE_INTEGRATED5_IT_ITE8212_DUAL_CHANNEL	0x8212 /*IT/ITE8212 Dual channel ATA RAID controller (PCI version seems to be IT8212, embedded seems to be ITE8212)*/
	#define SUBDEVICE_INTEGRATED5_IT_ITE8212_DUAL_CHANNEL	0x12830001 /*IT/ITE8212 Dual channel ATA RAID controller*/
#define DEVICE_INTEGRATED5_IT8330G2			0x8330 /*IT8330G*/
#define DEVICE_INTEGRATED5_IT8874F_PCI_DUAL		0x8872 /*IT8874F PCI Dual Serial Port Controller*/
#define DEVICE_INTEGRATED5_IT8888F_PCI_TO		0x8888 /*IT8888F PCI to ISA Bridge with SMB*/
#define DEVICE_INTEGRATED5_IT8889F_PCI_TO		0x8889 /*IT8889F PCI to ISA Bridge*/
#define DEVICE_INTEGRATED5_IT8330G3			0xe886 /*IT8330G*/
/* Vendor: 1284: Sahara Networks, Inc. */
/* Vendor: 1285: Platform Technologies, Inc. */
#define DEVICE_PLATFORM_AGOGO_SOUND_CHIP		0x0100 /*AGOGO sound chip (aka ESS Maestro 1)*/
/* Vendor: 1286: Mazet GmbH */
/* Vendor: 1287: M-Pact, Inc. */
#define DEVICE_M_PACT_LS220D_DVD_DECODER		0x001e /*LS220D DVD Decoder*/
#define DEVICE_M_PACT_LS220C_DVD_DECODER		0x001f /*LS220C DVD Decoder*/
/* Vendor: 1288: Timestep Corporation */
/* Vendor: 1289: AVC Technology, Inc. */
/* Vendor: 128a: Asante Technologies, Inc. */
/* Vendor: 128b: Transwitch Corporation */
/* Vendor: 128c: Retix Corporation */
/* Vendor: 128d: G2 Networks, Inc. */
#define DEVICE_G2_NETWORKS_ATM155_ADAPTER		0x0021 /*ATM155 Adapter*/
/* Vendor: 128e: Hoontech Corporation/Samho Multi Tech Ltd. */
#define DEVICE_HOONTECH_ST128_WSS_SB			0x0008 /*ST128 WSS/SB*/
#define DEVICE_HOONTECH_ST128_SAM9407			0x0009 /*ST128 SAM9407*/
#define DEVICE_HOONTECH_ST128_GAME_PORT			0x000a /*ST128 Game Port*/
#define DEVICE_HOONTECH_ST128_MPU_PORT			0x000b /*ST128 MPU Port*/
#define DEVICE_HOONTECH_ST128_CTRL_PORT			0x000c /*ST128 Ctrl Port*/
/* Vendor: 128f: Tateno Dennou, Inc. */
/* Vendor: 1290: Sord Computer Corporation */
/* Vendor: 1291: NCS Computer Italia */
/* Vendor: 1292: Tritech Microelectronics Inc */
/* Vendor: 1293: Media Reality Technology */
/* Vendor: 1294: Rhetorex, Inc. */
/* Vendor: 1295: Imagenation Corporation */
/* Vendor: 1296: Kofax Image Products */
/* Vendor: 1297: Holco Enterprise Co, Ltd/Shuttle Computer */
/* Vendor: 1298: Spellcaster Telecommunications Inc. */
/* Vendor: 1299: Knowledge Technology Lab. */
/* Vendor: 129a: VMetro, inc. */
#define DEVICE_VMETRO_PBT_615_PCI_X_BUS			0x0615 /*PBT-615 PCI-X Bus Analyzer*/
/* Vendor: 129b: Image Access */
/* Vendor: 129c: Jaycor */
/* Vendor: 129d: Compcore Multimedia, Inc. */
/* Vendor: 129e: Victor Company of Japan, Ltd. */
/* Vendor: 129f: OEC Medical Systems, Inc. */
/* Vendor: 12a0: Allen-Bradley Company */
/* Vendor: 12a1: Simpact Associates, Inc. */
/* Vendor: 12a2: Newgen Systems Corporation */
/* Vendor: 12a3: Lucent Technologies */
#define DEVICE_LUCENT_T8105_H100_DIGITAL		0x8105 /*T8105 H100 Digital Switch*/
/* Vendor: 12a4: NTT Electronics Technology Company */
/* Vendor: 12a5: Vision Dynamics Ltd. */
/* Vendor: 12a6: Scalable Networks, Inc. */
/* Vendor: 12a7: AMO GmbH */
/* Vendor: 12a8: News Datacom */
/* Vendor: 12a9: Xiotech Corporation */
/* Vendor: 12aa: SDL Communications, Inc. */
/* Vendor: 12ab: Yuan Yuan Enterprise Co., Ltd. */
#define DEVICE_YUAN_AU8830_VORTEX2_BASED		0x0002 /*AU8830 [Vortex2] Based Sound Card With A3D Support*/
#define DEVICE_YUAN_MPG_200C_PCI_DVD			0x3000 /*MPG-200C PCI DVD Decoder Card*/
/* Vendor: 12ac: Measurex Corporation */
/* Vendor: 12ad: Multidata GmbH */
/* Vendor: 12ae: Alteon Networks Inc. */
#define DEVICE_ALTEON_ACENIC_GIGABIT_ETHERNET		0x0001 /*AceNIC Gigabit Ethernet*/
	#define SUBDEVICE_ALTEON_GIGABIT_ETHERNET_SX_PCI	0x10140104 /*Gigabit Ethernet-SX PCI Adapter*/
	#define SUBDEVICE_ALTEON_GIGABIT_ETHERNET_SX_UNIVERSAL	0x12ae0001 /*Gigabit Ethernet-SX (Universal)*/
	#define SUBDEVICE_ALTEON_GIGABIT_ETHERNET_SX_PCI2	0x14100104 /*Gigabit Ethernet-SX PCI Adapter*/
#define DEVICE_ALTEON_ACENIC_GIGABIT_ETHERNET2		0x0002 /*AceNIC Gigabit Ethernet (Copper)*/
	#define SUBDEVICE_ALTEON_ACENIC_GIGABIT_ETHERNET	0x10a98002 /*Acenic Gigabit Ethernet*/
	#define SUBDEVICE_ALTEON_GIGABIT_ETHERNET_T_3C986_T	0x12ae0002 /*Gigabit Ethernet-T (3C986-T)*/
#define DEVICE_ALTEON_FARALLON_PN9100_T_GIGABIT		0x00fa /*Farallon PN9100-T Gigabit Ethernet*/
/* Vendor: 12af: TDK USA Corp */
/* Vendor: 12b0: Jorge Scientific Corp */
/* Vendor: 12b1: GammaLink */
/* Vendor: 12b2: General Signal Networks */
/* Vendor: 12b3: Inter-Face Co Ltd */
/* Vendor: 12b4: FutureTel Inc */
/* Vendor: 12b5: Granite Systems Inc. */
/* Vendor: 12b6: Natural Microsystems */
/* Vendor: 12b7: Cognex Modular Vision Systems Div. - Acumen Inc. */
/* Vendor: 12b8: Korg */
/* Vendor: 12b9: 3Com Corp, Modem Division (formerly US Robotics) */
#define DEVICE_3COM2_WINMODEM				0x1006 /*WinModem*/
	#define SUBDEVICE_3COM2_USR_56K_INTERNAL	0x12b9005c /*USR 56k Internal Voice WinModem (Model 3472)*/
	#define SUBDEVICE_3COM2_USR_56K_INTERNAL2	0x12b9005e /*USR 56k Internal WinModem (Models 662975)*/
	#define SUBDEVICE_3COM2_USR_56K_INTERNAL3	0x12b90062 /*USR 56k Internal Voice WinModem (Model 662978)*/
	#define SUBDEVICE_3COM2_USR_56K_INTERNAL4	0x12b90068 /*USR 56k Internal Voice WinModem (Model 5690)*/
	#define SUBDEVICE_3COM2_USR_56K_INTERNAL5	0x12b9007a /*USR 56k Internal Voice WinModem (Model 662974)*/
	#define SUBDEVICE_3COM2_USR_56K_INTERNAL6	0x12b9007f /*USR 56k Internal WinModem (Models 5698, 5699)*/
	#define SUBDEVICE_3COM2_USR_56K_INTERNAL7	0x12b90080 /*USR 56k Internal WinModem (Models 2975, 3528)*/
	#define SUBDEVICE_3COM2_USR_56K_INTERNAL8	0x12b90081 /*USR 56k Internal Voice WinModem (Models 2974, 3529)*/
	#define SUBDEVICE_3COM2_USR_56K_INTERNAL9	0x12b90091 /*USR 56k Internal Voice WinModem (Model 2978)*/
#define DEVICE_3COM2_USR_56K_INTERNAL			0x1007 /*USR 56k Internal WinModem*/
	#define SUBDEVICE_3COM2_USR_56K_INTERNAL10	0x12b900a3 /*USR 56k Internal WinModem (Model 3595)*/
#define DEVICE_3COM2_56K_FAXMODEM_MODEL			0x1008 /*56K FaxModem Model 5610*/
	#define SUBDEVICE_3COM2_USR_56K_INTERNAL11	0x12b900a2 /*USR 56k Internal FAX Modem (Model 2977)*/
	#define SUBDEVICE_3COM2_USR_56K_INTERNAL12	0x12b900aa /*USR 56k Internal Voice Modem (Model 2976)*/
	#define SUBDEVICE_3COM2_USR_56K_INTERNAL13	0x12b900ab /*USR 56k Internal Voice Modem (Model 5609)*/
	#define SUBDEVICE_3COM2_USR_56K_INTERNAL14	0x12b900ac /*USR 56k Internal Voice Modem (Model 3298)*/
	#define SUBDEVICE_3COM2_USR_56K_INTERNAL15	0x12b900ad /*USR 56k Internal FAX Modem (Model 5610)*/
/* Vendor: 12ba: BittWare, Inc. */
/* Vendor: 12bb: Nippon Unisoft Corporation */
/* Vendor: 12bc: Array Microsystems */
/* Vendor: 12bd: Computerm Corp. */
/* Vendor: 12be: Anchor Chips Inc. */
#define DEVICE_ANCHOR_AN3041Q_CO_MEM			0x3041 /*AN3041Q CO-MEM*/
#define DEVICE_ANCHOR_AN3042Q_CO_MEM_LITE		0x3042 /*AN3042Q CO-MEM Lite*/
	#define SUBDEVICE_ANCHOR_ANCHOR_CHIPS_LITE	0x12be3042 /*Anchor Chips Lite Evaluation Board*/
/* Vendor: 12bf: Fujifilm Microdevices */
/* Vendor: 12c0: Infimed */
/* Vendor: 12c1: GMM Research Corp */
/* Vendor: 12c2: Mentec Limited */
/* Vendor: 12c3: Holtek Microelectronics Inc */
#define DEVICE_HOLTEK_PCI_NE2K_ETHERNET			0x0058 /*PCI NE2K Ethernet*/
#define DEVICE_HOLTEK_PCI_NE2K_ETHERNET2		0x5598 /*PCI NE2K Ethernet*/
/* Vendor: 12c4: Connect Tech Inc */
/* Vendor: 12c5: Picture Elements Incorporated */
#define DEVICE_PICTURE2_IMAGING_SCANNING_SUBSYSTEM_ENGINE	0x007e /*Imaging/Scanning Subsystem Engine*/
#define DEVICE_PICTURE2_IMAGING_SCANNING_SUBSYSTEM_ENGINE2	0x007f /*Imaging/Scanning Subsystem Engine*/
#define DEVICE_PICTURE2_PCIVST_GRAYSCALE_THRESHOLDING	0x0081 /*PCIVST [Grayscale Thresholding Engine]*/
#define DEVICE_PICTURE2_VIDEO_SIMULATOR_SENDER		0x0085 /*Video Simulator/Sender*/
#define DEVICE_PICTURE2_THR2_MULTI_SCALE_THRESHOLDER	0x0086 /*THR2 Multi-scale Thresholder*/
/* Vendor: 12c6: Mitani Corporation */
/* Vendor: 12c7: Dialogic Corp */
/* Vendor: 12c8: G Force Co, Ltd */
/* Vendor: 12c9: Gigi Operations */
/* Vendor: 12ca: Integrated Computing Engines */
/* Vendor: 12cb: Antex Electronics Corporation */
/* Vendor: 12cc: Pluto Technologies International */
/* Vendor: 12cd: Aims Lab */
/* Vendor: 12ce: Netspeed Inc. */
/* Vendor: 12cf: Prophet Systems, Inc. */
/* Vendor: 12d0: GDE Systems, Inc. */
/* Vendor: 12d1: PSITech */
/* Vendor: 12d2: NVidia / SGS Thomson (Joint Venture) */
#define DEVICE_NVIDIA2_NV1				0x0008 /*NV1*/
#define DEVICE_NVIDIA2_DAC64				0x0009 /*DAC64*/
#define DEVICE_NVIDIA2_RIVA128				0x0018 /*Riva128*/
	#define SUBDEVICE_NVIDIA2_VICTORY_ERAZOR	0x10480c10 /*VICTORY Erazor*/
	#define SUBDEVICE_NVIDIA2_STB_VELOCITY_128	0x107b8030 /*STB Velocity 128*/
	#define SUBDEVICE_NVIDIA2_VIPER_V330		0x10920350 /*Viper V330*/
	#define SUBDEVICE_NVIDIA2_VIPER_V3302		0x10921092 /*Viper V330*/
	#define SUBDEVICE_NVIDIA2_STB_VELOCITY_1282	0x10b41b1b /*STB Velocity 128*/
	#define SUBDEVICE_NVIDIA2_STB_VELOCITY_1283	0x10b41b1d /*STB Velocity 128*/
	#define SUBDEVICE_NVIDIA2_STB_VELOCITY_1284	0x10b41b1e /*STB Velocity 128, PAL TV-Out*/
	#define SUBDEVICE_NVIDIA2_STB_VELOCITY_1285	0x10b41b20 /*STB Velocity 128 Sapphire*/
	#define SUBDEVICE_NVIDIA2_STB_VELOCITY_1286	0x10b41b21 /*STB Velocity 128*/
	#define SUBDEVICE_NVIDIA2_STB_VELOCITY_1287	0x10b41b22 /*STB Velocity 128 AGP, NTSC TV-Out*/
	#define SUBDEVICE_NVIDIA2_STB_VELOCITY_1288	0x10b41b23 /*STB Velocity 128 AGP, PAL TV-Out*/
	#define SUBDEVICE_NVIDIA2_STB_VELOCITY_1289	0x10b41b27 /*STB Velocity 128 DVD*/
	#define SUBDEVICE_NVIDIA2_MVP_PRO_128		0x10b41b88 /*MVP Pro 128*/
	#define SUBDEVICE_NVIDIA2_STB_VELOCITY_12810	0x10b4222a /*STB Velocity 128 AGP*/
	#define SUBDEVICE_NVIDIA2_STB_VELOCITY_12811	0x10b42230 /*STB Velocity 128*/
	#define SUBDEVICE_NVIDIA2_STB_VELOCITY_12812	0x10b42232 /*STB Velocity 128*/
	#define SUBDEVICE_NVIDIA2_STB_VELOCITY_12813	0x10b42235 /*STB Velocity 128 AGP*/
	#define SUBDEVICE_NVIDIA2_3DVISION_SAGP_3DEXPLORER	0x2a1554a3 /*3DVision-SAGP / 3DexPlorer 3000*/
#define DEVICE_NVIDIA2_RIVA128ZX			0x0019 /*Riva128ZX*/
#define DEVICE_NVIDIA2_TNT				0x0020 /*TNT*/
#define DEVICE_NVIDIA2_TNT2				0x0028 /*TNT2*/
#define DEVICE_NVIDIA2_UTNT2				0x0029 /*UTNT2*/
#define DEVICE_NVIDIA2_VTNT2				0x002c /*VTNT2*/
#define DEVICE_NVIDIA2_ITNT2				0x00a0 /*ITNT2*/
/* Vendor: 12d3: Vingmed Sound A/S */
/* Vendor: 12d4: Ulticom (Formerly DGM&S) */
#define DEVICE_ULTICOM_T1_CARD				0x0200 /*T1 Card*/
/* Vendor: 12d5: Equator Technologies Inc */
/* Vendor: 12d6: Analogic Corp */
/* Vendor: 12d7: Biotronic SRL */
/* Vendor: 12d8: Pericom Semiconductor */
/* Vendor: 12d9: Aculab PLC */
#define DEVICE_ACULAB_PCI_PROSODY			0x0002 /*PCI Prosody*/
#define DEVICE_ACULAB_CPCI_PROSODY			0x0004 /*cPCI Prosody*/
#define DEVICE_ACULAB_ACULAB_E1_T1_PCI			0x0005 /*Aculab E1/T1 PCI card*/
/* Vendor: 12da: True Time Inc. */
/* Vendor: 12db: Annapolis Micro Systems, Inc */
/* Vendor: 12dc: Symicron Computer Communication Ltd. */
/* Vendor: 12dd: Management Graphics */
/* Vendor: 12de: Rainbow Technologies */
#define DEVICE_RAINBOW_CRYPTOSWIFT_CS200		0x0200 /*CryptoSwift CS200*/
/* Vendor: 12df: SBS Technologies Inc */
/* Vendor: 12e0: Chase Research */
#define DEVICE_CHASE_ST16C654_QUAD_UART			0x0010 /*ST16C654 Quad UART*/
#define DEVICE_CHASE_ST16C654_QUAD_UART2		0x0020 /*ST16C654 Quad UART*/
#define DEVICE_CHASE_ST16C654_QUAD_UART3		0x0030 /*ST16C654 Quad UART*/
/* Vendor: 12e1: Nintendo Co, Ltd */
/* Vendor: 12e2: Datum Inc. Bancomm-Timing Division */
/* Vendor: 12e3: Imation Corp - Medical Imaging Systems */
/* Vendor: 12e4: Brooktrout Technology Inc */
/* Vendor: 12e5: Apex Semiconductor Inc */
/* Vendor: 12e6: Cirel Systems */
/* Vendor: 12e7: Sunsgroup Corporation */
/* Vendor: 12e8: Crisc Corp */
/* Vendor: 12e9: GE Spacenet */
/* Vendor: 12ea: Zuken */
/* Vendor: 12eb: Aureal Semiconductor */
#define DEVICE_AUREAL_VORTEX_1				0x0001 /*Vortex 1*/
	#define SUBDEVICE_AUREAL_AU8820_VORTEX_DIGITAL	0x104d8036 /*AU8820 Vortex Digital Audio Processor*/
	#define SUBDEVICE_AUREAL_SONIC_IMPACT_A3D	0x10922000 /*Sonic Impact A3D*/
	#define SUBDEVICE_AUREAL_SONIC_IMPACT_A3D2	0x10922100 /*Sonic Impact A3D*/
	#define SUBDEVICE_AUREAL_SONIC_IMPACT_A3D3	0x10922110 /*Sonic Impact A3D*/
	#define SUBDEVICE_AUREAL_SONIC_IMPACT_A3D4	0x10922200 /*Sonic Impact A3D*/
	#define SUBDEVICE_AUREAL_AU8820_VORTEX_DIGITAL2	0x122d1002 /*AU8820 Vortex Digital Audio Processor*/
	#define SUBDEVICE_AUREAL_AU8820_VORTEX_DIGITAL3	0x12eb0001 /*AU8820 Vortex Digital Audio Processor*/
	#define SUBDEVICE_AUREAL_MONTEGO		0x50533355 /*Montego*/
#define DEVICE_AUREAL_VORTEX_2				0x0002 /*Vortex 2*/
	#define SUBDEVICE_AUREAL_AU8830_VORTEX_3D	0x104d8049 /*AU8830 Vortex 3D Digital Audio Processor*/
	#define SUBDEVICE_AUREAL_AU8830_VORTEX_3D2	0x104d807b /*AU8830 Vortex 3D Digital Audio Processor*/
	#define SUBDEVICE_AUREAL_MONSTER_SOUND_II	0x10923000 /*Monster Sound II*/
	#define SUBDEVICE_AUREAL_MONSTER_SOUND_II2	0x10923001 /*Monster Sound II*/
	#define SUBDEVICE_AUREAL_MONSTER_SOUND_II3	0x10923002 /*Monster Sound II*/
	#define SUBDEVICE_AUREAL_MONSTER_SOUND_II4	0x10923003 /*Monster Sound II*/
	#define SUBDEVICE_AUREAL_MONSTER_SOUND_II5	0x10923004 /*Monster Sound II*/
	#define SUBDEVICE_AUREAL_AU8830_VORTEX_3D3	0x12eb0001 /*AU8830 Vortex 3D Digital Audio Processor*/
	#define SUBDEVICE_AUREAL_AU8830_VORTEX_3D4	0x12eb0002 /*AU8830 Vortex 3D Digital Audio Processor*/
	#define SUBDEVICE_AUREAL_AU8830_VORTEX_3D5	0x12eb0088 /*AU8830 Vortex 3D Digital Audio Processor*/
	#define SUBDEVICE_AUREAL_AU8830_VORTEX_3D6	0x144d3510 /*AU8830 Vortex 3D Digital Audio Processor*/
	#define SUBDEVICE_AUREAL_MONTEGO_II		0x50533356 /*Montego II*/
#define DEVICE_AUREAL_AU8810_VORTEX_DIGITAL		0x0003 /*AU8810 Vortex Digital Audio Processor*/
	#define SUBDEVICE_AUREAL_AU8810_VORTEX_DIGITAL	0x104d8049 /*AU8810 Vortex Digital Audio Processor*/
	#define SUBDEVICE_AUREAL_AU8810_VORTEX_DIGITAL2	0x104d8077 /*AU8810 Vortex Digital Audio Processor*/
	#define SUBDEVICE_AUREAL_AU8810_VORTEX_DIGITAL3	0x109f1000 /*AU8810 Vortex Digital Audio Processor*/
	#define SUBDEVICE_AUREAL_AU8810_VORTEX_DIGITAL4	0x12eb0003 /*AU8810 Vortex Digital Audio Processor*/
	#define SUBDEVICE_AUREAL_AU8810_VORTEX_DIGITAL5	0x14626780 /*AU8810 Vortex Digital Audio Processor*/
	#define SUBDEVICE_AUREAL_AU8810_VORTEX_DIGITAL6	0x14a42073 /*AU8810 Vortex Digital Audio Processor*/
	#define SUBDEVICE_AUREAL_AU8810_VORTEX_DIGITAL7	0x14a42091 /*AU8810 Vortex Digital Audio Processor*/
	#define SUBDEVICE_AUREAL_AU8810_VORTEX_DIGITAL8	0x14a42104 /*AU8810 Vortex Digital Audio Processor*/
	#define SUBDEVICE_AUREAL_AU8810_VORTEX_DIGITAL9	0x14a42106 /*AU8810 Vortex Digital Audio Processor*/
#define DEVICE_AUREAL_VORTEX_56K_SOFTWARE		0x8803 /*Vortex 56k Software Modem*/
	#define SUBDEVICE_AUREAL_VORTEX_56K_SOFTWARE	0x12eb8803 /*Vortex 56k Software Modem*/
/* Vendor: 12ec: 3A International, Inc. */
/* Vendor: 12ed: Optivision Inc. */
/* Vendor: 12ee: Orange Micro */
/* Vendor: 12ef: Vienna Systems */
/* Vendor: 12f0: Pentek */
/* Vendor: 12f1: Sorenson Vision Inc */
/* Vendor: 12f2: Gammagraphx, Inc. */
/* Vendor: 12f3: Radstone Technology */
/* Vendor: 12f4: Megatel */
/* Vendor: 12f5: Forks */
/* Vendor: 12f6: Dawson France */
/* Vendor: 12f7: Cognex */
/* Vendor: 12f8: Electronic Design GmbH */
#define DEVICE_ELECTRONIC_VIDEOMAKER			0x0002 /*VideoMaker*/
/* Vendor: 12f9: Four Fold Ltd */
/* Vendor: 12fb: Spectrum Signal Processing */
/* Vendor: 12fc: Capital Equipment Corp */
/* Vendor: 12fd: I2S */
/* Vendor: 12fe: ESD Electronic System Design GmbH */
/* Vendor: 12ff: Lexicon */
/* Vendor: 1300: Harman International Industries Inc */
/* Vendor: 1302: Computer Sciences Corp */
/* Vendor: 1303: Innovative Integration */
/* Vendor: 1304: Juniper Networks */
/* Vendor: 1305: Netphone, Inc */
/* Vendor: 1306: Duet Technologies */
/* Vendor: 1307: Measurement Computing */
#define DEVICE_MEASUREMENT_PCI_DAS1602_16		0x0001 /*PCI-DAS1602/16*/
#define DEVICE_MEASUREMENT_PCI_DIO48H			0x000b /*PCI-DIO48H*/
#define DEVICE_MEASUREMENT_PCI_PDISO8			0x000c /*PCI-PDISO8*/
#define DEVICE_MEASUREMENT_PCI_PDISO16			0x000d /*PCI-PDISO16*/
#define DEVICE_MEASUREMENT_PCI_DAS1200			0x000f /*PCI-DAS1200*/
#define DEVICE_MEASUREMENT_PCI_DAS1602_12		0x0010 /*PCI-DAS1602/12*/
#define DEVICE_MEASUREMENT_PCI_DIO24H			0x0014 /*PCI-DIO24H*/
#define DEVICE_MEASUREMENT_PCI_DIO24H_CTR3		0x0015 /*PCI-DIO24H/CTR3*/
#define DEVICE_MEASUREMENT_PCI_DIO48H_CTR15		0x0016 /*PCI-DIO48H/CTR15*/
#define DEVICE_MEASUREMENT_PCI_DIO96H			0x0017 /*PCI-DIO96H*/
#define DEVICE_MEASUREMENT_PCI_CTR05			0x0018 /*PCI-CTR05*/
#define DEVICE_MEASUREMENT_PCI_DAS1200_JR		0x0019 /*PCI-DAS1200/JR*/
#define DEVICE_MEASUREMENT_PCI_DAS1001			0x001a /*PCI-DAS1001*/
#define DEVICE_MEASUREMENT_PCI_DAS1002			0x001b /*PCI-DAS1002*/
#define DEVICE_MEASUREMENT_PCI_DAS1602JR_16		0x001c /*PCI-DAS1602JR/16*/
#define DEVICE_MEASUREMENT_PCI_DAS6402_16		0x001d /*PCI-DAS6402/16*/
#define DEVICE_MEASUREMENT_PCI_DAS6402_12		0x001e /*PCI-DAS6402/12*/
#define DEVICE_MEASUREMENT_PCI_DAS16_M1			0x001f /*PCI-DAS16/M1*/
#define DEVICE_MEASUREMENT_PCI_DDA02_12			0x0020 /*PCI-DDA02/12*/
#define DEVICE_MEASUREMENT_PCI_DDA04_12			0x0021 /*PCI-DDA04/12*/
#define DEVICE_MEASUREMENT_PCI_DDA08_12			0x0022 /*PCI-DDA08/12*/
#define DEVICE_MEASUREMENT_PCI_DDA02_16			0x0023 /*PCI-DDA02/16*/
#define DEVICE_MEASUREMENT_PCI_DDA04_16			0x0024 /*PCI-DDA04/16*/
#define DEVICE_MEASUREMENT_PCI_DDA08_16			0x0025 /*PCI-DDA08/16*/
#define DEVICE_MEASUREMENT_PCI_DAC04_12_HS		0x0026 /*PCI-DAC04/12-HS*/
#define DEVICE_MEASUREMENT_PCI_DAC04_16_HS		0x0027 /*PCI-DAC04/16-HS*/
#define DEVICE_MEASUREMENT_PCI_DIO24			0x0028 /*PCI-DIO24*/
#define DEVICE_MEASUREMENT_PCI_DAS08			0x0029 /*PCI-DAS08*/
#define DEVICE_MEASUREMENT_PCI_INT32			0x002c /*PCI-INT32*/
#define DEVICE_MEASUREMENT_PCI_DUAL_AC5			0x0033 /*PCI-DUAL-AC5*/
#define DEVICE_MEASUREMENT_PCI_DAS_TC			0x0034 /*PCI-DAS-TC*/
#define DEVICE_MEASUREMENT_PCI_DAS64_M1_16		0x0035 /*PCI-DAS64/M1/16*/
#define DEVICE_MEASUREMENT_PCI_DAS64_M2_16		0x0036 /*PCI-DAS64/M2/16*/
#define DEVICE_MEASUREMENT_PCI_DAS64_M3_16		0x0037 /*PCI-DAS64/M3/16*/
#define DEVICE_MEASUREMENT_PCI_DAS1000			0x004c /*PCI-DAS1000*/
#define DEVICE_MEASUREMENT_PCI_QUAD04			0x004d /*PCI-QUAD04*/
#define DEVICE_MEASUREMENT_PCI_DAS4020_12		0x0052 /*PCI-DAS4020/12*/
#define DEVICE_MEASUREMENT_PCI_DAS6025			0x005e /*PCI-DAS6025*/
/* Vendor: 1308: Jato Technologies Inc. */
#define DEVICE_JATO_NETCELERATOR_ADAPTER		0x0001 /*NetCelerator Adapter*/
	#define SUBDEVICE_JATO_NETCELERATOR_ADAPTER	0x13080001 /*NetCelerator Adapter*/
/* Vendor: 1309: AB Semiconductor Ltd */
/* Vendor: 130a: Mitsubishi Electric Microcomputer */
/* Vendor: 130b: Colorgraphic Communications Corp */
/* Vendor: 130c: Ambex Technologies, Inc */
/* Vendor: 130d: Accelerix Inc */
/* Vendor: 130e: Yamatake-Honeywell Co. Ltd */
/* Vendor: 130f: Advanet Inc */
/* Vendor: 1310: Gespac */
/* Vendor: 1311: Videoserver, Inc */
/* Vendor: 1312: Acuity Imaging, Inc */
/* Vendor: 1313: Yaskawa Electric Co. */
/* Vendor: 1316: Teradyne Inc */
/* Vendor: 1317: Linksys */
#define DEVICE_LINKSYS_21X4X_DEC_TULIP_COMPATIBLE	0x0981 /*21x4x DEC-Tulip compatible 10/100 Ethernet*/
#define DEVICE_LINKSYS_NC100_NETWORK_EVERYWHERE		0x0985 /*NC100 Network Everywhere Fast Ethernet 10/100*/
#define DEVICE_LINKSYS_21X4X_DEC_TULIP_COMPATIBLE2	0x1985 /*21x4x DEC-Tulip compatible 10/100 Ethernet*/
#define DEVICE_LINKSYS_HSP_MICROMODEM_56		0x2850 /*HSP MicroModem 56*/
#define DEVICE_LINKSYS_ADMTEK_ADM8211_802_11B		0x8201 /*ADMtek ADM8211 802.11b Wireless Interface*/
	#define SUBDEVICE_LINKSYS_SMC2635W_802_11B_11MBPS	0x10b82635 /*SMC2635W 802.11b (11Mbps) wireless lan pcmcia (cardbus) card*/
	#define SUBDEVICE_LINKSYS_SMC2635W_802_11B_11MBPS2	0x13178201 /*SMC2635W 802.11b (11mbps) wireless lan pcmcia (cardbus) card*/
#define DEVICE_LINKSYS_ADMTEK_ADM8211_802_11B2		0x8211 /*ADMtek ADM8211 802.11b Wireless Interface*/
#define DEVICE_LINKSYS_21X4X_DEC_TULIP_COMPATIBLE3	0x9511 /*21x4x DEC-Tulip compatible 10/100 Ethernet*/
/* Vendor: 1318: Packet Engines Inc. */
#define DEVICE_PACKET_GNIC_II_PCI_GIGABIT		0x0911 /*GNIC-II PCI Gigabit Ethernet [Hamachi]*/
/* Vendor: 1319: Fortemedia, Inc */
#define DEVICE_FORTEMEDIA_XWAVE_QS3000A_FM801		0x0801 /*Xwave QS3000A [FM801]*/
#define DEVICE_FORTEMEDIA_XWAVE_QS3000A_FM8012		0x0802 /*Xwave QS3000A [FM801 game port]*/
#define DEVICE_FORTEMEDIA_FM801_PCI_AUDIO		0x1000 /*FM801 PCI Audio*/
#define DEVICE_FORTEMEDIA_FM801_PCI_JOYSTICK		0x1001 /*FM801 PCI Joystick*/
/* Vendor: 131a: Finisar Corp. */
/* Vendor: 131c: Nippon Electro-Sensory Devices Corp */
/* Vendor: 131d: Sysmic, Inc. */
/* Vendor: 131e: Xinex Networks Inc */
/* Vendor: 131f: Siig Inc */
#define DEVICE_SIIG_CYBERSERIAL_1_PORT_16550		0x1000 /*CyberSerial (1-port) 16550*/
#define DEVICE_SIIG_CYBERSERIAL_1_PORT_16650		0x1001 /*CyberSerial (1-port) 16650*/
#define DEVICE_SIIG_CYBERSERIAL_1_PORT_16850		0x1002 /*CyberSerial (1-port) 16850*/
#define DEVICE_SIIG_DUET_1S_16550_1P			0x1010 /*Duet 1S(16550)+1P*/
#define DEVICE_SIIG_DUET_1S_16650_1P			0x1011 /*Duet 1S(16650)+1P*/
#define DEVICE_SIIG_DUET_1S_16850_1P			0x1012 /*Duet 1S(16850)+1P*/
#define DEVICE_SIIG_CYBERPARALLEL_1_PORT		0x1020 /*CyberParallel (1-port)*/
#define DEVICE_SIIG_CYBERPARALLEL_2_PORT		0x1021 /*CyberParallel (2-port)*/
#define DEVICE_SIIG_CYBERSERIAL_2_PORT_16550		0x1030 /*CyberSerial (2-port) 16550*/
#define DEVICE_SIIG_CYBERSERIAL_2_PORT_16650		0x1031 /*CyberSerial (2-port) 16650*/
#define DEVICE_SIIG_CYBERSERIAL_2_PORT_16850		0x1032 /*CyberSerial (2-port) 16850*/
#define DEVICE_SIIG_TRIO_2S_16550_1P			0x1034 /*Trio 2S(16550)+1P*/
#define DEVICE_SIIG_TRIO_2S_16650_1P			0x1035 /*Trio 2S(16650)+1P*/
#define DEVICE_SIIG_TRIO_2S_16850_1P			0x1036 /*Trio 2S(16850)+1P*/
#define DEVICE_SIIG_CYBERSERIAL_4_PORT_16550		0x1050 /*CyberSerial (4-port) 16550*/
#define DEVICE_SIIG_CYBERSERIAL_4_PORT_16650		0x1051 /*CyberSerial (4-port) 16650*/
#define DEVICE_SIIG_CYBERSERIAL_4_PORT_16850		0x1052 /*CyberSerial (4-port) 16850*/
#define DEVICE_SIIG_CYBERSERIAL_1_PORT_165502		0x2000 /*CyberSerial (1-port) 16550*/
#define DEVICE_SIIG_CYBERSERIAL_1_PORT_166502		0x2001 /*CyberSerial (1-port) 16650*/
#define DEVICE_SIIG_CYBERSERIAL_1_PORT_168502		0x2002 /*CyberSerial (1-port) 16850*/
#define DEVICE_SIIG_DUET_1S_16550_1P2			0x2010 /*Duet 1S(16550)+1P*/
#define DEVICE_SIIG_DUET_1S_16650_1P2			0x2011 /*Duet 1S(16650)+1P*/
#define DEVICE_SIIG_DUET_1S_16850_1P2			0x2012 /*Duet 1S(16850)+1P*/
#define DEVICE_SIIG_CYBERPARALLEL_1_PORT2		0x2020 /*CyberParallel (1-port)*/
#define DEVICE_SIIG_CYBERPARALLEL_2_PORT2		0x2021 /*CyberParallel (2-port)*/
#define DEVICE_SIIG_CYBERSERIAL_2_PORT_165502		0x2030 /*CyberSerial (2-port) 16550*/
	#define SUBDEVICE_SIIG_PCI_SERIAL_CARD		0x131f2030 /*PCI Serial Card*/
#define DEVICE_SIIG_CYBERSERIAL_2_PORT_166502		0x2031 /*CyberSerial (2-port) 16650*/
#define DEVICE_SIIG_CYBERSERIAL_2_PORT_168502		0x2032 /*CyberSerial (2-port) 16850*/
#define DEVICE_SIIG_TRIO_1S_16550_2P			0x2040 /*Trio 1S(16550)+2P*/
#define DEVICE_SIIG_TRIO_1S_16650_2P			0x2041 /*Trio 1S(16650)+2P*/
#define DEVICE_SIIG_TRIO_1S_16850_2P			0x2042 /*Trio 1S(16850)+2P*/
#define DEVICE_SIIG_CYBERSERIAL_4_PORT_165502		0x2050 /*CyberSerial (4-port) 16550*/
#define DEVICE_SIIG_CYBERSERIAL_4_PORT_166502		0x2051 /*CyberSerial (4-port) 16650*/
#define DEVICE_SIIG_CYBERSERIAL_4_PORT_168502		0x2052 /*CyberSerial (4-port) 16850*/
#define DEVICE_SIIG_TRIO_2S_16550_1P2			0x2060 /*Trio 2S(16550)+1P*/
#define DEVICE_SIIG_TRIO_2S_16650_1P2			0x2061 /*Trio 2S(16650)+1P*/
#define DEVICE_SIIG_TRIO_2S_16850_1P2			0x2062 /*Trio 2S(16850)+1P*/
#define DEVICE_SIIG_CYBERSERIAL_8_PORT_ST16654		0x2081 /*CyberSerial (8-port) ST16654*/
/* Vendor: 1320: Crypto AG */
/* Vendor: 1321: Arcobel Graphics BV */
/* Vendor: 1322: MTT Co., Ltd */
/* Vendor: 1323: Dome Inc */
/* Vendor: 1324: Sphere Communications */
/* Vendor: 1325: Salix Technologies, Inc */
/* Vendor: 1326: Seachange international */
/* Vendor: 1327: Voss scientific */
/* Vendor: 1328: quadrant international */
/* Vendor: 1329: Productivity Enhancement */
/* Vendor: 132a: Microcom Inc. */
/* Vendor: 132b: Broadband Technologies */
/* Vendor: 132c: Micrel Inc */
/* Vendor: 132d: Integrated Silicon Solution, Inc. */
/* Vendor: 1330: MMC Networks */
/* Vendor: 1331: Radisys Corp. */
#define DEVICE_RADISYS_ENP_2611				0x0030 /*ENP-2611*/
#define DEVICE_RADISYS_82600_HOST_BRIDGE		0x8200 /*82600 Host Bridge*/
#define DEVICE_RADISYS_82600_IDE			0x8201 /*82600 IDE*/
#define DEVICE_RADISYS_82600_USB			0x8202 /*82600 USB*/
#define DEVICE_RADISYS_82600_PCI_BRIDGE			0x8210 /*82600 PCI Bridge*/
/* Vendor: 1332: Micro Memory */
#define DEVICE_MICRO3_MM_5415CN_PCI_MEMORY		0x5415 /*MM-5415CN PCI Memory Module with Battery Backup*/
#define DEVICE_MICRO3_MM_5425CN_PCI_64_66		0x5425 /*MM-5425CN PCI 64/66 Memory Module with Battery Backup*/
/* Vendor: 1334: Redcreek Communications, Inc */
/* Vendor: 1335: Videomail, Inc */
/* Vendor: 1337: Third Planet Publishing */
/* Vendor: 1338: BT Electronics */
/* Vendor: 133a: Vtel Corp */
/* Vendor: 133b: Softcom Microsystems */
/* Vendor: 133c: Holontech Corp */
/* Vendor: 133d: SS Technologies */
/* Vendor: 133e: Virtual Computer Corp */
/* Vendor: 133f: SCM Microsystems */
/* Vendor: 1340: Atalla Corp */
/* Vendor: 1341: Kyoto Microcomputer Co */
/* Vendor: 1342: Promax Systems Inc */
/* Vendor: 1343: Phylon Communications Inc */
/* Vendor: 1344: Crucial Technology */
/* Vendor: 1345: Arescom Inc */
/* Vendor: 1347: Odetics */
/* Vendor: 1349: Sumitomo Electric Industries, Ltd. */
/* Vendor: 134a: DTC Technology Corp. */
#define DEVICE_DTC_DOMEX_536				0x0001 /*Domex 536*/
#define DEVICE_DTC_DOMEX_DMX3194UP_SCSI			0x0002 /*Domex DMX3194UP SCSI Adapter*/
/* Vendor: 134b: ARK Research Corp. */
/* Vendor: 134c: Chori Joho System Co. Ltd */
/* Vendor: 134d: PCTel Inc */
#define DEVICE_PCTEL_HSP56_MICROMODEM			0x2189 /*HSP56 MicroModem*/
#define DEVICE_PCTEL_2304WT_V_92_MDC			0x2486 /*2304WT V.92 MDC Modem*/
#define DEVICE_PCTEL_HSP_MICROMODEM_56			0x7890 /*HSP MicroModem 56*/
	#define SUBDEVICE_PCTEL_PCT789_ADAPTER		0x134d0001 /*PCT789 adapter*/
#define DEVICE_PCTEL_HSP_MICROMODEM_562			0x7891 /*HSP MicroModem 56*/
	#define SUBDEVICE_PCTEL_HSP_MICROMODEM_56	0x134d0001 /*HSP MicroModem 56*/
#define DEVICE_PCTEL_HSP_MICROMODEM_563			0x7892 /*HSP MicroModem 56*/
#define DEVICE_PCTEL_HSP_MICROMODEM_564			0x7893 /*HSP MicroModem 56*/
#define DEVICE_PCTEL_HSP_MICROMODEM_565			0x7894 /*HSP MicroModem 56*/
#define DEVICE_PCTEL_HSP_MICROMODEM_566			0x7895 /*HSP MicroModem 56*/
#define DEVICE_PCTEL_HSP_MICROMODEM_567			0x7896 /*HSP MicroModem 56*/
#define DEVICE_PCTEL_HSP_MICROMODEM_568			0x7897 /*HSP MicroModem 56*/
/* Vendor: 134e: CSTI */
/* Vendor: 134f: Algo System Co Ltd */
/* Vendor: 1350: Systec Co. Ltd */
/* Vendor: 1351: Sonix Inc */
/* Vendor: 1353: Thales Idatys */
#define DEVICE_THALES_PROSERVER				0x0002 /*Proserver*/
#define DEVICE_THALES_PCI_FUT				0x0003 /*PCI-FUT*/
#define DEVICE_THALES_PCI_S0				0x0004 /*PCI-S0*/
#define DEVICE_THALES_PCI_FUT_S0			0x0005 /*PCI-FUT-S0*/
/* Vendor: 1354: Dwave System Inc */
/* Vendor: 1355: Kratos Analytical Ltd */
/* Vendor: 1356: The Logical Co */
/* Vendor: 1359: Prisa Networks */
/* Vendor: 135a: Brain Boxes */
/* Vendor: 135b: Giganet Inc */
/* Vendor: 135c: Quatech Inc */
#define DEVICE_QUATECH_QSC_100				0x0010 /*QSC-100*/
#define DEVICE_QUATECH_DSC_100				0x0020 /*DSC-100*/
#define DEVICE_QUATECH_DSC_200_300			0x0030 /*DSC-200/300*/
#define DEVICE_QUATECH_QSC_200_300			0x0040 /*QSC-200/300*/
#define DEVICE_QUATECH_ESC_100D				0x0050 /*ESC-100D*/
#define DEVICE_QUATECH_ESC_100M				0x0060 /*ESC-100M*/
#define DEVICE_QUATECH_MPAC_100_SYNCRONOUS_SERIAL	0x00f0 /*MPAC-100 Syncronous Serial Card (Zilog 85230)*/
#define DEVICE_QUATECH_QSCLP_100			0x0170 /*QSCLP-100*/
#define DEVICE_QUATECH_DSCLP_100			0x0180 /*DSCLP-100*/
#define DEVICE_QUATECH_SSCLP_100			0x0190 /*SSCLP-100*/
#define DEVICE_QUATECH_QSCLP_200_300			0x01a0 /*QSCLP-200/300*/
#define DEVICE_QUATECH_DSCLP_200_300			0x01b0 /*DSCLP-200/300*/
#define DEVICE_QUATECH_SSCLP_200_300			0x01c0 /*SSCLP-200/300*/
/* Vendor: 135d: ABB Network Partner AB */
/* Vendor: 135e: Sealevel Systems Inc */
#define DEVICE_SEALEVEL_ROUTE_56_PCI			0x5101 /*Route 56.PCI - Multi-Protocol Serial Interface (Zilog Z16C32)*/
#define DEVICE_SEALEVEL_SINGLE_PORT_RS_232_422_485_530	0x7101 /*Single Port RS-232/422/485/530*/
#define DEVICE_SEALEVEL_DUAL_PORT_RS_232_422_485	0x7201 /*Dual Port RS-232/422/485 Interface*/
#define DEVICE_SEALEVEL_DUAL_PORT_RS_232		0x7202 /*Dual Port RS-232 Interface*/
#define DEVICE_SEALEVEL_FOUR_PORT_RS_232		0x7401 /*Four Port RS-232 Interface*/
#define DEVICE_SEALEVEL_FOUR_PORT_RS_422_485		0x7402 /*Four Port RS-422/485 Interface*/
#define DEVICE_SEALEVEL_EIGHT_PORT_RS_232		0x7801 /*Eight Port RS-232 Interface*/
#define DEVICE_SEALEVEL_8001_DIGITAL_I_O		0x8001 /*8001 Digital I/O Adapter*/
/* Vendor: 135f: I-Data International A-S */
/* Vendor: 1360: Meinberg Funkuhren */
#define DEVICE_MEINBERG_PCI32_DCF77_RADIO		0x0101 /*PCI32 DCF77 Radio Clock*/
#define DEVICE_MEINBERG_PCI509_DCF77_RADIO		0x0102 /*PCI509 DCF77 Radio Clock*/
#define DEVICE_MEINBERG_PCI510_DCF77_RADIO		0x0103 /*PCI510 DCF77 Radio Clock*/
#define DEVICE_MEINBERG_GPS167PCI_GPS_RECEIVER		0x0201 /*GPS167PCI GPS Receiver*/
#define DEVICE_MEINBERG_GPS168PCI_GPS_RECEIVER		0x0202 /*GPS168PCI GPS Receiver*/
#define DEVICE_MEINBERG_GPS169PCI_GPS_RECEIVER		0x0203 /*GPS169PCI GPS Receiver*/
#define DEVICE_MEINBERG_TCR510PCI_IRIG_RECEIVER		0x0301 /*TCR510PCI IRIG Receiver*/
/* Vendor: 1361: Soliton Systems K.K. */
/* Vendor: 1362: Fujifacom Corporation */
/* Vendor: 1363: Phoenix Technology Ltd */
/* Vendor: 1364: ATM Communications Inc */
/* Vendor: 1365: Hypercope GmbH */
/* Vendor: 1366: Teijin Seiki Co. Ltd */
/* Vendor: 1367: Hitachi Zosen Corporation */
/* Vendor: 1368: Skyware Corporation */
/* Vendor: 1369: Digigram */
/* Vendor: 136a: High Soft Tech */
/* Vendor: 136b: Kawasaki Steel Corporation */
#define DEVICE_KAWASAKI_KL5A72002_MOTION_JPEG		0xff01 /*KL5A72002 Motion JPEG*/
/* Vendor: 136c: Adtek System Science Co Ltd */
/* Vendor: 136d: Gigalabs Inc */
/* Vendor: 136f: Applied Magic Inc */
/* Vendor: 1370: ATL Products */
/* Vendor: 1371: CNet Technology Inc */
#define DEVICE_CNET_GIGACARD_NETWORK_ADAPTER		0x434e /*GigaCard Network Adapter*/
	#define SUBDEVICE_CNET_N_WAY_PCI_BUS_GIGA_CARD	0x1371434e /*N-Way PCI-Bus Giga-Card 1000/100/10Mbps(L)*/
/* Vendor: 1373: Silicon Vision Inc */
/* Vendor: 1374: Silicom Ltd */
/* Vendor: 1375: Argosystems Inc */
/* Vendor: 1376: LMC */
/* Vendor: 1377: Electronic Equipment Production & Distribution GmbH */
/* Vendor: 1378: Telemann Co. Ltd */
/* Vendor: 1379: Asahi Kasei Microsystems Co Ltd */
/* Vendor: 137a: Mark of the Unicorn Inc */
#define DEVICE_MARK_PCI_324_AUDIOWIRE_INTERFACE		0x0001 /*PCI-324 Audiowire Interface*/
/* Vendor: 137b: PPT Vision */
/* Vendor: 137c: Iwatsu Electric Co Ltd */
/* Vendor: 137d: Dynachip Corporation */
/* Vendor: 137e: Patriot Scientific Corporation */
/* Vendor: 137f: Japan Satellite Systems Inc */
/* Vendor: 1380: Sanritz Automation Co Ltd */
/* Vendor: 1381: Brains Co. Ltd */
/* Vendor: 1382: Marian - Electronic & Software */
#define DEVICE_MARIAN_ARC88_AUDIO_RECORDING		0x0001 /*ARC88 audio recording card*/
#define DEVICE_MARIAN_MARC_8_MIDI_8			0x2088 /*Marc-8 MIDI 8 channel audio card*/
/* Vendor: 1383: Controlnet Inc */
/* Vendor: 1384: Reality Simulation Systems Inc */
/* Vendor: 1385: Netgear */
#define DEVICE_NETGEAR_WG311T				0x0013 /*WG311T*/
#define DEVICE_NETGEAR_802_11B_WIRELESS_ADAPTER		0x4100 /*802.11b Wireless Adapter (MA301)*/
#define DEVICE_NETGEAR_MA311_802_11B_WIRELESS		0x4105 /*MA311 802.11b wireless adapter*/
#define DEVICE_NETGEAR_WAG511_802_11A_B_G_DUAL		0x4400 /*WAG511 802.11a/b/g Dual Band Wireless PC Card*/
#define DEVICE_NETGEAR_WAG511_802_11A_B_G_DUAL2		0x4600 /*WAG511 802.11a/b/g Dual Band Wireless PC Card*/
#define DEVICE_NETGEAR_WAG511_802_11A_B_G_DUAL3		0x4601 /*WAG511 802.11a/b/g Dual Band Wireless PC Card*/
#define DEVICE_NETGEAR_WAG511_802_11A_B_G_DUAL4		0x4610 /*WAG511 802.11a/b/g Dual Band Wireless PC Card*/
#define DEVICE_NETGEAR_WAG311_802_11A_G_WIRELESS	0x4a00 /*WAG311 802.11a/g Wireless PCI Adapter*/
#define DEVICE_NETGEAR_WG311V2_54_MBPS			0x4c00 /*WG311v2 54 Mbps Wireless PCI Adapter*/
#define DEVICE_NETGEAR_GA620_GIGABIT_ETHERNET		0x620a /*GA620 Gigabit Ethernet*/
#define DEVICE_NETGEAR_GA622				0x622a /*GA622*/
#define DEVICE_NETGEAR_GA630_GIGABIT_ETHERNET		0x630a /*GA630 Gigabit Ethernet*/
#define DEVICE_NETGEAR_FA310TX				0xf004 /*FA310TX*/
/* Vendor: 1386: Video Domain Technologies */
/* Vendor: 1387: Systran Corp */
/* Vendor: 1388: Hitachi Information Technology Co Ltd */
/* Vendor: 1389: Applicom International */
#define DEVICE_APPLICOM_PCI1500PFB_INTELLIGENT_FIELDBUS	0x0001 /*PCI1500PFB [Intelligent fieldbus adaptor]*/
/* Vendor: 138a: Fusion Micromedia Corp */
/* Vendor: 138b: Tokimec Inc */
/* Vendor: 138c: Silicon Reality */
/* Vendor: 138d: Future Techno Designs pte Ltd */
/* Vendor: 138e: Basler GmbH */
/* Vendor: 138f: Patapsco Designs Inc */
/* Vendor: 1390: Concept Development Inc */
/* Vendor: 1391: Development Concepts Inc */
/* Vendor: 1392: Medialight Inc */
/* Vendor: 1393: Moxa Technologies Co Ltd */
#define DEVICE_MOXA_SMARTIO_C104H_PCI			0x1040 /*Smartio C104H/PCI*/
#define DEVICE_MOXA_INDUSTRIO_CP_114			0x1141 /*Industrio CP-114*/
#define DEVICE_MOXA_SMARTIO_C168H_PCI			0x1680 /*Smartio C168H/PCI*/
#define DEVICE_MOXA_INTELLIO_CP_204J			0x2040 /*Intellio CP-204J*/
#define DEVICE_MOXA_INTELLIO_C218_TURBO			0x2180 /*Intellio C218 Turbo PCI*/
#define DEVICE_MOXA_INTELLIO_C320_TURBO			0x3200 /*Intellio C320 Turbo PCI*/
/* Vendor: 1394: Level One Communications */
#define DEVICE_LEVEL_LXT1001_GIGABIT_ETHERNET		0x0001 /*LXT1001 Gigabit Ethernet*/
	#define SUBDEVICE_LEVEL_NETCELERATOR_ADAPTER	0x13940001 /*NetCelerator Adapter*/
/* Vendor: 1395: Ambicom Inc */
/* Vendor: 1396: Cipher Systems Inc */
/* Vendor: 1397: Cologne Chip Designs GmbH */
#define DEVICE_COLOGNE_ISDN_NETWORK_CONTROLLER		0x2bd0 /*ISDN network controller [HFC-PCI]*/
	#define SUBDEVICE_COLOGNE_ISDN_BOARD		0x13972bd0 /*ISDN Board*/
	#define SUBDEVICE_COLOGNE_CI1_1_HARP		0xe4bf1000 /*CI1-1-Harp*/
/* Vendor: 1398: Clarion co. Ltd */
/* Vendor: 1399: Rios systems Co Ltd */
/* Vendor: 139a: Alacritech Inc */
#define DEVICE_ALACRITECH_QUAD_PORT_10_100		0x0001 /*Quad Port 10/100 Server Accelerator*/
#define DEVICE_ALACRITECH_SINGLE_PORT_10_100		0x0003 /*Single Port 10/100 Server Accelerator*/
#define DEVICE_ALACRITECH_SINGLE_PORT_GIGABIT		0x0005 /*Single Port Gigabit Server Accelerator*/
/* Vendor: 139b: Mediasonic Multimedia Systems Ltd */
/* Vendor: 139c: Quantum 3d Inc */
/* Vendor: 139d: EPL limited */
/* Vendor: 139e: Media4 */
/* Vendor: 139f: Aethra s.r.l. */
/* Vendor: 13a0: Crystal Group Inc */
/* Vendor: 13a1: Kawasaki Heavy Industries Ltd */
/* Vendor: 13a2: Ositech Communications Inc */
/* Vendor: 13a3: Hifn Inc. */
#define DEVICE_HIFN_7751_SECURITY_PROCESSOR		0x0005 /*7751 Security Processor*/
#define DEVICE_HIFN_6500_PUBLIC_KEY			0x0006 /*6500 Public Key Processor*/
#define DEVICE_HIFN_7811_SECURITY_PROCESSOR		0x0007 /*7811 Security Processor*/
#define DEVICE_HIFN_7951_SECURITY_PROCESSOR		0x0012 /*7951 Security Processor*/
#define DEVICE_HIFN_78XX_SECURITY_PROCESSOR		0x0014 /*78XX Security Processor*/
#define DEVICE_HIFN_8065_SECURITY_PROCESSOR		0x0016 /*8065 Security Processor*/
#define DEVICE_HIFN_8165_SECURITY_PROCESSOR		0x0017 /*8165 Security Processor*/
#define DEVICE_HIFN_8154_SECURITY_PROCESSOR		0x0018 /*8154 Security Processor*/
/* Vendor: 13a4: Rascom Inc */
/* Vendor: 13a5: Audio Digital Imaging Inc */
/* Vendor: 13a6: Videonics Inc */
/* Vendor: 13a7: Teles AG */
/* Vendor: 13a8: Exar Corp. */
#define DEVICE_EXAR_XR17C154_QUAD_UART			0x0154 /*XR17C154 Quad UART*/
#define DEVICE_EXAR_XR17C158_OCTAL_UART			0x0158 /*XR17C158 Octal UART*/
/* Vendor: 13a9: Siemens Medical Systems, Ultrasound Group */
/* Vendor: 13aa: Broadband Networks Inc */
/* Vendor: 13ab: Arcom Control Systems Ltd */
/* Vendor: 13ac: Motion Media Technology Ltd */
/* Vendor: 13ad: Nexus Inc */
/* Vendor: 13ae: ALD Technology Ltd */
/* Vendor: 13af: T.Sqware */
/* Vendor: 13b0: Maxspeed Corp */
/* Vendor: 13b1: Tamura corporation */
/* Vendor: 13b2: Techno Chips Co. Ltd */
/* Vendor: 13b3: Lanart Corporation */
/* Vendor: 13b4: Wellbean Co Inc */
/* Vendor: 13b5: ARM */
/* Vendor: 13b6: Dlog GmbH */
/* Vendor: 13b7: Logic Devices Inc */
/* Vendor: 13b8: Nokia Telecommunications oy */
/* Vendor: 13b9: Elecom Co Ltd */
/* Vendor: 13ba: Oxford Instruments */
/* Vendor: 13bb: Sanyo Technosound Co Ltd */
/* Vendor: 13bc: Bitran Corporation */
/* Vendor: 13bd: Sharp corporation */
/* Vendor: 13be: Miroku Jyoho Service Co. Ltd */
/* Vendor: 13bf: Sharewave Inc */
/* Vendor: 13c0: Microgate Corporation */
#define DEVICE_MICROGATE_SYNCLINK_ADAPTER_V1		0x0010 /*SyncLink Adapter v1*/
#define DEVICE_MICROGATE_SYNCLINK_SCC_ADAPTER		0x0020 /*SyncLink SCC Adapter*/
#define DEVICE_MICROGATE_SYNCLINK_MULTIPORT_ADAPTER	0x0030 /*SyncLink Multiport Adapter*/
#define DEVICE_MICROGATE_SYNCLINK_ADAPTER_V2		0x0210 /*SyncLink Adapter v2*/
/* Vendor: 13c1: 3ware Inc */
#define DEVICE_3WARE_3WARE_INC_3WARE			0x1000 /*3ware Inc 3ware 5xxx/6xxx-series PATA-RAID*/
#define DEVICE_3WARE_3WARE_INC_3WARE2			0x1001 /*3ware Inc 3ware 7xxx/8xxx-series PATA/SATA-RAID*/
	#define SUBDEVICE_3WARE_3WARE_INC_3WARE		0x13c11001 /*3ware Inc 3ware 7xxx/8xxx-series PATA/SATA-RAID*/
#define DEVICE_3WARE_3WARE_INC_3WARE3			0x1002 /*3ware Inc 3ware 9xxx-series SATA-RAID*/
/* Vendor: 13c2: Technotrend Systemtechnik GmbH */
/* Vendor: 13c3: Janz Computer AG */
/* Vendor: 13c4: Phase Metrics */
/* Vendor: 13c5: Alphi Technology Corp */
/* Vendor: 13c6: Condor Engineering Inc */
#define DEVICE_CONDOR_CEI_520_A429_CARD			0x0520 /*CEI-520 A429 Card*/
#define DEVICE_CONDOR_CEI_620_A429_CARD			0x0620 /*CEI-620 A429 Card*/
#define DEVICE_CONDOR_CEI_820_A429_CARD			0x0820 /*CEI-820 A429 Card*/
/* Vendor: 13c7: Blue Chip Technology Ltd */
/* Vendor: 13c8: Apptech Inc */
/* Vendor: 13c9: Eaton Corporation */
/* Vendor: 13ca: Iomega Corporation */
/* Vendor: 13cb: Yano Electric Co Ltd */
/* Vendor: 13cc: Metheus Corporation */
/* Vendor: 13cd: Compatible Systems Corporation */
/* Vendor: 13ce: Cocom A/S */
/* Vendor: 13cf: Studio Audio & Video Ltd */
/* Vendor: 13d0: Techsan Electronics Co Ltd */
#define DEVICE_TECHSAN_B2C2_FLEXCOPII_DVB		0x2103 /*B2C2 FlexCopII DVB chip / Technisat SkyStar2 DVB card*/
#define DEVICE_TECHSAN_B2C2_FLEXCOPIII_DVB		0x2200 /*B2C2 FlexCopIII DVB chip / Technisat SkyStar2 DVB card*/
/* Vendor: 13d1: Abocom Systems Inc */
#define DEVICE_ABOCOM_ADMTEK_CENTAUR_C_REV		0xab02 /*ADMtek Centaur-C rev 17 [D-Link DFE-680TX] CardBus Fast Ethernet Adapter*/
#define DEVICE_ABOCOM_21X4X_DEC_TULIP_COMPATIBLE	0xab03 /*21x4x DEC-Tulip compatible 10/100 Ethernet*/
#define DEVICE_ABOCOM_RTL8139_FE2000VX_CARDBUS		0xab06 /*RTL8139 [FE2000VX] CardBus Fast Ethernet Attached Port Adapter*/
#define DEVICE_ABOCOM_21X4X_DEC_TULIP_COMPATIBLE2	0xab08 /*21x4x DEC-Tulip compatible 10/100 Ethernet*/
/* Vendor: 13d2: Shark Multimedia Inc */
/* Vendor: 13d3: IMC Networks */
/* Vendor: 13d4: Graphics Microsystems Inc */
/* Vendor: 13d5: Media 100 Inc */
/* Vendor: 13d6: K.I. Technology Co Ltd */
/* Vendor: 13d7: Toshiba Engineering Corporation */
/* Vendor: 13d8: Phobos corporation */
/* Vendor: 13d9: Apex PC Solutions Inc */
/* Vendor: 13da: Intresource Systems pte Ltd */
/* Vendor: 13db: Janich & Klass Computertechnik GmbH */
/* Vendor: 13dc: Netboost Corporation */
/* Vendor: 13dd: Multimedia Bundle Inc */
/* Vendor: 13de: ABB Robotics Products AB */
/* Vendor: 13df: E-Tech Inc */
#define DEVICE_E_TECH_PCI56RVP_MODEM			0x0001 /*PCI56RVP Modem*/
	#define SUBDEVICE_E_TECH_PCI56RVP_MODEM		0x13df0001 /*PCI56RVP Modem*/
/* Vendor: 13e0: GVC Corporation */
/* Vendor: 13e1: Silicom Multimedia Systems Inc */
/* Vendor: 13e2: Dynamics Research Corporation */
/* Vendor: 13e3: Nest Inc */
/* Vendor: 13e4: Calculex Inc */
/* Vendor: 13e5: Telesoft Design Ltd */
/* Vendor: 13e6: Argosy research Inc */
/* Vendor: 13e7: NAC Incorporated */
/* Vendor: 13e8: Chip Express Corporation */
/* Vendor: 13e9: Intraserver Technology Inc */
/* Vendor: 13ea: Dallas Semiconductor */
/* Vendor: 13eb: Hauppauge Computer Works Inc */
/* Vendor: 13ec: Zydacron Inc */
/* Vendor: 13ed: Raytheion E-Systems */
/* Vendor: 13ee: Hayes Microcomputer Products Inc */
/* Vendor: 13ef: Coppercom Inc */
/* Vendor: 13f0: Sundance Technology Inc */
#define DEVICE_SUNDANCE_ST201_SUNDANCE_ETHERNET		0x0201 /*ST201 Sundance Ethernet*/
/* Vendor: 13f1: Oce' - Technologies B.V. */
/* Vendor: 13f2: Ford Microelectronics Inc */
/* Vendor: 13f3: Mcdata Corporation */
/* Vendor: 13f4: Troika Networks, Inc. */
#define DEVICE_TROIKA_ZENTAI_FIBRE_CHANNEL		0x1401 /*Zentai Fibre Channel Adapter*/
/* Vendor: 13f5: Kansai Electric Co. Ltd */
/* Vendor: 13f6: C-Media Electronics Inc */
#define DEVICE_C_MEDIA_CMI8738				0x0011 /*CMI8738*/
#define DEVICE_C_MEDIA_CM8338A				0x0100 /*CM8338A*/
	#define SUBDEVICE_C_MEDIA_CMI8338_C3DX_PCI_AUDIO	0x13f6ffff /*CMI8338/C3DX PCI Audio Device*/
#define DEVICE_C_MEDIA_CM8338B				0x0101 /*CM8338B*/
	#define SUBDEVICE_C_MEDIA_CMI8338_031_PCI_AUDIO	0x13f60101 /*CMI8338-031 PCI Audio Device*/
#define DEVICE_C_MEDIA_CM8738				0x0111 /*CM8738*/
	#define SUBDEVICE_C_MEDIA_P6STP_FL_MOTHERBOARD	0x10190970 /*P6STP-FL motherboard*/
	#define SUBDEVICE_C_MEDIA_CUSI_FX_MOTHERBOARD	0x10438035 /*CUSI-FX motherboard*/
	#define SUBDEVICE_C_MEDIA_CMI8738_6_CHANNEL_AUDIO	0x10438077 /*CMI8738 6-channel audio controller*/
	#define SUBDEVICE_C_MEDIA_CMI8738_6CH_MX	0x104380e2 /*CMI8738 6ch-MX*/
	#define SUBDEVICE_C_MEDIA_CMI8738_C3DX_PCI_AUDIO	0x13f60111 /*CMI8738/C3DX PCI Audio Device*/
	#define SUBDEVICE_C_MEDIA_GAMESURROUND_MUSE_XL	0x1681a000 /*Gamesurround MUSE XL*/
#define DEVICE_C_MEDIA_CM87382				0x0211 /*CM8738*/
/* Vendor: 13f7: Wildfire Communications */
/* Vendor: 13f8: Ad Lib Multimedia Inc */
/* Vendor: 13f9: NTT Advanced Technology Corp. */
/* Vendor: 13fa: Pentland Systems Ltd */
/* Vendor: 13fb: Aydin Corp */
/* Vendor: 13fc: Computer Peripherals International */
/* Vendor: 13fd: Micro Science Inc */
/* Vendor: 13fe: Advantech Co. Ltd */
#define DEVICE_ADVANTECH_PCI_1240_4_CHANNEL_STEPPER	0x1240 /*PCI-1240 4-channel stepper motor controller card w.  Nova Electronics MCX314*/
#define DEVICE_ADVANTECH_PCI_1612_4_PORT_RS_232_422_485	0x1600 /*PCI-1612 4-port RS-232/422/485 PCI Communication Card*/
#define DEVICE_ADVANTECH_PCI_1752			0x1752 /*PCI-1752*/
#define DEVICE_ADVANTECH_PCI_1754			0x1754 /*PCI-1754*/
#define DEVICE_ADVANTECH_PCI_1756			0x1756 /*PCI-1756*/
/* Vendor: 13ff: Silicon Spice Inc */
/* Vendor: 1400: Artx Inc */
#define DEVICE_ARTX_9432_TX				0x1401 /*9432 TX*/
/* Vendor: 1401: CR-Systems A/S */
/* Vendor: 1402: Meilhaus Electronic GmbH */
/* Vendor: 1403: Ascor Inc */
/* Vendor: 1404: Fundamental Software Inc */
/* Vendor: 1405: Excalibur Systems Inc */
/* Vendor: 1406: Oce' Printing Systems GmbH */
/* Vendor: 1407: Lava Computer mfg Inc */
#define DEVICE_LAVA_LAVA_DUAL_SERIAL			0x0100 /*Lava Dual Serial*/
#define DEVICE_LAVA_LAVA_QUATRO_A			0x0101 /*Lava Quatro A*/
#define DEVICE_LAVA_LAVA_QUATRO_B			0x0102 /*Lava Quatro B*/
#define DEVICE_LAVA_QUATTRO_PCI_A			0x0120 /*Quattro-PCI A*/
#define DEVICE_LAVA_QUATTRO_PCI_B			0x0121 /*Quattro-PCI B*/
#define DEVICE_LAVA_LAVA_OCTO_A				0x0180 /*Lava Octo A*/
#define DEVICE_LAVA_LAVA_OCTO_B				0x0181 /*Lava Octo B*/
#define DEVICE_LAVA_LAVA_PORT_PLUS			0x0200 /*Lava Port Plus*/
#define DEVICE_LAVA_LAVA_QUAD_A				0x0201 /*Lava Quad A*/
#define DEVICE_LAVA_LAVA_QUAD_B				0x0202 /*Lava Quad B*/
#define DEVICE_LAVA_LAVA_QUATTRO_PCI			0x0220 /*Lava Quattro PCI Ports A/B*/
#define DEVICE_LAVA_LAVA_QUATTRO_PCI2			0x0221 /*Lava Quattro PCI Ports C/D*/
#define DEVICE_LAVA_LAVA_SINGLE_SERIAL			0x0500 /*Lava Single Serial*/
#define DEVICE_LAVA_LAVA_PORT_650			0x0600 /*Lava Port 650*/
#define DEVICE_LAVA_LAVA_PARALLEL			0x8000 /*Lava Parallel*/
#define DEVICE_LAVA_DUAL_PARALLEL_PORT			0x8001 /*Dual parallel port controller A*/
#define DEVICE_LAVA_LAVA_DUAL_PARALLEL			0x8002 /*Lava Dual Parallel port A*/
#define DEVICE_LAVA_LAVA_DUAL_PARALLEL2			0x8003 /*Lava Dual Parallel port B*/
#define DEVICE_LAVA_BOCA_RESEARCH_IOPPAR		0x8800 /*BOCA Research IOPPAR*/
/* Vendor: 1408: Aloka Co. Ltd */
/* Vendor: 1409: Timedia Technology Co Ltd */
#define DEVICE_TIMEDIA_PCI2S550_DUAL_16550		0x7168 /*PCI2S550 (Dual 16550 UART)*/
/* Vendor: 140a: DSP Research Inc */
/* Vendor: 140b: Ramix Inc */
/* Vendor: 140c: Elmic Systems Inc */
/* Vendor: 140d: Matsushita Electric Works Ltd */
/* Vendor: 140e: Goepel Electronic GmbH */
/* Vendor: 140f: Salient Systems Corp */
/* Vendor: 1410: Midas lab Inc */
/* Vendor: 1411: Ikos Systems Inc */
/* Vendor: 1412: VIA Technologies Inc. */
#define DEVICE_VIA3_ICE1712_ENVY24_PCI			0x1712 /*ICE1712 [Envy24] PCI Multi-Channel I/O Controller*/
	#define SUBDEVICE_VIA3_M_AUDIO_DELTA_410	0x1412d638 /*M-Audio Delta 410*/
#define DEVICE_VIA3_VT1720_24_ENVY24PT_HT_PCI		0x1724 /*VT1720/24 [Envy24PT/HT] PCI Multi-Channel Audio Controller*/
/* Vendor: 1413: Addonics */
/* Vendor: 1414: Microsoft Corporation */
/* Vendor: 1415: Oxford Semiconductor Ltd */
#define DEVICE_OXFORD2_VSCOM_011H_EP1_1			0x8403 /*VScom 011H-EP1 1 port parallel adaptor*/
#define DEVICE_OXFORD2_OX16PCI954_QUAD_16950		0x9501 /*OX16PCI954 (Quad 16950 UART) function 0*/
	#define SUBDEVICE_OXFORD2_CYBERPRO_4_PORT	0x131f2050 /*CyberPro (4-port)*/
	#define SUBDEVICE_OXFORD2_MCCR_SERIAL_P0_3	0x15ed2000 /*MCCR Serial p0-3 of 8*/
	#define SUBDEVICE_OXFORD2_MCCR_SERIAL_P0_32	0x15ed2001 /*MCCR Serial p0-3 of 16*/
#define DEVICE_OXFORD2_EXSYS_EX_41092_DUAL		0x950a /*EXSYS EX-41092 Dual 16950 Serial adapter*/
#define DEVICE_OXFORD2_OXCB950_CARDBUS_16950		0x950b /*OXCB950 Cardbus 16950 UART*/
#define DEVICE_OXFORD2_OX16PCI954_QUAD_169502		0x9511 /*OX16PCI954 (Quad 16950 UART) function 1*/
	#define SUBDEVICE_OXFORD2_MCCR_SERIAL_P4_7	0x15ed2000 /*MCCR Serial p4-7 of 8*/
	#define SUBDEVICE_OXFORD2_MCCR_SERIAL_P4_15	0x15ed2001 /*MCCR Serial p4-15 of 16*/
#define DEVICE_OXFORD2_OX16PCI952_DUAL_16950		0x9521 /*OX16PCI952 (Dual 16950 UART)*/
/* Vendor: 1416: Multiwave Innovation pte Ltd */
/* Vendor: 1417: Convergenet Technologies Inc */
/* Vendor: 1418: Kyushu electronics systems Inc */
/* Vendor: 1419: Excel Switching Corp */
/* Vendor: 141a: Apache Micro Peripherals Inc */
/* Vendor: 141b: Zoom Telephonics Inc */
/* Vendor: 141d: Digitan Systems Inc */
/* Vendor: 141e: Fanuc Ltd */
/* Vendor: 141f: Visiontech Ltd */
/* Vendor: 1420: Psion Dacom plc */
#define DEVICE_PSION_GOLD_CARD_NETGLOBAL		0x8002 /*Gold Card NetGlobal 56k+10/100Mb CardBus (Ethernet part)*/
#define DEVICE_PSION_GOLD_CARD_NETGLOBAL2		0x8003 /*Gold Card NetGlobal 56k+10/100Mb CardBus (Modem part)*/
/* Vendor: 1421: Ads Technologies Inc */
/* Vendor: 1422: Ygrec Systems Co Ltd */
/* Vendor: 1423: Custom Technology Corp. */
/* Vendor: 1424: Videoserver Connections */
/* Vendor: 1425: Chelsio Communications Inc */
/* Vendor: 1426: Storage Technology Corp. */
/* Vendor: 1427: Better On-Line Solutions */
/* Vendor: 1428: Edec Co Ltd */
/* Vendor: 1429: Unex Technology Corp. */
/* Vendor: 142a: Kingmax Technology Inc */
/* Vendor: 142b: Radiolan */
/* Vendor: 142c: Minton Optic Industry Co Ltd */
/* Vendor: 142d: Pix stream Inc */
/* Vendor: 142e: Vitec Multimedia */
#define DEVICE_VITEC_VM2_2_VIDEO_MAKER			0x4020 /*VM2-2 [Video Maker 2] MPEG1/2 Encoder*/
/* Vendor: 142f: Radicom Research Inc */
/* Vendor: 1430: ITT Aerospace/Communications Division */
/* Vendor: 1431: Gilat Satellite Networks */
/* Vendor: 1432: Edimax Computer Co. */
#define DEVICE_EDIMAX_RTL81XX_FAST_ETHERNET		0x9130 /*RTL81xx Fast Ethernet*/
/* Vendor: 1433: Eltec Elektronik GmbH */
/* Vendor: 1435: Real Time Devices US Inc. */
/* Vendor: 1436: CIS Technology Inc */
/* Vendor: 1437: Nissin Inc Co */
/* Vendor: 1438: Atmel-dream */
/* Vendor: 1439: Outsource Engineering & Mfg. Inc */
/* Vendor: 143a: Stargate Solutions Inc */
/* Vendor: 143b: Canon Research Center, America */
/* Vendor: 143c: Amlogic Inc */
/* Vendor: 143d: Tamarack Microelectronics Inc */
/* Vendor: 143e: Jones Futurex Inc */
/* Vendor: 143f: Lightwell Co Ltd - Zax Division */
/* Vendor: 1440: ALGOL Corp. */
/* Vendor: 1441: AGIE Ltd */
/* Vendor: 1442: Phoenix Contact GmbH & Co. */
/* Vendor: 1443: Unibrain S.A. */
/* Vendor: 1444: TRW */
/* Vendor: 1445: Logical DO Ltd */
/* Vendor: 1446: Graphin Co Ltd */
/* Vendor: 1447: AIM GmBH */
/* Vendor: 1448: Alesis Studio Electronics */
/* Vendor: 1449: TUT Systems Inc */
/* Vendor: 144a: Adlink Technology */
#define DEVICE_ADLINK_PCI_7296				0x7296 /*PCI-7296*/
#define DEVICE_ADLINK_PCI_7432				0x7432 /*PCI-7432*/
#define DEVICE_ADLINK_PCI_7433				0x7433 /*PCI-7433*/
#define DEVICE_ADLINK_PCI_7434				0x7434 /*PCI-7434*/
#define DEVICE_ADLINK_PCI_7841				0x7841 /*PCI-7841*/
#define DEVICE_ADLINK_PCI_8133				0x8133 /*PCI-8133*/
#define DEVICE_ADLINK_PCI_8164				0x8164 /*PCI-8164*/
#define DEVICE_ADLINK_PCI_8554				0x8554 /*PCI-8554*/
#define DEVICE_ADLINK_PCI_9111				0x9111 /*PCI-9111*/
#define DEVICE_ADLINK_PCI_9113				0x9113 /*PCI-9113*/
#define DEVICE_ADLINK_PCI_9114				0x9114 /*PCI-9114*/
/* Vendor: 144b: Loronix Information Systems Inc */
/* Vendor: 144c: Catalina Research Inc */
/* Vendor: 144d: Samsung Electronics Co Ltd */
/* Vendor: 144e: OLITEC */
/* Vendor: 144f: Askey Computer Corp. */
/* Vendor: 1450: Octave Communications Ind. */
/* Vendor: 1451: SP3D Chip Design GmBH */
/* Vendor: 1453: MYCOM Inc */
/* Vendor: 1454: Altiga Networks */
/* Vendor: 1455: Logic Plus Plus Inc */
/* Vendor: 1456: Advanced Hardware Architectures */
/* Vendor: 1457: Nuera Communications Inc */
/* Vendor: 1458: Giga-byte Technology */
/* Vendor: 1459: DOOIN Electronics */
/* Vendor: 145a: Escalate Networks Inc */
/* Vendor: 145b: PRAIM SRL */
/* Vendor: 145c: Cryptek */
/* Vendor: 145d: Gallant Computer Inc */
/* Vendor: 145e: Aashima Technology B.V. */
/* Vendor: 145f: Baldor Electric Company */
#define DEVICE_BALDOR_NEXTMOVE_PCI			0x0001 /*NextMove PCI*/
/* Vendor: 1460: DYNARC INC */
/* Vendor: 1461: Avermedia Technologies Inc */
/* Vendor: 1462: Micro-Star International Co., Ltd. */
#define DEVICE_MICRO_STAR_PCI_CARD_WIRELESS		0x6825 /*PCI Card wireless 11g [PC54G]*/
#define DEVICE_MICRO_STAR_NVIDIA_NV25_GEFORCE4		0x8725 /*NVIDIA NV25 [GeForce4 Ti 4600] VGA Adapter*/
#define DEVICE_MICRO_STAR_NVIDIA_NV28_GEFORCE4		0x9000 /*NVIDIA NV28 [GeForce4 Ti 4800] VGA Adapter*/
#define DEVICE_MICRO_STAR_NVIDIA_NV31_GEFORCE		0x9119 /*NVIDIA NV31 [GeForce FX 5600XT] VGA Adapter*/
/* Vendor: 1463: Fast Corporation */
/* Vendor: 1464: Interactive Circuits & Systems Ltd */
/* Vendor: 1465: GN NETTEST Telecom DIV. */
/* Vendor: 1466: Designpro Inc. */
/* Vendor: 1467: DIGICOM SPA */
/* Vendor: 1468: AMBIT Microsystem Corp. */
/* Vendor: 1469: Cleveland Motion Controls */
/* Vendor: 146a: IFR */
/* Vendor: 146b: Parascan Technologies Ltd */
/* Vendor: 146c: Ruby Tech Corp. */
#define DEVICE_RUBY_FE_1430TX_FAST_ETHERNET		0x1430 /*FE-1430TX Fast Ethernet PCI Adapter*/
/* Vendor: 146d: Tachyon, INC. */
/* Vendor: 146e: Williams Electronics Games, Inc. */
/* Vendor: 146f: Multi Dimensional Consulting Inc */
/* Vendor: 1470: Bay Networks */
/* Vendor: 1471: Integrated Telecom Express Inc */
/* Vendor: 1472: DAIKIN Industries, Ltd */
/* Vendor: 1473: ZAPEX Technologies Inc */
/* Vendor: 1474: Doug Carson & Associates */
/* Vendor: 1475: PICAZO Communications */
/* Vendor: 1476: MORTARA Instrument Inc */
/* Vendor: 1477: Net Insight */
/* Vendor: 1478: DIATREND Corporation */
/* Vendor: 1479: TORAY Industries Inc */
/* Vendor: 147a: FORMOSA Industrial Computing */
/* Vendor: 147b: ABIT Computer Corp. */
/* Vendor: 147c: AWARE, Inc. */
/* Vendor: 147d: Interworks Computer Products */
/* Vendor: 147e: Matsushita Graphic Communication Systems, Inc. */
/* Vendor: 147f: NIHON UNISYS, Ltd. */
/* Vendor: 1480: SCII Telecom */
/* Vendor: 1481: BIOPAC Systems Inc */
/* Vendor: 1482: ISYTEC - Integrierte Systemtechnik GmBH */
/* Vendor: 1483: LABWAY Corporation */
/* Vendor: 1484: Logic Corporation */
/* Vendor: 1485: ERMA - Electronic GmBH */
/* Vendor: 1486: L3 Communications Telemetry & Instrumentation */
/* Vendor: 1487: MARQUETTE Medical Systems */
/* Vendor: 1488: KONTRON Electronik GmBH */
/* Vendor: 1489: KYE Systems Corporation */
/* Vendor: 148a: OPTO */
/* Vendor: 148b: INNOMEDIALOGIC Inc. */
/* Vendor: 148c: C.P. Technology Co. Ltd */
/* Vendor: 148d: DIGICOM Systems, Inc. */
#define DEVICE_DIGICOM3_HCF_56K_DATA_FAX		0x1003 /*HCF 56k Data/Fax Modem*/
/* Vendor: 148e: OSI Plus Corporation */
/* Vendor: 148f: Plant Equipment, Inc. */
/* Vendor: 1490: Stone Microsystems PTY Ltd. */
/* Vendor: 1491: ZEAL Corporation */
/* Vendor: 1492: Time Logic Corporation */
/* Vendor: 1493: MAKER Communications */
/* Vendor: 1494: WINTOP Technology, Inc. */
/* Vendor: 1495: TOKAI Communications Industry Co. Ltd */
/* Vendor: 1496: JOYTECH Computer Co., Ltd. */
/* Vendor: 1497: SMA Regelsysteme GmBH */
/* Vendor: 1498: TEWS Datentechnik GmBH */
#define DEVICE_TEWS_TPCI200				0x30c8 /*TPCI200*/
/* Vendor: 1499: EMTEC CO., Ltd */
/* Vendor: 149a: ANDOR Technology Ltd */
/* Vendor: 149b: SEIKO Instruments Inc */
/* Vendor: 149c: OVISLINK Corp. */
/* Vendor: 149d: NEWTEK Inc */
#define DEVICE_NEWTEK_VIDEO_TOASTER_FOR			0x0001 /*Video Toaster for PC*/
/* Vendor: 149e: Mapletree Networks Inc. */
/* Vendor: 149f: LECTRON Co Ltd */
/* Vendor: 14a0: SOFTING GmBH */
/* Vendor: 14a1: Systembase Co Ltd */
/* Vendor: 14a2: Millennium Engineering Inc */
/* Vendor: 14a3: Maverick Networks */
/* Vendor: 14a4: GVC/BCM Advanced Research */
/* Vendor: 14a5: XIONICS Document Technologies Inc */
/* Vendor: 14a6: INOVA Computers GmBH & Co KG */
/* Vendor: 14a7: MYTHOS Systems Inc */
/* Vendor: 14a8: FEATRON Technologies Corporation */
/* Vendor: 14a9: HIVERTEC Inc */
/* Vendor: 14aa: Advanced MOS Technology Inc */
/* Vendor: 14ab: Mentor Graphics Corp. */
/* Vendor: 14ac: Novaweb Technologies Inc */
/* Vendor: 14ad: Time Space Radio AB */
/* Vendor: 14ae: CTI, Inc */
/* Vendor: 14af: Guillemot Corporation */
#define DEVICE_GUILLEMOT_3D_PROPHET_II			0x7102 /*3D Prophet II MX*/
/* Vendor: 14b0: BST Communication Technology Ltd */
/* Vendor: 14b1: Nextcom K.K. */
/* Vendor: 14b2: ENNOVATE Networks Inc */
/* Vendor: 14b3: XPEED Inc */
#define DEVICE_XPEED_DSL_NIC				0x0000 /*DSL NIC*/
/* Vendor: 14b4: PHILIPS Business Electronics B.V. */
/* Vendor: 14b5: Creamware GmBH */
#define DEVICE_CREAMWARE_SCOPE				0x0200 /*Scope*/
#define DEVICE_CREAMWARE_PULSAR				0x0300 /*Pulsar*/
#define DEVICE_CREAMWARE_PULSARSRB			0x0400 /*PulsarSRB*/
#define DEVICE_CREAMWARE_PULSAR2			0x0600 /*Pulsar2*/
#define DEVICE_CREAMWARE_DSP_BOARD			0x0800 /*DSP-Board*/
#define DEVICE_CREAMWARE_DSP_BOARD2			0x0900 /*DSP-Board*/
#define DEVICE_CREAMWARE_DSP_BOARD3			0x0a00 /*DSP-Board*/
#define DEVICE_CREAMWARE_DSP_BOARD4			0x0b00 /*DSP-Board*/
/* Vendor: 14b6: Quantum Data Corp. */
/* Vendor: 14b7: PROXIM Inc */
#define DEVICE_PROXIM_SYMPHONY_4110			0x0001 /*Symphony 4110*/
/* Vendor: 14b8: Techsoft Technology Co Ltd */
/* Vendor: 14b9: AIRONET Wireless Communications */
#define DEVICE_AIRONET_PC4800				0x0001 /*PC4800*/
#define DEVICE_AIRONET_PC48002				0x0340 /*PC4800*/
#define DEVICE_AIRONET_PC48003				0x0350 /*PC4800*/
#define DEVICE_AIRONET_PC4500				0x4500 /*PC4500*/
#define DEVICE_AIRONET_CISCO_AIRONET_340		0x4800 /*Cisco Aironet 340 802.11b Wireless LAN Adapter/Aironet PC4800*/
#define DEVICE_AIRONET_CISCO_AIRONET_WIRELESS		0xa504 /*Cisco Aironet Wireless 802.11b*/
#define DEVICE_AIRONET_CISCO_AIRONET_CB20A		0xa505 /*Cisco Aironet CB20a 802.11a Wireless LAN Adapter*/
#define DEVICE_AIRONET_CISCO_AIRONET_MINI		0xa506 /*Cisco Aironet Mini PCI b/g*/
/* Vendor: 14ba: INTERNIX Inc. */
/* Vendor: 14bb: SEMTECH Corporation */
/* Vendor: 14bc: Globespan Semiconductor Inc. */
/* Vendor: 14bd: CARDIO Control N.V. */
/* Vendor: 14be: L3 Communications */
/* Vendor: 14bf: SPIDER Communications Inc. */
/* Vendor: 14c0: COMPAL Electronics Inc */
/* Vendor: 14c1: MYRICOM Inc. */
#define DEVICE_MYRICOM_MYRINET_2000_SCALABLE		0x8043 /*Myrinet 2000 Scalable Cluster Interconnect*/
/* Vendor: 14c2: DTK Computer */
/* Vendor: 14c3: MEDIATEK Corp. */
/* Vendor: 14c4: IWASAKI Information Systems Co Ltd */
/* Vendor: 14c5: Automation Products AB */
/* Vendor: 14c6: Data Race Inc */
/* Vendor: 14c7: Modular Technology Holdings Ltd */
/* Vendor: 14c8: Turbocomm Tech. Inc. */
/* Vendor: 14c9: ODIN Telesystems Inc */
/* Vendor: 14ca: PE Logic Corp. */
/* Vendor: 14cb: Billionton Systems Inc */
/* Vendor: 14cc: NAKAYO Telecommunications Inc */
/* Vendor: 14cd: Universal Scientific Ind. */
/* Vendor: 14ce: Whistle Communications */
/* Vendor: 14cf: TEK Microsystems Inc. */
/* Vendor: 14d0: Ericsson Axe R & D */
/* Vendor: 14d1: Computer Hi-Tech Co Ltd */
/* Vendor: 14d2: Titan Electronics Inc */
#define DEVICE_TITAN_VSCOM_010L_1			0x8001 /*VScom 010L 1 port parallel adaptor*/
#define DEVICE_TITAN_VSCOM_020L_2			0x8002 /*VScom 020L 2 port parallel adaptor*/
#define DEVICE_TITAN_VSCOM_100L_1			0x8010 /*VScom 100L 1 port serial adaptor*/
#define DEVICE_TITAN_VSCOM_110L_1			0x8011 /*VScom 110L 1 port serial and 1 port parallel adaptor*/
#define DEVICE_TITAN_VSCOM_200L_1			0x8020 /*VScom 200L 1 port serial adaptor*/
#define DEVICE_TITAN_VSCOM_210L_2			0x8021 /*VScom 210L 2 port serial and 1 port parallel adaptor*/
#define DEVICE_TITAN_VSCOM_400L_4			0x8040 /*VScom 400L 4 port serial adaptor*/
#define DEVICE_TITAN_VSCOM_800L_8			0x8080 /*VScom 800L 8 port serial adaptor*/
#define DEVICE_TITAN_VSCOM_010H_1			0xa000 /*VScom 010H 1 port parallel adaptor*/
#define DEVICE_TITAN_VSCOM_100H_1			0xa001 /*VScom 100H 1 port serial adaptor*/
#define DEVICE_TITAN_VSCOM_400H_4			0xa003 /*VScom 400H 4 port serial adaptor*/
#define DEVICE_TITAN_VSCOM_400HF1_4			0xa004 /*VScom 400HF1 4 port serial adaptor*/
#define DEVICE_TITAN_VSCOM_200H_2			0xa005 /*VScom 200H 2 port serial adaptor*/
#define DEVICE_TITAN_VSCOM_010HV2_1			0xe001 /*VScom 010HV2 1 port parallel adaptor*/
#define DEVICE_TITAN_VSCOM_100HV2_1			0xe010 /*VScom 100HV2 1 port serial adaptor*/
#define DEVICE_TITAN_VSCOM_200HV2_2			0xe020 /*VScom 200HV2 2 port serial adaptor*/
/* Vendor: 14d3: CIRTECH (UK) Ltd */
/* Vendor: 14d4: Panacom Technology Corp */
/* Vendor: 14d5: Nitsuko Corporation */
/* Vendor: 14d6: Accusys Inc */
/* Vendor: 14d7: Hirakawa Hewtech Corp */
/* Vendor: 14d8: HOPF Elektronik GmBH */
/* Vendor: 14d9: Alliance Semiconductor Corporation */
#define DEVICE_ALLIANCE2_AP1011_SP1011_HYPERTRANSPORT_PCI_BRIDGE	0x0010 /*AP1011/SP1011 HyperTransport-PCI Bridge [Sturgeon]*/
#define DEVICE_ALLIANCE2_AS90L10204_10208_HYPERTRANSPORT_TO	0x9000 /*AS90L10204/10208 HyperTransport to PCI-X Bridge*/
/* Vendor: 14da: National Aerospace Laboratories */
/* Vendor: 14db: AFAVLAB Technology Inc */
#define DEVICE_AFAVLAB_TK9902				0x2120 /*TK9902*/
/* Vendor: 14dc: Amplicon Liveline Ltd */
#define DEVICE_AMPLICON_PCI230				0x0000 /*PCI230*/
#define DEVICE_AMPLICON_PCI242				0x0001 /*PCI242*/
#define DEVICE_AMPLICON_PCI244				0x0002 /*PCI244*/
#define DEVICE_AMPLICON_PCI247				0x0003 /*PCI247*/
#define DEVICE_AMPLICON_PCI248				0x0004 /*PCI248*/
#define DEVICE_AMPLICON_PCI249				0x0005 /*PCI249*/
#define DEVICE_AMPLICON_PCI260				0x0006 /*PCI260*/
#define DEVICE_AMPLICON_PCI224				0x0007 /*PCI224*/
#define DEVICE_AMPLICON_PCI234				0x0008 /*PCI234*/
#define DEVICE_AMPLICON_PCI236				0x0009 /*PCI236*/
#define DEVICE_AMPLICON_PCI272				0x000a /*PCI272*/
#define DEVICE_AMPLICON_PCI215				0x000b /*PCI215*/
/* Vendor: 14dd: Boulder Design Labs Inc */
/* Vendor: 14de: Applied Integration Corporation */
/* Vendor: 14df: ASIC Communications Corp */
/* Vendor: 14e1: INVERTEX */
/* Vendor: 14e2: INFOLIBRIA */
/* Vendor: 14e3: AMTELCO */
/* Vendor: 14e4: Broadcom Corporation */
#define DEVICE_BROADCOM_SENTRY5_CHIPCOMMON_I_O		0x0800 /*Sentry5 Chipcommon I/O Controller*/
#define DEVICE_BROADCOM_SENTRY5_PCI_BRIDGE		0x0804 /*Sentry5 PCI Bridge*/
#define DEVICE_BROADCOM_SENTRY5_MIPS32_CPU		0x0805 /*Sentry5 MIPS32 CPU*/
#define DEVICE_BROADCOM_SENTRY5_ETHERNET_CONTROLLER	0x0806 /*Sentry5 Ethernet Controller*/
#define DEVICE_BROADCOM_SENTRY5_CRYPTO_ACCELERATOR	0x080b /*Sentry5 Crypto Accelerator*/
#define DEVICE_BROADCOM_SENTRY5_DDR_SDR_RAM		0x080f /*Sentry5 DDR/SDR RAM Controller*/
#define DEVICE_BROADCOM_SENTRY5_EXTERNAL_INTERFACE	0x0811 /*Sentry5 External Interface Core*/
#define DEVICE_BROADCOM_BCM3302_SENTRY5_MIPS32		0x0816 /*BCM3302 Sentry5 MIPS32 CPU*/
#define DEVICE_BROADCOM_NETXTREME_BCM5700_GIGABIT	0x1644 /*NetXtreme BCM5700 Gigabit Ethernet*/
	#define SUBDEVICE_BROADCOM_BROADCOM_VIGIL_B5700	0x10140277 /*Broadcom Vigil B5700 1000Base-T*/
	#define SUBDEVICE_BROADCOM_BROADCOM_BCM5700	0x102800d1 /*Broadcom BCM5700*/
	#define SUBDEVICE_BROADCOM_BROADCOM_BCM57002	0x10280106 /*Broadcom BCM5700*/
	#define SUBDEVICE_BROADCOM_BROADCOM_BCM5700_1000BASE_T	0x10280109 /*Broadcom BCM5700 1000Base-T*/
	#define SUBDEVICE_BROADCOM_BROADCOM_BCM5700_1000BASETX	0x1028010a /*Broadcom BCM5700 1000BaseTX*/
	#define SUBDEVICE_BROADCOM_3C996_T_1000BASE_T	0x10b71000 /*3C996-T 1000Base-T*/
	#define SUBDEVICE_BROADCOM_3C996B_T_1000BASE_T	0x10b71001 /*3C996B-T 1000Base-T*/
	#define SUBDEVICE_BROADCOM_3C996C_T_1000BASE_T	0x10b71002 /*3C996C-T 1000Base-T*/
	#define SUBDEVICE_BROADCOM_3C997_T_1000BASE_T_DUAL	0x10b71003 /*3C997-T 1000Base-T Dual Port*/
	#define SUBDEVICE_BROADCOM_3C996_SX_1000BASE_SX	0x10b71004 /*3C996-SX 1000Base-SX*/
	#define SUBDEVICE_BROADCOM_3C997_SX_1000BASE_SX_DUAL	0x10b71005 /*3C997-SX 1000Base-SX Dual Port*/
	#define SUBDEVICE_BROADCOM_3C942_GIGABIT_LOM	0x10b71008 /*3C942 Gigabit LOM (31X31)*/
	#define SUBDEVICE_BROADCOM_NETXTREME_1000BASE_SX	0x14e40002 /*NetXtreme 1000Base-SX*/
	#define SUBDEVICE_BROADCOM_NETXTREME_1000BASE_SX2	0x14e40003 /*NetXtreme 1000Base-SX*/
	#define SUBDEVICE_BROADCOM_NETXTREME_1000BASE_T	0x14e40004 /*NetXtreme 1000Base-T*/
	#define SUBDEVICE_BROADCOM_NETXTREME_1000BASETX	0x14e41028 /*NetXtreme 1000BaseTX*/
	#define SUBDEVICE_BROADCOM_BCM5700_1000BASE_T	0x14e41644 /*BCM5700 1000Base-T*/
#define DEVICE_BROADCOM_NETXTREME_BCM5701_GIGABIT	0x1645 /*NetXtreme BCM5701 Gigabit Ethernet*/
	#define SUBDEVICE_BROADCOM_NC7770_GIGABIT_SERVER	0x0e11007c /*NC7770 Gigabit Server Adapter (PCI-X, 10/100/1000-T)*/
	#define SUBDEVICE_BROADCOM_NC6770_GIGABIT_SERVER	0x0e11007d /*NC6770 Gigabit Server Adapter (PCI-X, 1000-SX)*/
	#define SUBDEVICE_BROADCOM_NC7780_GIGABIT_SERVER	0x0e110085 /*NC7780 Gigabit Server Adapter (embedded, WOL)*/
	#define SUBDEVICE_BROADCOM_NC7780_GIGABIT_SERVER2	0x0e110099 /*NC7780 Gigabit Server Adapter (embedded, WOL)*/
	#define SUBDEVICE_BROADCOM_NC7770_GIGABIT_SERVER2	0x0e11009a /*NC7770 Gigabit Server Adapter (PCI-X, 10/100/1000-T)*/
	#define SUBDEVICE_BROADCOM_NC6770_GIGABIT_SERVER2	0x0e1100c1 /*NC6770 Gigabit Server Adapter (PCI-X, 1000-SX)*/
	#define SUBDEVICE_BROADCOM_BROADCOM_BCM5701_1000BASE_T	0x10280121 /*Broadcom BCM5701 1000Base-T*/
	#define SUBDEVICE_BROADCOM_HP_1000BASE_T_PCI	0x103c128a /*HP 1000Base-T (PCI) [A7061A]*/
	#define SUBDEVICE_BROADCOM_HP_1000BASE_SX_PCI	0x103c128b /*HP 1000Base-SX (PCI) [A7073A]*/
	#define SUBDEVICE_BROADCOM_HP_CORE_LAN		0x103c12a4 /*HP Core Lan 1000Base-T*/
	#define SUBDEVICE_BROADCOM_HP_IOX_CORE		0x103c12c1 /*HP IOX Core Lan 1000Base-T [A7109AX]*/
	#define SUBDEVICE_BROADCOM_SGI_IO9_GIGABIT	0x10a98010 /*SGI IO9 Gigabit Ethernet (Copper)*/
	#define SUBDEVICE_BROADCOM_SGI_GIGABIT_ETHERNET	0x10a98011 /*SGI Gigabit Ethernet (Copper)*/
	#define SUBDEVICE_BROADCOM_SGI_GIGABIT_ETHERNET2	0x10a98012 /*SGI Gigabit Ethernet (Fiber)*/
	#define SUBDEVICE_BROADCOM_3C996_SX_1000BASE_SX2	0x10b71004 /*3C996-SX 1000Base-SX*/
	#define SUBDEVICE_BROADCOM_3C996B_T_1000BASE_T2	0x10b71006 /*3C996B-T 1000Base-T*/
	#define SUBDEVICE_BROADCOM_3C1000_T_1000BASE_T	0x10b71007 /*3C1000-T 1000Base-T*/
	#define SUBDEVICE_BROADCOM_3C940_BR01_1000BASE_T	0x10b71008 /*3C940-BR01 1000Base-T*/
	#define SUBDEVICE_BROADCOM_BCM5701_1000BASE_T	0x14e40001 /*BCM5701 1000Base-T*/
	#define SUBDEVICE_BROADCOM_BCM5701_1000BASE_T2	0x14e40005 /*BCM5701 1000Base-T*/
	#define SUBDEVICE_BROADCOM_BCM5701_1000BASE_T3	0x14e40006 /*BCM5701 1000Base-T*/
	#define SUBDEVICE_BROADCOM_BCM5701_1000BASE_SX	0x14e40007 /*BCM5701 1000Base-SX*/
	#define SUBDEVICE_BROADCOM_BCM5701_1000BASE_T4	0x14e40008 /*BCM5701 1000Base-T*/
	#define SUBDEVICE_BROADCOM_BCM5701_1000BASE_T5	0x14e48008 /*BCM5701 1000Base-T*/
#define DEVICE_BROADCOM_NETXTREME_BCM5702_GIGABIT	0x1646 /*NetXtreme BCM5702 Gigabit Ethernet*/
	#define SUBDEVICE_BROADCOM_NC7760_1000BASETX	0x0e1100bb /*NC7760 1000BaseTX*/
	#define SUBDEVICE_BROADCOM_BROADCOM_BCM5702_1000BASETX	0x10280126 /*Broadcom BCM5702 1000BaseTX*/
	#define SUBDEVICE_BROADCOM_BCM5702_1000BASETX	0x14e48009 /*BCM5702 1000BaseTX*/
#define DEVICE_BROADCOM_NETXTREME_BCM5703_GIGABIT	0x1647 /*NetXtreme BCM5703 Gigabit Ethernet*/
	#define SUBDEVICE_BROADCOM_NC7780_1000BASETX	0x0e110099 /*NC7780 1000BaseTX*/
	#define SUBDEVICE_BROADCOM_NC7770_1000BASETX	0x0e11009a /*NC7770 1000BaseTX*/
	#define SUBDEVICE_BROADCOM_SGI_IO9_GIGABIT2	0x10a98010 /*SGI IO9 Gigabit Ethernet (Copper)*/
	#define SUBDEVICE_BROADCOM_BCM5703_1000BASETX	0x14e40009 /*BCM5703 1000BaseTX*/
	#define SUBDEVICE_BROADCOM_BCM5703_1000BASESX	0x14e4000a /*BCM5703 1000BaseSX*/
	#define SUBDEVICE_BROADCOM_BCM5703_1000BASETX2	0x14e4000b /*BCM5703 1000BaseTX*/
	#define SUBDEVICE_BROADCOM_BCM5703_1000BASETX3	0x14e48009 /*BCM5703 1000BaseTX*/
	#define SUBDEVICE_BROADCOM_BCM5703_1000BASETX4	0x14e4800a /*BCM5703 1000BaseTX*/
#define DEVICE_BROADCOM_NETXTREME_BCM5704_GIGABIT	0x1648 /*NetXtreme BCM5704 Gigabit Ethernet*/
	#define SUBDEVICE_BROADCOM_NC7772_GIGABIT_SERVER	0x0e1100cf /*NC7772 Gigabit Server Adapter (PCI-X, 10,100,1000-T)*/
	#define SUBDEVICE_BROADCOM_NC7782_GIGABIT_SERVER	0x0e1100d0 /*NC7782 Gigabit Server Adapter (PCI-X, 10,100,1000-T)*/
	#define SUBDEVICE_BROADCOM_NC7783_GIGABIT_SERVER	0x0e1100d1 /*NC7783 Gigabit Server Adapter (PCI-X, 10,100,1000-T)*/
	#define SUBDEVICE_BROADCOM_3C998_T_DUAL_PORT	0x10b72000 /*3C998-T Dual Port 10/100/1000 PCI-X*/
	#define SUBDEVICE_BROADCOM_3C999_T_QUAD_PORT	0x10b73000 /*3C999-T Quad Port 10/100/1000 PCI-X*/
	#define SUBDEVICE_BROADCOM_NETXTREME_CIOB_E_1000BASE_T	0x11661648 /*NetXtreme CIOB-E 1000Base-T*/
#define DEVICE_BROADCOM_NETXTREME_II_BCM5706		0x164a /*NetXtreme II BCM5706 Gigabit Ethernet*/
#define DEVICE_BROADCOM_NETXTREME_BCM5702FE_GIGABIT	0x164d /*NetXtreme BCM5702FE Gigabit Ethernet*/
#define DEVICE_BROADCOM_NETXTREME_BCM5705_GIGABIT	0x1653 /*NetXtreme BCM5705 Gigabit Ethernet*/
	#define SUBDEVICE_BROADCOM_NC7761_GIGABIT_SERVER	0x0e1100e3 /*NC7761 Gigabit Server Adapter*/
#define DEVICE_BROADCOM_NETXTREME_BCM5705_2_GIGABIT	0x1654 /*NetXtreme BCM5705_2 Gigabit Ethernet*/
	#define SUBDEVICE_BROADCOM_NC7761_GIGABIT_SERVER2	0x0e1100e3 /*NC7761 Gigabit Server Adapter*/
	#define SUBDEVICE_BROADCOM_NC1020_HP_PROLIANT	0x103c3100 /*NC1020 HP ProLiant Gigabit Server Adapter 32 PCI*/
#define DEVICE_BROADCOM_NETXTREME_BCM5721_GIGABIT	0x1659 /*NetXtreme BCM5721 Gigabit Ethernet PCI Express*/
#define DEVICE_BROADCOM_NETXTREME_BCM5705M_GIGABIT	0x165d /*NetXtreme BCM5705M Gigabit Ethernet*/
#define DEVICE_BROADCOM_NETXTREME_BCM5705M_2_GIGABIT	0x165e /*NetXtreme BCM5705M_2 Gigabit Ethernet*/
	#define SUBDEVICE_BROADCOM_NC6000_LAPTOP	0x103c0890 /*NC6000 laptop*/
#define DEVICE_BROADCOM_570X_10_100_INTEGRATED		0x166e /*570x 10/100 Integrated Controller*/
#define DEVICE_BROADCOM_NETXTREME_BCM5751_GIGABIT	0x1677 /*NetXtreme BCM5751 Gigabit Ethernet PCI Express*/
	#define SUBDEVICE_BROADCOM_OPTIPLEX_GX280	0x10280179 /*Optiplex GX280*/
#define DEVICE_BROADCOM_NETXTREME_BCM5751M_GIGABIT	0x167d /*NetXtreme BCM5751M Gigabit Ethernet PCI Express*/
#define DEVICE_BROADCOM_NETXTREME_BCM5751F_FAST		0x167e /*NetXtreme BCM5751F Fast Ethernet PCI Express*/
#define DEVICE_BROADCOM_NETXTREME_BCM5782_GIGABIT	0x1696 /*NetXtreme BCM5782 Gigabit Ethernet*/
	#define SUBDEVICE_BROADCOM_HP_D530_CMT		0x103c12bc /*HP d530 CMT (DG746A)*/
	#define SUBDEVICE_BROADCOM_NETXTREME_BCM5782_1000BASE_T	0x14e4000d /*NetXtreme BCM5782 1000Base-T*/
#define DEVICE_BROADCOM_NETXTREME_BCM5788_GIGABIT	0x169c /*NetXtreme BCM5788 Gigabit Ethernet*/
#define DEVICE_BROADCOM_NETLINK_BCM5789_GIGABIT		0x169d /*NetLink BCM5789 Gigabit Ethernet PCI Express*/
#define DEVICE_BROADCOM_NETXTREME_BCM5702X_GIGABIT	0x16a6 /*NetXtreme BCM5702X Gigabit Ethernet*/
	#define SUBDEVICE_BROADCOM_NC7760_GIGABIT_SERVER	0x0e1100bb /*NC7760 Gigabit Server Adapter (PCI-X, 10/100/1000-T)*/
	#define SUBDEVICE_BROADCOM_BCM5702_1000BASE_T	0x10280126 /*BCM5702 1000Base-T*/
	#define SUBDEVICE_BROADCOM_BCM5702_1000BASE_T2	0x14e4000c /*BCM5702 1000Base-T*/
	#define SUBDEVICE_BROADCOM_BCM5702_1000BASE_T3	0x14e48009 /*BCM5702 1000Base-T*/
#define DEVICE_BROADCOM_NETXTREME_BCM5703X_GIGABIT	0x16a7 /*NetXtreme BCM5703X Gigabit Ethernet*/
	#define SUBDEVICE_BROADCOM_NC7771_GIGABIT_SERVER	0x0e1100ca /*NC7771 Gigabit Server Adapter (PCI-X, 10,100,1000-T)*/
	#define SUBDEVICE_BROADCOM_NC7781_GIGABIT_SERVER	0x0e1100cb /*NC7781 Gigabit Server Adapter (PCI-X, 10,100,1000-T)*/
	#define SUBDEVICE_BROADCOM_NETXTREME_BCM5703_1000BASE_T	0x14e40009 /*NetXtreme BCM5703 1000Base-T*/
	#define SUBDEVICE_BROADCOM_NETXTREME_BCM5703_1000BASE_SX	0x14e4000a /*NetXtreme BCM5703 1000Base-SX*/
	#define SUBDEVICE_BROADCOM_NETXTREME_BCM5703_1000BASE_T2	0x14e4000b /*NetXtreme BCM5703 1000Base-T*/
	#define SUBDEVICE_BROADCOM_NETXTREME_BCM5703_1000BASE_T3	0x14e4800a /*NetXtreme BCM5703 1000Base-T*/
#define DEVICE_BROADCOM_NETXTREME_BCM5704S_GIGABIT	0x16a8 /*NetXtreme BCM5704S Gigabit Ethernet*/
	#define SUBDEVICE_BROADCOM_3C998_SX_DUAL_PORT	0x10b72001 /*3C998-SX Dual Port 1000-SX PCI-X*/
#define DEVICE_BROADCOM_NETXTREME_II_BCM5706S		0x16aa /*NetXtreme II BCM5706S Gigabit Ethernet*/
#define DEVICE_BROADCOM_NETXTREME_BCM5702A3_GIGABIT	0x16c6 /*NetXtreme BCM5702A3 Gigabit Ethernet*/
	#define SUBDEVICE_BROADCOM_3C1000B_T_10_100_1000_PCI	0x10b71100 /*3C1000B-T 10/100/1000 PCI*/
	#define SUBDEVICE_BROADCOM_BCM5702_1000BASE_T4	0x14e4000c /*BCM5702 1000Base-T*/
	#define SUBDEVICE_BROADCOM_BCM5702_1000BASE_T5	0x14e48009 /*BCM5702 1000Base-T*/
#define DEVICE_BROADCOM_NETXTREME_BCM5703_GIGABIT2	0x16c7 /*NetXtreme BCM5703 Gigabit Ethernet*/
	#define SUBDEVICE_BROADCOM_NC7771_GIGABIT_SERVER2	0x0e1100ca /*NC7771 Gigabit Server Adapter (PCI-X, 10,100,1000-T)*/
	#define SUBDEVICE_BROADCOM_NC7781_GIGABIT_SERVER2	0x0e1100cb /*NC7781 Gigabit Server Adapter (PCI-X, 10,100,1000-T)*/
	#define SUBDEVICE_BROADCOM_HP_COMBO_FC_GIGE_SX	0x103c12c3 /*HP Combo FC/GigE-SX [A9782A]*/
	#define SUBDEVICE_BROADCOM_HP_COMBO_FC_GIGE_T	0x103c12ca /*HP Combo FC/GigE-T [A9784A]*/
	#define SUBDEVICE_BROADCOM_NETXTREME_BCM5703_1000BASE_T4	0x14e40009 /*NetXtreme BCM5703 1000Base-T*/
	#define SUBDEVICE_BROADCOM_NETXTREME_BCM5703_1000BASE_SX2	0x14e4000a /*NetXtreme BCM5703 1000Base-SX*/
#define DEVICE_BROADCOM_NETLINK_BCM5781_GIGABIT		0x16dd /*NetLink BCM5781 Gigabit Ethernet PCI Express*/
#define DEVICE_BROADCOM_NETXTREME_BCM5753_GIGABIT	0x16f7 /*NetXtreme BCM5753 Gigabit Ethernet PCI Express*/
#define DEVICE_BROADCOM_NETXTREME_BCM5753M_GIGABIT	0x16fd /*NetXtreme BCM5753M Gigabit Ethernet PCI Express*/
#define DEVICE_BROADCOM_NETXTREME_BCM5753F_FAST		0x16fe /*NetXtreme BCM5753F Fast Ethernet PCI Express*/
#define DEVICE_BROADCOM_BCM4401_B0_100BASE_TX		0x170c /*BCM4401-B0 100Base-TX*/
#define DEVICE_BROADCOM_NETXTREME_BCM5901_100BASE_TX	0x170d /*NetXtreme BCM5901 100Base-TX*/
	#define SUBDEVICE_BROADCOM_THINKPAD_R40E_2684_HVG	0x10140545 /*ThinkPad R40e (2684-HVG) builtin ethernet controller*/
#define DEVICE_BROADCOM_NETXTREME_BCM5901_100BASE_TX2	0x170e /*NetXtreme BCM5901 100Base-TX*/
#define DEVICE_BROADCOM_BCM3352				0x3352 /*BCM3352*/
#define DEVICE_BROADCOM_BCM3360				0x3360 /*BCM3360*/
#define DEVICE_BROADCOM_BCM4210_ILINE10_HOMEPNA		0x4210 /*BCM4210 iLine10 HomePNA 2.0*/
#define DEVICE_BROADCOM_BCM4211_ILINE10_HOMEPNA		0x4211 /*BCM4211 iLine10 HomePNA 2.0 + V.90 56k modem*/
#define DEVICE_BROADCOM_BCM4212_V_90_56K		0x4212 /*BCM4212 v.90 56k modem*/
#define DEVICE_BROADCOM_BCM4303_802_11B_WIRELESS	0x4301 /*BCM4303 802.11b Wireless LAN Controller*/
	#define SUBDEVICE_BROADCOM_TRUEMOBILE_1180_ONBOARD	0x10280407 /*TrueMobile 1180 Onboard WLAN*/
	#define SUBDEVICE_BROADCOM_WL_103B_WIRELESS_LAN	0x10430120 /*WL-103b Wireless LAN PC Card*/
#define DEVICE_BROADCOM_BCM4307_V_90_56K		0x4305 /*BCM4307 V.90 56k Modem*/
#define DEVICE_BROADCOM_BCM4307_ETHERNET_CONTROLLER	0x4306 /*BCM4307 Ethernet Controller*/
#define DEVICE_BROADCOM_BCM4307_802_11B_WIRELESS	0x4307 /*BCM4307 802.11b Wireless LAN Controller*/
#define DEVICE_BROADCOM_BCM4310_CHIPCOMMON_I_OCONTROLLER	0x4310 /*BCM4310 Chipcommon I/OController*/
#define DEVICE_BROADCOM_BCM4310_UART			0x4312 /*BCM4310 UART*/
#define DEVICE_BROADCOM_BCM4310_ETHERNET_CONTROLLER	0x4313 /*BCM4310 Ethernet Controller*/
#define DEVICE_BROADCOM_BCM4310_USB_CONTROLLER		0x4315 /*BCM4310 USB Controller*/
#define DEVICE_BROADCOM_BCM4306_802_11B_G_WIRELESS	0x4320 /*BCM4306 802.11b/g Wireless LAN Controller*/
	#define SUBDEVICE_BROADCOM_TRUEMOBILE_1300_WLAN	0x10280001 /*TrueMobile 1300 WLAN Mini-PCI Card*/
	#define SUBDEVICE_BROADCOM_WIRELESS_1350_WLAN	0x10280003 /*Wireless 1350 WLAN Mini-PCI Card*/
	#define SUBDEVICE_BROADCOM_WL_100G		0x1043100f /*WL-100G*/
	#define SUBDEVICE_BROADCOM_LINKSYS_WMP54G_PCI	0x14e44320 /*Linksys WMP54G PCI*/
	#define SUBDEVICE_BROADCOM_WPC54G		0x17374320 /*WPC54G*/
	#define SUBDEVICE_BROADCOM_BELKIN_F5D7010_54G	0x17997010 /*Belkin F5D7010 54g Wireless Network card*/
#define DEVICE_BROADCOM_BCM4306_802_11A_WIRELESS	0x4321 /*BCM4306 802.11a Wireless LAN Controller*/
#define DEVICE_BROADCOM_BCM4306_UART			0x4322 /*BCM4306 UART*/
#define DEVICE_BROADCOM_BCM4309_802_11A_B_G		0x4324 /*BCM4309 802.11a/b/g*/
	#define SUBDEVICE_BROADCOM_TRUEMOBILE_1400	0x10280001 /*Truemobile 1400*/
	#define SUBDEVICE_BROADCOM_TRUEMOBILE_1450_MINIPCI	0x10280003 /*Truemobile 1450 MiniPCI*/
#define DEVICE_BROADCOM_BCM43XG_802_11B_G		0x4325 /*BCM43xG 802.11b/g*/
	#define SUBDEVICE_BROADCOM_WIRELESS_NOTEBOOK_ADAPTER	0x14140003 /*Wireless Notebook Adapter MN-720*/
	#define SUBDEVICE_BROADCOM_WIRELESS_PCI_ADAPTER	0x14140004 /*Wireless PCI Adapter MN-730*/
#define DEVICE_BROADCOM_BCM4307_CHIPCOMMON_I_O		0x4326 /*BCM4307 Chipcommon I/O Controller?*/
#define DEVICE_BROADCOM_BCM4401_100BASE_T		0x4401 /*BCM4401 100Base-T*/
	#define SUBDEVICE_BROADCOM_A7V8X_MOTHERBOARD	0x104380a8 /*A7V8X motherboard*/
#define DEVICE_BROADCOM_BCM4402_INTEGRATED_10_100BASET	0x4402 /*BCM4402 Integrated 10/100BaseT*/
#define DEVICE_BROADCOM_BCM4402_V_90_56K		0x4403 /*BCM4402 V.90 56k Modem*/
#define DEVICE_BROADCOM_BCM4413_ILINE32_HOMEPNA		0x4410 /*BCM4413 iLine32 HomePNA 2.0*/
#define DEVICE_BROADCOM_BCM4413_V_90_56K		0x4411 /*BCM4413 V.90 56k modem*/
#define DEVICE_BROADCOM_BCM4412_10_100BASET		0x4412 /*BCM4412 10/100BaseT*/
#define DEVICE_BROADCOM_BCM44XX_CARDBUS_ILINE32		0x4430 /*BCM44xx CardBus iLine32 HomePNA 2.0*/
#define DEVICE_BROADCOM_BCM4432_CARDBUS_10_100BASET	0x4432 /*BCM4432 CardBus 10/100BaseT*/
#define DEVICE_BROADCOM_BCM4610_SENTRY5_PCI		0x4610 /*BCM4610 Sentry5 PCI to SB Bridge*/
#define DEVICE_BROADCOM_BCM4610_SENTRY5_ILINE32		0x4611 /*BCM4610 Sentry5 iLine32 HomePNA 1.0*/
#define DEVICE_BROADCOM_BCM4610_SENTRY5_V_90		0x4612 /*BCM4610 Sentry5 V.90 56k Modem*/
#define DEVICE_BROADCOM_BCM4610_SENTRY5_ETHERNET	0x4613 /*BCM4610 Sentry5 Ethernet Controller*/
#define DEVICE_BROADCOM_BCM4610_SENTRY5_EXTERNAL	0x4614 /*BCM4610 Sentry5 External Interface*/
#define DEVICE_BROADCOM_BCM4610_SENTRY5_USB		0x4615 /*BCM4610 Sentry5 USB Controller*/
#define DEVICE_BROADCOM_BCM4704_PCI_TO			0x4704 /*BCM4704 PCI to SB Bridge*/
#define DEVICE_BROADCOM_BCM4704_SENTRY5_802_11B		0x4705 /*BCM4704 Sentry5 802.11b Wireless LAN Controller*/
#define DEVICE_BROADCOM_BCM4704_SENTRY5_ETHERNET	0x4706 /*BCM4704 Sentry5 Ethernet Controller*/
#define DEVICE_BROADCOM_BCM4704_SENTRY5_USB		0x4707 /*BCM4704 Sentry5 USB Controller*/
#define DEVICE_BROADCOM_BCM4704_CRYPTO_ACCELERATOR	0x4708 /*BCM4704 Crypto Accelerator*/
#define DEVICE_BROADCOM_BCM4710_SENTRY5_PCI		0x4710 /*BCM4710 Sentry5 PCI to SB Bridge*/
#define DEVICE_BROADCOM_BCM47XX_SENTRY5_ILINE32		0x4711 /*BCM47xx Sentry5 iLine32 HomePNA 2.0*/
#define DEVICE_BROADCOM_BCM47XX_V_92_56K		0x4712 /*BCM47xx V.92 56k modem*/
#define DEVICE_BROADCOM_SENTRY5_ETHERNET_CONTROLLER2	0x4713 /*Sentry5 Ethernet Controller*/
#define DEVICE_BROADCOM_BCM47XX_SENTRY5_EXTERNAL	0x4714 /*BCM47xx Sentry5 External Interface*/
#define DEVICE_BROADCOM_SENTRY5_USB_CONTROLLER		0x4715 /*Sentry5 USB Controller*/
#define DEVICE_BROADCOM_BCM47XX_SENTRY5_USB		0x4716 /*BCM47xx Sentry5 USB Host Controller*/
#define DEVICE_BROADCOM_BCM47XX_SENTRY5_USB2		0x4717 /*BCM47xx Sentry5 USB Device Controller*/
#define DEVICE_BROADCOM_SENTRY5_CRYPTO_ACCELERATOR2	0x4718 /*Sentry5 Crypto Accelerator*/
#define DEVICE_BROADCOM_BCM4712_MIPS_CPU		0x4720 /*BCM4712 MIPS CPU*/
#define DEVICE_BROADCOM_BCM5365P_SENTRY5_HOST		0x5365 /*BCM5365P Sentry5 Host Bridge*/
#define DEVICE_BROADCOM_BCM5600_STRATASWITCH_24_2	0x5600 /*BCM5600 StrataSwitch 24+2 Ethernet Switch Controller*/
#define DEVICE_BROADCOM_BCM5605_STRATASWITCH_24_2	0x5605 /*BCM5605 StrataSwitch 24+2 Ethernet Switch Controller*/
#define DEVICE_BROADCOM_BCM5615_STRATASWITCH_24_2	0x5615 /*BCM5615 StrataSwitch 24+2 Ethernet Switch Controller*/
#define DEVICE_BROADCOM_BCM5625_STRATASWITCH_24_2	0x5625 /*BCM5625 StrataSwitch 24+2 Ethernet Switch Controller*/
#define DEVICE_BROADCOM_BCM5645_STRATASWITCH_24_2	0x5645 /*BCM5645 StrataSwitch 24+2 Ethernet Switch Controller*/
#define DEVICE_BROADCOM_BCM5670_8_PORT_10GE		0x5670 /*BCM5670 8-Port 10GE Ethernet Switch Fabric*/
#define DEVICE_BROADCOM_BCM5680_G_SWITCH_8		0x5680 /*BCM5680 G-Switch 8 Port Gigabit Ethernet Switch Controller*/
#define DEVICE_BROADCOM_BCM5690_12_PORT_MULTI_LAYER	0x5690 /*BCM5690 12-port Multi-Layer Gigabit Ethernet Switch*/
#define DEVICE_BROADCOM_BCM5691_GE_10GE_8_2		0x5691 /*BCM5691 GE/10GE 8+2 Gigabit Ethernet Switch Controller*/
#define DEVICE_BROADCOM_BCM5820_CRYPTO_ACCELERATOR	0x5820 /*BCM5820 Crypto Accelerator*/
#define DEVICE_BROADCOM_BCM5821_CRYPTO_ACCELERATOR	0x5821 /*BCM5821 Crypto Accelerator*/
#define DEVICE_BROADCOM_BCM5822_CRYPTO_ACCELERATOR	0x5822 /*BCM5822 Crypto Accelerator*/
#define DEVICE_BROADCOM_BCM5823_CRYPTO_ACCELERATOR	0x5823 /*BCM5823 Crypto Accelerator*/
#define DEVICE_BROADCOM_BCM5824_CRYPTO_ACCELERATOR	0x5824 /*BCM5824 Crypto Accelerator*/
#define DEVICE_BROADCOM_BCM5840_CRYPTO_ACCELERATOR	0x5840 /*BCM5840 Crypto Accelerator*/
#define DEVICE_BROADCOM_BCM5841_CRYPTO_ACCELERATOR	0x5841 /*BCM5841 Crypto Accelerator*/
#define DEVICE_BROADCOM_BCM5850_CRYPTO_ACCELERATOR	0x5850 /*BCM5850 Crypto Accelerator*/
/* Vendor: 14e5: Pixelfusion Ltd */
/* Vendor: 14e6: SHINING Technology Inc */
/* Vendor: 14e7: 3CX */
/* Vendor: 14e8: RAYCER Inc */
/* Vendor: 14e9: GARNETS System CO Ltd */
/* Vendor: 14ea: Planex Communications, Inc */
#define DEVICE_PLANEX_FNW_3603_TX_CARDBUS_FAST		0xab06 /*FNW-3603-TX CardBus Fast Ethernet*/
#define DEVICE_PLANEX_RTL81XX_REALTEK_ETHERNET		0xab07 /*RTL81xx RealTek Ethernet*/
/* Vendor: 14eb: SEIKO EPSON Corp */
/* Vendor: 14ec: ACQIRIS */
/* Vendor: 14ed: DATAKINETICS Ltd */
/* Vendor: 14ee: MASPRO KENKOH Corp */
/* Vendor: 14ef: CARRY Computer ENG. CO Ltd */
/* Vendor: 14f0: CANON RESEACH CENTRE FRANCE */
/* Vendor: 14f1: Conexant */
#define DEVICE_CONEXANT_HCF_56K_MODEM			0x1002 /*HCF 56k Modem*/
#define DEVICE_CONEXANT_HCF_56K_MODEM2			0x1003 /*HCF 56k Modem*/
#define DEVICE_CONEXANT_HCF_56K_MODEM3			0x1004 /*HCF 56k Modem*/
#define DEVICE_CONEXANT_HCF_56K_MODEM4			0x1005 /*HCF 56k Modem*/
#define DEVICE_CONEXANT_HCF_56K_MODEM5			0x1006 /*HCF 56k Modem*/
#define DEVICE_CONEXANT_HCF_56K_MODEM6			0x1022 /*HCF 56k Modem*/
#define DEVICE_CONEXANT_HCF_56K_MODEM7			0x1023 /*HCF 56k Modem*/
#define DEVICE_CONEXANT_HCF_56K_MODEM8			0x1024 /*HCF 56k Modem*/
#define DEVICE_CONEXANT_HCF_56K_MODEM9			0x1025 /*HCF 56k Modem*/
#define DEVICE_CONEXANT_HCF_56K_MODEM10			0x1026 /*HCF 56k Modem*/
#define DEVICE_CONEXANT_HCF_56K_MODEM11			0x1032 /*HCF 56k Modem*/
#define DEVICE_CONEXANT_HCF_56K_DATA_FAX		0x1033 /*HCF 56k Data/Fax Modem*/
	#define SUBDEVICE_CONEXANT_NEC			0x10338077 /*NEC*/
	#define SUBDEVICE_CONEXANT_DELL_ZEUS		0x122d4027 /*Dell Zeus - MDP3880-W(B) Data Fax Modem*/
	#define SUBDEVICE_CONEXANT_DELL_MERCURY		0x122d4030 /*Dell Mercury - MDP3880-U(B) Data Fax Modem*/
	#define SUBDEVICE_CONEXANT_DELL_THOR		0x122d4034 /*Dell Thor - MDP3880-W(U) Data Fax Modem*/
	#define SUBDEVICE_CONEXANT_DELL_COPPER		0x13e0020d /*Dell Copper*/
	#define SUBDEVICE_CONEXANT_DELL_SILVER		0x13e0020e /*Dell Silver*/
	#define SUBDEVICE_CONEXANT_IBM			0x13e00261 /*IBM*/
	#define SUBDEVICE_CONEXANT_COMPAQ_GOLDWING	0x13e00290 /*Compaq Goldwing*/
	#define SUBDEVICE_CONEXANT_IBM2			0x13e002a0 /*IBM*/
	#define SUBDEVICE_CONEXANT_IBM3			0x13e002b0 /*IBM*/
	#define SUBDEVICE_CONEXANT_COMPAQ_SCOOTER	0x13e002c0 /*Compaq Scooter*/
	#define SUBDEVICE_CONEXANT_IBM4			0x13e002d0 /*IBM*/
	#define SUBDEVICE_CONEXANT_IBM_P85_DF_1		0x144f1500 /*IBM P85-DF (1)*/
	#define SUBDEVICE_CONEXANT_IBM_P85_DF_2		0x144f1501 /*IBM P85-DF (2)*/
	#define SUBDEVICE_CONEXANT_IBM_P85_DF_3		0x144f150a /*IBM P85-DF (3)*/
	#define SUBDEVICE_CONEXANT_IBM_P85_DF_LOW	0x144f150b /*IBM P85-DF Low Profile (1)*/
	#define SUBDEVICE_CONEXANT_IBM_P85_DF_LOW2	0x144f1510 /*IBM P85-DF Low Profile (2)*/
#define DEVICE_CONEXANT_HCF_56K_DATA_FAX_VOICE		0x1034 /*HCF 56k Data/Fax/Voice Modem*/
#define DEVICE_CONEXANT_HCF_56K_DATA_FAX_VOICE_SPKP	0x1035 /*HCF 56k Data/Fax/Voice/Spkp (w/Handset) Modem*/
	#define SUBDEVICE_CONEXANT_FUJITSU_P85_DFSV	0x10cf1098 /*Fujitsu P85-DFSV*/
#define DEVICE_CONEXANT_HCF_56K_DATA_FAX_VOICE_SPKP2	0x1036 /*HCF 56k Data/Fax/Voice/Spkp Modem*/
	#define SUBDEVICE_CONEXANT_HCF_56K_MODEM	0x104d8067 /*HCF 56k Modem*/
	#define SUBDEVICE_CONEXANT_MDP3880SP_W		0x122d4029 /*MDP3880SP-W*/
	#define SUBDEVICE_CONEXANT_MDP3880SP_U		0x122d4031 /*MDP3880SP-U*/
	#define SUBDEVICE_CONEXANT_DELL_TITANIUM	0x13e00209 /*Dell Titanium*/
	#define SUBDEVICE_CONEXANT_DELL_GRAPHITE	0x13e0020a /*Dell Graphite*/
	#define SUBDEVICE_CONEXANT_GATEWAY_RED_OWL	0x13e00260 /*Gateway Red Owl*/
	#define SUBDEVICE_CONEXANT_GATEWAY_WHITE_HORSE	0x13e00270 /*Gateway White Horse*/
#define DEVICE_CONEXANT_HCF_56K_DATA_FAX2		0x1052 /*HCF 56k Data/Fax Modem (Worldwide)*/
#define DEVICE_CONEXANT_HCF_56K_DATA_FAX3		0x1053 /*HCF 56k Data/Fax Modem (Worldwide)*/
#define DEVICE_CONEXANT_HCF_56K_DATA_FAX_VOICE2		0x1054 /*HCF 56k Data/Fax/Voice Modem (Worldwide)*/
#define DEVICE_CONEXANT_HCF_56K_DATA_FAX_VOICE_SPKP3	0x1055 /*HCF 56k Data/Fax/Voice/Spkp (w/Handset) Modem (Worldwide)*/
#define DEVICE_CONEXANT_HCF_56K_DATA_FAX_VOICE_SPKP4	0x1056 /*HCF 56k Data/Fax/Voice/Spkp Modem (Worldwide)*/
#define DEVICE_CONEXANT_HCF_56K_DATA_FAX_VOICE_SPKP5	0x1057 /*HCF 56k Data/Fax/Voice/Spkp Modem (Worldwide)*/
#define DEVICE_CONEXANT_HCF_56K_DATA_FAX_VOICE3		0x1059 /*HCF 56k Data/Fax/Voice Modem (Worldwide)*/
#define DEVICE_CONEXANT_HCF_56K_DATA_FAX4		0x1063 /*HCF 56k Data/Fax Modem*/
#define DEVICE_CONEXANT_HCF_56K_DATA_FAX_VOICE4		0x1064 /*HCF 56k Data/Fax/Voice Modem*/
#define DEVICE_CONEXANT_HCF_56K_DATA_FAX_VOICE_SPKP6	0x1065 /*HCF 56k Data/Fax/Voice/Spkp (w/Handset) Modem*/
#define DEVICE_CONEXANT_HCF_56K_DATA_FAX_VOICE_SPKP7	0x1066 /*HCF 56k Data/Fax/Voice/Spkp Modem*/
	#define SUBDEVICE_CONEXANT_DELL_ATHENA		0x122d4033 /*Dell Athena - MDP3900V-U*/
#define DEVICE_CONEXANT_HCF_56K_DATA_FAX5		0x1433 /*HCF 56k Data/Fax Modem*/
#define DEVICE_CONEXANT_HCF_56K_DATA_FAX_VOICE5		0x1434 /*HCF 56k Data/Fax/Voice Modem*/
#define DEVICE_CONEXANT_HCF_56K_DATA_FAX_VOICE_SPKP8	0x1435 /*HCF 56k Data/Fax/Voice/Spkp (w/Handset) Modem*/
#define DEVICE_CONEXANT_HCF_56K_DATA_FAX6		0x1436 /*HCF 56k Data/Fax Modem*/
#define DEVICE_CONEXANT_HCF_56K_DATA_FAX7		0x1453 /*HCF 56k Data/Fax Modem*/
	#define SUBDEVICE_CONEXANT_IBM5			0x13e00240 /*IBM*/
	#define SUBDEVICE_CONEXANT_IBM6			0x13e00250 /*IBM*/
	#define SUBDEVICE_CONEXANT_IBM_P95_DF_1		0x144f1502 /*IBM P95-DF (1)*/
	#define SUBDEVICE_CONEXANT_IBM_P95_DF_2		0x144f1503 /*IBM P95-DF (2)*/
#define DEVICE_CONEXANT_HCF_56K_DATA_FAX_VOICE6		0x1454 /*HCF 56k Data/Fax/Voice Modem*/
#define DEVICE_CONEXANT_HCF_56K_DATA_FAX_VOICE_SPKP9	0x1455 /*HCF 56k Data/Fax/Voice/Spkp (w/Handset) Modem*/
#define DEVICE_CONEXANT_HCF_56K_DATA_FAX_VOICE_SPKP10	0x1456 /*HCF 56k Data/Fax/Voice/Spkp Modem*/
	#define SUBDEVICE_CONEXANT_DELL_EUROPA		0x122d4035 /*Dell Europa - MDP3900V-W*/
	#define SUBDEVICE_CONEXANT_DELL_MP3930V_W_C_MINIPCI	0x122d4302 /*Dell MP3930V-W(C) MiniPCI*/
#define DEVICE_CONEXANT_ADSL_ACCESSRUNNER_PCI		0x1610 /*ADSL AccessRunner PCI Arbitration Device*/
#define DEVICE_CONEXANT_ACCESSRUNNER_PCI_ADSL		0x1611 /*AccessRunner PCI ADSL Interface Device*/
#define DEVICE_CONEXANT_ADSL_ACCESSRUNNER_V2		0x1620 /*ADSL AccessRunner V2 PCI Arbitration Device*/
#define DEVICE_CONEXANT_ACCESSRUNNER_V2_PCI		0x1621 /*AccessRunner V2 PCI ADSL Interface Device*/
#define DEVICE_CONEXANT_ACCESSRUNNER_V2_PCI2		0x1622 /*AccessRunner V2 PCI ADSL Yukon WAN Adapter*/
#define DEVICE_CONEXANT_HCF_56K_MODEM12			0x1803 /*HCF 56k Modem*/
	#define SUBDEVICE_CONEXANT_623_LAN_GRIZZLY	0x0e110023 /*623-LAN Grizzly*/
	#define SUBDEVICE_CONEXANT_623_LAN_YOGI		0x0e110043 /*623-LAN Yogi*/
#define DEVICE_CONEXANT_HCF_56K_MODEM13			0x1815 /*HCF 56k Modem*/
	#define SUBDEVICE_CONEXANT_GRIZZLY		0x0e110022 /*Grizzly*/
	#define SUBDEVICE_CONEXANT_YOGI			0x0e110042 /*Yogi*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX		0x2003 /*HSF 56k Data/Fax Modem*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE		0x2004 /*HSF 56k Data/Fax/Voice Modem*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE_SPKP	0x2005 /*HSF 56k Data/Fax/Voice/Spkp (w/Handset) Modem*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE_SPKP2	0x2006 /*HSF 56k Data/Fax/Voice/Spkp Modem*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX2		0x2013 /*HSF 56k Data/Fax Modem*/
	#define SUBDEVICE_CONEXANT_BEAR			0x0e11b195 /*Bear*/
	#define SUBDEVICE_CONEXANT_SEMINOLE_1		0x0e11b196 /*Seminole 1*/
	#define SUBDEVICE_CONEXANT_SEMINOLE_2		0x0e11b1be /*Seminole 2*/
	#define SUBDEVICE_CONEXANT_ACER			0x10258013 /*Acer*/
	#define SUBDEVICE_CONEXANT_NEC2			0x1033809d /*NEC*/
	#define SUBDEVICE_CONEXANT_NEC3			0x103380bc /*NEC*/
	#define SUBDEVICE_CONEXANT_HP			0x155d6793 /*HP*/
	#define SUBDEVICE_CONEXANT_E_MACHINES		0x155d8850 /*E Machines*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE2		0x2014 /*HSF 56k Data/Fax/Voice Modem*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE_SPKP3	0x2015 /*HSF 56k Data/Fax/Voice/Spkp (w/Handset) Modem*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE_SPKP4	0x2016 /*HSF 56k Data/Fax/Voice/Spkp Modem*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX3		0x2043 /*HSF 56k Data/Fax Modem (WorldW SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE3		0x2044 /*HSF 56k Data/Fax/Voice Modem (WorldW SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE_SPKP5	0x2045 /*HSF 56k Data/Fax/Voice/Spkp (w/Handset) Modem (WorldW SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE_SPKP6	0x2046 /*HSF 56k Data/Fax/Voice/Spkp Modem (WorldW SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX4		0x2063 /*HSF 56k Data/Fax Modem (SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE4		0x2064 /*HSF 56k Data/Fax/Voice Modem (SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE_SPKP7	0x2065 /*HSF 56k Data/Fax/Voice/Spkp (w/Handset) Modem (SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE_SPKP8	0x2066 /*HSF 56k Data/Fax/Voice/Spkp Modem (SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_MODEM			0x2093 /*HSF 56k Modem*/
	#define SUBDEVICE_CONEXANT_LEGEND		0x155d2f07 /*Legend*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_CELL		0x2143 /*HSF 56k Data/Fax/Cell Modem (Mob WorldW SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE_CELL	0x2144 /*HSF 56k Data/Fax/Voice/Cell Modem (Mob WorldW SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE_SPKP9	0x2145 /*HSF 56k Data/Fax/Voice/Spkp (w/HS)/Cell Modem (Mob WorldW SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE_SPKP_CELL	0x2146 /*HSF 56k Data/Fax/Voice/Spkp/Cell Modem (Mob WorldW SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_CELL2		0x2163 /*HSF 56k Data/Fax/Cell Modem (Mob SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE_CELL2	0x2164 /*HSF 56k Data/Fax/Voice/Cell Modem (Mob SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE_SPKP10	0x2165 /*HSF 56k Data/Fax/Voice/Spkp (w/HS)/Cell Modem (Mob SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE_SPKP_CELL2	0x2166 /*HSF 56k Data/Fax/Voice/Spkp/Cell Modem (Mob SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX5		0x2343 /*HSF 56k Data/Fax CardBus Modem (Mob WorldW SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE5		0x2344 /*HSF 56k Data/Fax/Voice CardBus Modem (Mob WorldW SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE_SPKP11	0x2345 /*HSF 56k Data/Fax/Voice/Spkp (w/HS) CardBus Modem (Mob WorldW SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE_SPKP12	0x2346 /*HSF 56k Data/Fax/Voice/Spkp CardBus Modem (Mob WorldW SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX6		0x2363 /*HSF 56k Data/Fax CardBus Modem (Mob SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE6		0x2364 /*HSF 56k Data/Fax/Voice CardBus Modem (Mob SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE_SPKP13	0x2365 /*HSF 56k Data/Fax/Voice/Spkp (w/HS) CardBus Modem (Mob SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE_SPKP14	0x2366 /*HSF 56k Data/Fax/Voice/Spkp CardBus Modem (Mob SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX7		0x2443 /*HSF 56k Data/Fax Modem (Mob WorldW SmartDAA)*/
	#define SUBDEVICE_CONEXANT_MODEM		0x104d8075 /*Modem*/
	#define SUBDEVICE_CONEXANT_MODEM2		0x104d8083 /*Modem*/
	#define SUBDEVICE_CONEXANT_MODEM3		0x104d8097 /*Modem*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE7		0x2444 /*HSF 56k Data/Fax/Voice Modem (Mob WorldW SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE_SPKP15	0x2445 /*HSF 56k Data/Fax/Voice/Spkp (w/HS) Modem (Mob WorldW SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE_SPKP16	0x2446 /*HSF 56k Data/Fax/Voice/Spkp Modem (Mob WorldW SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX8		0x2463 /*HSF 56k Data/Fax Modem (Mob SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE8		0x2464 /*HSF 56k Data/Fax/Voice Modem (Mob SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE_SPKP17	0x2465 /*HSF 56k Data/Fax/Voice/Spkp (w/HS) Modem (Mob SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_DATA_FAX_VOICE_SPKP18	0x2466 /*HSF 56k Data/Fax/Voice/Spkp Modem (Mob SmartDAA)*/
#define DEVICE_CONEXANT_HSF_56K_HSFI			0x2f00 /*HSF 56k HSFi Modem*/
	#define SUBDEVICE_CONEXANT_IBM_HSFI_V_90	0x13e08d84 /*IBM HSFi V.90*/
	#define SUBDEVICE_CONEXANT_COMPAQ_STINGER	0x13e08d85 /*Compaq Stinger*/
	#define SUBDEVICE_CONEXANT_DYNALINK_56PMI	0x14f12004 /*Dynalink 56PMi*/
#define DEVICE_CONEXANT_HSF_56K_HSFI2			0x2f02 /*HSF 56k HSFi Data/Fax*/
#define DEVICE_CONEXANT_HSF_56K_HSFI3			0x2f11 /*HSF 56k HSFi Modem*/
#define DEVICE_CONEXANT_RS8234_ATM_SAR			0x8234 /*RS8234 ATM SAR Controller [ServiceSAR Plus]*/
#define DEVICE_CONEXANT_WINFAST_TV2000_XP		0x8800 /*Winfast TV2000 XP*/
/* Vendor: 14f2: MOBILITY Electronics */
#define DEVICE_MOBILITY_EV1000_BRIDGE			0x0120 /*EV1000 bridge*/
#define DEVICE_MOBILITY_EV1000_PARALLEL_PORT		0x0121 /*EV1000 Parallel port*/
#define DEVICE_MOBILITY_EV1000_SERIAL_PORT		0x0122 /*EV1000 Serial port*/
#define DEVICE_MOBILITY_EV1000_KEYBOARD_CONTROLLER	0x0123 /*EV1000 Keyboard controller*/
#define DEVICE_MOBILITY_EV1000_MOUSE_CONTROLLER		0x0124 /*EV1000 Mouse controller*/
/* Vendor: 14f3: BroadLogic */
#define DEVICE_BROADLOGIC_2030_DVB_S_SATELLITE		0x2030 /*2030 DVB-S Satellite Reciever*/
#define DEVICE_BROADLOGIC_2050_DVB_T_TERRESTRIAL	0x2050 /*2050 DVB-T Terrestrial (Cable) Reciever*/
#define DEVICE_BROADLOGIC_2060_ATSC_TERRESTRIAL		0x2060 /*2060 ATSC Terrestrial (Cable) Reciever*/
/* Vendor: 14f4: TOKYO Electronic Industry CO Ltd */
/* Vendor: 14f5: SOPAC Ltd */
/* Vendor: 14f6: COYOTE Technologies LLC */
/* Vendor: 14f7: WOLF Technology Inc */
/* Vendor: 14f8: AUDIOCODES Inc */
#define DEVICE_AUDIOCODES_TP_240_DUAL_SPAN		0x2077 /*TP-240 dual span E1 VoIP PCI card*/
/* Vendor: 14f9: AG COMMUNICATIONS */
/* Vendor: 14fa: WANDEL & GOCHERMANN */
/* Vendor: 14fb: TRANSAS MARINE (UK) Ltd */
/* Vendor: 14fc: Quadrics Ltd */
#define DEVICE_QUADRICS_QSNET_ELAN3_NETWORK		0x0000 /*QsNet Elan3 Network Adapter*/
#define DEVICE_QUADRICS_QSNETII_ELAN4_NETWORK		0x0001 /*QsNetII Elan4 Network Adapter*/
/* Vendor: 14fd: JAPAN Computer Industry Inc */
/* Vendor: 14fe: ARCHTEK TELECOM Corp */
/* Vendor: 14ff: TWINHEAD INTERNATIONAL Corp */
/* Vendor: 1500: DELTA Electronics, Inc */
#define DEVICE_DELTA_RTL81XX_REALTEK_ETHERNET		0x1360 /*RTL81xx RealTek Ethernet*/
/* Vendor: 1501: BANKSOFT CANADA Ltd */
/* Vendor: 1502: MITSUBISHI ELECTRIC LOGISTICS SUPPORT Co Ltd */
/* Vendor: 1503: KAWASAKI LSI USA Inc */
/* Vendor: 1504: KAISER Electronics */
/* Vendor: 1505: ITA INGENIEURBURO FUR TESTAUFGABEN GmbH */
/* Vendor: 1506: CHAMELEON Systems Inc */
/* Vendor: 1507: Motorola ?? / HTEC */
#define DEVICE_MOTOROLA3_MPC105_EAGLE			0x0001 /*MPC105 [Eagle]*/
#define DEVICE_MOTOROLA3_MPC106_GRACKLE			0x0002 /*MPC106 [Grackle]*/
#define DEVICE_MOTOROLA3_MPC8240_KAHLUA			0x0003 /*MPC8240 [Kahlua]*/
#define DEVICE_MOTOROLA3_MC145575_HFC_PCI		0x0100 /*MC145575 [HFC-PCI]*/
#define DEVICE_MOTOROLA3_KTI829C_100VG			0x0431 /*KTI829c 100VG*/
#define DEVICE_MOTOROLA3_RAVEN				0x4801 /*Raven*/
#define DEVICE_MOTOROLA3_FALCON				0x4802 /*Falcon*/
#define DEVICE_MOTOROLA3_HAWK				0x4803 /*Hawk*/
#define DEVICE_MOTOROLA3_CPX8216			0x4806 /*CPX8216*/
/* Vendor: 1508: HONDA CONNECTORS/MHOTRONICS Inc */
/* Vendor: 1509: FIRST INTERNATIONAL Computer Inc */
/* Vendor: 150a: FORVUS RESEARCH Inc */
/* Vendor: 150b: YAMASHITA Systems Corp */
/* Vendor: 150c: KYOPAL CO Ltd */
/* Vendor: 150d: WARPSPPED Inc */
/* Vendor: 150e: C-PORT Corp */
/* Vendor: 150f: INTEC GmbH */
/* Vendor: 1510: BEHAVIOR TECH Computer Corp */
/* Vendor: 1511: CENTILLIUM Technology Corp */
/* Vendor: 1512: ROSUN Technologies Inc */
/* Vendor: 1513: Raychem */
/* Vendor: 1514: TFL LAN Inc */
/* Vendor: 1515: Advent design */
/* Vendor: 1516: MYSON Technology Inc */
#define DEVICE_MYSON_MTD_8XX_100_10M_ETHERNET		0x0800 /*MTD-8xx 100/10M Ethernet PCI Adapter*/
#define DEVICE_MYSON_SURECOM_EP_320X_S_100_10M		0x0803 /*SURECOM EP-320X-S 100/10M Ethernet PCI Adapter*/
	#define SUBDEVICE_MYSON_SURECOM_EP_320X_S_100_10M	0x132010bd /*SURECOM EP-320X-S 100/10M Ethernet PCI Adapter*/
#define DEVICE_MYSON_MTD_8XX_100_10M_ETHERNET2		0x0891 /*MTD-8xx 100/10M Ethernet PCI Adapter*/
/* Vendor: 1517: ECHOTEK Corp */
/* Vendor: 1518: PEP MODULAR Computers GmbH */
/* Vendor: 1519: TELEFON AKTIEBOLAGET LM Ericsson */
/* Vendor: 151a: Globetek */
#define DEVICE_GLOBETEK_PCI_1002			0x1002 /*PCI-1002*/
#define DEVICE_GLOBETEK_PCI_1004			0x1004 /*PCI-1004*/
#define DEVICE_GLOBETEK_PCI_1008			0x1008 /*PCI-1008*/
/* Vendor: 151b: COMBOX Ltd */
/* Vendor: 151c: DIGITAL AUDIO LABS Inc */
/* Vendor: 151d: Fujitsu Computer Products Of America */
/* Vendor: 151e: MATRIX Corp */
/* Vendor: 151f: TOPIC SEMICONDUCTOR Corp */
#define DEVICE_TOPIC_TP560_DATA_FAX_VOICE_56K		0x0000 /*TP560 Data/Fax/Voice 56k modem*/
/* Vendor: 1520: CHAPLET System Inc */
/* Vendor: 1521: BELL Corp */
/* Vendor: 1522: MainPine Ltd */
#define DEVICE_MAINPINE_PCI_IOBUS			0x0100 /*PCI <-> IOBus Bridge*/
	#define SUBDEVICE_MAINPINE_ROCKFORCEDUO_2_PORT	0x15220200 /*RockForceDUO 2 Port V.92/V.44 Data/Fax/Voice Modem*/
	#define SUBDEVICE_MAINPINE_ROCKFORCEQUATRO_4_PORT	0x15220300 /*RockForceQUATRO 4 Port V.92/V.44 Data/Fax/Voice Modem*/
	#define SUBDEVICE_MAINPINE_ROCKFORCEDUO_2_PORT2	0x15220400 /*RockForceDUO+ 2 Port V.92/V.44 Data/Fax/Voice Modem*/
	#define SUBDEVICE_MAINPINE_ROCKFORCEQUATRO_4_PORT2	0x15220500 /*RockForceQUATRO+ 4 Port V.92/V.44 Data/Fax/Voice Modem*/
	#define SUBDEVICE_MAINPINE_ROCKFORCE_2_PORT	0x15220600 /*RockForce+ 2 Port V.90 Data/Fax/Voice Modem*/
	#define SUBDEVICE_MAINPINE_ROCKFORCE_4_PORT	0x15220700 /*RockForce+ 4 Port V.90 Data/Fax/Voice Modem*/
	#define SUBDEVICE_MAINPINE_ROCKFORCEOCTO_8_PORT	0x15220800 /*RockForceOCTO+ 8 Port V.92/V.44 Data/Fax/Voice Modem*/
	#define SUBDEVICE_MAINPINE_ROCKFORCEDUO_2_PORT3	0x15220c00 /*RockForceDUO+ 2 Port V.92/V.44 Data, V.34 Super-G3 Fax, Voice Modem*/
	#define SUBDEVICE_MAINPINE_ROCKFORCEQUATRO_4_PORT3	0x15220d00 /*RockForceQUATRO+ 4 Port V.92/V.44 Data, V.34 Super-G3 Fax, Voice Modem*/
	#define SUBDEVICE_MAINPINE_ROCKFORCEOCTO_8_PORT2	0x15221d00 /*RockForceOCTO+ 8 Port V.92/V.44 Data, V.34 Super-G3 Fax, Voice Modem*/
/* Vendor: 1523: MUSIC Semiconductors */
/* Vendor: 1524: ENE Technology Inc */
#define DEVICE_ENE_CB710_MEMORY_CARD			0x0510 /*CB710 Memory Card Reader Controller*/
#define DEVICE_ENE_PCI_SMART_CARD			0x0610 /*PCI Smart Card Reader Controller*/
#define DEVICE_ENE_CB1211_CARDBUS_CONTROLLER		0x1211 /*CB1211 Cardbus Controller*/
#define DEVICE_ENE_CB1225_CARDBUS_CONTROLLER		0x1225 /*CB1225 Cardbus Controller*/
#define DEVICE_ENE_CB1410_CARDBUS_CONTROLLER		0x1410 /*CB1410 Cardbus Controller*/
	#define SUBDEVICE_ENE_TRAVELMATE_290		0x1025005a /*TravelMate 290*/
#define DEVICE_ENE_CB_710_2_4_CARDBUS_CONTROLLER	0x1411 /*CB-710/2/4 Cardbus Controller*/
#define DEVICE_ENE_CB_712_4_CARDBUS_CONTROLLER		0x1412 /*CB-712/4 Cardbus Controller*/
#define DEVICE_ENE_CB1420_CARDBUS_CONTROLLER		0x1420 /*CB1420 Cardbus Controller*/
#define DEVICE_ENE_CB_720_2_4_CARDBUS_CONTROLLER	0x1421 /*CB-720/2/4 Cardbus Controller*/
#define DEVICE_ENE_CB_722_4_CARDBUS_CONTROLLER		0x1422 /*CB-722/4 Cardbus Controller*/
/* Vendor: 1525: IMPACT Technologies */
/* Vendor: 1526: ISS, Inc */
/* Vendor: 1527: SOLECTRON */
/* Vendor: 1528: ACKSYS */
/* Vendor: 1529: AMERICAN MICROSystems Inc */
/* Vendor: 152a: QUICKTURN DESIGN Systems */
/* Vendor: 152b: FLYTECH Technology CO Ltd */
/* Vendor: 152c: MACRAIGOR Systems LLC */
/* Vendor: 152d: QUANTA Computer Inc */
/* Vendor: 152e: MELEC Inc */
/* Vendor: 152f: PHILIPS - CRYPTO */
/* Vendor: 1530: ACQIS Technology Inc */
/* Vendor: 1531: CHRYON Corp */
/* Vendor: 1532: ECHELON Corp */
/* Vendor: 1533: BALTIMORE */
/* Vendor: 1534: ROAD Corp */
/* Vendor: 1535: EVERGREEN Technologies Inc */
/* Vendor: 1537: DATALEX COMMUNCATIONS */
/* Vendor: 1538: ARALION Inc */
#define DEVICE_ARALION_ARS106S_ULTRA_ATA		0x0303 /*ARS106S Ultra ATA 133/100/66 Host Controller*/
/* Vendor: 1539: ATELIER INFORMATIQUES et ELECTRONIQUE ETUDES S.A. */
/* Vendor: 153a: ONO SOKKI */
/* Vendor: 153b: TERRATEC Electronic GmbH */
#define DEVICE_TERRATEC_AUREON_5_1			0x1144 /*Aureon 5.1*/
#define DEVICE_TERRATEC_AUREON_5_1_SKY			0x1147 /*Aureon 5.1 Sky*/
#define DEVICE_TERRATEC_PHILIPS_SEMICONDUCTORS_SAA7134	0x1158 /*Philips Semiconductors SAA7134 (rev 01) [Terratec Cinergy 600 TV]*/
/* Vendor: 153c: ANTAL Electronic */
/* Vendor: 153d: FILANET Corp */
/* Vendor: 153e: TECHWELL Inc */
/* Vendor: 153f: MIPS DENMARK */
/* Vendor: 1540: PROVIDEO MULTIMEDIA Co Ltd */
/* Vendor: 1541: MACHONE Communications */
/* Vendor: 1542: VIVID Technology Inc */
/* Vendor: 1543: SILICON Laboratories */
#define DEVICE_SILICON11_INTEL_537_WINMODEM		0x3052 /*Intel 537 [Winmodem]*/
#define DEVICE_SILICON11_SI3036_MC_97_DAA		0x4c22 /*Si3036 MC'97 DAA*/
/* Vendor: 1544: DCM DATA Systems */
/* Vendor: 1545: VISIONTEK */
/* Vendor: 1546: IOI Technology Corp */
/* Vendor: 1547: MITUTOYO Corp */
/* Vendor: 1548: JET PROPULSION Laboratory */
/* Vendor: 1549: INTERCONNECT Systems Solutions */
/* Vendor: 154a: MAX Technologies Inc */
/* Vendor: 154b: COMPUTEX Co Ltd */
/* Vendor: 154c: VISUAL Technology Inc */
/* Vendor: 154d: PAN INTERNATIONAL Industrial Corp */
/* Vendor: 154e: SERVOTEST Ltd */
/* Vendor: 154f: STRATABEAM Technology */
/* Vendor: 1550: OPEN NETWORK Co Ltd */
/* Vendor: 1551: SMART Electronic DEVELOPMENT GmBH */
/* Vendor: 1552: RACAL AIRTECH Ltd */
/* Vendor: 1553: CHICONY Electronics Co Ltd */
/* Vendor: 1554: PROLINK Microsystems Corp */
/* Vendor: 1555: GESYTEC GmBH */
/* Vendor: 1556: PLD APPLICATIONS */
/* Vendor: 1557: MEDIASTAR Co Ltd */
/* Vendor: 1558: CLEVO/KAPOK Computer */
/* Vendor: 1559: SI LOGIC Ltd */
/* Vendor: 155a: INNOMEDIA Inc */
/* Vendor: 155b: PROTAC INTERNATIONAL Corp */
/* Vendor: 155c: Cemax-Icon Inc */
/* Vendor: 155d: Mac System Co Ltd */
/* Vendor: 155e: LP Elektronik GmbH */
/* Vendor: 155f: Perle Systems Ltd */
/* Vendor: 1560: Terayon Communications Systems */
/* Vendor: 1561: Viewgraphics Inc */
/* Vendor: 1562: Symbol Technologies */
/* Vendor: 1563: A-Trend Technology Co Ltd */
/* Vendor: 1564: Yamakatsu Electronics Industry Co Ltd */
/* Vendor: 1565: Biostar Microtech Int'l Corp */
/* Vendor: 1566: Ardent Technologies Inc */
/* Vendor: 1567: Jungsoft */
/* Vendor: 1568: DDK Electronics Inc */
/* Vendor: 1569: Palit Microsystems Inc. */
/* Vendor: 156a: Avtec Systems */
/* Vendor: 156b: 2wire Inc */
/* Vendor: 156c: Vidac Electronics GmbH */
/* Vendor: 156d: Alpha-Top Corp */
/* Vendor: 156e: Alfa Inc */
/* Vendor: 156f: M-Systems Flash Disk Pioneers Ltd */
/* Vendor: 1570: Lecroy Corp */
/* Vendor: 1571: Contemporary Controls */
#define DEVICE_CONTEMPORARY_CCSI_PCI20_485_ARCNET	0xa001 /*CCSI PCI20-485 ARCnet*/
#define DEVICE_CONTEMPORARY_CCSI_PCI20_485D_ARCNET	0xa002 /*CCSI PCI20-485D ARCnet*/
#define DEVICE_CONTEMPORARY_CCSI_PCI20_485X_ARCNET	0xa003 /*CCSI PCI20-485X ARCnet*/
#define DEVICE_CONTEMPORARY_CCSI_PCI20_CXB_ARCNET	0xa004 /*CCSI PCI20-CXB ARCnet*/
#define DEVICE_CONTEMPORARY_CCSI_PCI20_CXS_ARCNET	0xa005 /*CCSI PCI20-CXS ARCnet*/
#define DEVICE_CONTEMPORARY_CCSI_PCI20_FOG_SMA_ARCNET	0xa006 /*CCSI PCI20-FOG-SMA ARCnet*/
#define DEVICE_CONTEMPORARY_CCSI_PCI20_FOG_ST_ARCNET	0xa007 /*CCSI PCI20-FOG-ST ARCnet*/
#define DEVICE_CONTEMPORARY_CCSI_PCI20_TB5_ARCNET	0xa008 /*CCSI PCI20-TB5 ARCnet*/
#define DEVICE_CONTEMPORARY_CCSI_PCI20_5_485_5MBIT	0xa009 /*CCSI PCI20-5-485 5Mbit ARCnet*/
#define DEVICE_CONTEMPORARY_CCSI_PCI20_5_485D_5MBIT	0xa00a /*CCSI PCI20-5-485D 5Mbit ARCnet*/
#define DEVICE_CONTEMPORARY_CCSI_PCI20_5_485X_5MBIT	0xa00b /*CCSI PCI20-5-485X 5Mbit ARCnet*/
#define DEVICE_CONTEMPORARY_CCSI_PCI20_5_FOG_ST_5MBIT	0xa00c /*CCSI PCI20-5-FOG-ST 5Mbit ARCnet*/
#define DEVICE_CONTEMPORARY_CCSI_PCI20_5_FOG_SMA_5MBIT	0xa00d /*CCSI PCI20-5-FOG-SMA 5Mbit ARCnet*/
#define DEVICE_CONTEMPORARY_CCSI_PCI22_485_10MBIT	0xa201 /*CCSI PCI22-485 10Mbit ARCnet*/
#define DEVICE_CONTEMPORARY_CCSI_PCI22_485D_10MBIT	0xa202 /*CCSI PCI22-485D 10Mbit ARCnet*/
#define DEVICE_CONTEMPORARY_CCSI_PCI22_485X_10MBIT	0xa203 /*CCSI PCI22-485X 10Mbit ARCnet*/
#define DEVICE_CONTEMPORARY_CCSI_PCI22_CHB_10MBIT	0xa204 /*CCSI PCI22-CHB 10Mbit ARCnet*/
#define DEVICE_CONTEMPORARY_CCSI_PCI22_FOG_ST_10MBIT	0xa205 /*CCSI PCI22-FOG_ST 10Mbit ARCnet*/
#define DEVICE_CONTEMPORARY_CCSI_PCI22_THB_10MBIT	0xa206 /*CCSI PCI22-THB 10Mbit ARCnet*/
/* Vendor: 1572: Otis Elevator Company */
/* Vendor: 1573: Lattice - Vantis */
/* Vendor: 1574: Fairchild Semiconductor */
/* Vendor: 1575: Voltaire Advanced Data Security Ltd */
/* Vendor: 1576: Viewcast COM */
/* Vendor: 1578: HITT */
/* Vendor: 1579: Dual Technology Corp */
/* Vendor: 157a: Japan Elecronics Ind Inc */
/* Vendor: 157b: Star Multimedia Corp */
/* Vendor: 157c: Eurosoft (UK) */
#define DEVICE_EUROSOFT_FIX2000_PCI_Y2K			0x8001 /*Fix2000 PCI Y2K Compliance Card*/
/* Vendor: 157d: Gemflex Networks */
/* Vendor: 157e: Transition Networks */
/* Vendor: 157f: PX Instruments Technology Ltd */
/* Vendor: 1580: Primex Aerospace Co */
/* Vendor: 1581: SEH Computertechnik GmbH */
/* Vendor: 1582: Cytec Corp */
/* Vendor: 1583: Inet Technologies Inc */
/* Vendor: 1584: Uniwill Computer Corp */
/* Vendor: 1585: Logitron */
/* Vendor: 1586: Lancast Inc */
/* Vendor: 1587: Konica Corp */
/* Vendor: 1588: Solidum Systems Corp */
/* Vendor: 1589: Atlantek Microsystems Pty Ltd */
/* Vendor: 158a: Digalog Systems Inc */
/* Vendor: 158b: Allied Data Technologies */
/* Vendor: 158c: Hitachi Semiconductor & Devices Sales Co Ltd */
/* Vendor: 158d: Point Multimedia Systems */
/* Vendor: 158e: Lara Technology Inc */
/* Vendor: 158f: Ditect Coop */
/* Vendor: 1590: 3pardata Inc */
/* Vendor: 1591: ARN */
/* Vendor: 1592: Syba Tech Ltd */
#define DEVICE_SYBA_MULTI_IO_CARD			0x0781 /*Multi-IO Card*/
#define DEVICE_SYBA_PARALLEL_PORT_CARD			0x0782 /*Parallel Port Card 2xEPP*/
#define DEVICE_SYBA_MULTI_IO_CARD2			0x0783 /*Multi-IO Card*/
#define DEVICE_SYBA_MULTI_IO_CARD3			0x0785 /*Multi-IO Card*/
#define DEVICE_SYBA_MULTI_IO_CARD4			0x0786 /*Multi-IO Card*/
#define DEVICE_SYBA_MULTI_IO_CARD5			0x0787 /*Multi-IO Card*/
#define DEVICE_SYBA_MULTI_IO_CARD6			0x0788 /*Multi-IO Card*/
#define DEVICE_SYBA_MULTI_IO_CARD7			0x078a /*Multi-IO Card*/
/* Vendor: 1593: Bops Inc */
/* Vendor: 1594: Netgame Ltd */
/* Vendor: 1595: Diva Systems Corp */
/* Vendor: 1596: Folsom Research Inc */
/* Vendor: 1597: Memec Design Services */
/* Vendor: 1598: Granite Microsystems */
/* Vendor: 1599: Delta Electronics Inc */
/* Vendor: 159a: General Instrument */
/* Vendor: 159b: Faraday Technology Corp */
/* Vendor: 159c: Stratus Computer Systems */
/* Vendor: 159d: Ningbo Harrison Electronics Co Ltd */
/* Vendor: 159e: A-Max Technology Co Ltd */
/* Vendor: 159f: Galea Network Security */
/* Vendor: 15a0: Compumaster SRL */
/* Vendor: 15a1: Geocast Network Systems */
/* Vendor: 15a2: Catalyst Enterprises Inc */
#define DEVICE_CATALYST_TA700_PCI_BUS			0x0001 /*TA700 PCI Bus Analyzer/Exerciser*/
/* Vendor: 15a3: Italtel */
/* Vendor: 15a4: X-Net OY */
/* Vendor: 15a5: Toyota Macs Inc */
/* Vendor: 15a6: Sunlight Ultrasound Technologies Ltd */
/* Vendor: 15a7: SSE Telecom Inc */
/* Vendor: 15a8: Shanghai Communications Technologies Center */
/* Vendor: 15aa: Moreton Bay */
/* Vendor: 15ab: Bluesteel Networks Inc */
/* Vendor: 15ac: North Atlantic Instruments */
/* Vendor: 15ad: VMware Inc */
#define DEVICE_VMWARE__VMWARE_SVGA_II			0x0405 /*[VMware SVGA II] PCI Display Adapter*/
#define DEVICE_VMWARE_VIRTUAL_SVGA			0x0710 /*Virtual SVGA*/
#define DEVICE_VMWARE_VMWARE_HIGH_SPEED_VIRTUAL		0x0720 /*VMware High-Speed Virtual NIC [vmxnet]*/
/* Vendor: 15ae: Amersham Pharmacia Biotech */
/* Vendor: 15b0: Zoltrix International Ltd */
/* Vendor: 15b1: Source Technology Inc */
/* Vendor: 15b2: Mosaid Technologies Inc */
/* Vendor: 15b3: Mellanox Technologies */
#define DEVICE_MELLANOX_MT21108_INFINIBRIDGE		0x5274 /*MT21108 InfiniBridge*/
#define DEVICE_MELLANOX_MT23108_INFINIHOST		0x5a44 /*MT23108 InfiniHost*/
#define DEVICE_MELLANOX_MT23108_INFINIHOST_HCA		0x5a45 /*MT23108 [Infinihost HCA Flash Recovery]*/
#define DEVICE_MELLANOX_MT23108_PCI_BRIDGE		0x5a46 /*MT23108 PCI Bridge*/
#define DEVICE_MELLANOX_MT24204_INFINIHOST_III		0x5e8c /*MT24204 [InfiniHost III Lx HCA]*/
#define DEVICE_MELLANOX_MT24204_INFINIHOST_III2		0x5e8d /*MT24204 [InfiniHost III Lx HCA Flash Recovery]*/
#define DEVICE_MELLANOX_MT25208_INFINIHOST_III		0x6278 /*MT25208 InfiniHost III Ex (Tavor compatibility mode)*/
#define DEVICE_MELLANOX_MT25208_INFINIHOST_III2		0x6279 /*MT25208 [InfiniHost III Ex HCA Flash Recovery]*/
#define DEVICE_MELLANOX_MT25208_INFINIHOST_III3		0x6282 /*MT25208 InfiniHost III Ex*/
/* Vendor: 15b4: CCI/TRIAD */
/* Vendor: 15b5: Cimetrics Inc */
/* Vendor: 15b6: Texas Memory Systems Inc */
/* Vendor: 15b7: Sandisk Corp */
/* Vendor: 15b8: ADDI-DATA GmbH */
/* Vendor: 15b9: Maestro Digital Communications */
/* Vendor: 15ba: Impacct Technology Corp */
/* Vendor: 15bb: Portwell Inc */
/* Vendor: 15bc: Agilent Technologies */
#define DEVICE_AGILENT_64_BIT_133MHZ			0x2922 /*64 Bit, 133MHz PCI-X Exerciser & Protocol Checker*/
#define DEVICE_AGILENT_64_BIT_66MHZ			0x2928 /*64 Bit, 66MHz PCI Exerciser & Analyzer*/
#define DEVICE_AGILENT_64_BIT_133MHZ2			0x2929 /*64 Bit, 133MHz PCI-X Analyzer & Exerciser*/
/* Vendor: 15bd: DFI Inc */
/* Vendor: 15be: Sola Electronics */
/* Vendor: 15bf: High Tech Computer Corp (HTC) */
/* Vendor: 15c0: BVM Ltd */
/* Vendor: 15c1: Quantel */
/* Vendor: 15c2: Newer Technology Inc */
/* Vendor: 15c3: Taiwan Mycomp Co Ltd */
/* Vendor: 15c4: EVSX Inc */
/* Vendor: 15c5: Procomp Informatics Ltd */
#define DEVICE_PROCOMP_1394B_1394			0x8010 /*1394b - 1394 Firewire 3-Port Host Adapter Card*/
/* Vendor: 15c6: Technical University of Budapest */
/* Vendor: 15c7: Tateyama System Laboratory Co Ltd */
#define DEVICE_TATEYAMA_TATEYAMA_C_PCI_PLC_NC		0x0349 /*Tateyama C-PCI PLC/NC card Rev.01A*/
/* Vendor: 15c8: Penta Media Co Ltd */
/* Vendor: 15c9: Serome Technology Inc */
/* Vendor: 15ca: Bitboys OY */
/* Vendor: 15cb: AG Electronics Ltd */
/* Vendor: 15cc: Hotrail Inc */
/* Vendor: 15cd: Dreamtech Co Ltd */
/* Vendor: 15ce: Genrad Inc */
/* Vendor: 15cf: Hilscher GmbH */
/* Vendor: 15d1: Infineon Technologies AG */
/* Vendor: 15d2: FIC (First International Computer Inc) */
/* Vendor: 15d3: NDS Technologies Israel Ltd */
/* Vendor: 15d4: Iwill Corp */
/* Vendor: 15d5: Tatung Co */
/* Vendor: 15d6: Entridia Corp */
/* Vendor: 15d7: Rockwell-Collins Inc */
/* Vendor: 15d8: Cybernetics Technology Co Ltd */
/* Vendor: 15d9: Super Micro Computer Inc */
/* Vendor: 15da: Cyberfirm Inc */
/* Vendor: 15db: Applied Computing Systems Inc */
/* Vendor: 15dc: Litronic Inc */
#define DEVICE_LITRONIC_ARGUS_300_PCI			0x0001 /*Argus 300 PCI Cryptography Module*/
/* Vendor: 15dd: Sigmatel Inc */
/* Vendor: 15de: Malleable Technologies Inc */
/* Vendor: 15df: Infinilink Corp */
/* Vendor: 15e0: Cacheflow Inc */
/* Vendor: 15e1: Voice Technologies Group Inc */
/* Vendor: 15e2: Quicknet Technologies Inc */
/* Vendor: 15e3: Networth Technologies Inc */
/* Vendor: 15e4: VSN Systemen BV */
/* Vendor: 15e5: Valley technologies Inc */
/* Vendor: 15e6: Agere Inc */
/* Vendor: 15e7: Get Engineering Corp */
/* Vendor: 15e8: National Datacomm Corp */
#define DEVICE_NATIONAL4_WIRELESS_PCI_CARD		0x0130 /*Wireless PCI Card*/
/* Vendor: 15e9: Pacific Digital Corp */
#define DEVICE_PACIFIC_ADMA_100_DISCSTAQ_ATA		0x1841 /*ADMA-100 DiscStaQ ATA Controller*/
/* Vendor: 15ea: Tokyo Denshi Sekei K.K. */
/* Vendor: 15eb: Drsearch GmbH */
/* Vendor: 15ec: Beckhoff GmbH */
#define DEVICE_BECKHOFF_FC3101_PROFIBUS_DP		0x3101 /*FC3101 Profibus DP 1 Channel PCI*/
#define DEVICE_BECKHOFF_FC5102				0x5102 /*FC5102*/
/* Vendor: 15ed: Macrolink Inc */
/* Vendor: 15ee: In Win Development Inc */
/* Vendor: 15ef: Intelligent Paradigm Inc */
/* Vendor: 15f0: B-Tree Systems Inc */
/* Vendor: 15f1: Times N Systems Inc */
/* Vendor: 15f2: Diagnostic Instruments Inc */
/* Vendor: 15f3: Digitmedia Corp */
/* Vendor: 15f4: Valuesoft */
/* Vendor: 15f5: Power Micro Research */
/* Vendor: 15f6: Extreme Packet Device Inc */
/* Vendor: 15f7: Banctec */
/* Vendor: 15f8: Koga Electronics Co */
/* Vendor: 15f9: Zenith Electronics Corp */
/* Vendor: 15fa: J.P. Axzam Corp */
/* Vendor: 15fb: Zilog Inc */
/* Vendor: 15fc: Techsan Electronics Co Ltd */
/* Vendor: 15fd: N-CUBED.NET */
/* Vendor: 15fe: Kinpo Electronics Inc */
/* Vendor: 15ff: Fastpoint Technologies Inc */
/* Vendor: 1600: Northrop Grumman - Canada Ltd */
/* Vendor: 1601: Tenta Technology */
/* Vendor: 1602: Prosys-tec Inc */
/* Vendor: 1603: Nokia Wireless Communications */
/* Vendor: 1604: Central System Research Co Ltd */
/* Vendor: 1605: Pairgain Technologies */
/* Vendor: 1606: Europop AG */
/* Vendor: 1607: Lava Semiconductor Manufacturing Inc */
/* Vendor: 1608: Automated Wagering International */
/* Vendor: 1609: Scimetric Instruments Inc */
/* Vendor: 1612: Telesynergy Research Inc. */
/* Vendor: 1619: FarSite Communications Ltd */
#define DEVICE_FARSITE_FARSYNC_T2P_2			0x0400 /*FarSync T2P (2 port X.21/V.35/V.24)*/
#define DEVICE_FARSITE_FARSYNC_T4P_4			0x0440 /*FarSync T4P (4 port X.21/V.35/V.24)*/
/* Vendor: 161f: Rioworks */
/* Vendor: 1626: TDK Semiconductor Corp. */
#define DEVICE_TDK2_RTL81XX_FAST_ETHERNET		0x8410 /*RTL81xx Fast Ethernet*/
/* Vendor: 1629: Kongsberg Spacetec AS */
#define DEVICE_KONGSBERG_FORMAT_SYNCHRONIZER_V3_0	0x1003 /*Format synchronizer v3.0*/
#define DEVICE_KONGSBERG_FAST_UNIVERSAL_DATA		0x2002 /*Fast Universal Data Output*/
/* Vendor: 1637: Linksys */
#define DEVICE_LINKSYS2_LINKSYS_802_11B_WMP11		0x3874 /*Linksys 802.11b WMP11 PCI Wireless card*/
/* Vendor: 1638: Standard Microsystems Corp [SMC] */
#define DEVICE_STANDARD2_SMC2602W_EZCONNECT		0x1100 /*SMC2602W EZConnect / Addtron AWA-100 / Eumitcom PCI WL11000*/
/* Vendor: 163c: Smart Link Ltd. */
#define DEVICE_SMART3_SMARTLINK_SMARTPCI562_56K		0x3052 /*SmartLink SmartPCI562 56K Modem*/
#define DEVICE_SMART3_SMARTPCI561_MODEM			0x5449 /*SmartPCI561 Modem*/
/* Vendor: 1657: Brocade Communications Systems, Inc. */
/* Vendor: 165a: Epix Inc */
#define DEVICE_EPIX_PIXCI_R_CL1_CAMERA			0xc100 /*PIXCI(R) CL1 Camera Link Video Capture Board [custom QL5232]*/
#define DEVICE_EPIX_PIXCI_R_D2X_DIGITAL			0xd200 /*PIXCI(R) D2X Digital Video Capture Board [custom QL5232]*/
#define DEVICE_EPIX_PIXCI_R_D3X_DIGITAL			0xd300 /*PIXCI(R) D3X Digital Video Capture Board [custom QL5232]*/
/* Vendor: 165d: Hsing Tech. Enterprise Co., Ltd. */
/* Vendor: 1661: Worldspace Corp. */
/* Vendor: 1668: Actiontec Electronics Inc */
#define DEVICE_ACTIONTEC_MINI_PCI_BRIDGE		0x0100 /*Mini-PCI bridge*/
/* Vendor: 166d: Broadcom Corporation */
#define DEVICE_BROADCOM2_SIBYTE_BCM1125_1125H_1250_SYSTEM_ON_A_CHIP	0x0001 /*SiByte BCM1125/1125H/1250 System-on-a-Chip PCI*/
#define DEVICE_BROADCOM2_SIBYTE_BCM1125H_1250_SYSTEM_ON_A_CHIP	0x0002 /*SiByte BCM1125H/1250 System-on-a-Chip HyperTransport*/
/* Vendor: 1677: Bernecker + Rainer */
#define DEVICE_BERNECKER_5LS172_6_B_R_DUAL		0x104e /*5LS172.6 B&R Dual CAN Interface Card*/
#define DEVICE_BERNECKER_5LS172_61_B_R_DUAL		0x12d7 /*5LS172.61 B&R Dual CAN Interface Card*/
/* Vendor: 1681: Hercules */
#define DEVICE_HERCULES_HERCULES_3D_PROPHET		0x0010 /*Hercules 3d Prophet II Ultra 64MB [ 350 MHz NV15BR core, 128-bit DDR @ 460 MHz, 1.5v AGP4x  ]*/
/* Vendor: 1688: CastleNet Technology Inc. */
#define DEVICE_CASTLENET_WLAN_802_11B_CARD		0x1170 /*WLAN 802.11b card*/
/* Vendor: 168c: Atheros Communications, Inc. */
#define DEVICE_ATHEROS_AR5000_802_11A_WIRELESS		0x0007 /*AR5000 802.11a Wireless Adapter*/
#define DEVICE_ATHEROS_AR5210_802_11A_NIC		0x0011 /*AR5210 802.11a NIC*/
#define DEVICE_ATHEROS_AR5211_802_11AB_NIC		0x0012 /*AR5211 802.11ab NIC*/
#define DEVICE_ATHEROS_AR5212_802_11ABG_NIC		0x0013 /*AR5212 802.11abg NIC*/
	#define SUBDEVICE_ATHEROS_D_LINK_DWL_G650_B3	0x11863202 /*D-link DWL-G650 B3 Wireless cardbus adapter*/
	#define SUBDEVICE_ATHEROS_DWL_G520_WIRELESS_PCI	0x11863203 /*DWL-G520 Wireless PCI Adapter*/
	#define SUBDEVICE_ATHEROS_DWL_G520_WIRELESS_PCI2	0x11863a13 /*DWL-G520 Wireless PCI Adapter rev. B*/
	#define SUBDEVICE_ATHEROS_C54C_WIRELESS_801_11G	0x11863a94 /*C54C Wireless 801.11g cardbus*/
	#define SUBDEVICE_ATHEROS_NETGEAR_WG311T_WIRELESS	0x13854d00 /*Netgear WG311T Wireless PCI Adapter*/
	#define SUBDEVICE_ATHEROS_8482_WD_ORINOCO_11A_B_G	0x14b70a60 /*8482-WD ORiNOCO 11a/b/g Wireless PCI Adapter*/
	#define SUBDEVICE_ATHEROS_WG511T_WIRELESS_CARDBUS	0x168c0013 /*WG511T Wireless CardBus Adapter*/
	#define SUBDEVICE_ATHEROS_DWL_G650B2_WIRELESS_CARDBUS	0x168c1025 /*DWL-G650B2 Wireless CardBus Adapter*/
	#define SUBDEVICE_ATHEROS_NETGATE_5354MP_ARIES	0x168c2026 /*Netgate 5354MP ARIES a(108Mb turbo)/b/g MiniPCI Adapter*/
#define DEVICE_ATHEROS_AR5212_802_11ABG_NIC2		0x1014 /*AR5212 802.11abg NIC*/
/* Vendor: 16a5: Tekram Technology Co.,Ltd. */
/* Vendor: 16ab: Global Sun Technology Inc */
#define DEVICE_GLOBAL_GL24110P				0x1100 /*GL24110P*/
#define DEVICE_GLOBAL_PLX9052_PCMCIA_TO_PCI_WIRELESS	0x1101 /*PLX9052 PCMCIA-to-PCI Wireless LAN*/
#define DEVICE_GLOBAL_PCMCIA_TO_PCI_WIRELESS_NETWORK	0x1102 /*PCMCIA-to-PCI Wireless Network Bridge*/
/* Vendor: 16ae: Safenet Inc */
#define DEVICE_SAFENET_SAFEXCEL_1141			0x1141 /*SafeXcel-1141*/
/* Vendor: 16b4: Aspex Semiconductor Ltd */
/* Vendor: 16be: Creatix Polymedia GmbH */
/* Vendor: 16ca: CENATEK Inc */
#define DEVICE_CENATEK_ROCKET_DRIVE_DL			0x0001 /*Rocket Drive DL*/
/* Vendor: 16cd: Densitron Technologies */
/* Vendor: 16df: PIKA Technologies Inc. */
/* Vendor: 16e3: European Space Agency */
#define DEVICE_EUROPEAN_LEON2FT_PROCESSOR		0x1e0f /*LEON2FT Processor*/
/* Vendor: 16ec: U.S. Robotics */
#define DEVICE_U_S_USR997900_10_100_MBPS		0x00ff /*USR997900 10/100 Mbps PCI Network Card*/
#define DEVICE_U_S_WIRELESS_ACCESS_PCI			0x3685 /*Wireless Access PCI Adapter Model 022415*/
/* Vendor: 16ed: Sycron N. V. */
#define DEVICE_SYCRON_UMIO_COMMUNICATION_CARD		0x1001 /*UMIO communication card*/
/* Vendor: 16f3: Jetway Information Co., Ltd. */
/* Vendor: 16f4: Vweb Corp */
#define DEVICE_VWEB_VW2010				0x8000 /*VW2010*/
/* Vendor: 16f6: VideoTele.com, Inc. */
/* Vendor: 1702: Internet Machines Corporation (IMC) */
/* Vendor: 1705: Digital First, Inc. */
/* Vendor: 170b: NetOctave */
#define DEVICE_NETOCTAVE_NSP2000_SSL_CRYPTO_ACCELERATOR	0x0100 /*NSP2000-SSL crypto accelerator*/
/* Vendor: 170c: YottaYotta Inc. */
/* Vendor: 1725: Vitesse Semiconductor */
#define DEVICE_VITESSE2_VSC7174_PCI_PCI_X_SERIAL	0x7174 /*VSC7174 PCI/PCI-X Serial ATA Host Bus Controller*/
/* Vendor: 172a: Accelerated Encryption */
/* Vendor: 1734: Fujitsu Siemens Computer GmbH */
/* Vendor: 1737: Linksys */
#define DEVICE_LINKSYS3_WMP54G_WIRELESS_PCI		0x0013 /*WMP54G Wireless Pci Card*/
#define DEVICE_LINKSYS3_WMP54GS_WIRELESS_PCI		0x0015 /*WMP54GS Wireless Pci Card*/
#define DEVICE_LINKSYS3_GIGABIT_NETWORK_ADAPTER		0x1032 /*Gigabit Network Adapter*/
	#define SUBDEVICE_LINKSYS3_EG1032_V2_INSTANT	0x17370015 /*EG1032 v2 Instant Gigabit Network Adapter*/
#define DEVICE_LINKSYS3_GIGABIT_NETWORK_ADAPTER2	0x1064 /*Gigabit Network Adapter*/
	#define SUBDEVICE_LINKSYS3_EG1064_V2_INSTANT	0x17370016 /*EG1064 v2 Instant Gigabit Network Adapter*/
#define DEVICE_LINKSYS3_21X4X_DEC_TULIP_COMPATIBLE	0xab08 /*21x4x DEC-Tulip compatible 10/100 Ethernet*/
#define DEVICE_LINKSYS3_21X4X_DEC_TULIP_COMPATIBLE2	0xab09 /*21x4x DEC-Tulip compatible 10/100 Ethernet*/
/* Vendor: 173b: Altima (nee Broadcom) */
#define DEVICE_ALTIMA_AC1000_GIGABIT_ETHERNET		0x03e8 /*AC1000 Gigabit Ethernet*/
#define DEVICE_ALTIMA_AC1001_GIGABIT_ETHERNET		0x03e9 /*AC1001 Gigabit Ethernet*/
#define DEVICE_ALTIMA_AC9100_GIGABIT_ETHERNET		0x03ea /*AC9100 Gigabit Ethernet*/
	#define SUBDEVICE_ALTIMA_AC1002			0x173b0001 /*AC1002*/
#define DEVICE_ALTIMA_AC1003_GIGABIT_ETHERNET		0x03eb /*AC1003 Gigabit Ethernet*/
/* Vendor: 1743: Peppercon AG */
#define DEVICE_PEPPERCON_ROL_F_100_FAST_ETHERNET	0x8139 /*ROL/F-100 Fast Ethernet Adapter with ROL*/
/* Vendor: 1749: RLX Technologies */
/* Vendor: 174b: PC Partner Limited */
/* Vendor: 174d: WellX Telecom SA */
/* Vendor: 175c: AudioScience Inc */
/* Vendor: 175e: Sanera Systems, Inc. */
/* Vendor: 1787: Hightech Information System Ltd. */
/* Vendor: 1796: Research Centre Juelich */
#define DEVICE_RESEARCH_SIS1100_GIGABIT_LINK		0x0001 /*SIS1100 [Gigabit link]*/
#define DEVICE_RESEARCH_HOTLINK				0x0002 /*HOTlink*/
#define DEVICE_RESEARCH_COUNTER_TIMER			0x0003 /*Counter Timer*/
#define DEVICE_RESEARCH_CAMAC_CONTROLLER		0x0004 /*CAMAC Controller*/
#define DEVICE_RESEARCH_PROFIBUS			0x0005 /*PROFIBUS*/
#define DEVICE_RESEARCH_AMCC_HOTLINK			0x0006 /*AMCC HOTlink*/
/* Vendor: 1797: JumpTec h, GMBH */
/* Vendor: 1799: Belkin */
#define DEVICE_BELKIN_WIRELESS_PCI_CARD			0x6001 /*Wireless PCI Card - F5D6001*/
#define DEVICE_BELKIN_WIRELESS_PCMCIA_CARD		0x6020 /*Wireless PCMCIA Card - F5D6020*/
#define DEVICE_BELKIN_WIRELESS_PDA_CARD			0x6060 /*Wireless PDA Card - F5D6060*/
#define DEVICE_BELKIN_WIRELESS_PCI_CARD2		0x7000 /*Wireless PCI Card - F5D7000*/
/* Vendor: 17a0: Genesys Logic, Inc */
#define DEVICE_GENESYS_GL880S_USB_1_1			0x8033 /*GL880S USB 1.1 controller*/
#define DEVICE_GENESYS_GL880S_USB_2_0			0x8034 /*GL880S USB 2.0 controller*/
/* Vendor: 17af: Hightech Information System Ltd. */
/* Vendor: 17b3: Hawking Technologies */
#define DEVICE_HAWKING_PN672TX_10_100_ETHERNET		0xab08 /*PN672TX 10/100 Ethernet*/
/* Vendor: 17b4: Indra Networks, Inc. */
#define DEVICE_INDRA_WEBENHANCE_100_GZIP		0x0011 /*WebEnhance 100 GZIP Compression Card*/
/* Vendor: 17c0: Wistron Corp. */
/* Vendor: 17c2: Newisys, Inc. */
/* Vendor: 17cc: NetChip Technology, Inc */
#define DEVICE_NETCHIP_USB_2_0				0x2280 /*USB 2.0*/
/* Vendor: 17d5: S2io Inc. */
/* Vendor: 17ee: Connect Components Ltd */
/* Vendor: 17fe: Linksys, A Division of Cisco Systems */
#define DEVICE_LINKSYS4__AIRCONN_INPROCOMM_IPN		0x2220 /*[AirConn] INPROCOMM IPN 2220 Wireless LAN Adapter (rev 01)*/
/* Vendor: 1813: Ambient Technologies Inc */
#define DEVICE_AMBIENT_HAM_CONTROLLERLESS_MODEM		0x4000 /*HaM controllerless modem*/
	#define SUBDEVICE_AMBIENT_V9X_HAM_DATA		0x16be0001 /*V9x HAM Data Fax Modem*/
#define DEVICE_AMBIENT_HAM_PLUS_DATA			0x4100 /*HaM plus Data Fax Modem*/
	#define SUBDEVICE_AMBIENT_V9X_HAM_1394		0x16be0002 /*V9x HAM 1394*/
/* Vendor: 1814: RaLink */
#define DEVICE_RALINK_WIRELESS_PCI_ADPATOR		0x0101 /*Wireless PCI Adpator RT2400 / RT2460*/
#define DEVICE_RALINK_RALINK_RT2500_802_11		0x0201 /*Ralink RT2500 802.11 Cardbus Reference Card*/
	#define SUBDEVICE_RALINK_CWC_854_WIRELESS_G_CARDBUS	0x1371001e /*CWC-854 Wireless-G CardBus Adapter*/
	#define SUBDEVICE_RALINK_CWM_854_WIRELESS_G_MINI	0x1371001f /*CWM-854 Wireless-G Mini PCI Adapter*/
	#define SUBDEVICE_RALINK_CWP_854_WIRELESS_G_PCI	0x13710020 /*CWP-854 Wireless-G PCI Adapter*/
/* Vendor: 1820: InfiniCon Systems Inc. */
/* Vendor: 1822: Twinhan Technology Co. Ltd */
/* Vendor: 182d: SiteCom Europe BV */
#define DEVICE_SITECOM_ISDN_PCI_DC_105V2		0x3069 /*ISDN PCI DC-105V2*/
/* Vendor: 1830: Credence Systems Corporation */
/* Vendor: 183b: MikroM GmbH */
#define DEVICE_MIKROM_MVC100_DVI			0x08a7 /*MVC100 DVI*/
#define DEVICE_MIKROM_MVC101_SDI			0x08a8 /*MVC101 SDI*/
#define DEVICE_MIKROM_MVC102_DVI_AUDIO			0x08a9 /*MVC102 DVI+Audio*/
/* Vendor: 1849: ASRock Incorporation */
/* Vendor: 1851: Microtune, Inc. */
/* Vendor: 1852: Anritsu Corp. */
/* Vendor: 1867: Topspin Communications */
#define DEVICE_TOPSPIN_MT23108_PCI_X_HCA		0x5a44 /*MT23108 PCI-X HCA*/
#define DEVICE_TOPSPIN_MT23108_PCI_X_HCA2		0x5a45 /*MT23108 PCI-X HCA flash recovery*/
#define DEVICE_TOPSPIN_MT23108_PCI_X_HCA3		0x5a46 /*MT23108 PCI-X HCA bridge*/
#define DEVICE_TOPSPIN_MT25208_INFINIHOST_III		0x6278 /*MT25208 InfiniHost III Ex (Tavor compatibility mode)*/
#define DEVICE_TOPSPIN_MT25208_INFINIHOST_III2		0x6282 /*MT25208 InfiniHost III Ex*/
/* Vendor: 1888: Varisys Ltd */
#define DEVICE_VARISYS_VMFX1_FPGA_PMC			0x0301 /*VMFX1 FPGA PMC module*/
#define DEVICE_VARISYS_VSM2_DUAL_PMC			0x0601 /*VSM2 dual PMC carrier*/
#define DEVICE_VARISYS_VS14X_SERIES_POWERPC		0x0710 /*VS14x series PowerPC PCI board*/
#define DEVICE_VARISYS_VS24X_SERIES_POWERPC		0x0720 /*VS24x series PowerPC PCI board*/
/* Vendor: 1894: KNC One */
/* Vendor: 1896: B&B Electronics Manufacturing Company, Inc. */
/* Vendor: 18a1: Astute Networks Inc. */
/* Vendor: 18ac: DViCO Corporation */
#define DEVICE_DVICO_FUSIONHDTV_3_GOLD			0xd810 /*FusionHDTV 3 Gold*/
/* Vendor: 18bc: Info-Tek Corp. */
/* Vendor: 18c8: Cray Inc */
/* Vendor: 18c9: ARVOO Engineering BV */
/* Vendor: 18ca: XGI - Xabre Graphics Inc */
#define DEVICE_XGI_VOLARI_V8				0x0040 /*Volari V8*/
/* Vendor: 18e6: MPL AG */
#define DEVICE_MPL_OSCI_OCTAL_SERIAL			0x0001 /*OSCI [Octal Serial Communication Interface]*/
/* Vendor: 18f7: Commtech, Inc. */
#define DEVICE_COMMTECH_FASTCOM_ESCC_PCI_335		0x0001 /*Fastcom ESCC-PCI-335*/
#define DEVICE_COMMTECH_FASTCOM_422_4_PCI_335		0x0002 /*Fastcom 422/4-PCI-335*/
#define DEVICE_COMMTECH_FASTCOM_422_2_PCI_335		0x0004 /*Fastcom 422/2-PCI-335*/
#define DEVICE_COMMTECH_FASTCOM_IGESCC_PCI_ISO_1	0x0005 /*Fastcom IGESCC-PCI-ISO/1*/
#define DEVICE_COMMTECH_FASTCOM_232_4_PCI_335		0x000a /*Fastcom 232/4-PCI-335*/
/* Vendor: 18fb: Resilience Corporation */
/* Vendor: 1a08: Sierra semiconductor */
#define DEVICE_SIERRA3_SC15064				0x0000 /*SC15064*/
/* Vendor: 1b13: Jaton Corp */
/* Vendor: 1c1c: Symphony */
#define DEVICE_SYMPHONY2_82C101				0x0001 /*82C101*/
/* Vendor: 1d44: DPT */
#define DEVICE_DPT_PM2X24_PM3224			0xa400 /*PM2x24/PM3224*/
/* Vendor: 1de1: Tekram Technology Co.,Ltd. */
#define DEVICE_TEKRAM3_TRM_S1040			0x0391 /*TRM-S1040*/
#define DEVICE_TEKRAM3_DC_390				0x2020 /*DC-390*/
#define DEVICE_TEKRAM3_690C				0x690c /*690c*/
#define DEVICE_TEKRAM3_DC290				0xdc29 /*DC290*/
/* Vendor: 1fc0: Tumsan Oy */
#define DEVICE_TUMSAN_E2200_DUAL_E1_RAWPIPE		0x0300 /*E2200 Dual E1/Rawpipe Card*/
/* Vendor: 2000: Smart Link Ltd. */
/* Vendor: 2001: Temporal Research Ltd */
/* Vendor: 2003: Smart Link Ltd. */
/* Vendor: 2004: Smart Link Ltd. */
/* Vendor: 21c3: 21st Century Computer Corp. */
/* Vendor: 2348: Racore */
#define DEVICE_RACORE2_8142_100VG_ANYLAN		0x2010 /*8142 100VG/AnyLAN*/
/* Vendor: 2646: Kingston Technologies */
/* Vendor: 270b: Xantel Corporation */
/* Vendor: 270f: Chaintech Computer Co. Ltd */
/* Vendor: 2711: AVID Technology Inc. */
/* Vendor: 2a15: 3D Vision(?) */
/* Vendor: 3000: Hansol Electronics Inc. */
/* Vendor: 3142: Post Impression Systems. */
/* Vendor: 3388: Hint Corp */
#define DEVICE_HINT_HINT_HC4_PCI			0x0013 /*HiNT HC4 PCI to ISDN bridge, Multimedia audio controller*/
#define DEVICE_HINT_HINT_HC4_PCI2			0x0014 /*HiNT HC4 PCI to ISDN bridge, Network controller*/
#define DEVICE_HINT_HB6_UNIVERSAL_PCI_PCI		0x0020 /*HB6 Universal PCI-PCI bridge (transparent mode)*/
#define DEVICE_HINT_HB6_UNIVERSAL_PCI_PCI2		0x0021 /*HB6 Universal PCI-PCI bridge (non-transparent mode)*/
	#define SUBDEVICE_HINT_CT7_MAINBOARD		0x4c531050 /*CT7 mainboard*/
	#define SUBDEVICE_HINT_CT8_MAINBOARD		0x4c531080 /*CT8 mainboard*/
	#define SUBDEVICE_HINT_CA3_CR3_MAINBOARD	0x4c5310a0 /*CA3/CR3 mainboard*/
	#define SUBDEVICE_HINT_PPCI_MEZZANINE_32_BIT	0x4c533010 /*PPCI mezzanine (32-bit PMC)*/
	#define SUBDEVICE_HINT_PPCI_MEZZANINE_64_BIT	0x4c533011 /*PPCI mezzanine (64-bit PMC)*/
#define DEVICE_HINT_HINT_HB4_PCI_PCI			0x0022 /*HiNT HB4 PCI-PCI Bridge (PCI6150)*/
#define DEVICE_HINT_HB2_PCI_PCI_BRIDGE			0x0026 /*HB2 PCI-PCI Bridge*/
#define DEVICE_HINT_E_BAND_AUDIOTRAK_INCA88		0x101a /*E.Band [AudioTrak Inca88]*/
#define DEVICE_HINT_E_BAND_AUDIOTRAK_INCA882		0x101b /*E.Band [AudioTrak Inca88]*/
#define DEVICE_HINT_VXPRO_II_CHIPSET			0x8011 /*VXPro II Chipset*/
	#define SUBDEVICE_HINT_VXPRO_II_CHIPSET		0x33888011 /*VXPro II Chipset CPU to PCI Bridge*/
#define DEVICE_HINT_VXPRO_II_CHIPSET2			0x8012 /*VXPro II Chipset*/
	#define SUBDEVICE_HINT_VXPRO_II_CHIPSET2	0x33888012 /*VXPro II Chipset PCI to ISA Bridge*/
#define DEVICE_HINT_VXPRO_II_IDE			0x8013 /*VXPro II IDE*/
	#define SUBDEVICE_HINT_VXPRO_II_CHIPSET3	0x33888013 /*VXPro II Chipset EIDE Controller*/
/* Vendor: 3411: Quantum Designs (H.K.) Inc */
/* Vendor: 3513: ARCOM Control Systems Ltd */
/* Vendor: 3842: eVga.com. Corp. */
/* Vendor: 38ef: 4Links */
/* Vendor: 3d3d: 3DLabs */
#define DEVICE_3DLABS_GLINT_300SX			0x0001 /*GLINT 300SX*/
#define DEVICE_3DLABS_GLINT_500TX			0x0002 /*GLINT 500TX*/
#define DEVICE_3DLABS_GLINT_DELTA			0x0003 /*GLINT Delta*/
#define DEVICE_3DLABS_PERMEDIA				0x0004 /*Permedia*/
#define DEVICE_3DLABS_PERMEDIA2				0x0005 /*Permedia*/
#define DEVICE_3DLABS_GLINT_MX				0x0006 /*GLINT MX*/
#define DEVICE_3DLABS_3D_EXTREME			0x0007 /*3D Extreme*/
#define DEVICE_3DLABS_GLINT_GAMMA_G1			0x0008 /*GLINT Gamma G1*/
#define DEVICE_3DLABS_PERMEDIA_II_2D_3D			0x0009 /*Permedia II 2D+3D*/
	#define SUBDEVICE_3DLABS_ACCELSTAR_II		0x10400011 /*AccelStar II*/
	#define SUBDEVICE_3DLABS_6221L_4U		0x13e91000 /*6221L-4U*/
	#define SUBDEVICE_3DLABS_ACCELSTAR_II_3D	0x3d3d0100 /*AccelStar II 3D Accelerator*/
	#define SUBDEVICE_3DLABS_PERMEDIA_3_16		0x3d3d0111 /*Permedia 3:16*/
	#define SUBDEVICE_3DLABS_SANTA_ANA		0x3d3d0114 /*Santa Ana*/
	#define SUBDEVICE_3DLABS_OXYGEN_GVX1		0x3d3d0116 /*Oxygen GVX1*/
	#define SUBDEVICE_3DLABS_SCIROCCO		0x3d3d0119 /*Scirocco*/
	#define SUBDEVICE_3DLABS_SANTA_ANA_PCL		0x3d3d0120 /*Santa Ana PCL*/
	#define SUBDEVICE_3DLABS_OXYGEN_VX1		0x3d3d0125 /*Oxygen VX1*/
	#define SUBDEVICE_3DLABS_PERMEDIA3_CREATE	0x3d3d0127 /*Permedia3 Create!*/
#define DEVICE_3DLABS_GLINT_R3				0x000a /*GLINT R3*/
	#define SUBDEVICE_3DLABS_OXYGEN_VX12		0x3d3d0121 /*Oxygen VX1*/
#define DEVICE_3DLABS_GLINT_R3_OXYGEN			0x000c /*GLINT R3 [Oxygen VX1]*/
	#define SUBDEVICE_3DLABS_OXYGEN_VX1_4X_AGP	0x3d3d0144 /*Oxygen VX1-4X AGP [Permedia 4]*/
#define DEVICE_3DLABS_GLINT_R4_REV			0x000d /*GLint R4 rev A*/
#define DEVICE_3DLABS_GLINT_R4_REV2			0x0011 /*GLint R4 rev B*/
#define DEVICE_3DLABS_GLINT_R5_REV			0x0012 /*GLint R5 rev A*/
#define DEVICE_3DLABS_GLINT_R5_REV2			0x0013 /*GLint R5 rev B*/
#define DEVICE_3DLABS_VP10_VISUAL_PROCESSOR		0x0020 /*VP10 visual processor*/
#define DEVICE_3DLABS_VP10_VISUAL_PROCESSOR2		0x0022 /*VP10 visual processor*/
#define DEVICE_3DLABS_VP9_VISUAL_PROCESSOR		0x0024 /*VP9 visual processor*/
#define DEVICE_3DLABS_PERMEDIA_II_2D_3D2		0x0100 /*Permedia II 2D+3D*/
#define DEVICE_3DLABS_WILDCAT_III_6210			0x07a1 /*Wildcat III 6210*/
#define DEVICE_3DLABS_SUN_XVR_500_GRAPHICS		0x07a2 /*Sun XVR-500 Graphics Accelerator*/
#define DEVICE_3DLABS_WILDCAT_IV_7210			0x07a3 /*Wildcat IV 7210*/
#define DEVICE_3DLABS_PERMEDIA3				0x1004 /*Permedia*/
#define DEVICE_3DLABS_PERMEDIA4				0x3d04 /*Permedia*/
#define DEVICE_3DLABS_GLINT_VGA				0xffff /*Glint VGA*/
/* Vendor: 4005: Avance Logic Inc. */
#define DEVICE_AVANCE2_ALS300_PCI_AUDIO			0x0300 /*ALS300 PCI Audio Device*/
#define DEVICE_AVANCE2_ALS300_PCI_AUDIO2		0x0308 /*ALS300+ PCI Audio Device*/
#define DEVICE_AVANCE2_PCI_INPUT_CONTROLLER		0x0309 /*PCI Input Controller*/
#define DEVICE_AVANCE2_ALG_2064				0x1064 /*ALG-2064*/
#define DEVICE_AVANCE2_ALG_2064I			0x2064 /*ALG-2064i*/
#define DEVICE_AVANCE2_ALG_2364A_GUI_ACCELERATOR	0x2128 /*ALG-2364A GUI Accelerator*/
#define DEVICE_AVANCE2_ALG_2301				0x2301 /*ALG-2301*/
#define DEVICE_AVANCE2_ALG_2302				0x2302 /*ALG-2302*/
#define DEVICE_AVANCE2_AVG_2302_GUI_ACCELERATOR		0x2303 /*AVG-2302 GUI Accelerator*/
#define DEVICE_AVANCE2_ALG_2364A			0x2364 /*ALG-2364A*/
#define DEVICE_AVANCE2_ALG_2464				0x2464 /*ALG-2464*/
#define DEVICE_AVANCE2_ALG_2564A_25128A			0x2501 /*ALG-2564A/25128A*/
#define DEVICE_AVANCE2_ALS4000_AUDIO_CHIPSET		0x4000 /*ALS4000 Audio Chipset*/
	#define SUBDEVICE_AVANCE2_ALS4000_AUDIO_CHIPSET	0x40054000 /*ALS4000 Audio Chipset*/
#define DEVICE_AVANCE2_ALC200_200P			0x4710 /*ALC200/200P*/
/* Vendor: 4033: Addtron Technology Co, Inc. */
#define DEVICE_ADDTRON_RTL8139_ETHERNET			0x1360 /*RTL8139 Ethernet*/
/* Vendor: 4143: Digital Equipment Corp */
/* Vendor: 4144: Alpha Data */
/* Vendor: 416c: Aladdin Knowledge Systems */
#define DEVICE_ALADDIN_ALADDINCARD			0x0100 /*AladdinCARD*/
#define DEVICE_ALADDIN_CPC				0x0200 /*CPC*/
/* Vendor: 4444: Internext Compression Inc */
#define DEVICE_INTERNEXT_ITVC16_CX23416_MPEG_2		0x0016 /*iTVC16 (CX23416) MPEG-2 Encoder*/
	#define SUBDEVICE_INTERNEXT_WINTV_PVR_250	0x00704009 /*WinTV PVR 250*/
#define DEVICE_INTERNEXT_ITVC15_MPEG_2_ENCODER		0x0803 /*iTVC15 MPEG-2 Encoder*/
	#define SUBDEVICE_INTERNEXT_WINTV_PVR_350	0x00704000 /*WinTV PVR-350*/
	#define SUBDEVICE_INTERNEXT_WINTV_PVR_2502	0x00704001 /*WinTV PVR-250*/
/* Vendor: 4468: Bridgeport machines */
/* Vendor: 4594: Cogetec Informatique Inc */
/* Vendor: 45fb: Baldor Electric Company */
/* Vendor: 4680: Umax Computer Corp */
/* Vendor: 4843: Hercules Computer Technology Inc */
/* Vendor: 4916: RedCreek Communications Inc */
#define DEVICE_REDCREEK2_REDCREEK_PCI_ADAPTER		0x1960 /*RedCreek PCI adapter*/
/* Vendor: 4943: Growth Networks */
/* Vendor: 494f: ACCES I/O Products, Inc. */
#define DEVICE_ACCES_LPCI_COM_8SM			0x10e8 /*LPCI-COM-8SM*/
/* Vendor: 4978: Axil Computer Inc */
/* Vendor: 4a14: NetVin */
#define DEVICE_NETVIN_NV5000SC				0x5000 /*NV5000SC*/
	#define SUBDEVICE_NETVIN_RT8029_BASED_ETHERNET_ADAPTER	0x4a145000 /*RT8029-Based Ethernet Adapter*/
/* Vendor: 4b10: Buslogic Inc. */
/* Vendor: 4c48: LUNG HWA Electronics */
/* Vendor: 4c53: SBS Technologies */
#define DEVICE_SBS3_PLUSTEST_DEVICE			0x0000 /*PLUSTEST device*/
	#define SUBDEVICE_SBS3_PLUSTEST_CARD_PC104	0x4c533000 /*PLUSTEST card (PC104+)*/
	#define SUBDEVICE_SBS3_PLUSTEST_CARD_PMC	0x4c533001 /*PLUSTEST card (PMC)*/
#define DEVICE_SBS3_PLUSTEST_MM_DEVICE			0x0001 /*PLUSTEST-MM device*/
	#define SUBDEVICE_SBS3_PLUSTEST_MM_CARD_PMC	0x4c533002 /*PLUSTEST-MM card (PMC)*/
/* Vendor: 4ca1: Seanix Technology Inc */
/* Vendor: 4d51: MediaQ Inc. */
#define DEVICE_MEDIAQ_MQ_200				0x0200 /*MQ-200*/
/* Vendor: 4d54: Microtechnica Co Ltd */
/* Vendor: 4ddc: ILC Data Device Corp */
#define DEVICE_ILC_DD_42924I5_300_ARINC_429		0x0100 /*DD-42924I5-300 (ARINC 429 Data Bus)*/
#define DEVICE_ILC_BU_65570I1_MIL_STD_1553_TEST		0x0801 /*BU-65570I1 MIL-STD-1553 Test and Simulation*/
#define DEVICE_ILC_BU_65570I2_MIL_STD_1553_TEST		0x0802 /*BU-65570I2 MIL-STD-1553 Test and Simulation*/
#define DEVICE_ILC_BU_65572I1_MIL_STD_1553_TEST		0x0811 /*BU-65572I1 MIL-STD-1553 Test and Simulation*/
#define DEVICE_ILC_BU_65572I2_MIL_STD_1553_TEST		0x0812 /*BU-65572I2 MIL-STD-1553 Test and Simulation*/
#define DEVICE_ILC_BU_65570T1_MIL_STD_1553_TEST		0x0881 /*BU-65570T1 MIL-STD-1553 Test and Simulation*/
#define DEVICE_ILC_BU_65570T2_MIL_STD_1553_TEST		0x0882 /*BU-65570T2 MIL-STD-1553 Test and Simulation*/
#define DEVICE_ILC_BU_65572T1_MIL_STD_1553_TEST		0x0891 /*BU-65572T1 MIL-STD-1553 Test and Simulation*/
#define DEVICE_ILC_BU_65572T2_MIL_STD_1553_TEST		0x0892 /*BU-65572T2 MIL-STD-1553 Test and Simulation*/
#define DEVICE_ILC_BU_65565C1_MIL_STD_1553_DATA		0x0901 /*BU-65565C1 MIL-STD-1553 Data Bus*/
#define DEVICE_ILC_BU_65565C2_MIL_STD_1553_DATA		0x0902 /*BU-65565C2 MIL-STD-1553 Data Bus*/
#define DEVICE_ILC_BU_65565C3_MIL_STD_1553_DATA		0x0903 /*BU-65565C3 MIL-STD-1553 Data Bus*/
#define DEVICE_ILC_BU_65565C4_MIL_STD_1553_DATA		0x0904 /*BU-65565C4 MIL-STD-1553 Data Bus*/
#define DEVICE_ILC_BU_65569I1_MIL_STD_1553_DATA		0x0b01 /*BU-65569I1 MIL-STD-1553 Data Bus*/
#define DEVICE_ILC_BU_65569I2_MIL_STD_1553_DATA		0x0b02 /*BU-65569I2 MIL-STD-1553 Data Bus*/
#define DEVICE_ILC_BU_65569I3_MIL_STD_1553_DATA		0x0b03 /*BU-65569I3 MIL-STD-1553 Data Bus*/
#define DEVICE_ILC_BU_65569I4_MIL_STD_1553_DATA		0x0b04 /*BU-65569I4 MIL-STD-1553 Data Bus*/
/* Vendor: 5046: GemTek Technology Corporation */
#define DEVICE_GEMTEK_PCI_RADIO				0x1001 /*PCI Radio*/
/* Vendor: 5053: Voyetra Technologies */
#define DEVICE_VOYETRA_DAYTONA_AUDIO_ADAPTER		0x2010 /*Daytona Audio Adapter*/
/* Vendor: 5136: S S Technologies */
/* Vendor: 5143: Qualcomm Inc */
/* Vendor: 5145: Ensoniq (Old) */
#define DEVICE_ENSONIQ2_CONCERT_AUDIOPCI		0x3031 /*Concert AudioPCI*/
/* Vendor: 5168: Animation Technologies Inc. */
/* Vendor: 5301: Alliance Semiconductor Corp. */
#define DEVICE_ALLIANCE3_PROMOTION_AT3D			0x0001 /*ProMotion aT3D*/
/* Vendor: 5333: S3 Inc. */
#define DEVICE_S3_INC_PLATO_PX_SYSTEM			0x0551 /*Plato/PX (system)*/
#define DEVICE_S3_INC_86C325_VIRGE			0x5631 /*86c325 [ViRGE]*/
#define DEVICE_S3_INC_86C866_VISION_866			0x8800 /*86c866 [Vision 866]*/
#define DEVICE_S3_INC_86C964_VISION_964			0x8801 /*86c964 [Vision 964]*/
#define DEVICE_S3_INC_86C764_0_TRIO_32			0x8810 /*86c764_0 [Trio 32 vers 0]*/
#define DEVICE_S3_INC_86C764_765_TRIO32_64_64V		0x8811 /*86c764/765 [Trio32/64/64V+]*/
#define DEVICE_S3_INC_86CM65_AURORA64V			0x8812 /*86cM65 [Aurora64V+]*/
#define DEVICE_S3_INC_86C764_3_TRIO_32_64		0x8813 /*86c764_3 [Trio 32/64 vers 3]*/
#define DEVICE_S3_INC_86C767_TRIO_64UV			0x8814 /*86c767 [Trio 64UV+]*/
#define DEVICE_S3_INC_86CM65_AURORA_128			0x8815 /*86cM65 [Aurora 128]*/
#define DEVICE_S3_INC_86C988_VIRGE_VX			0x883d /*86c988 [ViRGE/VX]*/
#define DEVICE_S3_INC_FIREGL				0x8870 /*FireGL*/
#define DEVICE_S3_INC_86C868_VISION_868			0x8880 /*86c868 [Vision 868 VRAM] vers 0*/
#define DEVICE_S3_INC_86C868_VISION_8682		0x8881 /*86c868 [Vision 868 VRAM] vers 1*/
#define DEVICE_S3_INC_86C868_VISION_8683		0x8882 /*86c868 [Vision 868 VRAM] vers 2*/
#define DEVICE_S3_INC_86C868_VISION_8684		0x8883 /*86c868 [Vision 868 VRAM] vers 3*/
#define DEVICE_S3_INC_86C928_VISION_928			0x88b0 /*86c928 [Vision 928 VRAM] vers 0*/
#define DEVICE_S3_INC_86C928_VISION_9282		0x88b1 /*86c928 [Vision 928 VRAM] vers 1*/
#define DEVICE_S3_INC_86C928_VISION_9283		0x88b2 /*86c928 [Vision 928 VRAM] vers 2*/
#define DEVICE_S3_INC_86C928_VISION_9284		0x88b3 /*86c928 [Vision 928 VRAM] vers 3*/
#define DEVICE_S3_INC_86C864_VISION_864			0x88c0 /*86c864 [Vision 864 DRAM] vers 0*/
#define DEVICE_S3_INC_86C864_VISION_8642		0x88c1 /*86c864 [Vision 864 DRAM] vers 1*/
#define DEVICE_S3_INC_86C864_VISION_864_P		0x88c2 /*86c864 [Vision 864-P DRAM] vers 2*/
#define DEVICE_S3_INC_86C864_VISION_864_P2		0x88c3 /*86c864 [Vision 864-P DRAM] vers 3*/
#define DEVICE_S3_INC_86C964_VISION_9642		0x88d0 /*86c964 [Vision 964 VRAM] vers 0*/
#define DEVICE_S3_INC_86C964_VISION_9643		0x88d1 /*86c964 [Vision 964 VRAM] vers 1*/
#define DEVICE_S3_INC_86C964_VISION_964_P		0x88d2 /*86c964 [Vision 964-P VRAM] vers 2*/
#define DEVICE_S3_INC_86C964_VISION_964_P2		0x88d3 /*86c964 [Vision 964-P VRAM] vers 3*/
#define DEVICE_S3_INC_86C968_VISION_968			0x88f0 /*86c968 [Vision 968 VRAM] rev 0*/
#define DEVICE_S3_INC_86C968_VISION_9682		0x88f1 /*86c968 [Vision 968 VRAM] rev 1*/
#define DEVICE_S3_INC_86C968_VISION_9683		0x88f2 /*86c968 [Vision 968 VRAM] rev 2*/
#define DEVICE_S3_INC_86C968_VISION_9684		0x88f3 /*86c968 [Vision 968 VRAM] rev 3*/
#define DEVICE_S3_INC_86C755_TRIO_64V2_DX		0x8900 /*86c755 [Trio 64V2/DX]*/
	#define SUBDEVICE_S3_INC_86C775_TRIO64V2_DX	0x53338900 /*86C775 Trio64V2/DX*/
#define DEVICE_S3_INC_86C775_86C785_TRIO_64V2_DX	0x8901 /*86c775/86c785 [Trio 64V2/DX or /GX]*/
	#define SUBDEVICE_S3_INC_86C775_TRIO64V2_DX_86C785	0x53338901 /*86C775 Trio64V2/DX, 86C785 Trio64V2/GX*/
#define DEVICE_S3_INC_PLATO_PX				0x8902 /*Plato/PX*/
#define DEVICE_S3_INC_TRIO_3D_BUSINESS			0x8903 /*Trio 3D business multimedia*/
#define DEVICE_S3_INC_TRIO_64_3D			0x8904 /*Trio 64 3D*/
	#define SUBDEVICE_S3_INC_INTEGRATED_TRIO3D	0x101400db /*Integrated Trio3D*/
	#define SUBDEVICE_S3_INC_86C365_TRIO3D_AGP	0x53338904 /*86C365 Trio3D AGP*/
#define DEVICE_S3_INC_TRIO_64V_FAMILY			0x8905 /*Trio 64V+ family*/
#define DEVICE_S3_INC_TRIO_64V_FAMILY2			0x8906 /*Trio 64V+ family*/
#define DEVICE_S3_INC_TRIO_64V_FAMILY3			0x8907 /*Trio 64V+ family*/
#define DEVICE_S3_INC_TRIO_64V_FAMILY4			0x8908 /*Trio 64V+ family*/
#define DEVICE_S3_INC_TRIO_64V_FAMILY5			0x8909 /*Trio 64V+ family*/
#define DEVICE_S3_INC_TRIO_64V_FAMILY6			0x890a /*Trio 64V+ family*/
#define DEVICE_S3_INC_TRIO_64V_FAMILY7			0x890b /*Trio 64V+ family*/
#define DEVICE_S3_INC_TRIO_64V_FAMILY8			0x890c /*Trio 64V+ family*/
#define DEVICE_S3_INC_TRIO_64V_FAMILY9			0x890d /*Trio 64V+ family*/
#define DEVICE_S3_INC_TRIO_64V_FAMILY10			0x890e /*Trio 64V+ family*/
#define DEVICE_S3_INC_TRIO_64V_FAMILY11			0x890f /*Trio 64V+ family*/
#define DEVICE_S3_INC_VIRGE_DX_OR_GX			0x8a01 /*ViRGE/DX or /GX*/
	#define SUBDEVICE_S3_INC_VIRGE_GX		0x0e11b032 /*ViRGE/GX*/
	#define SUBDEVICE_S3_INC_NITRO_3D		0x10b41617 /*Nitro 3D*/
	#define SUBDEVICE_S3_INC_NITRO_3D2		0x10b41717 /*Nitro 3D*/
	#define SUBDEVICE_S3_INC_VIRGE_DX		0x53338a01 /*ViRGE/DX*/
#define DEVICE_S3_INC_VIRGE_GX2				0x8a10 /*ViRGE/GX2*/
	#define SUBDEVICE_S3_INC_STEALTH_3D_4000	0x10928a10 /*Stealth 3D 4000*/
#define DEVICE_S3_INC_86C368_TRIO_3D_2X			0x8a13 /*86c368 [Trio 3D/2X]*/
	#define SUBDEVICE_S3_INC_TRIO3D_2X		0x53338a13 /*Trio3D/2X*/
#define DEVICE_S3_INC_86C794_SAVAGE_3D			0x8a20 /*86c794 [Savage 3D]*/
	#define SUBDEVICE_S3_INC_86C391_SAVAGE3D	0x53338a20 /*86C391 Savage3D*/
#define DEVICE_S3_INC_86C390_SAVAGE_3D_MV		0x8a21 /*86c390 [Savage 3D/MV]*/
	#define SUBDEVICE_S3_INC_86C390_SAVAGE3D_MV	0x53338a21 /*86C390 Savage3D/MV*/
#define DEVICE_S3_INC_SAVAGE_4				0x8a22 /*Savage 4*/
	#define SUBDEVICE_S3_INC_SAVAGE_4		0x10338068 /*Savage 4*/
	#define SUBDEVICE_S3_INC_SAVAGE_42		0x10338069 /*Savage 4*/
	#define SUBDEVICE_S3_INC_SAVAGE4_LT		0x10338110 /*Savage4 LT*/
	#define SUBDEVICE_S3_INC_SR9_8MB_SDRAM		0x105d0018 /*SR9 8Mb SDRAM*/
	#define SUBDEVICE_S3_INC_SR9_PRO_16MB		0x105d002a /*SR9 Pro 16Mb SDRAM*/
	#define SUBDEVICE_S3_INC_SR9_PRO_32MB		0x105d003a /*SR9 Pro 32Mb SDRAM*/
	#define SUBDEVICE_S3_INC_SR9_PRO_16MB2		0x105d092f /*SR9 Pro+ 16Mb SGRAM*/
	#define SUBDEVICE_S3_INC_STEALTH_III_S540	0x10924207 /*Stealth III S540*/
	#define SUBDEVICE_S3_INC_STEALTH_III_S5402	0x10924800 /*Stealth III S540*/
	#define SUBDEVICE_S3_INC_SPEEDSTAR_A90		0x10924807 /*SpeedStar A90*/
	#define SUBDEVICE_S3_INC_STEALTH_III_S5403	0x10924808 /*Stealth III S540*/
	#define SUBDEVICE_S3_INC_STEALTH_III_S5404	0x10924809 /*Stealth III S540*/
	#define SUBDEVICE_S3_INC_STEALTH_III_S5405	0x1092480e /*Stealth III S540*/
	#define SUBDEVICE_S3_INC_STEALTH_III_S520	0x10924904 /*Stealth III S520*/
	#define SUBDEVICE_S3_INC_SPEEDSTAR_A200		0x10924905 /*SpeedStar A200*/
	#define SUBDEVICE_S3_INC_STEALTH_III_S5406	0x10924a09 /*Stealth III S540*/
	#define SUBDEVICE_S3_INC_STEALTH_III_S5407	0x10924a0b /*Stealth III S540 Xtreme*/
	#define SUBDEVICE_S3_INC_STEALTH_III_S5408	0x10924a0f /*Stealth III S540*/
	#define SUBDEVICE_S3_INC_STEALTH_III_S5409	0x10924e01 /*Stealth III S540*/
	#define SUBDEVICE_S3_INC_3D_BLASTER_SAVAGE	0x1102101d /*3d Blaster Savage 4*/
	#define SUBDEVICE_S3_INC_3D_BLASTER_SAVAGE2	0x1102101e /*3d Blaster Savage 4*/
	#define SUBDEVICE_S3_INC_86C394_397_SAVAGE4_SDRAM	0x53338100 /*86C394-397 Savage4 SDRAM 100*/
	#define SUBDEVICE_S3_INC_86C394_397_SAVAGE4_SDRAM2	0x53338110 /*86C394-397 Savage4 SDRAM 110*/
	#define SUBDEVICE_S3_INC_86C394_397_SAVAGE4_SDRAM3	0x53338125 /*86C394-397 Savage4 SDRAM 125*/
	#define SUBDEVICE_S3_INC_86C394_397_SAVAGE4_SDRAM4	0x53338143 /*86C394-397 Savage4 SDRAM 143*/
	#define SUBDEVICE_S3_INC_86C394_397_SAVAGE4	0x53338a22 /*86C394-397 Savage4*/
	#define SUBDEVICE_S3_INC_86C394_397_SAVAGE4_32BIT	0x53338a2e /*86C394-397 Savage4 32bit*/
	#define SUBDEVICE_S3_INC_86C394_397_SAVAGE4_SGRAM	0x53339125 /*86C394-397 Savage4 SGRAM 125*/
	#define SUBDEVICE_S3_INC_86C394_397_SAVAGE4_SGRAM2	0x53339143 /*86C394-397 Savage4 SGRAM 143*/
#define DEVICE_S3_INC_SAVAGE_42				0x8a23 /*Savage 4*/
#define DEVICE_S3_INC_PROSAVAGE_PM133			0x8a25 /*ProSavage PM133*/
#define DEVICE_S3_INC_PROSAVAGE_KM133			0x8a26 /*ProSavage KM133*/
#define DEVICE_S3_INC_VIRGE_M3				0x8c00 /*ViRGE/M3*/
#define DEVICE_S3_INC_VIRGE_MX				0x8c01 /*ViRGE/MX*/
	#define SUBDEVICE_S3_INC_VIRGE_MX		0x11790001 /*ViRGE/MX*/
#define DEVICE_S3_INC_VIRGE_MX2				0x8c02 /*ViRGE/MX+*/
#define DEVICE_S3_INC_VIRGE_MX_MV			0x8c03 /*ViRGE/MX+MV*/
#define DEVICE_S3_INC_86C270_294_SAVAGE_MX_MV		0x8c10 /*86C270-294 Savage/MX-MV*/
#define DEVICE_S3_INC_82C270_294_SAVAGE_MX		0x8c11 /*82C270-294 Savage/MX*/
#define DEVICE_S3_INC_86C270_294_SAVAGE_IX_MV		0x8c12 /*86C270-294 Savage/IX-MV*/
	#define SUBDEVICE_S3_INC_THINKPAD_T20		0x1014017f /*ThinkPad T20*/
#define DEVICE_S3_INC_86C270_294_SAVAGE_IX		0x8c13 /*86C270-294 Savage/IX*/
	#define SUBDEVICE_S3_INC_MAGNIA_Z310		0x11790001 /*Magnia Z310*/
#define DEVICE_S3_INC_SUPERSAVAGE_MX_128		0x8c22 /*SuperSavage MX/128*/
#define DEVICE_S3_INC_SUPERSAVAGE_MX_64			0x8c24 /*SuperSavage MX/64*/
#define DEVICE_S3_INC_SUPERSAVAGE_MX_64C		0x8c26 /*SuperSavage MX/64C*/
#define DEVICE_S3_INC_SUPERSAVAGE_IX_128_SDR		0x8c2a /*SuperSavage IX/128 SDR*/
#define DEVICE_S3_INC_SUPERSAVAGE_IX_128_DDR		0x8c2b /*SuperSavage IX/128 DDR*/
#define DEVICE_S3_INC_SUPERSAVAGE_IX_64_SDR		0x8c2c /*SuperSavage IX/64 SDR*/
#define DEVICE_S3_INC_SUPERSAVAGE_IX_64_DDR		0x8c2d /*SuperSavage IX/64 DDR*/
#define DEVICE_S3_INC_SUPERSAVAGE_IX_C_SDR		0x8c2e /*SuperSavage IX/C SDR*/
	#define SUBDEVICE_S3_INC_THINKPAD_T23_2647_4MG	0x101401fc /*ThinkPad T23 (2647-4MG)*/
#define DEVICE_S3_INC_SUPERSAVAGE_IX_C_DDR		0x8c2f /*SuperSavage IX/C DDR*/
#define DEVICE_S3_INC_86C380_PROSAVAGEDDR_K4M266	0x8d01 /*86C380 [ProSavageDDR K4M266]*/
#define DEVICE_S3_INC_VT8636A_PROSAVAGE_KN133		0x8d02 /*VT8636A [ProSavage KN133] AGP4X VGA Controller (TwisterK)*/
#define DEVICE_S3_INC_VT8751_PROSAVAGEDDR_P4M266	0x8d03 /*VT8751 [ProSavageDDR P4M266]*/
#define DEVICE_S3_INC_VT8375_PROSAVAGE8_KM266_KL266	0x8d04 /*VT8375 [ProSavage8 KM266/KL266]*/
#define DEVICE_S3_INC_86C410_SAVAGE_2000		0x9102 /*86C410 Savage 2000*/
	#define SUBDEVICE_S3_INC_VIPER_II_Z200		0x10925932 /*Viper II Z200*/
	#define SUBDEVICE_S3_INC_VIPER_II_Z2002		0x10925934 /*Viper II Z200*/
	#define SUBDEVICE_S3_INC_VIPER_II_Z2003		0x10925952 /*Viper II Z200*/
	#define SUBDEVICE_S3_INC_VIPER_II_Z2004		0x10925954 /*Viper II Z200*/
	#define SUBDEVICE_S3_INC_VIPER_II_Z2005		0x10925a35 /*Viper II Z200*/
	#define SUBDEVICE_S3_INC_VIPER_II_Z2006		0x10925a37 /*Viper II Z200*/
	#define SUBDEVICE_S3_INC_VIPER_II_Z2007		0x10925a55 /*Viper II Z200*/
	#define SUBDEVICE_S3_INC_VIPER_II_Z2008		0x10925a57 /*Viper II Z200*/
#define DEVICE_S3_INC_SONICVIBES			0xca00 /*SonicVibes*/
/* Vendor: 544c: Teralogic Inc */
#define DEVICE_TERALOGIC_TL880_BASED_HDTV_ATSC_TUNER	0x0350 /*TL880-based HDTV/ATSC tuner*/
/* Vendor: 5455: Technische University Berlin */
#define DEVICE_TECHNISCHE_S5933				0x4458 /*S5933*/
/* Vendor: 5519: Cnet Technologies, Inc. */
/* Vendor: 5544: Dunord Technologies */
#define DEVICE_DUNORD_I_30XX_SCANNER_INTERFACE		0x0001 /*I-30xx Scanner Interface*/
/* Vendor: 5555: Genroco, Inc */
#define DEVICE_GENROCO_TURBOSTOR_HFP_832_HIPPI		0x0003 /*TURBOstor HFP-832 [HiPPI NIC]*/
/* Vendor: 5654: VoiceTronix Pty Ltd */
#define DEVICE_VOICETRONIX_OPENSWITCH12			0x3132 /*OpenSwitch12*/
/* Vendor: 5700: Netpower */
/* Vendor: 5851: Exacq Technologies */
/* Vendor: 6356: UltraStor */
/* Vendor: 6374: c't Magazin für Computertechnik */
#define DEVICE_C_T_GPPCI					0x6773 /*GPPCI*/
/* Vendor: 6409: Logitec Corp. */
/* Vendor: 6666: Decision Computer International Co. */
#define DEVICE_DECISION_PCCOM4				0x0001 /*PCCOM4*/
#define DEVICE_DECISION_PCCOM8				0x0002 /*PCCOM8*/
/* Vendor: 7604: O.N. Electronic Co Ltd. */
/* Vendor: 7bde: MIDAC Corporation */
/* Vendor: 7fed: PowerTV */
/* Vendor: 8008: Quancom Electronic GmbH */
#define DEVICE_QUANCOM_WDOG1_PCI_WATCHDOG_1		0x0010 /*WDOG1 [PCI-Watchdog 1]*/
#define DEVICE_QUANCOM_PWDOG2_PCI_WATCHDOG_2		0x0011 /*PWDOG2 [PCI-Watchdog 2]*/
/* Vendor: 807d: Asustek Computer, Inc. */
/* Vendor: 8086: Intel Corp. */
#define DEVICE_INTEL_82379AB				0x0007 /*82379AB*/
#define DEVICE_INTEL_EXTENDED_EXPRESS_SYSTEM		0x0008 /*Extended Express System Support Controller*/
	#define SUBDEVICE_INTEL_WORLDMARK_4300_INCA	0x00081000 /*WorldMark 4300 INCA ASIC*/
#define DEVICE_INTEL_21145_FAST_ETHERNET		0x0039 /*21145 Fast Ethernet*/
#define DEVICE_INTEL_82437FX				0x0122 /*82437FX*/
#define DEVICE_INTEL_80303_I_O_PROCESSOR		0x0309 /*80303 I/O Processor PCI-to-PCI Bridge*/
#define DEVICE_INTEL_80312_I_O_COMPANION		0x030d /*80312 I/O Companion Chip PCI-to-PCI Bridge*/
#define DEVICE_INTEL_6700_6702PXH_I_OXAPIC_INTERRUPT	0x0326 /*6700/6702PXH I/OxAPIC Interrupt Controller A*/
#define DEVICE_INTEL_6700PXH_I_OXAPIC_INTERRUPT		0x0327 /*6700PXH I/OxAPIC Interrupt Controller B*/
#define DEVICE_INTEL_6700PXH_PCI_EXPRESS_TO_PCI		0x0329 /*6700PXH PCI Express-to-PCI Bridge A*/
#define DEVICE_INTEL_6700PXH_PCI_EXPRESS_TO_PCI2	0x032a /*6700PXH PCI Express-to-PCI Bridge B*/
#define DEVICE_INTEL_6702PXH_PCI_EXPRESS_TO_PCI		0x032c /*6702PXH PCI Express-to-PCI Bridge A*/
#define DEVICE_INTEL_80332_DOBSON_I_O			0x0330 /*80332 [Dobson] I/O processor*/
#define DEVICE_INTEL_80332_DOBSON_I_O2			0x0331 /*80332 [Dobson] I/O processor*/
#define DEVICE_INTEL_80332_DOBSON_I_O3			0x0332 /*80332 [Dobson] I/O processor*/
#define DEVICE_INTEL_80332_DOBSON_I_O4			0x0333 /*80332 [Dobson] I/O processor*/
#define DEVICE_INTEL_80332_DOBSON_I_O5			0x0334 /*80332 [Dobson] I/O processor*/
#define DEVICE_INTEL_80331_LINDSAY_I_O			0x0335 /*80331 [Lindsay] I/O processor*/
#define DEVICE_INTEL_80331_LINDSAY_I_O2			0x0336 /*80331 [Lindsay] I/O processor*/
#define DEVICE_INTEL_41210_LANAI_SERIAL			0x0340 /*41210 [Lanai] Serial to Parallel PCI Bridge*/
#define DEVICE_INTEL_41210_LANAI_SERIAL2		0x0341 /*41210 [Lanai] Serial to Parallel PCI Bridge*/
#define DEVICE_INTEL_82375EB_SB_PCI_TO			0x0482 /*82375EB/SB PCI to EISA Bridge*/
#define DEVICE_INTEL_82424TX_ZX_SATURN_CPU		0x0483 /*82424TX/ZX [Saturn] CPU to PCI bridge*/
#define DEVICE_INTEL_82378ZB_IB_82379AB_SIO		0x0484 /*82378ZB/IB, 82379AB (SIO, SIO.A) PCI to ISA Bridge*/
#define DEVICE_INTEL_82425EX_ZX_ARIES_PCISET		0x0486 /*82425EX/ZX [Aries] PCIset with ISA bridge*/
#define DEVICE_INTEL_82434LX_NX_MERCURY_NEPTUNE_PROCESSOR	0x04a3 /*82434LX/NX [Mercury/Neptune] Processor to PCI bridge*/
#define DEVICE_INTEL_82437FX_TRITON_FX			0x04d0 /*82437FX [Triton FX]*/
#define DEVICE_INTEL_E8870_PROCESSOR_BUS		0x0500 /*E8870 Processor bus control*/
#define DEVICE_INTEL_E8870_MEMORY_CONTROLLER		0x0501 /*E8870 Memory controller*/
#define DEVICE_INTEL_E8870_SCALABILITY_PORT		0x0502 /*E8870 Scalability Port 0*/
#define DEVICE_INTEL_E8870_SCALABILITY_PORT2		0x0503 /*E8870 Scalability Port 1*/
#define DEVICE_INTEL_E8870IO_HUB_INTERFACE		0x0510 /*E8870IO Hub Interface Port 0 registers (8-bit compatibility port)*/
#define DEVICE_INTEL_E8870IO_HUB_INTERFACE2		0x0511 /*E8870IO Hub Interface Port 1 registers*/
#define DEVICE_INTEL_E8870IO_HUB_INTERFACE3		0x0512 /*E8870IO Hub Interface Port 2 registers*/
#define DEVICE_INTEL_E8870IO_HUB_INTERFACE4		0x0513 /*E8870IO Hub Interface Port 3 registers*/
#define DEVICE_INTEL_E8870IO_HUB_INTERFACE5		0x0514 /*E8870IO Hub Interface Port 4 registers*/
#define DEVICE_INTEL_E8870IO_GENERAL_SIOH		0x0515 /*E8870IO General SIOH registers*/
#define DEVICE_INTEL_E8870IO_RAS_REGISTERS		0x0516 /*E8870IO RAS registers*/
#define DEVICE_INTEL_E8870SP_SCALABILITY_PORT		0x0530 /*E8870SP Scalability Port 0 registers*/
#define DEVICE_INTEL_E8870SP_SCALABILITY_PORT2		0x0531 /*E8870SP Scalability Port 1 registers*/
#define DEVICE_INTEL_E8870SP_SCALABILITY_PORT3		0x0532 /*E8870SP Scalability Port 2 registers*/
#define DEVICE_INTEL_E8870SP_SCALABILITY_PORT4		0x0533 /*E8870SP Scalability Port 3 registers*/
#define DEVICE_INTEL_E8870SP_SCALABILITY_PORT5		0x0534 /*E8870SP Scalability Port 4 registers*/
#define DEVICE_INTEL_E8870SP_SCALABILITY_PORT6		0x0535 /*E8870SP Scalability Port 5 registers*/
#define DEVICE_INTEL_E8870SP_INTERLEAVE_REGISTERS	0x0536 /*E8870SP Interleave registers 0 and 1*/
#define DEVICE_INTEL_E8870SP_INTERLEAVE_REGISTERS2	0x0537 /*E8870SP Interleave registers 2 and 3*/
#define DEVICE_INTEL_RAID_CONTROLLER			0x0600 /*RAID Controller*/
	#define SUBDEVICE_INTEL_ICP_VORTEX_GDT8546RZ	0x808601c1 /*ICP Vortex GDT8546RZ*/
	#define SUBDEVICE_INTEL_SCRU32			0x808601f7 /*SCRU32*/
#define DEVICE_INTEL_80303_I_O_PROCESSOR2		0x061f /*80303 I/O Processor*/
#define DEVICE_INTEL_80960RP_I960_RP			0x0960 /*80960RP [i960 RP Microprocessor/Bridge]*/
#define DEVICE_INTEL_80960RM_I960RM_BRIDGE		0x0962 /*80960RM [i960RM Bridge]*/
#define DEVICE_INTEL_80960RP_I960_RP2			0x0964 /*80960RP [i960 RP Microprocessor/Bridge]*/
#define DEVICE_INTEL_82542_GIGABIT_ETHERNET		0x1000 /*82542 Gigabit Ethernet Controller*/
	#define SUBDEVICE_INTEL_NC1632_GIGABIT_ETHERNET	0x0e11b0df /*NC1632 Gigabit Ethernet Adapter (1000-SX)*/
	#define SUBDEVICE_INTEL_NC1633_GIGABIT_ETHERNET	0x0e11b0e0 /*NC1633 Gigabit Ethernet Adapter (1000-LX)*/
	#define SUBDEVICE_INTEL_NC1634_GIGABIT_ETHERNET	0x0e11b123 /*NC1634 Gigabit Ethernet Adapter (1000-SX)*/
	#define SUBDEVICE_INTEL_NETFINITY_GIGABIT_ETHERNET	0x10140119 /*Netfinity Gigabit Ethernet SX Adapter*/
	#define SUBDEVICE_INTEL_PRO_1000_GIGABIT_SERVER	0x80861000 /*PRO/1000 Gigabit Server Adapter*/
#define DEVICE_INTEL_82543GC_GIGABIT_ETHERNET		0x1001 /*82543GC Gigabit Ethernet Controller (Fiber)*/
	#define SUBDEVICE_INTEL_NC6136_GIGABIT_SERVER	0x0e11004a /*NC6136 Gigabit Server Adapter*/
	#define SUBDEVICE_INTEL_NETFINITY_GIGABIT_ETHERNET2	0x101401ea /*Netfinity Gigabit Ethernet SX Adapter*/
	#define SUBDEVICE_INTEL_PRO_1000_F_SERVER	0x80861002 /*PRO/1000 F Server Adapter*/
	#define SUBDEVICE_INTEL_PRO_1000_F_SERVER2	0x80861003 /*PRO/1000 F Server Adapter*/
#define DEVICE_INTEL_PRO_100_LAN_MODEM			0x1002 /*Pro 100 LAN+Modem 56 Cardbus II*/
	#define SUBDEVICE_INTEL_PRO_100_LAN_MODEM	0x8086200e /*Pro 100 LAN+Modem 56 Cardbus II*/
	#define SUBDEVICE_INTEL_PRO_100_SR		0x80862013 /*Pro 100 SR Mobile Combo Adapter*/
	#define SUBDEVICE_INTEL_PRO_100_S		0x80862017 /*Pro 100 S Combo Mobile Adapter*/
#define DEVICE_INTEL_82543GC_GIGABIT_ETHERNET2		0x1004 /*82543GC Gigabit Ethernet Controller (Copper)*/
	#define SUBDEVICE_INTEL_NC7132_GIGABIT_UPGRADE	0x0e110049 /*NC7132 Gigabit Upgrade Module*/
	#define SUBDEVICE_INTEL_NC7131_GIGABIT_SERVER	0x0e11b1a4 /*NC7131 Gigabit Server Adapter*/
	#define SUBDEVICE_INTEL_GIGABIT_ETHERNET_SERVER	0x101410f2 /*Gigabit Ethernet Server Adapter*/
	#define SUBDEVICE_INTEL_PRO_1000_T_SERVER	0x80861004 /*PRO/1000 T Server Adapter*/
	#define SUBDEVICE_INTEL_PRO_1000_T_SERVER2	0x80862004 /*PRO/1000 T Server Adapter*/
#define DEVICE_INTEL_82544EI_GIGABIT_ETHERNET		0x1008 /*82544EI Gigabit Ethernet Controller (Copper)*/
	#define SUBDEVICE_INTEL_ISERIES_1000_100_10_ETHERNET	0x10140269 /*iSeries 1000/100/10 Ethernet Adapter*/
	#define SUBDEVICE_INTEL_PRO_1000_XT_NETWORK	0x1028011c /*PRO/1000 XT Network Connection*/
	#define SUBDEVICE_INTEL_PRO_1000_XT_SERVER	0x80861107 /*PRO/1000 XT Server Adapter*/
	#define SUBDEVICE_INTEL_PRO_1000_XT_SERVER2	0x80862107 /*PRO/1000 XT Server Adapter*/
	#define SUBDEVICE_INTEL_PRO_1000_XT_SERVER3	0x80862110 /*PRO/1000 XT Server Adapter*/
	#define SUBDEVICE_INTEL_PRO_1000_XT_NETWORK2	0x80863108 /*PRO/1000 XT Network Connection*/
#define DEVICE_INTEL_82544EI_GIGABIT_ETHERNET2		0x1009 /*82544EI Gigabit Ethernet Controller (Fiber)*/
	#define SUBDEVICE_INTEL_ISERIES_GIGABIT_ETHERNET	0x10140268 /*iSeries Gigabit Ethernet Adapter*/
	#define SUBDEVICE_INTEL_PRO_1000_XF_SERVER	0x80861109 /*PRO/1000 XF Server Adapter*/
	#define SUBDEVICE_INTEL_PRO_1000_XF_SERVER2	0x80862109 /*PRO/1000 XF Server Adapter*/
#define DEVICE_INTEL_82544GC_GIGABIT_ETHERNET		0x100c /*82544GC Gigabit Ethernet Controller (Copper)*/
	#define SUBDEVICE_INTEL_PRO_1000_T_DESKTOP	0x80861112 /*PRO/1000 T Desktop Adapter*/
	#define SUBDEVICE_INTEL_PRO_1000_T_DESKTOP2	0x80862112 /*PRO/1000 T Desktop Adapter*/
#define DEVICE_INTEL_82544GC_GIGABIT_ETHERNET2		0x100d /*82544GC Gigabit Ethernet Controller (LOM)*/
	#define SUBDEVICE_INTEL_PRO_1000_XT_NETWORK3	0x10280123 /*PRO/1000 XT Network Connection*/
	#define SUBDEVICE_INTEL_82544GC_BASED_NETWORK	0x1079891f /*82544GC Based Network Connection*/
	#define SUBDEVICE_INTEL_CT8_MAINBOARD		0x4c531080 /*CT8 mainboard*/
	#define SUBDEVICE_INTEL_82544GC_BASED_NETWORK2	0x8086110d /*82544GC Based Network Connection*/
#define DEVICE_INTEL_82540EM_GIGABIT_ETHERNET		0x100e /*82540EM Gigabit Ethernet Controller*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_NETWORK	0x10140265 /*PRO/1000 MT Network Connection*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_NETWORK2	0x10140267 /*PRO/1000 MT Network Connection*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_NETWORK3	0x1014026a /*PRO/1000 MT Network Connection*/
	#define SUBDEVICE_INTEL_OPTIPLEX_GX260		0x1028002e /*Optiplex GX260*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_NETWORK4	0x10280151 /*PRO/1000 MT Network Connection*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_DESKTOP	0x107b8920 /*PRO/1000 MT Desktop Adapter*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_DESKTOP2	0x8086001e /*PRO/1000 MT Desktop Adapter*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_DESKTOP3	0x8086002e /*PRO/1000 MT Desktop Adapter*/
#define DEVICE_INTEL_82545EM_GIGABIT_ETHERNET		0x100f /*82545EM Gigabit Ethernet Controller (Copper)*/
	#define SUBDEVICE_INTEL_ISERIES_1000_100_10_ETHERNET2	0x10140269 /*iSeries 1000/100/10 Ethernet Adapter*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_NETWORK5	0x1014028e /*PRO/1000 MT Network Connection*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_NETWORK6	0x80861000 /*PRO/1000 MT Network Connection*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_SERVER	0x80861001 /*PRO/1000 MT Server Adapter*/
#define DEVICE_INTEL_82546EB_GIGABIT_ETHERNET		0x1010 /*82546EB Gigabit Ethernet Controller (Copper)*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_DUAL	0x1014027c /*PRO/1000 MT Dual Port Network Adapter*/
	#define SUBDEVICE_INTEL_RESLINK_X		0x18fb7872 /*RESlink-X*/
	#define SUBDEVICE_INTEL_CT8_MAINBOARD2		0x4c531080 /*CT8 mainboard*/
	#define SUBDEVICE_INTEL_CA3_CR3_MAINBOARD	0x4c5310a0 /*CA3/CR3 mainboard*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_DUAL2	0x80861011 /*PRO/1000 MT Dual Port Server Adapter*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_DUAL3	0x8086101a /*PRO/1000 MT Dual Port Network Adapter*/
	#define SUBDEVICE_INTEL_SE7501HG2_MAINBOARD	0x80863424 /*SE7501HG2 Mainboard*/
#define DEVICE_INTEL_82545EM_GIGABIT_ETHERNET2		0x1011 /*82545EM Gigabit Ethernet Controller (Fiber)*/
	#define SUBDEVICE_INTEL_ISERIES_GIGABIT_ETHERNET2	0x10140268 /*iSeries Gigabit Ethernet Adapter*/
	#define SUBDEVICE_INTEL_PRO_1000_MF_SERVER	0x80861002 /*PRO/1000 MF Server Adapter*/
	#define SUBDEVICE_INTEL_PRO_1000_MF_SERVER2	0x80861003 /*PRO/1000 MF Server Adapter (LX)*/
#define DEVICE_INTEL_82546EB_GIGABIT_ETHERNET2		0x1012 /*82546EB Gigabit Ethernet Controller (Fiber)*/
	#define SUBDEVICE_INTEL_PRO_1000_MF_DUAL	0x80861012 /*PRO/1000 MF Dual Port Server Adapter*/
#define DEVICE_INTEL_82541EI_GIGABIT_ETHERNET		0x1013 /*82541EI Gigabit Ethernet Controller (Copper)*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_NETWORK7	0x80860013 /*PRO/1000 MT Network Connection*/
	#define SUBDEVICE_INTEL_IBM_THINKCENTRE_NETWORK	0x80861013 /*IBM ThinkCentre Network Card*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_DESKTOP4	0x80861113 /*PRO/1000 MT Desktop Adapter*/
#define DEVICE_INTEL_82541ER_GIGABIT_ETHERNET		0x1014 /*82541ER Gigabit Ethernet Controller*/
#define DEVICE_INTEL_82540EM_GIGABIT_ETHERNET2		0x1015 /*82540EM Gigabit Ethernet Controller (LOM)*/
#define DEVICE_INTEL_82540EP_GIGABIT_ETHERNET		0x1016 /*82540EP Gigabit Ethernet Controller (LOM)*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_MOBILE	0x1014052c /*PRO/1000 MT Mobile Connection*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_MOBILE2	0x11790001 /*PRO/1000 MT Mobile Connection*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_MOBILE3	0x80861016 /*PRO/1000 MT Mobile Connection*/
#define DEVICE_INTEL_82540EP_GIGABIT_ETHERNET2		0x1017 /*82540EP Gigabit Ethernet Controller (LOM)*/
	#define SUBDEVICE_INTEL_PR0_1000_MT_DESKTOP	0x80861017 /*PR0/1000 MT Desktop Connection*/
#define DEVICE_INTEL_82541EI_GIGABIT_ETHERNET2		0x1018 /*82541EI Gigabit Ethernet Controller*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_DESKTOP5	0x80861018 /*PRO/1000 MT Desktop Adapter*/
#define DEVICE_INTEL_82547EI_GIGABIT_ETHERNET		0x1019 /*82547EI Gigabit Ethernet Controller (LOM)*/
	#define SUBDEVICE_INTEL_GA_8IPE1000_PRO2_MOTHERBOARD	0x14581019 /*GA-8IPE1000 Pro2 motherboard (865PE)*/
	#define SUBDEVICE_INTEL_PRO_1000_CT_DESKTOP	0x80861019 /*PRO/1000 CT Desktop Connection*/
	#define SUBDEVICE_INTEL_D865PERL_MAINBOARD	0x8086301f /*D865PERL mainboard*/
	#define SUBDEVICE_INTEL_S875WP1_E_MAINBOARD	0x80863427 /*S875WP1-E mainboard*/
#define DEVICE_INTEL_82546EB_GIGABIT_ETHERNET3		0x101d /*82546EB Gigabit Ethernet Controller*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_QUAD	0x80861000 /*PRO/1000 MT Quad Port Server Adapter*/
#define DEVICE_INTEL_82540EP_GIGABIT_ETHERNET3		0x101e /*82540EP Gigabit Ethernet Controller (Mobile)*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_MOBILE4	0x10140549 /*PRO/1000 MT Mobile Connection*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_MOBILE5	0x11790001 /*PRO/1000 MT Mobile Connection*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_MOBILE6	0x8086101e /*PRO/1000 MT Mobile Connection*/
#define DEVICE_INTEL_82545GM_GIGABIT_ETHERNET		0x1026 /*82545GM Gigabit Ethernet Controller*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_SERVER2	0x80861000 /*PRO/1000 MT Server Connection*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_SERVER3	0x80861001 /*PRO/1000 MT Server Adapter*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_SERVER4	0x80861002 /*PRO/1000 MT Server Adapter*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_SERVER5	0x80861026 /*PRO/1000 MT Server Connection*/
#define DEVICE_INTEL_82545GM_GIGABIT_ETHERNET2		0x1027 /*82545GM Gigabit Ethernet Controller*/
	#define SUBDEVICE_INTEL_PRO_1000_MF_SERVER3	0x80861001 /*PRO/1000 MF Server Adapter(LX)*/
	#define SUBDEVICE_INTEL_PRO_1000_MF_SERVER4	0x80861002 /*PRO/1000 MF Server Adapter(LX)*/
	#define SUBDEVICE_INTEL_PRO_1000_MF_SERVER5	0x80861003 /*PRO/1000 MF Server Adapter(LX)*/
	#define SUBDEVICE_INTEL_PRO_1000_MF_SERVER6	0x80861027 /*PRO/1000 MF Server Adapter*/
#define DEVICE_INTEL_82545GM_GIGABIT_ETHERNET3		0x1028 /*82545GM Gigabit Ethernet Controller*/
	#define SUBDEVICE_INTEL_PRO_1000_MB_SERVER	0x80861028 /*PRO/1000 MB Server Adapter*/
#define DEVICE_INTEL_82559_ETHERNET_CONTROLLER		0x1029 /*82559 Ethernet Controller*/
#define DEVICE_INTEL_82559_INBUSINESS_10_100		0x1030 /*82559 InBusiness 10/100*/
#define DEVICE_INTEL_82801CAM_ICH3_PRO_100		0x1031 /*82801CAM (ICH3) PRO/100 VE (LOM) Ethernet Controller*/
	#define SUBDEVICE_INTEL_THINKPAD_A_T_X_SERIES	0x10140209 /*ThinkPad A/T/X Series*/
	#define SUBDEVICE_INTEL_VAIO_PCG_GR214EP_GR214MP_GR215MP_GR314MP_GR315MP	0x104d80e7 /*Vaio PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VE	0x107b5350 /*EtherExpress PRO/100 VE*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VE2	0x11790001 /*EtherExpress PRO/100 VE*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VE3	0x144dc000 /*EtherExpress PRO/100 VE*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VE4	0x144dc001 /*EtherExpress PRO/100 VE*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VE5	0x144dc003 /*EtherExpress PRO/100 VE*/
	#define SUBDEVICE_INTEL_VPR_MATRIX_170B4	0x144dc006 /*vpr Matrix 170B4*/
#define DEVICE_INTEL_82801CAM_ICH3_PRO_1002		0x1032 /*82801CAM (ICH3) PRO/100 VE Ethernet Controller*/
#define DEVICE_INTEL_82801CAM_ICH3_PRO_1003		0x1033 /*82801CAM (ICH3) PRO/100 VM (LOM) Ethernet Controller*/
#define DEVICE_INTEL_82801CAM_ICH3_PRO_1004		0x1034 /*82801CAM (ICH3) PRO/100 VM Ethernet Controller*/
#define DEVICE_INTEL_82801CAM_ICH3_82562EH_LOM		0x1035 /*82801CAM (ICH3)/82562EH (LOM)  Ethernet Controller*/
#define DEVICE_INTEL_82801CAM_ICH3_82562EH		0x1036 /*82801CAM (ICH3) 82562EH Ethernet Controller*/
#define DEVICE_INTEL_82801CAM_ICH3_CHIPSET		0x1037 /*82801CAM (ICH3) Chipset Ethernet Controller*/
#define DEVICE_INTEL_82801CAM_ICH3_PRO_1005		0x1038 /*82801CAM (ICH3) PRO/100 VM (KM) Ethernet Controller*/
#define DEVICE_INTEL_82801DB_PRO_100_VE			0x1039 /*82801DB PRO/100 VE (LOM) Ethernet Controller*/
	#define SUBDEVICE_INTEL_NETVISTA_A30P		0x10140267 /*NetVista A30p*/
#define DEVICE_INTEL_82801DB_PRO_100_VE2		0x103a /*82801DB PRO/100 VE (CNR) Ethernet Controller*/
#define DEVICE_INTEL_82801DB_PRO_100_VM			0x103b /*82801DB PRO/100 VM (LOM) Ethernet Controller*/
#define DEVICE_INTEL_82801DB_PRO_100_VM2		0x103c /*82801DB PRO/100 VM (CNR) Ethernet Controller*/
#define DEVICE_INTEL_82801DB_PRO_100_VE3		0x103d /*82801DB PRO/100 VE (MOB) Ethernet Controller*/
#define DEVICE_INTEL_82801DB_PRO_100_VM3		0x103e /*82801DB PRO/100 VM (MOB) Ethernet Controller*/
#define DEVICE_INTEL_536EP_DATA_FAX			0x1040 /*536EP Data Fax Modem*/
	#define SUBDEVICE_INTEL_V_9X_DSP_DATA		0x16be1040 /*V.9X DSP Data Fax Modem*/
#define DEVICE_INTEL_PRO_WIRELESS_LAN_2100		0x1043 /*PRO/Wireless LAN 2100 3B Mini PCI Adapter*/
	#define SUBDEVICE_INTEL_MIM2000_CENTRINO	0x80862527 /*MIM2000/Centrino*/
#define DEVICE_INTEL_PRO_10GBE_LR_SERVER		0x1048 /*PRO/10GbE LR Server Adapter*/
	#define SUBDEVICE_INTEL_PRO_10GBE_LR_SERVER	0x8086a01f /*PRO/10GbE LR Server Adapter*/
	#define SUBDEVICE_INTEL_PRO_10GBE_LR_SERVER2	0x8086a11f /*PRO/10GbE LR Server Adapter*/
#define DEVICE_INTEL_82562EZ_10_100_ETHERNET		0x1050 /*82562EZ 10/100 Ethernet Controller*/
	#define SUBDEVICE_INTEL_865PE_NEO2_MS_6728	0x1462728c /*865PE Neo2 (MS-6728)*/
	#define SUBDEVICE_INTEL_MS_6758_875P_NEO	0x1462758c /*MS-6758 (875P Neo)*/
	#define SUBDEVICE_INTEL_S875WP1_E_MAINBOARD2	0x80863427 /*S875WP1-E mainboard*/
#define DEVICE_INTEL_82801EB_ER_ICH5_ICH5R_INTEGRATED	0x1051 /*82801EB/ER (ICH5/ICH5R) integrated LAN Controller*/
#define DEVICE_INTEL_82551QM_ETHERNET_CONTROLLER	0x1059 /*82551QM Ethernet Controller*/
#define DEVICE_INTEL_82562ET_EZ_GT_GZ_PRO_100		0x1064 /*82562ET/EZ/GT/GZ - PRO/100 VE (LOM) Ethernet Controller*/
#define DEVICE_INTEL_82562ET_EZ_GT_GZ_PRO_1002		0x1065 /*82562ET/EZ/GT/GZ - PRO/100 VE Ethernet Controller*/
#define DEVICE_INTEL_82562_EM_EX_GX			0x1066 /*82562 EM/EX/GX - PRO/100 VM (LOM) Ethernet Controller*/
#define DEVICE_INTEL_82562_EM_EX_GX2			0x1067 /*82562 EM/EX/GX - PRO/100 VM Ethernet Controller*/
#define DEVICE_INTEL_82562ET_EZ_GT_GZ_PRO_1003		0x1068 /*82562ET/EZ/GT/GZ - PRO/100 VE (LOM) Ethernet Controller Mobile*/
#define DEVICE_INTEL_82562_EM_EX_GX3			0x1069 /*82562 EM/EX/GX - PRO/100 VM (LOM) Ethernet Controller Mobile*/
#define DEVICE_INTEL_82562G_T_PRO_100			0x106a /*82562G \t- PRO/100 VE (LOM) Ethernet Controller*/
#define DEVICE_INTEL_82562G_T_PRO_1002			0x106b /*82562G \t- PRO/100 VE Ethernet Controller Mobile*/
#define DEVICE_INTEL_82547GI_GIGABIT_ETHERNET		0x1075 /*82547GI Gigabit Ethernet Controller*/
	#define SUBDEVICE_INTEL_POWEREDGE_750		0x10280165 /*PowerEdge 750*/
	#define SUBDEVICE_INTEL_PRO_1000_CT_NETWORK	0x80860075 /*PRO/1000 CT Network Connection*/
	#define SUBDEVICE_INTEL_PRO_1000_CT_NETWORK2	0x80861075 /*PRO/1000 CT Network Connection*/
#define DEVICE_INTEL_82541GI_PI_GIGABIT_ETHERNET	0x1076 /*82541GI/PI Gigabit Ethernet Controller*/
	#define SUBDEVICE_INTEL_POWEREDGE_7502		0x10280165 /*PowerEdge 750*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_NETWORK8	0x80860076 /*PRO/1000 MT Network Connection*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_NETWORK9	0x80861076 /*PRO/1000 MT Network Connection*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_DESKTOP6	0x80861176 /*PRO/1000 MT Desktop Adapter*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_DESKTOP7	0x80861276 /*PRO/1000 MT Desktop Adapter*/
#define DEVICE_INTEL_82541GI_GIGABIT_ETHERNET		0x1077 /*82541GI Gigabit Ethernet Controller*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_MOBILE7	0x11790001 /*PRO/1000 MT Mobile Connection*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_MOBILE8	0x80860077 /*PRO/1000 MT Mobile Connection*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_MOBILE9	0x80861077 /*PRO/1000 MT Mobile Connection*/
#define DEVICE_INTEL_82541EI_GIGABIT_ETHERNET3		0x1078 /*82541EI Gigabit Ethernet Controller*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_NETWORK10	0x80861078 /*PRO/1000 MT Network Connection*/
#define DEVICE_INTEL_82546GB_GIGABIT_ETHERNET		0x1079 /*82546GB Gigabit Ethernet Controller*/
	#define SUBDEVICE_INTEL_HP_DUAL_PORT		0x103c12a6 /*HP Dual Port 1000Base-T [A9900A]*/
	#define SUBDEVICE_INTEL_HP_CORE_DUAL		0x103c12cf /*HP Core Dual Port 1000Base-T [AB352A]*/
	#define SUBDEVICE_INTEL_CX9_VX9			0x4c531090 /*Cx9 / Vx9 mainboard*/
	#define SUBDEVICE_INTEL_CL9_MAINBOARD		0x4c5310b0 /*CL9 mainboard*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_DUAL4	0x80860079 /*PRO/1000 MT Dual Port Network Connection*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_DUAL5	0x80861079 /*PRO/1000 MT Dual Port Network Connection*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_DUAL6	0x80861179 /*PRO/1000 MT Dual Port Network Connection*/
	#define SUBDEVICE_INTEL_PRO_1000_MT_DUAL7	0x8086117a /*PRO/1000 MT Dual Port Server Adapter*/
#define DEVICE_INTEL_82546GB_GIGABIT_ETHERNET2		0x107a /*82546GB Gigabit Ethernet Controller*/
	#define SUBDEVICE_INTEL_HP_DUAL_PORT2		0x103c12a8 /*HP Dual Port 1000base-SX [A9899A]*/
	#define SUBDEVICE_INTEL_PRO_1000_MF_DUAL2	0x8086107a /*PRO/1000 MF Dual Port Server Adapter*/
	#define SUBDEVICE_INTEL_PRO_1000_MF_DUAL3	0x8086127a /*PRO/1000 MF Dual Port Server Adapter*/
#define DEVICE_INTEL_82546GB_GIGABIT_ETHERNET3		0x107b /*82546GB Gigabit Ethernet Controller*/
	#define SUBDEVICE_INTEL_PRO_1000_MB_DUAL	0x8086007b /*PRO/1000 MB Dual Port Server Connection*/
	#define SUBDEVICE_INTEL_PRO_1000_MB_DUAL2	0x8086107b /*PRO/1000 MB Dual Port Server Connection*/
#define DEVICE_INTEL_PRO_1000_MF_SERVER			0x1107 /*PRO/1000 MF Server Adapter (LX)*/
#define DEVICE_INTEL_82815_815_CHIPSET			0x1130 /*82815 815 Chipset Host Bridge and Memory Controller Hub*/
	#define SUBDEVICE_INTEL_TRAVELMATE_612_TX	0x10251016 /*Travelmate 612 TX*/
	#define SUBDEVICE_INTEL_TUSL2_C_MAINBOARD	0x10438027 /*TUSL2-C Mainboard*/
	#define SUBDEVICE_INTEL_VAIO_PCG_FX403		0x104d80df /*Vaio PCG-FX403*/
	#define SUBDEVICE_INTEL_D815EEA2_MAINBOARD	0x80864532 /*D815EEA2 mainboard*/
	#define SUBDEVICE_INTEL_D815EGEW_MAINBOARD	0x80864557 /*D815EGEW Mainboard*/
#define DEVICE_INTEL_82815_815_CHIPSET2			0x1131 /*82815 815 Chipset AGP Bridge*/
#define DEVICE_INTEL_82815_CGC_CHIPSET			0x1132 /*82815 CGC [Chipset Graphics Controller]*/
	#define SUBDEVICE_INTEL_TRAVELMATE_612_TX2	0x10251016 /*Travelmate 612 TX*/
	#define SUBDEVICE_INTEL_VAIO_PCG_FX4032		0x104d80df /*Vaio PCG-FX403*/
	#define SUBDEVICE_INTEL_D815EEA2_MAINBOARD2	0x80864532 /*D815EEA2 Mainboard*/
	#define SUBDEVICE_INTEL_D815EGEW_MAINBOARD2	0x80864557 /*D815EGEW Mainboard*/
#define DEVICE_INTEL_82806AA_PCI64_HUB			0x1161 /*82806AA PCI64 Hub Advanced Programmable Interrupt Controller*/
	#define SUBDEVICE_INTEL_82806AA_PCI64_HUB	0x80861161 /*82806AA PCI64 Hub APIC*/
#define DEVICE_INTEL_XSCALE_80200_BIG			0x1162 /*Xscale 80200 Big Endian Companion Chip*/
#define DEVICE_INTEL_INTEL_IXP1200_NETWORK		0x1200 /*Intel IXP1200 Network Processor*/
	#define SUBDEVICE_INTEL_AEP_SSL_ACCELERATOR	0x172a0000 /*AEP SSL Accelerator*/
#define DEVICE_INTEL_8255XER_82551IT_FAST_ETHERNET	0x1209 /*8255xER/82551IT Fast Ethernet Controller*/
	#define SUBDEVICE_INTEL_CT7_MAINBOARD		0x4c531050 /*CT7 mainboard*/
	#define SUBDEVICE_INTEL_CE7_MAINBOARD		0x4c531051 /*CE7 mainboard*/
	#define SUBDEVICE_INTEL_PC6_MAINBOARD		0x4c531070 /*PC6 mainboard*/
#define DEVICE_INTEL_82092AA_PCI_TO			0x1221 /*82092AA PCI to PCMCIA Bridge*/
#define DEVICE_INTEL_82092AA_IDE_CONTROLLER		0x1222 /*82092AA IDE Controller*/
#define DEVICE_INTEL_SAA7116				0x1223 /*SAA7116*/
#define DEVICE_INTEL_82452KX_GX_ORION			0x1225 /*82452KX/GX [Orion]*/
#define DEVICE_INTEL_82596_PRO_10_PCI			0x1226 /*82596 PRO/10 PCI*/
#define DEVICE_INTEL_82865_ETHEREXPRESS_PRO_100A	0x1227 /*82865 EtherExpress PRO/100A*/
#define DEVICE_INTEL_82556_ETHEREXPRESS_PRO_100		0x1228 /*82556 EtherExpress PRO/100 Smart*/
#define DEVICE_INTEL_82557_8_9_ETHERNET_PRO		0x1229 /*82557/8/9 [Ethernet Pro 100]*/
	#define SUBDEVICE_INTEL_82559_FAST_ETHERNET	0x0e113001 /*82559 Fast Ethernet LOM with Alert on LAN**/
	#define SUBDEVICE_INTEL_82559_FAST_ETHERNET2	0x0e113002 /*82559 Fast Ethernet LOM with Alert on LAN**/
	#define SUBDEVICE_INTEL_82559_FAST_ETHERNET3	0x0e113003 /*82559 Fast Ethernet LOM with Alert on LAN**/
	#define SUBDEVICE_INTEL_82559_FAST_ETHERNET4	0x0e113004 /*82559 Fast Ethernet LOM with Alert on LAN**/
	#define SUBDEVICE_INTEL_82559_FAST_ETHERNET5	0x0e113005 /*82559 Fast Ethernet LOM with Alert on LAN**/
	#define SUBDEVICE_INTEL_82559_FAST_ETHERNET6	0x0e113006 /*82559 Fast Ethernet LOM with Alert on LAN**/
	#define SUBDEVICE_INTEL_82559_FAST_ETHERNET7	0x0e113007 /*82559 Fast Ethernet LOM with Alert on LAN**/
	#define SUBDEVICE_INTEL_NC3120_FAST_ETHERNET	0x0e11b01e /*NC3120 Fast Ethernet NIC*/
	#define SUBDEVICE_INTEL_NC3122_FAST_ETHERNET	0x0e11b01f /*NC3122 Fast Ethernet NIC (dual port)*/
	#define SUBDEVICE_INTEL_NC1120_ETHERNET_NIC	0x0e11b02f /*NC1120 Ethernet NIC*/
	#define SUBDEVICE_INTEL_NETELLIGENT_10_100TX_NIC	0x0e11b04a /*Netelligent 10/100TX NIC with Wake on LAN*/
	#define SUBDEVICE_INTEL_NC3161_FAST_ETHERNET	0x0e11b0c6 /*NC3161 Fast Ethernet NIC (embedded, WOL)*/
	#define SUBDEVICE_INTEL_NC3160_FAST_ETHERNET	0x0e11b0c7 /*NC3160 Fast Ethernet NIC (embedded)*/
	#define SUBDEVICE_INTEL_NC3121_FAST_ETHERNET	0x0e11b0d7 /*NC3121 Fast Ethernet NIC (WOL)*/
	#define SUBDEVICE_INTEL_NC3131_FAST_ETHERNET	0x0e11b0dd /*NC3131 Fast Ethernet NIC (dual port)*/
	#define SUBDEVICE_INTEL_NC3132_FAST_ETHERNET	0x0e11b0de /*NC3132 Fast Ethernet Module (dual port)*/
	#define SUBDEVICE_INTEL_NC3133_FAST_ETHERNET	0x0e11b0e1 /*NC3133 Fast Ethernet Module (100-FX)*/
	#define SUBDEVICE_INTEL_NC3163_FAST_ETHERNET	0x0e11b134 /*NC3163 Fast Ethernet NIC (embedded, WOL)*/
	#define SUBDEVICE_INTEL_NC3162_FAST_ETHERNET	0x0e11b13c /*NC3162 Fast Ethernet NIC (embedded)*/
	#define SUBDEVICE_INTEL_NC3123_FAST_ETHERNET	0x0e11b144 /*NC3123 Fast Ethernet NIC (WOL)*/
	#define SUBDEVICE_INTEL_NC3134_FAST_ETHERNET	0x0e11b163 /*NC3134 Fast Ethernet NIC (dual port)*/
	#define SUBDEVICE_INTEL_NC3135_FAST_ETHERNET	0x0e11b164 /*NC3135 Fast Ethernet Upgrade Module (dual port)*/
	#define SUBDEVICE_INTEL_NC7131_GIGABIT_SERVER2	0x0e11b1a4 /*NC7131 Gigabit Server Adapter*/
	#define SUBDEVICE_INTEL_82558B_ETHERNET_PRO	0x1014005c /*82558B Ethernet Pro 10/100*/
	#define SUBDEVICE_INTEL_82559_FAST_ETHERNET8	0x101401bc /*82559 Fast Ethernet LAN On Motherboard*/
	#define SUBDEVICE_INTEL_10_100_ETHERNET_SERVER	0x101401f1 /*10/100 Ethernet Server Adapter*/
	#define SUBDEVICE_INTEL_10_100_ETHERNET_SERVER2	0x101401f2 /*10/100 Ethernet Server Adapter*/
	#define SUBDEVICE_INTEL_ETHERNET_PRO_100_S	0x10140207 /*Ethernet Pro/100 S*/
	#define SUBDEVICE_INTEL_10_100_DUAL_PORT	0x10140232 /*10/100 Dual Port Server Adapter*/
	#define SUBDEVICE_INTEL_THINKPAD_R30		0x1014023a /*ThinkPad R30*/
	#define SUBDEVICE_INTEL_NETFINITY_10_100	0x1014105c /*Netfinity 10/100*/
	#define SUBDEVICE_INTEL_THINKPAD_A22P		0x10142205 /*ThinkPad A22p*/
	#define SUBDEVICE_INTEL_10_100_ETHERJET_MANAGEMENT	0x1014305c /*10/100 EtherJet Management Adapter*/
	#define SUBDEVICE_INTEL_10_100_ETHERJET_ADAPTER	0x1014405c /*10/100 EtherJet Adapter with Alert on LAN*/
	#define SUBDEVICE_INTEL_10_100_ETHERJET_SECURE	0x1014505c /*10/100 EtherJet Secure Management Adapter*/
	#define SUBDEVICE_INTEL_10_100_ETHERJET_SECURE2	0x1014605c /*10/100 EtherJet Secure Management Adapter*/
	#define SUBDEVICE_INTEL_10_100_NETFINITY_10_100	0x1014705c /*10/100 Netfinity 10/100 Ethernet Security Adapter*/
	#define SUBDEVICE_INTEL_10_100_NETFINITY_10_1002	0x1014805c /*10/100 Netfinity 10/100 Ethernet Security Adapter*/
	#define SUBDEVICE_INTEL_POWEREDGE_2500_2550	0x1028009b /*PowerEdge 2500/2550*/
	#define SUBDEVICE_INTEL_POWEREDGE_1400		0x102800ce /*PowerEdge 1400*/
	#define SUBDEVICE_INTEL_PC_9821X_B06		0x10338000 /*PC-9821X-B06*/
	#define SUBDEVICE_INTEL_PK_UG_X006		0x10338016 /*PK-UG-X006*/
	#define SUBDEVICE_INTEL_PK_UG_X0062		0x1033801f /*PK-UG-X006*/
	#define SUBDEVICE_INTEL_PK_UG_X0063		0x10338026 /*PK-UG-X006*/
	#define SUBDEVICE_INTEL_82559_BASED_FAST_ETHERNET	0x10338063 /*82559-based Fast Ethernet Adapter*/
	#define SUBDEVICE_INTEL_82559_BASED_FAST_ETHERNET2	0x10338064 /*82559-based Fast Ethernet Adapter*/
	#define SUBDEVICE_INTEL_NETSERVER_10_100TX	0x103c10c0 /*NetServer 10/100TX*/
	#define SUBDEVICE_INTEL_NETSERVER_10_100TX2	0x103c10c3 /*NetServer 10/100TX*/
	#define SUBDEVICE_INTEL_NETSERVER_10_100TX3	0x103c10ca /*NetServer 10/100TX*/
	#define SUBDEVICE_INTEL_NETSERVER_10_100TX4	0x103c10cb /*NetServer 10/100TX*/
	#define SUBDEVICE_INTEL_NETSERVER_10_100TX5	0x103c10e3 /*NetServer 10/100TX*/
	#define SUBDEVICE_INTEL_NETSERVER_10_100TX6	0x103c10e4 /*NetServer 10/100TX*/
	#define SUBDEVICE_INTEL_NETSERVER_10_100TX7	0x103c1200 /*NetServer 10/100TX*/
	#define SUBDEVICE_INTEL_SMARTETHER100_SC1100	0x10c31100 /*SmartEther100 SC1100*/
	#define SUBDEVICE_INTEL_8255X_BASED_ETHERNET_ADAPTER	0x10cf1115 /*8255x-based Ethernet Adapter (10/100)*/
	#define SUBDEVICE_INTEL_8255X_BASED_ETHERNET_ADAPTER2	0x10cf1143 /*8255x-based Ethernet Adapter (10/100)*/
	#define SUBDEVICE_INTEL_8255X_BASED_ETHERNET_ADAPTER3	0x11790001 /*8255x-based Ethernet Adapter (10/100)*/
	#define SUBDEVICE_INTEL_PCI_FASTETHER_LAN	0x11790002 /*PCI FastEther LAN on Docker*/
	#define SUBDEVICE_INTEL_8255X_BASED_FAST_ETHERNET	0x11790003 /*8255x-based Fast Ethernet*/
	#define SUBDEVICE_INTEL_AT_2560_100		0x12592560 /*AT-2560 100*/
	#define SUBDEVICE_INTEL_AT_2560_100_FX		0x12592561 /*AT-2560 100 FX Ethernet Adapter*/
	#define SUBDEVICE_INTEL_NE10_100_ADAPTER	0x12660001 /*NE10/100 Adapter*/
	#define SUBDEVICE_INTEL_6221L_4U		0x13e91000 /*6221L-4U*/
	#define SUBDEVICE_INTEL_SEM_2000_MINIPCI_LAN	0x144d2501 /*SEM-2000 MiniPCI LAN Adapter*/
	#define SUBDEVICE_INTEL_SEM_2100IL_MINIPCI_LAN	0x144d2502 /*SEM-2100IL MiniPCI LAN Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100B_TX	0x16681100 /*EtherExpress PRO/100B (TX) (MiniPCI Ethernet+Modem)*/
	#define SUBDEVICE_INTEL_CT8_MAINBOARD3		0x4c531080 /*CT8 mainboard*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100B_TX2	0x80860001 /*EtherExpress PRO/100B (TX)*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100B_T4	0x80860002 /*EtherExpress PRO/100B (T4)*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_10	0x80860003 /*EtherExpress PRO/10+*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_WFM	0x80860004 /*EtherExpress PRO/100 WfM*/
	#define SUBDEVICE_INTEL_82557_10_100		0x80860005 /*82557 10/100*/
	#define SUBDEVICE_INTEL_82557_10_100_WITH	0x80860006 /*82557 10/100 with Wake on LAN*/
	#define SUBDEVICE_INTEL_82558_10_100_ADAPTER	0x80860007 /*82558 10/100 Adapter*/
	#define SUBDEVICE_INTEL_82558_10_100_WITH	0x80860008 /*82558 10/100 with Wake on LAN*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100	0x80860009 /*EtherExpress PRO/100+*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_MANAGEMENT	0x8086000a /*EtherExpress PRO/100+ Management Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_1002	0x8086000b /*EtherExpress PRO/100+*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_MANAGEMENT2	0x8086000c /*EtherExpress PRO/100+ Management Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_ALERT	0x8086000d /*EtherExpress PRO/100+ Alert On LAN II* Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_MANAGEMENT3	0x8086000e /*EtherExpress PRO/100+ Management Adapter with Alert On LAN**/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_DESKTOP	0x8086000f /*EtherExpress PRO/100 Desktop Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_S	0x80860010 /*EtherExpress PRO/100 S Management Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_S2	0x80860011 /*EtherExpress PRO/100 S Management Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_S3	0x80860012 /*EtherExpress PRO/100 S Advanced Management Adapter (D)*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_S4	0x80860013 /*EtherExpress PRO/100 S Advanced Management Adapter (E)*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_1003	0x80860030 /*EtherExpress PRO/100  Management Adapter with Alert On LAN* GC*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_DESKTOP2	0x80860031 /*EtherExpress PRO/100 Desktop Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_S5	0x80860040 /*EtherExpress PRO/100 S Desktop Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_S6	0x80860041 /*EtherExpress PRO/100 S Desktop Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_DESKTOP3	0x80860042 /*EtherExpress PRO/100 Desktop Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_S7	0x80860050 /*EtherExpress PRO/100 S Desktop Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_SERVER	0x80861009 /*EtherExpress PRO/100+ Server Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_SERVER2	0x8086100c /*EtherExpress PRO/100+ Server Adapter (PILA8470B)*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_S8	0x80861012 /*EtherExpress PRO/100 S Server Adapter (D)*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_S9	0x80861013 /*EtherExpress PRO/100 S Server Adapter (E)*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_S10	0x80861015 /*EtherExpress PRO/100 S Dual Port Server Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_DUAL	0x80861017 /*EtherExpress PRO/100+ Dual Port Server Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_MANAGEMENT4	0x80861030 /*EtherExpress PRO/100+ Management Adapter with Alert On LAN* G Server*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_S11	0x80861040 /*EtherExpress PRO/100 S Server Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_S12	0x80861041 /*EtherExpress PRO/100 S Server Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_SERVER3	0x80861042 /*EtherExpress PRO/100 Server Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_S13	0x80861050 /*EtherExpress PRO/100 S Server Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_SERVER4	0x80861051 /*EtherExpress PRO/100 Server Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_SERVER5	0x80861052 /*EtherExpress PRO/100 Server Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_DUAL2	0x808610f0 /*EtherExpress PRO/100+ Dual Port Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_S14	0x80862009 /*EtherExpress PRO/100 S Mobile Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_CARDBUS	0x8086200d /*EtherExpress PRO/100 Cardbus*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_LAN_V90	0x8086200e /*EtherExpress PRO/100 LAN+V90 Cardbus Modem*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_SR	0x8086200f /*EtherExpress PRO/100 SR Mobile Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_S15	0x80862010 /*EtherExpress PRO/100 S Mobile Combo Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_SR2	0x80862013 /*EtherExpress PRO/100 SR Mobile Combo Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_S16	0x80862016 /*EtherExpress PRO/100 S Mobile Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_S17	0x80862017 /*EtherExpress PRO/100 S Combo Mobile Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_SR3	0x80862018 /*EtherExpress PRO/100 SR Mobile Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_SR4	0x80862019 /*EtherExpress PRO/100 SR Combo Mobile Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_P	0x80862101 /*EtherExpress PRO/100 P Mobile Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_SP	0x80862102 /*EtherExpress PRO/100 SP Mobile Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_SP2	0x80862103 /*EtherExpress PRO/100 SP Mobile Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_SP3	0x80862104 /*EtherExpress PRO/100 SP Mobile Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_SP4	0x80862105 /*EtherExpress PRO/100 SP Mobile Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_P2	0x80862106 /*EtherExpress PRO/100 P Mobile Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_NETWORK	0x80862107 /*EtherExpress PRO/100 Network Connection*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_NETWORK2	0x80862108 /*EtherExpress PRO/100 Network Connection*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_P3	0x80862200 /*EtherExpress PRO/100 P Mobile Combo Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_P4	0x80862201 /*EtherExpress PRO/100 P Mobile Combo Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_SP5	0x80862202 /*EtherExpress PRO/100 SP Mobile Combo Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_MINIPCI	0x80862203 /*EtherExpress PRO/100+ MiniPCI*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_MINIPCI2	0x80862204 /*EtherExpress PRO/100+ MiniPCI*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_SP6	0x80862205 /*EtherExpress PRO/100 SP Mobile Combo Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_SP7	0x80862206 /*EtherExpress PRO/100 SP Mobile Combo Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_SP8	0x80862207 /*EtherExpress PRO/100 SP Mobile Combo Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_P5	0x80862208 /*EtherExpress PRO/100 P Mobile Combo Adapter*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_MINIPCI3	0x80862402 /*EtherExpress PRO/100+ MiniPCI*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_MINIPCI4	0x80862407 /*EtherExpress PRO/100+ MiniPCI*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_MINIPCI5	0x80862408 /*EtherExpress PRO/100+ MiniPCI*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_MINIPCI6	0x80862409 /*EtherExpress PRO/100+ MiniPCI*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_MINIPCI7	0x8086240f /*EtherExpress PRO/100+ MiniPCI*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_MINIPCI8	0x80862410 /*EtherExpress PRO/100+ MiniPCI*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_MINIPCI9	0x80862411 /*EtherExpress PRO/100+ MiniPCI*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_MINIPCI10	0x80862412 /*EtherExpress PRO/100+ MiniPCI*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_MINIPCI11	0x80862413 /*EtherExpress PRO/100+ MiniPCI*/
	#define SUBDEVICE_INTEL_82559_FAST_ETHERNET9	0x80863000 /*82559 Fast Ethernet LAN on Motherboard*/
	#define SUBDEVICE_INTEL_82559_FAST_ETHERNET10	0x80863001 /*82559 Fast Ethernet LOM with Basic Alert on LAN**/
	#define SUBDEVICE_INTEL_82559_FAST_ETHERNET11	0x80863002 /*82559 Fast Ethernet LOM with Alert on LAN II**/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_S18	0x80863006 /*EtherExpress PRO/100 S Network Connection*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_S19	0x80863007 /*EtherExpress PRO/100 S Network Connection*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_NETWORK3	0x80863008 /*EtherExpress PRO/100 Network Connection*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_S20	0x80863010 /*EtherExpress PRO/100 S Network Connection*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_S21	0x80863011 /*EtherExpress PRO/100 S Network Connection*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_NETWORK4	0x80863012 /*EtherExpress PRO/100 Network Connection*/
	#define SUBDEVICE_INTEL_SDS2_MAINBOARD		0x80863411 /*SDS2 Mainboard*/
#define DEVICE_INTEL_430FX_82437FX			0x122d /*430FX - 82437FX TSC [Triton I]*/
#define DEVICE_INTEL_82371FB_PIIX_ISA			0x122e /*82371FB PIIX ISA [Triton I]*/
#define DEVICE_INTEL_82371FB_PIIX_IDE			0x1230 /*82371FB PIIX IDE [Triton I]*/
#define DEVICE_INTEL_DSVD_MODEM				0x1231 /*DSVD Modem*/
#define DEVICE_INTEL_430MX_82371MX			0x1234 /*430MX - 82371MX Mobile PCI I/O IDE Xcelerator (MPIIX)*/
#define DEVICE_INTEL_430MX_82437MX			0x1235 /*430MX - 82437MX Mob. System Ctrlr (MTSC) & 82438MX Data Path (MTDP)*/
#define DEVICE_INTEL_440FX_82441FX			0x1237 /*440FX - 82441FX PMC [Natoma]*/
#define DEVICE_INTEL_82371FB_PIIX_IDE2			0x1239 /*82371FB PIIX IDE Interface*/
#define DEVICE_INTEL_82380PB_PCI_TO			0x123b /*82380PB PCI to PCI Docking Bridge*/
#define DEVICE_INTEL_82380AB_MISA_MOBILE		0x123c /*82380AB (MISA) Mobile PCI-to-ISA Bridge*/
#define DEVICE_INTEL_683053_PROGRAMMABLE_INTERRUPT	0x123d /*683053 Programmable Interrupt Device*/
#define DEVICE_INTEL_82466GX_IHPC_INTEGRATED		0x123e /*82466GX (IHPC) Integrated Hot-Plug Controller*/
#define DEVICE_INTEL_82466GX_INTEGRATED_HOT_PLUG	0x123f /*82466GX Integrated Hot-Plug Controller (IHPC)*/
#define DEVICE_INTEL_82752_752_AGP			0x1240 /*82752 (752) AGP Graphics Accelerator*/
#define DEVICE_INTEL_82380FB_MPCI2_MOBILE		0x124b /*82380FB (MPCI2) Mobile Docking Controller*/
#define DEVICE_INTEL_430HX_82439HX			0x1250 /*430HX - 82439HX TXC [Triton II]*/
#define DEVICE_INTEL_82806AA_PCI64_HUB2			0x1360 /*82806AA PCI64 Hub PCI Bridge*/
#define DEVICE_INTEL_82806AA_PCI64_HUB3			0x1361 /*82806AA PCI64 Hub Controller (HRes)*/
	#define SUBDEVICE_INTEL_82806AA_PCI64_HUB2	0x80861361 /*82806AA PCI64 Hub Controller (HRes)*/
	#define SUBDEVICE_INTEL_82806AA_PCI64_HUB3	0x80868000 /*82806AA PCI64 Hub Controller (HRes)*/
#define DEVICE_INTEL_82870P2_P64H2_HUB			0x1460 /*82870P2 P64H2 Hub PCI Bridge*/
#define DEVICE_INTEL_82870P2_P64H2_I_OXAPIC		0x1461 /*82870P2 P64H2 I/OxAPIC*/
	#define SUBDEVICE_INTEL_P4DP6			0x15d93480 /*P4DP6*/
	#define SUBDEVICE_INTEL_CX9_VX92		0x4c531090 /*Cx9 / Vx9 mainboard*/
#define DEVICE_INTEL_82870P2_P64H2_HOT			0x1462 /*82870P2 P64H2 Hot Plug Controller*/
#define DEVICE_INTEL_80960RP_I960RP_MICROPROCESSOR	0x1960 /*80960RP [i960RP Microprocessor]*/
	#define SUBDEVICE_INTEL_MEGARAID_431_RAID	0x101e0431 /*MegaRAID 431 RAID Controller*/
	#define SUBDEVICE_INTEL_MEGARAID_438_ULTRA2	0x101e0438 /*MegaRAID 438 Ultra2 LVD RAID Controller*/
	#define SUBDEVICE_INTEL_MEGARAID_466_EXPRESS	0x101e0466 /*MegaRAID 466 Express Plus RAID Controller*/
	#define SUBDEVICE_INTEL_MEGARAID_467_ENTERPRISE	0x101e0467 /*MegaRAID 467 Enterprise 1500 RAID Controller*/
	#define SUBDEVICE_INTEL_MEGARAID_490_EXPRESS	0x101e0490 /*MegaRAID 490 Express 300 RAID Controller*/
	#define SUBDEVICE_INTEL_MEGARAID_762_EXPRESS	0x101e0762 /*MegaRAID 762 Express RAID Controller*/
	#define SUBDEVICE_INTEL_POWEREDGE_EXPANDABLE_RAID	0x101e09a0 /*PowerEdge Expandable RAID Controller 2/SC*/
	#define SUBDEVICE_INTEL_POWEREDGE_EXPANDABLE_RAID2	0x10280467 /*PowerEdge Expandable RAID Controller 2/DC*/
	#define SUBDEVICE_INTEL_POWEREDGE_EXPANDABLE_RAID3	0x10281111 /*PowerEdge Expandable RAID Controller 2/SC*/
	#define SUBDEVICE_INTEL_MEGARAID		0x103c03a2 /*MegaRAID*/
	#define SUBDEVICE_INTEL_MEGARAID_438_HP		0x103c10c6 /*MegaRAID 438, HP NetRAID-3Si*/
	#define SUBDEVICE_INTEL_MEGARAID_T5_INTEGRATED	0x103c10c7 /*MegaRAID T5, Integrated HP NetRAID*/
	#define SUBDEVICE_INTEL_MEGARAID_INTEGRATED_HP	0x103c10cc /*MegaRAID, Integrated HP NetRAID*/
	#define SUBDEVICE_INTEL_HP_NETRAID_1SI		0x103c10cd /*HP NetRAID-1Si*/
	#define SUBDEVICE_INTEL_SUPERTRAK		0x105a0000 /*SuperTrak*/
	#define SUBDEVICE_INTEL_SUPERTRAK_PRO		0x105a2168 /*SuperTrak Pro*/
	#define SUBDEVICE_INTEL_SUPERTRAK66_100		0x105a5168 /*SuperTrak66/100*/
	#define SUBDEVICE_INTEL_MEGARAID_466_POWEREDGE	0x11111111 /*MegaRAID 466, PowerEdge Expandable RAID Controller 2/SC*/
	#define SUBDEVICE_INTEL_POWEREDGE_EXPANDABLE_RAID4	0x11111112 /*PowerEdge Expandable RAID Controller 2/SC*/
	#define SUBDEVICE_INTEL_MEGARAID2		0x113c03a2 /*MegaRAID*/
	#define SUBDEVICE_INTEL_CG1_RADIO		0xe4bf1010 /*CG1-RADIO*/
	#define SUBDEVICE_INTEL_CU2_QUARTET		0xe4bf1020 /*CU2-QUARTET*/
	#define SUBDEVICE_INTEL_CU1_CHORUS		0xe4bf1040 /*CU1-CHORUS*/
	#define SUBDEVICE_INTEL_CX1_BAND		0xe4bf3100 /*CX1-BAND*/
#define DEVICE_INTEL_80960RM_I960RM_MICROPROCESSOR	0x1962 /*80960RM [i960RM Microprocessor]*/
	#define SUBDEVICE_INTEL_SUPERTRAK_SX6000_I2O	0x105a0000 /*SuperTrak SX6000 I2O CPU*/
#define DEVICE_INTEL_82840_840_CARMEL			0x1a21 /*82840 840 (Carmel) Chipset Host Bridge (Hub A)*/
#define DEVICE_INTEL_82840_840_CARMEL2			0x1a23 /*82840 840 (Carmel) Chipset AGP Bridge*/
#define DEVICE_INTEL_82840_840_CARMEL3			0x1a24 /*82840 840 (Carmel) Chipset PCI Bridge (Hub B)*/
#define DEVICE_INTEL_82845_845_BROOKDALE		0x1a30 /*82845 845 (Brookdale) Chipset Host Bridge*/
	#define SUBDEVICE_INTEL_OPTIPLEX_GX240		0x1028010e /*Optiplex GX240*/
#define DEVICE_INTEL_82845_845_BROOKDALE2		0x1a31 /*82845 845 (Brookdale) Chipset AGP Bridge*/
#define DEVICE_INTEL_82801AA_ISA_BRIDGE			0x2410 /*82801AA ISA Bridge (LPC)*/
#define DEVICE_INTEL_82801AA_IDE			0x2411 /*82801AA IDE*/
#define DEVICE_INTEL_82801AA_USB			0x2412 /*82801AA USB*/
#define DEVICE_INTEL_82801AA_SMBUS			0x2413 /*82801AA SMBus*/
#define DEVICE_INTEL_82801AA_AC_97_AUDIO		0x2415 /*82801AA AC'97 Audio*/
	#define SUBDEVICE_INTEL_PRECISION_WORKSTATION_220	0x10280095 /*Precision Workstation 220 Integrated Digital Audio*/
	#define SUBDEVICE_INTEL_SOUNDMAX_INTEGRATED_DIGITAL	0x11d40040 /*SoundMAX Integrated Digital Audio*/
	#define SUBDEVICE_INTEL_SOUNDMAX_INTEGRATED_DIGITAL2	0x11d40048 /*SoundMAX Integrated Digital Audio*/
	#define SUBDEVICE_INTEL_SOUNDMAX_INTEGRATED_DIGITAL3	0x11d45340 /*SoundMAX Integrated Digital Audio*/
#define DEVICE_INTEL_82801AA_AC_97_MODEM		0x2416 /*82801AA AC'97 Modem*/
#define DEVICE_INTEL_82801AA_PCI_BRIDGE			0x2418 /*82801AA PCI Bridge*/
#define DEVICE_INTEL_82801AB_ISA_BRIDGE			0x2420 /*82801AB ISA Bridge (LPC)*/
#define DEVICE_INTEL_82801AB_IDE			0x2421 /*82801AB IDE*/
#define DEVICE_INTEL_82801AB_USB			0x2422 /*82801AB USB*/
#define DEVICE_INTEL_82801AB_SMBUS			0x2423 /*82801AB SMBus*/
#define DEVICE_INTEL_82801AB_AC_97_AUDIO		0x2425 /*82801AB AC'97 Audio*/
	#define SUBDEVICE_INTEL_SOUNDMAX_INTEGRATED_DIGITAL4	0x11d40040 /*SoundMAX Integrated Digital Audio*/
	#define SUBDEVICE_INTEL_SOUNDMAX_INTEGRATED_DIGITAL5	0x11d40048 /*SoundMAX Integrated Digital Audio*/
#define DEVICE_INTEL_82801AB_AC_97_MODEM		0x2426 /*82801AB AC'97 Modem*/
#define DEVICE_INTEL_82801AB_PCI_BRIDGE			0x2428 /*82801AB PCI Bridge*/
#define DEVICE_INTEL_82801BA_ISA_BRIDGE			0x2440 /*82801BA ISA Bridge (LPC)*/
#define DEVICE_INTEL_82801BA_BAM_USB_HUB		0x2442 /*82801BA/BAM USB (Hub #1)*/
	#define SUBDEVICE_INTEL_NETVISTA_A40_A40P	0x101401c6 /*Netvista A40/A40p*/
	#define SUBDEVICE_INTEL_TRAVELMATE_612_TX3	0x10251016 /*Travelmate 612 TX*/
	#define SUBDEVICE_INTEL_OPTIPLEX_GX2402		0x1028010e /*Optiplex GX240*/
	#define SUBDEVICE_INTEL_TUSL2_C_MAINBOARD2	0x10438027 /*TUSL2-C Mainboard*/
	#define SUBDEVICE_INTEL_VAIO_PCG_FX4033		0x104d80df /*Vaio PCG-FX403*/
	#define SUBDEVICE_INTEL_TH7II_RAID		0x147b0507 /*TH7II-RAID*/
	#define SUBDEVICE_INTEL_D815EEA2_MAINBOARD3	0x80864532 /*D815EEA2 mainboard*/
	#define SUBDEVICE_INTEL_D815EGEW_MAINBOARD3	0x80864557 /*D815EGEW Mainboard*/
#define DEVICE_INTEL_82801BA_BAM_SMBUS			0x2443 /*82801BA/BAM SMBus*/
	#define SUBDEVICE_INTEL_NETVISTA_A40_A40P2	0x101401c6 /*Netvista A40/A40p*/
	#define SUBDEVICE_INTEL_TRAVELMATE_612_TX4	0x10251016 /*Travelmate 612 TX*/
	#define SUBDEVICE_INTEL_OPTIPLEX_GX2403		0x1028010e /*Optiplex GX240*/
	#define SUBDEVICE_INTEL_TUSL2_C_MAINBOARD3	0x10438027 /*TUSL2-C Mainboard*/
	#define SUBDEVICE_INTEL_VAIO_PCG_FX4034		0x104d80df /*Vaio PCG-FX403*/
	#define SUBDEVICE_INTEL_TH7II_RAID2		0x147b0507 /*TH7II-RAID*/
	#define SUBDEVICE_INTEL_D815EEA2_MAINBOARD4	0x80864532 /*D815EEA2 mainboard*/
	#define SUBDEVICE_INTEL_D815EGEW_MAINBOARD4	0x80864557 /*D815EGEW Mainboard*/
#define DEVICE_INTEL_82801BA_BAM_USB_HUB2		0x2444 /*82801BA/BAM USB (Hub #2)*/
	#define SUBDEVICE_INTEL_TRAVELMATE_612_TX5	0x10251016 /*Travelmate 612 TX*/
	#define SUBDEVICE_INTEL_OPTIPLEX_GX2404		0x1028010e /*Optiplex GX240*/
	#define SUBDEVICE_INTEL_TUSL2_C_MAINBOARD4	0x10438027 /*TUSL2-C Mainboard*/
	#define SUBDEVICE_INTEL_VAIO_PCG_FX4035		0x104d80df /*Vaio PCG-FX403*/
	#define SUBDEVICE_INTEL_TH7II_RAID3		0x147b0507 /*TH7II-RAID*/
	#define SUBDEVICE_INTEL_D815EEA2_MAINBOARD5	0x80864532 /*D815EEA2 mainboard*/
#define DEVICE_INTEL_82801BA_BAM_AC_97_AUDIO		0x2445 /*82801BA/BAM AC'97 Audio*/
	#define SUBDEVICE_INTEL_NETVISTA_A40_A40P3	0x101401c6 /*Netvista A40/A40p*/
	#define SUBDEVICE_INTEL_TRAVELMATE_612_TX6	0x10251016 /*Travelmate 612 TX*/
	#define SUBDEVICE_INTEL_VAIO_PCG_FX4036		0x104d80df /*Vaio PCG-FX403*/
	#define SUBDEVICE_INTEL_STAC9721_AC		0x14623370 /*STAC9721 AC*/
	#define SUBDEVICE_INTEL_TH7II_RAID4		0x147b0507 /*TH7II-RAID*/
	#define SUBDEVICE_INTEL_D815EGEW_MAINBOARD5	0x80864557 /*D815EGEW Mainboard*/
#define DEVICE_INTEL_82801BA_BAM_AC_97_MODEM		0x2446 /*82801BA/BAM AC'97 Modem*/
	#define SUBDEVICE_INTEL_TRAVELMATE_612_TX7	0x10251016 /*Travelmate 612 TX*/
	#define SUBDEVICE_INTEL_VAIO_PCG_FX4037		0x104d80df /*Vaio PCG-FX403*/
#define DEVICE_INTEL_82801_MOBILE_PCI			0x2448 /*82801 Mobile PCI Bridge*/
#define DEVICE_INTEL_82801BA_BAM_CA_CAM_ETHERNET_CONTROLLER	0x2449 /*82801BA/BAM/CA/CAM Ethernet Controller*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VM	0x0e110012 /*EtherExpress PRO/100 VM*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VE6	0x0e110091 /*EtherExpress PRO/100 VE*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VE7	0x101401ce /*EtherExpress PRO/100 VE*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VE8	0x101401dc /*EtherExpress PRO/100 VE*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VE9	0x101401eb /*EtherExpress PRO/100 VE*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VE10	0x101401ec /*EtherExpress PRO/100 VE*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VE11	0x10140202 /*EtherExpress PRO/100 VE*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VE12	0x10140205 /*EtherExpress PRO/100 VE*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VE13	0x10140217 /*EtherExpress PRO/100 VE*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VE14	0x10140234 /*EtherExpress PRO/100 VE*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VE15	0x1014023d /*EtherExpress PRO/100 VE*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VE16	0x10140244 /*EtherExpress PRO/100 VE*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VE17	0x10140245 /*EtherExpress PRO/100 VE*/
	#define SUBDEVICE_INTEL_PRO_100_VE_DESKTOP	0x10140265 /*PRO/100 VE Desktop Connection*/
	#define SUBDEVICE_INTEL_PRO_100_VE_DESKTOP2	0x10140267 /*PRO/100 VE Desktop Connection*/
	#define SUBDEVICE_INTEL_PRO_100_VE_DESKTOP3	0x1014026a /*PRO/100 VE Desktop Connection*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VE18	0x109f315d /*EtherExpress PRO/100 VE*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VE19	0x109f3181 /*EtherExpress PRO/100 VE*/
	#define SUBDEVICE_INTEL_PRO_100_VE_NETWORK	0x1179ff01 /*PRO/100 VE Network Connection*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VE20	0x11867801 /*EtherExpress PRO/100 VE*/
	#define SUBDEVICE_INTEL_HOMEPNA_1M_CNR		0x144d2602 /*HomePNA 1M CNR*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VE21	0x80863010 /*EtherExpress PRO/100 VE*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VM2	0x80863011 /*EtherExpress PRO/100 VM*/
	#define SUBDEVICE_INTEL_82562EH_BASED_PHONELINE	0x80863012 /*82562EH based Phoneline*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VE22	0x80863013 /*EtherExpress PRO/100 VE*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_VM3	0x80863014 /*EtherExpress PRO/100 VM*/
	#define SUBDEVICE_INTEL_82562EH_BASED_PHONELINE2	0x80863015 /*82562EH based Phoneline*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_P6	0x80863016 /*EtherExpress PRO/100 P Mobile Combo*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_P7	0x80863017 /*EtherExpress PRO/100 P Mobile*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_1004	0x80863018 /*EtherExpress PRO/100*/
#define DEVICE_INTEL_82801BAM_IDE_U100			0x244a /*82801BAM IDE U100*/
	#define SUBDEVICE_INTEL_TRAVELMATE_612TX	0x10251016 /*Travelmate 612TX*/
	#define SUBDEVICE_INTEL_VAIO_PCG_FX4038		0x104d80df /*Vaio PCG-FX403*/
#define DEVICE_INTEL_82801BA_IDE_U100			0x244b /*82801BA IDE U100*/
	#define SUBDEVICE_INTEL_NETVISTA_A40_A40P4	0x101401c6 /*Netvista A40/A40p*/
	#define SUBDEVICE_INTEL_OPTIPLEX_GX2405		0x1028010e /*Optiplex GX240*/
	#define SUBDEVICE_INTEL_TUSL2_C_MAINBOARD5	0x10438027 /*TUSL2-C Mainboard*/
	#define SUBDEVICE_INTEL_TH7II_RAID5		0x147b0507 /*TH7II-RAID*/
	#define SUBDEVICE_INTEL_D815EEA2_MAINBOARD6	0x80864532 /*D815EEA2 mainboard*/
	#define SUBDEVICE_INTEL_D815EGEW_MAINBOARD6	0x80864557 /*D815EGEW Mainboard*/
#define DEVICE_INTEL_82801BAM_ISA_BRIDGE		0x244c /*82801BAM ISA Bridge (LPC)*/
#define DEVICE_INTEL_82801_PCI_BRIDGE			0x244e /*82801 PCI Bridge*/
	#define SUBDEVICE_INTEL_NETVISTA_A30P2		0x10140267 /*NetVista A30p*/
#define DEVICE_INTEL_82801E_ISA_BRIDGE			0x2450 /*82801E ISA Bridge (LPC)*/
#define DEVICE_INTEL_82801E_USB				0x2452 /*82801E USB*/
#define DEVICE_INTEL_82801E_SMBUS			0x2453 /*82801E SMBus*/
#define DEVICE_INTEL_82801E_ETHERNET_CONTROLLER		0x2459 /*82801E Ethernet Controller 0*/
#define DEVICE_INTEL_82801E_IDE_U100			0x245b /*82801E IDE U100*/
#define DEVICE_INTEL_82801E_ETHERNET_CONTROLLER2	0x245d /*82801E Ethernet Controller 1*/
#define DEVICE_INTEL_82801E_PCI_BRIDGE			0x245e /*82801E PCI Bridge*/
#define DEVICE_INTEL_82801CA_LPC_INTERFACE		0x2480 /*82801CA LPC Interface Controller*/
#define DEVICE_INTEL_82801CA_CAM_USB_HUB		0x2482 /*82801CA/CAM USB (Hub #1)*/
	#define SUBDEVICE_INTEL_THINKPAD_A_T_X_SERIES2	0x10140220 /*ThinkPad A/T/X Series*/
	#define SUBDEVICE_INTEL_VAIO_PCG_GR214EP_GR214MP_GR215MP_GR314MP_GR315MP2	0x104d80e7 /*VAIO PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP*/
	#define SUBDEVICE_INTEL_P4DP62			0x15d93480 /*P4DP6*/
	#define SUBDEVICE_INTEL_VPR_MATRIX_170B42	0x80861958 /*vpr Matrix 170B4*/
	#define SUBDEVICE_INTEL_SE7501HG2_MAINBOARD2	0x80863424 /*SE7501HG2 Mainboard*/
	#define SUBDEVICE_INTEL_LATITUDE_C640		0x80864541 /*Latitude C640*/
#define DEVICE_INTEL_82801CA_CAM_SMBUS_CONTROLLER	0x2483 /*82801CA/CAM SMBus Controller*/
	#define SUBDEVICE_INTEL_THINKPAD_A_T_X_SERIES3	0x10140220 /*ThinkPad A/T/X Series*/
	#define SUBDEVICE_INTEL_VAIO_PCG_GR214EP_GR214MP_GR215MP_GR314MP_GR315MP3	0x104d80e7 /*VAIO PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP*/
	#define SUBDEVICE_INTEL_P4DP63			0x15d93480 /*P4DP6*/
	#define SUBDEVICE_INTEL_VPR_MATRIX_170B43	0x80861958 /*vpr Matrix 170B4*/
#define DEVICE_INTEL_82801CA_CAM_USB_HUB2		0x2484 /*82801CA/CAM USB (Hub #2)*/
	#define SUBDEVICE_INTEL_THINKPAD_A_T_X_SERIES4	0x10140220 /*ThinkPad A/T/X Series*/
	#define SUBDEVICE_INTEL_VAIO_PCG_GR214EP_GR214MP_GR215MP_GR314MP_GR315MP4	0x104d80e7 /*VAIO PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP*/
	#define SUBDEVICE_INTEL_P4DP64			0x15d93480 /*P4DP6*/
	#define SUBDEVICE_INTEL_VPR_MATRIX_170B44	0x80861958 /*vpr Matrix 170B4*/
#define DEVICE_INTEL_82801CA_CAM_AC_97_AUDIO		0x2485 /*82801CA/CAM AC'97 Audio Controller*/
	#define SUBDEVICE_INTEL_CRYSTAL_WMD_AUDIO	0x10135959 /*Crystal WMD Audio Codec*/
	#define SUBDEVICE_INTEL_THINKPAD_T23_2647_4MG	0x10140222 /*ThinkPad T23 (2647-4MG) or A30/A30p (2652/2653)*/
	#define SUBDEVICE_INTEL_THINKPAD_T30		0x10140508 /*ThinkPad T30*/
	#define SUBDEVICE_INTEL_THINKPAD_A_T_X_SERIES5	0x1014051c /*ThinkPad A/T/X Series*/
	#define SUBDEVICE_INTEL_VAIO_PCG_GR214EP_GR214MP_GR215MP_GR314MP_GR315MP5	0x104d80e7 /*VAIO PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP*/
	#define SUBDEVICE_INTEL_VPR_MATRIX_170B45	0x144dc006 /*vpr Matrix 170B4*/
#define DEVICE_INTEL_82801CA_CAM_AC_97_MODEM		0x2486 /*82801CA/CAM AC'97 Modem Controller*/
	#define SUBDEVICE_INTEL_THINKPAD_A_T_X_SERIES6	0x10140223 /*ThinkPad A/T/X Series*/
	#define SUBDEVICE_INTEL_THINKPAD_R31_2656BBG	0x10140503 /*ThinkPad R31 2656BBG*/
	#define SUBDEVICE_INTEL_THINKPAD_A_T_X_SERIES7	0x1014051a /*ThinkPad A/T/X Series*/
	#define SUBDEVICE_INTEL_ACER_620_SERIES		0x101f1025 /*Acer 620 Series*/
	#define SUBDEVICE_INTEL_VAIO_PCG_GR214EP_GR214MP_GR215MP_GR314MP_GR315MP6	0x104d80e7 /*VAIO PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP*/
	#define SUBDEVICE_INTEL_TOSHIBA_SATELLITE_1110	0x11790001 /*Toshiba Satellite 1110 Z15 internal Modem*/
	#define SUBDEVICE_INTEL_DELL_INSPIRON_2100	0x134d4c21 /*Dell Inspiron 2100 internal modem*/
	#define SUBDEVICE_INTEL_VPR_MATRIX_170B46	0x144d2115 /*vpr Matrix 170B4 internal modem*/
	#define SUBDEVICE_INTEL_MD56ORD_V_92_MDC	0x14f15421 /*MD56ORD V.92 MDC Modem*/
#define DEVICE_INTEL_82801CA_CAM_USB_HUB3		0x2487 /*82801CA/CAM USB (Hub #3)*/
	#define SUBDEVICE_INTEL_THINKPAD_A_T_X_SERIES8	0x10140220 /*ThinkPad A/T/X Series*/
	#define SUBDEVICE_INTEL_VAIO_PCG_GR214EP_GR214MP_GR215MP_GR314MP_GR315MP7	0x104d80e7 /*VAIO PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP*/
	#define SUBDEVICE_INTEL_P4DP65			0x15d93480 /*P4DP6*/
	#define SUBDEVICE_INTEL_VPR_MATRIX_170B47	0x80861958 /*vpr Matrix 170B4*/
#define DEVICE_INTEL_82801CAM_IDE_U100			0x248a /*82801CAM IDE U100*/
	#define SUBDEVICE_INTEL_THINKPAD_A_T_X_SERIES9	0x10140220 /*ThinkPad A/T/X Series*/
	#define SUBDEVICE_INTEL_VAIO_PCG_GR214EP_GR214MP_GR215MP_GR314MP_GR315MP8	0x104d80e7 /*VAIO PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP*/
	#define SUBDEVICE_INTEL_VPR_MATRIX_170B48	0x80861958 /*vpr Matrix 170B4*/
	#define SUBDEVICE_INTEL_LATITUDE_C6402		0x80864541 /*Latitude C640*/
#define DEVICE_INTEL_82801CA_ULTRA_ATA			0x248b /*82801CA Ultra ATA Storage Controller*/
	#define SUBDEVICE_INTEL_P4DP66			0x15d93480 /*P4DP6*/
#define DEVICE_INTEL_82801CAM_ISA_BRIDGE		0x248c /*82801CAM ISA Bridge (LPC)*/
#define DEVICE_INTEL_82801DB_DBL_ICH4_ICH4_L_LPC	0x24c0 /*82801DB/DBL (ICH4/ICH4-L) LPC Interface Bridge*/
	#define SUBDEVICE_INTEL_NETVISTA_A30P3		0x10140267 /*NetVista A30p*/
	#define SUBDEVICE_INTEL_845PE_MAX_MS_6580	0x14625800 /*845PE Max (MS-6580)*/
#define DEVICE_INTEL_82801DBL_ICH4_L_IDE		0x24c1 /*82801DBL (ICH4-L) IDE Controller*/
#define DEVICE_INTEL_82801DB_DBL_DBM_ICH4_ICH4_L_ICH4_M_USB	0x24c2 /*82801DB/DBL/DBM (ICH4/ICH4-L/ICH4-M) USB UHCI Controller #1*/
	#define SUBDEVICE_INTEL_NETVISTA_A30P4		0x10140267 /*NetVista A30p*/
	#define SUBDEVICE_INTEL_TRAVELMATE_290		0x1025005a /*TravelMate 290*/
	#define SUBDEVICE_INTEL_OPTIPLEX_GX2602		0x10280126 /*Optiplex GX260*/
	#define SUBDEVICE_INTEL_LATITUDE_D505		0x10280163 /*Latitude D505*/
	#define SUBDEVICE_INTEL_NC6000_LAPTOP		0x103c0890 /*NC6000 laptop*/
	#define SUBDEVICE_INTEL_MIM2000			0x10718160 /*MIM2000*/
	#define SUBDEVICE_INTEL_845PE_MAX_MS_65802	0x14625800 /*845PE Max (MS-6580)*/
	#define SUBDEVICE_INTEL_AVERATEC_5110H_LAPTOP	0x15092990 /*Averatec 5110H laptop*/
	#define SUBDEVICE_INTEL_CX9_VX93		0x4c531090 /*Cx9 / Vx9 mainboard*/
#define DEVICE_INTEL_82801DB_DBL_DBM_ICH4_ICH4_L_ICH4_M_SMBUS	0x24c3 /*82801DB/DBL/DBM (ICH4/ICH4-L/ICH4-M) SMBus Controller*/
	#define SUBDEVICE_INTEL_NETVISTA_A30P5		0x10140267 /*NetVista A30p*/
	#define SUBDEVICE_INTEL_TRAVELMATE_2902		0x1025005a /*TravelMate 290*/
	#define SUBDEVICE_INTEL_OPTIPLEX_GX2603		0x10280126 /*Optiplex GX260*/
	#define SUBDEVICE_INTEL_NC6000_LAPTOP2		0x103c0890 /*NC6000 laptop*/
	#define SUBDEVICE_INTEL_MIM20002		0x10718160 /*MIM2000*/
	#define SUBDEVICE_INTEL_GA_8PE667_ULTRA		0x145824c2 /*GA-8PE667 Ultra*/
	#define SUBDEVICE_INTEL_845PE_MAX_MS_65803	0x14625800 /*845PE Max (MS-6580)*/
	#define SUBDEVICE_INTEL_CX9_VX94		0x4c531090 /*Cx9 / Vx9 mainboard*/
#define DEVICE_INTEL_82801DB_DBL_DBM_ICH4_ICH4_L_ICH4_M_USB2	0x24c4 /*82801DB/DBL/DBM (ICH4/ICH4-L/ICH4-M) USB UHCI Controller #2*/
	#define SUBDEVICE_INTEL_NETVISTA_A30P6		0x10140267 /*NetVista A30p*/
	#define SUBDEVICE_INTEL_TRAVELMATE_2903		0x1025005a /*TravelMate 290*/
	#define SUBDEVICE_INTEL_OPTIPLEX_GX2604		0x10280126 /*Optiplex GX260*/
	#define SUBDEVICE_INTEL_LATITUDE_D5052		0x10280163 /*Latitude D505*/
	#define SUBDEVICE_INTEL_NC6000_LAPTOP3		0x103c0890 /*NC6000 laptop*/
	#define SUBDEVICE_INTEL_MIM20003		0x10718160 /*MIM2000*/
	#define SUBDEVICE_INTEL_845PE_MAX_MS_65804	0x14625800 /*845PE Max (MS-6580)*/
	#define SUBDEVICE_INTEL_AVERATEC_5110H		0x15092990 /*Averatec 5110H*/
	#define SUBDEVICE_INTEL_CX9_VX95		0x4c531090 /*Cx9 / Vx9 mainboard*/
#define DEVICE_INTEL_82801DB_DBL_DBM_ICH4_ICH4_L_ICH4_M_AC_97	0x24c5 /*82801DB/DBL/DBM (ICH4/ICH4-L/ICH4-M) AC'97 Audio Controller*/
	#define SUBDEVICE_INTEL_ANALOG_DEVICES_INC	0x0e1100b8 /*Analog Devices Inc. codec [SoundMAX]*/
	#define SUBDEVICE_INTEL_NETVISTA_A30P7		0x10140267 /*NetVista A30p*/
	#define SUBDEVICE_INTEL_TRAVELMATE_2904		0x1025005a /*TravelMate 290*/
	#define SUBDEVICE_INTEL_LATITUDE_D5053		0x10280163 /*Latitude D505*/
	#define SUBDEVICE_INTEL_NC6000_LAPTOP4		0x103c0890 /*NC6000 laptop*/
	#define SUBDEVICE_INTEL_MIM20004		0x10718160 /*MIM2000*/
	#define SUBDEVICE_INTEL_GA_8PE667_ULTRA2	0x1458a002 /*GA-8PE667 Ultra*/
	#define SUBDEVICE_INTEL_845PE_MAX_MS_65805	0x14625800 /*845PE Max (MS-6580)*/
#define DEVICE_INTEL_82801DB_DBL_DBM_ICH4_ICH4_L_ICH4_M_AC_972	0x24c6 /*82801DB/DBL/DBM (ICH4/ICH4-L/ICH4-M) AC'97 Modem Controller*/
	#define SUBDEVICE_INTEL_TRAVELMATE_2905		0x1025005a /*TravelMate 290*/
	#define SUBDEVICE_INTEL_NC6000_LAPTOP5		0x103c0890 /*NC6000 laptop*/
	#define SUBDEVICE_INTEL_MIM20005		0x10718160 /*MIM2000*/
#define DEVICE_INTEL_82801DB_DBL_DBM_ICH4_ICH4_L_ICH4_M_USB3	0x24c7 /*82801DB/DBL/DBM (ICH4/ICH4-L/ICH4-M) USB UHCI Controller #3*/
	#define SUBDEVICE_INTEL_NETVISTA_A30P8		0x10140267 /*NetVista A30p*/
	#define SUBDEVICE_INTEL_TRAVELMATE_2906		0x1025005a /*TravelMate 290*/
	#define SUBDEVICE_INTEL_OPTIPLEX_GX2605		0x10280126 /*Optiplex GX260*/
	#define SUBDEVICE_INTEL_LATITUDE_D5054		0x10280163 /*Latitude D505*/
	#define SUBDEVICE_INTEL_NC6000_LAPTOP6		0x103c0890 /*NC6000 laptop*/
	#define SUBDEVICE_INTEL_MIM20006		0x10718160 /*MIM2000*/
	#define SUBDEVICE_INTEL_845PE_MAX_MS_65806	0x14625800 /*845PE Max (MS-6580)*/
	#define SUBDEVICE_INTEL_AVERATEC_5110H2		0x15092990 /*Averatec 5110H*/
	#define SUBDEVICE_INTEL_CX9_VX96		0x4c531090 /*Cx9 / Vx9 mainboard*/
#define DEVICE_INTEL_82801DBM_ICH4_M_IDE		0x24ca /*82801DBM (ICH4-M) IDE Controller*/
	#define SUBDEVICE_INTEL_TRAVELMATE_2907		0x1025005a /*TravelMate 290*/
	#define SUBDEVICE_INTEL_LATITUDE_D5055		0x10280163 /*Latitude D505*/
	#define SUBDEVICE_INTEL_NC6000_LAPTOP7		0x103c0890 /*NC6000 laptop*/
	#define SUBDEVICE_INTEL_MIM20007		0x10718160 /*MIM2000*/
#define DEVICE_INTEL_82801DB_ICH4_IDE			0x24cb /*82801DB (ICH4) IDE Controller*/
	#define SUBDEVICE_INTEL_NETVISTA_A30P9		0x10140267 /*NetVista A30p*/
	#define SUBDEVICE_INTEL_OPTIPLEX_GX2606		0x10280126 /*Optiplex GX260*/
	#define SUBDEVICE_INTEL_GA_8PE667_ULTRA3	0x145824c2 /*GA-8PE667 Ultra*/
	#define SUBDEVICE_INTEL_845PE_MAX_MS_65807	0x14625800 /*845PE Max (MS-6580)*/
	#define SUBDEVICE_INTEL_CX9_VX97		0x4c531090 /*Cx9 / Vx9 mainboard*/
#define DEVICE_INTEL_82801DBM_ICH4_M_LPC		0x24cc /*82801DBM (ICH4-M) LPC Interface Bridge*/
#define DEVICE_INTEL_82801DB_DBM_ICH4_ICH4_M_USB2	0x24cd /*82801DB/DBM (ICH4/ICH4-M) USB2 EHCI Controller*/
	#define SUBDEVICE_INTEL_NETVISTA_A30P10		0x10140267 /*NetVista A30p*/
	#define SUBDEVICE_INTEL_TRAVELMATE_2908		0x1025005a /*TravelMate 290*/
	#define SUBDEVICE_INTEL_OPTIPLEX_GX2607		0x10280126 /*Optiplex GX260*/
	#define SUBDEVICE_INTEL_LATITUDE_D5056		0x10280163 /*Latitude D505*/
	#define SUBDEVICE_INTEL_NC6000_LAPTOP8		0x103c0890 /*NC6000 laptop*/
	#define SUBDEVICE_INTEL_MIM20008		0x10718160 /*MIM2000*/
	#define SUBDEVICE_INTEL_845PE_MAX_MS_65808	0x14623981 /*845PE Max (MS-6580)*/
	#define SUBDEVICE_INTEL_AVERATEC_5110H3		0x15091968 /*Averatec 5110H*/
	#define SUBDEVICE_INTEL_CX9_VX98		0x4c531090 /*Cx9 / Vx9 mainboard*/
#define DEVICE_INTEL_82801EB_ER_ICH5_ICH5R_LPC		0x24d0 /*82801EB/ER (ICH5/ICH5R) LPC Interface Bridge*/
#define DEVICE_INTEL_82801EB_ICH5_SATA			0x24d1 /*82801EB (ICH5) SATA Controller*/
	#define SUBDEVICE_INTEL_D530_CMT_DG746A		0x103c12bc /*d530 CMT (DG746A)*/
	#define SUBDEVICE_INTEL_GA_8IPE1000_PRO2_MOTHERBOARD2	0x145824d1 /*GA-8IPE1000 Pro2 motherboard (865PE)*/
	#define SUBDEVICE_INTEL_865PE_NEO2_MS_67282	0x14627280 /*865PE Neo2 (MS-6728)*/
	#define SUBDEVICE_INTEL_S875WP1_E_MAINBOARD3	0x80863427 /*S875WP1-E mainboard*/
	#define SUBDEVICE_INTEL_D865PERL_MAINBOARD2	0x8086524c /*D865PERL mainboard*/
#define DEVICE_INTEL_82801EB_ER_ICH5_ICH5R_USB		0x24d2 /*82801EB/ER (ICH5/ICH5R) USB UHCI Controller #1*/
	#define SUBDEVICE_INTEL_D530_CMT_DG746A2	0x103c12bc /*d530 CMT (DG746A)*/
	#define SUBDEVICE_INTEL_P4P800_MAINBOARD	0x104380a6 /*P4P800 Mainboard*/
	#define SUBDEVICE_INTEL_GA_8KNXP_MOTHERBOARD_875P	0x145824d2 /*GA-8KNXP motherboard (875P)*/
	#define SUBDEVICE_INTEL_865PE_NEO2_MS_67283	0x14627280 /*865PE Neo2 (MS-6728)*/
	#define SUBDEVICE_INTEL_S875WP1_E_MAINBOARD4	0x80863427 /*S875WP1-E mainboard*/
	#define SUBDEVICE_INTEL_D865PERL_MAINBOARD3	0x8086524c /*D865PERL mainboard*/
#define DEVICE_INTEL_82801EB_ER_ICH5_ICH5R_SMBUS	0x24d3 /*82801EB/ER (ICH5/ICH5R) SMBus Controller*/
	#define SUBDEVICE_INTEL_P4P800_MAINBOARD2	0x104380a6 /*P4P800 Mainboard*/
	#define SUBDEVICE_INTEL_GA_8IPE1000_PRO2_MOTHERBOARD3	0x145824d2 /*GA-8IPE1000 Pro2 motherboard (865PE)*/
	#define SUBDEVICE_INTEL_865PE_NEO2_MS_67284	0x14627280 /*865PE Neo2 (MS-6728)*/
	#define SUBDEVICE_INTEL_S875WP1_E_MAINBOARD5	0x80863427 /*S875WP1-E mainboard*/
	#define SUBDEVICE_INTEL_D865PERL_MAINBOARD4	0x8086524c /*D865PERL mainboard*/
#define DEVICE_INTEL_82801EB_ER_ICH5_ICH5R_USB2		0x24d4 /*82801EB/ER (ICH5/ICH5R) USB UHCI Controller #2*/
	#define SUBDEVICE_INTEL_D530_CMT_DG746A3	0x103c12bc /*d530 CMT (DG746A)*/
	#define SUBDEVICE_INTEL_P4P800_MAINBOARD3	0x104380a6 /*P4P800 Mainboard*/
	#define SUBDEVICE_INTEL_GA_8IPE1000_PRO2_MOTHERBOARD4	0x145824d2 /*GA-8IPE1000 Pro2 motherboard (865PE)*/
	#define SUBDEVICE_INTEL_865PE_NEO2_MS_67285	0x14627280 /*865PE Neo2 (MS-6728)*/
	#define SUBDEVICE_INTEL_S875WP1_E_MAINBOARD6	0x80863427 /*S875WP1-E mainboard*/
	#define SUBDEVICE_INTEL_D865PERL_MAINBOARD5	0x8086524c /*D865PERL mainboard*/
#define DEVICE_INTEL_82801EB_ER_ICH5_ICH5R_AC_97	0x24d5 /*82801EB/ER (ICH5/ICH5R) AC'97 Audio Controller*/
	#define SUBDEVICE_INTEL_ANALOG_DEVICES_CODEC	0x103c12bc /*Analog Devices codec [SoundMAX Integrated Digital Audio]*/
	#define SUBDEVICE_INTEL_P4P800_MAINBOARD4	0x104380f3 /*P4P800 Mainboard*/
	#define SUBDEVICE_INTEL_GA_8KNXP_MOTHERBOARD_875P2	0x1458a002 /*GA-8KNXP motherboard (875P)*/
	#define SUBDEVICE_INTEL_865PE_NEO2_MS_67286	0x14627280 /*865PE Neo2 (MS-6728)*/
	#define SUBDEVICE_INTEL_D865PERL_MAINBOARD6	0x8086a000 /*D865PERL mainboard*/
#define DEVICE_INTEL_82801EB_ER_ICH5_ICH5R_AC_972	0x24d6 /*82801EB/ER (ICH5/ICH5R) AC'97 Modem Controller*/
#define DEVICE_INTEL_82801EB_ER_ICH5_ICH5R_USB3		0x24d7 /*82801EB/ER (ICH5/ICH5R) USB UHCI #3*/
	#define SUBDEVICE_INTEL_D530_CMT_DG746A4	0x103c12bc /*d530 CMT (DG746A)*/
	#define SUBDEVICE_INTEL_P4P800_MAINBOARD5	0x104380a6 /*P4P800 Mainboard*/
	#define SUBDEVICE_INTEL_GA_8IPE1000_PRO2_MOTHERBOARD5	0x145824d2 /*GA-8IPE1000 Pro2 motherboard (865PE)*/
	#define SUBDEVICE_INTEL_865PE_NEO2_MS_67287	0x14627280 /*865PE Neo2 (MS-6728)*/
	#define SUBDEVICE_INTEL_S875WP1_E_MAINBOARD7	0x80863427 /*S875WP1-E mainboard*/
	#define SUBDEVICE_INTEL_D865PERL_MAINBOARD7	0x8086524c /*D865PERL mainboard*/
#define DEVICE_INTEL_82801EB_ER_ICH5_ICH5R_IDE		0x24db /*82801EB/ER (ICH5/ICH5R) IDE Controller*/
	#define SUBDEVICE_INTEL_D530_CMT_DG746A5	0x103c12bc /*d530 CMT (DG746A)*/
	#define SUBDEVICE_INTEL_P4P800_MAINBOARD6	0x104380a6 /*P4P800 Mainboard*/
	#define SUBDEVICE_INTEL_GA_8IPE1000_PRO2_MOTHERBOARD6	0x145824d2 /*GA-8IPE1000 Pro2 motherboard (865PE)*/
	#define SUBDEVICE_INTEL_865PE_NEO2_MS_67288	0x14627280 /*865PE Neo2 (MS-6728)*/
	#define SUBDEVICE_INTEL_MSI_875P		0x14627580 /*MSI 875P*/
	#define SUBDEVICE_INTEL_S875WP1_E_MAINBOARD8	0x80863427 /*S875WP1-E mainboard*/
	#define SUBDEVICE_INTEL_D865PERL_MAINBOARD8	0x8086524c /*D865PERL mainboard*/
#define DEVICE_INTEL_82801EB_ICH5_LPC			0x24dc /*82801EB (ICH5) LPC Interface Bridge*/
#define DEVICE_INTEL_82801EB_ER_ICH5_ICH5R_USB22	0x24dd /*82801EB/ER (ICH5/ICH5R) USB2 EHCI Controller*/
	#define SUBDEVICE_INTEL_D530_CMT_DG746A6	0x103c12bc /*d530 CMT (DG746A)*/
	#define SUBDEVICE_INTEL_P4P800_MAINBOARD7	0x104380a6 /*P4P800 Mainboard*/
	#define SUBDEVICE_INTEL_GA_8IPE1000_PRO2_MOTHERBOARD7	0x14585006 /*GA-8IPE1000 Pro2 motherboard (865PE)*/
	#define SUBDEVICE_INTEL_865PE_NEO2_MS_67289	0x14627280 /*865PE Neo2 (MS-6728)*/
	#define SUBDEVICE_INTEL_S875WP1_E_MAINBOARD9	0x80863427 /*S875WP1-E mainboard*/
	#define SUBDEVICE_INTEL_D865PERL_MAINBOARD9	0x8086524c /*D865PERL mainboard*/
#define DEVICE_INTEL_82801EB_ER_ICH5_ICH5R_USB4		0x24de /*82801EB/ER (ICH5/ICH5R) USB UHCI Controller #4*/
	#define SUBDEVICE_INTEL_P4P800_MAINBOARD8	0x104380a6 /*P4P800 Mainboard*/
	#define SUBDEVICE_INTEL_GA_8IPE1000_PRO2_MOTHERBOARD8	0x145824d2 /*GA-8IPE1000 Pro2 motherboard (865PE)*/
	#define SUBDEVICE_INTEL_865PE_NEO2_MS_672810	0x14627280 /*865PE Neo2 (MS-6728)*/
	#define SUBDEVICE_INTEL_S875WP1_E_MAINBOARD10	0x80863427 /*S875WP1-E mainboard*/
	#define SUBDEVICE_INTEL_D865PERL_MAINBOARD10	0x8086524c /*D865PERL mainboard*/
#define DEVICE_INTEL_82801ER_ICH5R_SATA			0x24df /*82801ER (ICH5R) SATA Controller*/
#define DEVICE_INTEL_82820_820_CAMINO			0x2500 /*82820 820 (Camino) Chipset Host Bridge (MCH)*/
	#define SUBDEVICE_INTEL_PRECISION_WORKSTATION_2202	0x10280095 /*Precision Workstation 220 Chipset*/
	#define SUBDEVICE_INTEL_P3C_2000_SYSTEM_CHIPSET	0x1043801c /*P3C-2000 system chipset*/
#define DEVICE_INTEL_82820_820_CAMINO2			0x2501 /*82820 820 (Camino) Chipset Host Bridge (MCH)*/
	#define SUBDEVICE_INTEL_P3C_2000_SYSTEM_CHIPSET2	0x1043801c /*P3C-2000 system chipset*/
#define DEVICE_INTEL_82820_820_CAMINO3			0x250b /*82820 820 (Camino) Chipset Host Bridge*/
#define DEVICE_INTEL_82820_820_CAMINO4			0x250f /*82820 820 (Camino) Chipset AGP Bridge*/
#define DEVICE_INTEL_82805AA_MTH_MEMORY			0x2520 /*82805AA MTH Memory Translator Hub*/
#define DEVICE_INTEL_82804AA_MRH_S_MEMORY		0x2521 /*82804AA MRH-S Memory Repeater Hub for SDRAM*/
#define DEVICE_INTEL_82850_850_TEHAMA			0x2530 /*82850 850 (Tehama) Chipset Host Bridge (MCH)*/
	#define SUBDEVICE_INTEL_TH7II_RAID6		0x147b0507 /*TH7II-RAID*/
#define DEVICE_INTEL_82860_860_WOMBAT			0x2531 /*82860 860 (Wombat) Chipset Host Bridge (MCH)*/
#define DEVICE_INTEL_82850_850_TEHAMA2			0x2532 /*82850 850 (Tehama) Chipset AGP Bridge*/
#define DEVICE_INTEL_82860_860_WOMBAT2			0x2533 /*82860 860 (Wombat) Chipset AGP Bridge*/
#define DEVICE_INTEL_82860_860_WOMBAT3			0x2534 /*82860 860 (Wombat) Chipset PCI Bridge*/
#define DEVICE_INTEL_E7500_MEMORY_CONTROLLER		0x2540 /*E7500 Memory Controller Hub*/
	#define SUBDEVICE_INTEL_P4DP67			0x15d93480 /*P4DP6*/
#define DEVICE_INTEL_E7500_E7501_HOST_RASUM		0x2541 /*E7500/E7501 Host RASUM Controller*/
	#define SUBDEVICE_INTEL_P4DP68			0x15d93480 /*P4DP6*/
	#define SUBDEVICE_INTEL_CX9_VX99		0x4c531090 /*Cx9 / Vx9 mainboard*/
	#define SUBDEVICE_INTEL_SE7501HG2_MAINBOARD3	0x80863424 /*SE7501HG2 Mainboard*/
#define DEVICE_INTEL_E7500_E7501_HUB_INTERFACE		0x2543 /*E7500/E7501 Hub Interface B PCI-to-PCI Bridge*/
#define DEVICE_INTEL_E7500_E7501_HUB_INTERFACE2		0x2544 /*E7500/E7501 Hub Interface B RASUM Controller*/
	#define SUBDEVICE_INTEL_CX9_VX910		0x4c531090 /*Cx9 / Vx9 mainboard*/
#define DEVICE_INTEL_E7500_E7501_HUB_INTERFACE3		0x2545 /*E7500/E7501 Hub Interface C PCI-to-PCI Bridge*/
#define DEVICE_INTEL_E7500_E7501_HUB_INTERFACE4		0x2546 /*E7500/E7501 Hub Interface C RASUM Controller*/
#define DEVICE_INTEL_E7500_E7501_HUB_INTERFACE5		0x2547 /*E7500/E7501 Hub Interface D PCI-to-PCI Bridge*/
#define DEVICE_INTEL_E7500_E7501_HUB_INTERFACE6		0x2548 /*E7500/E7501 Hub Interface D RASUM Controller*/
#define DEVICE_INTEL_E7501_MEMORY_CONTROLLER		0x254c /*E7501 Memory Controller Hub*/
	#define SUBDEVICE_INTEL_CX9_VX911		0x4c531090 /*Cx9 / Vx9 mainboard*/
	#define SUBDEVICE_INTEL_SE7501HG2_MAINBOARD4	0x80863424 /*SE7501HG2 Mainboard*/
#define DEVICE_INTEL_E7505_MEMORY_CONTROLLER		0x2550 /*E7505 Memory Controller Hub*/
#define DEVICE_INTEL_E7505_E7205_SERIES_RAS		0x2551 /*E7505/E7205 Series RAS Controller*/
#define DEVICE_INTEL_E7505_E7205_PCI_TO_AGP_BRIDGE	0x2552 /*E7505/E7205 PCI-to-AGP Bridge*/
#define DEVICE_INTEL_E7505_HUB_INTERFACE		0x2553 /*E7505 Hub Interface B PCI-to-PCI Bridge*/
#define DEVICE_INTEL_E7505_HUB_INTERFACE2		0x2554 /*E7505 Hub Interface B PCI-to-PCI Bridge RAS Controller*/
#define DEVICE_INTEL_E7205_MEMORY_CONTROLLER		0x255d /*E7205 Memory Controller Hub*/
#define DEVICE_INTEL_82845G_GL_BROOKDALE_G_GE_PE_DRAM_CONTROLLER_HOST_HUB	0x2560 /*82845G/GL[Brookdale-G]/GE/PE DRAM Controller/Host-Hub Interface*/
	#define SUBDEVICE_INTEL_OPTIPLEX_GX2608		0x10280126 /*Optiplex GX260*/
	#define SUBDEVICE_INTEL_GA_8PE667_ULTRA4	0x14582560 /*GA-8PE667 Ultra*/
	#define SUBDEVICE_INTEL_845PE_MAX_MS_65809	0x14625800 /*845PE Max (MS-6580)*/
#define DEVICE_INTEL_82845G_GL_BROOKDALE_G_GE_PE_HOST_TO_AGP_BRIDGE	0x2561 /*82845G/GL[Brookdale-G]/GE/PE Host-to-AGP Bridge*/
#define DEVICE_INTEL_82845G_GL_BROOKDALE_G_GE_CHIPSET_INTEGRATED	0x2562 /*82845G/GL[Brookdale-G]/GE Chipset Integrated Graphics Device*/
	#define SUBDEVICE_INTEL_NETVISTA_A30P11		0x10140267 /*NetVista A30p*/
#define DEVICE_INTEL_82865G_PE_P_DRAM_CONTROLLER_HOST_HUB	0x2570 /*82865G/PE/P DRAM Controller/Host-Hub Interface*/
	#define SUBDEVICE_INTEL_P4P800_MAINBOARD9	0x104380f2 /*P4P800 Mainboard*/
	#define SUBDEVICE_INTEL_GA_8IPE1000_PRO2_MOTHERBOARD9	0x14582570 /*GA-8IPE1000 Pro2 motherboard (865PE)*/
#define DEVICE_INTEL_82865G_PE_P_PCI_TO			0x2571 /*82865G/PE/P PCI to AGP Controller*/
#define DEVICE_INTEL_82865G_INTEGRATED_GRAPHICS		0x2572 /*82865G Integrated Graphics Device*/
#define DEVICE_INTEL_82865G_PE_P_PCI_TO2		0x2573 /*82865G/PE/P PCI to CSA Bridge*/
#define DEVICE_INTEL_82865G_PE_P_PROCESSOR_TO		0x2576 /*82865G/PE/P Processor to I/O Memory Interface*/
#define DEVICE_INTEL_82875P_E7210_MEMORY_CONTROLLER	0x2578 /*82875P/E7210 Memory Controller Hub*/
	#define SUBDEVICE_INTEL_GA_8KNXP_MOTHERBOARD_875P3	0x14582578 /*GA-8KNXP motherboard (875P)*/
	#define SUBDEVICE_INTEL_MS_6758_875P_NEO2	0x14627580 /*MS-6758 (875P Neo)*/
	#define SUBDEVICE_INTEL_SUPER_MICRO_COMPUTER	0x15d94580 /*Super Micro Computer Inc. P4SCE*/
#define DEVICE_INTEL_82875P_PROCESSOR_TO		0x2579 /*82875P Processor to AGP Controller*/
#define DEVICE_INTEL_82875P_E7210_PROCESSOR_TO		0x257b /*82875P/E7210 Processor to PCI to CSA Bridge*/
#define DEVICE_INTEL_82875P_E7210_PROCESSOR_TO2		0x257e /*82875P/E7210 Processor to I/O Memory Interface*/
#define DEVICE_INTEL_915G_P_GV_PROCESSOR_TO		0x2580 /*915G/P/GV Processor to I/O Controller*/
#define DEVICE_INTEL_915G_P_GV_PCI_EXPRESS		0x2581 /*915G/P/GV PCI Express Root Port*/
#define DEVICE_INTEL_82915G_EXPRESS_CHIPSET		0x2582 /*82915G Express Chipset Family Graphics Controller*/
	#define SUBDEVICE_INTEL_OPTIPLEX_GX280		0x10281079 /*Optiplex GX280*/
#define DEVICE_INTEL_925X_XE_MEMORY_CONTROLLER		0x2584 /*925X/XE Memory Controller Hub*/
#define DEVICE_INTEL_925X_XE_PCI_EXPRESS		0x2585 /*925X/XE PCI Express Root Port*/
#define DEVICE_INTEL_E7220_E7221_MEMORY_CONTROLLER	0x2588 /*E7220/E7221 Memory Controller Hub*/
#define DEVICE_INTEL_E7220_E7221_PCI_EXPRESS		0x2589 /*E7220/E7221 PCI Express Root Port*/
#define DEVICE_INTEL_E7221_INTEGRATED_GRAPHICS		0x258a /*E7221 Integrated Graphics Controller*/
#define DEVICE_INTEL_MOBILE_MEMORY_CONTROLLER		0x2590 /*Mobile Memory Controller Hub*/
#define DEVICE_INTEL_MOBILE_MEMORY_CONTROLLER2		0x2591 /*Mobile Memory Controller Hub PCI Express Port*/
#define DEVICE_INTEL_MOBILE_GRAPHICS_CONTROLLER		0x2592 /*Mobile Graphics Controller*/
#define DEVICE_INTEL_6300ESB_LPC_INTERFACE		0x25a1 /*6300ESB LPC Interface Controller*/
#define DEVICE_INTEL_6300ESB_PATA_STORAGE		0x25a2 /*6300ESB PATA Storage Controller*/
	#define SUBDEVICE_INTEL_CL9_MAINBOARD2		0x4c5310b0 /*CL9 mainboard*/
#define DEVICE_INTEL_6300ESB_SATA_STORAGE		0x25a3 /*6300ESB SATA Storage Controller*/
	#define SUBDEVICE_INTEL_CL9_MAINBOARD3		0x4c5310b0 /*CL9 mainboard*/
#define DEVICE_INTEL_6300ESB_SMBUS_CONTROLLER		0x25a4 /*6300ESB SMBus Controller*/
	#define SUBDEVICE_INTEL_CL9_MAINBOARD4		0x4c5310b0 /*CL9 mainboard*/
#define DEVICE_INTEL_6300ESB_AC_97_AUDIO		0x25a6 /*6300ESB AC'97 Audio Controller*/
	#define SUBDEVICE_INTEL_CL9_MAINBOARD5		0x4c5310b0 /*CL9 mainboard*/
#define DEVICE_INTEL_6300ESB_AC_97_MODEM		0x25a7 /*6300ESB AC'97 Modem Controller*/
#define DEVICE_INTEL_6300ESB_USB_UNIVERSAL		0x25a9 /*6300ESB USB Universal Host Controller*/
	#define SUBDEVICE_INTEL_CL9_MAINBOARD6		0x4c5310b0 /*CL9 mainboard*/
#define DEVICE_INTEL_6300ESB_USB_UNIVERSAL2		0x25aa /*6300ESB USB Universal Host Controller*/
	#define SUBDEVICE_INTEL_CL9_MAINBOARD7		0x4c5310b0 /*CL9 mainboard*/
#define DEVICE_INTEL_6300ESB_WATCHDOG_TIMER		0x25ab /*6300ESB Watchdog Timer*/
	#define SUBDEVICE_INTEL_CL9_MAINBOARD8		0x4c5310b0 /*CL9 mainboard*/
#define DEVICE_INTEL_6300ESB_I_O_ADVANCED		0x25ac /*6300ESB I/O Advanced Programmable Interrupt Controller*/
	#define SUBDEVICE_INTEL_CL9_MAINBOARD9		0x4c5310b0 /*CL9 mainboard*/
#define DEVICE_INTEL_6300ESB_USB2_ENHANCED		0x25ad /*6300ESB USB2 Enhanced Host Controller*/
#define DEVICE_INTEL_6300ESB_64_BIT_PCI_X		0x25ae /*6300ESB 64-bit PCI-X Bridge*/
#define DEVICE_INTEL_6300ESB_SATA_RAID			0x25b0 /*6300ESB SATA RAID Controller*/
#define DEVICE_INTEL_SERVER_HUB_INTERFACE		0x2600 /*Server Hub Interface*/
#define DEVICE_INTEL_SERVER_HUB_PCI			0x2601 /*Server Hub PCI Express x4 Port D*/
#define DEVICE_INTEL_SERVER_HUB_PCI2			0x2602 /*Server Hub PCI Express x4 Port C0*/
#define DEVICE_INTEL_SERVER_HUB_PCI3			0x2603 /*Server Hub PCI Express x4 Port C1*/
#define DEVICE_INTEL_SERVER_HUB_PCI4			0x2604 /*Server Hub PCI Express x4 Port B0*/
#define DEVICE_INTEL_SERVER_HUB_PCI5			0x2605 /*Server Hub PCI Express x4 Port B1*/
#define DEVICE_INTEL_SERVER_HUB_PCI6			0x2606 /*Server Hub PCI Express x4 Port A0*/
#define DEVICE_INTEL_SERVER_HUB_PCI7			0x2607 /*Server Hub PCI Express x4 Port A1*/
#define DEVICE_INTEL_SERVER_HUB_PCI8			0x2608 /*Server Hub PCI Express x8 Port C*/
#define DEVICE_INTEL_SERVER_HUB_PCI9			0x2609 /*Server Hub PCI Express x8 Port B*/
#define DEVICE_INTEL_SERVER_HUB_PCI10			0x260a /*Server Hub PCI Express x8 Port A*/
#define DEVICE_INTEL_SERVER_HUB_IMI			0x260c /*Server Hub IMI Registers*/
#define DEVICE_INTEL_SERVER_HUB_SYSTEM			0x2610 /*Server Hub System Bus, Boot, and Interrupt Registers*/
#define DEVICE_INTEL_SERVER_HUB_ADDRESS			0x2611 /*Server Hub Address Mapping Registers*/
#define DEVICE_INTEL_SERVER_HUB_RAS			0x2612 /*Server Hub RAS Registers*/
#define DEVICE_INTEL_SERVER_HUB_RESERVED		0x2613 /*Server Hub Reserved Registers*/
#define DEVICE_INTEL_SERVER_HUB_RESERVED2		0x2614 /*Server Hub Reserved Registers*/
#define DEVICE_INTEL_SERVER_HUB_MISCELLANEOUS		0x2615 /*Server Hub Miscellaneous Registers*/
#define DEVICE_INTEL_SERVER_HUB_RESERVED3		0x2617 /*Server Hub Reserved Registers*/
#define DEVICE_INTEL_SERVER_HUB_RESERVED4		0x2618 /*Server Hub Reserved Registers*/
#define DEVICE_INTEL_SERVER_HUB_RESERVED5		0x2619 /*Server Hub Reserved Registers*/
#define DEVICE_INTEL_SERVER_HUB_RESERVED6		0x261a /*Server Hub Reserved Registers*/
#define DEVICE_INTEL_SERVER_HUB_RESERVED7		0x261b /*Server Hub Reserved Registers*/
#define DEVICE_INTEL_SERVER_HUB_RESERVED8		0x261c /*Server Hub Reserved Registers*/
#define DEVICE_INTEL_SERVER_HUB_RESERVED9		0x261d /*Server Hub Reserved Registers*/
#define DEVICE_INTEL_SERVER_HUB_RESERVED10		0x261e /*Server Hub Reserved Registers*/
#define DEVICE_INTEL_EXTERNAL_MEMORY_BRIDGE		0x2620 /*External Memory Bridge*/
#define DEVICE_INTEL_EXTERNAL_MEMORY_BRIDGE2		0x2621 /*External Memory Bridge Control Registers*/
#define DEVICE_INTEL_EXTERNAL_MEMORY_BRIDGE3		0x2622 /*External Memory Bridge Memory Interleaving Registers*/
#define DEVICE_INTEL_EXTERNAL_MEMORY_BRIDGE4		0x2623 /*External Memory Bridge DDR Initialization and Calibration*/
#define DEVICE_INTEL_EXTERNAL_MEMORY_BRIDGE5		0x2624 /*External Memory Bridge Reserved Registers*/
#define DEVICE_INTEL_EXTERNAL_MEMORY_BRIDGE6		0x2625 /*External Memory Bridge Reserved Registers*/
#define DEVICE_INTEL_EXTERNAL_MEMORY_BRIDGE7		0x2626 /*External Memory Bridge Reserved Registers*/
#define DEVICE_INTEL_EXTERNAL_MEMORY_BRIDGE8		0x2627 /*External Memory Bridge Reserved Registers*/
#define DEVICE_INTEL_82801FB_FR_ICH6_ICH6R_LPC		0x2640 /*82801FB/FR (ICH6/ICH6R) LPC Interface Bridge*/
#define DEVICE_INTEL_82801FBM_ICH6M_LPC			0x2641 /*82801FBM (ICH6M) LPC Interface Bridge*/
#define DEVICE_INTEL_82801FW_FRW_ICH6W_ICH6RW_LPC	0x2642 /*82801FW/FRW (ICH6W/ICH6RW) LPC Interface Bridge*/
#define DEVICE_INTEL_82801FB_FW_ICH6_ICH6W_SATA		0x2651 /*82801FB/FW (ICH6/ICH6W) SATA Controller*/
	#define SUBDEVICE_INTEL_OPTIPLEX_GX2802		0x10280179 /*Optiplex GX280*/
#define DEVICE_INTEL_82801FR_FRW_ICH6R_ICH6RW_SATA	0x2652 /*82801FR/FRW (ICH6R/ICH6RW) SATA Controller*/
#define DEVICE_INTEL_82801FBM_ICH6M_SATA		0x2653 /*82801FBM (ICH6M) SATA Controller*/
#define DEVICE_INTEL_82801FB_FBM_FR_FW_FRW_ICH6_FAMILY	0x2658 /*82801FB/FBM/FR/FW/FRW (ICH6 Family) USB UHCI #1*/
	#define SUBDEVICE_INTEL_OPTIPLEX_GX2803		0x10280179 /*Optiplex GX280*/
#define DEVICE_INTEL_82801FB_FBM_FR_FW_FRW_ICH6_FAMILY2	0x2659 /*82801FB/FBM/FR/FW/FRW (ICH6 Family) USB UHCI #2*/
	#define SUBDEVICE_INTEL_OPTIPLEX_GX2804		0x10280179 /*Optiplex GX280*/
#define DEVICE_INTEL_82801FB_FBM_FR_FW_FRW_ICH6_FAMILY3	0x265a /*82801FB/FBM/FR/FW/FRW (ICH6 Family) USB UHCI #3*/
	#define SUBDEVICE_INTEL_OPTIPLEX_GX2805		0x10280179 /*Optiplex GX280*/
#define DEVICE_INTEL_82801FB_FBM_FR_FW_FRW_ICH6_FAMILY4	0x265b /*82801FB/FBM/FR/FW/FRW (ICH6 Family) USB UHCI #4*/
	#define SUBDEVICE_INTEL_OPTIPLEX_GX2806		0x10280179 /*Optiplex GX280*/
#define DEVICE_INTEL_82801FB_FBM_FR_FW_FRW_ICH6_FAMILY5	0x265c /*82801FB/FBM/FR/FW/FRW (ICH6 Family) USB2 EHCI Controller*/
	#define SUBDEVICE_INTEL_OPTIPLEX_GX2807		0x10280179 /*Optiplex GX280*/
#define DEVICE_INTEL_82801FB_FBM_FR_FW_FRW_ICH6_FAMILY6	0x2660 /*82801FB/FBM/FR/FW/FRW (ICH6 Family) PCI Express Port 1*/
#define DEVICE_INTEL_82801FB_FBM_FR_FW_FRW_ICH6_FAMILY7	0x2662 /*82801FB/FBM/FR/FW/FRW (ICH6 Family) PCI Express Port 2*/
#define DEVICE_INTEL_82801FB_FBM_FR_FW_FRW_ICH6_FAMILY8	0x2664 /*82801FB/FBM/FR/FW/FRW (ICH6 Family) PCI Express Port 3*/
#define DEVICE_INTEL_82801FB_FBM_FR_FW_FRW_ICH6_FAMILY9	0x2666 /*82801FB/FBM/FR/FW/FRW (ICH6 Family) PCI Express Port 4*/
#define DEVICE_INTEL_82801FB_FBM_FR_FW_FRW_ICH6_FAMILY10	0x2668 /*82801FB/FBM/FR/FW/FRW (ICH6 Family) High Definition Audio Controller*/
#define DEVICE_INTEL_82801FB_FBM_FR_FW_FRW_ICH6_FAMILY11	0x266a /*82801FB/FBM/FR/FW/FRW (ICH6 Family) SMBus Controller*/
	#define SUBDEVICE_INTEL_OPTIPLEX_GX2808		0x10280179 /*Optiplex GX280*/
#define DEVICE_INTEL_82801FB_FBM_FR_FW_FRW_ICH6_FAMILY12	0x266c /*82801FB/FBM/FR/FW/FRW (ICH6 Family) LAN Controller*/
#define DEVICE_INTEL_82801FB_FBM_FR_FW_FRW_ICH6_FAMILY13	0x266d /*82801FB/FBM/FR/FW/FRW (ICH6 Family) AC'97 Modem Controller*/
#define DEVICE_INTEL_82801FB_FBM_FR_FW_FRW_ICH6_FAMILY14	0x266e /*82801FB/FBM/FR/FW/FRW (ICH6 Family) AC'97 Audio Controller*/
	#define SUBDEVICE_INTEL_OPTIPLEX_GX2809		0x10280179 /*Optiplex GX280*/
#define DEVICE_INTEL_82801FB_FBM_FR_FW_FRW_ICH6_FAMILY15	0x266f /*82801FB/FBM/FR/FW/FRW (ICH6 Family) IDE Controller*/
#define DEVICE_INTEL_82915G_EXPRESS_CHIPSET2		0x2782 /*82915G Express Chipset Family Graphics Controller*/
#define DEVICE_INTEL_MOBILE_GRAPHICS_CONTROLLER2	0x2792 /*Mobile Graphics Controller*/
#define DEVICE_INTEL_INTEGRATED_RAID			0x3092 /*Integrated RAID*/
#define DEVICE_INTEL_GD31244_PCI_X_SATA			0x3200 /*GD31244 PCI-X SATA HBA*/
#define DEVICE_INTEL_82855PM_PROCESSOR_TO		0x3340 /*82855PM Processor to I/O Controller*/
	#define SUBDEVICE_INTEL_TRAVELMATE_2909		0x1025005a /*TravelMate 290*/
	#define SUBDEVICE_INTEL_NC6000_LAPTOP9		0x103c0890 /*NC6000 laptop*/
#define DEVICE_INTEL_82855PM_PROCESSOR_TO2		0x3341 /*82855PM Processor to AGP Controller*/
#define DEVICE_INTEL_82830_830_CHIPSET			0x3575 /*82830 830 Chipset Host Bridge*/
	#define SUBDEVICE_INTEL_THINKPAD_A_T_X_SERIES10	0x1014021d /*ThinkPad A/T/X Series*/
	#define SUBDEVICE_INTEL_VAIO_PCG_GR214EP_GR214MP_GR215MP_GR314MP_GR315MP9	0x104d80e7 /*VAIO PCG-GR214EP/GR214MP/GR215MP/GR314MP/GR315MP*/
#define DEVICE_INTEL_82830_830_CHIPSET2			0x3576 /*82830 830 Chipset AGP Bridge*/
#define DEVICE_INTEL_82830_CGC_CHIPSET			0x3577 /*82830 CGC [Chipset Graphics Controller]*/
	#define SUBDEVICE_INTEL_THINKPAD_A_T_X_SERIES11	0x10140513 /*ThinkPad A/T/X Series*/
#define DEVICE_INTEL_82830_830_CHIPSET3			0x3578 /*82830 830 Chipset Host Bridge*/
#define DEVICE_INTEL_82852_82855_GM_GME_PM_GMV_PROCESSOR	0x3580 /*82852/82855 GM/GME/PM/GMV Processor to I/O Controller*/
	#define SUBDEVICE_INTEL_LATITUDE_D5057		0x10280163 /*Latitude D505*/
	#define SUBDEVICE_INTEL_CL9_MAINBOARD10		0x4c5310b0 /*CL9 mainboard*/
#define DEVICE_INTEL_82852_82855_GM_GME_PM_GMV_PROCESSOR2	0x3581 /*82852/82855 GM/GME/PM/GMV Processor to AGP Controller*/
#define DEVICE_INTEL_82852_855GM_INTEGRATED_GRAPHICS	0x3582 /*82852/855GM Integrated Graphics Device*/
	#define SUBDEVICE_INTEL_LATITUDE_D5058		0x10280163 /*Latitude D505*/
	#define SUBDEVICE_INTEL_CL9_MAINBOARD11		0x4c5310b0 /*CL9 mainboard*/
#define DEVICE_INTEL_82852_82855_GM_GME_PM_GMV_PROCESSOR3	0x3584 /*82852/82855 GM/GME/PM/GMV Processor to I/O Controller*/
	#define SUBDEVICE_INTEL_LATITUDE_D5059		0x10280163 /*Latitude D505*/
	#define SUBDEVICE_INTEL_CL9_MAINBOARD12		0x4c5310b0 /*CL9 mainboard*/
#define DEVICE_INTEL_82852_82855_GM_GME_PM_GMV_PROCESSOR4	0x3585 /*82852/82855 GM/GME/PM/GMV Processor to I/O Controller*/
	#define SUBDEVICE_INTEL_LATITUDE_D50510		0x10280163 /*Latitude D505*/
	#define SUBDEVICE_INTEL_CL9_MAINBOARD13		0x4c5310b0 /*CL9 mainboard*/
#define DEVICE_INTEL_E7520_MEMORY_CONTROLLER		0x3590 /*E7520 Memory Controller Hub*/
#define DEVICE_INTEL_E7525_E7520_ERROR_REPORTING	0x3591 /*E7525/E7520 Error Reporting Registers*/
#define DEVICE_INTEL_E7320_MEMORY_CONTROLLER		0x3592 /*E7320 Memory Controller Hub*/
#define DEVICE_INTEL_E7320_ERROR_REPORTING		0x3593 /*E7320 Error Reporting Registers*/
#define DEVICE_INTEL_E7520_DMA_CONTROLLER		0x3594 /*E7520 DMA Controller*/
#define DEVICE_INTEL_E7525_E7520_E7320_PCI_EXPRESS	0x3595 /*E7525/E7520/E7320 PCI Express Port A*/
#define DEVICE_INTEL_E7525_E7520_E7320_PCI_EXPRESS2	0x3596 /*E7525/E7520/E7320 PCI Express Port A1*/
#define DEVICE_INTEL_E7525_E7520_PCI_EXPRESS		0x3597 /*E7525/E7520 PCI Express Port B*/
#define DEVICE_INTEL_E7520_PCI_EXPRESS			0x3598 /*E7520 PCI Express Port B1*/
#define DEVICE_INTEL_E7520_PCI_EXPRESS2			0x3599 /*E7520 PCI Express Port C*/
#define DEVICE_INTEL_E7520_PCI_EXPRESS3			0x359a /*E7520 PCI Express Port C1*/
#define DEVICE_INTEL_E7525_E7520_E7320_EXTENDED_CONFIGURATION	0x359b /*E7525/E7520/E7320 Extended Configuration Registers*/
#define DEVICE_INTEL_E7525_MEMORY_CONTROLLER		0x359e /*E7525 Memory Controller Hub*/
#define DEVICE_INTEL_PRO_WIRELESS_2200BG		0x4220 /*PRO/Wireless 2200BG*/
#define DEVICE_INTEL_PRO_WIRELESS_2915ABG_MINIPCI	0x4223 /*PRO/Wireless 2915ABG MiniPCI Adapter*/
#define DEVICE_INTEL_ETHEREXPRESS_PRO_100_INTELLIGENT	0x5200 /*EtherExpress PRO/100 Intelligent Server*/
#define DEVICE_INTEL_ETHEREXPRESS_PRO_100_INTELLIGENT2	0x5201 /*EtherExpress PRO/100 Intelligent Server*/
	#define SUBDEVICE_INTEL_ETHEREXPRESS_PRO_100_SERVER6	0x80860001 /*EtherExpress PRO/100 Server Ethernet Adapter*/
#define DEVICE_INTEL_80310_IOP_IO			0x530d /*80310 IOP [IO Processor]*/
#define DEVICE_INTEL_82371SB_PIIX3_ISA			0x7000 /*82371SB PIIX3 ISA [Natoma/Triton II]*/
#define DEVICE_INTEL_82371SB_PIIX3_IDE			0x7010 /*82371SB PIIX3 IDE [Natoma/Triton II]*/
#define DEVICE_INTEL_82371SB_PIIX3_USB			0x7020 /*82371SB PIIX3 USB [Natoma/Triton II]*/
#define DEVICE_INTEL_430VX_82437VX			0x7030 /*430VX - 82437VX TVX [Triton VX]*/
#define DEVICE_INTEL_INTEL_INTERCAST_VIDEO		0x7050 /*Intel Intercast Video Capture Card*/
#define DEVICE_INTEL_430TX_82439TX			0x7100 /*430TX - 82439TX MTXC*/
#define DEVICE_INTEL_82371AB_EB_MB_PIIX4_ISA		0x7110 /*82371AB/EB/MB PIIX4 ISA*/
	#define SUBDEVICE_INTEL_VIRTUALHW_V3		0x15ad1976 /*virtualHW v3*/
#define DEVICE_INTEL_82371AB_EB_MB_PIIX4_IDE		0x7111 /*82371AB/EB/MB PIIX4 IDE*/
	#define SUBDEVICE_INTEL_VIRTUALHW_V32		0x15ad1976 /*virtualHW v3*/
#define DEVICE_INTEL_82371AB_EB_MB_PIIX4_USB		0x7112 /*82371AB/EB/MB PIIX4 USB*/
	#define SUBDEVICE_INTEL_VIRTUALHW_V33		0x15ad1976 /*virtualHW v3*/
#define DEVICE_INTEL_82371AB_EB_MB_PIIX4_ACPI		0x7113 /*82371AB/EB/MB PIIX4 ACPI*/
	#define SUBDEVICE_INTEL_VIRTUALHW_V34		0x15ad1976 /*virtualHW v3*/
#define DEVICE_INTEL_82810_GMCH_GRAPHICS		0x7120 /*82810 GMCH [Graphics Memory Controller Hub]*/
	#define SUBDEVICE_INTEL_CL7_MAINBOARD		0x4c531040 /*CL7 mainboard*/
	#define SUBDEVICE_INTEL_PC7_MAINBOARD		0x4c531060 /*PC7 mainboard*/
#define DEVICE_INTEL_82810_CGC_CHIPSET			0x7121 /*82810 CGC [Chipset Graphics Controller]*/
	#define SUBDEVICE_INTEL_CL7_MAINBOARD2		0x4c531040 /*CL7 mainboard*/
	#define SUBDEVICE_INTEL_PC7_MAINBOARD2		0x4c531060 /*PC7 mainboard*/
	#define SUBDEVICE_INTEL_CAYMAN_CA810_MAINBOARD	0x80864341 /*Cayman (CA810) Mainboard*/
#define DEVICE_INTEL_82810_DC_100_GMCH			0x7122 /*82810 DC-100 GMCH [Graphics Memory Controller Hub]*/
#define DEVICE_INTEL_82810_DC_100_CGC			0x7123 /*82810 DC-100 CGC [Chipset Graphics Controller]*/
#define DEVICE_INTEL_82810E_DC_133_GMCH			0x7124 /*82810E DC-133 GMCH [Graphics Memory Controller Hub]*/
#define DEVICE_INTEL_82810E_DC_133_CGC			0x7125 /*82810E DC-133 CGC [Chipset Graphics Controller]*/
#define DEVICE_INTEL_82810_DC_133_SYSTEM		0x7126 /*82810 DC-133 System and Graphics Controller*/
#define DEVICE_INTEL_82810_M_DC_100_SYSTEM		0x7128 /*82810-M DC-100 System and Graphics Controller*/
#define DEVICE_INTEL_82810_M_DC_133_SYSTEM		0x712a /*82810-M DC-133 System and Graphics Controller*/
#define DEVICE_INTEL_440LX_EX_82443LX_EX		0x7180 /*440LX/EX - 82443LX/EX Host bridge*/
#define DEVICE_INTEL_440LX_EX_82443LX_EX2		0x7181 /*440LX/EX - 82443LX/EX AGP bridge*/
#define DEVICE_INTEL_440BX_ZX_DX_82443BX_ZX_DX		0x7190 /*440BX/ZX/DX - 82443BX/ZX/DX Host bridge*/
	#define SUBDEVICE_INTEL_ARMADA_1750_LAPTOP	0x0e110500 /*Armada 1750 Laptop System Chipset*/
	#define SUBDEVICE_INTEL_ARMADA_M700_E500	0x0e11b110 /*Armada M700/E500*/
	#define SUBDEVICE_INTEL_TOSHIBA_TECRA_8100	0x11790001 /*Toshiba Tecra 8100 Laptop System Chipset*/
	#define SUBDEVICE_INTEL_VIRTUALHW_V35		0x15ad1976 /*virtualHW v3*/
	#define SUBDEVICE_INTEL_CT7_MAINBOARD2		0x4c531050 /*CT7 mainboard*/
	#define SUBDEVICE_INTEL_CE7_MAINBOARD2		0x4c531051 /*CE7 mainboard*/
#define DEVICE_INTEL_440BX_ZX_DX_82443BX_ZX_DX2		0x7191 /*440BX/ZX/DX - 82443BX/ZX/DX AGP bridge*/
#define DEVICE_INTEL_440BX_ZX_DX_82443BX_ZX_DX3		0x7192 /*440BX/ZX/DX - 82443BX/ZX/DX Host bridge (AGP disabled)*/
	#define SUBDEVICE_INTEL_ARMADA_1700_LAPTOP	0x0e110460 /*Armada 1700 Laptop System Chipset*/
	#define SUBDEVICE_INTEL_CC7_CR7_CP7_VC7_VP7_VR7_MAINBOARD	0x4c531000 /*CC7/CR7/CP7/VC7/VP7/VR7 mainboard*/
#define DEVICE_INTEL_82440MX_HOST_BRIDGE		0x7194 /*82440MX Host Bridge*/
	#define SUBDEVICE_INTEL_VERSA_NOTE_VXI		0x10330000 /*Versa Note Vxi*/
	#define SUBDEVICE_INTEL_CA3_CR3_MAINBOARD2	0x4c5310a0 /*CA3/CR3 mainboard*/
#define DEVICE_INTEL_82440MX_AC_97_AUDIO		0x7195 /*82440MX AC'97 Audio Controller*/
	#define SUBDEVICE_INTEL_VERSA_NOTE_VXI2		0x103380cc /*Versa Note VXi*/
	#define SUBDEVICE_INTEL_QSOUND_SIGMATEL_STAC97_PCI	0x10cf1099 /*QSound_SigmaTel Stac97 PCI Audio*/
	#define SUBDEVICE_INTEL_SOUNDMAX_INTEGRATED_DIGITAL6	0x11d40040 /*SoundMAX Integrated Digital Audio*/
	#define SUBDEVICE_INTEL_SOUNDMAX_INTEGRATED_DIGITAL7	0x11d40048 /*SoundMAX Integrated Digital Audio*/
#define DEVICE_INTEL_82440MX_AC_97_MODEM		0x7196 /*82440MX AC'97 Modem Controller*/
#define DEVICE_INTEL_82440MX_ISA_BRIDGE			0x7198 /*82440MX ISA Bridge*/
#define DEVICE_INTEL_82440MX_EIDE_CONTROLLER		0x7199 /*82440MX EIDE Controller*/
#define DEVICE_INTEL_82440MX_USB_UNIVERSAL		0x719a /*82440MX USB Universal Host Controller*/
#define DEVICE_INTEL_82440MX_POWER_MANAGEMENT		0x719b /*82440MX Power Management Controller*/
#define DEVICE_INTEL_440GX_82443GX			0x71a0 /*440GX - 82443GX Host bridge*/
	#define SUBDEVICE_INTEL_CT7_MAINBOARD3		0x4c531050 /*CT7 mainboard*/
	#define SUBDEVICE_INTEL_CE7_MAINBOARD3		0x4c531051 /*CE7 mainboard*/
#define DEVICE_INTEL_440GX_82443GX2			0x71a1 /*440GX - 82443GX AGP bridge*/
#define DEVICE_INTEL_440GX_82443GX3			0x71a2 /*440GX - 82443GX Host bridge (AGP disabled)*/
	#define SUBDEVICE_INTEL_CC7_CR7_CP7_VC7_VP7_VR7_MAINBOARD2	0x4c531000 /*CC7/CR7/CP7/VC7/VP7/VR7 mainboard*/
#define DEVICE_INTEL_82372FB_PIIX5_ISA			0x7600 /*82372FB PIIX5 ISA*/
#define DEVICE_INTEL_82372FB_PIIX5_IDE			0x7601 /*82372FB PIIX5 IDE*/
#define DEVICE_INTEL_82372FB_PIIX5_USB			0x7602 /*82372FB PIIX5 USB*/
#define DEVICE_INTEL_82372FB_PIIX5_SMBUS		0x7603 /*82372FB PIIX5 SMBus*/
#define DEVICE_INTEL_82740_I740_AGP			0x7800 /*82740 (i740) AGP Graphics Accelerator*/
	#define SUBDEVICE_INTEL_STARFIGHTER_AGP		0x003d0008 /*Starfighter AGP*/
	#define SUBDEVICE_INTEL_STARFIGHTER_AGP2	0x003d000b /*Starfighter AGP*/
	#define SUBDEVICE_INTEL_STEALTH_II_G460		0x10920100 /*Stealth II G460*/
	#define SUBDEVICE_INTEL_LIGHTSPEED_740		0x10b4201a /*Lightspeed 740*/
	#define SUBDEVICE_INTEL_LIGHTSPEED_7402		0x10b4202f /*Lightspeed 740*/
	#define SUBDEVICE_INTEL_TERMINATOR_2X_I		0x80860000 /*Terminator 2x/i*/
	#define SUBDEVICE_INTEL_INTEL740_GRAPHICS_ACCELERATOR	0x80860100 /*Intel740 Graphics Accelerator*/
#define DEVICE_INTEL_450KX_GX_ORION			0x84c4 /*450KX/GX [Orion] - 82454KX/GX PCI bridge*/
#define DEVICE_INTEL_450KX_GX_ORION2			0x84c5 /*450KX/GX [Orion] - 82453KX/GX Memory controller*/
#define DEVICE_INTEL_450NX_82451NX			0x84ca /*450NX - 82451NX Memory & I/O Controller*/
#define DEVICE_INTEL_450NX_82454NX_84460GX		0x84cb /*450NX - 82454NX/84460GX PCI Expander Bridge*/
#define DEVICE_INTEL_460GX_84460GX			0x84e0 /*460GX - 84460GX System Address Controller (SAC)*/
#define DEVICE_INTEL_460GX_84460GX2			0x84e1 /*460GX - 84460GX System Data Controller (SDC)*/
#define DEVICE_INTEL_460GX_84460GX3			0x84e2 /*460GX - 84460GX AGP Bridge (GXB function 2)*/
#define DEVICE_INTEL_460GX_84460GX4			0x84e3 /*460GX - 84460GX Memory Address Controller (MAC)*/
#define DEVICE_INTEL_460GX_84460GX5			0x84e4 /*460GX - 84460GX Memory Data Controller (MDC)*/
#define DEVICE_INTEL_460GX_82466GX			0x84e6 /*460GX - 82466GX Wide and fast PCI eXpander Bridge (WXB)*/
#define DEVICE_INTEL_460GX_84460GX6			0x84ea /*460GX - 84460GX AGP Bridge (GXB function 1)*/
#define DEVICE_INTEL_IXP4XX_INTEL			0x8500 /*IXP4XX - Intel Network Processor family. IXP420, IXP421, IXP422, IXP425 and IXC1100*/
#define DEVICE_INTEL_IXP2000_FAMILY_NETWORK		0x9000 /*IXP2000 Family Network Processor*/
#define DEVICE_INTEL_IXP2400_NETWORK_PROCESSOR		0x9001 /*IXP2400 Network Processor*/
#define DEVICE_INTEL_IXP2800_NETWORK_PROCESSOR		0x9004 /*IXP2800 Network Processor*/
#define DEVICE_INTEL_INTEGRATED_RAID2			0x9621 /*Integrated RAID*/
#define DEVICE_INTEL_INTEGRATED_RAID3			0x9622 /*Integrated RAID*/
#define DEVICE_INTEL_INTEGRATED_RAID4			0x9641 /*Integrated RAID*/
#define DEVICE_INTEL_INTEGRATED_RAID5			0x96a1 /*Integrated RAID*/
#define DEVICE_INTEL_PRO_10GBE_LR_SERVER2		0xa01f /*PRO/10GbE LR Server Adapter*/
#define DEVICE_INTEL_PRO_10GBE_LR_SERVER3		0xa11f /*PRO/10GbE LR Server Adapter*/
#define DEVICE_INTEL_21152_PCI_TO_PCI_BRIDGE		0xb152 /*21152 PCI-to-PCI Bridge*/
#define DEVICE_INTEL_21154_PCI_TO_PCI_BRIDGE		0xb154 /*21154 PCI-to-PCI Bridge*/
#define DEVICE_INTEL_21555_NON_TRANSPARENT		0xb555 /*21555 Non transparent PCI-to-PCI Bridge*/
	#define SUBDEVICE_INTEL_PCI_VOIP_GATEWAY	0x12d9000a /*PCI VoIP Gateway*/
	#define SUBDEVICE_INTEL_CT7_MAINBOARD4		0x4c531050 /*CT7 mainboard*/
	#define SUBDEVICE_INTEL_CE7_MAINBOARD4		0x4c531051 /*CE7 mainboard*/
	#define SUBDEVICE_INTEL_CC8_1_BLUES		0xe4bf1000 /*CC8-1-BLUES*/
#define DEVICE_INTEL_450NX_GX_ORION			0xffff /*450NX/GX [Orion] - 82453KX/GX Memory controller [BUG]*/
/* Vendor: 8401: TRENDware International Inc. */
/* Vendor: 8800: Trigem Computer Inc. */
#define DEVICE_TRIGEM2_VIDEO_ASSISTENT_COMPONENT	0x2008 /*Video assistent component*/
/* Vendor: 8866: T-Square Design Inc. */
/* Vendor: 8888: Silicon Magic */
/* Vendor: 8c4a: Winbond */
#define DEVICE_WINBOND2_W89C940_MISPROGRAMMED_NE2K	0x1980 /*W89C940 misprogrammed [ne2k]*/
/* Vendor: 8e0e: Computone Corporation */
/* Vendor: 8e2e: KTI */
#define DEVICE_KTI_ET32P2				0x3000 /*ET32P2*/
/* Vendor: 9004: Adaptec */
#define DEVICE_ADAPTEC2_AHA_2940U_CN			0x0078 /*AHA-2940U_CN*/
#define DEVICE_ADAPTEC2_AIC_7810			0x1078 /*AIC-7810*/
#define DEVICE_ADAPTEC2_AIC_1160_FAMILY_FIBRE		0x1160 /*AIC-1160 [Family Fibre Channel Adapter]*/
#define DEVICE_ADAPTEC2_AIC_7821			0x2178 /*AIC-7821*/
#define DEVICE_ADAPTEC2_AHA_2930CU			0x3860 /*AHA-2930CU*/
#define DEVICE_ADAPTEC2_AHA_4844W_4844UW		0x3b78 /*AHA-4844W/4844UW*/
#define DEVICE_ADAPTEC2_AIC_755X			0x5075 /*AIC-755x*/
#define DEVICE_ADAPTEC2_AHA_7850			0x5078 /*AHA-7850*/
	#define SUBDEVICE_ADAPTEC2_AHA_2904_INTEGRATED_AIC_7850	0x90047850 /*AHA-2904/Integrated AIC-7850*/
#define DEVICE_ADAPTEC2_AIC_755X2			0x5175 /*AIC-755x*/
#define DEVICE_ADAPTEC2_AIC_7851			0x5178 /*AIC-7851*/
#define DEVICE_ADAPTEC2_AIC_755X3			0x5275 /*AIC-755x*/
#define DEVICE_ADAPTEC2_AIC_7852			0x5278 /*AIC-7852*/
#define DEVICE_ADAPTEC2_AIC_755X4			0x5375 /*AIC-755x*/
#define DEVICE_ADAPTEC2_AIC_7850			0x5378 /*AIC-7850*/
#define DEVICE_ADAPTEC2_AIC_755X5			0x5475 /*AIC-755x*/
#define DEVICE_ADAPTEC2_AIC_78502			0x5478 /*AIC-7850*/
#define DEVICE_ADAPTEC2_AVA_2930			0x5575 /*AVA-2930*/
#define DEVICE_ADAPTEC2_AIC_7855			0x5578 /*AIC-7855*/
#define DEVICE_ADAPTEC2_ANA_7711_TCP_OFFLOAD		0x5647 /*ANA-7711 TCP Offload Engine*/
	#define SUBDEVICE_ADAPTEC2_ANA_7711F_TCP_OFFLOAD	0x90047710 /*ANA-7711F TCP Offload Engine - Optical*/
	#define SUBDEVICE_ADAPTEC2_ANA_7711LP_TCP_OFFLOAD	0x90047711 /*ANA-7711LP TCP Offload Engine - Copper*/
#define DEVICE_ADAPTEC2_AIC_755X6			0x5675 /*AIC-755x*/
#define DEVICE_ADAPTEC2_AIC_7856			0x5678 /*AIC-7856*/
#define DEVICE_ADAPTEC2_AIC_755X7			0x5775 /*AIC-755x*/
#define DEVICE_ADAPTEC2_AIC_78503			0x5778 /*AIC-7850*/
#define DEVICE_ADAPTEC2_AIC_5800			0x5800 /*AIC-5800*/
#define DEVICE_ADAPTEC2_ANA_5910_5930_5940_ATM155	0x5900 /*ANA-5910/5930/5940 ATM155 & 25 LAN Adapter*/
#define DEVICE_ADAPTEC2_ANA_5910A_5930A_5940A_ATM_ADAPTER	0x5905 /*ANA-5910A/5930A/5940A ATM Adapter*/
#define DEVICE_ADAPTEC2_AIC_3860			0x6038 /*AIC-3860*/
#define DEVICE_ADAPTEC2_AIC_1480_APA_1480		0x6075 /*AIC-1480 / APA-1480*/
	#define SUBDEVICE_ADAPTEC2_AIC_1480_APA_1480	0x90047560 /*AIC-1480 / APA-1480 Cardbus*/
#define DEVICE_ADAPTEC2_AIC_7860			0x6078 /*AIC-7860*/
#define DEVICE_ADAPTEC2_AIC_7861			0x6178 /*AIC-7861*/
	#define SUBDEVICE_ADAPTEC2_AHA_2940AU_SINGLE	0x90047861 /*AHA-2940AU Single*/
#define DEVICE_ADAPTEC2_AIC_78602			0x6278 /*AIC-7860*/
#define DEVICE_ADAPTEC2_AIC_78603			0x6378 /*AIC-7860*/
#define DEVICE_ADAPTEC2_AIC_786X			0x6478 /*AIC-786x*/
#define DEVICE_ADAPTEC2_AIC_786X2			0x6578 /*AIC-786x*/
#define DEVICE_ADAPTEC2_AIC_786X3			0x6678 /*AIC-786x*/
#define DEVICE_ADAPTEC2_AIC_786X4			0x6778 /*AIC-786x*/
#define DEVICE_ADAPTEC2_ANA620XX_ANA69011A		0x6915 /*ANA620xx/ANA69011A*/
	#define SUBDEVICE_ADAPTEC2_ANA69011A_TX_10_100	0x90040008 /*ANA69011A/TX 10/100*/
	#define SUBDEVICE_ADAPTEC2_ANA69011A_TX_10_1002	0x90040009 /*ANA69011A/TX 10/100*/
	#define SUBDEVICE_ADAPTEC2_ANA62022_2_PORT_10_100	0x90040010 /*ANA62022 2-port 10/100*/
	#define SUBDEVICE_ADAPTEC2_ANA62044_4_PORT_10_100	0x90040018 /*ANA62044 4-port 10/100*/
	#define SUBDEVICE_ADAPTEC2_ANA62044_4_PORT_10_1002	0x90040019 /*ANA62044 4-port 10/100*/
	#define SUBDEVICE_ADAPTEC2_ANA62022_2_PORT_10_1002	0x90040020 /*ANA62022 2-port 10/100*/
	#define SUBDEVICE_ADAPTEC2_ANA69011A_TX_10_1003	0x90040028 /*ANA69011A/TX 10/100*/
	#define SUBDEVICE_ADAPTEC2_ANA69011A_TX_64_BIT	0x90048008 /*ANA69011A/TX 64 bit 10/100*/
	#define SUBDEVICE_ADAPTEC2_ANA69011A_TX_64_BIT2	0x90048009 /*ANA69011A/TX 64 bit 10/100*/
	#define SUBDEVICE_ADAPTEC2_ANA62022_2_PORT_64	0x90048010 /*ANA62022 2-port 64 bit 10/100*/
	#define SUBDEVICE_ADAPTEC2_ANA62044_4_PORT_64	0x90048018 /*ANA62044 4-port 64 bit 10/100*/
	#define SUBDEVICE_ADAPTEC2_ANA62044_4_PORT_642	0x90048019 /*ANA62044 4-port 64 bit 10/100*/
	#define SUBDEVICE_ADAPTEC2_ANA62022_2_PORT_642	0x90048020 /*ANA62022 2-port 64 bit 10/100*/
	#define SUBDEVICE_ADAPTEC2_ANA69011A_TX_64_BIT3	0x90048028 /*ANA69011A/TX 64 bit 10/100*/
#define DEVICE_ADAPTEC2_AHA_294X_AIC_7870		0x7078 /*AHA-294x / AIC-7870*/
#define DEVICE_ADAPTEC2_AHA_2940_2940W_AIC_7871		0x7178 /*AHA-2940/2940W / AIC-7871*/
#define DEVICE_ADAPTEC2_AHA_3940_3940W_AIC_7872		0x7278 /*AHA-3940/3940W / AIC-7872*/
#define DEVICE_ADAPTEC2_AHA_3985_AIC_7873		0x7378 /*AHA-3985 / AIC-7873*/
#define DEVICE_ADAPTEC2_AHA_2944_2944W_AIC_7874		0x7478 /*AHA-2944/2944W / AIC-7874*/
#define DEVICE_ADAPTEC2_AHA_3944_3944W_AIC_7875		0x7578 /*AHA-3944/3944W / AIC-7875*/
#define DEVICE_ADAPTEC2_AHA_4944W_UW_AIC_7876		0x7678 /*AHA-4944W/UW / AIC-7876*/
#define DEVICE_ADAPTEC2_ANA_7711F_NETWORK_ACCELERATOR	0x7710 /*ANA-7711F Network Accelerator Card (NAC) - Optical*/
#define DEVICE_ADAPTEC2_ANA_7711C_NETWORK_ACCELERATOR	0x7711 /*ANA-7711C Network Accelerator Card (NAC) - Copper*/
#define DEVICE_ADAPTEC2_AIC_787X			0x7778 /*AIC-787x*/
#define DEVICE_ADAPTEC2_AIC_78102			0x7810 /*AIC-7810*/
#define DEVICE_ADAPTEC2_AIC_7815_RAID_MEMORY_CONTROLLER	0x7815 /*AIC-7815 RAID+Memory Controller IC*/
	#define SUBDEVICE_ADAPTEC2_ARO_1130U2_RAID_CONTROLLER	0x90047815 /*ARO-1130U2 RAID Controller*/
	#define SUBDEVICE_ADAPTEC2_AIC_7815_RAID_MEMORY_CONTROLLER	0x90047840 /*AIC-7815 RAID+Memory Controller IC*/
#define DEVICE_ADAPTEC2_AIC_78504			0x7850 /*AIC-7850*/
#define DEVICE_ADAPTEC2_AHA_2930			0x7855 /*AHA-2930*/
#define DEVICE_ADAPTEC2_AIC_78604			0x7860 /*AIC-7860*/
#define DEVICE_ADAPTEC2_AIC_7870			0x7870 /*AIC-7870*/
#define DEVICE_ADAPTEC2_AHA_2940			0x7871 /*AHA-2940*/
#define DEVICE_ADAPTEC2_AHA_3940			0x7872 /*AHA-3940*/
#define DEVICE_ADAPTEC2_AHA_3980			0x7873 /*AHA-3980*/
#define DEVICE_ADAPTEC2_AHA_2944			0x7874 /*AHA-2944*/
#define DEVICE_ADAPTEC2_AIC_7880P			0x7880 /*AIC-7880P*/
#define DEVICE_ADAPTEC2_AIC_7890			0x7890 /*AIC-7890*/
#define DEVICE_ADAPTEC2_AIC_789X			0x7891 /*AIC-789x*/
#define DEVICE_ADAPTEC2_AIC_789X2			0x7892 /*AIC-789x*/
#define DEVICE_ADAPTEC2_AIC_789X3			0x7893 /*AIC-789x*/
#define DEVICE_ADAPTEC2_AIC_789X4			0x7894 /*AIC-789x*/
#define DEVICE_ADAPTEC2_AHA_2940U_UW_AHA_39XX		0x7895 /*AHA-2940U/UW / AHA-39xx / AIC-7895*/
	#define SUBDEVICE_ADAPTEC2_AHA_2940U_2940UW_DUAL_AHA_394XAU_AUW_AUWD	0x90047890 /*AHA-2940U/2940UW Dual AHA-394xAU/AUW/AUWD AIC-7895B*/
	#define SUBDEVICE_ADAPTEC2_AHA_2940U_2940UW_DUAL	0x90047891 /*AHA-2940U/2940UW Dual*/
	#define SUBDEVICE_ADAPTEC2_AHA_3940AU_AUW_AUWD_UWD	0x90047892 /*AHA-3940AU/AUW/AUWD/UWD*/
	#define SUBDEVICE_ADAPTEC2_AHA_3944AUWD		0x90047894 /*AHA-3944AUWD*/
	#define SUBDEVICE_ADAPTEC2_AHA_2940U_2940UW_DUAL_AHA_394XAU_AUW_AUWD2	0x90047895 /*AHA-2940U/2940UW Dual AHA-394xAU/AUW/AUWD AIC-7895B*/
	#define SUBDEVICE_ADAPTEC2_AHA_2940U_2940UW_DUAL_AHA_394XAU_AUW_AUWD3	0x90047896 /*AHA-2940U/2940UW Dual AHA-394xAU/AUW/AUWD AIC-7895B*/
	#define SUBDEVICE_ADAPTEC2_AHA_2940U_2940UW_DUAL_AHA_394XAU_AUW_AUWD4	0x90047897 /*AHA-2940U/2940UW Dual AHA-394xAU/AUW/AUWD AIC-7895B*/
#define DEVICE_ADAPTEC2_AIC_789X5			0x7896 /*AIC-789x*/
#define DEVICE_ADAPTEC2_AIC_789X6			0x7897 /*AIC-789x*/
#define DEVICE_ADAPTEC2_AIC_7880U			0x8078 /*AIC-7880U*/
	#define SUBDEVICE_ADAPTEC2_AIC_7880P_ULTRA_ULTRA_WIDE	0x90047880 /*AIC-7880P Ultra/Ultra Wide SCSI Chipset*/
#define DEVICE_ADAPTEC2_AHA_2940U_UW_D_AIC_7881U	0x8178 /*AHA-2940U/UW/D / AIC-7881U*/
	#define SUBDEVICE_ADAPTEC2_AHA_2940UW_SCSI_HOST	0x90047881 /*AHA-2940UW SCSI Host Adapter*/
#define DEVICE_ADAPTEC2_AHA_3940U_UW_UWD_AIC_7882U	0x8278 /*AHA-3940U/UW/UWD / AIC-7882U*/
#define DEVICE_ADAPTEC2_AHA_3940U_UW_AIC_7883U		0x8378 /*AHA-3940U/UW / AIC-7883U*/
#define DEVICE_ADAPTEC2_AHA_2944UW_AIC_7884U		0x8478 /*AHA-2944UW / AIC-7884U*/
#define DEVICE_ADAPTEC2_AHA_3944U_UWD_AIC_7885		0x8578 /*AHA-3944U/UWD / AIC-7885*/
#define DEVICE_ADAPTEC2_AHA_4944UW_AIC_7886		0x8678 /*AHA-4944UW / AIC-7886*/
#define DEVICE_ADAPTEC2_AHA_2940UW_PRO			0x8778 /*AHA-2940UW Pro / AIC-788x*/
	#define SUBDEVICE_ADAPTEC2_2940UW_PRO_ULTRA_WIDE	0x90047887 /*2940UW Pro Ultra-Wide SCSI Controller*/
#define DEVICE_ADAPTEC2_AHA_2930UW_AIC_7888		0x8878 /*AHA-2930UW / AIC-7888*/
	#define SUBDEVICE_ADAPTEC2_AHA_2930UW_SCSI_CONTROLLER	0x90047888 /*AHA-2930UW SCSI Controller*/
#define DEVICE_ADAPTEC2_ABA_1030			0x8b78 /*ABA-1030*/
#define DEVICE_ADAPTEC2_AHA_4944W_UW			0xec78 /*AHA-4944W/UW*/
/* Vendor: 9005: Adaptec */
#define DEVICE_ADAPTEC3_AHA_2940U2_U2W			0x0010 /*AHA-2940U2/U2W*/
	#define SUBDEVICE_ADAPTEC3_AHA_2940U2_SCSI_CONTROLLER	0x90052180 /*AHA-2940U2 SCSI Controller*/
	#define SUBDEVICE_ADAPTEC3_AHA_2940U2B_SCSI_CONTROLLER	0x90058100 /*AHA-2940U2B SCSI Controller*/
	#define SUBDEVICE_ADAPTEC3_AHA_2940U2B_SCSI_CONTROLLER2	0x9005a100 /*AHA-2940U2B SCSI Controller*/
	#define SUBDEVICE_ADAPTEC3_AHA_2940U2W_SCSI_CONTROLLER	0x9005a180 /*AHA-2940U2W SCSI Controller*/
	#define SUBDEVICE_ADAPTEC3_AHA_2950U2B_SCSI_CONTROLLER	0x9005e100 /*AHA-2950U2B SCSI Controller*/
#define DEVICE_ADAPTEC3_AHA_2930U2			0x0011 /*AHA-2930U2*/
#define DEVICE_ADAPTEC3_78902				0x0013 /*78902*/
	#define SUBDEVICE_ADAPTEC3_AAA_131U2_ARRAY1000_1	0x90050003 /*AAA-131U2 Array1000 1 Channel RAID Controller*/
	#define SUBDEVICE_ADAPTEC3_AIC7890_ARO		0x9005000f /*AIC7890_ARO*/
#define DEVICE_ADAPTEC3_AHA_2940U2_U2W_7890_7891	0x001f /*AHA-2940U2/U2W / 7890/7891*/
	#define SUBDEVICE_ADAPTEC3_2940U2W_SCSI_CONTROLLER	0x9005000f /*2940U2W SCSI Controller*/
	#define SUBDEVICE_ADAPTEC3_2940U2W_SCSI_CONTROLLER2	0x9005a180 /*2940U2W SCSI Controller*/
#define DEVICE_ADAPTEC3_AIC_7890			0x0020 /*AIC-7890*/
#define DEVICE_ADAPTEC3_AIC_78902			0x002f /*AIC-7890*/
#define DEVICE_ADAPTEC3_AIC_78903			0x0030 /*AIC-7890*/
#define DEVICE_ADAPTEC3_AIC_78904			0x003f /*AIC-7890*/
#define DEVICE_ADAPTEC3_AHA_3940U2X_395U2X		0x0050 /*AHA-3940U2x/395U2x*/
	#define SUBDEVICE_ADAPTEC3_AHA_3950U2B		0x9005f500 /*AHA-3950U2B*/
	#define SUBDEVICE_ADAPTEC3_AHA_3950U2B2		0x9005ffff /*AHA-3950U2B*/
#define DEVICE_ADAPTEC3_AHA_3950U2D			0x0051 /*AHA-3950U2D*/
	#define SUBDEVICE_ADAPTEC3_AHA_3950U2D		0x9005b500 /*AHA-3950U2D*/
#define DEVICE_ADAPTEC3_AIC_7896_SCSI_CONTROLLER	0x0053 /*AIC-7896 SCSI Controller*/
	#define SUBDEVICE_ADAPTEC3_AIC_7896_SCSI_CONTROLLER	0x9005ffff /*AIC-7896 SCSI Controller mainboard implementation*/
#define DEVICE_ADAPTEC3_AIC_7896U2_7897U2		0x005f /*AIC-7896U2/7897U2*/
#define DEVICE_ADAPTEC3_AIC_7892A_U160_M		0x0080 /*AIC-7892A U160/m*/
	#define SUBDEVICE_ADAPTEC3_COMPAQ_64_BIT_66MHZ_WIDE	0x0e11e2a0 /*Compaq 64-Bit/66MHz Wide Ultra3 SCSI Adapter*/
	#define SUBDEVICE_ADAPTEC3_AHA_29160C		0x90056220 /*AHA-29160C*/
	#define SUBDEVICE_ADAPTEC3_29160N_ULTRA160_SCSI	0x900562a0 /*29160N Ultra160 SCSI Controller*/
	#define SUBDEVICE_ADAPTEC3_29160LP_LOW_PROFILE	0x9005e220 /*29160LP Low Profile Ultra160 SCSI Controller*/
	#define SUBDEVICE_ADAPTEC3_29160_ULTRA160_SCSI	0x9005e2a0 /*29160 Ultra160 SCSI Controller*/
#define DEVICE_ADAPTEC3_AIC_7892B_U160_M		0x0081 /*AIC-7892B U160/m*/
	#define SUBDEVICE_ADAPTEC3_19160_ULTRA160_SCSI	0x900562a1 /*19160 Ultra160 SCSI Controller*/
#define DEVICE_ADAPTEC3_AIC_7892D_U160_M		0x0083 /*AIC-7892D U160/m*/
#define DEVICE_ADAPTEC3_AIC_7892P_U160_M		0x008f /*AIC-7892P U160/m*/
	#define SUBDEVICE_ADAPTEC3_MAGNIA_Z310		0x11790001 /*Magnia Z310*/
	#define SUBDEVICE_ADAPTEC3_ONBOARD_SCSI_HOST	0x15d99005 /*Onboard SCSI Host Adapter*/
#define DEVICE_ADAPTEC3_AHA_3960D_AIC_7899A		0x00c0 /*AHA-3960D / AIC-7899A U160/m*/
	#define SUBDEVICE_ADAPTEC3_COMPAQ_64_BIT_66MHZ_DUAL	0x0e11f620 /*Compaq 64-Bit/66MHz Dual Channel Wide Ultra3 SCSI Adapter*/
	#define SUBDEVICE_ADAPTEC3_AHA_3960D_U160_M	0x9005f620 /*AHA-3960D U160/m*/
#define DEVICE_ADAPTEC3_AIC_7899B_U160_M		0x00c1 /*AIC-7899B U160/m*/
#define DEVICE_ADAPTEC3_AIC_7899D_U160_M		0x00c3 /*AIC-7899D U160/m*/
#define DEVICE_ADAPTEC3_RAID_SUBSYSTEM_HBA		0x00c5 /*RAID subsystem HBA*/
	#define SUBDEVICE_ADAPTEC3_POWEREDGE_2400_2500_2550_4400	0x102800c5 /*PowerEdge 2400,2500,2550,4400*/
#define DEVICE_ADAPTEC3_AIC_7899P_U160_M		0x00cf /*AIC-7899P U160/m*/
	#define SUBDEVICE_ADAPTEC3_POWEREDGE_1400	0x102800ce /*PowerEdge 1400*/
	#define SUBDEVICE_ADAPTEC3_POWEREDGE_2550	0x102800d1 /*PowerEdge 2550*/
	#define SUBDEVICE_ADAPTEC3_POWEREDGE_2500	0x102800d9 /*PowerEdge 2500*/
	#define SUBDEVICE_ADAPTEC3_THUNDER_K7_S2462	0x10f12462 /*Thunder K7 S2462*/
	#define SUBDEVICE_ADAPTEC3_ONBOARD_SCSI_HOST2	0x15d99005 /*Onboard SCSI Host Adapter*/
	#define SUBDEVICE_ADAPTEC3_SDS2_MAINBOARD	0x80863411 /*SDS2 Mainboard*/
#define DEVICE_ADAPTEC3_SERVERAID_CONTROLLER		0x0250 /*ServeRAID Controller*/
	#define SUBDEVICE_ADAPTEC3_SERVERAID_XX		0x10140279 /*ServeRAID-xx*/
	#define SUBDEVICE_ADAPTEC3_SERVERAID_XX2	0x1014028c /*ServeRAID-xx*/
#define DEVICE_ADAPTEC3_SERVERAID_6M			0x0279 /*ServeRAID 6M*/
#define DEVICE_ADAPTEC3_AAC_RAID			0x0283 /*AAC-RAID*/
	#define SUBDEVICE_ADAPTEC3_CATAPULT		0x90050283 /*Catapult*/
#define DEVICE_ADAPTEC3_AAC_RAID2			0x0284 /*AAC-RAID*/
	#define SUBDEVICE_ADAPTEC3_TOMCAT		0x90050284 /*Tomcat*/
#define DEVICE_ADAPTEC3_AAC_RAID3			0x0285 /*AAC-RAID*/
	#define SUBDEVICE_ADAPTEC3_SATA_6CH_BEARCAT	0x0e110295 /*SATA 6Ch (Bearcat)*/
	#define SUBDEVICE_ADAPTEC3_POWEREDGE_EXPANDABLE_RAID	0x10280287 /*PowerEdge Expandable RAID Controller 320/DC*/
	#define SUBDEVICE_ADAPTEC3_CERC_SATA_RAID	0x10280291 /*CERC SATA RAID 2 PCI SATA 6ch (DellCorsair)*/
	#define SUBDEVICE_ADAPTEC3_LEGEND_S220_LEGEND	0x17aa0286 /*Legend S220 (Legend Crusader)*/
	#define SUBDEVICE_ADAPTEC3_LEGEND_S230_LEGEND	0x17aa0287 /*Legend S230 (Legend Vulcan)*/
	#define SUBDEVICE_ADAPTEC3_2200S_VULCAN		0x90050285 /*2200S (Vulcan)*/
	#define SUBDEVICE_ADAPTEC3_2120S_CRUSADER	0x90050286 /*2120S (Crusader)*/
	#define SUBDEVICE_ADAPTEC3_2200S_VULCAN_2M	0x90050287 /*2200S (Vulcan-2m)*/
	#define SUBDEVICE_ADAPTEC3_3230S_HARRIER	0x90050288 /*3230S (Harrier)*/
	#define SUBDEVICE_ADAPTEC3_3240S_TORNADO	0x90050289 /*3240S (Tornado)*/
	#define SUBDEVICE_ADAPTEC3_ASR_2020S_PCI_X_ZCR	0x9005028a /*ASR-2020S PCI-X ZCR (Skyhawk)*/
	#define SUBDEVICE_ADAPTEC3_ASR_2020S_SO_DIMM_PCI_X	0x9005028b /*ASR-2020S SO-DIMM PCI-X ZCR (Terminator)*/
	#define SUBDEVICE_ADAPTEC3_AAR_2410SA_PCI_SATA	0x90050290 /*AAR-2410SA PCI SATA 4ch (Jaguar II)*/
	#define SUBDEVICE_ADAPTEC3_AAR_2810SA_PCI_SATA	0x90050292 /*AAR-2810SA PCI SATA 8ch (Corsair-8)*/
	#define SUBDEVICE_ADAPTEC3_AAR_21610SA_PCI_SATA	0x90050293 /*AAR-21610SA PCI SATA 16ch (Corsair-16)*/
	#define SUBDEVICE_ADAPTEC3_ESD_SO_DIMM_PCI_X	0x90050294 /*ESD SO-DIMM PCI-X SATA ZCR (Prowler)*/
#define DEVICE_ADAPTEC3_AAC_RAID_ROCKET			0x0286 /*AAC-RAID (Rocket)*/
	#define SUBDEVICE_ADAPTEC3_ASR_2230S_ASR_2230SLP	0x9005028c /*ASR-2230S + ASR-2230SLP PCI-X (Lancer)*/
#define DEVICE_ADAPTEC3_ASC_29320A_U320			0x8000 /*ASC-29320A U320*/
#define DEVICE_ADAPTEC3_AIC_7901_U320			0x800f /*AIC-7901 U320*/
#define DEVICE_ADAPTEC3_ASC_39320_U320			0x8010 /*ASC-39320 U320*/
#define DEVICE_ADAPTEC3_ASC_32320D_U320			0x8011 /*ASC-32320D U320*/
	#define SUBDEVICE_ADAPTEC3_ASC_39320D_U320	0x0e1100ac /*ASC-39320D U320*/
	#define SUBDEVICE_ADAPTEC3_ASC_39320D_U3202	0x90050041 /*ASC-39320D U320*/
#define DEVICE_ADAPTEC3_ASC_29320_U320			0x8012 /*ASC-29320 U320*/
#define DEVICE_ADAPTEC3_ASC_29320B_U320			0x8013 /*ASC-29320B U320*/
#define DEVICE_ADAPTEC3_ASC_29320LP_U320		0x8014 /*ASC-29320LP U320*/
#define DEVICE_ADAPTEC3_ASC_39320B_U320			0x8015 /*ASC-39320B U320*/
#define DEVICE_ADAPTEC3_ASC_39320A_U320			0x8016 /*ASC-39320A U320*/
#define DEVICE_ADAPTEC3_ASC_29320ALP_U320		0x8017 /*ASC-29320ALP U320*/
#define DEVICE_ADAPTEC3_ASC_39320D_U320			0x801c /*ASC-39320D U320*/
#define DEVICE_ADAPTEC3_AIC_7902B_U320			0x801d /*AIC-7902B U320*/
#define DEVICE_ADAPTEC3_AIC_7901A_U320			0x801e /*AIC-7901A U320*/
#define DEVICE_ADAPTEC3_AIC_7902_U320			0x801f /*AIC-7902 U320*/
#define DEVICE_ADAPTEC3_ASC_29320A_U320_W_HOSTRAID	0x8080 /*ASC-29320A U320 w/HostRAID*/
#define DEVICE_ADAPTEC3_AIC_7901_U320_W_HOSTRAID	0x808f /*AIC-7901 U320 w/HostRAID*/
#define DEVICE_ADAPTEC3_ASC_39320_U320_W_HOSTRAID	0x8090 /*ASC-39320 U320 w/HostRAID*/
#define DEVICE_ADAPTEC3_ASC_39320D_U320_W_HOSTRAID	0x8091 /*ASC-39320D U320 w/HostRAID*/
#define DEVICE_ADAPTEC3_ASC_29320_U320_W_HOSTRAID	0x8092 /*ASC-29320 U320 w/HostRAID*/
#define DEVICE_ADAPTEC3_ASC_29320B_U320_W_HOSTRAID	0x8093 /*ASC-29320B U320 w/HostRAID*/
#define DEVICE_ADAPTEC3_ASC_29320LP_U320_W_HOSTRAID	0x8094 /*ASC-29320LP U320 w/HostRAID*/
#define DEVICE_ADAPTEC3_ASC_39320_B_U320_W_HOSTRAID	0x8095 /*ASC-39320(B) U320 w/HostRAID*/
#define DEVICE_ADAPTEC3_ASC_39320A_U320_W_HOSTRAID	0x8096 /*ASC-39320A U320 w/HostRAID*/
#define DEVICE_ADAPTEC3_ASC_29320ALP_U320_W_HOSTRAID	0x8097 /*ASC-29320ALP U320 w/HostRAID*/
#define DEVICE_ADAPTEC3_ASC_39320D_B_U320_W_HOSTRAID	0x809c /*ASC-39320D(B) U320 w/HostRAID*/
#define DEVICE_ADAPTEC3_AIC_7902_B_U320_W_HOSTRAID	0x809d /*AIC-7902(B) U320 w/HostRAID*/
#define DEVICE_ADAPTEC3_AIC_7901A_U320_W_HOSTRAID	0x809e /*AIC-7901A U320 w/HostRAID*/
#define DEVICE_ADAPTEC3_AIC_7902_U320_W_HOSTRAID	0x809f /*AIC-7902 U320 w/HostRAID*/
/* Vendor: 907f: Atronics */
#define DEVICE_ATRONICS_IDE_2015PL			0x2015 /*IDE-2015PL*/
/* Vendor: 919a: Gigapixel Corp */
/* Vendor: 9412: Holtek */
#define DEVICE_HOLTEK2_6565				0x6565 /*6565*/
/* Vendor: 9699: Omni Media Technology Inc */
#define DEVICE_OMNI_6565					0x6565 /*6565*/
/* Vendor: 9710: NetMos Technology */
#define DEVICE_NETMOS_USB_IRDA_PORT			0x7780 /*USB IRDA-port*/
#define DEVICE_NETMOS_PCI_9815_MULTI_I_O		0x9815 /*PCI 9815 Multi-I/O Controller*/
	#define SUBDEVICE_NETMOS_2P0S_2_PORT		0x10000020 /*2P0S (2 port parallel adaptor)*/
#define DEVICE_NETMOS_PCI_9835_MULTI_I_O		0x9835 /*PCI 9835 Multi-I/O Controller*/
	#define SUBDEVICE_NETMOS_2S_16C550_UART		0x10000002 /*2S (16C550 UART)*/
	#define SUBDEVICE_NETMOS_1P2S			0x10000012 /*1P2S*/
#define DEVICE_NETMOS_PCI_9845_MULTI_I_O		0x9845 /*PCI 9845 Multi-I/O Controller*/
	#define SUBDEVICE_NETMOS_0P4S_4_PORT		0x10000004 /*0P4S (4 port 16550A serial card)*/
	#define SUBDEVICE_NETMOS_0P6S_6_PORT		0x10000006 /*0P6S (6 port 16550a serial card)*/
#define DEVICE_NETMOS_PCI_9855_MULTI_I_O		0x9855 /*PCI 9855 Multi-I/O Controller*/
	#define SUBDEVICE_NETMOS_1P4S			0x10000014 /*1P4S*/
/* Vendor: 9902: Stargen Inc. */
#define DEVICE_STARGEN_SG2010_PCI_OVER			0x0001 /*SG2010 PCI over Starfabric Bridge*/
#define DEVICE_STARGEN_SG2010_PCI_TO			0x0002 /*SG2010 PCI to Starfabric Gateway*/
#define DEVICE_STARGEN_SG1010_STARFABRIC_SWITCH		0x0003 /*SG1010 Starfabric Switch and PCI Bridge*/
/* Vendor: a0a0: AOPEN Inc. */
/* Vendor: a0f1: UNISYS Corporation */
/* Vendor: a200: NEC Corporation */
/* Vendor: a259: Hewlett Packard */
/* Vendor: a25b: Hewlett Packard GmbH PL24-MKT */
/* Vendor: a304: Sony */
/* Vendor: a727: 3Com Corporation */
#define DEVICE_3COM3_3CRPAG175_WIRELESS_PC		0x0013 /*3CRPAG175 Wireless PC Card*/
/* Vendor: aa42: Scitex Digital Video */
/* Vendor: ac1e: Digital Receiver Technology Inc */
/* Vendor: ac3d: Actuality Systems */
/* Vendor: aecb: Adrienne Electronics Corporation */
/* Vendor: b1b3: Shiva Europe Limited */
/* Vendor: bd11: Pinnacle Systems, Inc. (Wrong ID) */
/* Vendor: c001: TSI Telsys */
/* Vendor: c0a9: Micron/Crucial Technology */
/* Vendor: c0de: Motorola */
/* Vendor: c0fe: Motion Engineering, Inc. */
/* Vendor: ca50: Varian Australia Pty Ltd */
/* Vendor: cafe: Chrysalis-ITS */
/* Vendor: cccc: Catapult Communications */
/* Vendor: cddd: Tyzx, Inc. */
#define DEVICE_TYZX_DEEPSEA_1_HIGH			0x0101 /*DeepSea 1 High Speed Stereo Vision Frame Grabber*/
#define DEVICE_TYZX_DEEPSEA_2_HIGH			0x0200 /*DeepSea 2 High Speed Stereo Vision Frame Grabber*/
/* Vendor: d4d4: Dy4 Systems Inc */
#define DEVICE_DY4_PCI_MEZZANINE_CARD			0x0601 /*PCI Mezzanine Card*/
/* Vendor: d531: I+ME ACTIA GmbH */
/* Vendor: d84d: Exsys */
/* Vendor: dead: Indigita Corporation */
/* Vendor: e000: Winbond */
#define DEVICE_WINBOND3_W89C940				0xe000 /*W89C940*/
/* Vendor: e159: Tiger Jet Network Inc. */
#define DEVICE_TIGER2_TIGER3XX_MODEM_ISDN_INTERFACE	0x0001 /*Tiger3XX Modem/ISDN interface*/
	#define SUBDEVICE_TIGER2_128K_ISDN_S_T_ADAPTER	0x00590001 /*128k ISDN-S/T Adapter*/
	#define SUBDEVICE_TIGER2_128K_ISDN_U_ADAPTER	0x00590003 /*128k ISDN-U Adapter*/
#define DEVICE_TIGER2_TIGER100APC_ISDN_CHIPSET		0x0002 /*Tiger100APC ISDN chipset*/
/* Vendor: e4bf: EKF Elektronik GmbH */
/* Vendor: e55e: Essence Technology, Inc. */
/* Vendor: ea01: Eagle Technology */
/* Vendor: ea60: RME */
#define DEVICE_RME_DIGI32				0x9896 /*Digi32*/
#define DEVICE_RME_DIGI32_PRO				0x9897 /*Digi32 Pro*/
#define DEVICE_RME_DIGI32_8				0x9898 /*Digi32/8*/
/* Vendor: eabb: Aashima Technology B.V. */
/* Vendor: eace: Endace Measurement Systems, Ltd */
#define DEVICE_ENDACE_DAG_3_10_OC_3_OC_12		0x3100 /*DAG 3.10 OC-3/OC-12*/
#define DEVICE_ENDACE_DAG_3_2X_OC_3_OC_12		0x3200 /*DAG 3.2x OC-3/OC-12*/
#define DEVICE_ENDACE_DAG_3_2E_FAST			0x320e /*DAG 3.2E Fast Ethernet*/
#define DEVICE_ENDACE_DAG_3_4E_FAST			0x340e /*DAG 3.4E Fast Ethernet*/
#define DEVICE_ENDACE_DAG_3_41E_FAST			0x341e /*DAG 3.41E Fast Ethernet*/
#define DEVICE_ENDACE_DAG_3_5_OC_3_OC_12		0x3500 /*DAG 3.5 OC-3/OC-12*/
#define DEVICE_ENDACE_DAG_3_5ECM_FAST			0x351c /*DAG 3.5ECM Fast Ethernet*/
#define DEVICE_ENDACE_DAG_4_10_OC_48			0x4100 /*DAG 4.10 OC-48*/
#define DEVICE_ENDACE_DAG_4_11_OC_48			0x4110 /*DAG 4.11 OC-48*/
#define DEVICE_ENDACE_DAG_4_2_OC_48			0x4220 /*DAG 4.2 OC-48*/
#define DEVICE_ENDACE_DAG_4_2E_DUAL			0x422e /*DAG 4.2E Dual Gigabit Ethernet*/
/* Vendor: ec80: Belkin Corporation */
#define DEVICE_BELKIN2_F5D6000				0xec00 /*F5D6000*/
/* Vendor: ecc0: Echo Digital Audio Corporation */
#define DEVICE_ECHO_GINA24_301				0x0050 /*Gina24_301*/
#define DEVICE_ECHO_GINA24_361				0x0051 /*Gina24_361*/
#define DEVICE_ECHO_LAYLA24				0x0060 /*Layla24*/
#define DEVICE_ECHO_MONA_301_80				0x0070 /*Mona_301_80*/
#define DEVICE_ECHO_MONA_301_66				0x0071 /*Mona_301_66*/
#define DEVICE_ECHO_MONA_361				0x0072 /*Mona_361*/
#define DEVICE_ECHO_MIA					0x0080 /*Mia*/
/* Vendor: edd8: ARK Logic Inc */
#define DEVICE_ARK2_1000PV_STINGRAY			0xa091 /*1000PV [Stingray]*/
#define DEVICE_ARK2_2000PV_STINGRAY			0xa099 /*2000PV [Stingray]*/
#define DEVICE_ARK2_2000MT				0xa0a1 /*2000MT*/
#define DEVICE_ARK2_2000MI				0xa0a9 /*2000MI*/
/* Vendor: f1d0: AJA Video */
#define DEVICE_AJA_KONA_SD_SMPTE			0xcafe /*KONA SD SMPTE 259M I/O*/
#define DEVICE_AJA_KONA_SD_SMPTE2			0xefac /*KONA SD SMPTE 259M I/O*/
#define DEVICE_AJA_KONA_HD_SMPTE			0xfacd /*KONA HD SMPTE 292M I/O*/
/* Vendor: fa57: Interagon AS */
#define DEVICE_INTERAGON_PMC_PATTERN_MATCHING		0x0001 /*PMC [Pattern Matching Chip]*/
/* Vendor: febd: Ultraview Corp. */
/* Vendor: feda: Broadcom Inc (nee Epigram) */
#define DEVICE_BROADCOM3_BCM4210_ILINE10_HOMEPNA	0xa0fa /*BCM4210 iLine10 HomePNA 2.0*/
#define DEVICE_BROADCOM3_BCM4230_ILINE10_HOMEPNA	0xa10e /*BCM4230 iLine10 HomePNA 2.0*/
/* Vendor: fede: Fedetec Inc. */
#define DEVICE_FEDETEC_TABIC_PCI_V3			0x0003 /*TABIC PCI v3*/
/* Vendor: fffe: VMWare Inc */
#define DEVICE_VMWARE2_VIRTUAL_SVGA_4_0			0x0405 /*Virtual SVGA 4.0*/
#define DEVICE_VMWARE2_VIRTUAL_SVGA			0x0710 /*Virtual SVGA*/
/* Vendor: ffff: Illegal Vendor ID */

#endif/*PCI_IDS_INCLUDED*/
